/*************************************************************************************************

    Copyright (C) 2009-2019    Jean de Montigny (jean.demontigny@hotmail.fr)
                               Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __BDM_DOMAIN_COMMUNICATION_H__
#define __BDM_DOMAIN_COMMUNICATION_H__
//=================================================================================================
inline
bool isInList(const std::vector<libMesh::Point>& list,
              const libMesh::Point& coord)
{
  for (libMesh::Point origin : list) {
    if (origin == coord)
      return 1;
  }
  return 0;
}
inline
bool isInList(const std::vector<std::array<double, 3>>& list,
              const std::array<double, 3>& coord)
{
  for (std::array<double, 3> origin : list) {
    if (origin == coord)
      return 1;
  }
  return 0;
}
//-------------------------------------------------------------------------------------------------
inline
std::vector<std::array<double, 3>> getRveOriginList(std::vector<BDM_Domain>& rve_list)
{
  std::vector<std::array<double, 3>> list;
  for (BDM_Domain this_rve : rve_list) {
    list.push_back(this_rve.origin());
  }
  return list;
}
//-------------------------------------------------------------------------------------------------
inline
bool isInsidefeb3boundaries(const std::array<double, 3>& coord,
                            const double feb3_min_bound = -1000.0,
                            const double feb3_max_bound = +1000.0)
{
  return ( coord[0] >= feb3_min_bound && coord[0] <= feb3_max_bound &&
           coord[1] >= feb3_min_bound && coord[1] <= feb3_max_bound &&
           coord[2] >= feb3_min_bound && coord[2] <= feb3_max_bound );
}
//-------------------------------------------------------------------------------------------------
// get cell coord corresponding to its position to invide neighbour cube
inline
std::array<double, 3> getInvadingCoord(const std::array<double, 3>& cell_coord,
                                       double min_coord, double max_coord)
{
  std::array<double, 3> invading_coord = {0.0, 0.0, 0.0};

  const double margin = (max_coord - min_coord) * 0.01;

  for (unsigned int i=0; i<3; i++) {
    if (cell_coord[i] <= min_coord)
      invading_coord[i] = max_coord - margin;
    else if (cell_coord[i] >= max_coord)
      invading_coord[i] = min_coord + margin;
    else
      invading_coord[i] = cell_coord[i];
  }
  return invading_coord;
}
//-------------------------------------------------------------------------------------------------
// get receiving cube coordinates corresponding to one single migrating cell
inline
std::array<double, 3> getInvadedCubeCoord(const std::array<double, 3>& our_origin,
                                          const std::array<double, 3>& cell_coord,
                                          double min_coord, double max_coord,
                                          double bdm_cube_size)
{
  std::array<double, 3> cube_coord = {0.0, 0.0, 0.0};

  for (unsigned int i=0; i<3; i++) {
    if (cell_coord[i] <= min_coord)
      cube_coord[i] = our_origin[i] - bdm_cube_size;
    else if (cell_coord[i] >= max_coord)
      cube_coord[i] = our_origin[i] + bdm_cube_size;
    else
      cube_coord[i] = our_origin[i];
  }
  return cube_coord;
}
//-------------------------------------------------------------------------------------------------
inline
BDM_Domain* getRveToCreateCellTo(const std::array<double, 3>& cell_escape_position,
                                 std::vector<BDM_Domain>& rve_list,
                                 const libMesh::Parameters& params,
                                 const std::array<double, 3>& rve_position)
{
  const double minCoord = params.get<double>("BioDynaMo/bounding_box/floor");
  const double maxCoord = params.get<double>("BioDynaMo/bounding_box/ceil");
  const double bdm_cube_size = params.get<double>("BioDynaMo/bounding_box/size");

  const std::array<double, 3>& invaded_cube_origin =
    getInvadedCubeCoord(rve_position, cell_escape_position, minCoord, maxCoord, bdm_cube_size);

  for (unsigned int r=0; r<rve_list.size(); r++) {
    if (invaded_cube_origin == rve_list[r].origin())
      return &rve_list[r];
  }
  return 0;
}
//-------------------------------------------------------------------------------------------------
inline
std::vector<std::array<double, 3>> getCoordCubesToCreate(
                                    const std::vector<std::array<double, 3>>& rve_origin_list,
                                    const BDM_Domain& our_rve,
                                    const libMesh::Parameters& params)
{
  const double minCoord = params.get<double>("BioDynaMo/bounding_box/floor");
  const double maxCoord = params.get<double>("BioDynaMo/bounding_box/ceil");
  const double bdm_cube_size = params.get<double>("BioDynaMo/bounding_box/size");

  const std::array<double, 3>& our_origin = our_rve.origin();

  std::vector<std::array<double, 3>> receiving_cube_coord;

  for (MigratingCell cell : our_rve.escape_cells_xyz) {
    std::array<double, 3> cell_coord_escape = cell.position;

    const std::array<double, 3>& invaded_cube =
      getInvadedCubeCoord(our_origin, cell_coord_escape, minCoord, maxCoord, bdm_cube_size);

    if ( isInsidefeb3boundaries(invaded_cube)         &&
       ! isInList(receiving_cube_coord, invaded_cube) &&
       ! isInList(rve_origin_list, invaded_cube)      ) {
      receiving_cube_coord.push_back(invaded_cube);
    }
  }
  return receiving_cube_coord;
}
//=================================================================================================
#endif // __BDM_DOMAIN_COMMUNICATION_H__
//=================================================================================================
