/*************************************************************************************************

    Copyright (C) 2009-2019    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __BDM_INTERFACE_H__
#define __BDM_INTERFACE_H__
//=================================================================================================
#include "feb3/utils.h"
#include "biodynamo.h"
#include "libmesh/parameters.h"
//=================================================================================================
// forward class declaration
class MigratingCell;
//-------------------------------------------------------------------------------------------------
// compact type definitions
typedef std::array<double, 3>  RealVector3d;
//=================================================================================================
class InterfaceData {
public:
    InterfaceData () {}
    InterfaceData (unsigned int id, const RealVector3d& xyz, unsigned int npt, unsigned int nch, unsigned int nme) {
        this->init(id, xyz, npt, nch, nme);
    }
    ~InterfaceData () {}
    //
    inline
    void init (unsigned int id, const RealVector3d& xyz, unsigned int npt, unsigned int nch, unsigned int nme) {
        this->_id = id;
        this->_origin = xyz;
        this->_n_phenotypes = npt;
        this->_n_biochem = nch;
        this->_n_biomech = nme;
        this->cells_mass.assign(this->_n_phenotypes, 0.0);
        this->cells_population.assign(this->_n_phenotypes, 0);
        this->cells_population_increment.assign(this->_n_phenotypes, 0);
        this->oxyg      = 0.0;
        this->oxyg_grad = {0.0, 0.0, 0.0};
        this->ecm      = 0.0;
        this->ecm_grad = {0.0, 0.0, 0.0};
        this->cells_volume_fraction = 0.0;
        this->biochem.assign(this->_n_biochem, 0.0);
        this->biomech.assign(this->_n_biomech, 0.0);
        this->biochem_grad.assign(this->_n_biochem, {0.0, 0.0, 0.0});
        this->biomech_grad.assign(this->_n_biomech, {0.0, 0.0, 0.0});
        this->interstitial_fluid_velocity = {0.0, 0.0, 0.0};
        this->escape_cells_per_face.assign(this->_n_phenotypes, {0, 0, 0, 0, 0, 0});
        this->income_cells_per_face.assign(this->_n_phenotypes, {0, 0, 0, 0, 0, 0});
        this->cell_dynamics.assign(this->_n_phenotypes, {0.0, 0.0, 0.0});
        this->cells_xyz.clear();
    }
    inline
    unsigned int id () const { return this->_id; }
    inline
    const RealVector3d& origin () const { return this->_origin; }
    inline
    unsigned int n_phenotypes () const { return this->_n_phenotypes; }
    inline
    unsigned int n_biochem () const { return this->_n_biochem; }
    inline
    unsigned int n_biomech () const { return this->_n_biomech; }
public:
    std::vector<Real> cells_mass;
    std::vector<unsigned int> cells_population;
    std::vector<int> cells_population_increment;
    Real         oxyg,      ecm;
    RealVector3d oxyg_grad, ecm_grad;
    std::vector<Real>         biochem,      biomech;
    std::vector<RealVector3d> biochem_grad, biomech_grad;
    Real cells_volume_fraction;
    RealVector3d interstitial_fluid_velocity;
    std::vector<std::array<int, 6>> escape_cells_per_face,
                                    income_cells_per_face;
    std::vector<MigratingCell> escape_cells_xyz,
                               income_cells_xyz;
    std::vector<RealVector3d> cell_dynamics;
    std::vector<RealVector3d> cells_xyz;
private:
    unsigned int _id, // simulation index
                 _n_phenotypes, // number of phenotypes (cell types)
                 _n_biochem, _n_biomech; // number of biochemical & biomechanical variables
    RealVector3d _origin; // coordinates of the origin (point)
};
//-------------------------------------------------------------------------------------------------
class BDM_Domain : public InterfaceData {
public:
    BDM_Domain () : _is_init(false) {}
    BDM_Domain (unsigned int id, const RealVector3d& xyz, unsigned int npt, unsigned int nch, unsigned int nme) {
        BDM_Domain::init(id, xyz, npt, nch, nme);
    }
    ~BDM_Domain () {}
    //
    inline
    void init (unsigned int id, const RealVector3d& xyz, unsigned int npt, unsigned int nch, unsigned int nme) {
        this->_is_init = true;
        InterfaceData::init(id, xyz, npt, nch, nme);
    }
    inline
    bool is_init () const { return this->_is_init; }
private:
    bool _is_init;
public:
    bdm::Simulation<>* sim;
};
//=================================================================================================
#include "feb3/bdm_migrating_cell.h"
#include "feb3/bdm_domain_communication.h"
//=================================================================================================
#endif // __BDM_INTERFACE_H__
//=================================================================================================
