/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __RVE_H__
#define __RVE_H__
//=================================================================================================
#include "libmesh/mesh.h"
#include "libmesh/elem.h"
#include "libmesh/point.h"
#include "feb3/utils.h"
//=================================================================================================
class RVE : public Mesh {
// member functions:
public:
    explicit RVE (const Parallel::Communicator& comm_in) : Mesh(comm_in, SP1D) {}
    ~RVE () {}
    //
    void generate (const unsigned int nfib, const Real cmin =0., const Real cmax =1.);
    void write (const std::string& fname ="rve.msh") const;
    void reinit ();
    //
    // public member functions to read-only stuff
    bool node_is_boundary (const dof_id_type& id, int& f) const;
    inline
    std::map<dof_id_type, int>::const_iterator bnode_map_begin () const { return this->_bnode.begin(); }
    inline
    std::map<dof_id_type, int>::const_iterator bnode_map_end   () const { return this->_bnode.end(); }
    inline
    std::map<dof_id_type, std::set<dof_id_type>>::const_iterator bnode_neighbor_begin () const { return this->_bnode_neighbor.begin(); }
    inline
    std::map<dof_id_type, std::set<dof_id_type>>::const_iterator bnode_neighbor_end   () const { return this->_bnode_neighbor.end(); }
    inline
    std::map<dof_id_type, std::set<dof_id_type>>::const_iterator bnode_neighbor_begin (int f) const { return this->_bnode_neighbor_face[f].begin(); }
    inline
    std::map<dof_id_type, std::set<dof_id_type>>::const_iterator bnode_neighbor_end   (int f) const { return this->_bnode_neighbor_face[f].end(); }
    // geometric traits of the representative volume element (RVE)
    inline
    Real length () const { return fabs(max_coord()-min_coord()); }
    inline
    Real volume () const { return _volume; }
    inline
    Real surface () const { return _surface; }
    inline
    Real side_surface () const { return _side_surface; }
    inline
    Real fibres_length () const { return _length; }
    // vertex point(s) of the RVE cuboid
    inline
    const Point& vertex (int i) const { return _vertex.at(i); }
    inline
    const std::vector<Point>& vertices () const { return _vertex; }
    inline
    std::vector<Point> face_vertices (int f) const {
        std::vector<Point> v(4);
        const std::vector<int> n = vertices_of_HEX_face(f);
        v[0] = _vertex[n[0]]; v[1] = _vertex[n[1]]; v[2] = _vertex[n[2]]; v[3] = _vertex[n[3]];
        return v;
    }
    // minimum and maximum (local) coordinate of the RVE cuboid
    inline
    const Real min_coord () const { return _vertex[0](2); }
    inline
    const Real max_coord () const { return _vertex[6](2); }
    // geometric centre of the RVE, and outward unit normal vectors for each facet
    inline
    Point centre () const { return 0.5*(max_coord()+min_coord())*Point(1,1,1); }
    inline
    const Point& normal (int f) const { return _normal[f]; }
    // Gauss quadrature points needed for the surface integrals calculations
    inline
    const std::vector<Point>& quadrature (int f) const { return _quad_pnt[f]; }
    inline
    Real jacobian () const { return _jac; }
private:
    // private member functions to set-up stuff
    std::string _gen_rve (const unsigned int nfib, const Real cmin, const Real cmax);
    void _setup_stuff ();
// member objects:
private:
    // RVE volume, external surface area, side/face surface area, total fibre length
    Real _volume, _surface, _side_surface, _length;
    // list of cubic RVE - 8 in total by definition
    std::vector<Point> _vertex;
    // list of outward unit normal vectors per face - 6 in total by definition
    Point _normal[6];
    // quadrature (rule) points per face - 6 in total by definition
    std::vector<Point> _quad_pnt[6];
    // jacobian for quadrature (rule)
    Real _jac;
    // map of boundary nodes: node ID / face ID
    std::map<dof_id_type, int> _bnode;
    // map of boundary nodes: node ID / adjacent node ID
    std::map<dof_id_type, std::set<dof_id_type>> _bnode_neighbor;
    // map of boundary nodes: node ID / adjacent node ID [face ID] - 6 in total by definition
    std::map<dof_id_type, std::set<dof_id_type>> _bnode_neighbor_face[6];
};
//-------------------------------------------------------------------------------------------------
// compact type definition
typedef RVE   MeshRVE;
//=================================================================================================
#endif // __RVE_H__
//=================================================================================================
