/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __TUMOUR_ANGIOGENESIS_CONSTITUTIVE_H__
#define __TUMOUR_ANGIOGENESIS_CONSTITUTIVE_H__
//=================================================================================================
struct TumourAngiogenesisConstitutive {
    Real (*blood_vessel_hydraulic_conductivity) ( const VascularInfo* vi,
                                                  const MaterialInfo* mi__vnr,
                                                  const MaterialInfo* mi__dmn,
                                                  FieldInfo* fld );
    Real (*lymph_vessel_hydraulic_conductivity) ( const VascularInfo* vi,
                                                  const MaterialInfo* mi__vnr,
                                                  const MaterialInfo* mi__dmn,
                                                  FieldInfo* fld );
    Real (*blood_vessel_hydraulic_permeability) ( const VascularInfo* vi,
                                                  const MaterialInfo* mi__vnr,
                                                  const MaterialInfo* mi__dmn,
                                                  FieldInfo* fld );
    Real (*lymph_vessel_hydraulic_permeability) ( const VascularInfo* vi,
                                                  const MaterialInfo* mi__vnr,
                                                  const MaterialInfo* mi__dmn,
                                                  FieldInfo* fld );
    Real (*vascular_hydraulic_permeability_2_drug) ( const VascularInfo* vi,
                                                     const MaterialInfo* mi__vnr,
                                                     const MaterialInfo* mi__dmn,
                                                     FieldInfo* fld,
                                                     const unsigned int drug_id );
    Real (*vascular_reflection_coefficient_2_drug) ( const VascularInfo* vi,
                                                     const MaterialInfo* mi__vnr,
                                                     const MaterialInfo* mi__dmn,
                                                     FieldInfo* fld,
                                                     const unsigned int drug_id );
    Real (*osmotic_pressure_difference) ( const VascularInfo* vi,
                                          const MaterialInfo* mi__vnr,
                                          const MaterialInfo* mi__dmn,
                                          FieldInfo* fld );
    Real (*blood_vessel_stroma_tissue_ratio) ( const MaterialInfo* mi__vnr,
                                               const MaterialInfo* mi__dmn,
                                               FieldInfo* fld );
    Real (*lymph_vessel_stroma_tissue_ratio) ( const MaterialInfo* mi__vnr,
                                               const MaterialInfo* mi__dmn,
                                               FieldInfo* fld );
    Real (*stroma_cross_section) ( const MaterialInfo* mi__vnr,
                                   const MaterialInfo* mi__dmn,
                                   FieldInfo* fld );
    Dyadic3D (*stroma_hydraulic_conductivity) ( const MaterialInfo* mi__vnr,
                                                const MaterialInfo* mi__dmn,
                                                FieldInfo* fld );
    Dyadic3D (*stroma_diffusion_coefficient_2_drug) ( const MaterialInfo* mi__vnr,
                                                      const MaterialInfo* mi__dmn,
                                                      FieldInfo* fld,
                                                      const unsigned int drug_id );
};
//=================================================================================================
#ifdef _VARIABILITY_
#   error "macro defined elsewhere!!!"
#else
#   define _VARIABILITY_   normal_distro(1., 0.0111)
#endif
//=================================================================================================
// constitutive independent function
inline
Real vascular_length ( const Node* nod,     const Point nod_U,
                       const Node* c_nod[], const Point c_nod_U[] ) {
    Real length = 0.0;
    for (unsigned int l=0; l<3; l++) {
        if ( NULL == c_nod[l] ) continue;
        length += Point((*c_nod[l])+c_nod_U[l]-(*nod)-nod_U).norm();
    }
    return length;
}
inline
Real blood_viscosity (const MaterialInfo* mi__vnr) {
    return mi__vnr->get_biochem_param(0)*_VARIABILITY_;
}
inline
Real lymph_viscosity (const MaterialInfo* mi__vnr) {
    return mi__vnr->get_biochem_param(1)*_VARIABILITY_;
}
//-------------------------------------------------------------------------------------------------
inline
Real plasma_viscosity_498 ( const MaterialInfo* mi__vnr, const MaterialInfo* mi__dmn,
                            FieldInfo* fld )
{
    const unsigned int nvars = 4;
    //
    const Real temperature_load = fld->biochem[nvars+8],
               ref_viscosity = mi__vnr->get_biochem_param(2)*_VARIABILITY_,
               coefficient = mi__vnr->get_biochem_param(4);
    return coefficient>0.0
         ? ref_viscosity*exp(-coefficient*temperature_load) : ref_viscosity;
}
inline
Real interstitial_fluid_viscosity_498 ( const MaterialInfo* mi__vnr, const MaterialInfo* mi__dmn,
                                        FieldInfo* fld )
{
    const unsigned int nvars = 4;
    //
    const Real temperature_load = fld->biochem[nvars+8], // Kelvin
               ref_viscosity = mi__vnr->get_biochem_param(3)*_VARIABILITY_,
               coefficient = mi__vnr->get_biochem_param(4);
    return coefficient>0.0
         ? ref_viscosity*exp(-coefficient*temperature_load) : ref_viscosity;
}
//-------------------------------------------------------------------------------------------------
inline
Real reference_temperature (const MaterialInfo* mi__vnr) {
    return mi__vnr->get_biochem_param(9)>0.0
         ? mi__vnr->get_biochem_param(9)*_VARIABILITY_ : 310.65;
}
//
inline
Real vascular_circumference (const VascularInfo* vi) {
    const Real vessel_radius = vi->radius;
    return m2PI*vessel_radius;
}
inline
Real vascular_cross_section (const VascularInfo* vi) {
    const Real vessel_radius = vi->radius;
    return mPI*pow2(vessel_radius);
}
//-------------------------------------------------------------------------------------------------
// constitutive functions that return the hydraulic conductivity of the vascular wall
inline
Real blood_vessel_hydraulic_conductivity_498 ( const VascularInfo* vi, const MaterialInfo* mi__vnr, const MaterialInfo* mi__dmn,
                                               FieldInfo* fld )
{
    const Real vessel_radius = vi->radius,
               blood_dynamic_viscosity = blood_viscosity(mi__vnr);
    return frac_1o8*pow2(vessel_radius)/blood_dynamic_viscosity;
}
inline
Real lymph_vessel_hydraulic_conductivity_498 ( const VascularInfo* vi, const MaterialInfo* mi__vnr, const MaterialInfo* mi__dmn,
                                               FieldInfo* fld )
{
    const Real vessel_radius = vi->radius,
               lymph_dynamic_viscosity = lymph_viscosity(mi__vnr);
    return frac_1o8*pow2(vessel_radius)/lymph_dynamic_viscosity;
}
//-------------------------------------------------------------------------------------------------
// constitutive functions that return the permeability of the vascular wall
inline
Real blood_vessel_hydraulic_permeability_498 ( const VascularInfo* vi, const MaterialInfo* mi__vnr, const MaterialInfo* mi__dmn,
                                               FieldInfo* fld )
{
    if ( VascularInfo::COLLAPSED == vi->status ) return 0.0;
    //
    const Real vessel_thickness = vi->thickness,
               vessel_pore_size = vi->pore_radius,
               vascular_pores_fraction = vi->pores_fraction,
               fluid_viscosity = plasma_viscosity_498(mi__vnr, mi__dmn, fld);
    return frac_1o8*vascular_pores_fraction*pow2(vessel_pore_size)
          /(fluid_viscosity*vessel_thickness);
}
inline
Real lymph_vessel_hydraulic_permeability_498 ( const VascularInfo* vi, const MaterialInfo* mi__vnr, const MaterialInfo* mi__dmn,
                                               FieldInfo* fld )
{
    if ( VascularInfo::COLLAPSED == vi->status ) return 0.0;
    //
    const Real vessel_thickness = vi->thickness,
               vessel_pore_size = vi->pore_radius,
               vascular_pores_fraction = vi->pores_fraction,
               fluid_viscosity = interstitial_fluid_viscosity_498(mi__vnr, mi__dmn, fld);
    return frac_1o8*vascular_pores_fraction*pow2(vessel_pore_size)
          /(fluid_viscosity*vessel_thickness);
}
//-------------------------------------------------------------------------------------------------
// constitutive functions that return the permeability of the vascular wall with respect to
// a drug molecule or a nanoparticle
inline
Real vascular_hydraulic_permeability_2_drug_498 ( const VascularInfo* vi, const MaterialInfo* mi__vnr, const MaterialInfo* mi__dmn,
                                                  FieldInfo* fld, const unsigned int drug_id )
{
    const unsigned int nvars = 4;
    //
    if ( VascularInfo::COLLAPSED == vi->status ) return 0.0;
    // see: https://link.springer.com/article/10.1007/s10237-015-0682-0
    const Real thickness = vi->thickness,
               pores_fraction = vi->pores_fraction,
               pore_radius = vi->pore_radius;
    //
    Real molecule_radius = 0.0;
    if      ( drug_id == 10 ) molecule_radius = mi__vnr->get_biochem_param(210)*_VARIABILITY_;
    else if ( drug_id == 11 ) molecule_radius = mi__vnr->get_biochem_param(211)*_VARIABILITY_;
    else if ( drug_id == 12 ) molecule_radius = mi__vnr->get_biochem_param(212)*_VARIABILITY_;
    else if ( drug_id == 13 ) molecule_radius = mi__vnr->get_biochem_param(213)*_VARIABILITY_;
    else libmesh_error();
    //
    const Real lambda = molecule_radius / pore_radius;
    if ( lambda >= 1.0 ) return 0.0;
    const Real lambda_[] = { 1.0, lambda, pow2(lambda), pow3(lambda), pow4(lambda) };
    const Real delta_[] = { 1.0, (1.0-lambda), pow2(1.0-lambda) };
    //
    const Real F = delta_[2];
    const Real a_[] = { 0, -1.216667E+00,  1.533591E+00, -2.250830E+01, -5.617000E+00, -3.363000E-01, -1.216000E+00, 1.647000E+00 };
    const Real b_[] = { 0,  1.166667E-01, -4.418651E-02,  4.018000E+00, -3.978800E+00, -1.921500E+00,  4.392000E+00, 5.006000E+00 };
    const Real c_ = 31.404888898374958/pow(1.0-lambda, 1.5);
    //
    const Real K_t = c_*(1.0+a_[1]*delta_[1]+a_[2]*delta_[2])
                   + (a_[3]*lambda_[0]+a_[4]*lambda_[1]+a_[5]*lambda_[2]+a_[6]*lambda_[3]+a_[7]*lambda_[4]);
    // hydrodynamic interactions of the drug
    const Real H = m6PI*F/K_t;
    //
    const Real temperature_load = fld->biochem[nvars+8], // Kelvin
               temperature = reference_temperature(mi__vnr) + temperature_load;
    Real fluid_viscosity;
    if      ( VascularInfo::BLOOD == vi->flow ) fluid_viscosity = blood_viscosity(mi__vnr);
    else if ( VascularInfo::LYMPH == vi->flow ) fluid_viscosity = lymph_viscosity(mi__vnr);
    else libmesh_error();
    // Stokes-Einstein equation: drug diffusion coefficient in free-solution
    const Real D_o = Boltzmann_constant*temperature
                   / (m6PI*fluid_viscosity*molecule_radius);
    //
    return pores_fraction*H*D_o/thickness;
}
//-------------------------------------------------------------------------------------------------
// constitutive functions that return the reflection coefficient of the vascular wall with respect to
// a drug molecule or a nanoparticle
inline
Real vascular_reflection_coefficient_2_drug_498 ( const VascularInfo* vi, const MaterialInfo* mi__vnr, const MaterialInfo* mi__dmn,
                                                  FieldInfo* fld, const unsigned int drug_id )
{
    if ( VascularInfo::COLLAPSED == vi->status ) return 1.0;
    // see: https://link.springer.com/article/10.1007/s10237-015-0682-0
    const Real pore_radius = vi->pore_radius;
    //
    Real molecule_radius = 0.0;
    if      ( drug_id == 10 ) molecule_radius = mi__vnr->get_biochem_param(210)*_VARIABILITY_;
    else if ( drug_id == 11 ) molecule_radius = mi__vnr->get_biochem_param(211)*_VARIABILITY_;
    else if ( drug_id == 12 ) molecule_radius = mi__vnr->get_biochem_param(212)*_VARIABILITY_;
    else if ( drug_id == 13 ) molecule_radius = mi__vnr->get_biochem_param(213)*_VARIABILITY_;
    else libmesh_error();
    //
    const Real lambda = molecule_radius / pore_radius;
    if ( lambda >= 1.0 ) return 1.0;
    const Real lambda_[] = { 1.0, lambda, pow2(lambda), pow3(lambda), pow4(lambda) };
    const Real delta_[] = { 1.0, (1.0-lambda), pow2(1.0-lambda) };
    //
    const Real F = delta_[2];
    const Real a_[] = { 0, -1.216667E+00,  1.533591E+00, -2.250830E+01, -5.617000E+00, -3.363000E-01, -1.216000E+00, 1.647000E+00 };
    const Real b_[] = { 0,  1.166667E-01, -4.418651E-02,  4.018000E+00, -3.978800E+00, -1.921500E+00,  4.392000E+00, 5.006000E+00 };
    const Real c_ = 31.404888898374958/pow(1.0-lambda, 1.5);
    //
    const Real K_t = c_*(1.0+a_[1]*delta_[1]+a_[2]*delta_[2])
                   + (a_[3]*lambda_[0]+a_[4]*lambda_[1]+a_[5]*lambda_[2]+a_[6]*lambda_[3]+a_[7]*lambda_[4]);
    const Real K_s = c_*(1.0+b_[1]*delta_[1]+b_[2]*delta_[2])
                   + (b_[3]*lambda_[0]+b_[4]*lambda_[1]+b_[5]*lambda_[2]+b_[6]*lambda_[3]+b_[7]*lambda_[4]);
    // electrostatic interactions of the drug
    const Real W = 0.5*(2.0-F)*F*K_s/K_t;
    //
    if ( W > 1.0 ) return 0.0;
    return 1.0-W;
}
//-------------------------------------------------------------------------------------------------
// constitutive functions that return the pressure difference between the interstitial fluid
// and the plasma fluid
inline
Real osmotic_pressure_difference_498 ( const VascularInfo* vi, const MaterialInfo* mi__vnr, const MaterialInfo* mi__dmn,
                                       FieldInfo* fld )
{
    const Real osmotic_reflection_coeff = VascularInfo::COLLAPSED == vi->status
                                        ? 0.0 : mi__dmn->get_biochem_param(60),
               osmotic_press__plasma = mi__dmn->get_biochem_param(61),
               osmotic_press__intfld = mi__dmn->get_biochem_param(62);
    return osmotic_reflection_coeff*(osmotic_press__plasma-osmotic_press__intfld);
}
//-------------------------------------------------------------------------------------------------
// constitutive functions that return the vascular density over (a finite volume of) the stroma
inline
Real blood_vessel_stroma_ratio_498 ( const MaterialInfo* mi__vnr, const MaterialInfo* mi__dmn,
                                     FieldInfo* fld )
{
    return mi__dmn->get_biochem_param(63)*_VARIABILITY_;
}
//
inline
Real lymph_vessel_stroma_ratio_498 ( const MaterialInfo* mi__vnr, const MaterialInfo* mi__dmn,
                                     FieldInfo* fld )
{
    return mi__dmn->get_biochem_param(64)*_VARIABILITY_;
}
//-------------------------------------------------------------------------------------------------
inline
Real stroma_cross_section_498 ( const MaterialInfo* mi__vnr, const MaterialInfo* mi__dmn,
                                FieldInfo* fld )
{
    const Real vascular_mean_radius = 0.5*(mi__vnr->get_biochem_param(52)
                                          +mi__vnr->get_biochem_param(53)),
               vascular_density_ratio = blood_vessel_stroma_ratio_498(mi__vnr, mi__dmn, fld);
    return m2PI*vascular_mean_radius/vascular_density_ratio;
}
//-------------------------------------------------------------------------------------------------
// constitutive functions that return the hydraulic conductivity of the stroma
inline
Dyadic3D stroma_hydraulic_conductivity_498 ( const MaterialInfo* mi__vnr, const MaterialInfo* mi__dmn,
                                             FieldInfo* fld )
{
    // const unsigned int nvars = 4;
    //
    const Real kappa_0 = mi__dmn->get_biochem_param(70),
               scale   = mi__dmn->get_biochem_param(71),
               kappa_T = mi__dmn->get_biochem_param(72)!=0.0 ? mi__dmn->get_biochem_param(72) : kappa_0*0.999999,
               kappa_C = mi__dmn->get_biochem_param(73)!=0.0 ? mi__dmn->get_biochem_param(73) : kappa_0*0.999999;
    const int dependence = mi__dmn->get_biochem_param(74);
    const Real ecm = apply_lbound(1.0e-6, fld->biochem[3]);
    //
    std::vector<Real> kappa(SP3D);
    std::vector<Vector3D> n; // principal directions (unit) vectors
    if ( 0 == dependence ) {
        n.push_back( set_vector(1., 0., 0.) );
        n.push_back( set_vector(0., 1., 0.) );
        n.push_back( set_vector(0., 0., 1.) );
        //
        for (int i=0; i<SP3D; i++) {
            kappa[i] = kappa_0; // baseline value
        }
    } else if ( 1 == dependence ) {
        std::vector<Real> epsilon; // Green strain (principal values)
        eigen_solve(fld->strain, epsilon, n);
        //
        for (int i=0; i<SP3D; i++) {
            const Real lambda = sqrt(2.0*epsilon[i]+1.0); // stretch ratio
            if ( lambda >= 2.0 ) { // tension... too much of it
                kappa[i] = kappa_T;
            } else if ( lambda >= 1.0 ) { // tension
                kappa[i] = (0.5*(kappa_0-kappa_T)) * cos(mPI*(1.0-lambda))
                         + (0.5*(kappa_0+kappa_T));
            } else { // compression
                kappa[i] = (0.5*(kappa_0-kappa_C)) * cos(mPI*(1.0-lambda))
                         + (0.5*(kappa_0+kappa_C));
            }
        }
    } else libmesh_error();
    //
    for (int i=0; i<SP3D; i++) {
        if      ( scale > 0.0 ) kappa[i] *= ( scale*ecm)*_VARIABILITY_;
        else if ( scale < 0.0 ) kappa[i] *= (-scale/ecm)*_VARIABILITY_;
        else                    kappa[i] *= _VARIABILITY_;
    }
    //
    Dyadic3D K;
    K.comp = kappa[0] * tensor(n[0], n[0]).comp
           + kappa[1] * tensor(n[1], n[1]).comp
           + kappa[2] * tensor(n[2], n[2]).comp;
    return K;
}
//-------------------------------------------------------------------------------------------------
// constitutive functions that return the diffusion coefficient of a drug molecule or a nanoparticle
// with respect to the stroma
inline
Dyadic3D stroma_diffusion_coefficient_2_drug_498 ( const MaterialInfo* mi__vnr, const MaterialInfo* mi__dmn,
                                                   FieldInfo* fld, const unsigned int drug_id )
{
    const unsigned int nvars = 4;
    //
    Real molecule_radius = 0.0;
    if      ( drug_id == 10 ) molecule_radius = mi__vnr->get_biochem_param(210)*_VARIABILITY_;
    else if ( drug_id == 11 ) molecule_radius = mi__vnr->get_biochem_param(211)*_VARIABILITY_;
    else if ( drug_id == 12 ) molecule_radius = mi__vnr->get_biochem_param(212)*_VARIABILITY_;
    else if ( drug_id == 13 ) molecule_radius = mi__vnr->get_biochem_param(213)*_VARIABILITY_;
    else libmesh_error();
    //
    const Real temperature_load = fld->biochem[nvars+8], // Kelvin
               temperature = reference_temperature(mi__vnr) + temperature_load;
    const Real fluid_viscosity = interstitial_fluid_viscosity_498(mi__vnr, mi__dmn, fld);
    const Real fibre_volume_fraction = mi__dmn->get_biochem_param(300),
               matrix_porosity       = 1.0 - fibre_volume_fraction,
               fibre_radius          = mi__dmn->get_biochem_param(301)*_VARIABILITY_;
    const Real diffusion_factor_coefficient = mi__dmn->get_biochem_param(302),
               Kozeny_factor_coefficient    = mi__dmn->get_biochem_param(303);
    const int dependence = mi__dmn->get_biochem_param(304);
    //
    std::vector<Real> diff(SP3D);
    std::vector<Vector3D> n; // principal directions (unit) vectors
    if ( 0 == dependence ) {
        n.push_back( set_vector(1., 0., 0.) );
        n.push_back( set_vector(0., 1., 0.) );
        n.push_back( set_vector(0., 0., 1.) );
        //
        for (int i=0; i<SP3D; i++) {
            diff[i] = 1.0; // baseline value
        }
    } else if ( 1 == dependence ) {
        std::vector<Real> epsilon; // Green strain (principal values)
        eigen_solve(fld->strain, epsilon, n);
        // make these sanity checks
        if ( diffusion_factor_coefficient<1.0 ) libmesh_error();
        //
        for (int i=0; i<SP3D; i++) {
            const Real lambda = sqrt(2.0*epsilon[i]+1.0); // stretch ratio
            if ( lambda > 2.0 ) { // tension... too much of it
                diff[i] = 1.0/diffusion_factor_coefficient;
            } else if ( lambda > 1.0 ) { // tension
                diff[i] = pow(diffusion_factor_coefficient, 1.0-lambda);
            } else if ( lambda > 0.0 ) { // compression
                diff[i] = pow(diffusion_factor_coefficient, lambda-1.0);
            } else libmesh_error();
        }
    } else libmesh_error();
    //
    // Stokes-Einstein equation [https://en.wikipedia.org/wiki/Einstein_relation_(kinetic_theory)]
    Real D_o = Boltzmann_constant*temperature
             / (m6PI*fluid_viscosity*molecule_radius);
    // definition of the equations used, refer to: Pluen et al. (1999) Biophys J. 77(1): 542–552.
    // [https://www.ncbi.nlm.nih.gov/pmc/articles/PMC1300351/]
    if        ( fibre_volume_fraction > 0.0 && Kozeny_factor_coefficient == 0.0 ) {
        // Ogston model [https://www.jstor.org/stable/78361?seq=1#page_scan_tab_contents]
        D_o *= exp(-sqrt(fibre_volume_fraction)*molecule_radius/fibre_radius);
    } else if ( fibre_volume_fraction > 0.0 && Kozeny_factor_coefficient >  0.0 ) {
        const Real f = (1.0-matrix_porosity), // coeff.
                   f2 = pow2(f), // coeff.
                   c = 2.0*pow3(matrix_porosity)/f, // coeff.
                   l = log(1.0/f), // coeff.
                   k_p = c/(l+4.0*f-f2-3.0), // parallel fibres coeff.
                   k_v = c/(l-(1.0-f2)/(1.0+f2)), // vertical fibres coeff.
                   k = Kozeny_factor_coefficient*(2.0*k_v+k_p)/3.0; // average fibres coef.
        const Real kappa = matrix_porosity*pow2(molecule_radius)/k;
        // Phillips et al. (1989) AIChE J. 35:1761–1769. DOI:10.1002/aic.690351102
        D_o /= (1.0+sqrt(pow2(molecule_radius)/kappa)+(pow2(molecule_radius)/kappa)/3.0);
    }
    // now update the diffusion coefficient for all 3 principal directions
    for (int i=0; i<SP3D; i++) {
        diff[i] *= D_o*_VARIABILITY_;
    }
    //
    Dyadic3D D;
    D.comp = diff[0] * tensor(n[0], n[0]).comp
           + diff[1] * tensor(n[1], n[1]).comp
           + diff[2] * tensor(n[2], n[2]).comp;
    return D;
}
//=================================================================================================
#undef _VARIABILITY_
//=================================================================================================
#endif // __TUMOUR_ANGIOGENESIS_CONSTITUTIVE_H__
//=================================================================================================
