/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __UTILS__ALGEBRA_2D_H__
#define __UTILS__ALGEBRA_2D_H__
//=================================================================================================
inline
Dual<int> indices_Voigt_2d (const int i) {
    switch ( i ) {
    case 0 : return Dual<int>(0, 0);
    case 1 : return Dual<int>(1, 1);
    case 2 : return Dual<int>(0, 1);
    default : libmesh_error();
    }
}
//-------------------------------------------------------------------------------------------------
inline
Real trace (const Dyadic2D& A) {
    return A(0,0)+A(1,1);
}
inline
Real determinant (const Dyadic2D& A) {
    return A(0,0)*A(1,1)-A(0,1)*A(1,0);
}
inline
Dyadic2D transpose (const Dyadic2D& A) {
    Dyadic2D A_t;
    A_t.comp = A(0,0), A(1,0),
               A(0,1), A(1,1);
    return A_t;
}
inline
Dyadic2D inverse (const Dyadic2D& A) {
    const Real det_A = determinant(A);
    libmesh_assert( det_A > 1.0e-12 );
    Dyadic2D A_inv;
    A_inv.comp = A(1,1)/det_A,-A(0,1)/det_A,
                -A(1,0)/det_A, A(0,0)/det_A;
    return A_inv;
}
inline
Real product (const Vector2D& L, const Dyadic2D& D, const Vector2D& R) {
    return L(0)*(D(0,0)*R(0)+D(0,1)*R(1))
          +L(1)*(D(1,0)*R(0)+D(1,1)*R(1));
}
inline
Real product (const Vector2D& A, const Vector2D& B) {
    return A(0)*B(0)+A(1)*B(1);
}
inline
Real operator* (const Vector2D& A, const Vector2D& B) {
    return product(A, B);
}
inline
Real magnitude (const Vector2D& A) {
    return sqrt(product(A, A));
}
inline
Vector2D unit (const Vector2D& A) {
    const Real mag_A = magnitude(A);
    Vector2D R;
    R.comp = A(0)/mag_A, A(1)/mag_A;
    return R;
}
inline
Real inner_product (const Dyadic2D& A, const Dyadic2D& B) {
    return A(0,0)*B(0,0)+A(0,1)*B(1,0)
          +A(1,0)*B(0,1)+A(1,1)*B(1,1);
}
inline
Real outer_product (const Dyadic2D& A, const Dyadic2D& B) {
    return A(0,0)*B(0,0)+A(0,1)*B(0,1)
          +A(1,0)*B(1,0)+A(1,1)*B(1,1);
}
inline
Dyadic2D scale (const Real& s, const Dyadic2D& A) {
    Dyadic2D R;
    R.comp = s*A(0,0), s*A(0,1),
             s*A(1,0), s*A(1,1);
    return R;
}
inline
Vector2D set_vector (const Real& x0, const Real& x1) {
    Vector2D R;
    R.comp = x0, x1;
    return R;
}
inline
bool are_equal (const Vector2D& A, const Vector2D& B, Real tol =1.0e-6) {
    return ( fabs(A(0)-B(0))<=tol && fabs(A(1)-B(1))<=tol );
}
inline
bool are_equal (const Dyadic2D& A, const Dyadic2D& B, Real tol =1.0e-6) {
    return ( fabs(A(0,0)-B(0,0))<=tol && fabs(A(0,1)-B(0,1))<=tol &&
             fabs(A(1,0)-B(1,0))<=tol && fabs(A(1,1)-B(1,1))<=tol );
}
//-------------------------------------------------------------------------------------------------
inline
Dyadic2D tensor (const Vector2D& A, const Vector2D& B) {
    Dyadic2D R;
    R.comp = A(0)*B(0), A(0)*B(1),
             A(1)*B(0), A(1)*B(1);
    return R;
}
inline
Dyadic2D tensor (const Vector2D& A) {
    return tensor(A, A);
}
inline
Triadic2D tensor (const Vector2D& A, const Vector2D& B, const Vector2D& C) {
    blitz::firstIndex  i;
    blitz::secondIndex j;
    blitz::thirdIndex  k;
    Triadic2D R;
    R.comp = A.comp(i)*B.comp(j)*C.comp(k);
    return R;
}
inline
Quindic2D tensor (const Vector2D& A, const Vector2D& B, const Vector2D& C, const Vector2D& D) {
    blitz::firstIndex  i;
    blitz::secondIndex j;
    blitz::thirdIndex  k;
    blitz::fourthIndex l;
    Quindic2D R;
    R.comp = A.comp(i)*B.comp(j)*C.comp(k)*D.comp(l);
    return R;
}
inline
Quindic2D tensor (const Dyadic2D& A, const Dyadic2D& B) {
    blitz::firstIndex  i;
    blitz::secondIndex j;
    blitz::thirdIndex  k;
    blitz::fourthIndex l;
    Quindic2D R;
    R.comp = A.comp(i,j)*B.comp(k,l);
    return R;
}
inline
Dyadic2D symm (const Dyadic2D& A) {
    Dyadic2D R;
    R.comp = A(0,0),              0.5*(A(0,1)+A(1,0)),
             0.5*(A(1,0)+A(0,1)), A(1,1);
    return R;
}
inline
Dyadic2D skew (const Dyadic2D& A) {
    Dyadic2D R;
    R.comp = 0.0,                 0.5*(A(0,1)-A(1,0)),
             0.5*(A(1,0)-A(0,1)), 0.0;
    return R;
}
inline
Quindic2D symm_tensor (const Dyadic2D& A) {
    blitz::firstIndex  i;
    blitz::secondIndex j;
    blitz::thirdIndex  k;
    blitz::fourthIndex l;
    Quindic2D R;
    R.comp = 0.5*(A.comp(i,k)*A.comp(j,l) + A.comp(i,l)*A.comp(j,k));
    return R;
}
inline
Quindic2D op_symm_tensor (const Dyadic2D& A) {
    blitz::firstIndex  i;
    blitz::secondIndex j;
    blitz::thirdIndex  k;
    blitz::fourthIndex l;
    Quindic2D R;
    R.comp = -0.5*(A.comp(i,k)*A.comp(j,l) + A.comp(i,l)*A.comp(j,k));
    return R;
}
//-------------------------------------------------------------------------------------------------
inline
Dyadic2D mul_im_mj (const Dyadic2D& A, const Dyadic2D& B) {
    Dyadic2D R;
    blitz::firstIndex  i;
    blitz::secondIndex j;
    blitz::thirdIndex  k;
    R.comp = sum(A.comp(i,k)*B.comp(k,j), k);
    return R;
}
inline
Dyadic2D mul_mi_mj (const Dyadic2D& A, const Dyadic2D& B) {
    Dyadic2D R;
    blitz::firstIndex  i;
    blitz::secondIndex j;
    blitz::thirdIndex  k;
    R.comp = sum(A.comp(k,i)*B.comp(k,j), k);
    return R;
}
inline
Dyadic2D mul_im_jm (const Dyadic2D& A, const Dyadic2D& B) {
    Dyadic2D R;
    blitz::firstIndex  i;
    blitz::secondIndex j;
    blitz::thirdIndex  k;
    R.comp = sum(A.comp(i,k)*B.comp(j,k), k);
    return R;
}
inline
Dyadic2D mul_mi_jm (const Dyadic2D& A, const Dyadic2D& B) {
    Dyadic2D R;
    blitz::firstIndex  i;
    blitz::secondIndex j;
    blitz::thirdIndex  k;
    R.comp = sum(A.comp(k,i)*B.comp(j,k), k);
    return R;
}
//-------------------------------------------------------------------------------------------------
inline
Vector2D mul_im_m (const Dyadic2D& A, const Vector2D& B) {
    Vector2D R;
    blitz::firstIndex  i;
    blitz::secondIndex j;
    R.comp = sum(A.comp(i,j)*B.comp(j), j);
    return R;
}
inline
Vector2D mul_m_mi (const Vector2D& B, const Dyadic2D& A) {
    Vector2D R;
    blitz::firstIndex  i;
    blitz::secondIndex j;
    R.comp = sum(B.comp(j)*A.comp(j,i), j);
    return R;
}
//-------------------------------------------------------------------------------------------------
inline
Dyadic2D mul_ijmn_mn (const Quindic2D& Q, const Dyadic2D& D) {
    Dyadic2D R;
    for (int i=0; i<SP2D; i++) {
        for (int j=0; j<SP2D; j++) {
            Dyadic2D Qd;
            Qd.comp = Q(i,j,0,0), Q(i,j,0,1),
                      Q(i,j,1,0), Q(i,j,1,1);
            R.comp(i,j) = outer_product(Qd, D);
        }
    }
    return R;
}
inline
Dyadic2D mul_mn_mnij (const Dyadic2D& D, const Quindic2D& Q) {
    Dyadic2D R;
    for (int i=0; i<SP2D; i++) {
        for (int j=0; j<SP2D; j++) {
            Dyadic2D Qd;
            Qd.comp = Q(0,0,i,j), Q(0,1,i,j),
                      Q(1,0,i,j), Q(1,1,i,j);
            R.comp(i,j) = outer_product(D, Qd);
        }
    }
    return R;
}
//-------------------------------------------------------------------------------------------------
inline
Quindic2D mul_mjkl_im (const Quindic2D& Q, const Dyadic2D& D) {
    Quindic2D R;
    for (int i=0; i<SP2D; i++)
        for (int j=0; j<SP2D; j++)
            for (int k=0; k<SP2D; k++)
                for (int l=0; l<SP2D; l++)
                    R.comp(i,j,k,l) = Q(0,j,k,l)*D(i,0)+Q(1,j,k,l)*D(i,1);
    return R;
}
inline
Quindic2D mul_imkl_jm (const Quindic2D& Q, const Dyadic2D& D) {
    Quindic2D R;
    for (int i=0; i<SP2D; i++)
        for (int j=0; j<SP2D; j++)
            for (int k=0; k<SP2D; k++)
                for (int l=0; l<SP2D; l++)
                    R.comp(i,j,k,l) = Q(i,0,k,l)*D(j,0)+Q(i,1,k,l)*D(j,1);
    return R;
}
inline
Quindic2D mul_ijml_km (const Quindic2D& Q, const Dyadic2D& D) {
    Quindic2D R;
    for (int i=0; i<SP2D; i++)
        for (int j=0; j<SP2D; j++)
            for (int k=0; k<SP2D; k++)
                for (int l=0; l<SP2D; l++)
                    R.comp(i,j,k,l) = Q(i,j,0,l)*D(k,0)+Q(i,j,1,l)*D(k,1);
    return R;
}
inline
Quindic2D mul_ijkm_lm (const Quindic2D& Q, const Dyadic2D& D) {
    Quindic2D R;
    for (int i=0; i<SP2D; i++)
        for (int j=0; j<SP2D; j++)
            for (int k=0; k<SP2D; k++)
                for (int l=0; l<SP2D; l++)
                    R.comp(i,j,k,l) = Q(i,j,k,0)*D(l,0)+Q(i,j,k,1)*D(l,1);
    return R;
}
inline
Quindic2D mul_ijmn_mnkl (const Quindic2D& A, const Quindic2D& B) {
    Quindic2D R;
    for (int i=0; i<SP2D; i++) {
        for (int j=0; j<SP2D; j++) {
            Dyadic2D Ad;
            Ad.comp = A(i,j,0,0), A(i,j,0,1),
                      A(i,j,1,0), A(i,j,1,1);
            for (int k=0; k<SP2D; k++) {
                for (int l=0; l<SP2D; l++) {
                    Dyadic2D Bd;
                    Bd.comp = B(0,0,k,l), B(0,1,k,l),
                              B(1,0,k,l), B(1,1,k,l);
                    R.comp(i,j,k,l) = outer_product(Ad, Bd);
                }
            }
        }
    }
    return R;
}
//-------------------------------------------------------------------------------------------------
inline
Dyadic2D mul_m_imjn_n (const Vector2D& A, const Quindic2D& Q, const Vector2D& B) {
    Dyadic2D R;
    for (int i=0; i<SP2D; i++) {
        for (int j=0; j<SP2D; j++) {
            R.comp(i,j) = A(0)*Q(i,0,j,0)*B(0)+A(0)*Q(i,0,j,1)*B(1)
                         +A(1)*Q(i,1,j,0)*B(0)+A(1)*Q(i,1,j,1)*B(1);
        }
    }
    return R;
}
inline
Dyadic2D mul_ijm_m (const Triadic2D& T, const Vector2D& V) {
    Dyadic2D O;
    O.comp(0,0) = T(0,0,0)*V(0)+T(0,0,1)*V(1);
    O.comp(0,1) = T(0,1,0)*V(0)+T(0,1,1)*V(1);
    O.comp(1,0) = T(1,0,0)*V(0)+T(1,0,1)*V(1);
    O.comp(1,1) = T(1,1,0)*V(0)+T(1,1,1)*V(1);
    return O;
}
inline
Dyadic2D mul_imj_m (const Triadic2D& T, const Vector2D& V) {
    Dyadic2D O;
    O.comp(0,0) = T(0,0,0)*V(0)+T(0,1,0)*V(1);
    O.comp(0,1) = T(0,0,1)*V(0)+T(0,1,1)*V(1);
    O.comp(1,0) = T(1,0,0)*V(0)+T(1,1,0)*V(1);
    O.comp(1,1) = T(1,0,1)*V(0)+T(1,1,1)*V(1);
    return O;
}
inline
Dyadic2D mul_mij_m (const Triadic2D& T, const Vector2D& V) {
    Dyadic2D O;
    O.comp(0,0) = T(0,0,0)*V(0)+T(1,0,0)*V(1);
    O.comp(0,1) = T(0,0,1)*V(0)+T(1,0,1)*V(1);
    O.comp(1,0) = T(0,1,0)*V(0)+T(1,1,0)*V(1);
    O.comp(1,1) = T(0,1,1)*V(0)+T(1,1,1)*V(1);
    return O;
}
//-------------------------------------------------------------------------------------------------
inline
Triadic2D mul_ijm_mk (const Triadic2D& C, const Dyadic2D& A) {
    Triadic2D R;
    for (int i=0; i<SP2D; i++)
        for (int j=0; j<SP2D; j++)
            for (int k=0; k<SP2D; k++)
                R.comp(i,j,k) = C(i,j,0)*A(0,k)+C(i,j,1)*A(1,k);
    return R;
}
inline
Triadic2D mul_ijm_km (const Triadic2D& C, const Dyadic2D& A) {
    Triadic2D R;
    for (int i=0; i<SP2D; i++)
        for (int j=0; j<SP2D; j++)
            for (int k=0; k<SP2D; k++)
                R.comp(i,j,k) = C(i,j,0)*A(k,0)+C(i,j,1)*A(k,1);
    return R;
}
inline
Vector2D mul_mn_mni (const Dyadic2D& A, const Triadic2D& C) {
    Vector2D R;
    for (int i=0; i<SP2D; i++) {
        Dyadic2D Cd;
        Cd.comp = C(0,0,i), C(0,1,i),
                  C(1,0,i), C(1,1,i);
        R.comp(i) = outer_product(A, Cd);
    }
    return R;
}
inline
Vector2D mul_imn_mn (const Triadic2D& C, const Dyadic2D& A) {
    Vector2D R;
    for (int i=0; i<SP2D; i++) {
        Dyadic2D Cd;
        Cd.comp = C(i,0,0), C(i,0,1),
                  C(i,1,0), C(i,1,1);
        R.comp(i) = outer_product(Cd, A);
    }
    return R;
}
inline
Triadic2D mul_imn_mnjk (const Triadic2D& C, const Quindic2D& Q) {
    Triadic2D R;
    for (int i=0; i<SP2D; i++) {
        Dyadic2D Cd;
        Cd.comp = C(i,0,0), C(i,0,1),
                  C(i,1,0), C(i,1,1);
        for (int j=0; j<SP2D; j++) {
            for (int k=0; k<SP2D; k++) {
                Dyadic2D Qd;
                Qd.comp = Q(0,0,j,k), Q(0,1,j,k),
                          Q(1,0,j,k), Q(1,1,j,k);
                R.comp(i,j,k) = outer_product(Cd, Qd);
            }
        }
    }
    return R;
}
inline
Triadic2D mul_mni_mnjk (const Triadic2D& C, const Quindic2D& Q) {
    Triadic2D R;
    for (int i=0; i<SP2D; i++) {
        Dyadic2D Cd;
        Cd.comp = C(0,0,i), C(0,1,i),
                  C(1,0,i), C(1,1,i);
        for (int j=0; j<SP2D; j++) {
            for (int k=0; k<SP2D; k++) {
                Dyadic2D Qd;
                Qd.comp = Q(0,0,j,k), Q(0,1,j,k),
                          Q(1,0,j,k), Q(1,1,j,k);
                R.comp(i,j,k) = outer_product(Cd, Qd);
            }
        }
    }
    return R;
}
inline
Dyadic2D mul_imn_mnj (const Triadic2D& A, const Triadic2D& B) {
    Dyadic2D R;
    for (int i=0; i<SP2D; i++) {
        Dyadic2D Ad;
        Ad.comp = A(i,0,0), A(i,0,1),
                  A(i,1,0), A(i,1,1);
        for (int j=0; j<SP2D; j++) {
            Dyadic2D Bd;
            Bd.comp = B(0,0,j), B(0,1,j),
                      B(1,0,j), B(1,1,j);
            R.comp(i,j) = outer_product(Ad, Bd);
        }
    }
    return R;
}
inline
Dyadic2D squared (const Dyadic2D& A) {
    return mul_im_mj(A, A);
}
//-------------------------------------------------------------------------------------------------
inline
Quindic2D mul_ikmjln_mn (const Sextic2D& S, const Dyadic2D& D) {
    Quindic2D R;
    for (int i=0; i<SP2D; i++) {
        for (int j=0; j<SP2D; j++) {
            for (int k=0; k<SP2D; k++) {
                for (int l=0; l<SP2D; l++) {
                    Dyadic2D Sd;
                    Sd.comp = S(i,k,0,j,l,0), S(i,k,0,j,l,1),
                              S(i,k,1,j,l,0), S(i,k,1,j,l,1);
                    R.comp(i,j,k,l) = outer_product(Sd, D);
                }
            }
        }
    }
    return R;
}
inline
Quindic2D mul_ilmjkn_mn (const Sextic2D& S, const Dyadic2D& D) {
    Quindic2D R;
    for (int i=0; i<SP2D; i++) {
        for (int j=0; j<SP2D; j++) {
            for (int k=0; k<SP2D; k++) {
                for (int l=0; l<SP2D; l++) {
                    Dyadic2D Sd;
                    Sd.comp = S(i,l,0,j,k,0), S(i,l,0,j,k,1),
                              S(i,l,1,j,k,0), S(i,l,1,j,k,1);
                    R.comp(i,j,k,l) = outer_product(Sd, D);
                }
            }
        }
    }
    return R;
}
inline
Quindic2D mul_jkmiln_mn (const Sextic2D& S, const Dyadic2D& D) {
    Quindic2D R;
    for (int i=0; i<SP2D; i++) {
        for (int j=0; j<SP2D; j++) {
            for (int k=0; k<SP2D; k++) {
                for (int l=0; l<SP2D; l++) {
                    Dyadic2D Sd;
                    Sd.comp = S(j,k,0,i,l,0), S(j,k,0,i,l,1),
                              S(j,k,1,i,l,0), S(j,k,1,i,l,1);
                    R.comp(i,j,k,l) = outer_product(Sd, D);
                }
            }
        }
    }
    return R;
}
inline
Quindic2D mul_jlmikn_mn (const Sextic2D& S, const Dyadic2D& D) {
    Quindic2D R;
    for (int i=0; i<SP2D; i++) {
        for (int j=0; j<SP2D; j++) {
            for (int k=0; k<SP2D; k++) {
                for (int l=0; l<SP2D; l++) {
                    Dyadic2D Sd;
                    Sd.comp = S(j,l,0,i,k,0), S(j,l,0,i,k,1),
                              S(j,l,1,i,k,0), S(j,l,1,i,k,1);
                    R.comp(i,j,k,l) = outer_product(Sd, D);
                }
            }
        }
    }
    return R;
}
//-------------------------------------------------------------------------------------------------
inline
Dyadic2D mul_mi_mn_nj (const Dyadic2D& L, const Dyadic2D& D, const Dyadic2D& R) {
    Dyadic2D O;
    for (int i=0; i<SP2D; i++) {
        for (int j=0; j<SP2D; j++) {
            O.comp(i,j) = L(0,i)*D(0,0)*R(0,j)+L(0,i)*D(0,1)*R(1,j)
                         +L(1,i)*D(1,0)*R(0,j)+L(1,i)*D(1,1)*R(1,j);
        }
    }
    return O;
}
inline
Dyadic2D mul_mi_mn_jn (const Dyadic2D& L, const Dyadic2D& D, const Dyadic2D& R) {
    Dyadic2D O;
    for (int i=0; i<SP2D; i++) {
        for (int j=0; j<SP2D; j++) {
            O.comp(i,j) = L(0,i)*D(0,0)*R(j,0)+L(0,i)*D(0,1)*R(j,1)
                         +L(1,i)*D(1,0)*R(j,0)+L(1,i)*D(1,1)*R(j,1);
        }
    }
    return O;
}
inline
Dyadic2D mul_mj_mn_ni (const Dyadic2D& L, const Dyadic2D& D, const Dyadic2D& R) {
    Dyadic2D O;
    for (int i=0; i<SP2D; i++) {
        for (int j=0; j<SP2D; j++) {
            O.comp(i,j) = L(0,j)*D(0,0)*R(0,i)+L(0,j)*D(0,1)*R(1,i)
                         +L(1,j)*D(1,0)*R(0,i)+L(1,j)*D(1,1)*R(1,i);
        }
    }
    return O;
}
inline
Dyadic2D mul_mj_mn_in (const Dyadic2D& L, const Dyadic2D& D, const Dyadic2D& R) {
    Dyadic2D O;
    for (int i=0; i<SP2D; i++) {
        for (int j=0; j<SP2D; j++) {
            O.comp(i,j) = L(0,j)*D(0,0)*R(i,0)+L(0,j)*D(0,1)*R(i,1)
                         +L(1,j)*D(1,0)*R(i,0)+L(1,j)*D(1,1)*R(i,1);
        }
    }
    return O;
}
//-------------------------------------------------------------------------------------------------
inline
Dyadic2D volumetric (const Dyadic2D& A) {
    const Real tr_A = frac_1o2*(A(0,0)+A(1,1));
    Dyadic2D A_vol;
    A_vol.comp = tr_A, 0.0,
                 0.0,  tr_A;
    return A_vol;
}
inline
Dyadic2D deviatoric (const Dyadic2D& A) {
    Dyadic2D A_dev;
    A_dev.comp = A.comp - volumetric(A).comp;
    return A_dev;
}
inline
Dyadic2D identity_2d () {
    Dyadic2D I;
    I.comp = 1.0, 0.0,
             0.0, 1.0;
    return I;
}
inline
Dyadic2D random_diagonal_2d (const Real vMin, const Real vMax) {
    Dyadic2D R;
    R.comp = uniform_distro(vMin, vMax), 0.0,
             0.0, uniform_distro(vMin, vMax);
    return R;
}
//=================================================================================================
#endif // __UTILS__ALGEBRA_2D_H__
//=================================================================================================
