/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __UTILS__ALGEBRA_MORE_H__
#define __UTILS__ALGEBRA_MORE_H__
//=================================================================================================
class Vector3DGradient {
public:
    Vector3DGradient () { _grad[0] = _grad[1] = _grad[2] = RealGradient(0.0, 0.0, 0.0); }
    ~Vector3DGradient () {}
    //
    inline
    Real operator() (unsigned int i, unsigned int Dx) const { return _grad[i](Dx); }
    inline
    Vector3D     vector   (unsigned int Dx) const { Vector3D v; v.comp=_grad[0](Dx),_grad[1](Dx),_grad[2](Dx); return v; }
    inline
    RealGradient gradient (unsigned int i) const { return _grad[i]; }
    //
    inline
    Real& operator() (unsigned int i, unsigned int Dx) { return _grad[i](Dx); }
    inline
    RealGradient& operator() (unsigned int i) { return _grad[i]; }
private:
    RealGradient _grad[SP3D];
};
//=================================================================================================
inline
Vector3D cross (const Vector3D& u, const Vector3D& v) {
    Vector3D p;
    p.comp = (u(1)*v(2)-u(2)*v(1)), -(u(0)*v(2)-u(2)*v(0)), (u(0)*v(1)-u(1)*v(0));
    return p;
}
inline
Vector3D vec (const RealGradient& grad) {
    Vector3D R;
    R.comp = grad(0), grad(1), grad(2);
    return R;
}
inline
RealGradient vec (const Vector3D& v) {
    RealGradient R(v(0), v(1), v(2));
    return R;
}
inline
Point coordinates (const Vector3D& v) {
    Point p(v(0), v(1), v(2));
    return p;
}
inline
Dyadic3D tensor (const RealGradient& A, const RealGradient& B) {
    Dyadic3D R;
    R.comp = A(0)*B(0), A(0)*B(1), A(0)*B(2),
             A(1)*B(0), A(1)*B(1), A(1)*B(2),
             A(2)*B(0), A(2)*B(1), A(2)*B(2);
    return R;
}
inline
Dyadic3D tensor (const RealGradient& A) {
    return tensor(A, A);
}
//-------------------------------------------------------------------------------------------------
inline
Real first_invariant (const Dyadic3D& A) {
    return trace(A);
}
inline
Real second_invariant (const Dyadic3D& A) {
    return 0.5*(pow2(first_invariant(A))-inner_product(A, A));
}
inline
Real third_invariant (const Dyadic3D& A) {
    return determinant(A);
}
//-------------------------------------------------------------------------------------------------
inline
Triadic3D permutation () {
    const Point e0(1.0, 0.0, 0.0),
                e1(0.0, 1.0, 0.0),
                e2(0.0, 0.0, 1.0);
    Triadic3D E;
    E.comp(0,0,0) = e0*(e0.cross(e0)); E.comp(0,0,1) = e0*(e0.cross(e1)); E.comp(0,0,2) = e0*(e0.cross(e2));
    E.comp(0,1,0) = e0*(e1.cross(e0)); E.comp(0,1,1) = e0*(e1.cross(e1)); E.comp(0,1,2) = e0*(e1.cross(e2));
    E.comp(0,2,0) = e0*(e2.cross(e0)); E.comp(0,2,1) = e0*(e2.cross(e1)); E.comp(0,2,2) = e0*(e2.cross(e2));
    E.comp(1,0,0) = e1*(e0.cross(e0)); E.comp(1,0,1) = e1*(e0.cross(e1)); E.comp(1,0,2) = e1*(e0.cross(e2));
    E.comp(1,1,0) = e1*(e1.cross(e0)); E.comp(1,1,1) = e1*(e1.cross(e1)); E.comp(1,1,2) = e1*(e1.cross(e2));
    E.comp(1,2,0) = e1*(e2.cross(e0)); E.comp(1,2,1) = e1*(e2.cross(e1)); E.comp(1,2,2) = e1*(e2.cross(e2));
    E.comp(2,0,0) = e2*(e0.cross(e0)); E.comp(2,0,1) = e2*(e0.cross(e1)); E.comp(2,0,2) = e2*(e0.cross(e2));
    E.comp(2,1,0) = e2*(e1.cross(e0)); E.comp(2,1,1) = e2*(e1.cross(e1)); E.comp(2,1,2) = e2*(e1.cross(e2));
    E.comp(2,2,0) = e2*(e2.cross(e0)); E.comp(2,2,1) = e2*(e2.cross(e1)); E.comp(2,2,2) = e2*(e2.cross(e2));
    return E;
}
//-------------------------------------------------------------------------------------------------
inline
Real min_eigen (const Dyadic3D& A, Vector3D& v) {
    std::vector<Real> eigval;
    std::vector<Vector3D> eigvec;
    eigen_solve(A, eigval, eigvec);
    v = eigvec[2];
    return eigval[2];
}
inline
Vector3D min_eigen (const Dyadic3D& A) {
    Vector3D v;
    Real l = min_eigen(A, v);
    v.comp *= l;
    return v;
}
inline
Real max_eigen (const Dyadic3D& A, Vector3D& v) {
    std::vector<Real> eigval;
    std::vector<Vector3D> eigvec;
    eigen_solve(A, eigval, eigvec);
    v = eigvec[0];
    return eigval[0];
}
inline
Vector3D max_eigen (const Dyadic3D& A) {
    Vector3D v;
    Real l = max_eigen(A, v);
    v.comp *= l;
    return v;
}
//-------------------------------------------------------------------------------------------------
inline
Vector3D operator* (const Real s, const Vector3D& v) {
    Vector3D R;
    R.comp = s*v(0), s*v(1), s*v(2);
    return R;
}
inline
Vector3D operator* (const Vector3D& v, const Real s) {
    Vector3D R;
    R.comp = s*v(0), s*v(1), s*v(2);
    return R;
}
inline
Dyadic3D operator* (const Real s, const Dyadic3D& m) {
    Dyadic3D R;
    R.comp = s*m(0,0), s*m(0,1), s*m(0,2),
             s*m(1,0), s*m(1,1), s*m(1,2),
             s*m(2,0), s*m(2,1), s*m(2,2);
    return R;
}
inline
Dyadic3D operator* (const Dyadic3D& m, const Real s) {
    Dyadic3D R;
    R.comp = s*m(0,0), s*m(0,1), s*m(0,2),
             s*m(1,0), s*m(1,1), s*m(1,2),
             s*m(2,0), s*m(2,1), s*m(2,2);
    return R;
}
inline
Point operator* (const Dyadic3D& m, const Point& v) {
    Point r;
    r(0) = m(0,0)*v(0)+m(0,1)*v(1)+m(0,2)*v(2);
    r(1) = m(1,0)*v(0)+m(1,1)*v(1)+m(1,2)*v(2);
    r(2) = m(2,0)*v(0)+m(2,1)*v(1)+m(2,2)*v(2);
    return r;
}
inline
Point operator* (const Point& v, const Dyadic3D& m) {
    Point r;
    r(0) = m(0,0)*v(0)+m(1,0)*v(1)+m(2,0)*v(2);
    r(1) = m(0,1)*v(0)+m(1,1)*v(1)+m(2,1)*v(2);
    r(2) = m(0,2)*v(0)+m(1,2)*v(1)+m(2,2)*v(2);
    return r;
}
inline
Point operator/ (const Point& v, const Dyadic3D& m) {
    const Real d = determinant(m);
    libmesh_assert( d > 1.0e-12 );
    Point r;
    r(0) =  (m(1,0)*m(2,1)*v(2)-m(1,0)*m(2,2)*v(1)-m(1,1)*m(2,0)*v(2)+m(1,1)*m(2,2)*v(0)+m(1,2)*m(2,0)*v(1)-m(1,2)*m(2,1)*v(0))/d;
    r(1) = -(m(0,0)*m(2,1)*v(2)-m(0,0)*m(2,2)*v(1)-m(0,1)*m(2,0)*v(2)+m(0,1)*m(2,2)*v(0)+m(0,2)*m(2,0)*v(1)-m(0,2)*m(2,1)*v(0))/d;
    r(2) =  (m(0,0)*m(1,1)*v(2)-m(0,0)*m(1,2)*v(1)-m(0,1)*m(1,0)*v(2)+m(0,1)*m(1,2)*v(0)+m(0,2)*m(1,0)*v(1)-m(0,2)*m(1,1)*v(0))/d;
    return r;
}
inline
Vector3D operator* (const Dyadic3D& m, const Vector3D& v) {
    Vector3D r;
    r.comp(0) = m(0,0)*v(0)+m(0,1)*v(1)+m(0,2)*v(2);
    r.comp(1) = m(1,0)*v(0)+m(1,1)*v(1)+m(1,2)*v(2);
    r.comp(2) = m(2,0)*v(0)+m(2,1)*v(1)+m(2,2)*v(2);
    return r;
}
inline
Vector3D operator* (const Vector3D& v, const Dyadic3D& m) {
    Vector3D r;
    r.comp(0) = m(0,0)*v(0)+m(1,0)*v(1)+m(2,0)*v(2);
    r.comp(1) = m(0,1)*v(0)+m(1,1)*v(1)+m(2,1)*v(2);
    r.comp(2) = m(0,2)*v(0)+m(1,2)*v(1)+m(2,2)*v(2);
    return r;
}
inline
Vector3D operator/ (const Vector3D& v, const Dyadic3D& m) {
    const Real d = determinant(m);
    libmesh_assert( d > 1.0e-12 );
    Vector3D r;
    r.comp(0) =  (m(1,0)*m(2,1)*v(2)-m(1,0)*m(2,2)*v(1)-m(1,1)*m(2,0)*v(2)+m(1,1)*m(2,2)*v(0)+m(1,2)*m(2,0)*v(1)-m(1,2)*m(2,1)*v(0))/d;
    r.comp(1) = -(m(0,0)*m(2,1)*v(2)-m(0,0)*m(2,2)*v(1)-m(0,1)*m(2,0)*v(2)+m(0,1)*m(2,2)*v(0)+m(0,2)*m(2,0)*v(1)-m(0,2)*m(2,1)*v(0))/d;
    r.comp(2) =  (m(0,0)*m(1,1)*v(2)-m(0,0)*m(1,2)*v(1)-m(0,1)*m(1,0)*v(2)+m(0,1)*m(1,2)*v(0)+m(0,2)*m(1,0)*v(1)-m(0,2)*m(1,1)*v(0))/d;
    return r;
}
//-------------------------------------------------------------------------------------------------
inline
Dyadic3D rotation (const Real& theta_x, const Real& theta_y, const Real& theta_z) {
    const Real S_x = sin(theta_x), C_x = cos(theta_x),
               S_y = sin(theta_y), C_y = cos(theta_y),
               S_z = sin(theta_z), C_z = cos(theta_z);
    Dyadic3D w;
    w.comp(0,0) = C_y*C_z;
    w.comp(0,1) = C_z*S_x*S_y-C_x*S_z;
    w.comp(0,2) = S_x*S_z+C_x*C_z*S_y;
    w.comp(1,0) = C_y*S_z;
    w.comp(1,1) = C_x*C_z+S_x*S_y*S_z;
    w.comp(1,2) = C_x*S_y*S_z-C_z*S_x;
    w.comp(2,0) = -S_y;
    w.comp(2,1) = C_y*S_x;
    w.comp(2,2) = C_x*C_y;
    return w;
}
//-------------------------------------------------------------------------------------------------
inline
void inverse (const Blitz_DyadicReal& src, Blitz_DyadicReal& res, const Real epsilon) {
    const int N =src.extent(blitz::firstDim);
    if ( src.extent(blitz::secondDim) != N ) libmesh_error();
    if ( &src == &res ) libmesh_error();
    Real* Mout =(Real*) malloc(sizeof(Real)*N*N);
    const int actualsize =N;
    int i, j, k;
    Real sum, x;
    res.resize(N, N);
    // copy from source
    k = 0;
    for (i=0; i<N; i++) {
        for (j=0; j<N; j++) {
            Mout[k] = src(i,j);
            ++k;
        }
    }
    // Add small value to diagonal if diagonal is zero
    for (i=0; i<actualsize; i++) {
        j = i*actualsize+i;
        if (Mout[j]<epsilon && Mout[j]>-epsilon) Mout[j] = epsilon;
    }
    // normalize row with respect to [0]
    for (i=1; i<actualsize; i++) {
        Mout[i] /= Mout[0];
    }
    for (i=1; i<actualsize; i++) {
        // do a column of L
        for (j=i; j<actualsize; j++) {
            sum = 0.0;
            for (k=0; k<i; k++) {
                sum += Mout[j*actualsize+k] * Mout[k*actualsize+i];
            }
            Mout[j*actualsize+i] -= sum;
        }
        if (i == actualsize-1) continue;
        // do a row of U
        for (j=i+1; j<actualsize; j++) {
            sum = 0.0;
            for (k=0; k < i; k++) {
                sum += Mout[i*actualsize+k] * Mout[k*actualsize+j];
            }
            Mout[i*actualsize+j] = (Mout[i*actualsize+j] - sum) / Mout[i*actualsize+i];
        }
    }
    // invert L
    for (i=0; i<actualsize; i++) {
        for (j=i; j<actualsize; j++) {
            x = 1.0;
            if ( i != j ) {
                x = 0.0;
                for ( k = i; k < j; k++ ) {
                    x -= Mout[j*actualsize+k] * Mout[k*actualsize+i];
                }
            }
            Mout[j*actualsize+i] = x / Mout[j*actualsize+j];
        }
    }
    // invert U
    for (i=0; i<actualsize; i++) {
        for (j=i; j<actualsize; j++) {
            if ( i == j ) continue;
            sum = 0.0;
            for (k=i; k<j; k++ ) {
                sum += Mout[k*actualsize+j] * (i==k ? 1.0 : Mout[i*actualsize+k]);
            }
            Mout[i*actualsize+j] = -sum;
        }
    }
    // final inversion
    for (i=0; i<actualsize; i++) {
        for (j=0; j<actualsize; j++) {
            sum = 0.0;
            for (k=((i>j)?i:j); k<actualsize; k++) {
                sum += (j==k ? 1.0 : Mout[j*actualsize+k]) * Mout[k*actualsize+i];
            }
            Mout[j*actualsize+i] = sum;
        }
    }
    // copy result
    k = 0;
    for (i=0; i<N; i++) {
        for (j=0; j<N; j++) {
            res(i,j) = Mout[k];
            ++k;
        }
    }
    // free memory
    free(Mout);
}
//-------------------------------------------------------------------------------------------------
template<int N>
inline
void inverse (const Dyadic<Real, N>& A, Dyadic<Real, N>& A_inv, const Real epsilon) {
    libmesh_assert( &A != &A_inv );
    const int actualsize = N;
    int i, j, k;
    Real sum, x;
    k = 0;
    Real* Mout =(Real*) malloc(sizeof(Real)*N*N); // copy from source
    for (i=0; i<N; i++) {
        for (j=0; j<N; j++) {
            Mout[k] = A.comp(i, j);
            ++k;
        }
    }
    for (i=0; i<actualsize; i++) { // add small value to diagonal if diagonal is zero
        j = i * actualsize + i;
        if ( Mout[j] < epsilon && Mout[j] > -epsilon ) Mout[j] = epsilon;
    }
    for (i=1; i<actualsize; i++) { // normalize row with respect to [0]
        Mout[i] /= Mout[0];
    }
    for (i=1; i<actualsize; i++) {
        for (j=i; j<actualsize; j++) { // do a column of L
            sum = 0.0;
            for (k=0; k<i; k++) {
                sum += Mout[j*actualsize+k] * Mout[k*actualsize+i];
            }
            Mout[j*actualsize+i] -= sum;
        }
        if (i == actualsize-1) continue;
        for (j=i+1; j<actualsize; j++) { // do a row of U
            sum = 0.0;
            for (k=0; k<i; k++) {
                sum += Mout[i*actualsize+k] * Mout[k*actualsize+j];
            }
            Mout[i*actualsize+j] = (Mout[i*actualsize+j] - sum) / Mout[i*actualsize+i];
        }
    }
    for (i=0; i<actualsize; i++) { // invert L
        for (j=i; j<actualsize; j++) {
            x = 1.0;
            if ( i != j ) {
                x = 0.0;
                for (k=i; k<j; k++ ) {
                    x -= Mout[j*actualsize+k] * Mout[k*actualsize+i];
                }
            }
            Mout[j*actualsize+i] = x / Mout[j*actualsize+j];
        }
    }
    for (i=0; i<actualsize; i++) { // invert U
        for (j=i; j<actualsize; j++) {
            if ( i == j ) continue;
            sum = 0.0;
            for (k=i; k<j; k++ ) {
                sum += Mout[k*actualsize+j] * (i==k ? 1.0 : Mout[i*actualsize+k]);
            }
            Mout[i*actualsize+j] = -sum;
        }
    }
    for (i=0; i<actualsize; i++) { // final inversion
        for (j=0; j<actualsize; j++) {
            sum = 0.0;
            for (k=((i>j)?i:j); k<actualsize; k++) {
                sum += (j==k ? 1.0 : Mout[j*actualsize+k]) * Mout[k*actualsize+i];
            }
            Mout[j*actualsize+i] = sum;
        }
    }
    k = 0; // copy result
    for (i=0; i<N; i++) {
        for (j=0; j<N; j++) {
            A_inv.comp(i, j) = Mout[k];
            ++k;
        }
    }
    free(Mout); // free memory
}
//-------------------------------------------------------------------------------------------------
template<int N>
inline
void eigen_solve (const Dyadic<Real, N>& A, std::vector<Real>& l, std::vector< Vector<Real, N> >& v) {
    gsl_matrix*  mat = gsl_matrix_alloc(N, N);
    for (int i=0; i<N; i++) {
        for (int j=0; j<N; j++) {
            gsl_matrix_set(mat, i, j, A(i, j));
        }
    }
    for (int i=0; i<N; i++) {
        for (int j=0; j<N; j++) {
            if (j == i) continue;
            const Real a = gsl_matrix_get(mat, i, j),
                       b = gsl_matrix_get(mat, j, i);
            if (fabs(a-b)>1.0e-6) libmesh_error();
        }
    }
    gsl_vector*  eval = gsl_vector_alloc(N);
    gsl_matrix*  evec = gsl_matrix_alloc(N, N);
    gsl_eigen_symmv_workspace* wrk = gsl_eigen_symmv_alloc(N);
    gsl_eigen_symmv(mat, eval, evec, wrk);
    gsl_eigen_symmv_free(wrk);
    gsl_matrix_free(mat);
    gsl_eigen_symmv_sort(eval, evec, GSL_EIGEN_SORT_VAL_DESC);
    l.resize(N);
    v.resize(N);
    for (int i=0; i<N; i++) {
        l[i] = gsl_vector_get(eval, i);
        for (int j=0; j<N; j++)
            v[i].comp(j) = gsl_matrix_get(evec, j, i);
    }
    gsl_vector_free(eval);
    gsl_matrix_free(evec);
}
//=================================================================================================
const Point     zero_coordinates = Point(0.0, 0.0, 0.0);
const Vector3D  zero_vector = Vector3D();
const Dyadic3D  zero_matrix = Dyadic3D();
const Dyadic3D  identity_matrix = identity_3d();
const Quindic3D identity_tensor = symm_tensor(identity_matrix);
const Triadic3D permutation_tensor = permutation();
//=================================================================================================
#endif // __UTILS__ALGEBRA_MORE_H__
//=================================================================================================
