/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __UTILS_INVARIANTS_H__
#define __UTILS_INVARIANTS_H__
//=================================================================================================
class InvariantsBasicInfo {
public:
    InvariantsBasicInfo (const Dyadic3D& C_in, const Vector3D* N);
    ~InvariantsBasicInfo () {}
public:
    // invariants and pseudo-invariants of tensor 'C' with 'N[3]' (direction unit) vectors
    Real I1, I2, I3, I4, I6, I8,
         J1, J2, J3, J4, J6, J8;
    // derivatives of the invariants and pseudo-invariants
    Dyadic3D dI1_dC, dI2_dC, dI3_dC, dI4_dC, dI6_dC, dI8_dC,
             dJ1_dC, dJ2_dC, dJ3_dC, dJ4_dC, dJ6_dC, dJ8_dC;
    // right_Cauchy-Green deformation tensor 'C' and its inverse
    Dyadic3D C, Ci;
    // tensor product of fibre orientation vectors
    Dyadic3D N_N[SP3D];
};
//-------------------------------------------------------------------------------------------------
class InvariantsInfo :
public InvariantsBasicInfo {
public:
    InvariantsInfo (const Dyadic3D& C_in, const Vector3D* N);
    ~InvariantsInfo () {}
public:
    // pseudo-invariants of tensor 'C' with 'N[3]' (direction unit) vectors
    Real I5, I7, I9,
         J5, J7, J9;
    // derivatives of the invariants and pseudo-invariants
    Dyadic3D dI5_dC, dI7_dC, dI9_dC,
             dJ5_dC, dJ7_dC, dJ9_dC;
    // derivatives of the invariants and pseudo-invariants
    Quindic3D d2I1_dCdC, d2I2_dCdC, d2I3_dCdC, d2I4_dCdC, d2I6_dCdC, d2I8_dCdC,
              d2J1_dCdC, d2J2_dCdC, d2J3_dCdC, d2J4_dCdC, d2J6_dCdC, d2J8_dCdC;
    Quindic3D d2I5_dCdC, d2I7_dCdC, d2I9_dCdC,
              d2J5_dCdC, d2J7_dCdC, d2J9_dCdC;
    // 'C' squared and tensor derivative of inverse with respect to 'C'
    Dyadic3D C2;
    Quindic3D Ci_s;
    // tensor products of fibre orientation vectors wrt matrix 'C' and its inverse
    Vector3D CN[SP3D], NC[SP3D];
    Vector3D CiN[SP3D], NCi[SP3D];
    Dyadic3D N_CN[SP3D], NC_N[SP3D];
    Dyadic3D N_CiN[SP3D], NCi_N[SP3D];
};
//-------------------------------------------------------------------------------------------------
class InvariantsEulerInfo {
public:
    InvariantsEulerInfo (const Dyadic3D& b_in, const Vector3D* n);
    ~InvariantsEulerInfo () {}
public:
    // invariants, pseudo-invariants of tensor 'b' (Finger tensor)
    Real I1, I2, J1, J2, J;
    // ...and their derivatives
    Dyadic3D     FF__dI1dC,      FF__dI2dC,      FF__dJ1dC,      FF__dJ2dC,      FF__dJdC;
    Quindic3D FFFF__d2I1dCdC, FFFF__d2I2dCdC, FFFF__d2J1dCdC, FFFF__d2J2dCdC, FFFF__d2JdCdC;
    // left_Cauchy-Green deformation tensor 'b' and its squared form
    Dyadic3D b, b2;
};
//=================================================================================================
//
//
//
//=================================================================================================
inline
InvariantsBasicInfo::InvariantsBasicInfo (const Dyadic3D& C_in, const Vector3D* N) {
    libmesh_assert(N != 0);
    //
    const Dyadic3D& I = identity_matrix;
    //
    C = C_in;
    Ci = inverse(C);
    //
    N_N[0] = tensor(N[0]);
    N_N[1] = tensor(N[1]);
    N_N[2] = tensor(N[2]);
    //
    blitz::firstIndex  i;
    blitz::secondIndex j;
    //
    I1 = first_invariant(C);
    I2 = second_invariant(C);
    I3 = third_invariant(C);
    I4 = product(N[0], C, N[0]);
    I6 = product(N[1], C, N[1]);
    I8 = product(N[2], C, N[2]);
    //
    dI1_dC.comp = I.comp(i,j);
    dI2_dC.comp = I1*I.comp(i,j) - C.comp(i,j);
    dI3_dC.comp = I3*Ci.comp(i,j);
    dI4_dC.comp = N_N[0].comp(i,j);
    dI6_dC.comp = N_N[1].comp(i,j);
    dI8_dC.comp = N_N[2].comp(i,j);
    //
    const Real recip_I3p1o2 = 1.0/pow(I3, frac_1o2),
               recip_I3p1o3 = 1.0/pow(I3, frac_1o3),
               recip_I3p2o3 = 1.0/pow(I3, frac_2o3),
               recip_I3p4o3 = 1.0/pow(I3, frac_4o3),
               recip_I3p5o3 = 1.0/pow(I3, frac_5o3);
    //
    J1 = I1*recip_I3p1o3;
    J2 = I2*recip_I3p2o3;
    J3 = 1./recip_I3p1o2;
    J4 = I4*recip_I3p1o3;
    J6 = I6*recip_I3p1o3;
    J8 = I8*recip_I3p1o3;
    //
    dJ1_dC.comp = recip_I3p1o3*dI1_dC.comp(i,j) - (frac_1o3*I1*recip_I3p4o3)*dI3_dC.comp(i,j);
    dJ2_dC.comp = recip_I3p2o3*dI2_dC.comp(i,j) - (frac_2o3*I2*recip_I3p5o3)*dI3_dC.comp(i,j);
    dJ3_dC.comp = (frac_1o2*recip_I3p1o2)*dI3_dC.comp(i,j);
    dJ4_dC.comp = recip_I3p1o3*dI4_dC.comp(i,j) - (frac_1o3*I4*recip_I3p4o3)*dI3_dC.comp(i,j);
    dJ6_dC.comp = recip_I3p1o3*dI6_dC.comp(i,j) - (frac_1o3*I6*recip_I3p4o3)*dI3_dC.comp(i,j);
    dJ8_dC.comp = recip_I3p1o3*dI8_dC.comp(i,j) - (frac_1o3*I8*recip_I3p4o3)*dI3_dC.comp(i,j);
}
//-------------------------------------------------------------------------------------------------
inline
InvariantsInfo::InvariantsInfo (const Dyadic3D& C_in, const Vector3D* N) :
InvariantsBasicInfo(C_in, N) {
    libmesh_assert(N != 0);
    //
    const Dyadic3D& I = identity_matrix;
    const Quindic3D& I_s = identity_tensor;
    //
    C2 = squared(C);
    Ci_s = op_symm_tensor(Ci);
    //
    CN[0] = mul_im_m(C, N[0]); CiN[0] = mul_im_m(Ci, N[0]);
    CN[1] = mul_im_m(C, N[1]); CiN[1] = mul_im_m(Ci, N[1]);
    CN[2] = mul_im_m(C, N[2]); CiN[2] = mul_im_m(Ci, N[2]);
    NC[0] = mul_m_mi(N[0], C); NCi[0] = mul_m_mi(N[0], Ci);
    NC[1] = mul_m_mi(N[1], C); NCi[1] = mul_m_mi(N[1], Ci);
    NC[2] = mul_m_mi(N[2], C); NCi[2] = mul_m_mi(N[2], Ci);
    N_CN[0] = tensor(N[0], CN[0]); N_CiN[0] = tensor(N[0], CiN[0]);
    N_CN[1] = tensor(N[1], CN[1]); N_CiN[1] = tensor(N[1], CiN[1]);
    N_CN[2] = tensor(N[2], CN[2]); N_CiN[2] = tensor(N[2], CiN[2]);
    NC_N[0] = tensor(NC[0], N[0]); NCi_N[0] = tensor(NCi[0], N[0]);
    NC_N[1] = tensor(NC[1], N[1]); NCi_N[1] = tensor(NCi[1], N[1]);
    NC_N[2] = tensor(NC[2], N[2]); NCi_N[2] = tensor(NCi[2], N[2]);
    //
    blitz::firstIndex  i;
    blitz::secondIndex j;
    blitz::thirdIndex  k;
    blitz::fourthIndex l;
    //
    I5 = product(N[0], C2, N[0]);
    I7 = product(N[1], C2, N[1]);
    I9 = product(N[2], C2, N[2]);
    //
    dI5_dC.comp = N_CN[0].comp(i,j) + NC_N[0].comp(i,j);
    dI7_dC.comp = N_CN[1].comp(i,j) + NC_N[1].comp(i,j);
    dI9_dC.comp = N_CN[2].comp(i,j) + NC_N[2].comp(i,j);
    //
    d2I1_dCdC.comp = 0.0;
    d2I2_dCdC.comp = I.comp(i,j)*I.comp(k,l) - I_s.comp(i,j,k,l);
    d2I3_dCdC.comp = I3*Ci.comp(i,j)*Ci.comp(k,l) + I3*Ci_s.comp(i,j,k,l);
    d2I4_dCdC.comp = 0.0;
    d2I6_dCdC.comp = 0.0;
    d2I8_dCdC.comp = 0.0;
    d2I5_dCdC.comp = 0.5*(N[0].comp(i)*I.comp(j,k)*N[0].comp(l)+N[0].comp(i)*I.comp(j,l)*N[0].comp(k)
                         +N[0].comp(j)*I.comp(i,k)*N[0].comp(l)+N[0].comp(j)*I.comp(i,l)*N[0].comp(k));
    d2I7_dCdC.comp = 0.5*(N[1].comp(i)*I.comp(j,k)*N[1].comp(l)+N[1].comp(i)*I.comp(j,l)*N[1].comp(k)
                         +N[1].comp(j)*I.comp(i,k)*N[1].comp(l)+N[1].comp(j)*I.comp(i,l)*N[1].comp(k));
    d2I9_dCdC.comp = 0.5*(N[2].comp(i)*I.comp(j,k)*N[2].comp(l)+N[2].comp(i)*I.comp(j,l)*N[2].comp(k)
                         +N[2].comp(j)*I.comp(i,k)*N[2].comp(l)+N[2].comp(j)*I.comp(i,l)*N[2].comp(k));
    //
    const Real recip_I3p3o2 = 1.0/pow(I3, frac_3o2),
               recip_I3p1o2 = 1.0/pow(I3, frac_1o2),
               recip_I3p1o3 = 1.0/pow(I3, frac_1o3),
               recip_I3p2o3 = 1.0/pow(I3, frac_2o3),
               recip_I3p4o3 = 1.0/pow(I3, frac_4o3),
               recip_I3p5o3 = 1.0/pow(I3, frac_5o3),
               recip_I3p7o3 = 1.0/pow(I3, frac_7o3),
               recip_I3p8o3 = 1.0/pow(I3, frac_8o3);
    //
    J5 = I5*recip_I3p2o3;
    J7 = I7*recip_I3p2o3;
    J9 = I9*recip_I3p2o3;
    //
    dJ5_dC.comp = recip_I3p2o3*dI5_dC.comp(i,j) - (frac_2o3*I5*recip_I3p5o3)*dI3_dC.comp(i,j);
    dJ7_dC.comp = recip_I3p2o3*dI7_dC.comp(i,j) - (frac_2o3*I7*recip_I3p5o3)*dI3_dC.comp(i,j);
    dJ9_dC.comp = recip_I3p2o3*dI9_dC.comp(i,j) - (frac_2o3*I9*recip_I3p5o3)*dI3_dC.comp(i,j);
    //
    d2J1_dCdC.comp = recip_I3p1o3*d2I1_dCdC.comp(i,j,k,l)
                   - (frac_1o3*recip_I3p4o3)*(dI1_dC.comp(i,j)*dI3_dC.comp(k,l)+dI3_dC.comp(i,j)*dI1_dC.comp(k,l))
                   - (frac_1o3*I1*recip_I3p4o3)*d2I3_dCdC.comp(i,j,k,l)
                   + (frac_4o9*I1*recip_I3p7o3)*dI3_dC.comp(i,j)*dI3_dC.comp(k,l);
    d2J2_dCdC.comp = recip_I3p2o3*d2I2_dCdC.comp(i,j,k,l)
                   - (frac_2o3*recip_I3p5o3)*(dI2_dC.comp(i,j)*dI3_dC.comp(k,l)+dI3_dC.comp(i,j)*dI2_dC.comp(k,l))
                   - (frac_2o3*I2*recip_I3p5o3)*d2I3_dCdC.comp(i,j,k,l)
                   + (frac_10o9*I2*recip_I3p8o3)*dI3_dC.comp(i,j)*dI3_dC.comp(k,l);
    d2J3_dCdC.comp = (frac_1o2*recip_I3p1o2)*d2I3_dCdC.comp(i,j,k,l)
                   - (frac_1o4*recip_I3p3o2)*dI3_dC.comp(i,j)*dI3_dC.comp(k,l);
    d2J4_dCdC.comp = recip_I3p1o3*d2I4_dCdC.comp(i,j,k,l)
                   - (frac_1o3*recip_I3p4o3)*(dI4_dC.comp(i,j)*dI3_dC.comp(k,l)+dI3_dC.comp(i,j)*dI4_dC.comp(k,l))
                   - (frac_1o3*I4*recip_I3p4o3)*d2I3_dCdC.comp(i,j,k,l)
                   + (frac_4o9*I4*recip_I3p7o3)*dI3_dC.comp(i,j)*dI3_dC.comp(k,l);
    d2J6_dCdC.comp = recip_I3p1o3*d2I6_dCdC.comp(i,j,k,l)
                   - (frac_1o3*recip_I3p4o3)*(dI6_dC.comp(i,j)*dI3_dC.comp(k,l)+dI3_dC.comp(i,j)*dI6_dC.comp(k,l))
                   - (frac_1o3*I6*recip_I3p4o3)*d2I3_dCdC.comp(i,j,k,l)
                   + (frac_4o9*I6*recip_I3p7o3)*dI3_dC.comp(i,j)*dI3_dC.comp(k,l);
    d2J8_dCdC.comp = recip_I3p1o3*d2I8_dCdC.comp(i,j,k,l)
                   - (frac_1o3*recip_I3p4o3)*(dI8_dC.comp(i,j)*dI3_dC.comp(k,l)+dI3_dC.comp(i,j)*dI8_dC.comp(k,l))
                   - (frac_1o3*I8*recip_I3p4o3)*d2I3_dCdC.comp(i,j,k,l)
                   + (frac_4o9*I8*recip_I3p7o3)*dI3_dC.comp(i,j)*dI3_dC.comp(k,l);
    d2J5_dCdC.comp = recip_I3p2o3*d2I5_dCdC.comp(i,j,k,l)
                   - (frac_2o3*recip_I3p5o3)*(dI5_dC.comp(i,j)*dI3_dC.comp(k,l)+dI3_dC.comp(i,j)*dI5_dC.comp(k,l))
                   - (frac_2o3*I5*recip_I3p5o3)*d2I3_dCdC.comp(i,j,k,l)
                   + (frac_10o9*I5*recip_I3p8o3)*dI3_dC.comp(i,j)*dI3_dC.comp(k,l);
    d2J7_dCdC.comp = recip_I3p2o3*d2I7_dCdC.comp(i,j,k,l)
                   - (frac_2o3*recip_I3p5o3)*(dI7_dC.comp(i,j)*dI3_dC.comp(k,l)+dI3_dC.comp(i,j)*dI7_dC.comp(k,l))
                   - (frac_2o3*I7*recip_I3p5o3)*d2I3_dCdC.comp(i,j,k,l)
                   + (frac_10o9*I7*recip_I3p8o3)*dI3_dC.comp(i,j)*dI3_dC.comp(k,l);
    d2J9_dCdC.comp = recip_I3p2o3*d2I9_dCdC.comp(i,j,k,l)
                   - (frac_2o3*recip_I3p5o3)*(dI9_dC.comp(i,j)*dI3_dC.comp(k,l)+dI3_dC.comp(i,j)*dI9_dC.comp(k,l))
                   - (frac_2o3*I9*recip_I3p5o3)*d2I3_dCdC.comp(i,j,k,l)
                   + (frac_10o9*I9*recip_I3p8o3)*dI3_dC.comp(i,j)*dI3_dC.comp(k,l);
}
//-------------------------------------------------------------------------------------------------
inline
InvariantsEulerInfo::InvariantsEulerInfo (const Dyadic3D& b_in, const Vector3D* n) {
    libmesh_assert(n);
    const Dyadic3D& I = identity_matrix;
    const Quindic3D& I_s = identity_tensor;
    //
    b = b_in;
    b2 = squared(b);
    const Quindic3D B_s = symm_tensor(b);
    //
    blitz::firstIndex  i;
    blitz::secondIndex j;
    blitz::thirdIndex  k;
    blitz::fourthIndex l;
    //
    I1 = trace(b);
    I2 = inner_product(b, b);
    J  = sqrt(determinant(b));
    const Real Jp2o3 = pow(J, frac_2o3),
               Jp4o3 = pow(J, frac_4o3);
    J1 = I1/Jp2o3;
    J2 = I2/Jp4o3;
    //
    FF__dJdC.comp = (0.5*J)*I.comp(i,j);
    FF__dI1dC.comp = b.comp(i,j);
    FF__dI2dC.comp = 2.0*b2.comp(i,j);
    FF__dJ1dC.comp = (1.0/Jp2o3)*b.comp(i,j)-(frac_1o3*J1)*I.comp(i,j);
    FF__dJ2dC.comp = (2.0/Jp4o3)*b2.comp(i,j)-(frac_2o3*J2)*I.comp(i,j);
    //
    FFFF__d2JdCdC.comp = (frac_1o4*J)*I.comp(i,j)*I.comp(k,l)-(frac_1o2*J)*I_s.comp(i,j,k,l);
    FFFF__d2I1dCdC.comp = 0.0;
    FFFF__d2I2dCdC.comp = 2.0*B_s.comp(i,j,k,l);
    FFFF__d2J1dCdC.comp = (frac_1o3*J1)*I_s.comp(i,j,k,l)
                        - frac_1o3*I.comp(i,j)*FF__dJ1dC.comp(k,l) - (frac_1o3/Jp2o3)*b.comp(i,j)*I.comp(k,l);
    FFFF__d2J2dCdC.comp = (frac_2o3*J2)*I_s.comp(i,j,k,l) - frac_2o3*I.comp(i,j)*FF__dJ2dC.comp(k,l)
                        + (2.0/Jp4o3)*B_s.comp(i,j,k,l) - (frac_4o3/Jp4o3)*b.comp(i,j)*I.comp(k,l);
}
//=================================================================================================
#endif // __UTILS_INVARIANTS_H__
//=================================================================================================
