/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __UTILS_REGION_INFO_H__
#define __UTILS_REGION_INFO_H__
//=================================================================================================
class RegionInfo {
public:
    RegionInfo ();
    ~RegionInfo () {}
    //
    void set_biomech_2d  (void (*f_ptr)(const MaterialInfo&, const FieldInfo&, Dyadic2D&, Quindic2D&));
    void set_biomech_3d  (void (*f_ptr)(const MaterialInfo&, const FieldInfo&, Dyadic3D&, Quindic3D&, Dyadic3D&, Real&, Dyadic3D&, Real&));
    void set_biochem     (void (*f_ptr)(const MaterialInfo&, FieldInfo&, Blitz_VectorReal&, Blitz_VectorGradient&, Blitz_DyadicReal&, Blitz_VectorPoint&, Blitz_DyadicReal&, Blitz_DyadicGradient&));
    void set_damage      (void (*f_ptr)(const MaterialInfo&, FieldInfo&));
    void set_non_damage  (void (*f_ptr)(const MaterialInfo&, FieldInfo&));
    void set_coupled_2d  (void (*f_ptr)(const MaterialInfo&, const FieldInfo&, Dyadic2D&, Quindic2D&));
    void set_coupled_3d  (void (*f_ptr)(const MaterialInfo&, const FieldInfo&, Dyadic3D&, Quindic3D&, Dyadic3D&, Real&, Dyadic3D&, Real&));
    void set_multiscale  (void (*f_ptr)(const MaterialInfo&, const FieldInfo&, EquationSystems&, std::vector<FieldInfo>&, Dyadic3D&, Vector3D&));
    void set_plastic_deform (void (*f_ptr)(const MaterialInfo&, const FieldInfo&));
    void calc_biomech_2d  (const FieldInfo& fld, Dyadic2D& S, Quindic2D& D) const;
    void calc_biomech_3d  (const FieldInfo& fld, Dyadic3D& S, Quindic3D& Duu, Dyadic3D& Dup, Real& P, Dyadic3D& Dpu, Real& Dpp) const;
    void calc_biochem     (FieldInfo& fld, Blitz_VectorReal& F, Blitz_VectorGradient& G, Blitz_DyadicReal& K, Blitz_VectorPoint& V, Blitz_DyadicReal& D, Blitz_DyadicGradient& X) const;
    void calc_damage      (FieldInfo& fld) const;
    void calc_coupled_2d  (const FieldInfo& fld, Dyadic2D& S, Quindic2D& D) const;
    void calc_coupled_3d  (const FieldInfo& fld, Dyadic3D& S, Quindic3D& Duu, Dyadic3D& Dup, Real& P, Dyadic3D& Dpu, Real& Dpp) const;
    void calc_multiscale  (const FieldInfo& fld, EquationSystems& es_m, std::vector<FieldInfo>& fld_m, Dyadic3D& S, Vector3D& Q) const;
    void calc_plastic_deform (const FieldInfo& fld) const;
private:
    void (*_biomech_2d)  (const MaterialInfo&, const FieldInfo&, Dyadic2D&, Quindic2D&);
    void (*_biomech_3d)  (const MaterialInfo&, const FieldInfo&, Dyadic3D&, Quindic3D&, Dyadic3D&, Real&, Dyadic3D&, Real&);
    void (*_biochem)     (const MaterialInfo&, FieldInfo&, Blitz_VectorReal&, Blitz_VectorGradient&, Blitz_DyadicReal&, Blitz_VectorPoint&, Blitz_DyadicReal&, Blitz_DyadicGradient&);
    void (*_damage)      (const MaterialInfo&, FieldInfo&);
    void (*_non_damage)  (const MaterialInfo&, FieldInfo&);
    void (*_coupled_2d)  (const MaterialInfo&, const FieldInfo&, Dyadic2D&, Quindic2D&);
    void (*_coupled_3d)  (const MaterialInfo&, const FieldInfo&, Dyadic3D&, Quindic3D&, Dyadic3D&, Real&, Dyadic3D&, Real&);
    void (*_multiscale)  (const MaterialInfo&, const FieldInfo&, EquationSystems&, std::vector<FieldInfo>&, Dyadic3D&, Vector3D&);
    void (*_plastic_deform) (const MaterialInfo&, const FieldInfo&);
public:
    // flags to designate if region is membrane (2D) or is in "damaged"/"non-healthy" state
    bool is_membrane, is_damaged;
    // pointer to a material information class object (containing model parameters)
    MaterialInfo* mat;
    // list of nodes of the region (global and local ones)
    std::set<unsigned int> node, node_local;
    // damage factor (constant value)
    Real damage_factor_const;
};
//=================================================================================================
//
//
//
//=================================================================================================
// damage factor margins (constant parameters)
const Real material__min_damage_factor = 0.0000,
           material__max_damage_factor = 0.9999;
//=================================================================================================
inline
void damage_model__nothing (const MaterialInfo& mat, FieldInfo& fld) {
    ; // do nothing here...
}
inline
void damage_model__max (const MaterialInfo& mat, FieldInfo& fld) {
    fld.biomech[0] = material__max_damage_factor;
}
inline
void damage_model__min (const MaterialInfo& mat, FieldInfo& fld) {
    fld.biomech[0] = material__min_damage_factor;
}
//-------------------------------------------------------------------------------------------------
inline
void zero_permanent_strain (const MaterialInfo& mat, const FieldInfo& fld) {
    // deformation gradient: inelastic (plastic)
    fld.deform_grad_inel = identity_matrix;
    // deformation gradient: elastic
    fld.deform_grad_elas = fld.deform_gradient;
}
//=================================================================================================
inline
RegionInfo::RegionInfo () : is_membrane(false), is_damaged(false) {
    this->damage_factor_const = material__min_damage_factor;
    this->_damage = &damage_model__max;
    this->_non_damage = &damage_model__min;
    // default pointer to function initialisation
    this->set_plastic_deform(zero_permanent_strain);
}
//-------------------------------------------------------------------------------------------------
inline
void RegionInfo::set_biomech_2d (void (*f_ptr)(const MaterialInfo&, const FieldInfo&, Dyadic2D&, Quindic2D&)) {
    libmesh_assert( this->is_membrane );
    this->_biomech_2d = f_ptr;
}
inline
void RegionInfo::set_biomech_3d (void (*f_ptr)(const MaterialInfo&, const FieldInfo&, Dyadic3D&, Quindic3D&, Dyadic3D&, Real&, Dyadic3D&, Real&)) {
    libmesh_assert( !this->is_membrane );
    this->_biomech_3d = f_ptr;
}
inline
void RegionInfo::set_biochem (void (*f_ptr)(const MaterialInfo&, FieldInfo&, Blitz_VectorReal&, Blitz_VectorGradient&, Blitz_DyadicReal&, Blitz_VectorPoint&, Blitz_DyadicReal&, Blitz_DyadicGradient&)) {
    this->_biochem = f_ptr;
}
inline
void RegionInfo::set_damage (void (*f_ptr)(const MaterialInfo&, FieldInfo&)) {
    this->_damage = f_ptr;
}
inline
void RegionInfo::set_non_damage (void (*f_ptr)(const MaterialInfo&, FieldInfo&)) {
    this->_non_damage = f_ptr;
}
inline
void RegionInfo::set_coupled_2d (void (*f_ptr)(const MaterialInfo&, const FieldInfo&, Dyadic2D&, Quindic2D&)) {
    libmesh_assert( this->is_membrane );
    this->_coupled_2d = f_ptr;
}
inline
void RegionInfo::set_coupled_3d (void (*f_ptr)(const MaterialInfo&, const FieldInfo&, Dyadic3D&, Quindic3D&, Dyadic3D&, Real&, Dyadic3D&, Real&)) {
    libmesh_assert( !this->is_membrane );
    this->_coupled_3d = f_ptr;
}
inline
void RegionInfo::set_multiscale (void (*f_ptr)(const MaterialInfo&, const FieldInfo&, EquationSystems&, std::vector<FieldInfo>&, Dyadic3D&, Vector3D&)) {
    libmesh_assert( !this->is_membrane );
    this->_multiscale = f_ptr;
}
inline
void RegionInfo::set_plastic_deform (void (*f_ptr)(const MaterialInfo&, const FieldInfo&)) {
    this->_plastic_deform = f_ptr;
}
//-------------------------------------------------------------------------------------------------
inline
void RegionInfo::calc_biomech_2d (const FieldInfo& fld, Dyadic2D& S, Quindic2D& D) const {
    this->_biomech_2d(*(this->mat), fld, S, D);
}
inline
void RegionInfo::calc_biomech_3d (const FieldInfo& fld, Dyadic3D& S, Quindic3D& Duu, Dyadic3D& Dup, Real& P, Dyadic3D& Dpu, Real& Dpp) const {
    this->_biomech_3d(*(this->mat), fld, S, Duu, Dup, P, Dpu, Dpp);
}
inline
void RegionInfo::calc_biochem (FieldInfo& fld, Blitz_VectorReal& F, Blitz_VectorGradient& G, Blitz_DyadicReal& K, Blitz_VectorPoint& V, Blitz_DyadicReal& D, Blitz_DyadicGradient& X) const {
    this->_biochem(*(this->mat), fld, F, G, K, V, D, X);
}
inline
void RegionInfo::calc_damage (FieldInfo& fld) const {
    if ( this->is_damaged ) this->_damage(*(this->mat), fld);
    else this->_non_damage(*(this->mat), fld);
}
inline
void RegionInfo::calc_coupled_2d (const FieldInfo& fld, Dyadic2D& S, Quindic2D& D) const {
    this->_coupled_2d(*(this->mat), fld, S, D);
}
inline
void RegionInfo::calc_coupled_3d (const FieldInfo& fld, Dyadic3D& S, Quindic3D& Duu, Dyadic3D& Dup, Real& P, Dyadic3D& Dpu, Real& Dpp) const {
    this->_coupled_3d(*(this->mat), fld, S, Duu, Dup, P, Dpu, Dpp);
}
inline
void RegionInfo::calc_multiscale (const FieldInfo& fld, EquationSystems& es_m, std::vector<FieldInfo>& fld_m, Dyadic3D& S, Vector3D& Q) const {
    this->_multiscale(*(this->mat), fld, es_m, fld_m, S, Q);
}
inline
void RegionInfo::calc_plastic_deform (const FieldInfo& fld) const {
    this->_plastic_deform(*(this->mat), fld);
}
//=================================================================================================
#endif // __UTILS_REGION_INFO_H__
//=================================================================================================
