/*************************************************************************************************

    Copyright (C) 2009-2019    Jean de Montigny (jean.demontigny@hotmail.fr)
                               Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#define BDM_VERBOSE
#undef BDM_VERBOSE
//=================================================================================================
// BioDynaMo environment parameters that will be used for all simulations
#ifdef SET_BIODYNAMO_PARAMETERS
# error "macro \"SET_BIODYNAMO_PARAMETERS\" has been defined elsewhere"
#else
# undef  SET_BIODYNAMO_PARAMETERS_
# define SET_BIODYNAMO_PARAMETERS_(_params_) \
  auto set_bdm_params = [&] (bdm::Param* param) \
  { \
    const std::vector<std::string> _svn {"diameter_", "phenotype_", "phenotype2transform_", "age_", \
                                         "convection_", "oxygen_", "oxygen_grad_", "ecm_", "ecm_grad_", \
                                         "cc_grow_", "cc_divide_", "cc_migrate_", "cc_transform_"}; \
    const std::vector<std::string> _svt {"Float64", "Int32", "Int32", "Int32", \
                                         "Float64_3", "Float64", "Float64_3", "Float64", "Float64_3", \
                                         "Float64", "Float64", "Float64", "Float64"}; \
    _params_.set<unsigned int>("BioDynaMo/simulation_output_variables_in_total") = _svn.size(); \
    for (unsigned int I=0; I<_svn.size(); I++) \
    { \
      _params_.set<std::string>("BioDynaMo/simulation_output_variable_"+std::to_string(I)+"/name") = _svn[I]; \
      _params_.set<std::string>("BioDynaMo/simulation_output_variable_"+std::to_string(I)+"/type") = _svt[I]; \
    } \
    if ( _svn.size() != _svt.size() ) libmesh_error(); \
    \
    param->simulation_time_step_ = _params_.get<double>("BioDynaMo/simulation_time_step"); \
    param->bound_space_ = true; \
    param->min_bound_ = _params_.get<Real>("BioDynaMo/bounding_box/min"); \
    param->max_bound_ = _params_.get<Real>("BioDynaMo/bounding_box/max"); \
    param->run_mechanical_interactions_ = true; \
    param->simulation_max_displacement_ = _params_.get<double>("BioDynaMo/cells_repulsion_max_displacement"); \
    param->show_simulation_step_ = false; \
    param->calculate_gradients_ = false; \
    param->export_visualization_ = _params_.set<bool>("BioDynaMo/use_default_visualization_tools"); \
    param->output_dir_ = _params_.get<std::string>("BioDynaMo/visualization_output_path"); \
    param->visualization_export_interval_ = _params_.get<int>("BioDynaMo/visualization_output_interval"); \
    param->visualize_sim_objects_["MyCell"] = std::set<std::string>(_svn.begin(), _svn.end()); \
  }
#endif
//=================================================================================================
#include <sys/stat.h>
#include "feb3/domain_data.h"
#include "biodynamo.h"
#include "mycell.h"
#include "feb3/bdm_interface.h"
#include "feb3/bdm_migrating_cell.h"
#include "./params.h"
#include "libmesh/perf_log.h"
//=================================================================================================
// external variable - perfomance log for BioDynaMo
extern
PerfLog bdm_log;
//=================================================================================================
//
//
//
//=================================================================================================
namespace bdm {
//=================================================================================================
// define compile-time parameter
BDM_CTPARAM() {
  //
  BDM_CTPARAM_HEADER();
  //
  // list objects we use and their biology modules
  using SimObjectTypes = CTList<MyCell>;
  BDM_CTPARAM_FOR(bdm, MyCell) {
    using BiologyModules = CTList<MyCellBiologyModule>;
  };
};
//=================================================================================================
void update_cells_position (BDM_Domain& rve)
{
  //bdm_log.push("bdm::update_cells_position");

  // clear this container first
  rve.cells_xyz.clear();
  // access all cells of this representative volume element
  auto all_cells = rve.sim->GetResourceManager()->template Get<MyCell>();
  // loop for all cells
  for (unsigned int j=0; j<all_cells->size(); j++) {
    auto cell = (*all_cells)[j];
    // insert the current coordinates of the cell in container
    rve.cells_xyz.push_back(cell.GetPosition());
  }

  //bdm_log.pop("bdm::update_cells_position");
}
//-------------------------------------------------------------------------------------------------
template <typename TSimulation = Simulation<>>
void create_cells ( BDM_Domain& rve, const libMesh::Parameters& params,
                    const int phenotype, const int n_cells,
                    const libMesh::BoundingBox& bb, const int pattern )
{
  //bdm_log.push("bdm::create_cells");

  // sanity check
  if (phenotype >= (int)rve.n_phenotypes()) libmesh_error();

  const Real safe_dist = params.get<Real>("BioDynaMo/cells_distance_to_avoid_overlap");

  const Real x_min = bb.min()(0), x_max = bb.max()(0),
             y_min = bb.min()(1), y_max = bb.max()(1),
             z_min = bb.min()(2), z_max = bb.max()(2);

  auto* sim = TSimulation::GetActive();

  // loop to generate the number of cells required
  for (int icell=0; icell<n_cells; icell++) {
    // obtain a random position (within user-defined bounds)
    const RealVector3d xyz = { sim->GetRandom()->Uniform(x_min, x_max) ,
                               sim->GetRandom()->Uniform(y_min, y_max) ,
                               sim->GetRandom()->Uniform(z_min, z_max) };
    //
    //if ( ! phenotype ) libmesh_error();
    if ( -1 == pattern ) {
      ; // do nothing here...
    } else if ( 0 == pattern ) {
      // initialise cells in a 3D hexagonal domain
      ; // do nothing here...
    } else if ( 1 == pattern ) {
      // initialise cells in a 3D ellipsoidal domain
      const double a = 0.5*(x_max-x_min), xo = 0.5*(x_max+x_min),
                   b = 0.5*(y_max-y_min), yo = 0.5*(y_max+y_min),
                   c = 0.5*(z_max-z_min), zo = 0.5*(z_max+z_min);
      const double E3D = pow2((xyz[0]-xo)/a) // Ellipsoid in 3D
                       + pow2((xyz[1]-yo)/b)
                       + pow2((xyz[2]-zo)/c);
      if ( E3D>1.0 ) {
        icell--;
        continue;
      }
    } else libmesh_error();
    //
    // check if proposed space vector does not place new cell
    // "overlapping" to another existing one
    bool it_overlaps = false;
    if ( -1 != pattern )
    // do iterate wrt to the existing cells found in the RVE
    for (unsigned int l=0; l<rve.cells_xyz.size();  l++) {
      const RealVector3d& other_xyz = rve.cells_xyz[l];
      //
      const Real DMag = sqrt(pow2(other_xyz[0]-xyz[0])
                            +pow2(other_xyz[1]-xyz[1])
                            +pow2(other_xyz[2]-xyz[2]));
      if ( DMag > safe_dist ) continue;
      it_overlaps = true;
      break;
    }
    if ( it_overlaps ) {
      icell--;
      continue;
    }
    //
    // ...otherwise do accept the proposed initial position ;)
    rve.cells_xyz.push_back(xyz);

    const int other_phenotype =
      params.get<int>("BioDynaMo/cell_"+std::to_string(phenotype)+"/transform_to_another_phenotype");
    const double diameter =
      params.get<Real>("BioDynaMo/cell_"+std::to_string(phenotype)+"/diameter/ref");

    // create the cell at desired position
    MyCell cell(xyz);
    // define the cell phenotype
    cell.SetPhenotype(phenotype);
    // define the cell phenotype this one could transform into
    cell.SetPhenotype2Transform(other_phenotype);
    // set the size (average diameter) of the cell
    cell.SetDiameter(diameter);
    // set the cell density
    cell.SetDensity(1.0);
    // set the pointer to the (model) parameter data
    cell.SetParametersPointer(&params);
    // set the age of the cell of course (default =0)
    cell.SetAge();
    // set cell convection field (interstitial fluid velocity)
    cell.SetConvection(rve.interstitial_fluid_velocity);
    // set the oxygen (saturation) level and its gradient
    cell.SetOxygen(rve.oxyg);
    cell.SetOxygenGrad(rve.oxyg_grad);
    // set the extracellular matrix (normalized) density and its gradient
    cell.SetECM(rve.ecm);
    cell.SetECMGrad(rve.ecm_grad);
    // check cell state again wrt O2 if only it's in under normoxia|hypoxia
    if ( cell.GetPhenotype() ) {
      // oxygen levels...
      const Real oxygenHypoxia =
        params.get<Real>("BioDynaMo/cell_"+std::to_string(cell.GetPhenotype())+"/O2_limit/to_hypoxia"),
                 oxygenNecrosi =
        params.get<Real>("BioDynaMo/cell_"+std::to_string(cell.GetPhenotype())+"/O2_limit/to_necrosis");
      // based on the oxygen level confirm or update the state of the cell
      unsigned int up_phenotype = 0;
      if ( cell.GetPhenotype() % 2 ) {
        // currently, the cell is in normoxic state, but...
        if      (cell.GetOxygen() > oxygenHypoxia) up_phenotype = cell.GetPhenotype();
        else if (cell.GetOxygen() > oxygenNecrosi) up_phenotype = cell.GetPhenotype()+1;
        else                                       up_phenotype = 0;
      } else {
        // currently, the cell is in hypoxic state, but...
        if      (cell.GetOxygen() > oxygenHypoxia) up_phenotype = cell.GetPhenotype()-1;
        else if (cell.GetOxygen() > oxygenNecrosi) up_phenotype = cell.GetPhenotype();
        else                                       up_phenotype = 0;
      }
      // update the cell phenotype
      cell.SetPhenotype(up_phenotype);
    }
    // set the chemical cue values responsible for enhanced growth, division, migration and
    // transformation (from this phenotype to another)
    if ( cell.GetPhenotype() ) {
      // set the chemical cue value
      const int index__cc_grow =
        params.get<int>("BioDynaMo/cell_"+std::to_string(cell.GetPhenotype())+"/chemical_cue_for_growth/FEB3_communicated_field");
      if ( -1 != index__cc_grow )
        cell.SetChemicalCue_Grow(rve.biochem[index__cc_grow]);
      // set the chemical cue value
      const int index__cc_divide =
        params.get<int>("BioDynaMo/cell_"+std::to_string(cell.GetPhenotype())+"/chemical_cue_for_division/FEB3_communicated_field");
      if ( -1 != index__cc_divide )
        cell.SetChemicalCue_Divide(rve.biochem[index__cc_divide]);
      // set the chemical cue value
      const int index__cc_migrate =
        params.get<int>("BioDynaMo/cell_"+std::to_string(cell.GetPhenotype())+"/chemical_cue_for_migration/FEB3_communicated_field");
      if ( -1 != index__cc_migrate )
        cell.SetChemicalCue_Migrate(rve.biochem[index__cc_migrate]);
      // set the chemical cue value
      const int index__cc_transform =
        params.get<int>("BioDynaMo/cell_"+std::to_string(cell.GetPhenotype())+"/chemical_cue_for_transformation/FEB3_communicated_field");
      if ( -1 != index__cc_transform )
        cell.SetChemicalCue_Transform(rve.biochem[index__cc_transform]);
    }
    //
    // plug in the cell biological module
    cell.AddBiologyModule(MyCellBiologyModule());

    // insert cell to the resource manager
    sim->GetResourceManager()->push_back(cell);
  } // end for nb of cell to create

  //bdm_log.pop("bdm::create_cells");
}
//-------------------------------------------------------------------------------------------------
template <typename TSimulation = Simulation<>>
void migrate_cells ( MigratingCell& incoming_cell,
                     const BDM_Domain& rve, const libMesh::Parameters& params )
{
  //bdm_log.push("bdm::migrate_cells");

  auto* sim = TSimulation::GetActive();

  //
  // NOTE: not all information are passed to the cell that has just migrated (invading),
  //       as we are doing with the create_cell routine above. for example we do not
  //       provide the biochemical info, the microstructural info, as well as the state
  //       of the cell. also, as we see (where we call this function in the simulation below)
  //       the migration is permitted to the cell of phenotype =1 only!
  //

  const int phenotype = incoming_cell.phenotype;
  const int other_phenotype =
    params.get<int>("BioDynaMo/cell_"+std::to_string(phenotype)+"/transform_to_another_phenotype");
  const double diameter = incoming_cell.diameter;
  const int age = incoming_cell.age;

  // create the cell at desired position
  MyCell cell(incoming_cell.position);
  // set the size (average diameter) of the cell
  cell.SetDiameter(diameter);
  // set the cell density
  cell.SetDensity(1.0);
  // set the pointer to the (model) parameter data
  cell.SetParametersPointer(&params);
  // define the cell phenotype
  cell.SetPhenotype(phenotype);
  // define the cell phenotype this one could transform into
  cell.SetPhenotype2Transform(other_phenotype);
  // set the age of the cell of course
  cell.SetAge(age);
  // set cell convection field (interstitial fluid velocity)
  cell.SetConvection(rve.interstitial_fluid_velocity);
  // set the oxygen (saturation) level and its gradient
  cell.SetOxygen(rve.oxyg);
  cell.SetOxygenGrad(rve.oxyg_grad);
  // set the extracellular matrix (normalized) density and its gradient
  cell.SetECM(rve.ecm);
  cell.SetECMGrad(rve.ecm_grad);
  // check cell state again wrt O2 if only it's in under normoxia|hypoxia
  if ( cell.GetPhenotype() ) {
    // oxygen levels...
    const Real oxygenHypoxia =
      params.get<Real>("BioDynaMo/cell_"+std::to_string(cell.GetPhenotype())+"/O2_limit/to_hypoxia"),
               oxygenNecrosi =
      params.get<Real>("BioDynaMo/cell_"+std::to_string(cell.GetPhenotype())+"/O2_limit/to_necrosis");
    // based on the oxygen level confirm or update the state of the cell
    unsigned int up_phenotype = 0;
    if ( cell.GetPhenotype() % 2 ) {
      // currently, the cell is in normoxic state, but...
      if      (cell.GetOxygen() > oxygenHypoxia) up_phenotype = cell.GetPhenotype();
      else if (cell.GetOxygen() > oxygenNecrosi) up_phenotype = cell.GetPhenotype()+1;
      else                                       up_phenotype = 0;
    } else {
      // currently, the cell is in hypoxic state, but...
      if      (cell.GetOxygen() > oxygenHypoxia) up_phenotype = cell.GetPhenotype()-1;
      else if (cell.GetOxygen() > oxygenNecrosi) up_phenotype = cell.GetPhenotype();
      else                                       up_phenotype = 0;
    }
    // update the cell phenotype
    cell.SetPhenotype(up_phenotype);
  }
  // set the chemical cue values responsible for enhanced growth, division, migration and
  // transformation (from this phenotype to another)
  if ( cell.GetPhenotype() ) {
    // set the chemical cue value
    const int index__cc_grow =
      params.get<int>("BioDynaMo/cell_"+std::to_string(cell.GetPhenotype())+"/chemical_cue_for_growth/FEB3_communicated_field");
    if ( -1 != index__cc_grow )
      cell.SetChemicalCue_Grow(rve.biochem[index__cc_grow]);
    // set the chemical cue value
    const int index__cc_divide =
      params.get<int>("BioDynaMo/cell_"+std::to_string(cell.GetPhenotype())+"/chemical_cue_for_division/FEB3_communicated_field");
    if ( -1 != index__cc_divide )
      cell.SetChemicalCue_Divide(rve.biochem[index__cc_divide]);
    // set the chemical cue value
    const int index__cc_migrate =
      params.get<int>("BioDynaMo/cell_"+std::to_string(cell.GetPhenotype())+"/chemical_cue_for_migration/FEB3_communicated_field");
    if ( -1 != index__cc_migrate )
      cell.SetChemicalCue_Migrate(rve.biochem[index__cc_migrate]);
    // set the chemical cue value
    const int index__cc_transform =
      params.get<int>("BioDynaMo/cell_"+std::to_string(cell.GetPhenotype())+"/chemical_cue_for_transformation/FEB3_communicated_field");
    if ( -1 != index__cc_transform )
      cell.SetChemicalCue_Transform(rve.biochem[index__cc_transform]);
  }
  //
  // plug in the cell biological module
  cell.AddBiologyModule(MyCellBiologyModule());

  // insert cell to the resource manager
  sim->GetResourceManager()->push_back(cell);

  //bdm_log.pop("bdm::migrate_cells");
}
//-------------------------------------------------------------------------------------------------
template <typename TSimulation = Simulation<>>
void add_face_incoming_cells (BDM_Domain& rve, const libMesh::Parameters& params)
{
  //bdm_log.push("bdm::add_face_incoming_cells");

  // auto* sim = TSimulation::GetActive();

  // volume fractions...
  const Real VF_ECM = rve.ecm,
             VF_IF  = params.get<double>("BioDynaMo/interstitial_fluid_volume_fraction"),
             VF_Cell= rve.cells_volume_fraction;

  // if the sum of volume fractions is above an alarming level then exit calculations
  if ( VF_ECM + VF_IF + VF_Cell >= 1.0 - params.get<double>("BioDynaMo/volume_fraction/alarm_margin_to_stop") )
    return;

  // distinguish amongst the different solvers of FEB3 coupled with BioDynaMo
  const std::string solver_ID =
    params.get<std::string>("ID of FEB3 solver to communicate with BioDynaMo");
  if ( "FEM_BIOCHEMICS_SOLVER-13-1-2" == solver_ID ) {
    //
    for (int ivar=0; ivar<params.get<int>("BioDynaMo/FEB3_communicated_fields_number"); ivar++) {
      const std::string pn = "BioDynaMo/FEB3_communicated_field_"+std::to_string(ivar);
      //
      // check if we should consider this state-variable or not
      if ( -1 == params.get<int>(pn) ) continue;
      // obtain the counterpart phenotype index for this variable of the reaction-diffusion
      // solver - note that this variable pairing is predefined in the "src/input.C" file!
      const unsigned int t = static_cast<unsigned int>(params.get<int>(pn));
      //
      // if supposedly to income cell being a necrotic one then simply ignore
      // adding it into the list... just ignore it!
      if ( 0 == t ) continue;
      //
      const unsigned int in_cell_min =
        params.get<unsigned int>("BioDynaMo/cell_"+std::to_string(t)+"/face_incoming_cell_threshold/min");
      const unsigned int in_cell_max =
        params.get<unsigned int>("BioDynaMo/cell_"+std::to_string(t)+"/face_incoming_cell_threshold/max");
      //
      const Real&               cellX = rve.biochem[ivar];
      const RealVector3d&  Grad_cellX = rve.biochem_grad[ivar];
      // simply skip the following if the macroscopic solver has predicted zero
      // or negative values for this state-variable...
      if ( cellX <= 0.0 ) continue;
      // if the cell population predicted by BioDynaMo (@previous step) is
      // smaller than the corresponding then surely we have some cells incoming...
      Real Delta_cellX = cellX - rve.cells_population[t];
      if ( Delta_cellX < in_cell_min ) continue;
      if ( Delta_cellX > in_cell_max ) Delta_cellX = in_cell_max;
      //
      // const libMesh::Point v_rand = libMesh::Point( sim->GetRandom()->Uniform(-1.0, +1.0) ,
      //                                               sim->GetRandom()->Uniform(-1.0, +1.0) ,
      //                                               sim->GetRandom()->Uniform(-1.0, +1.0) );
      const libMesh::Point v_cell = libMesh::Point( rve.cell_dynamics[t][0] ,
                                                    rve.cell_dynamics[t][1] ,
                                                    rve.cell_dynamics[t][2] );
      const libMesh::Point v_grad = libMesh::Point( Grad_cellX[0] ,
                                                    Grad_cellX[1] ,
                                                    Grad_cellX[2] );
      //
      // direction to which cells migrate, thus, dictate how they will enter the RVE
      // and from which face (this is calculated below)
      const libMesh::Point dir = (v_cell.norm() <= 1.0e-6 ? v_grad.unit() : v_cell.unit());
      //
      const libMesh::Point eta[] = { libMesh::Point(+1.0,  0.0,  0.0) , // face 0
                                     libMesh::Point(-1.0,  0.0,  0.0) , // face 1
                                     libMesh::Point( 0.0, +1.0,  0.0) , // face 2
                                     libMesh::Point( 0.0, -1.0,  0.0) , // face 3
                                     libMesh::Point( 0.0,  0.0, +1.0) , // face 4
                                     libMesh::Point( 0.0,  0.0, -1.0) };// face 5
      // calculate the dot product of the cells gradient wrt to the "inward"
      // normal vector of each face of the cubic `BDM_Domain`
      std::vector<Real> weight;
      for (unsigned int f=0; f<6; f++) {
        const Real cos_w = dir * eta[f];
        weight.push_back( cos_w < 0.0 ? 0.0 : cos_w );
      }
      const Real sum_w = std::accumulate(weight.begin(), weight.end(), 0.0);
      if ( fabs(sum_w) <= 1.0e-6 ) continue;
      //
      for (unsigned int f=0; f<6; f++) {
        const Real pct = weight[f] / sum_w;
        const unsigned int in_cell = std::ceil(pct * Delta_cellX);
        //
        rve.income_cells_per_face[t][f] = 0;
        //
        if ( in_cell < in_cell_min ) continue;
        //
        rve.income_cells_per_face[t][f] = (in_cell > in_cell_max ? in_cell_max : in_cell);
      }
      // ...end of communicating variables loop
    }
    //
    #ifdef BDM_VERBOSE
    if ( 0 )
    {/////////////////////////////////////////
      std::cout << "   " << rve.id() << "   "
                << "   " << rve.n_phenotypes() << std::endl;
      for (unsigned int P=0; P<rve.n_phenotypes(); P++) {
          std::cout << "\n phenotype = " << P << " ... income_cells_per_face = {";
          for (int i=0; i<6; i++)
              std::cout << ' ' << rve.income_cells_per_face[P][i];
          std::cout << "}" << std::endl;
          for (int i=0; i<6; i++) {
              if ( rve.income_cells_per_face[P][i] )
                libmesh_error();
          }
      }
      std::cout << std::endl;
    }/////////////////////////////////////////
    #endif
    //
  } else if ( "NULL" == solver_ID ) {
    //
    ;
    //
  } else libmesh_error();

  // const int pattern = params.get<bool>("BioDynaMo/random_generator_fix_seed")
  //                   ? -1 : 0;
  const int pattern = -1;
  const Real minCRD_ = params.get<Real>("BioDynaMo/bounding_box/floor"),
             maxCRD_ = params.get<Real>("BioDynaMo/bounding_box/ceil"),
             minCRD0 = params.get<Real>("BioDynaMo/bounding_box/floor (higher)"),
             maxCRD0 = params.get<Real>("BioDynaMo/bounding_box/ceil (lower)");
  //
  // create cells coming from the neighbour - consider the 6 faces of the domain
  for (unsigned int t=0; t<rve.n_phenotypes(); t++) {
    // add incoming cells (face 0, x=min)
    if ( rve.income_cells_per_face[t][0] > 0 ) {
      const libMesh::BoundingBox bb( libMesh::Point(minCRD_, minCRD_, minCRD_) ,
                                     libMesh::Point(minCRD0, maxCRD_, maxCRD_) );
      bdm::create_cells(rve, params, t, rve.income_cells_per_face[t][0], bb, pattern);
    }
    // add incoming cells (face 1, x=max)
    if ( rve.income_cells_per_face[t][1] > 0 ) {
      const libMesh::BoundingBox bb( libMesh::Point(maxCRD0, minCRD_, minCRD_) ,
                                     libMesh::Point(maxCRD_, maxCRD_, maxCRD_) );
      bdm::create_cells(rve, params, t, rve.income_cells_per_face[t][1], bb, pattern);
    }
    // add incoming cells (face 2, y=min)
    if ( rve.income_cells_per_face[t][2] > 0 ) {
      const libMesh::BoundingBox bb( libMesh::Point(minCRD_, minCRD_, minCRD_) ,
                                     libMesh::Point(maxCRD_, minCRD0, maxCRD_) );
      bdm::create_cells(rve, params, t, rve.income_cells_per_face[t][2], bb, pattern);
    }
    // add incoming cells (face 3, y=max)
    if ( rve.income_cells_per_face[t][3] > 0 ) {
      const libMesh::BoundingBox bb( libMesh::Point(minCRD_, maxCRD0, minCRD_) ,
                                     libMesh::Point(maxCRD_, maxCRD_, maxCRD_) );
      bdm::create_cells(rve, params, t, rve.income_cells_per_face[t][3], bb, pattern);
    }
    // add incoming cells (face 4, z=min)
    if ( rve.income_cells_per_face[t][4] > 0 ) {
      const libMesh::BoundingBox bb( libMesh::Point(minCRD_, minCRD_, minCRD_) ,
                                     libMesh::Point(maxCRD_, maxCRD_, minCRD0) );
      bdm::create_cells(rve, params, t, rve.income_cells_per_face[t][4], bb, pattern);
    }
    // add incoming cells (face 5, z=max)
    if ( rve.income_cells_per_face[t][5] > 0 ) {
      const libMesh::BoundingBox bb( libMesh::Point(minCRD_, minCRD_, maxCRD0) ,
                                     libMesh::Point(maxCRD_, maxCRD_, maxCRD_) );
      bdm::create_cells(rve, params, t, rve.income_cells_per_face[t][5], bb, pattern);
    }
  }

  //bdm_log.pop("bdm::add_face_incoming_cells");
}
//-------------------------------------------------------------------------------------------------
template <typename TResourceManager = ResourceManager<>>
void remove_migrating_cells (BDM_Domain& rve, const libMesh::Parameters& params)
{
  //bdm_log.push("bdm::remove_migrating_cells");

  const Real minCRD_ = params.get<Real>("BioDynaMo/bounding_box/min"),
             maxCRD_ = params.get<Real>("BioDynaMo/bounding_box/max");

  // access all cells of this representative volume element
  auto all_cells = rve.sim->GetResourceManager()->template Get<MyCell>();
  // iterate for all cells and check if anyone attempts to escape
  // the representative volume element
  for (unsigned int j=0; j<all_cells->size(); j++) {
    auto cell = (*all_cells)[j];
    //
    // create a MigratingCell object containing all info for cell migration
    MigratingCell escaping_cell;
    escaping_cell.position = cell.GetPosition();
    escaping_cell.diameter = cell.GetDiameter();
    escaping_cell.phenotype = cell.GetPhenotype();
    escaping_cell.age = cell.GetAge();
    //
    const std::array<double, 3> xyz = cell.GetPosition();
    //
    if ( xyz[0] < minCRD_ || xyz[0] > maxCRD_ ||
         xyz[1] < minCRD_ || xyz[1] > maxCRD_ ||
         xyz[2] < minCRD_ || xyz[2] > maxCRD_ ) {
      // make sure to add this migrating cell into the list
      rve.escape_cells_xyz.push_back(escaping_cell);
      // ...and now remove this cell from the BioDynaMo simulation
      cell.RemoveFromSimulation();
    }
  }

  //bdm_log.pop("bdm::remove_migrating_cells");
}
//-------------------------------------------------------------------------------------------------
template <typename TResourceManager = ResourceManager<>>
void remove_face_escaping_cells (BDM_Domain& rve, const libMesh::Parameters& params)
{
  //bdm_log.push("bdm::remove_face_escaping_cells");

  const Real minCRD_ = params.get<Real>("BioDynaMo/bounding_box/min"),
             maxCRD_ = params.get<Real>("BioDynaMo/bounding_box/max");

  const libMesh::Point vertex[] = { libMesh::Point(minCRD_, minCRD_, minCRD_) ,
                                    libMesh::Point(maxCRD_, minCRD_, minCRD_) ,
                                    libMesh::Point(maxCRD_, maxCRD_, minCRD_) ,
                                    libMesh::Point(minCRD_, minCRD_, minCRD_) ,
                                    libMesh::Point(minCRD_, minCRD_, maxCRD_) ,
                                    libMesh::Point(maxCRD_, minCRD_, maxCRD_) ,
                                    libMesh::Point(maxCRD_, maxCRD_, maxCRD_) ,
                                    libMesh::Point(minCRD_, minCRD_, maxCRD_) };

  const TripleUInt vertex2face[] = { TripleUInt(0, 2, 4) ,
                                     TripleUInt(1, 2, 4) ,
                                     TripleUInt(1, 3, 4) ,
                                     TripleUInt(0, 3, 4) ,
                                     TripleUInt(0, 2, 5) ,
                                     TripleUInt(1, 2, 5) ,
                                     TripleUInt(1, 3, 5) ,
                                     TripleUInt(0, 3, 5) };

  const Real tol = params.get<Real>("BioDynaMo/bounding_box/tol/0.1pct");

  // access all cells of this representative volume element (RVE)
  auto all_cells = rve.sim->GetResourceManager()->template Get<MyCell>();
  // iterate for all cells and check if anyone attempts to escape
  // the representative volume element
  for (unsigned int j=0; j<all_cells->size(); j++) {
    auto cell = (*all_cells)[j];
    //
    const unsigned int t = cell.GetPhenotype();
    const libMesh::Point xyz(cell.GetPosition()[0], cell.GetPosition()[1], cell.GetPosition()[2]);
    //
    // if this is a newly-created cell that happens to live at the edge of the
    // boundary then it might have been positioned (after dublication) there
    // due to the "cell repulsion" mechanism, then simply ignore removing it from
    // the simulation
    if ( cell.GetAge() < 2 ) continue;
    //
    if        ( xyz(0) < minCRD_+tol ) {
      // make sure to increment this index
      rve.escape_cells_per_face[t][0] += 1;
      // ...and now remove this cell from the BioDynaMo simulation
      cell.RemoveFromSimulation();
    } else if ( xyz(0) > maxCRD_-tol ) {
      // make sure to increment this index
      rve.escape_cells_per_face[t][1] += 1;
      // ...and now remove this cell from the BioDynaMo simulation
      cell.RemoveFromSimulation();
    } else if ( xyz(1) < minCRD_+tol ) {
      // make sure to increment this index
      rve.escape_cells_per_face[t][2] += 1;
      // ...and now remove this cell from the BioDynaMo simulation
      cell.RemoveFromSimulation();
    } else if ( xyz(1) > maxCRD_-tol ) {
      // make sure to increment this index
      rve.escape_cells_per_face[t][3] += 1;
      // ...and now remove this cell from the BioDynaMo simulation
      cell.RemoveFromSimulation();
    } else if ( xyz(2) < minCRD_+tol ) {
      // make sure to increment this index
      rve.escape_cells_per_face[t][4] += 1;
      // ...and now remove this cell from the BioDynaMo simulation
      cell.RemoveFromSimulation();
    } else if ( xyz(2) > maxCRD_-tol ) {
      // make sure to increment this index
      rve.escape_cells_per_face[t][5] += 1;
      // ...and now remove this cell from the BioDynaMo simulation
      cell.RemoveFromSimulation();
    } else {
      // now check in case any of the cells attempt to escape at the
      // vertices of the cubic RVE domain; if so then simply randomly
      // remove the cell to any of the corresponding faces of that vertex
      for (unsigned int v=0; v<8; v++) {
        // check if the cell is close to this vertex point
        const libMesh::Point diff = xyz - vertex[v];
        if ( diff.norm() > tol ) continue;
        // if yes, then select any of the 3 faces randomly!
        const int f = vertex2face[v]( uniform_distro(0, 2) );
        //
        // make sure to increment this index
        rve.escape_cells_per_face[t][f] += 1;
        // ...and now remove this cell from the BioDynaMo simulation
        cell.RemoveFromSimulation();
        // now break free from this internal loop
        break;
      }
      // ...end of checks
    }
    // ...end of cells loop
  }

  //bdm_log.pop("bdm::remove_face_escaping_cells");
}
//-------------------------------------------------------------------------------------------------
template <typename TResourceManager = ResourceManager<>>
void init_cells ( BDM_Domain& rve, const libMesh::Parameters& params,
                  const std::vector<int>& n_cells_per_phenotype )
{
  //bdm_log.push("bdm::init_cells");

  rve.sim->Activate();

  // sanity checks
  if (n_cells_per_phenotype.empty()) libmesh_error();
  if (n_cells_per_phenotype.size() != rve.n_phenotypes()) libmesh_error();
  // create cells wrt their type
  for (unsigned int t=0; t<rve.n_phenotypes(); t++) {
    const int phenotype = t;
    const int n_cells = n_cells_per_phenotype[t];
    const Real minCRDx = params.get<double>("BioDynaMo/cell_"+std::to_string(t)+"/bounding_box/X/min"),
               maxCRDx = params.get<double>("BioDynaMo/cell_"+std::to_string(t)+"/bounding_box/X/max");
    const Real minCRDy = params.get<double>("BioDynaMo/cell_"+std::to_string(t)+"/bounding_box/Y/min"),
               maxCRDy = params.get<double>("BioDynaMo/cell_"+std::to_string(t)+"/bounding_box/Y/max");
    const Real minCRDz = params.get<double>("BioDynaMo/cell_"+std::to_string(t)+"/bounding_box/Z/min"),
               maxCRDz = params.get<double>("BioDynaMo/cell_"+std::to_string(t)+"/bounding_box/Z/max");
    const libMesh::BoundingBox bb( libMesh::Point(minCRDx, minCRDy, minCRDz) ,
                                   libMesh::Point(maxCRDx, maxCRDy, maxCRDz) );
    const int pattern = params.get<int>("BioDynaMo/cell_"+std::to_string(t)+"/initialisation_pattern");
    //
    if ( ! n_cells ) continue;
    // now create the cells ;)
    bdm::create_cells(rve, params, phenotype, n_cells, bb, pattern);
  }

  const Real RVE_vol = pow3( params.get<Real>("BioDynaMo/bounding_box/size") );

  // access all cells of this representative volume element
  auto all_cells = rve.sim->GetResourceManager()->template Get<MyCell>();
  // now initialise all data necessary, for all cell types and for all cells requested
  for (unsigned int t=0; t<rve.n_phenotypes(); t++) {
    rve.cells_mass[t] = 0.0;
    rve.cells_population[t] = 0;
  }
  rve.cells_volume_fraction = 0.0;
  for (unsigned int icell=0; icell<all_cells->size(); icell++) {
    auto cell = (*all_cells)[icell];
    const int t = cell.GetPhenotype();
    //
    rve.cells_mass[t] += cell.GetMass();
    rve.cells_population[t] += 1;
    rve.cells_volume_fraction += cell.GetVolume();
  }
  // calculate the cells volume fraction in the RVE
  rve.cells_volume_fraction /= RVE_vol;
  // reset these counters
  for (unsigned int t=0; t<rve.n_phenotypes(); t++) {
    for (unsigned int f=0; f<6; f++) {
      rve.escape_cells_per_face[t][f] = 0;
      rve.income_cells_per_face[t][f] = 0;
    }
    for (unsigned int d=0; d<3; d++) {
      rve.cell_dynamics[t][d] = 0.0;
    }
  }
  // clear counters for incoming & escaping cells
  rve.income_cells_xyz.clear();
  rve.escape_cells_xyz.clear();
  //
  bdm::update_cells_position(rve);

  //bdm_log.pop("bdm::init_cells");
}
//-------------------------------------------------------------------------------------------------
template <typename TResourceManager = ResourceManager<>>
void init_bdm_domains ( std::vector<BDM_Domain>& rve_list, libMesh::Parameters& params,
                        std::vector<libMesh::Point>& rve_preset_xyz )
{
  //bdm_log.push("bdm::init_bdm_domains");

  SET_BIODYNAMO_PARAMETERS_(params);

  const int n_phenotypes = params.get<unsigned int>("BioDynaMo/number_of_phenotypes"),
            n_biochem = params.get<unsigned int>("BioDynaMo/number_of_biochemical_parameters"),
            n_mstruct = params.get<unsigned int>("BioDynaMo/number_of_biomechanical_parameters");

  std::vector<int> n_cells_per_phenotype(n_phenotypes);

  if ( ! rve_preset_xyz.empty() ) libmesh_error();

  const double Delta = params.get<Real>("BioDynaMo/bounding_box/size");
  // number of BioDynaMo simulations: RVE domains
  for (int id=0; id<params.get<int>("BioDynaMo/number_of_RVEs"); id++) {
    libMesh::Point xyz;
    // random distribution of initial location
    do {
       xyz = libMesh::Point( Delta * int(rand()%100) ,
                             Delta * int(rand()%100) ,
                             Delta * int(rand()%100) );
    } while ( isInList(rve_preset_xyz, xyz) );
    // save these coordinates
    rve_preset_xyz.push_back(xyz);
  }

  // number of BioDynaMo simulations: RVE domains
  for (int id=0; id<params.get<int>("BioDynaMo/number_of_RVEs"); id++) {
    // random distribution of initial RVEs location
    const RealVector3d xyz = { rve_preset_xyz[id](0) ,
                               rve_preset_xyz[id](1) ,
                               rve_preset_xyz[id](2) };
    //
    BDM_Domain rve(id, xyz, n_phenotypes, n_biochem, n_mstruct);
    // set some initial values for these data
    for (int t=0; t<n_phenotypes; t++) {
      rve.cells_population[t] = params.get<int>("BioDynaMo/cell_"+std::to_string(t)+"/initial_population");
      //
      n_cells_per_phenotype[t] = params.get<int>("BioDynaMo/cell_"+std::to_string(t)+"/initial_population");
    }
    rve.oxyg = params.get<double>("BioDynaMo/oxygen");
    for (unsigned int s=0; s<3; s++) {
      const std::string v_comp = "BioDynaMo/oxygen_grad/component_"+std::to_string(s);
      rve.oxyg_grad[s] = params.get<double>(v_comp);
    }
    for (unsigned int s=0; s<3; s++) {
      const std::string v_comp = "BioDynaMo/interstitial_fluid_velocity/component_"+std::to_string(s);
      rve.interstitial_fluid_velocity[s] = params.get<double>(v_comp);
    }
    // create a new empty BioDynaMo simulation
    const std::string sname = id
                            ? ("BDM_p"+std::to_string(libMesh::global_processor_id())+"_d")
                            : ("BDM_p"+std::to_string(libMesh::global_processor_id())+"_d0");
    rve.sim = new bdm::Simulation<>(sname, set_bdm_params);

    rve_list.push_back( rve );
    // ...end of RVEs (to create) loop
  }

  // initialise data for all BioDynaMo simulations, i.e. per each BDM_Domain
  for (unsigned int j=0; j<rve_list.size(); j++) {
    BDM_Domain& rve = rve_list[j];
    //
    bdm::init_cells(rve, params, n_cells_per_phenotype);
    // ...end of RVEs loop
  }

  //bdm_log.pop("bdm::init_bdm_domains");
}
//-------------------------------------------------------------------------------------------------
template <typename TResourceManager = ResourceManager<>>
void init_migrating_bdm_domains (std::vector<BDM_Domain>& rve_list, libMesh::Parameters& params)
{
  //bdm_log.push("bdm::init_migrating_bdm_domains");

  if ( params.get<bool>("BioDynaMo/cell_migration/is_face_model") ) return;
  // ...new cube creation due to cell migration

  SET_BIODYNAMO_PARAMETERS_(params);

  // get vector containing coordinates of all bdm cubes in simulation
  std::vector<std::array<double, 3>> rve_coord_list = getRveOriginList(rve_list);

  // for cubes in simulation
  for (unsigned int j=0; j<rve_list.size(); j++) {
    BDM_Domain& existing_rve = rve_list[j];
    // if no cells escaped this cube then skip following computations...
    if ( existing_rve.escape_cells_xyz.empty() ) continue;

    // get coord of cube to create, if needs to create new cube
    //TODO: perf improvement: need to check only cubes that don't have 6 neighbours
    std::vector<std::array<double, 3>> coord_cubes_to_create =
      getCoordCubesToCreate(rve_coord_list, existing_rve, params);

    // if escaped cells triggered new cube creation
    if ( ! coord_cubes_to_create.empty() ) {
      // for each new cube to create
      for (std::array<double, 3> coord_to_create : coord_cubes_to_create) {

        if ( isInList(rve_coord_list, coord_to_create) ) continue;

        // add new cube origin to rve_coord_list
        rve_coord_list.push_back(coord_to_create);

        const int n_phenotypes = params.get<unsigned int>("BioDynaMo/number_of_phenotypes"),
                  n_biochem = params.get<unsigned int>("BioDynaMo/number_of_biochemical_parameters"),
                  n_mstruct = params.get<unsigned int>("BioDynaMo/number_of_biomechanical_parameters");

        std::vector<int> n_cells_per_phenotype(n_phenotypes);

        BDM_Domain rve(rve_list.size(), coord_to_create, n_phenotypes, n_biochem, n_mstruct);
        // set some initial values for these data
        for (int t=0; t<n_phenotypes; t++) {
          rve.cells_population[t] = params.get<int>("BioDynaMo/cell_"+std::to_string(t)+"/initial_population");
          //
          n_cells_per_phenotype[t] = params.get<int>("BioDynaMo/cell_"+std::to_string(t)+"/initial_population");
        }
        //
        rve.oxyg = params.get<double>("BioDynaMo/oxygen");
        for (unsigned int s=0; s<3; s++) {
          const std::string v_comp = "BioDynaMo/oxygen_grad/component_"+std::to_string(s);
          rve.oxyg_grad[s] = params.get<double>(v_comp);
        }
        for (unsigned int s=0; s<3; s++) {
          const std::string v_comp = "BioDynaMo/interstitial_fluid_velocity/component_"+std::to_string(s);
          rve.interstitial_fluid_velocity[s] = params.get<double>(v_comp);
        }
        // create a new empty BioDynaMo simulation
        const std::string sname = ("BDM_p"+std::to_string(libMesh::global_processor_id())+"_d");
        rve.sim = new bdm::Simulation<>(sname, set_bdm_params);

        rve_list.push_back( rve );

        // initialise bdm simulation wrt info contained in the corresponding BDM_Domain
        bdm::init_cells(rve_list[rve_list.size()-1], params, n_cells_per_phenotype);

      } // end for each RVE to create

    } // end new RVE creation

    // cell migration...
    const double minCoord = params.get<Real>("BioDynaMo/bounding_box/floor");
    const double maxCoord = params.get<Real>("BioDynaMo/bounding_box/ceil");
    // we know that all cubes receiving cells exist, so just transmit cells
    // update income_cells_xyz of corresponding receiving RVE
    for (MigratingCell cell : existing_rve.escape_cells_xyz) {
      // update cell position to its new invading position
      BDM_Domain* invaded_rve = getRveToCreateCellTo(cell.position,
                                                     rve_list, params,
                                                     existing_rve.origin());
      if ( ! invaded_rve ) {
        continue;
      } else {
        cell.position = getInvadingCoord(cell.position, minCoord, maxCoord);
        //
        invaded_rve->income_cells_xyz.push_back(cell);
      }
    }
    // ...end of (existing) RVEs loop
  }

  //bdm_log.pop("bdm::init_migrating_bdm_domains");
}
//-------------------------------------------------------------------------------------------------
template <typename TResourceManager = ResourceManager<>>
void reinit_cells (BDM_Domain& rve, const libMesh::Parameters& params)
{
  //bdm_log.push("bdm::reinit_cells");

  const Real RVE_vol = pow3( params.get<Real>("BioDynaMo/bounding_box/size") );
  const Real SpaceScale = params.get<Real>("BioDynaMo/micro2macro/space_scaling");
  const Real Length = params.get<Real>("BioDynaMo/bounding_box/size")*SpaceScale,
             Area = params.get<Real>("BioDynaMo/bounding_box/face")*pow2(SpaceScale),
             Time = params.get<Real>("BioDynaMo/simulation_time_step_between_iterations");
  //  const Real Coeff = 0.5 * (Area/(Length*Time));
  const Real Coeff = 0.5 * (params.get<Real>("BioDynaMo/micro2macro/space_scaling")/Time);
  //
  auto all_cells = rve.sim->GetResourceManager()->template Get<MyCell>();
  for (unsigned int t=0; t<rve.n_phenotypes(); t++) {
    rve.cells_mass[t] = 0.0;
    rve.cells_population[t] = 0;
  }
  rve.cells_volume_fraction = 0.0;
  for (unsigned int j=0; j<all_cells->size(); j++) {
    auto cell = (*all_cells)[j];
    const unsigned int t = cell.GetPhenotype();
    //
    rve.cells_mass[t] += cell.GetMass();
    rve.cells_population[t] += 1;
    rve.cells_volume_fraction += cell.GetVolume();
  }
  // calculate the cells volume fraction in the RVE
  rve.cells_volume_fraction /= RVE_vol;
  //
  for (unsigned int t=0; t<rve.n_phenotypes(); t++) {
    //
    const Real max_cell_dynamics =
      params.get<Real>("BioDynaMo/cell_"+std::to_string(t)+"/face_outgoing_cell_ratio/max");
    //
    // calculate the (incremental) change of the cell population
    rve.cells_population_increment[t] = rve.cells_population[t]
                                      - rve.cells_population_increment[t];
    // calculate the average rate of the cell population
    const std::array<int, 6> delta_cells_per_face =
      { rve.income_cells_per_face[t][0] - rve.escape_cells_per_face[t][0] ,
        rve.income_cells_per_face[t][1] - rve.escape_cells_per_face[t][1] ,
        rve.income_cells_per_face[t][2] - rve.escape_cells_per_face[t][2] ,
        rve.income_cells_per_face[t][3] - rve.escape_cells_per_face[t][3] ,
        rve.income_cells_per_face[t][4] - rve.escape_cells_per_face[t][4] ,
        rve.income_cells_per_face[t][5] - rve.escape_cells_per_face[t][5] };
    rve.cell_dynamics[t][0] =
    rve.cell_dynamics[t][1] =
    rve.cell_dynamics[t][2] = 0.0;
    if ( delta_cells_per_face[0] != delta_cells_per_face[1] ) {
      rve.cell_dynamics[t][0] = Coeff*(delta_cells_per_face[0]-delta_cells_per_face[1]);
    // } else if ( delta_cells_per_face[0] == delta_cells_per_face[1] &&
    //            (delta_cells_per_face[0] || delta_cells_per_face[1]) ) {
    //   rve.cell_dynamics[t][0] = Coeff*uniform_distro(-0.01, +0.01);
    }
    if ( delta_cells_per_face[2] != delta_cells_per_face[3] ) {
      rve.cell_dynamics[t][1] = Coeff*(delta_cells_per_face[2]-delta_cells_per_face[3]);
    // } else if ( delta_cells_per_face[2] == delta_cells_per_face[3] &&
    //            (delta_cells_per_face[2] || delta_cells_per_face[3]) ) {
    //   rve.cell_dynamics[t][1] = Coeff*uniform_distro(-0.01, +0.01);
    }
    if ( delta_cells_per_face[4] != delta_cells_per_face[5] ) {
      rve.cell_dynamics[t][2] = Coeff*(delta_cells_per_face[4]-delta_cells_per_face[5]);
    // } else if ( delta_cells_per_face[4] == delta_cells_per_face[5] &&
    //            (delta_cells_per_face[4] || delta_cells_per_face[5]) ) {
    //   rve.cell_dynamics[t][2] = Coeff*uniform_distro(-0.01, +0.01);
    }
    //
    if ( max_cell_dynamics < rve.cell_dynamics[t][0] )
        rve.cell_dynamics[t][0] = max_cell_dynamics;
    if ( max_cell_dynamics < rve.cell_dynamics[t][1] )
        rve.cell_dynamics[t][1] = max_cell_dynamics;
    if ( max_cell_dynamics < rve.cell_dynamics[t][2] )
        rve.cell_dynamics[t][2] = max_cell_dynamics;
    //
    // final check whether cell dynamics has overshoot
    const Real v_max = params.get<double>("BioDynaMo/cell_migration/max_speed");
    const Real v_norm = sqrt(pow2(rve.cell_dynamics[t][0])
                            +pow2(rve.cell_dynamics[t][1])
                            +pow2(rve.cell_dynamics[t][2]));
    if ( v_norm > v_max ) {
      const Real s = v_max / v_norm;
      rve.cell_dynamics[t][0] *= s;
      rve.cell_dynamics[t][1] *= s;
      rve.cell_dynamics[t][2] *= s;
    }
    // ...end of phenotypes loop
  }
  //
  bdm::update_cells_position(rve);

  //bdm_log.pop("bdm::reinit_cells");
}
//-------------------------------------------------------------------------------------------------
template <typename TResourceManager = ResourceManager<>>
void pre_simulation (BDM_Domain& rve, const libMesh::Parameters& params)
{
  //bdm_log.push("bdm::pre_simulation");

  const Real RVE_vol = pow3( params.get<Real>("BioDynaMo/bounding_box/size") );
  //
  // initialize some of the counters / indices of the BDM_Domain class
  for (unsigned int t=0; t<rve.n_phenotypes(); t++) {
    rve.cells_population_increment[t] = rve.cells_population[t];
  }
  //
  // current cell invasion/migration model mechanism assumes either
  // the RVE trasfer model or the face invasion model
  if ( params.get<bool>("BioDynaMo/cell_migration/is_face_model") ) {
    bdm::add_face_incoming_cells(rve, params);
  } else {
    for (MigratingCell incoming_cell : rve.income_cells_xyz) {
      bdm::migrate_cells(incoming_cell, rve, params);
    }
  }
  //
  // access all cells of this representative volume element
  auto all_cells = rve.sim->GetResourceManager()->template Get<MyCell>();
  //
  // iterate for all cells and update their traits (phenotype, oxygen, etc.)
  // of course where necessary
  for (unsigned int j=0; j<all_cells->size(); j++) {
    auto cell = (*all_cells)[j];
    //
    // set cell convection field (interstitial fluid velocity)
    cell.SetConvection(rve.interstitial_fluid_velocity);
    // set the oxygen (saturation) level and its gradient
    cell.SetOxygen(rve.oxyg);
    cell.SetOxygenGrad(rve.oxyg_grad);
    // set the extracellular matrix (normalized) density and its gradient
    cell.SetECM(rve.ecm);
    cell.SetECMGrad(rve.ecm_grad);
    // check cell state again wrt O2 if only it's in under normoxia|hypoxia
    if ( cell.GetPhenotype() ) {
      // oxygen levels...
      const Real oxygenHypoxia =
        params.get<Real>("BioDynaMo/cell_"+std::to_string(cell.GetPhenotype())+"/O2_limit/to_hypoxia"),
                 oxygenNecrosi =
        params.get<Real>("BioDynaMo/cell_"+std::to_string(cell.GetPhenotype())+"/O2_limit/to_necrosis");
      // based on the oxygen level confirm or update the state of the cell
      unsigned int up_phenotype = 0;
      if ( cell.GetPhenotype() % 2 ) {
        // currently, the cell is in normoxic state, but...
        if      (cell.GetOxygen() > oxygenHypoxia) up_phenotype = cell.GetPhenotype();
        else if (cell.GetOxygen() > oxygenNecrosi) up_phenotype = cell.GetPhenotype()+1;
        else                                       up_phenotype = 0;
      } else {
        // currently, the cell is in hypoxic state, but...
        if      (cell.GetOxygen() > oxygenHypoxia) up_phenotype = cell.GetPhenotype()-1;
        else if (cell.GetOxygen() > oxygenNecrosi) up_phenotype = cell.GetPhenotype();
        else                                       up_phenotype = 0;
      }
      // update the cell phenotype
      cell.SetPhenotype(up_phenotype);
    }
    // ...end of cells loop
  }
  //
  // initialize some of the counters / indices of the BDM_Domain class
  for (unsigned int t=0; t<rve.n_phenotypes(); t++) {
    rve.cells_mass[t] = 0.0;
    rve.cells_population[t] = 0;
  }
  rve.cells_volume_fraction = 0.0;
  for (unsigned int j=0; j<all_cells->size(); j++) {
    auto cell = (*all_cells)[j];
    const unsigned int t = cell.GetPhenotype();
    rve.cells_mass[t] += cell.GetMass();
    rve.cells_population[t] += 1;
    rve.cells_volume_fraction += cell.GetVolume();
  }
  // calculate the cells volume fraction in the RVE
  rve.cells_volume_fraction /= RVE_vol;
  // reset these counters
  for (unsigned int t=0; t<rve.n_phenotypes(); t++) {
    for (unsigned int f=0; f<6; f++) {
      rve.escape_cells_per_face[t][f] = 0;
      rve.income_cells_per_face[t][f] = 0;
    }
  }
  // clear counters for incoming & escaping cells
  rve.income_cells_xyz.clear();
  rve.escape_cells_xyz.clear();
  //
  // update the coordinates of all cells in the RVE
  bdm::update_cells_position(rve);

  //bdm_log.pop("bdm::pre_simulation");
}
//-------------------------------------------------------------------------------------------------
template <typename TResourceManager = ResourceManager<>>
void simulate_bdm_domains (std::vector<BDM_Domain>& rve_list, libMesh::Parameters& params)
{
  //bdm_log.push("bdm::simulate_bdm_domains");

  const int n_time_steps = params.get<int>("BioDynaMo/number_of_time_steps");
  const bool face_invasion = params.get<bool>("BioDynaMo/cell_migration/is_face_model");

  for (unsigned int j=0; j<rve_list.size(); j++) {
    BDM_Domain& rve = rve_list[j];
    //
    rve.sim->Activate();
    //
    #ifdef BDM_VERBOSE
    std::cout << " RVE: " << rve.id() << std::endl;
    #endif
    bdm::pre_simulation(rve, params);
    //
    // create temporary copy of the cells volume fraction for this RVE only!!!
    params.set<double>("BioDynaMo/cells_volume_fraction") = rve.cells_volume_fraction;
    //
    // loop for all simulation steps
    for (int istep=0; istep<n_time_steps; istep++) {
      // perform the BioDynaMo simulation now
      rve.sim->GetScheduler()->Simulate(1);
      // run either the cube transfer model or the face invasion model
      // and remove cells that cross the RVE boundary
      if ( ! face_invasion )
        bdm::remove_migrating_cells(rve, params);
      else
        bdm::remove_face_escaping_cells(rve, params);
    }
    // re-initialise data for the representative volume element
    bdm::reinit_cells(rve, params);
    //
    #ifdef BDM_VERBOSE
    for (unsigned int t=0; t<rve.n_phenotypes(); t++) {
      std::cout << "\n ===" << std::flush;
      std::cout << " type: " << t << std::flush;
      std::cout << "  ---" << std::flush;
      std::cout << " number of cells: " << rve.cells_population[t] << std::flush;
      std::cout << "  ---" << std::flush;
      std::cout << " number of cells increment: " << rve.cells_population_increment[t] << std::flush;
      std::cout << " ---" << std::flush;
      std::cout << " cell dynamics: " << std::flush;
      for (unsigned int d=0; d<3; d++)
        std::cout << " " << rve.cell_dynamics[t][d] << std::flush;
      std::cout << " ---" << std::flush;
    }
    std::cout << std::endl;
    #endif
    // ...end of RVEs loop
  }

  //bdm_log.pop("bdm::simulate_bdm_domains");
}
//-------------------------------------------------------------------------------------------------
template <typename TResourceManager = ResourceManager<>>
void clean_bdm_domains (std::vector<BDM_Domain>& rve_list)
{
  //bdm_log.push("bdm::clean_bdm_domains");

  for (unsigned int j=0; j<rve_list.size(); j++) {
    BDM_Domain& rve = rve_list[j];
    //
    rve.sim->Activate();
    // delete this pointer now
    delete rve.sim;
    // ...end of RVEs loop
  }
  // and of course empty this vector
  rve_list.clear();

  //bdm_log.pop("bdm::clean_bdm_domains");
}
//=================================================================================================
} // end namespace bdm
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
//=================================================================================================
// external function
extern
void read_BDM_parameters (GetPot&, std::set<std::string>&, libMesh::Parameters&);
//=================================================================================================
void DomainData::init_bdm () {
    bdm_log.push("DomainData::init_bdm");
    //
    Parameters& params = this->_equation_systems.parameters;
    //
    // check we if plan to continue computations or exit normally
    if ( ! params.get<bool>("reaction-diffusion solver: BioDynaMo") ) return;
    //
    // sanity checks
    if ( ! params.get<bool>("BioDynaMo/cell_migration/is_face_model") ) libmesh_error();
    //
    const unsigned int nvars_RDS   = params.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = params.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       //ninvars_NCS = params.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_AUX   = params.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    //
    // pre-set parameters
    {
      params.set<bool>("BioDynaMo/use_default_visualization_tools") = false;
      params.set<std::string>("BioDynaMo/visualization_output_path") = params.get<std::string>("results directory: paraview") + "/BDM";
      //
      params.set<int>("BioDynaMo/number_of_RVEs") = 0;
      params.set<unsigned int>("BioDynaMo/number_of_biochemical_parameters") = nvars_RDS;
      params.set<unsigned int>("BioDynaMo/number_of_biomechanical_parameters") = nvars_AUX;
      //
      params.set<int>("BioDynaMo/number_of_iterations") = 0;
      params.set<double>("BioDynaMo/simulation_time_step") = params.get<Real>("reaction-diffusion solver: time step")
                                                           / params.get<int>("BioDynaMo/number_of_time_steps");
      //
      params.set<bool>("BioDynaMo/cell_migration/is_face_model") = true;
      //
      if ( params.get<double>("BioDynaMo/cell_migration/max_speed") <= 0.0 ) {
        params.set<double>("BioDynaMo/cell_migration/max_speed") =
          (params.get<Real>("finite element grid: min edge length")/params.get<Real>("BioDynaMo/bounding_box/size"))
          / params.get<Real>("reaction-diffusion solver: time step");
      }
      //
      params.set<double>("BioDynaMo/oxygen") = 0.0;
      params.set<double>("BioDynaMo/oxygen_grad/component_0") = 0.0;
      params.set<double>("BioDynaMo/oxygen_grad/component_1") = 0.0;
      params.set<double>("BioDynaMo/oxygen_grad/component_2") = 0.0;
      //
      params.set<double>("BioDynaMo/interstitial_fluid_velocity/component_0") = 0.0;
      params.set<double>("BioDynaMo/interstitial_fluid_velocity/component_1") = 0.0;
      params.set<double>("BioDynaMo/interstitial_fluid_velocity/component_2") = 0.0;
      //
      // check if we utilize any of FEB3's solvers - coupled with BioDynaMo
      switch ( params.get<ProblemCase>("solver") ) {
        case NO_SOLVER :
          //
          params.set<std::string>("ID of FEB3 solver to communicate with BioDynaMo") =
            "NULL";
          //
          break;
        case FEM_BIOCHEMICS_SOLVER :
          //
          if ( params.get<unsigned int>("reaction-diffusion solver: dimension")               ==13 &&
               params.get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 1 &&
               params.get<unsigned int>("reaction-diffusion solver: auxiliary variables")     == 2 ) {
            params.set<std::string>("ID of FEB3 solver to communicate with BioDynaMo") =
              "FEM_BIOCHEMICS_SOLVER-13-1-2";
            //
            params.set<int>("BioDynaMo/FEB3_communicated_fields_number") = 13;
            params.set<int>("BioDynaMo/FEB3_communicated_field_0") = -1;
            params.set<int>("BioDynaMo/FEB3_communicated_field_1") =  0;
            params.set<int>("BioDynaMo/FEB3_communicated_field_2") =  1;
            params.set<int>("BioDynaMo/FEB3_communicated_field_3") =  2;
            params.set<int>("BioDynaMo/FEB3_communicated_field_4") =  3;
            params.set<int>("BioDynaMo/FEB3_communicated_field_5") =  4;
            params.set<int>("BioDynaMo/FEB3_communicated_field_6") =  5;
            params.set<int>("BioDynaMo/FEB3_communicated_field_7") =  6;
            params.set<int>("BioDynaMo/FEB3_communicated_field_8") = -1;
            params.set<int>("BioDynaMo/FEB3_communicated_field_9") = -1;
            params.set<int>("BioDynaMo/FEB3_communicated_field_10")= -1;
            params.set<int>("BioDynaMo/FEB3_communicated_field_11")= -1;
            params.set<int>("BioDynaMo/FEB3_communicated_field_12")= -1;
          } else libmesh_error();
          //
          break;
        default :
          libmesh_error();
      }
      //
    }
    //
    const unsigned int npheno_BDM = params.get<unsigned int>("BioDynaMo/number_of_phenotypes"),
                       nbchem_BDM = params.get<unsigned int>("BioDynaMo/number_of_biochemical_parameters"),
                       nbmech_BDM = params.get<unsigned int>("BioDynaMo/number_of_biomechanical_parameters");

    SET_BIODYNAMO_PARAMETERS_(params);

    // index to track the ID of the (local) BioDynaMo representative volume
    // elements (RVEs) to initialize for the entire simulation
    unsigned int bdm_id = 0;
    //
    // initialize the BioDynaMo data wrt the solver employed
    if ( FEM_BIOCHEMICS_SOLVER == params.get<ProblemCase>("solver") &&
         nvars_RDS == 13 && ndcvars_RDS == 1 && nvars_AUX == 2 )
    {
        std::string sname;
        // sanity checks
        if (nvars_RDS!=nbchem_BDM) libmesh_error();
        if (nvars_AUX!=nbmech_BDM) libmesh_error();
        if (7!=npheno_BDM) libmesh_error();
        //
        #ifdef BDM_VERBOSE
        sname = params.get<std::string>("BioDynaMo/visualization_output_path")
              + "/BDM.res-" + std::to_string(libMesh::global_processor_id());
        std::ofstream fout(sname);
        //
        std::cout << __FILE__ << ' ' << __LINE__ << std::endl;
        std::cout << "this->elem_qi_local.size() = " << this->elem_qi_local.size() << std::endl;
        #endif
        // iterate for all (local) finite elements
        for (auto citer=this->elem_qi_local.begin(); citer!=this->elem_qi_local.end(); citer++) {
            const dof_id_type elem_id = citer->first;
            const QuadratureInfo& qi = citer->second;
            const unsigned int n_qp = qi.n_points();
            //
            // vector of RVEs for this FE we wish to store
            std::vector<BDM_Domain> rve_list_2_add;
            //
            std::vector<std::vector<int>> n_cells_per_phenotype_list;
            //
            #ifdef BDM_VERBOSE
            std::cout << "\telem_ID = " << citer->first << "\tn_qp = " << n_qp << std::endl;
            #endif
            // iterate for all (local) quadrature points
            for (unsigned int qp=0; qp<n_qp; qp++) {
                //==================================
                //
                std::vector<int> n_cells_per_phenotype(npheno_BDM);
                //
                sname = bdm_id
                      ? (".BDM_p"+std::to_string(libMesh::global_processor_id())+"_d")
                      : (".BDM_p"+std::to_string(libMesh::global_processor_id())+"_d0");
                //
                //==================================
                #ifdef BDM_VERBOSE
                const Real RAD = sqrt(pow2(qi.XYZ[qp](0))+pow2(qi.XYZ[qp](1))+pow2(qi.XYZ[qp](2)));
                //
                fout << std::setw(10) << bdm_id
                     << ' ' << std::setw(14) << std::scientific << RAD
                     << ' ' << std::setw(14) << std::scientific << qi.XYZ[qp](0)
                     << ' ' << std::setw(14) << std::scientific << qi.XYZ[qp](1)
                     << ' ' << std::setw(14) << std::scientific << qi.XYZ[qp](2)
                     << std::endl;
                //
                std::cout << "\t\tbdm_id = " << bdm_id << " - "
                          << "\tnpheno_BDM = " << npheno_BDM << "\tnbchem_BDM = " << nbchem_BDM << "\tnbmech_BDM = " << nbmech_BDM << std::endl;
                #endif
                //==================================
                //
                // create and initialize the BioDynaMo domain
                BDM_Domain rve( bdm_id, { qi.XYZ[qp](0), qi.XYZ[qp](1), qi.XYZ[qp](2) },
                                npheno_BDM, nbchem_BDM, nbmech_BDM );
                //
                rve.oxyg      = qi.biochem[qp][0];
                rve.oxyg_grad = { qi.grad_biochem[qp][0](0) ,
                                  qi.grad_biochem[qp][0](1) ,
                                  qi.grad_biochem[qp][0](2) };
                rve.ecm      = qi.biochem[qp][12];
                rve.ecm_grad = { qi.grad_biochem[qp][12](0) ,
                                 qi.grad_biochem[qp][12](1) ,
                                 qi.grad_biochem[qp][12](2) };
                for (unsigned int i=0; i<nbchem_BDM; i++) {
                    rve.biochem[i]      = qi.biochem[qp][i];
                    rve.biochem_grad[i] = { qi.grad_biochem[qp][i](0) ,
                                            qi.grad_biochem[qp][i](1) ,
                                            qi.grad_biochem[qp][i](2) };
                }
                for (unsigned int i=0; i<nbmech_BDM; i++) {
                    rve.biomech[i]      = qi.auxiliary[qp][i];
                    rve.biomech_grad[i] = { qi.grad_auxiliary[qp][i](0) ,
                                            qi.grad_auxiliary[qp][i](1) ,
                                            qi.grad_auxiliary[qp][i](2) };
                }
                rve.interstitial_fluid_velocity = { qi.fluid_velocity[qp](0) ,
                                                    qi.fluid_velocity[qp](1) ,
                                                    qi.fluid_velocity[qp](2) };
                rve.sim = new bdm::Simulation<>(sname, set_bdm_params);
                //
                //==================================
                //
                n_cells_per_phenotype[0] = static_cast<int>( qi.biochem[qp][1] );
                n_cells_per_phenotype[1] = static_cast<int>( qi.biochem[qp][2] );
                n_cells_per_phenotype[2] = static_cast<int>( qi.biochem[qp][3] );
                n_cells_per_phenotype[3] = static_cast<int>( qi.biochem[qp][4] );
                n_cells_per_phenotype[4] = static_cast<int>( qi.biochem[qp][5] );
                n_cells_per_phenotype[5] = static_cast<int>( qi.biochem[qp][6] );
                n_cells_per_phenotype[6] = static_cast<int>( qi.biochem[qp][7] );
                //
                n_cells_per_phenotype_list.push_back( n_cells_per_phenotype );
                //
                //==================================
                //
                // insert this representative volume element (otherwise referred as
                // BioDynaMo domain of analysis) in the corresponding container
                rve_list_2_add.push_back( rve );
                //
                //==================================
                //
                // increment this index...
                ++bdm_id;
                // ...end of quadrature points loop
            }
            //
            // now store the list of representative volume elements in the global container
            this->elem_rve_local.insert( std::make_pair(elem_id, rve_list_2_add) );
            //
            std::vector<BDM_Domain>& rve_list = this->elem_rve_local.find(elem_id)->second;
            //
            // initialise data for all BioDynaMo simulations, i.e. per each BDM_Domain
            const unsigned int n_rve = rve_list.size();
            if ( n_rve != n_cells_per_phenotype_list.size() ) libmesh_error();
            #ifdef BDM_VERBOSE
            std::cout << " DomainData::initialise RVEs - start " << std::endl;
            #endif
            for (unsigned int j=0; j<n_rve; j++) {
              bdm::init_cells(rve_list[j], params, n_cells_per_phenotype_list[j]);
              //
              #ifdef BDM_VERBOSE
              std::cout << " RVE: " << rve_list[j].id() << std::flush;
              for (unsigned int t=0; t<rve_list[j].n_phenotypes(); t++) {
                std::cout << "\n ===" << std::flush;
                std::cout << " type: " << t << std::flush;
                std::cout << "  ---" << std::flush;
                std::cout << " number of cells: " << rve_list[j].cells_population[t] << std::flush;
                std::cout << " ---" << std::flush;
              }
              std::cout << std::endl;
              #endif
              //
              // set up the seed for BioDynaMo's random generator seed
              auto* random = rve_list[j].sim->GetRandom();
              random->SetSeed( params.get<bool>("BioDynaMo/random_generator_fix_seed")
                             ? 2444 : uniform_distro(1, 9999) );
              // ...end of RVEs loop
            }
            #ifdef BDM_VERBOSE
            std::cout << " DomainData::initialise RVEs - end " << std::endl;
            #endif
            // ...end of (local) finite elements loop
        }
        //
        // exception is caught for the solver utilized
    } else libmesh_error();
    //
    bdm_log.pop("DomainData::init_bdm");
}
//=================================================================================================
void DomainData::run_bdm () {
    bdm_log.push("DomainData::run_bdm");
    //
    Parameters& params = this->_equation_systems.parameters;
    //
    // check we if plan to continue computations or exit normally
    if ( ! params.get<bool>("reaction-diffusion solver: BioDynaMo") ) return;
    //
    const unsigned int nvars_RDS   = params.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = params.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       //ninvars_NCS = params.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_AUX   = params.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    //
    const unsigned int npheno_BDM = params.get<unsigned int>("BioDynaMo/number_of_phenotypes"),
                       nbchem_BDM = params.get<unsigned int>("BioDynaMo/number_of_biochemical_parameters"),
                       nbmech_BDM = params.get<unsigned int>("BioDynaMo/number_of_biomechanical_parameters");
    //
    // initialize the BioDynaMo data wrt the solver employed
    if ( FEM_BIOCHEMICS_SOLVER == params.get<ProblemCase>("solver") &&
         nvars_RDS == 13 && ndcvars_RDS == 1 && nvars_AUX == 2 )
    {
        // sanity checks
        if (nvars_RDS!=nbchem_BDM) libmesh_error();
        if (nvars_AUX!=nbmech_BDM) libmesh_error();
        if (7!=npheno_BDM) libmesh_error();
        //
        // iterate for all (local) finite elements
        for (auto citer=this->elem_qi_local.begin(); citer!=this->elem_qi_local.end(); citer++) {
            const dof_id_type elem_id = citer->first;
            const QuadratureInfo& qi = citer->second;
            const unsigned int n_qp = qi.n_points();
            //
            std::vector<BDM_Domain>& rve_list = this->elem_rve_local.find(elem_id)->second;
            //
            // iterate for all (local) quadrature points
            for (unsigned int qp=0; qp<n_qp; qp++) {
                // now update all data relevant to this RVE
                BDM_Domain& rve = rve_list[qp];
                //
                rve.oxyg      = qi.biochem[qp][0];
                rve.oxyg_grad = { qi.grad_biochem[qp][0](0) ,
                                  qi.grad_biochem[qp][0](1) ,
                                  qi.grad_biochem[qp][0](2) };
                rve.ecm      = qi.biochem[qp][12];
                rve.ecm_grad = { qi.grad_biochem[qp][12](0) ,
                                 qi.grad_biochem[qp][12](1) ,
                                 qi.grad_biochem[qp][12](2) };
                for (unsigned int i=0; i<nbchem_BDM; i++) {
                    rve.biochem[i]      = qi.biochem[qp][i];
                    rve.biochem_grad[i] = { qi.grad_biochem[qp][i](0) ,
                                            qi.grad_biochem[qp][i](1) ,
                                            qi.grad_biochem[qp][i](2) };
                }
                for (unsigned int i=0; i<nbmech_BDM; i++) {
                    rve.biomech[i]      = qi.auxiliary[qp][i];
                    rve.biomech_grad[i] = { qi.grad_auxiliary[qp][i](0) ,
                                            qi.grad_auxiliary[qp][i](1) ,
                                            qi.grad_auxiliary[qp][i](2) };
                }
                rve.interstitial_fluid_velocity = { qi.fluid_velocity[qp](0) ,
                                                    qi.fluid_velocity[qp](1) ,
                                                    qi.fluid_velocity[qp](2) };
                // ...end of quadrature points loop
            }
            // ...end of (local) finite elements loop
        }
        //
        // exception is caught for the solver utilized
    } else libmesh_error();
    //
    // simulate for all BioDynaMo data
    for ( std::map<dof_id_type, std::vector<BDM_Domain>>::iterator
          citer=this->elem_rve_local.begin(); citer!=this->elem_rve_local.end(); citer++ )
    {
        #ifdef BDM_VERBOSE
        std::cout << " FE-ID " << citer->first << std::endl;
        #endif
        std::vector<BDM_Domain>& rve_list = citer->second;
        //
        // run BioDynaMo simulations for all RVEs
        bdm::simulate_bdm_domains(rve_list, params);
        //
        #ifdef BDM_VERBOSE
        for (unsigned int j=0; j<rve_list.size(); j++) {
          std::cout << " RVE: " << rve_list[j].id() << std::flush;
          for (unsigned int t=0; t<rve_list[j].n_phenotypes(); t++) {
            std::cout << "\n ===" << std::flush;
            std::cout << " type: " << t << std::flush;
            std::cout << "  ---" << std::flush;
            std::cout << " number of cells: " << rve_list[j].cells_population[t] << std::flush;
            std::cout << "  ---" << std::flush;
            std::cout << " number of cells increment: " << rve_list[j].cells_population_increment[t] << std::flush;
            std::cout << " ---" << std::flush;
            std::cout << " cell dynamics: " << std::flush;
            for (unsigned int d=0; d<3; d++)
              std::cout << " " << rve_list[j].cell_dynamics[t][d] << std::flush;
            std::cout << " ---" << std::flush;
          }
          std::cout << std::endl;
        }
        #endif
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    bdm_log.pop("DomainData::run_bdm");
}
//=================================================================================================
void DomainData::clear_bdm () {
    bdm_log.push("DomainData::clear_bdm");
    //
    const Parameters& params = this->_equation_systems.parameters;
    //
    // check we if plan to continue computations or exit normally
    if ( ! params.get<bool>("reaction-diffusion solver: BioDynaMo") ) return;
    //
    // clear this global container for ever
    for ( std::map<dof_id_type, std::vector<BDM_Domain>>::iterator
          citer=this->elem_rve_local.begin(); citer!=this->elem_rve_local.end(); citer++ )
    {
        std::vector<BDM_Domain>& rve_list = citer->second;
        //
        // finally do memory clean-up
        bdm::clean_bdm_domains(rve_list);
    }
    this->elem_rve_local.clear();
    //
    bdm_log.pop("DomainData::clear_bdm");
}
//=================================================================================================
void DomainData::write_bdm (const char* fname) const {
    bdm_log.push("DomainData::write_bdm");
    //
    const Parameters& params = this->_equation_systems.parameters;
    //
    const MeshBase::const_element_iterator begin_el = this->_equation_systems.get_mesh().active_elements_begin();
    const MeshBase::const_element_iterator end_el   = this->_equation_systems.get_mesh().active_elements_end();
    //
    std::ofstream fout;
    //
    if ( libMesh::global_processor_id() == 0 ) {
        // only the 0-ranked processor prints this file
        fout.open(params.get<std::string>("results directory: paraview")+"/"+std::string(fname)+".pvtu");
        // print out the parallel XML-formatted VTU file
        fout << "<?xml version=\"1.0\"?>" << std::endl;
        fout << "<VTKFile type=\"PUnstructuredGrid\" version=\"0.1\" byte_order=\"LittleEndian\" >" << std::endl;
        fout << "  <PUnstructuredGrid GhostLevel=\"0\">" << std::endl;
        fout << "    <PPoints>" << std::endl;
        fout << "      <PDataArray type=\"Float64\" NumberOfComponents=\"3\"/>" << std::endl;
        fout << "    </PPoints>" << std::endl;
        fout << "    <PPointData>" << std::endl;
        fout << "      <PDataArray type=\"Float64\" Name=\"ORIGIN\" NumberOfComponents=\"3\"/>" << std::endl;
        fout << "      <PDataArray type=\"Int32\"   Name=\"BDM_ID\"/>" << std::endl;
        fout << "      <PDataArray type=\"Float64\" Name=\"BDM_diameter\"/>" << std::endl;
        fout << "      <PDataArray type=\"Int32\"   Name=\"BDM_phenotype\"/>" << std::endl;
        fout << "      <PDataArray type=\"Int32\"   Name=\"BDM_age\"/>" << std::endl;
        fout << "      <PDataArray type=\"Float64\" Name=\"BDM_convection\" NumberOfComponents=\"3\"/>" << std::endl;
        fout << "      <PDataArray type=\"Float64\" Name=\"BDM_oxygen\"/>" << std::endl;
        fout << "      <PDataArray type=\"Float64\" Name=\"BDM_oxygen_grad\" NumberOfComponents=\"3\"/>" << std::endl;
        fout << "      <PDataArray type=\"Float64\" Name=\"BDM_ecm\"/>" << std::endl;
        fout << "      <PDataArray type=\"Float64\" Name=\"BDM_ecm_grad\" NumberOfComponents=\"3\"/>" << std::endl;
        fout << "      <PDataArray type=\"Float64\" Name=\"BDM_cc_grow\"/>" << std::endl;
        fout << "      <PDataArray type=\"Float64\" Name=\"BDM_cc_divide\"/>" << std::endl;
        fout << "      <PDataArray type=\"Float64\" Name=\"BDM_cc_migrate\"/>" << std::endl;
        fout << "      <PDataArray type=\"Float64\" Name=\"BDM_cc_transform\"/>" << std::endl;
        fout << "    </PPointData>" << std::endl;
        fout << "    <PCellData>" << std::endl;
        fout << "      <PDataArray type=\"Int32\"   Name=\"FE_ID\"/>" << std::endl;
        fout << "    </PCellData>" << std::endl;
        for (MeshBase::const_element_iterator ci=begin_el; ci!=end_el; ci++) {
            const dof_id_type elem_ID = (*ci)->id();
            //
            fout << "    <Piece Source=\"BDM/" << fname << elem_ID << ".vtu\"/>" << std::endl;
        }
        fout << "  </PUnstructuredGrid>" << std::endl;
        fout << "</VTKFile>" << std::endl;
        // now close the file safely
        fout.close();
    }
    //
    // iterate for all BioDynaMo data
    for ( std::map<dof_id_type, std::vector<BDM_Domain>>::const_iterator
          citer=this->elem_rve_local.begin(); citer!=this->elem_rve_local.end(); citer++ )
    {
        const dof_id_type elem_ID = citer->first;
        const std::vector<BDM_Domain>& rve_list = citer->second;
        //
        const std::string vtu_fname = params.get<std::string>("results directory: paraview")
                                    + "/BDM/" + std::string(fname) + std::to_string(elem_ID) + ".vtu";
        // open the VTU file to store results
        fout.open(vtu_fname);
        //
        unsigned int n_points = 0;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            // increment this index wrt the total number of cells for this RVE
            n_points += all_cells->size();
        }
        unsigned int n_cells = rve_list.size();
        //
        fout << "<?xml version=\"1.0\"?>" << std::endl;
        fout << "<VTKFile type=\"UnstructuredGrid\" version=\"0.1\" byte_order=\"LittleEndian\">" << std::endl;
        fout << "  <UnstructuredGrid>" << std::endl;
        fout << "    <Piece NumberOfPoints=\"" << n_points << "\" NumberOfCells=\"" << n_cells << "\">" << std::endl;
        fout << "      <Points>" << std::endl;
        fout << "        <DataArray type=\"Float64\" NumberOfComponents=\"3\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            for (unsigned int j=0; j<all_cells->size(); j++) {
                const std::array<double, 3> position = (*all_cells)[j].GetPosition();
                fout << ' ' << position[0] << ' ' << position[1] << ' ' << position[2];
            }
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "      </Points>" << std::endl;
        fout << "      <PointData>" << std::endl;
        fout << "        <DataArray type=\"Float64\" Name=\"ORIGIN\" NumberOfComponents=\"3\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            const std::array<double, 3> origin = rve_list[i].origin();
            for (unsigned int j=0; j<all_cells->size(); j++) {
                fout << ' ' << origin[0] << ' ' << origin[1] << ' ' << origin[2];
            }
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "        <DataArray type=\"Int32\" Name=\"BDM_ID\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            for (unsigned int j=0; j<all_cells->size(); j++) {
                fout << ' ' << rve_list[i].id();
            }
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "        <DataArray type=\"Float64\" Name=\"BDM_diameter\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            for (unsigned int j=0; j<all_cells->size(); j++) {
                fout << ' ' << (*all_cells)[j].GetDiameter();
            }
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "        <DataArray type=\"Int32\" Name=\"BDM_phenotype\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            for (unsigned int j=0; j<all_cells->size(); j++) {
                fout << ' ' << (*all_cells)[j].GetPhenotype();
            }
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "        <DataArray type=\"Int32\" Name=\"BDM_age\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            for (unsigned int j=0; j<all_cells->size(); j++) {
                fout << ' ' << (*all_cells)[j].GetAge();
            }
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "        <DataArray type=\"Float64\" Name=\"BDM_convection\" NumberOfComponents=\"3\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            for (unsigned int j=0; j<all_cells->size(); j++) {
                const std::array<double, 3> convection = (*all_cells)[j].GetConvection();
                fout << ' ' << convection[0] << ' ' << convection[1] << ' ' << convection[2];
            }
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "        <DataArray type=\"Float64\" Name=\"BDM_oxygen\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            for (unsigned int j=0; j<all_cells->size(); j++) {
                fout << ' ' << (*all_cells)[j].GetOxygen();
            }
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "        <DataArray type=\"Float64\" Name=\"BDM_oxygen_grad\" NumberOfComponents=\"3\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            for (unsigned int j=0; j<all_cells->size(); j++) {
                const std::array<double, 3> oxygen_grad = (*all_cells)[j].GetOxygenGrad();
                fout << ' ' << oxygen_grad[0] << ' ' << oxygen_grad[1] << ' ' << oxygen_grad[2];
            }
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "        <DataArray type=\"Float64\" Name=\"BDM_ecm\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            for (unsigned int j=0; j<all_cells->size(); j++) {
                fout << ' ' << (*all_cells)[j].GetECM();
            }
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "        <DataArray type=\"Float64\" Name=\"BDM_ecm_grad\" NumberOfComponents=\"3\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            for (unsigned int j=0; j<all_cells->size(); j++) {
                const std::array<double, 3> ecm_grad = (*all_cells)[j].GetECMGrad();
                fout << ' ' << ecm_grad[0] << ' ' << ecm_grad[1] << ' ' << ecm_grad[2];
            }
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "        <DataArray type=\"Float64\" Name=\"BDM_cc_grow\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            for (unsigned int j=0; j<all_cells->size(); j++) {
                fout << ' ' << (*all_cells)[j].GetChemicalCue_Grow();
            }
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "        <DataArray type=\"Float64\" Name=\"BDM_cc_divide\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            for (unsigned int j=0; j<all_cells->size(); j++) {
                fout << ' ' << (*all_cells)[j].GetChemicalCue_Divide();
            }
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "        <DataArray type=\"Float64\" Name=\"BDM_cc_migrate\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            for (unsigned int j=0; j<all_cells->size(); j++) {
                fout << ' ' << (*all_cells)[j].GetChemicalCue_Migrate();
            }
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "        <DataArray type=\"Float64\" Name=\"BDM_cc_transform\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
            auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
            for (unsigned int j=0; j<all_cells->size(); j++) {
                fout << ' ' << (*all_cells)[j].GetChemicalCue_Transform();
            }
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "      </PointData>" << std::endl;
        fout << "      <Cells>" << std::endl;
        fout << "        <DataArray type=\"Int32\" Name=\"connectivity\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        for (unsigned int ipnt=0; ipnt<n_points; ipnt++) {
          fout << ' ' << ipnt;
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "        <DataArray type=\"Int32\" Name=\"offsets\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        int offset = 0;
        for (unsigned int i=0; i<rve_list.size(); i++) {
          auto all_cells = rve_list[i].sim->GetResourceManager()->template Get<bdm::MyCell>();
          offset += all_cells->size();
          fout << ' ' << offset;
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "        <DataArray type=\"Int32\" Name=\"types\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
          fout << ' ' << 2;
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "      </Cells>" << std::endl;
        fout << "      <CellData>" << std::endl;
        fout << "        <DataArray type=\"Int32\" Name=\"FE_ID\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        for (unsigned int i=0; i<rve_list.size(); i++) {
          fout << ' ' << elem_ID;
        }
        fout << std::endl;
        fout << "        </DataArray>" << std::endl;
        fout << "      </CellData>" << std::endl;
        fout << "    </Piece>" << std::endl;
        fout << "  </UnstructuredGrid>" << std::endl;
        fout << "</VTKFile>" << std::endl;
        //
        // now close the file safely
        fout.close();
        // ...end of (local) finite elements loop
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    bdm_log.pop("DomainData::write_bdm");
}
//=================================================================================================
#undef SET_BIODYNAMO_PARAMETERS_
//=================================================================================================
#undef BDM_VERBOSE
//=================================================================================================
//
//
//
//=================================================================================================
#include "cancer_growth_test.h"
//=================================================================================================
