/*************************************************************************************************

    Copyright (C) 2009-2019    Jean de Montigny (jean.demontigny@hotmail.fr)
                               Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#ifndef __MYCELL_H__
#define __MYCELL_H__
//=================================================================================================
#include "feb3/utils.h"
#include "libmesh/parameters.h"
#include "biodynamo.h"
//=================================================================================================
#ifdef _SOME_VARIABILITY_
    #error "an unexpected error occurred!"
#else
    #define _SOME_VARIABILITY_  random->Uniform(0.98, 1.02)
#endif
//=================================================================================================
namespace bdm {
//=================================================================================================
BDM_SIM_OBJECT(MyCell, Cell) {
  //
  BDM_SIM_OBJECT_HEADER(MyCell, Cell, 1, params_pt_, phenotype_, phenotype2transform_, age_,
                        convection_, oxygen_, oxygen_grad_, ecm_, ecm_grad_,
                        cc_grow_, cc_divide_, cc_migrate_, cc_transform_);
  //
  public:
    // constructors, copy constructor
    MyCellExt () {}
    MyCellExt (const std::array<double, 3>& position) : Base(position) {}
    template <typename TMother>
    MyCellExt (const CellDivisionEvent& event, TMother* mother) : Base(event, mother)
    { // if cell divides, then daughter copies data from mother cell
      params_pt_[kIdx] = mother->params_pt_[mother->kIdx];
      phenotype_[kIdx] = mother->phenotype_[mother->kIdx];
      phenotype2transform_[kIdx] = mother->phenotype2transform_[mother->kIdx];
      age_[kIdx] = 0; // does not copy age from mother cell
      convection_[kIdx] = mother->convection_[mother->kIdx];
      oxygen_[kIdx] = mother->oxygen_[mother->kIdx];
      oxygen_grad_[kIdx] = mother->oxygen_grad_[mother->kIdx];
      ecm_[kIdx] = mother->ecm_[mother->kIdx];
      ecm_grad_[kIdx] = mother->ecm_grad_[mother->kIdx];
      cc_grow_[kIdx]    = mother->cc_grow_[mother->kIdx];
      cc_divide_[kIdx]  = mother->cc_divide_[mother->kIdx];
      cc_migrate_[kIdx] = mother->cc_migrate_[mother->kIdx];
      cc_transform_[kIdx] = mother->cc_transform_[mother->kIdx];
    }
    // if cell divides, then daughter keeps the state from its mother
    template <typename TDaughter>
    void EventHandler (const CellDivisionEvent& event, TDaughter* daughter) { Base::EventHandler(event, daughter); }

    inline
    void SetParametersPointer (const libMesh::Parameters* p) { params_pt_[kIdx] = p; }
    inline
    const libMesh::Parameters* GetParametersPointer () const { return params_pt_[kIdx]; }

    inline
    void SetPhenotype (int p) { phenotype_[kIdx] = p; }
    inline
    int GetPhenotype () const { return phenotype_[kIdx]; }

    inline
    void SetPhenotype2Transform (int p) { phenotype2transform_[kIdx] = p; }
    inline
    int GetPhenotype2Transform () const { return phenotype2transform_[kIdx]; }

    inline
    void SetAge (int a = 0) { age_[kIdx] = a; }
    inline
    int GetAge () const { return age_[kIdx]; }
    inline
    void IncrementAge () { age_[kIdx] += 1; }

    inline
    void SetConvection (const RealVector3d& v) { convection_[kIdx] = v; }
    inline
    RealVector3d GetConvection () const { return convection_[kIdx]; }
    inline
    double GetConvection (unsigned int i) const { return convection_[kIdx][i]; }

    inline
    void SetOxygen (double o) { oxygen_[kIdx] = o; }
    inline
    double GetOxygen () const { return oxygen_[kIdx]; }

    inline
    void SetOxygenGrad (RealVector3d og) { oxygen_grad_[kIdx] = og; }
    inline
    RealVector3d GetOxygenGrad () const { return oxygen_grad_[kIdx]; }
    inline
    double GetOxygenGrad (unsigned int i) const { return oxygen_grad_[kIdx][i]; }

    inline
    void SetECM (double e) { ecm_[kIdx] = e; }
    inline
    double GetECM () const { return ecm_[kIdx]; }

    inline
    void SetECMGrad (RealVector3d eg) { ecm_grad_[kIdx] = eg; }
    inline
    RealVector3d GetECMGrad () const { return ecm_grad_[kIdx]; }
    inline
    double GetECMGrad (unsigned int i) const { return ecm_grad_[kIdx][i]; }

    inline
    void SetChemicalCue_Grow (double cc) { cc_grow_[kIdx] = cc; }
    inline
    double GetChemicalCue_Grow () const { return cc_grow_[kIdx]; }
    inline
    void SetChemicalCue_Divide (double cc) { cc_divide_[kIdx] = cc; }
    inline
    double GetChemicalCue_Divide () const { return cc_divide_[kIdx]; }
    inline
    void SetChemicalCue_Migrate (double cc) { cc_migrate_[kIdx] = cc; }
    inline
    double GetChemicalCue_Migrate () const { return cc_migrate_[kIdx]; }
    inline
    void SetChemicalCue_Transform (double cc) { cc_transform_[kIdx] = cc; }
    inline
    double GetChemicalCue_Transform () const { return cc_transform_[kIdx]; }
  //
  private:
    vec<const libMesh::Parameters*> params_pt_;
    vec<int> phenotype_, phenotype2transform_, age_;
    vec<RealVector3d> convection_; // interstitial (bio)fluid velocity
    vec<double> oxygen_; // O2 saturation level (must be within [0, 1]) and gradient below
    vec<RealVector3d> oxygen_grad_;
    vec<double> ecm_; // ECM density (must be within [0, 1]) and gradient below
    vec<RealVector3d> ecm_grad_;
    // chemical cues regulating cell growth, division, migration and transformation
    vec<double> cc_grow_, cc_divide_, cc_migrate_, cc_transform_;
  //
};
//-------------------------------------------------------------------------------------------------
struct MyCellBiologyModule : public BaseBiologyModule {
public:
    MyCellBiologyModule () : BaseBiologyModule(gAllEventIds) {}
    //
    template <typename TEvent, typename TBm>
    MyCellBiologyModule (const TEvent& event, TBm* other, uint64_t new_oid = 0) : BaseBiologyModule(event, other, new_oid) {}
    //
    template <typename T, typename TSimulation = Simulation<>>
    void Run (T* cell)
    {
      auto* sim = TSimulation::GetActive();
      auto* random = sim->GetRandom();
      // access the model parameters
      const libMesh::Parameters* pp_ = cell->GetParametersPointer();

      // obtain the oxygen (saturation) level
      const Real O2_level = cell->GetOxygen();
      // obtain the extracelular matrix (normalized) density
      const Real ECM_density = apply_bounds(0.0, cell->GetECM(), 1.0);
      //
      const Real CC_Grow_level    = cell->GetChemicalCue_Grow(),
                 CC_Divide_level  = cell->GetChemicalCue_Divide(),
                 CC_Migrate_level = cell->GetChemicalCue_Migrate(),
                 CC_Transf_level  = cell->GetChemicalCue_Transform();
      // volume fractions...
      const Real VF_ECM = ECM_density,
                 VF_IF  = pp_->get<double>("BioDynaMo/interstitial_fluid_volume_fraction"),
                 VF_Cell= pp_->get<double>("BioDynaMo/cells_volume_fraction");
      //
      // convective transport (due to interstitial fluid flow)
      // note that the vector is in scaled form
      RealVector3d flow = {0.0, 0.0, 0.0};
      { //******************
        const Real dt_ = pp_->get<double>("BioDynaMo/simulation_time_step"),
                   M2m_ = pp_->get<Real>("BioDynaMo/macro2micro/space_scaling");
        //
        // form the vector now...
        flow = Math::ScalarMult(M2m_*dt_, cell->GetConvection());
      } //******************
      //
      // by definition "0" phenotype desigates the cell is in necrotic
      // state regardless its original condition or phenotype...
      if ( 0 == cell->GetPhenotype() ) {
        //
        const Real deathProbability = pp_->get<Real>("BioDynaMo/cell_0/death_probability");
        if ( random->Uniform(0.0, 1.0) < deathProbability ) {
          // now delete this cell
          cell->RemoveFromSimulation();
          // ...and that's it!
          return;
        }
        // cell moves
        cell->UpdatePosition(flow);
        // ...and that's it!
        return;
      }
      //
      // cell gets older
      cell->IncrementAge();
      //
      const unsigned int currentPhenotype = cell->GetPhenotype();
      if (0 == currentPhenotype) libmesh_error();
      //
      const Real maxCellSize = _SOME_VARIABILITY_
                             * pp_->get<Real>("BioDynaMo/cell_"+std::to_string(currentPhenotype)+"/diameter/max");
      //
      const Real growthSpeed[] = { _SOME_VARIABILITY_ * ( currentPhenotype % 2 // below for normoxia
                                   ? pp_->get<Real>("BioDynaMo/cell_"+std::to_string(currentPhenotype  )+"/growth_speed")
                                   : pp_->get<Real>("BioDynaMo/cell_"+std::to_string(currentPhenotype-1)+"/growth_speed") ) ,
                                   _SOME_VARIABILITY_ * ( currentPhenotype % 2 // below for hypoxia
                                   ? pp_->get<Real>("BioDynaMo/cell_"+std::to_string(currentPhenotype+1)+"/growth_speed")
                                   : pp_->get<Real>("BioDynaMo/cell_"+std::to_string(currentPhenotype  )+"/growth_speed") ) };
      //
      const Real divideProbability[] = { ( currentPhenotype % 2 // below for normoxia
                                         ? pp_->get<Real>("BioDynaMo/cell_"+std::to_string(currentPhenotype  )+"/division_probability")
                                         : pp_->get<Real>("BioDynaMo/cell_"+std::to_string(currentPhenotype-1)+"/division_probability") ) ,
                                         ( currentPhenotype % 2 // below for hypoxia
                                         ? pp_->get<Real>("BioDynaMo/cell_"+std::to_string(currentPhenotype+1)+"/division_probability")
                                         : pp_->get<Real>("BioDynaMo/cell_"+std::to_string(currentPhenotype  )+"/division_probability") ) };
      //
      const Real deathProbability[] = { ( currentPhenotype % 2 // below for normoxia
                                        ? pp_->get<Real>("BioDynaMo/cell_"+std::to_string(currentPhenotype  )+"/death_probability")
                                        : pp_->get<Real>("BioDynaMo/cell_"+std::to_string(currentPhenotype-1)+"/death_probability") ) ,
                                        ( currentPhenotype % 2 // below for hypoxia
                                        ? pp_->get<Real>("BioDynaMo/cell_"+std::to_string(currentPhenotype+1)+"/death_probability")
                                        : pp_->get<Real>("BioDynaMo/cell_"+std::to_string(currentPhenotype  )+"/death_probability") ) };
      //
      const int oldEnough2Die[] = { ( currentPhenotype % 2 // below for normoxia
                                    ? pp_->get<int>("BioDynaMo/cell_"+std::to_string(currentPhenotype  )+"/old_enough_to_die")
                                    : pp_->get<int>("BioDynaMo/cell_"+std::to_string(currentPhenotype-1)+"/old_enough_to_die") ) ,
                                    ( currentPhenotype % 2 // below for hypoxia
                                    ? pp_->get<int>("BioDynaMo/cell_"+std::to_string(currentPhenotype+1)+"/old_enough_to_die")
                                    : pp_->get<int>("BioDynaMo/cell_"+std::to_string(currentPhenotype  )+"/old_enough_to_die") ) };
      //
      const Real oxygen2Hypoxia = _SOME_VARIABILITY_
                                * pp_->get<Real>("BioDynaMo/cell_"+std::to_string(currentPhenotype)+"/O2_limit/to_hypoxia"),
                 oxygen2Necrosi = _SOME_VARIABILITY_
                                * pp_->get<Real>("BioDynaMo/cell_"+std::to_string(currentPhenotype)+"/O2_limit/to_necrosis");
      //
      // based on the oxygen level confirm or update the state of the cell
      unsigned int updatedPhenotype = 0;
      if ( currentPhenotype % 2 ) {
        // currently, the cell is in normoxic state, but...
        if      (O2_level > oxygen2Hypoxia) updatedPhenotype = currentPhenotype;
        else if (O2_level > oxygen2Necrosi) updatedPhenotype = currentPhenotype+1;
        else                                updatedPhenotype = 0;
      } else {
        // currently, the cell is in hypoxic state, but...
        if      (O2_level > oxygen2Hypoxia) updatedPhenotype = currentPhenotype-1;
        else if (O2_level > oxygen2Necrosi) updatedPhenotype = currentPhenotype;
        else                                updatedPhenotype = 0;
      }
      //
      // update now the cell phenotyupe
      cell->SetPhenotype(updatedPhenotype);
      //
      // check if the cell has transioned into necrosis
      if ( 0 == updatedPhenotype ) {
        // if oxygen level for necrosis has positive value then
        // the cell may get into a necrotic state, otherwise it simply dies
        if ( oxygen2Necrosi > 0.0 ) {
          // cell moves
          cell->UpdatePosition(flow);
        } else {
          // do delete this cell
          cell->RemoveFromSimulation();
        }
        // ...and that's it!
        return;
      }
      //
      const unsigned int updatedPhenotypeIndex = ( updatedPhenotype % 2 ? 0 : 1 );
      //
      // cell dies when it reaches a certain age
      if ( cell->GetAge() >= oldEnough2Die[updatedPhenotypeIndex] )
      if ( random->Uniform(0.0, 1.0) < deathProbability[updatedPhenotypeIndex] ) {
        // now delete this cell
        cell->RemoveFromSimulation();
        // ...and that's it!
        return;
      }
      //
      const RealVector3d grad_O2 = cell->GetOxygenGrad();
      const Real grad_O2_magn = sqrt(pow2(grad_O2[0])+pow2(grad_O2[1])+pow2(grad_O2[2]));
      //
      // evaluate the vector for the O2-dependent chemotaxis of the cell
      std::array<double, 3> chemotaxis_O2 = {0.0, 0.0, 0.0};
      if ( pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/migration_wrt_oxygen_grad/threshold")
           <= grad_O2_magn )
      { //******************
        const Real min_ = pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/migration_wrt_oxygen_grad/min"),
                   max_ = pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/migration_wrt_oxygen_grad/max");
        //
        // form the vector now...
        chemotaxis_O2 = { (grad_O2[0]/grad_O2_magn)*random->Uniform(min_, max_) ,
                          (grad_O2[1]/grad_O2_magn)*random->Uniform(min_, max_) ,
                          (grad_O2[2]/grad_O2_magn)*random->Uniform(min_, max_) };
      } //******************
      //
      std::array<double, 3> chemotaxis_CC = {0.0, 0.0, 0.0};
      if ( pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/chemical_cue_for_migration/threshold") > 0 &&
           pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/chemical_cue_for_migration/threshold") < CC_Migrate_level )
      { //******************
        const Real min_ = pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/migration_wrt_chemical_cue/min"),
                   max_ = pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/migration_wrt_chemical_cue/max");
        //
        // form the vector now...
        chemotaxis_CC = { random->Uniform(min_, max_) ,
                          random->Uniform(min_, max_) ,
                          random->Uniform(min_, max_) };
      } //******************
      //
      // evaluate the vector for the Brownian motion of the cell
      RealVector3d diffuse = {0.0, 0.0, 0.0};
      { //******************
        const Real min_ = pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/migration_wrt_Brownian/min"),
                   max_ = pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/migration_wrt_Brownian/max");
        //
        // form the vector now...
        diffuse = { random->Uniform(min_, max_) ,
                    random->Uniform(min_, max_) ,
                    random->Uniform(min_, max_) };
      } //******************
      //
      const RealVector3d grad_ECM = cell->GetECMGrad();
      const Real grad_ECM_magn = sqrt(pow2(grad_ECM[0])+pow2(grad_ECM[1])+pow2(grad_ECM[2]));
      //
      // evaluate the vector for the ECM-dependent haptotaxis of the cell
      std::array<double, 3> haptotaxis_ECM = {0.0, 0.0, 0.0};
      if ( ECM_density > 1.0e-6 )
      { //******************
        const Real min_ = pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/migration_wrt_ecm_grad/min"),
                   max_ = pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/migration_wrt_ecm_grad/max");
        const Real c_ecm_ = pow(ECM_density, pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/migration_wrt_ecm_grad/exponent"));
        //
        // form the vector now...
        haptotaxis_ECM = { (-c_ecm_)*(grad_ECM[0]/grad_ECM_magn)*random->Uniform(min_, max_) ,
                           (-c_ecm_)*(grad_ECM[1]/grad_ECM_magn)*random->Uniform(min_, max_) ,
                           (-c_ecm_)*(grad_ECM[2]/grad_ECM_magn)*random->Uniform(min_, max_) };
      } //******************
      //
      //
      //
      if ( VF_ECM + VF_IF + VF_Cell < 1.0 - pp_->get<double>("BioDynaMo/volume_fraction/alarm_margin_to_stop") )
      { //******************
      // if only the sum of volume fractions is below an alarming level then
      // permit division or growth for the cell...
      if ( cell->GetDiameter() < maxCellSize ) {
        // ...cell grows until it reaches a certain size threshold
        double SFactor = 1.0;
        if ( pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/chemical_cue_for_growth/threshold") > 0 &&
             pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/chemical_cue_for_growth/threshold") < CC_Grow_level )
        {
          SFactor =
            pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/chemical_cue_for_growth/scaling");
        }
        //
        cell->ChangeVolume(SFactor*growthSpeed[updatedPhenotypeIndex]);
        //
      } else {
        // ...otherwise it has a chance to divide
        double SFactor = 1.0;
        if ( pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/chemical_cue_for_division/threshold") > 0 &&
             pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/chemical_cue_for_division/threshold") < CC_Divide_level )
        {
          SFactor =
            pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/chemical_cue_for_division/scaling");
        }
        //
        if ( random->Uniform(0.0, 1.0) < SFactor*divideProbability[updatedPhenotypeIndex] ) {
          cell->Divide();
        }
        //
      }
      // ...end of volume fraction if-conditional
      } //******************
      //
      //
      //
      std::array<double, 3> displacement = {0.0, 0.0, 0.0};
      // permit (for normoxic or hypoxic state only) for cell to move/migrate
      // due to interstitial (bio)fluid flow or due to Brownian motion (diffusion)
      displacement = Math::Add(diffuse, displacement);
      displacement = Math::Add(flow, displacement);
      // cell can also sense local gradients of chemical agents (e.g., oxygen)
      // and move based on these (towards higher or lower concentrations), as
      // well as a chemical cue that promotes cell (random) migration
      displacement = Math::Add(chemotaxis_O2, displacement);
      displacement = Math::Add(chemotaxis_CC, displacement);
      // ...and for haptotaxis due to ECM local variations
      displacement = Math::Add(haptotaxis_ECM, displacement);
      // cell moves
      cell->UpdatePosition(displacement);
      //
      //
      //
      if ( pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/chemical_cue_for_transformation/threshold") > 0 &&
           pp_->get<Real>("BioDynaMo/cell_"+std::to_string(updatedPhenotype)+"/chemical_cue_for_transformation/threshold") < CC_Transf_level )
      { //******************
        const int phenotype = cell->GetPhenotype2Transform();
        const int other_phenotype =
          pp_->get<int>("BioDynaMo/cell_"+std::to_string(phenotype)+"/transform_to_another_phenotype");
        const double diameter =
          pp_->get<Real>("BioDynaMo/cell_"+std::to_string(phenotype)+"/diameter/ref");

        // re-define the cell phenotype
        cell->SetPhenotype(phenotype);
        // re-define the cell phenotype this one could transform into
        cell->SetPhenotype2Transform(other_phenotype);
        // re-set the size (average diameter) of the cell
        cell->SetDiameter(diameter);
        // re-set the age of the cell of course (default =0)
        cell->SetAge();
      } //******************
      // end of MyCellBiologyModule::Run
    }
    //
    BDM_CLASS_DEF_NV(MyCellBiologyModule, 1);
};
//=================================================================================================
} // end namespace bdm
//=================================================================================================
#undef _SOME_VARIABILITY_
//=================================================================================================
#endif // __MYCELL_H__
//=================================================================================================
