/*************************************************************************************************

    Copyright (C) 2009-2019    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include <sys/stat.h>
#include "feb3/domain_data.h"
#include "biodynamo.h"
#include "mycell.h"
#include "feb3/bdm_interface.h"
#include "feb3/bdm_migrating_cell.h"
#include "./params.h"
//=================================================================================================
void read_BDM_parameters ( GetPot& in, std::set<std::string>& in_p,
                           libMesh::Parameters& params )
{

  std::string param_name;
  //
  param_name = "BioDynaMo/visualization_output_path";
  params.set<std::string>(param_name) = get_param<std::string>(in, in_p, param_name, "./");
  //
  param_name = "BioDynaMo/micro2macro/space_scaling";
  params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
  // the opposite space scaling is the reciprocal of the above
  param_name = "BioDynaMo/macro2micro/space_scaling";
  params.set<Real>(param_name) = 1.0
                               / params.get<Real>("BioDynaMo/micro2macro/space_scaling");
  //
  param_name = "BioDynaMo/number_of_RVEs";
  params.set<int>(param_name) = get_param<int>(in, in_p, param_name, 0);
  param_name = "BioDynaMo/number_of_iterations";
  params.set<int>(param_name) = get_param<int>(in, in_p, param_name, 0);
  param_name = "BioDynaMo/number_of_time_steps";
  params.set<int>(param_name) = get_param<int>(in, in_p, param_name);
  param_name = "BioDynaMo/number_of_phenotypes";
  params.set<unsigned int>(param_name) = get_param<unsigned int>(in, in_p, param_name);
  param_name = "BioDynaMo/number_of_biochemical_parameters";
  params.set<unsigned int>(param_name) = get_param<unsigned int>(in, in_p, param_name, 1);
  param_name = "BioDynaMo/number_of_biomechanical_parameters";
  params.set<unsigned int>(param_name) = get_param<unsigned int>(in, in_p, param_name, 1);
  //
  param_name = "BioDynaMo/simulation_time_step";
  params.set<double>(param_name) = get_param<double>(in, in_p, param_name, 1.0);
  //
  param_name = "BioDynaMo/simulation_time_step_between_iterations";
  params.set<double>(param_name) = params.get<double>("BioDynaMo/simulation_time_step")
                                 * params.get<int>("BioDynaMo/number_of_time_steps");
  //
  param_name = "BioDynaMo/visualization_output_interval";
  params.set<int>(param_name) = get_param<int>(in, in_p, param_name);
  //
  param_name = "BioDynaMo/bounding_box/size";
  params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
  //
  param_name = "BioDynaMo/bounding_box/min";
  params.set<Real>(param_name) = -0.5*params.get<Real>("BioDynaMo/bounding_box/size");
  param_name = "BioDynaMo/bounding_box/max";
  params.set<Real>(param_name) = +0.5*params.get<Real>("BioDynaMo/bounding_box/size");
  //
  // checks to verify input
  if ( params.get<unsigned int>("BioDynaMo/number_of_phenotypes") == 0 ) libmesh_error();
  if ( params.get<unsigned int>("BioDynaMo/number_of_phenotypes") % 2 == 0 ) libmesh_error();
  if ( params.get<unsigned int>("BioDynaMo/number_of_biochemical_parameters") == 0 ) libmesh_error();
  if ( params.get<unsigned int>("BioDynaMo/number_of_biomechanical_parameters") == 0 ) libmesh_error();
  if ( params.get<int>("BioDynaMo/visualization_output_interval") <= 0 ) libmesh_error();
  if ( params.get<Real>("BioDynaMo/bounding_box/size") <= 0.0 ) libmesh_error();
  //
  param_name = "BioDynaMo/bounding_box/tol/0.1pct";
  params.set<Real>(param_name) = params.get<Real>("BioDynaMo/bounding_box/size") * 0.001;
  param_name = "BioDynaMo/bounding_box/tol/1pct";
  params.set<Real>(param_name) = params.get<Real>("BioDynaMo/bounding_box/size") * 0.01;
  param_name = "BioDynaMo/bounding_box/tol/5pct";
  params.set<Real>(param_name) = params.get<Real>("BioDynaMo/bounding_box/size") * 0.05;
  //
  param_name = "BioDynaMo/bounding_box/face";
  params.set<Real>(param_name) = pow2( params.get<Real>("BioDynaMo/bounding_box/size") );
  //
  param_name = "BioDynaMo/bounding_box/floor";
  params.set<Real>(param_name) = params.get<Real>("BioDynaMo/bounding_box/min")
                               + params.get<Real>("BioDynaMo/bounding_box/tol/1pct");
  param_name = "BioDynaMo/bounding_box/floor (higher)";
  params.set<Real>(param_name) = params.get<Real>("BioDynaMo/bounding_box/min")
                               + params.get<Real>("BioDynaMo/bounding_box/tol/5pct");
  //
  param_name = "BioDynaMo/bounding_box/ceil";
  params.set<Real>(param_name) = params.get<Real>("BioDynaMo/bounding_box/max")
                               - params.get<Real>("BioDynaMo/bounding_box/tol/1pct");
  param_name = "BioDynaMo/bounding_box/ceil (lower)";
  params.set<Real>(param_name) = params.get<Real>("BioDynaMo/bounding_box/max")
                               - params.get<Real>("BioDynaMo/bounding_box/tol/5pct");
  //
  param_name = "BioDynaMo/cell_migration/is_face_model";
  params.set<bool>(param_name) = get_param<bool>(in, in_p, param_name, true);
  //
  param_name = "BioDynaMo/cell_migration/max_speed";
  params.set<double>(param_name) = get_param<double>(in, in_p, param_name, 0.0);
  //
  param_name = "BioDynaMo/random_generator_fix_seed";
  params.set<bool>(param_name) = get_param<bool>(in, in_p, param_name, false);
  //
  param_name = "BioDynaMo/oxygen";
  params.set<double>(param_name) = get_param<double>(in, in_p, param_name, 0.0);
  param_name = "BioDynaMo/oxygen_grad/component_0";
  params.set<double>(param_name) = get_param<double>(in, in_p, param_name, 0.0);
  param_name = "BioDynaMo/oxygen_grad/component_1";
  params.set<double>(param_name) = get_param<double>(in, in_p, param_name, 0.0);
  param_name = "BioDynaMo/oxygen_grad/component_2";
  params.set<double>(param_name) = get_param<double>(in, in_p, param_name, 0.0);
  //
  param_name = "BioDynaMo/interstitial_fluid_velocity/component_0";
  params.set<double>(param_name) = get_param<double>(in, in_p, param_name, 0.0);
  param_name = "BioDynaMo/interstitial_fluid_velocity/component_1";
  params.set<double>(param_name) = get_param<double>(in, in_p, param_name, 0.0);
  param_name = "BioDynaMo/interstitial_fluid_velocity/component_2";
  params.set<double>(param_name) = get_param<double>(in, in_p, param_name, 0.0);
  //
  param_name = "BioDynaMo/interstitial_fluid_volume_fraction";
  params.set<double>(param_name) = get_param<double>(in, in_p, param_name);
  //
  param_name = "BioDynaMo/cells_repulsion_max_displacement";
  params.set<double>(param_name) = get_param<double>(in, in_p, param_name);
  //
  param_name = "BioDynaMo/cells_distance_to_avoid_overlap";
  params.set<double>(param_name) = get_param<double>(in, in_p, param_name);
  //
  param_name = "BioDynaMo/volume_fraction/alarm_margin_to_stop";
  params.set<double>(param_name) = get_param<double>(in, in_p, param_name);
  //
  params.set<std::string>("ID of FEB3 solver to communicate with BioDynaMo") = "NULL";
  //
  // our convention assumes that the "0" phenotype corresponds to the necrotic cells
  // while odd-numbered phenotypes denote the normoxic state of a cell whereas the
  // even-numbered phenotypes denote the hypoxic state of the cell's counterparts!
  // note that for the necrotic cells we may be ignoring readins some of the parameters
  //
  for (unsigned int t=0; t<params.get<unsigned int>("BioDynaMo/number_of_phenotypes"); t++) {
    //
    param_name = "BioDynaMo/cell_" + std::to_string(t) + "/initialisation_pattern";
    if ( get_param<int>(in, in_p, param_name) >= 0 ) {
      // set the initialization pattern of the cells' distribution (in 3D)
      params.set<int>(param_name) = get_param<int>(in, in_p, param_name);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/initial_population";
      params.set<int>(param_name) = get_param<int>(in, in_p, param_name, 0);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/bounding_box/X/min";
      params.set<double>(param_name) = get_param<double>(in, in_p, param_name);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/bounding_box/X/max";
      params.set<double>(param_name) = get_param<double>(in, in_p, param_name);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/bounding_box/Y/min";
      params.set<double>(param_name) = get_param<double>(in, in_p, param_name);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/bounding_box/Y/max";
      params.set<double>(param_name) = get_param<double>(in, in_p, param_name);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/bounding_box/Z/min";
      params.set<double>(param_name) = get_param<double>(in, in_p, param_name);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/bounding_box/Z/max";
      params.set<double>(param_name) = get_param<double>(in, in_p, param_name);
      //
    } else {
      // set the initialization pattern of the cells' distribution (in 3D) to default
      params.set<int>(param_name) = 0;
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/initial_population";
      params.set<int>(param_name) = 0;
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/bounding_box/X/min";
      params.set<double>(param_name) = params.get<double>("BioDynaMo/bounding_box/floor");
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/bounding_box/X/max";
      params.set<double>(param_name) = params.get<double>("BioDynaMo/bounding_box/ceil");
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/bounding_box/Y/min";
      params.set<double>(param_name) = params.get<double>("BioDynaMo/bounding_box/floor");
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/bounding_box/Y/max";
      params.set<double>(param_name) = params.get<double>("BioDynaMo/bounding_box/ceil");
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/bounding_box/Z/min";
      params.set<double>(param_name) = params.get<double>("BioDynaMo/bounding_box/floor");
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/bounding_box/Z/max";
      params.set<double>(param_name) = params.get<double>("BioDynaMo/bounding_box/ceil");
      //
    }
    //
    // current phenotype is under normoxia
    if ( 0 != t && 0 != t%2 ) {
      // copy parameters from normoxic cell to its hypoxic counterpart ;)
      std::string new_p_name;
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/diameter/ref";
      new_p_name = "BioDynaMo/cell_" +std::to_string(t+1)+ "/diameter/ref";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      params.set<Real>(new_p_name) = get_param<Real>(in, in_p, param_name);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/diameter/max";
      new_p_name = "BioDynaMo/cell_" +std::to_string(t+1)+ "/diameter/max";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      params.set<Real>(new_p_name) = get_param<Real>(in, in_p, param_name);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/O2_limit/to_hypoxia";
      new_p_name = "BioDynaMo/cell_" +std::to_string(t+1)+ "/O2_limit/to_hypoxia";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      params.set<Real>(new_p_name) = get_param<Real>(in, in_p, param_name);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/O2_limit/to_necrosis";
      new_p_name = "BioDynaMo/cell_" +std::to_string(t+1)+ "/O2_limit/to_necrosis";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      params.set<Real>(new_p_name) = get_param<Real>(in, in_p, param_name);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/transform_to_another_phenotype";
      new_p_name = "BioDynaMo/cell_" +std::to_string(t+1)+ "/transform_to_another_phenotype";
      params.set<int>(param_name) = get_param<int>(in, in_p, param_name);
      params.set<int>(new_p_name) = get_param<int>(in, in_p, param_name);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_growth/FEB3_communicated_field";
      new_p_name = "BioDynaMo/cell_" +std::to_string(t+1)+ "/chemical_cue_for_growth/FEB3_communicated_field";
      params.set<int>(param_name) = get_param<int>(in, in_p, param_name, -1);
      params.set<int>(new_p_name) = get_param<int>(in, in_p, param_name, -1);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_division/FEB3_communicated_field";
      new_p_name = "BioDynaMo/cell_" +std::to_string(t+1)+ "/chemical_cue_for_division/FEB3_communicated_field";
      params.set<int>(param_name) = get_param<int>(in, in_p, param_name, -1);
      params.set<int>(new_p_name) = get_param<int>(in, in_p, param_name, -1);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_migration/FEB3_communicated_field";
      new_p_name = "BioDynaMo/cell_" +std::to_string(t+1)+ "/chemical_cue_for_migration/FEB3_communicated_field";
      params.set<int>(param_name) = get_param<int>(in, in_p, param_name, -1);
      params.set<int>(new_p_name) = get_param<int>(in, in_p, param_name, -1);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_transformation/FEB3_communicated_field";
      new_p_name = "BioDynaMo/cell_" +std::to_string(t+1)+ "/chemical_cue_for_transformation/FEB3_communicated_field";
      params.set<int>(param_name) = get_param<int>(in, in_p, param_name, -1);
      params.set<int>(new_p_name) = get_param<int>(in, in_p, param_name, -1);
      //
      // checks to verify input
      if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/diameter/ref") >=
           params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/diameter/max") )
        libmesh_error();
      if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/O2_limit/to_hypoxia") <= 0.0 ||
           params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/O2_limit/to_hypoxia") >= 1.0 )
        libmesh_error();
      if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/O2_limit/to_necrosis") <= 0.0 ||
           params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/O2_limit/to_necrosis") >= 1.0 )
        libmesh_error();
      if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/O2_limit/to_hypoxia") <=
           params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/O2_limit/to_necrosis") )
        libmesh_error();
      if ( params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/transform_to_another_phenotype") < -1 )
        libmesh_error();
      if ( params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/transform_to_another_phenotype") == 0 )
        libmesh_error();
      if ( params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/transform_to_another_phenotype") == (int) t   )
        libmesh_error();
      if ( params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/transform_to_another_phenotype") == (int) t+1 )
        libmesh_error();
      if ( params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/transform_to_another_phenotype") >=
           (int) params.get<unsigned int>("BioDynaMo/number_of_phenotypes") )
        libmesh_error();
      //
      if ( params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_growth/FEB3_communicated_field") < -1 ||
           params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_growth/FEB3_communicated_field") == 0 )
        libmesh_error();
      if ( params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_division/FEB3_communicated_field") < -1 ||
           params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_division/FEB3_communicated_field") == 0 )
        libmesh_error();
      if ( params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_migration/FEB3_communicated_field") < -1 ||
           params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_migration/FEB3_communicated_field") == 0 )
        libmesh_error();
      if ( params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_transformation/FEB3_communicated_field") < -1 ||
           params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_transformation/FEB3_communicated_field") == 0 )
        libmesh_error();
      if ( params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_growth/FEB3_communicated_field") >=
           (int) params.get<unsigned int>("reaction-diffusion solver: dimension") )
        libmesh_error();
      if ( params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_division/FEB3_communicated_field") >=
           (int) params.get<unsigned int>("reaction-diffusion solver: dimension") )
        libmesh_error();
      if ( params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_migration/FEB3_communicated_field") >=
           (int) params.get<unsigned int>("reaction-diffusion solver: dimension") )
        libmesh_error();
      if ( params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_transformation/FEB3_communicated_field") >=
           (int) params.get<unsigned int>("reaction-diffusion solver: dimension") )
        libmesh_error();
      //
    }
    //
    // current phenotype is under normoxia or hypoxia
    if ( 0 != t ) {
      //
      const Real RVE_Area = params.get<Real>("BioDynaMo/bounding_box/face");
      const Real Cell_XSection = 0.25 * mPI
                               * pow2(params.get<Real>("BioDynaMo/cell_"+std::to_string(t)+"/diameter/max"));
      //
      // threshold value to permit cell invasion/migration;  can take any >0 real value
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/face_incoming_cell_threshold";
      params.set<unsigned int>(param_name+"/min") = (unsigned int) get_param<int>(in, in_p, param_name, +1);
      //
      // checks to verify input
      if ( params.get<unsigned int>("BioDynaMo/cell_"+std::to_string(t)+"/face_incoming_cell_threshold/min") == 0 ||
           params.get<unsigned int>("BioDynaMo/cell_"+std::to_string(t)+"/face_incoming_cell_threshold/min") > 10 )
        libmesh_error();
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/face_incoming_cell_threshold/pct";
      const Real Cell_PCT = get_param<Real>(in, in_p, param_name, 0.6666666666666);
      if ( Cell_PCT < 0.20 ) libmesh_error();
      if ( Cell_PCT > 0.90 ) libmesh_error();
      // only a percentage of cells of this type is permitted to invade/migrate into (per face)
      const unsigned int NCells = static_cast<unsigned int>(Cell_PCT * RVE_Area / Cell_XSection);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/face_incoming_cell_threshold/max";
      params.set<unsigned int>(param_name) = (unsigned int) get_param<int>(in, in_p, param_name, NCells);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/face_outgoing_cell_threshold/max";
      params.set<unsigned int>(param_name) = (unsigned int) get_param<int>(in, in_p, param_name, NCells);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/face_outgoing_cell_ratio/max";
      const Real ratio = Cell_PCT * (1.0/params.get<Real>("BioDynaMo/bounding_box/size"))
                       * params.get<Real>("BioDynaMo/micro2macro/space_scaling");
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name, ratio);
      //
    } else {
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/face_outgoing_cell_ratio/max";
      const Real ratio = 0.0;
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name, ratio);
      //
    }
    //
    // current phenotype is under normoxia or hypoxia
    if ( 0 != t ) {
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/growth_speed";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/division_probability";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/old_enough_to_die";
      params.set<int>(param_name) = get_param<int>(in, in_p, param_name);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_transformation/threshold";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name, -1.0);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_growth/threshold";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name, -1.0);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_division/threshold";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name, -1.0);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_migration/threshold";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name, -1.0);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_growth/scaling";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name, 1.);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_division/scaling";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name, 1.);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_chemical_cue/min";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_chemical_cue/max";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_Brownian/min";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_Brownian/max";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_oxygen_grad/min";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_oxygen_grad/max";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_oxygen_grad/threshold";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      //
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_ecm_grad/min";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_ecm_grad/max";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      param_name = "BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_ecm_grad/exponent";
      params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
      //
      // checks to verify input
      if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/growth_speed") < 0.0 )
        libmesh_error();
      if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/division_probability") < 0.0 ||
           params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/division_probability") > 1.0 )
        libmesh_error();
      if ( params.get<int>("BioDynaMo/cell_" + std::to_string(t) + "/old_enough_to_die") <= 0 )
        libmesh_error();
      if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_growth/scaling") <= 0.0 )
        libmesh_error();
      if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/chemical_cue_for_division/scaling") <= 0.0 )
        libmesh_error();
      if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_chemical_cue/min") >=
           params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_chemical_cue/max") )
        libmesh_error();
      if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_Brownian/min") >=
           params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_Brownian/max") )
        libmesh_error();
      if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_oxygen_grad/min") >=
           params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_oxygen_grad/max") )
        libmesh_error();
      if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_oxygen_grad/threshold") < 0.0 )
        libmesh_error();
      if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_ecm_grad/min") >=
           params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_ecm_grad/max") )
        libmesh_error();
      if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_ecm_grad/exponent") < 0.0 )
        libmesh_error();
      if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/migration_wrt_ecm_grad/exponent") > 10. )
        libmesh_error();
      //
    }
    //
    param_name = "BioDynaMo/cell_" + std::to_string(t) + "/death_probability";
    params.set<Real>(param_name) = get_param<Real>(in, in_p, param_name);
    //
    // checks to verify input
    if ( params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/death_probability") <  0.0 ||
         params.get<Real>("BioDynaMo/cell_" + std::to_string(t) + "/death_probability") >= 1.0 )
      libmesh_error();
    //
    // end of cell type loop...
  }

}
//=================================================================================================
