/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include <ctime>
#include "feb3/domain_data.h"
#include "libmesh/gmsh_io.h"
#include "feb3/fem-context.h"
#include "feb3/iga-context.h"
#include "feb3/paraview_io.h"
#include "feb3/mesh_iga.h"
#include "feb3/rve.h"
#include <algorithm>
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
extern
libMesh::Parallel::Communicator* lcomm_ptr;
extern // list of parameters inserted for the analysis
std::set<std::string> in_pars;
extern // list of constrained nodes (temporary container/prior mesh preparation)
std::list< my_triple<Point, BoundaryConditionInfo, my_pair<std::string, std::string> > > nodal_bcs__unsorted;
//=================================================================================================
// local variables
static
char err_msg[buffer_len];
static // nodal ID (before mesh preparation) and associated pointer to node
std::map<dof_id_type, Node*> nodal_id2ptr_3d__map;
static // elemental ID (before mesh preparation) and associated pointer to element
std::map<dof_id_type, Elem*> eleml_id2ptr_3d__map;
static // nodal ID (before vascular network mesh preparation) and associated pointer to node
std::map<dof_id_type, Node*> nodal_id2ptr_vn__map;
//=================================================================================================
// external routines
extern
void inspect_all_systems (const EquationSystems& , std::ofstream& );
extern
void check_and_fix_vascular_network (const char*);
//=================================================================================================
#include "feb3/input_inl.h"
#include "feb3/domain_data_inl.h"
//=================================================================================================
//
//
//
//=================================================================================================
DomainData::~DomainData () {
    // clear these local (static) containers...
    nodal_id2ptr_3d__map.clear();
    eleml_id2ptr_3d__map.clear();
    nodal_id2ptr_vn__map.clear();
    // ...and these external containers too
    nodal_bcs__unsorted.clear();
    in_pars.clear();
    // clear all data associated with BioDynaMo (if necessary)
    this->clear_bdm();
}
//=================================================================================================
void DomainData::prepare_mesh () {
    MeshBase& msh = this->_equation_systems.get_mesh();
    const Parameters& par = this->_equation_systems.parameters;
    char txt[buffer_len];
    //
    // read the 3D finite element mesh
    flog.print(" >> Read the 3D mesh\n", false);
    GmshIO(msh).read(par.get<std::string>("input mesh file").c_str());
    //
    // initialise these local maps - prior to preparing the mesh for use
    nodal_id2ptr_3d__map.clear();
    int node_min_id = -1;
    for (MeshBase::node_iterator i=msh.nodes_begin(); i!=msh.nodes_end(); i++) {
        Node* node = (*i);
        nodal_id2ptr_3d__map.insert( std::make_pair(node->id(), node) );
        if ( -1 == node_min_id ) node_min_id = (int) node->id();
        else                     node_min_id = std::min((int) node_min_id, (int) node->id());
    }
    eleml_id2ptr_3d__map.clear();
    int elem_min_id = -1;
    for (MeshBase::element_iterator i=msh.elements_begin(); i!=msh.elements_end(); i++) {
        Elem* elem = (*i);
        eleml_id2ptr_3d__map.insert( std::make_pair(elem->id(), elem) );
        if ( -1 == elem_min_id ) elem_min_id = (int) elem->id();
        else                     elem_min_id = std::min((int) elem_min_id, (int) elem->id());
    }
    //
    flog.print(" >> Prepare 3D mesh for use\n", false);
    //
    // find the elements to delete
    flog.print("  > Find elements to delete (if necessary)\n", false);
    std::vector<Elem*> elems2delete;
    // check if to remove elements
    if ( par.get<bool>("delete unwanted elements/nodes") )
    for (MeshBase::element_iterator i=msh.elements_begin(); i!=msh.elements_end(); i++) {
        Elem* elem = (*i);
        if ( this->regn.end() == this->regn.find( elem->subdomain_id() ) ) {
            elems2delete.push_back( elem );
        }
        // end of elements list loop...
    }
    // check if to remove elements
    if ( ! elems2delete.empty() ) {
        // now remove the elements only, which may produce isolated nodes
        sprintf(txt, "  > Remove %lu elements\n", elems2delete.size());
        flog.print(txt, false);
        for ( std::vector<Elem*>::const_iterator
              ci=elems2delete.begin(); ci!=elems2delete.end(); ci++ ) {
            msh.delete_elem(*ci);
        }
        // find the nodes to keep
        flog.print("  > Find nodes to keep\n", false);
        std::set<dof_id_type> nodes2keep;
        for (MeshBase::element_iterator i=msh.elements_begin(); i!=msh.elements_end(); i++) {
            Elem* elem = (*i);
            for (unsigned int n=0; n<elem->n_nodes(); n++) {
                nodes2keep.insert( elem->node_id(n) );
            }
            // end of elements list loop...
        }
        // find the nodes to delete
        flog.print("  > Find nodes to delete\n", false);
        std::vector<Node*> nodes2delete;
        for (MeshBase::node_iterator i=msh.nodes_begin(); i!=msh.nodes_end(); i++) {
            Node* node = (*i);
            if ( nodes2keep.end() == nodes2keep.find( node->id() ) ) {
                nodes2delete.push_back( node );
            }
            // end of nodes list loop...
        }
        // now remove the nodes
        sprintf(txt, "  > Remove %lu nodes\n", nodes2delete.size());
        flog.print(txt, false);
        for ( std::vector<Node*>::const_iterator
              ci=nodes2delete.begin(); ci!=nodes2delete.end(); ci++ ) {
            msh.delete_node(*ci);
        }
        // create a new boundary condition surface; traction-free by default
        flog.print("  > Add default BC info\n", false);
        BoundaryConditionInfo bc_info;
        bc_info.type = BoundaryConditionInfo::NEUMANN_NORMAL;
        bc_info.value.comp(0) = 0.0;
        bc_info.set_time_function(&flat_function);
        bc_info.time_params.assign(2, 0.0);
        // give a new boundary condition ID too
        const boundary_id_type bc_id = (--this->bcs.end())->first+1;
        // insert data into corresponding container but make this check first
        ASSERT_(this->bcs.end() == this->bcs.find(bc_id), "boundary surface has already been inserted");
        this->bcs.insert( std::make_pair(bc_id, bc_info) );
        //
        for (MeshBase::element_iterator i=msh.elements_begin(); i!=msh.elements_end(); i++) {
            Elem* elem = (*i);
            // loop for all element sides
            for (unsigned int s=0; s<elem->n_sides(); s++) {
                // check if element has this side on boundary
                if ( elem->neighbor_ptr(s) != NULL ) continue;
                // get the boundary ID for side of element
                if ( BoundaryInfo::invalid_id == msh.boundary_info->boundary_id(elem, s) ) {
                    msh.boundary_info->add_side(elem, s, bc_id);
                }
            }
            // end of elements loop...
        }
        // end of if case...
    }
    //
    if ( ! par.get<bool>("enforce input mesh order") ) libmesh_error();
    //
    flog.print("  > Enforce mesh preparation\n", false);
    // prepare the mesh, which could involve also node/element renumbering
    msh.prepare_for_use(par.get<bool>("skip input mesh renumbering"));
    // check if we run a FEM simulation
    if ( par.get<bool>("FEM framework") ) {
        //
        // if order of Lagrange basis is 2 then update the mesh accordingly
        if ( SECOND == par.get<Order>("Lagrange basis order") )
            msh.all_second_order(par.get<bool>("enforce input mesh 2nd order to full"));
        //
        MeshTools::build_nodes_to_elem_map(msh, this->_nodes_to_elem_map);
        //
    // check if we run an IGA simulation
    } else if ( par.get<bool>("IGA framework") ) {
        //
        ; // ...nothing to do here so far...
        //
    } else libmesh_error();
    // end of FE mesh manipulation
    flog.print("                                                \n", false);
    msh.print_info();
    msh.print_info(flog.stream());
    //msh.print_info(ferr.stream());
    //msh.boundary_info->print_info(flog.stream());
    //msh.boundary_info->print_info(ferr.stream());
    flog.print("                                                \n", false);
    // store a copy of the input Gmsh formatted mesh file
    if ( libMesh::global_processor_id() == 0 ) {
        const std::string fname = par.get<std::string>("results directory") + "/"
                                + par.get<std::string>("output file name") + "-initial.msh";
        GmshIO(msh).write(fname);
    }
    // write boundary conditions (labels)
    if ( true ) {
        const std::string fname = par.get<std::string>("results directory") + "/paraview/"
                                + par.get<std::string>("output file name") + "-initial-boundary.vtu";
        ParaviewIO(msh).write_boundary(fname);
    }
    //
    std::cout << "\n  > Min. nodal / elemental IDs of mesh inserted: "
              << node_min_id << " / " << elem_min_id << "\n" << std::endl;
    //
    if ( true ) {
        const std::string fname = par.get<std::string>("results directory") + "/temporary/"
                                + par.get<std::string>("output file name") + "-node2node.txt";
        std::ofstream fout(fname);
        for ( std::map<dof_id_type, Node*>::const_iterator
              ci=nodal_id2ptr_3d__map.begin(); ci!=nodal_id2ptr_3d__map.end(); ci++ ) {
            fout << ci->first-node_min_id << ' ' << ci->second->id() << std::endl;
        }
    }
    //
    if ( true ) {
        const std::string fname = par.get<std::string>("results directory") + "/temporary/"
                                + par.get<std::string>("output file name") + "-elem2elem.txt";
        std::ofstream fout(fname);
        for ( std::map<dof_id_type, Elem*>::const_iterator
              ci=eleml_id2ptr_3d__map.begin(); ci!=eleml_id2ptr_3d__map.end(); ci++ ) {
            fout << ci->first-elem_min_id << ' ' << ci->second->id() << std::endl;
        }
    }
    //
    // check if we run an IGA simulation
    if ( par.get<bool>("IGA framework") ) {
        const unsigned int n_par_iga_msh = par.get<unsigned int>("number of IGA parametric mesh(es)");
        //
        std::map<std::string, MeshIGA>* msh_iga =
            this->_equation_systems.parameters.set<std::map<std::string, MeshIGA>*>("pointer to IGA parametric mesh(es)");
        //
        for (unsigned int ii=0; ii<n_par_iga_msh; ii++) {
            sprintf(txt, "IGA parametric mesh label #%d", ii);
            const std::string label = par.get<std::string>(txt);
            //
            msh_iga->insert( std::make_pair(label, MeshIGA(msh)) );
        }
        if ( msh_iga->size() != n_par_iga_msh ) libmesh_error();
        //
        WARNING_("inserted NURBS elements are not checked for consistency; left at user's own risk");
        //
        for (unsigned int ii=0; ii<n_par_iga_msh; ii++) {
            sprintf(txt, "IGA parametric mesh label #%d", ii);
            const std::string label = par.get<std::string>(txt);
            MeshIGA& msh_iga_p = msh_iga->find(label)->second;
            // set the label of IGA mesh
            msh_iga_p.set_label(label);
            // set the NURBS by reading the corresponding input file
            sprintf(txt, "input NURBS info file name #%d", ii);
            const std::string file = par.get<std::string>(txt);
            msh_iga_p.set_nurbs(file);
        }
        // enforce to prepare control mesh for use
        msh.prepare_for_use();
        // enforce to decompose the (control) mesh in one part
        msh.partition(1);
        //
        for (unsigned int ii=0; ii<n_par_iga_msh; ii++) {
            sprintf(txt, "IGA parametric mesh label #%d", ii);
            const std::string label = par.get<std::string>(txt);
            MeshIGA& msh_iga_p = msh_iga->find(label)->second;
            // set the parametric IGA mesh and distribute data
            msh_iga_p.set_parametric_mesh();
            // ...end of IGA mesh manipulation
            flog.print("\n", false);
            msh_iga_p.get_parametric_mesh().print_info();
            msh_iga_p.print_info(flog.stream());
        }
        // end of if case...
    }
}
//=================================================================================================
void DomainData::init () {
    const MeshBase& msh = this->_equation_systems.get_mesh();
    const Parameters& par = this->_equation_systems.parameters;
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    flog.print(" >> Init domain data\n", false);
    const MeshBase::const_element_iterator begin_el = msh.active_elements_begin();
    const MeshBase::const_element_iterator end_el   = msh.active_elements_end();
    //
    // calculate the min/max edge length of all active 3D finite elements
    this->_equation_systems.parameters.set<Real>("finite element grid: min edge length") = +1.0e+22;
    this->_equation_systems.parameters.set<Real>("finite element grid: max edge length") = -1.0e+22;
    for (MeshBase::const_element_iterator ci=begin_el; ci!=end_el; ci++) {
        const Elem* elem = (*ci);
        for (unsigned int e=0; e<elem->n_edges(); e++) {
            UniquePtr<const Elem> edge = elem->build_edge_ptr(e);
            const Real length = edge->volume();
            //
            this->_equation_systems.parameters.set<Real>("finite element grid: min edge length") =
                std::min(length, par.get<Real>("finite element grid: min edge length"));
            //
            this->_equation_systems.parameters.set<Real>("finite element grid: max edge length") =
                std::max(length, par.get<Real>("finite element grid: max edge length"));
        }
        // end of active elements loop...
    }
    // run a sanity check
    for (MeshBase::const_element_iterator ci=begin_el; ci!=end_el; ci++) {
        const Elem* elem = (*ci);
        // check if this finite element is assigned to a region, else we have a problem
        if ( this->regn.end() == this->regn.find(elem->subdomain_id()) ) libmesh_error();
        //
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            if ( elem->neighbor_ptr(s) != NULL ) continue;
            const boundary_id_type bc_id = msh.boundary_info->boundary_id(elem, s);
            if ( BoundaryInfo::invalid_id == bc_id ) libmesh_error();
            if ( this->bcs.end() == this->bcs.find(bc_id) ) libmesh_error();
            //
            std::map<region_id_type, RegionInfo>::const_iterator citer_regn = this->regn.find(bc_id);
            if ( this->regn.end() != citer_regn ) {
                if ( ! citer_regn->second.is_membrane ) libmesh_error();
            }
            // end of side loop...
        }
        // end of active elements loop...
    }
    // clear container first
    this->node2elem.clear();
    // loop over all active finite elements
    for (MeshBase::const_element_iterator ci=begin_el; ci!=end_el; ci++) {
        const Elem* elem = (*ci);
        //
        std::map<region_id_type, RegionInfo>::const_iterator citer_regn = this->regn.find(elem->subdomain_id());
        if ( this->regn.end() == citer_regn ) libmesh_error();
        //
        for (unsigned int i=0; i<elem->n_nodes(); i++) {
            std::map<dof_id_type, std::set<dof_id_type> >::iterator it = this->node2elem.find(elem->node_id(i));
            if ( this->node2elem.end() == it ) {
                SetUInt st;
                st.insert( elem->id() );
                //
                this->node2elem.insert( std::make_pair(elem->node_id(i), st) );
            } else {
                it->second.insert( elem->id() );
            }
            // end of element nodes loop...
        }
        // end of active elements loop...
    }
    // clear container first
    this->node2elemside.clear();
    // loop over all active finite elements
    if ( par.get<bool>("membrane elements present") )
    for (MeshBase::const_element_iterator ci=begin_el; ci!=end_el; ci++) {
        const Elem* elem = (*ci);
        //
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            // check if element has this side on boundary
            if ( elem->neighbor_ptr(s) != NULL ) continue;
            //
            const boundary_id_type bc_id = msh.boundary_info->boundary_id(elem, s);
            if ( bc_id == BoundaryInfo::invalid_id ) libmesh_error();
            //
            if ( this->bcs.end() == this->bcs.find(bc_id) ) libmesh_error();
            const BoundaryConditionInfo* bci_ptr = &(this->bcs.find(bc_id)->second);
            if ( NULL == bci_ptr ) libmesh_error();
            //
            std::map<region_id_type, RegionInfo>::const_iterator citer_regn = this->regn.find(bc_id);
            if ( this->regn.end() == citer_regn ) continue;
            //
            const RegionInfo* ri_ptr = &(citer_regn->second);
            if ( ! ri_ptr->is_membrane ) libmesh_error();
            // create a side element
            UniquePtr<const Elem> side = elem->build_side_ptr(s);
            //
            for (unsigned int i=0; i<side->n_nodes(); i++) {
                std::map<dof_id_type, std::map<dof_id_type, SetUInt> >::iterator
                    it = this->node2elemside.find(side->node_id(i));
                //
                if ( this->node2elemside.end() == it ) {
                    SetUInt st;
                    st.insert( s );
                    //
                    std::map<dof_id_type, SetUInt> mp;
                    mp.insert( std::make_pair(elem->id(), st) );
                    //
                    this->node2elemside.insert( std::make_pair(side->node_id(i), mp) );
                } else {
                    std::map<dof_id_type, SetUInt>::iterator jt = it->second.find(elem->id());
                    if ( it->second.end() == jt ) {
                        SetUInt st;
                        st.insert( s );
                        //
                        it->second.insert( std::make_pair(elem->id(), st) );
                    } else {
                        jt->second.insert( s );
                    }
                }
                // end of side nodes loop...
            }
            // end of side loop...
        }
        // end of active elements loop...
    }
    // loop over all active finite elements
    for (MeshBase::const_element_iterator ci=begin_el; ci!=end_el; ci++) {
        const Elem* elem = (*ci);
        //
        std::map<region_id_type, RegionInfo>::iterator iter_regn = this->regn.find(elem->subdomain_id());
        if ( this->regn.end() == iter_regn ) libmesh_error();
        //
        RegionInfo* ri_ptr = &(iter_regn->second);
        // loop over all nodes of the finite element
        for (unsigned int i=0; i<elem->n_nodes(); i++) {
            ri_ptr->node.insert( elem->node_id(i) );
            if ( elem->processor_id() == libMesh::global_processor_id() )
                ri_ptr->node_local.insert( elem->node_id(i) );
            // end of finite element nodes loop...
        }
        // end of active elements loop...
    }
    //
    //
    // do initialise the quadrature information for all active local elements (and side elements too if necessary)
    this->init_quadrature();
    // now, initialise the mesh for all RVEs corresponding to the 3D quadrature of the problem
    this->init_rve();
    // // ...and, finally, initialize the data associated with BioDynaMo (if necessary)
    // this->init_bdm();
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    //
    // clear containers first
    this->damaged_node.clear();
    this->damaged_node_local.clear();
    // loop over all active finite elements
    for (MeshBase::const_element_iterator ci=begin_el; ci!=end_el; ci++) {
        const Elem* elem = (*ci);
        //
        std::map<region_id_type, RegionInfo>::const_iterator citer_regn = this->regn.find(elem->subdomain_id());
        if ( this->regn.end() == citer_regn ) libmesh_error();
        const RegionInfo* ri_ptr = &(citer_regn->second);
        if ( ! ri_ptr->is_damaged ) continue;
        // loop over all nodes of the element
        for (unsigned int i=0; i<elem->n_nodes(); i++) {
            this->damaged_node.insert( elem->node_id(i) );
            if ( elem->processor_id() == libMesh::global_processor_id() )
                this->damaged_node_local.insert( elem->node_id(i) );
            // end of element nodes loop...
        }
        //
        if ( par.get<bool>("membrane elements present") )
        // loop for all element sides
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            // check if element has this side on boundary
            if ( elem->neighbor_ptr(s) != NULL ) continue;
            // get the boundary ID for side of element
            const boundary_id_type bc_id = msh.boundary_info->boundary_id(elem, s);
            if ( BoundaryInfo::invalid_id == bc_id ) libmesh_error();
            // get access to the surface region if present
            std::map<region_id_type, RegionInfo>::const_iterator citer_regn_m = this->regn.find(bc_id);
            if ( this->regn.end() == citer_regn_m ) continue;
            else if ( ! citer_regn_m->second.is_membrane ) libmesh_error();
            // create a side element
            UniquePtr<const Elem> side = elem->build_side_ptr(s);
            // loop over all nodes of the element
            for (unsigned int i=0; i<side->n_nodes(); i++) {
                this->damaged_node.insert( side->node_id(i) );
                if ( elem->processor_id() == libMesh::global_processor_id() )
                    this->damaged_node_local.insert( side->node_id(i) );
                // end of element nodes loop...
            }
            // end of element sides loop...
        }
        // end of active local elements loop...
    }
    //
    // clear containers first
    this->healthy_node.clear();
    this->healthy_node_local.clear();
    this->interface_node.clear();
    this->interface_node_local.clear();
    // loop over all active finite elements
    for (MeshBase::const_element_iterator ci=begin_el; ci!=end_el; ci++) {
        const Elem* elem = (*ci);
        //
        std::map<region_id_type, RegionInfo>::const_iterator citer_regn = this->regn.find(elem->subdomain_id());
        if ( this->regn.end() == citer_regn ) libmesh_error();
        //
        const RegionInfo* ri_ptr = &(citer_regn->second);
        if ( ri_ptr->is_damaged ) continue;
        // loop over all nodes of the element
        for (unsigned int i=0; i<elem->n_nodes(); i++) {
            if ( this->damaged_node.end() != this->damaged_node.find( elem->node_id(i) ) ) {
                this->interface_node.insert( elem->node_id(i) );
                if ( elem->processor_id() == libMesh::global_processor_id() )
                    this->interface_node_local.insert( elem->node_id(i) );
            } else {
                this->healthy_node.insert( elem->node_id(i) );
                if ( elem->processor_id() == libMesh::global_processor_id() )
                    this->healthy_node_local.insert( elem->node_id(i) );
            }
            // end of element nodes loop...
        }
        //
        if ( par.get<bool>("membrane elements present") )
        // loop for all element sides
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            // check if element has this side on boundary
            if ( elem->neighbor_ptr(s) != NULL ) continue;
            // get the boundary ID for side of element
            const boundary_id_type bc_id = msh.boundary_info->boundary_id(elem, s);
            if ( BoundaryInfo::invalid_id == bc_id ) libmesh_error();
            // get access to the surface region if present
            std::map<region_id_type, RegionInfo>::const_iterator citer_regn_m = this->regn.find(bc_id);
            if ( this->regn.end() == citer_regn_m ) continue;
            else if ( ! citer_regn_m->second.is_membrane ) libmesh_error();
            // create a side element
            UniquePtr<const Elem> side = elem->build_side_ptr(s);
            // loop over all nodes of the element
            for (unsigned int i=0; i<side->n_nodes(); i++) {
                if ( this->damaged_node.end() != this->damaged_node.find( side->node_id(i) ) ) {
                    this->interface_node.insert( side->node_id(i) );
                    if ( elem->processor_id() == libMesh::global_processor_id() )
                        this->interface_node_local.insert( side->node_id(i) );
                } else {
                    this->healthy_node.insert( side->node_id(i) );
                    if ( elem->processor_id() == libMesh::global_processor_id() )
                        this->healthy_node_local.insert( side->node_id(i) );
                }
                // end of element nodes loop...
            }
            // end of element sides loop...
        }
        // end of active elements loop...
    }
    //
    // clear container first
    this->nodal_bcs.clear();
    unsigned int kpnt = 0;
    // form a list of boundary nodes only and loop for these active nodes in finding the constrained nodes
    std::map<dof_id_type, const Node*> active_boundary_nodes;
    // loop over all active finite elements
    for (MeshBase::const_element_iterator ci=begin_el; ci!=end_el; ci++) {
        const Elem* elem = (*ci);
        // loop for all element sides
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            // check if element has this side on boundary
            if ( elem->neighbor_ptr(s) != NULL ) continue;
            // create a side element
            UniquePtr<const Elem> side = elem->build_side_ptr(s);
            // loop over all nodes of the element
            for (unsigned int i=0; i<side->n_nodes(); i++) {
                const Node* node = side->node_ptr(i);
                if ( active_boundary_nodes.end() != active_boundary_nodes.find(node->id()) ) continue;
                active_boundary_nodes.insert( std::make_pair(node->id(), node) );
                // end of element nodes loop...
            }
            // end of element sides loop...
        }
        // end of active elements loop...
    }
    // iterate for the constrained nodes where BCs are prescribed
    for ( std::list< my_triple<Point, BoundaryConditionInfo, my_pair<std::string, std::string> > >::const_iterator
          citer_nbc=nodal_bcs__unsorted.begin(); citer_nbc!=nodal_bcs__unsorted.end(); citer_nbc++ )
    {
        const my_triple<Point, BoundaryConditionInfo, my_pair<std::string, std::string> >* t_ptr = &(*citer_nbc);
        //
        const Point& p = t_ptr->get_first();
        const Real gtol = par.get<Real>("constrained boundary nodes geometric tolerance");
        BoundaryConditionInfo bc_info = t_ptr->get_second();
        //
        const unsigned int n_t_params = bc_info.time_params.size();
        if        ( "flat_time_function"             == t_ptr->get_third().get_first() ) {
            bc_info.set_time_function(&flat_function);
            ASSERT_(n_t_params == 1, "wrong number of time parameters inserted");
        } else if ( "linear_time_function"           == t_ptr->get_third().get_first() ) {
            bc_info.set_time_function(&linear_function);
            ASSERT_(n_t_params == 1, "wrong number of time parameters inserted");
        } else if ( "linear_time_function__periodic" == t_ptr->get_third().get_first() ) {
            bc_info.set_time_function(&linear_function__periodic);
            ASSERT_(n_t_params == 2, "wrong number of time parameters inserted");
        } else if ( "ramp_time_function"             == t_ptr->get_third().get_first() ) {
            bc_info.set_time_function(&ramp_function);
            ASSERT_(n_t_params == 4, "wrong number of time parameters inserted");
        } else if ( "ramp_time_function__periodic"   == t_ptr->get_third().get_first() ) {
            bc_info.set_time_function(&ramp_function__periodic);
            ASSERT_(n_t_params == 5, "wrong number of time parameters inserted");
        } else if ( "steep_time_function"            == t_ptr->get_third().get_first() ) {
            bc_info.set_time_function(&steep_function);
            ASSERT_(n_t_params == 6, "wrong number of time parameters inserted");
        } else if ( "steep_time_function__periodic"  == t_ptr->get_third().get_first() ) {
            bc_info.set_time_function(&steep_function__periodic);
            ASSERT_(n_t_params == 7, "wrong number of time parameters inserted");
        } else if ( "exp_time_function"              == t_ptr->get_third().get_first() ) {
            bc_info.set_time_function(&exp_function);
            ASSERT_(n_t_params == 5, "wrong number of time parameters inserted");
        } else if ( "exp_time_function__periodic"    == t_ptr->get_third().get_first() ) {
            bc_info.set_time_function(&exp_function__periodic);
            ASSERT_(n_t_params == 6, "wrong number of time parameters inserted");
        } else {
            sprintf(err_msg, "exception caught for time function of point with coords (%f;%f;%f)", p(0), p(1), p(2));
            ERROR_(err_msg);
        }
        //
        const unsigned int n_a_params = bc_info.active_params.size();
        if        ( "flat_time_function"             == t_ptr->get_third().get_second() ) {
            bc_info.set_active_function(&flat_function);
            ASSERT_(n_a_params == 1, "wrong number of active parameters inserted");
        } else if ( "linear_time_function"           == t_ptr->get_third().get_second() ) {
            bc_info.set_active_function(&linear_function);
            ASSERT_(n_a_params == 1, "wrong number of active parameters inserted");
        } else if ( "linear_time_function__periodic" == t_ptr->get_third().get_second() ) {
            bc_info.set_active_function(&linear_function__periodic);
            ASSERT_(n_a_params == 2, "wrong number of active parameters inserted");
        } else if ( "ramp_time_function"             == t_ptr->get_third().get_second() ) {
            bc_info.set_active_function(&ramp_function);
            ASSERT_(n_a_params == 4, "wrong number of active parameters inserted");
        } else if ( "ramp_time_function__periodic"   == t_ptr->get_third().get_second() ) {
            bc_info.set_active_function(&ramp_function__periodic);
            ASSERT_(n_a_params == 5, "wrong number of active parameters inserted");
        } else if ( "steep_time_function"            == t_ptr->get_third().get_second() ) {
            bc_info.set_active_function(&steep_function);
            ASSERT_(n_a_params == 6, "wrong number of active parameters inserted");
        } else if ( "steep_time_function__periodic"  == t_ptr->get_third().get_second() ) {
            bc_info.set_active_function(&steep_function__periodic);
            ASSERT_(n_a_params == 7, "wrong number of active parameters inserted");
        } else if ( "exp_time_function"              == t_ptr->get_third().get_second() ) {
            bc_info.set_active_function(&exp_function);
            ASSERT_(n_a_params == 5, "wrong number of active parameters inserted");
        } else if ( "exp_time_function__periodic"    == t_ptr->get_third().get_second() ) {
            bc_info.set_active_function(&exp_function__periodic);
            ASSERT_(n_a_params == 6, "wrong number of active parameters inserted");
        } else if ( "step_time_function"             == t_ptr->get_third().get_second() ) {
            bc_info.set_active_function(&step_function);
            ASSERT_(n_a_params == 3, "wrong number of active parameters inserted");
        } else {
            sprintf(err_msg, "exception caught for active function of point with coords (%f;%f;%f)", p(0), p(1), p(2));
            ERROR_(err_msg);
        }
        //
        if ( p(0) != BoundaryConditionInfo::generic_coord &&
             p(1) != BoundaryConditionInfo::generic_coord &&
             p(2) != BoundaryConditionInfo::generic_coord ) {
            //
            dof_id_type node_id;
            bool node_found = false;
            for ( std::map<dof_id_type, const Node*>::const_iterator
                  citer_nd=active_boundary_nodes.begin(); citer_nd!=active_boundary_nodes.end(); citer_nd++ ) {
                const Node* node = citer_nd->second;
                // check coordinates (think something smarter)
                if ( fabs((*node)(0)-p(0)) <= gtol &&
                     fabs((*node)(1)-p(1)) <= gtol &&
                     fabs((*node)(2)-p(2)) <= gtol ) {
                    node_id = node->id();
                    node_found = true;
                }
                // end of active boundary nodes loop..
            }
            sprintf(err_msg, "unable to locate constrained point #%d with coords (%f;%f;%f)", kpnt, p(0), p(1), p(2));
            ASSERT_(node_found, err_msg);
            // insert data into corresponding container but check first
            if ( this->nodal_bcs.end() == this->nodal_bcs.find(node_id) ) {
                this->nodal_bcs.insert( std::make_pair(node_id, bc_info) );
                sprintf(err_msg, "constrained point #%d with coords (%f;%f;%f) has functional ID %d", kpnt, p(0), p(1), p(2), node_id+1);
                WARNING_(err_msg);
                ++kpnt;
            }
            //
        } else if ( p(0) == BoundaryConditionInfo::generic_coord &&
                    p(1) == BoundaryConditionInfo::generic_coord ) {
            //
            dof_id_type node_id;
            bool node_found = false;
            for ( std::map<dof_id_type, const Node*>::const_iterator
                  citer_nd=active_boundary_nodes.begin(); citer_nd!=active_boundary_nodes.end(); citer_nd++ ) {
                const Node* node = citer_nd->second;
                // check coordinates (think something smarter)
                if ( fabs((*node)(2)-p(2)) <= gtol ) {
                    node_id = node->id();
                    node_found = true;
                    // insert data into corresponding container but check first
                    if ( this->nodal_bcs.end() == this->nodal_bcs.find(node_id) ) {
                        this->nodal_bcs.insert( std::make_pair(node_id, bc_info) );
                        sprintf(err_msg, "constrained point #%d with coords (%f;%f;%f) has functional ID %d", kpnt, p(0), p(1), p(2), node_id+1);
                        WARNING_(err_msg);
                        ++kpnt;
                    }
                }
                // end of active boundary nodes loop..
            }
            sprintf(err_msg, "unable to locate constrained point #%d with coords (%f;%f;%f)", kpnt, p(0), p(1), p(2));
            ASSERT_(node_found, err_msg);
            //
        } else if ( p(0) == BoundaryConditionInfo::generic_coord &&
                    p(2) == BoundaryConditionInfo::generic_coord ) {
            //
            dof_id_type node_id;
            bool node_found = false;
            for ( std::map<dof_id_type, const Node*>::const_iterator
                  citer_nd=active_boundary_nodes.begin(); citer_nd!=active_boundary_nodes.end(); citer_nd++ ) {
                const Node* node = citer_nd->second;
                // check coordinates (think something smarter)
                if ( fabs((*node)(1)-p(1)) <= gtol ) {
                    node_id = node->id();
                    node_found = true;
                    // insert data into corresponding container but check first
                    if ( this->nodal_bcs.end() == this->nodal_bcs.find(node_id) ) {
                        this->nodal_bcs.insert( std::make_pair(node_id, bc_info) );
                        sprintf(err_msg, "constrained point #%d with coords (%f;%f;%f) has functional ID %d", kpnt, p(0), p(1), p(2), node_id+1);
                        WARNING_(err_msg);
                        ++kpnt;
                    }
                }
                // end of active boundary nodes loop..
            }
            sprintf(err_msg, "unable to locate constrained point #%d with coords (%f;%f;%f)", kpnt, p(0), p(1), p(2));
            ASSERT_(node_found, err_msg);
            //
        } else if ( p(1) == BoundaryConditionInfo::generic_coord &&
                    p(2) == BoundaryConditionInfo::generic_coord ) {
            //
            dof_id_type node_id;
            bool node_found = false;
            for ( std::map<dof_id_type, const Node*>::const_iterator
                  citer_nd=active_boundary_nodes.begin(); citer_nd!=active_boundary_nodes.end(); citer_nd++ ) {
                const Node* node = citer_nd->second;
                // check coordinates (think something smarter)
                if ( fabs((*node)(0)-p(0)) <= gtol ) {
                    node_id = node->id();
                    node_found = true;
                    // insert data into corresponding container but check first
                    if ( this->nodal_bcs.end() == this->nodal_bcs.find(node_id) ) {
                        this->nodal_bcs.insert( std::make_pair(node_id, bc_info) );
                        sprintf(err_msg, "constrained point #%d with coords (%f;%f;%f) has functional ID %d", kpnt, p(0), p(1), p(2), node_id+1);
                        WARNING_(err_msg);
                        ++kpnt;
                    }
                }
                // end of active boundary nodes loop..
            }
            sprintf(err_msg, "unable to locate constrained point #%d with coords (%f;%f;%f)", kpnt, p(0), p(1), p(2));
            ASSERT_(node_found, err_msg);
            //
        } else if ( p(0) == BoundaryConditionInfo::generic_coord ) {
            //
            dof_id_type node_id;
            bool node_found = false;
            for ( std::map<dof_id_type, const Node*>::const_iterator
                  citer_nd=active_boundary_nodes.begin(); citer_nd!=active_boundary_nodes.end(); citer_nd++ ) {
                const Node* node = citer_nd->second;
                // check coordinates (think something smarter)
                if ( fabs((*node)(1)-p(1)) <= gtol &&
                     fabs((*node)(2)-p(2)) <= gtol ) {
                    node_id = node->id();
                    node_found = true;
                    // insert data into corresponding container but check first
                    if ( this->nodal_bcs.end() == this->nodal_bcs.find(node_id) ) {
                        this->nodal_bcs.insert( std::make_pair(node_id, bc_info) );
                        sprintf(err_msg, "constrained point #%d with coords (%f;%f;%f) has functional ID %d", kpnt, p(0), p(1), p(2), node_id+1);
                        WARNING_(err_msg);
                        ++kpnt;
                    }
                }
                // end of active boundary nodes loop..
            }
            sprintf(err_msg, "unable to locate constrained point #%d with coords (%f;%f;%f)", kpnt, p(0), p(1), p(2));
            ASSERT_(node_found, err_msg);
            //
        } else if ( p(1) == BoundaryConditionInfo::generic_coord ) {
            //
            dof_id_type node_id;
            bool node_found = false;
            for ( std::map<dof_id_type, const Node*>::const_iterator
                  citer_nd=active_boundary_nodes.begin(); citer_nd!=active_boundary_nodes.end(); citer_nd++ ) {
                const Node* node = citer_nd->second;
                // check coordinates (think something smarter)
                if ( fabs((*node)(0)-p(0)) <= gtol &&
                     fabs((*node)(2)-p(2)) <= gtol ) {
                    node_id = node->id();
                    node_found = true;
                    // insert data into corresponding container but check first
                    if ( this->nodal_bcs.end() == this->nodal_bcs.find(node_id) ) {
                        this->nodal_bcs.insert( std::make_pair(node_id, bc_info) );
                        sprintf(err_msg, "constrained point #%d with coords (%f;%f;%f) has functional ID %d", kpnt, p(0), p(1), p(2), node_id+1);
                        WARNING_(err_msg);
                        ++kpnt;
                    }
                }
                // end of active boundary nodes loop..
            }
            sprintf(err_msg, "unable to locate constrained point #%d with coords (%f;%f;%f)", kpnt, p(0), p(1), p(2));
            ASSERT_(node_found, err_msg);
            //
        } else if ( p(2) == BoundaryConditionInfo::generic_coord ) {
            //
            dof_id_type node_id;
            bool node_found = false;
            for ( std::map<dof_id_type, const Node*>::const_iterator
                  citer_nd=active_boundary_nodes.begin(); citer_nd!=active_boundary_nodes.end(); citer_nd++ ) {
                const Node* node = citer_nd->second;
                // check coordinates (think something smarter)
                if ( fabs((*node)(0)-p(0)) <= gtol &&
                     fabs((*node)(1)-p(1)) <= gtol ) {
                    node_id = node->id();
                    node_found = true;
                    // insert data into corresponding container but check first
                    if ( this->nodal_bcs.end() == this->nodal_bcs.find(node_id) ) {
                        this->nodal_bcs.insert( std::make_pair(node_id, bc_info) );
                        sprintf(err_msg, "constrained point #%d with coords (%f;%f;%f) has functional ID %d", kpnt, p(0), p(1), p(2), node_id+1);
                        WARNING_(err_msg);
                        ++kpnt;
                    }
                }
                // end of active boundary nodes loop..
            }
            sprintf(err_msg, "unable to locate constrained point #%d with coords (%f;%f;%f)", kpnt, p(0), p(1), p(2));
            ASSERT_(node_found, err_msg);
            //
        } else ERROR_("exception is caught here for constrained point insertion");
        // end of nodal BCs loop...
    }
    // force to clear container
    active_boundary_nodes.clear();
    //
    //
    // clear container first
    this->nforce.node_local.clear();
    //
    // check if we run a FEM simulation
    if ( par.get<bool>("FEM framework") ) {
        // we run a pure FEM analysis
        std::cout << "  > FE (3D) grid mininum edge length = " << std::scientific << par.get<Real>("finite element grid: min edge length") << std::endl;
        flog.stream() << "  > FE (3D) grid mininum edge length = " << std::scientific << par.get<Real>("finite element grid: min edge length") << std::endl;
        std::cout << "  > FE (3D) grid maximum edge length = " << std::scientific << par.get<Real>("finite element grid: max edge length") << std::endl;
        flog.stream() << "  > FE (3D) grid maximum edge length = " << std::scientific << par.get<Real>("finite element grid: max edge length") << std::endl;
        //
        int n_entities = 0;
        n_entities = this->_equation_systems.parameters.get<int>("nodal force evaluation: number of external surfaces");
        for (int isurf=0; isurf<n_entities; isurf++) {
            char str[buffer_len];
            sprintf(str, "nodal force evaluation: external surface %d", isurf);
            const boundary_id_type bc_id = this->_equation_systems.parameters.get<boundary_id_type>(str);
            //
            // loop over all active local elements
            MeshBase::const_element_iterator citer_el_end = msh.active_local_elements_end();
            for ( MeshBase::const_element_iterator
                  citer_el=msh.active_local_elements_begin(); citer_el!=citer_el_end; citer_el++ ) {
                const Elem* elem = *citer_el;
                // insert essential boundary conditions
                for (unsigned int s=0; s<elem->n_sides(); s++) {
                    // check if element has this side on the boundary
                    if ( elem->neighbor_ptr(s) != NULL ) continue;
                    // check if this boundary has the desired ID
                    if ( bc_id != msh.boundary_info->boundary_id(elem, s) ) continue;
                    // create a side element
                    UniquePtr<const Elem> side = elem->build_side_ptr(s);
                    // add the local nodes into list
                    for (unsigned int ns=0; ns<side->n_nodes(); ns++) {
                        this->nforce.node_local.insert( side->node_id(ns) );
                    }
                    // end of element sides...
                }
                // end of active local elements loop...
            }
            // end of external surfaces loop...
        }
        //
        n_entities = this->_equation_systems.parameters.get<int>("nodal force evaluation: number of internal surfaces");
        for (int isurf=0; isurf<n_entities; isurf++) {
            char str[buffer_len];
            sprintf(str, "nodal force evaluation: region-0 %d", isurf);
            const region_id_type r0_id = this->_equation_systems.parameters.get<region_id_type>(str);
            sprintf(str, "nodal force evaluation: region-1 %d", isurf);
            const region_id_type r1_id = this->_equation_systems.parameters.get<region_id_type>(str);
            //
            std::set<dof_id_type> node_local__r0;
            std::set<dof_id_type> node_local__r1;
            //
            // loop over all active local elements
            MeshBase::const_element_iterator citer_el_end = msh.active_local_elements_end();
            for ( MeshBase::const_element_iterator
                  citer_el=msh.active_local_elements_begin(); citer_el!=citer_el_end; citer_el++ ) {
                const Elem* elem = *citer_el;
                // check if this element is in each region
                if        ( r0_id == elem->subdomain_id() ) {
                    // add the local nodes into list
                    for (unsigned int ne=0; ne<elem->n_nodes(); ne++) {
                        node_local__r0.insert( elem->node_id(ne) );
                    }
                } else if ( r1_id == elem->subdomain_id() ) {
                    // add the local nodes into list
                    for (unsigned int ne=0; ne<elem->n_nodes(); ne++) {
                        node_local__r1.insert( elem->node_id(ne) );
                    }
                }
                // end of active local elements loop...
            }
            // check for common local nodes
            for (std::set<unsigned int>::const_iterator ci=node_local__r0.begin(); ci!=node_local__r0.end(); ci++) {
                if ( node_local__r1.end() != node_local__r1.find(*ci) ) {
                    this->nforce.node_local.insert( *ci );
                }
            }
            // end of internal surfaces loop...
        }
        //
        n_entities = this->_equation_systems.parameters.get<int>("nodal force evaluation: number of regions");
        for (int inode=0; inode<n_entities; inode++) {
            char str[buffer_len];
            sprintf(str, "nodal force evaluation: region %d", inode);
            const region_id_type r_id = this->_equation_systems.parameters.get<region_id_type>(str);
            //
            // loop over all active local elements
            MeshBase::const_element_iterator citer_el_end = msh.active_local_elements_end();
            for ( MeshBase::const_element_iterator
                  citer_el=msh.active_local_elements_begin(); citer_el!=citer_el_end; citer_el++ ) {
                const Elem* elem = *citer_el;
                // check if this element is in the same region
                if ( r_id != elem->subdomain_id() ) continue;
                // add the local nodes into list
                for (unsigned int ne=0; ne<elem->n_nodes(); ne++) {
                    this->nforce.node_local.insert( elem->node_id(ne) );
                }
                // end of active local elements loop...
            }
            // end of region loop...
        }
        //  if all regions are considered then force to clear container
        n_entities = this->_equation_systems.parameters.get<int>("nodal force evaluation: number of regions");
        if ( this->regn.size() == n_entities ) {
            this->nforce.node_local.clear();
        }
        //
    // check if we run an IGA simulation
    } else if ( par.get<bool>("IGA framework") ) {
        // we run a pure IGA analysis
        std::cout << "  > Control grid mininum edge length = " << std::scientific << par.get<Real>("finite element grid: min edge length") << std::endl;
        flog.stream() << "  > Control grid mininum edge length = " << std::scientific << par.get<Real>("finite element grid: min edge length") << std::endl;
        std::cout << "  > Control grid maximum edge length = " << std::scientific << par.get<Real>("finite element grid: max edge length") << std::endl;
        flog.stream() << "  > Control grid maximum edge length = " << std::scientific << par.get<Real>("finite element grid: max edge length") << std::endl;
        //
        ;
        // excepetion is caught...
    } else libmesh_error();
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // finally run this inspection routine
    inspect_all_systems(this->_equation_systems, flog.stream());
}
//=================================================================================================
void DomainData::init_quadrature () {
    flog.print("  > Init quadrature data\n", false);
    //
    const MeshBase& msh = this->_equation_systems.get_mesh();
    const Parameters& par = this->_equation_systems.parameters;
    // access the list of IGA meshes
    std::map<std::string, MeshIGA>& msh_iga =
        *(this->_equation_systems.parameters.set<std::map<std::string, MeshIGA>*>("pointer to IGA parametric mesh(es)"));
    // sanity check
    ASSERT_(SP3D==msh.mesh_dimension(), "fatal internal error occurred");
    // get the initial values for the corresponding fields of the reaction-diffusion solver
    std::vector<Real> initial_field_value__H__RDS,
                      initial_field_value__D__RDS;
    do {
        const unsigned int nvars_RDS = par.get<unsigned int>("reaction-diffusion solver: dimension");
        initial_field_value__H__RDS.resize(nvars_RDS);
        initial_field_value__D__RDS.resize(nvars_RDS);
        for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
            char field_name[buffer_len];
            sprintf(field_name, "reaction-diffusion solver: healthy domain initial field %d", ivar);
            initial_field_value__H__RDS[ivar] = par.get<Real>(field_name);
            sprintf(field_name, "reaction-diffusion solver: damaged domain initial field %d", ivar);
            initial_field_value__D__RDS[ivar] = par.get<Real>(field_name);
        }
    } while ( false );

    // set default quadrature information first
    //
    // check if we run a FEM simulation
    if ( par.get<bool>("FEM framework") ) {
        // we run a pure FEM analysis
        const System& sys = this->_equation_systems.get_system<System>(0);
        //
        std::vector<unsigned short int> var;
        FEM_Context fem_ctx((System&)sys);
        fem_ctx.get_var_list(var);
        //
        const unsigned int nvars_RDS   = par.get<unsigned int>("reaction-diffusion solver: dimension"),
                           ndcvars_RDS = par.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                           ninvars_NCS = par.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                           nvars_AUX   = par.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
        const Order gqo = par.get<Order>("Gauss quadrature order");
        //
        // clear containers first
        this->elem_qi_local.clear();
        this->elem_side_qi_local.clear();
        //
        const MeshBase::const_element_iterator begin_el = msh.active_local_elements_begin();
        const MeshBase::const_element_iterator end_el = msh.active_local_elements_end();
        // loop over all active local elements
        for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
            const Elem* elem = *citer_el;
            //
            QGauss quad_volu(SP3D, gqo);
            UniquePtr<FEBase> fe_volu = FEBase::build(SP3D, FEType(sys.variable_type(var[0])));
            fe_volu->attach_quadrature_rule(&quad_volu);
            //
            // initialize and get access to the FE stuff
            const std::vector< Point >& qpnt = fe_volu->get_xyz();
            //const std::vector< std::vector<RealGradient> >& dPhi_dX = fe_volu->get_dphi();
            fe_volu->reinit(elem);
            // get access to the region
            if ( this->regn.end() == this->regn.find(elem->subdomain_id()) ) libmesh_error();
            const RegionInfo* ri_ptr = &(this->regn.find(elem->subdomain_id())->second);
            //
            const unsigned int n_qp = quad_volu.n_points();
            QuadratureInfo elem_quad(n_qp, nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
            // parametric coordinates of the quadrature points
            //const std::vector< Point >& qpnt_p = quad_volu.get_points();
            //
            if ( begin_el == citer_el ) {
                std::cout << "  > Volume integration points / 3D finite element = " << n_qp << std::endl;
                flog.stream() << "  > Volume integration points / 3D finite element = " << n_qp << std::endl;
            }
            // set the quadrature points material and spatial location
            for (unsigned int qp=0; qp<n_qp; qp++) {
                elem_quad.XYZ[qp] = qpnt[qp];
                elem_quad.xyz[qp] = qpnt[qp];
                //
                elem_quad.ORIENT[0][qp] = orientation_of_3D_element(*citer_el, 0);
                elem_quad.ORIENT[1][qp] = orientation_of_3D_element(*citer_el, 1);
                elem_quad.ORIENT[2][qp] = orientation_of_3D_element(*citer_el, 2);
            }
            // set the internal biomechanical and biochemical internal variables
            if ( ri_ptr->is_damaged ) {
                for (unsigned int qp=0; qp<n_qp; qp++) {
                    if ( ninvars_NCS == 0 ) {
                        ; // nothing to do here...
                    } else if ( ninvars_NCS == 3 ) {
                        elem_quad.biomech[qp][0] = material__max_damage_factor;
                        elem_quad.biomech[qp][1] = ri_ptr->mat->get_biomech_param(0) * (1.0-elem_quad.biomech[qp][0]);
                        elem_quad.biomech[qp][2] = 0.0;
                    } else libmesh_error();
                    for (unsigned int i=0; i<nvars_RDS; i++) {
                        elem_quad.biochem[qp][i] = initial_field_value__D__RDS[i];
                    }
                }
            } else {
                for (unsigned int qp=0; qp<n_qp; qp++) {
                    if ( ninvars_NCS == 0 ) {
                        ; // nothing to do here...
                    } else if ( ninvars_NCS == 3 ) {
                        elem_quad.biomech[qp][0] = material__min_damage_factor;
                        elem_quad.biomech[qp][1] = ri_ptr->mat->get_biomech_param(0) * (1.0-elem_quad.biomech[qp][0]);
                        elem_quad.biomech[qp][2] = 0.0;
                    } else libmesh_error();
                    for (unsigned int i=0; i<nvars_RDS; i++) {
                        elem_quad.biochem[qp][i] = initial_field_value__H__RDS[i];
                    }
                }
            }
            // set the fibres orientation vector in reference configuration
            for (unsigned int qp=0; qp<n_qp; qp++) {
                elem_quad.fibre_ref[0][qp] = vec(Point(uniform_distro(-1., +1.), uniform_distro(-1., +1.), uniform_distro(-1., +1.)).unit());
                elem_quad.fibre_ref[1][qp] = vec(Point(uniform_distro(-1., +1.), uniform_distro(-1., +1.), uniform_distro(-1., +1.)).unit());
                elem_quad.fibre_ref[2][qp] = vec(Point(uniform_distro(-1., +1.), uniform_distro(-1., +1.), uniform_distro(-1., +1.)).unit());
                elem_quad.fibre_curr[0][qp] = elem_quad.fibre_ref[0][qp];
                elem_quad.fibre_curr[1][qp] = elem_quad.fibre_ref[1][qp];
                elem_quad.fibre_curr[2][qp] = elem_quad.fibre_ref[2][qp];
            }
            // store data into the container
            this->elem_qi_local.insert( std::make_pair(elem->id(), elem_quad) );
            // end of active local elements loop...
        }
        // check if membrane elements are present in analysis
        if ( par.get<bool>("membrane elements present") )
        // loop over all active local elements
        for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
            const Elem* elem = *citer_el;
            //
            std::map<unsigned int, QuadratureInfo> elem_quad;
            // loop for all element sides
            for (unsigned int s=0; s<elem->n_sides(); s++) {
                // check if element has this side on boundary
                if ( elem->neighbor_ptr(s) != NULL ) continue;
                // get the boundary ID for side of element
                const boundary_id_type bc_id = msh.boundary_info->boundary_id(elem, s);
                if ( BoundaryInfo::invalid_id == bc_id ) libmesh_error();
                // get access to the region
                std::map<region_id_type, RegionInfo>::const_iterator ci = this->regn.find(bc_id);
                if ( this->regn.end() == ci ) continue;
                const RegionInfo* ri_ptr = &(ci->second);
                if ( ! ri_ptr->is_membrane ) continue;
                // create a side element
                UniquePtr<const Elem> side = elem->build_side_ptr(s);
                //
                QGauss quad_surf(SP2D, gqo);
                UniquePtr<FEBase> fe_surf = FEBase::build(SP3D, FEType(sys.variable_type(var[0])));
                fe_surf->attach_quadrature_rule(&quad_surf);
                //
                // initialize and get access to the FE stuff
                const std::vector< Point >& qpnt = fe_surf->get_xyz();
                //const std::vector< std::vector<RealGradient> >& dPhi_dX = fe_surf->get_dphi();
                //const std::vector< Point >& qeta = fe_surf->get_normals();
                fe_surf->reinit(elem, s);
                //
                const unsigned int n_qp = quad_surf.n_points();
                QuadratureInfo side_quad(n_qp, nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
                //
                if ( begin_el == citer_el ) {
                    std::cout << "  > Surface integration points / 2D finite element = " << n_qp << std::endl;
                    flog.stream() << "  > Surface integration points / 2D finite element = " << n_qp << std::endl;
                }
                // set the quadrature points material and spatial location
                for (unsigned int qp=0; qp<n_qp; qp++) {
                    side_quad.XYZ[qp] = qpnt[qp];
                    side_quad.xyz[qp] = qpnt[qp];
                }
                // set the internal biomechanical and biochemical internal variables
                if ( ri_ptr->is_damaged ) {
                    for (unsigned int qp=0; qp<n_qp; qp++) {
                        if ( ninvars_NCS == 0 ) {
                            ; // nothing to do here...
                        } else if ( ninvars_NCS == 3 ) {
                            side_quad.biomech[qp][0] = material__max_damage_factor;
                            side_quad.biomech[qp][1] = ri_ptr->mat->get_biomech_param(0) * (1.0-side_quad.biomech[qp][0]);
                            side_quad.biomech[qp][2] = 0.0;
                        } else libmesh_error();
                        for (unsigned int i=0; i<nvars_RDS; i++) {
                            side_quad.biochem[qp][i] = initial_field_value__D__RDS[i];
                        }
                    }
                } else {
                    for (unsigned int qp=0; qp<n_qp; qp++) {
                        if ( ninvars_NCS == 0 ) {
                            ; // nothing to do here...
                        } else if ( ninvars_NCS == 3 ) {
                            side_quad.biomech[qp][0] = material__min_damage_factor;
                            side_quad.biomech[qp][1] = ri_ptr->mat->get_biomech_param(0) * (1.0-side_quad.biomech[qp][0]);
                            side_quad.biomech[qp][2] = 0.0;
                        } else libmesh_error();
                        for (unsigned int i=0; i<nvars_RDS; i++) {
                            side_quad.biochem[qp][i] = initial_field_value__H__RDS[i];
                        }
                    }
                }
                // set the fibres orientation vector in reference configuration
                for (unsigned int qp=0; qp<n_qp; qp++) {
                    side_quad.fibre_ref[0][qp] = vec(Point(uniform_distro(-1., +1.), uniform_distro(-1., +1.), uniform_distro(-1., +1.)).unit());
                    side_quad.fibre_ref[1][qp] = vec(Point(uniform_distro(-1., +1.), uniform_distro(-1., +1.), uniform_distro(-1., +1.)).unit());
                    side_quad.fibre_ref[2][qp] = vec(Point(uniform_distro(-1., +1.), uniform_distro(-1., +1.), uniform_distro(-1., +1.)).unit());
                    side_quad.fibre_curr[0][qp] = side_quad.fibre_ref[0][qp];
                    side_quad.fibre_curr[1][qp] = side_quad.fibre_ref[1][qp];
                    side_quad.fibre_curr[2][qp] = side_quad.fibre_ref[2][qp];
                }
                // store data into the container
                elem_quad.insert( std::make_pair(s, side_quad) );
                // end of element sides loop...
            }
            // store data into the container
            if ( ! elem_quad.empty() )
                this->elem_side_qi_local.insert( std::make_pair(elem->id(), elem_quad) );
            // end of active local elements loop...
        }
        //
    // check if we run an IGA simulation
    } else if ( par.get<bool>("IGA framework") ) {
        //const System& sys = this->_equation_systems.get_system<System>(0);
        //
        const std::string label =
            par.get<std::string>("IGA parametric mesh to use for analysis");
        const MeshIGA& msh_iga_p = msh_iga.find(label)->second;
        //
        const unsigned int nvars_RDS   = par.get<unsigned int>("reaction-diffusion solver: dimension"),
                           ndcvars_RDS = par.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                           ninvars_NCS = par.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                           nvars_AUX   = par.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
        const Order gqo = par.get<Order>("Gauss quadrature order");
        const bool adaptive_integration = par.get<bool>("adaptive Gauss quadrature order");
        //
        // clear containers first
        this->elem_qi_local.clear();
        this->elem_side_qi_local.clear();
        //
        const MeshBase::const_element_iterator begin_el = msh_iga_p.get_parametric_mesh().active_local_elements_begin();
        const MeshBase::const_element_iterator end_el = msh_iga_p.get_parametric_mesh().active_local_elements_end();
        // loop over all active local (parametric) elements
        for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
            const NURBS* elem = &(msh_iga_p.get_nurbs().at((*citer_el)->subdomain_id()));
            //
            NonUniform_QGauss quad_volu( SP3D,
                                         static_cast<Order>((adaptive_integration ? elem->order(0) : 0)+(int)gqo),
                                         static_cast<Order>((adaptive_integration ? elem->order(1) : 0)+(int)gqo),
                                         static_cast<Order>((adaptive_integration ? elem->order(2) : 0)+(int)gqo) );
            //
            // access the quadrature, initialize and get access to the IGA stuff
            const std::vector< Point >& Qp = quad_volu.get_points();
            Point alpha, beta;
            iga_transf_3d(*citer_el, alpha, beta);
            // get access to the region
            if ( this->regn.end() == this->regn.find(elem->subdomain_id()) ) libmesh_error();
            const RegionInfo* ri_ptr = &(this->regn.find(elem->subdomain_id())->second);
            //
            const unsigned int n_qp = quad_volu.n_points();
            QuadratureInfo elem_quad(n_qp, nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
            // parametric coordinates of the quadrature points
            //const std::vector< Point >& qpnt_p = quad_volu.get_points();
            //
            if ( begin_el == citer_el ) {
                std::cout << "  > Volume integration points / 3D NURBS parametric element = " << n_qp << std::endl;
                flog.stream() << "  > Volume integration points / 3D NURBS parametric element = " << n_qp << std::endl;
            }
            // set the quadrature points material and spatial location
            for (unsigned int qp=0; qp<n_qp; qp++) {
                const Point Z = self_scale(alpha, Qp[qp]) + beta;
                const Point& qpnt = elem->get_xyz();
                elem->reinit(Z);
                //
                elem_quad.XYZ[qp] = qpnt;
                elem_quad.xyz[qp] = qpnt;
                //
                elem_quad.ORIENT[0][qp] = orientation_of_3D_element(*citer_el, 0);
                elem_quad.ORIENT[1][qp] = orientation_of_3D_element(*citer_el, 1);
                elem_quad.ORIENT[2][qp] = orientation_of_3D_element(*citer_el, 2);
            }
            // set the internal biomechanical and biochemical internal variables
            if ( ri_ptr->is_damaged ) {
                for (unsigned int qp=0; qp<n_qp; qp++) {
                    if ( ninvars_NCS == 0 ) {
                        ; // nothing to do here...
                    } else if ( ninvars_NCS == 3 ) {
                        elem_quad.biomech[qp][0] = material__max_damage_factor;
                        elem_quad.biomech[qp][1] = ri_ptr->mat->get_biomech_param(0) * (1.0-elem_quad.biomech[qp][0]);
                        elem_quad.biomech[qp][2] = 0.0;
                    } else libmesh_error();
                    for (unsigned int i=0; i<nvars_RDS; i++) {
                        elem_quad.biochem[qp][i] = initial_field_value__D__RDS[i];
                    }
                }
            } else {
                for (unsigned int qp=0; qp<n_qp; qp++) {
                    if ( ninvars_NCS == 0 ) {
                        ; // nothing to do here...
                    } else if ( ninvars_NCS == 3 ) {
                        elem_quad.biomech[qp][0] = material__min_damage_factor;
                        elem_quad.biomech[qp][1] = ri_ptr->mat->get_biomech_param(0) * (1.0-elem_quad.biomech[qp][0]);
                        elem_quad.biomech[qp][2] = 0.0;
                    } else libmesh_error();
                    for (unsigned int i=0; i<nvars_RDS; i++) {
                        elem_quad.biochem[qp][i] = initial_field_value__H__RDS[i];
                    }
                }
            }
            // set the fibres orientation vector in reference configuration
            for (unsigned int qp=0; qp<n_qp; qp++) {
                elem_quad.fibre_ref[0][qp] = vec(Point(uniform_distro(-1., +1.), uniform_distro(-1., +1.), uniform_distro(-1., +1.)).unit());
                elem_quad.fibre_ref[1][qp] = vec(Point(uniform_distro(-1., +1.), uniform_distro(-1., +1.), uniform_distro(-1., +1.)).unit());
                elem_quad.fibre_ref[2][qp] = vec(Point(uniform_distro(-1., +1.), uniform_distro(-1., +1.), uniform_distro(-1., +1.)).unit());
                elem_quad.fibre_curr[0][qp] = elem_quad.fibre_ref[0][qp];
                elem_quad.fibre_curr[1][qp] = elem_quad.fibre_ref[1][qp];
                elem_quad.fibre_curr[2][qp] = elem_quad.fibre_ref[2][qp];
            }
            // store data into the container
            this->elem_qi_local.insert( std::make_pair((*citer_el)->id(), elem_quad) );
            // end of active local (parametric) elements loop...
        }
        // check if membrane elements are present in analysis
        if ( par.get<bool>("membrane elements present") )
        // loop over all active local (parametric) elements
        for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
            const NURBS* elem = &(msh_iga_p.get_nurbs().at((*citer_el)->subdomain_id()));
            //
            std::map<unsigned int, QuadratureInfo> elem_quad;
            // loop for all element sides
            for (unsigned int s=0; s<(*citer_el)->n_sides(); s++) {
                // check if element has this side on boundary
                if ( (*citer_el)->neighbor_ptr(s) != NULL ) continue;
                // get the ID of the boundary
                const boundary_id_type bc_id = elem->boundary_id(s);
                if ( BoundaryInfo::invalid_id == bc_id ) libmesh_error();
                // get access to the region
                std::map<region_id_type, RegionInfo>::const_iterator ci = this->regn.find(bc_id);
                if ( this->regn.end() == ci ) continue;
                const RegionInfo* ri_ptr = &(ci->second);
                if ( ! ri_ptr->is_membrane ) continue;
                // create a side element
                UniquePtr<const NURBS> side = elem->build_side_ptr(s);
                //
                NonUniform_QGauss quad_surf( SP2D,
                                             static_cast<Order>((adaptive_integration ? side->order(0) : 0)+(int)gqo),
                                             static_cast<Order>((adaptive_integration ? side->order(1) : 0)+(int)gqo) );
                //
                // access the quadrature, initialize and get access to the IGA stuff
                const std::vector< Point >& Qp = quad_surf.get_points();
                Point alpha, beta;
                iga_transf_3d(*citer_el, alpha, beta, s);
                //
                const unsigned int n_qp = quad_surf.n_points();
                QuadratureInfo side_quad(n_qp, nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
                //
                if ( begin_el == citer_el ) {
                    std::cout << "  > Surface integration points / 2D NURBS parametric element = " << n_qp << std::endl;
                    flog.stream() << "  > Surface integration points / 2D NURBS parametric element = " << n_qp << std::endl;
                }
                // set the quadrature points material and spatial location
                for (unsigned int qp=0; qp<n_qp; qp++) {
                    const Point Z = self_scale(alpha, Qp[qp]) + beta;
                    const Point& qpnt = side->get_xyz();
                    side->reinit(Z);
                    //
                    side_quad.XYZ[qp] = qpnt;
                    side_quad.xyz[qp] = qpnt;
                }
                // set the internal biomechanical and biochemical internal variables
                if ( ri_ptr->is_damaged ) {
                    for (unsigned int qp=0; qp<n_qp; qp++) {
                        if ( ninvars_NCS == 0 ) {
                            ; // nothing to do here...
                        } else if ( ninvars_NCS == 3 ) {
                            side_quad.biomech[qp][0] = material__max_damage_factor;
                            side_quad.biomech[qp][1] = ri_ptr->mat->get_biomech_param(0) * (1.0-side_quad.biomech[qp][0]);
                            side_quad.biomech[qp][2] = 0.0;
                        } else libmesh_error();
                        for (unsigned int i=0; i<nvars_RDS; i++) {
                            side_quad.biochem[qp][i] = initial_field_value__D__RDS[i];
                        }
                    }
                } else {
                    for (unsigned int qp=0; qp<n_qp; qp++) {
                        if ( ninvars_NCS == 0 ) {
                            ; // nothing to do here...
                        } else if ( ninvars_NCS == 3 ) {
                            side_quad.biomech[qp][0] = material__min_damage_factor;
                            side_quad.biomech[qp][1] = ri_ptr->mat->get_biomech_param(0) * (1.0-side_quad.biomech[qp][0]);
                            side_quad.biomech[qp][2] = 0.0;
                        } else libmesh_error();
                        for (unsigned int i=0; i<nvars_RDS; i++) {
                            side_quad.biochem[qp][i] = initial_field_value__H__RDS[i];
                        }
                    }
                }
                // set the fibres orientation vector in reference configuration
                for (unsigned int qp=0; qp<n_qp; qp++) {
                    side_quad.fibre_ref[0][qp] = vec(Point(uniform_distro(-1., +1.), uniform_distro(-1., +1.), uniform_distro(-1., +1.)).unit());
                    side_quad.fibre_ref[1][qp] = vec(Point(uniform_distro(-1., +1.), uniform_distro(-1., +1.), uniform_distro(-1., +1.)).unit());
                    side_quad.fibre_ref[2][qp] = vec(Point(uniform_distro(-1., +1.), uniform_distro(-1., +1.), uniform_distro(-1., +1.)).unit());
                    side_quad.fibre_curr[0][qp] = side_quad.fibre_ref[0][qp];
                    side_quad.fibre_curr[1][qp] = side_quad.fibre_ref[1][qp];
                    side_quad.fibre_curr[2][qp] = side_quad.fibre_ref[2][qp];
                }
                // store data into the container
                elem_quad.insert( std::make_pair(s, side_quad) );
                // end of element sides loop...
            }
            // store data into the container
            if ( ! elem_quad.empty() )
                this->elem_side_qi_local.insert( std::make_pair((*citer_el)->id(), elem_quad) );
            // end of active local (parametric) elements loop...
        }
        // exception is caught...
    } else libmesh_error();
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // print on the screen the total number of (volume & surface) integration points
    {
        const int n_qp_local = this->elem_qi_local.size();
        int n_qp_global;
        MPI_Allreduce(&n_qp_local, &n_qp_global, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
        std::cout << "  > Volume integration points in total = " << n_qp_global << std::endl;
        flog.stream() << "  > Volume integration points in total = " << n_qp_global << std::endl;
        //
        this->_equation_systems.parameters.set<unsigned int>("volume integration points (local)") =
            static_cast<unsigned int>(n_qp_local);
        this->_equation_systems.parameters.set<unsigned int>("volume integration points (global)") =
            static_cast<unsigned int>(n_qp_global);
    }
    if ( par.get<bool>("membrane elements present") )
    {
        const int n_qp_local = this->elem_side_qi_local.size();
        int n_qp_global;
        MPI_Allreduce(&n_qp_local, &n_qp_global, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
        std::cout << "  > Surface integration points in total = " << n_qp_global << std::endl;
        flog.stream() << "  > Surface integration points in total = " << n_qp_global << std::endl;
        //
        this->_equation_systems.parameters.set<unsigned int>("surface integration points (local)") =
            static_cast<unsigned int>(n_qp_local);
        this->_equation_systems.parameters.set<unsigned int>("surface integration points (global)") =
            static_cast<unsigned int>(n_qp_global);
    }
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // check if to read (overwrite) quadrature information from binary file
    if ( par.get<unsigned int>("input Quadrature from file") ) {
        const unsigned int n_files = par.get<unsigned int>("input Quadrature from file");
        // loop over all binary files to read
        for (unsigned int ifile=0; ifile<n_files; ifile++) {
            char file_name[buffer_len];
            sprintf(file_name, "input Quadrature file name #%d", ifile);
            const std::string name = par.get<std::string>(file_name);
            //
            std::ifstream file(name);
            //
            // read the total number of active local elements' quadrature info to write
            unsigned int nelem = 0;
            file.read((char*)&nelem, sizeof(unsigned int));
            // loop over all active local elements
            for (unsigned int ielem=0; ielem<nelem; ielem++) {
                // read the element ID of these quadrature points
                unsigned int elem_id;
                file.read((char*)&elem_id, sizeof(unsigned int));
                // read the side of the element (only for 2D elements)
                int elem_side = -1;
                file.read((char*)&elem_side, sizeof(int));
                libmesh_assert( elem_side == -1 );
                //
                QuadratureInfo elem_quad;
                // now read the quadrature data from the binary file
                elem_quad.load(file, QuadratureInfo::REFERENCE, QuadratureInfo::LEVEL_3);
                // store data into the container
                this->elem_qi_local.insert( std::make_pair(elem_id, elem_quad) );
                // end of active local elements loop...
            }
            //
            // check if membrane elements are present in analysis
            if ( par.get<bool>("membrane elements present") )
                WARNING_("this part of code is under development");
            // end of binary files loop...
        }
        // end of if case...
    }
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // ...exit this function normally now
}
//=================================================================================================
void DomainData::init_rve () {
    flog.print("  > Init representative volume elements\n", false);
    //
    EquationSystems& es = this->get_equation_systems();
    const Parameters& par = es.parameters;
    //
    if ( ! par.get<bool>("initialise RVEs") ) return;
    //
    // access the list of IGA meshes
    std::map<std::string, MeshIGA>& msh_iga =
        *(es.parameters.set<std::map<std::string, MeshIGA>*>("pointer to IGA parametric mesh(es)"));
    // access the list of RVE meshes
    std::map<std::string, MeshRVE>& msh_rve =
        *(es.parameters.set<std::map<std::string, MeshRVE>*>("pointer to RVE parametric mesh(es)"));
    // access the list of RVE equation systems
    std::map<std::string, EquationSystems>& es_rve =
        *(es.parameters.set<std::map<std::string, EquationSystems>*>("pointer to RVE equation system(s)"));
    //
    // Now set-up the meshes for the RVEs of the multiscale problem. Each quadrature point
    // (corresponding to an finite element in the FEM or a parametric element in the IGA)
    // will have a unique RVE. Introduce some default general parameters for the RVE meshes,
    // applicable to all in the analysis, but could be changed later on.
    //
    // make sure you clean the memory first
    msh_rve.clear();
    es_rve.clear();
    //
    // check if we run a FEM simulation
    if ( par.get<bool>("FEM framework") ) {
        const MeshBase& msh = es.get_mesh();
        //
        // access the iterators for the finite elements of the FEM mesh
        MeshBase::const_element_iterator begin_el = msh.active_local_elements_begin();
        MeshBase::const_element_iterator end_el   = msh.active_local_elements_end();
        // loop over all active finite elements
        for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
            const Elem* elem = *citer_el;
            // access the material region
            const RegionInfo* ri = &(this->regn.find(elem->subdomain_id())->second);
            // retrieve the total number of quadrature point for this element
            ASSERT_(this->elem_qi_local.end()!=this->elem_qi_local.find(elem->id()), "fatal internal error occurred");
            const QuadratureInfo* qpi_ptr = &(this->elem_qi_local.find(elem->id())->second);
            const unsigned int n_qp = qpi_ptr->n_points();
            //
            // now check if this region facilitates a multiscale model
            const int multiscale_model = ri->mat->get_biomech_param(100);
            if ( 0 == multiscale_model ) continue;
            //
            // loop for all quadrature points
            for (unsigned int iqp=0; iqp<n_qp; iqp++) {
                // key to insert item into map
                const std::string key = number2string(elem->id()) + "_" + number2string(iqp);
                //
                // insert the RVE mesh into the map container
                msh_rve.insert( std::make_pair(key, MeshRVE(*lcomm_ptr)) );
                // access the recently-added mesh
                MeshRVE& msh_ = msh_rve.find(key)->second;
                // insert the RVE equations system into the map container
                es_rve.insert( std::make_pair(key, EquationSystems(msh_)) );
                // access the recently-added equation systems
                EquationSystems& es_ = es_rve.find(key)->second;
                // set-up some very important parameters
                es_.parameters.set<RVE*>("pointer to RVE mesh") = &msh_;
                es_.parameters.set<const std::vector<Real>*>("pointer to RVE parameters") = &ri->mat->get_biomech_param();
                es_.parameters.set<Point>("RVE: origin point space vector") = qpi_ptr->XYZ[iqp];
                es_.parameters.set<Point>("RVE: orientation vector #0") = qpi_ptr->ORIENT[0][iqp];
                es_.parameters.set<Point>("RVE: orientation vector #1") = qpi_ptr->ORIENT[1][iqp];
                es_.parameters.set<Point>("RVE: orientation vector #2") = qpi_ptr->ORIENT[2][iqp];
                // end of quadrature points loop...
            }
            // end of active finite elements loop...
        }
    // check if we run an IGA simulation
    } else if ( par.get<bool>("IGA framework") ) {
        const std::string label =
            par.get<std::string>("IGA parametric mesh to use for analysis");
        const MeshIGA& msh_iga_p = msh_iga.find(label)->second;
        //
        // access the iterators for the parametric elements of the IGA mesh
        const MeshBase::const_element_iterator begin_el = msh_iga_p.get_parametric_mesh().active_local_elements_begin();
        const MeshBase::const_element_iterator end_el   = msh_iga_p.get_parametric_mesh().active_local_elements_end();
        // loop over all active parametric elements of a NURBS element
        for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
            const NURBS* elem = &(msh_iga_p.get_nurbs().at((*citer_el)->subdomain_id()));
            // access the material region
            const RegionInfo* ri = &(this->regn.find(elem->subdomain_id())->second);
            // retrieve the total number of quadrature point for this element
            ASSERT_(this->elem_qi_local.end()!=this->elem_qi_local.find((*citer_el)->id()), "fatal internal error occurred");
            const QuadratureInfo* qpi_ptr = &(this->elem_qi_local.find((*citer_el)->id())->second);
            const unsigned int n_qp = qpi_ptr->n_points();
            //
            // now check if this region facilitates a multiscale model
            const int multiscale_model = ri->mat->get_biomech_param(100);
            if ( 0 == multiscale_model ) continue;
            //
            // loop for all quadrature points
            for (unsigned int iqp=0; iqp<n_qp; iqp++) {
                // key to insert item into map
                const std::string key = number2string((*citer_el)->id()) + "_" + number2string(iqp);
                //
                // insert the RVE mesh into the map container
                msh_rve.insert( std::make_pair(key, MeshRVE(*lcomm_ptr)) );
                // access the recently-added mesh
                MeshRVE& msh_ = msh_rve.find(key)->second;
                // insert the RVE equations system into the map container
                es_rve.insert( std::make_pair(key, EquationSystems(msh_)) );
                // access the recently-added equation systems
                EquationSystems& es_ = es_rve.find(key)->second;
                // set-up some very important parameters
                es_.parameters.set<RVE*>("pointer to RVE mesh") = &msh_;
                es_.parameters.set<const std::vector<Real>*>("pointer to RVE parameters") = &ri->mat->get_biomech_param();
                es_.parameters.set<Point>("RVE: origin point space vector") = qpi_ptr->XYZ[iqp];
                es_.parameters.set<Point>("RVE: orientation vector #0") = qpi_ptr->ORIENT[0][iqp];
                es_.parameters.set<Point>("RVE: orientation vector #1") = qpi_ptr->ORIENT[1][iqp];
                es_.parameters.set<Point>("RVE: orientation vector #2") = qpi_ptr->ORIENT[2][iqp];
                // end of quadrature points loop...
            }
            // end of active parametric elements loop...
        }
    } else libmesh_error();
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    // Now access all RVEs and initialise their meshes and prepare them for the analysis. Here,
    // to simplify the multiscale modelling, we assume that all quadrature points of all elements
    // of a particular (material or subdomain) region employ the same 1D-mesh for an RVE.
    //
    // check if we run a FEM simulation
    if ( par.get<bool>("FEM framework") ) {
        char par_name[buffer_len];
        const MeshBase& msh = es.get_mesh();
        //
        // access the iterators for the finite elements of the FEM mesh
        const MeshBase::const_element_iterator begin_el = msh.active_local_elements_begin();
        const MeshBase::const_element_iterator end_el   = msh.active_local_elements_end();
        // loop over all active finite elements
        for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
            const Elem* elem = *citer_el;
            // access the material region
            const RegionInfo* ri = &(this->regn.find(elem->subdomain_id())->second);
            // retrieve the total number of quadrature point for this element
            const unsigned int n_qp = this->elem_qi_local.find(elem->id())->second.n_points();
            //
            // now check if this region facilitates a multiscale model
            const int multiscale_model = ri->mat->get_biomech_param(100);
            if ( 0 == multiscale_model ) continue;
            //
            // loop for all quadrature points
            for (unsigned int iqp=0; iqp<n_qp; iqp++) {
                // key to access item from map
                const std::string key = number2string(elem->id()) + "_" + number2string(iqp);
                //
                // now distinguish amongst the types of multiscale models available
                if ( 0 == multiscale_model ) {
                    ; // do nothing here...
                } else if ( 1 == multiscale_model ) {
                    // access the recently-added mesh and equation systems
                    MeshRVE& msh_ = msh_rve.find(key)->second;
                    EquationSystems& es_ = es_rve.find(key)->second;
                    // obtain the corresponding filename of the Gmsh file
                    sprintf(par_name, "region %d: input RVE mesh file", elem->subdomain_id());
                    const std::string fname = par.get<std::string>(par_name);
                    // now read the Gmsh file and prepare it for use
                    GmshIO(msh_).read(fname);
                    msh_.reinit();
                    // and finally initialise the equation systems
                    collagen_extracellular_matrix::multiscale_3___setup_system(es_);
                } else libmesh_error();
                // end of quadrature points loop...
            }
            // end of active finite elements loop...
        }
    // check if we run an IGA simulation
    } else if ( par.get<bool>("IGA framework") ) {
        char par_name[buffer_len];
        const std::string label =
            par.get<std::string>("IGA parametric mesh to use for analysis");
        const MeshIGA& msh_iga_p = msh_iga.find(label)->second;
        //
        // access the iterators for the parametric elements of the IGA mesh
        const MeshBase::const_element_iterator begin_el = msh_iga_p.get_parametric_mesh().active_local_elements_begin();
        const MeshBase::const_element_iterator end_el   = msh_iga_p.get_parametric_mesh().active_local_elements_end();
        // loop over all active parametric elements of a NURBS element
        for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
            const NURBS* elem = &(msh_iga_p.get_nurbs().at((*citer_el)->subdomain_id()));
            // access the material region
            const RegionInfo* ri = &(this->regn.find(elem->subdomain_id())->second);
            // retrieve the total number of quadrature points for this element
            const unsigned int n_qp = this->elem_qi_local.find((*citer_el)->id())->second.n_points();
            //
            // now check if this region facilitates a multiscale model
            const int multiscale_model = ri->mat->get_biomech_param(100);
            if ( 0 == multiscale_model ) continue;
            //
            // loop for all quadrature points
            for (unsigned int iqp=0; iqp<n_qp; iqp++) {
                // key to access item from map
                const std::string key = number2string((*citer_el)->id()) + "_" + number2string(iqp);
                //
                // now distinguish amongst the types of multiscale models available
                if ( 0 == multiscale_model ) {
                    ; // do nothing here...
                } else if ( 1 == multiscale_model ) {
                    // access the recently-added mesh and equation systems
                    MeshRVE& msh_ = msh_rve.find(key)->second;
                    EquationSystems& es_ = es_rve.find(key)->second;
                    // obtain the corresponding filename of the Gmsh file
                    sprintf(par_name, "region %d: input RVE mesh file", elem->subdomain_id());
                    const std::string fname = par.get<std::string>(par_name);
                    // now read the Gmsh file and prepare it for use
                    GmshIO(msh_).read(fname);
                    msh_.reinit();
                    // and finally initialise the equation systems
                    collagen_extracellular_matrix::multiscale_3___setup_system(es_);
                } else libmesh_error();
                // end of quadrature points loop...
            }
            // end of active parametric elements loop...
        }
    } else libmesh_error();
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // ...exit this function normally now
}
//=================================================================================================
#ifndef BIODYNAMO_ENABLED
//=================================================================================================
void DomainData::init_bdm () {
    //
    ;
    //
}
//=================================================================================================
void DomainData::run_bdm () {
    //
    ;
    //
}
//=================================================================================================
void DomainData::clear_bdm () {
    //
    ;
    //
}
//=================================================================================================
void DomainData::write_bdm (const char* fname) const {
    //
    ;
    //
}
//=================================================================================================
#endif
//=================================================================================================
//
//
//
//=================================================================================================
void DomainData_VascularNetwork::prepare_mesh () {
    const Parameters& par = this->_equation_systems_3d.parameters;
    // access the vascular network 1D mesh
    MeshBase& msh_vn = this->_equation_systems.get_mesh();
    //
    // read the vascular network 1D finite element mesh
    flog.print(" >> Read the vascular network mesh\n", false);
    GmshIO(msh_vn).read(par.get<std::string>("input vascular network file").c_str());
    //
    // initialise this local map - prior to preparing the mesh for use
    nodal_id2ptr_vn__map.clear();
    for (MeshBase::node_iterator n_iter=msh_vn.nodes_begin(); n_iter!=msh_vn.nodes_end(); n_iter++) {
        Node* node = (*n_iter);
        nodal_id2ptr_vn__map.insert( std::make_pair(node->id(), node) );
    }
    //
    // prepare the vascular network mesh
    flog.print(" >> Prepare vascular network mesh for use\n", false);
    msh_vn.prepare_for_use(par.get<bool>("input vascular network domain data: skip renumbering"));
    // end of FE mesh manipulation
    flog.print("                                                \n", false);
    msh_vn.print_info();
    msh_vn.print_info(flog.stream());
    //msh_vn.print_info(ferr.stream());
    //msh_vn.boundary_info->print_info(flog.stream());
    //msh_vn.boundary_info->print_info(ferr.stream());
    flog.print("                                                \n", false);
}
//=================================================================================================
void DomainData_VascularNetwork::init () {
    const MeshBase& msh_vn = this->_equation_systems.get_mesh();
    const Parameters& par = this->_equation_systems_3d.parameters;
    //
    flog.print(" >> Init domain data for vascular network\n", false);
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    MeshBase::const_element_iterator elem_begin = msh_vn.active_elements_begin();
    MeshBase::const_element_iterator elem_end   = msh_vn.active_elements_end();
    // calculate the min/max segment length of all active 1D finite elements
    this->_equation_systems_3d.parameters.set<Real>("vascular network: min segment length") = +1.0e+22;
    this->_equation_systems_3d.parameters.set<Real>("vascular network: max segment length") = -1.0e+22;
    for (MeshBase::const_element_iterator citer_el=elem_begin; citer_el!=elem_end; citer_el++) {
        const Elem* elem = *citer_el;
        ASSERT_(libMesh::EDGE2==elem->type(), "fatal internal error occurred");
        //
        const Real length = Point(elem->point(0)-elem->point(1)).norm();
        //
        this->_equation_systems_3d.parameters.set<Real>("vascular network: min segment length") =
            std::min(length, par.get<Real>("vascular network: min segment length"));
        //
        this->_equation_systems_3d.parameters.set<Real>("vascular network: max segment length") =
            std::max(length, par.get<Real>("vascular network: max segment length"));
        // end of active elements loop...
    }
    // print out the characteristic lengths of the vascular network
    std::cout << "  > Vascular Network mininum segment length = " << std::scientific << par.get<Real>("vascular network: min segment length") << std::endl;
    flog.stream() << "  > Vascular Network mininum segment length = " << std::scientific << par.get<Real>("vascular network: min segment length") << std::endl;
    std::cout << "  > Vascular Network maximum segment length = " << std::scientific << par.get<Real>("vascular network: max segment length") << std::endl;
    flog.stream() << "  > Vascular Network maximum segment length = " << std::scientific << par.get<Real>("vascular network: max segment length") << std::endl;
    //
    // initialise the mesh by either doing this from scratch or from user-provided data
    if ( "NA" == par.get<std::string>("input vascular network domain data file") ) {
        this->_init_self();
    } else {
        this->_init_extern();
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // store a copy of the input Gmsh formatted mesh file
    if ( libMesh::global_processor_id() == 0 ) {
        const std::string fname = par.get<std::string>("results directory") + "/"
                                + par.get<std::string>("output file name") + "-vn-initial.msh";
        this->save_mesh(fname.c_str());
    }
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
}
//=================================================================================================
void DomainData_VascularNetwork::_init_self () {
    const Parameters& par = this->_equation_systems_3d.parameters;
    std::string fname = par.get<std::string>("results directory: temporary") + "/"
                      + par.get<std::string>("output file name")
                      + "-initial_vascular_network.dmn";
    //
    if ( 0 == libMesh::global_processor_id() ) {
        this->_init_self_p0();
        this->save_all(fname.c_str());
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    if ( 0 != libMesh::global_processor_id() ) {
        this->load_all(fname.c_str());
    }
}
//=================================================================================================
void DomainData_VascularNetwork::_init_self_p0 () {
    const MeshBase& msh_3d = this->_equation_systems_3d.get_mesh();
    MeshBase&       msh_vn = this->_equation_systems.get_mesh();
    const Parameters& par = this->_equation_systems_3d.parameters;
    //
    MeshBase::const_element_iterator elem3d_begin = msh_3d.active_elements_begin();
    MeshBase::const_element_iterator elem3d_end   = msh_3d.active_elements_end();
    //
    MeshBase::const_node_iterator node_begin = msh_vn.nodes_begin();
    MeshBase::const_node_iterator node_end   = msh_vn.nodes_end();
    //
    MeshBase::const_element_iterator elem_begin = msh_vn.active_elements_begin();
    MeshBase::const_element_iterator elem_end   = msh_vn.active_elements_end();
    //
    flog.print("  > Init vascular network... 1/7                                     \r", false);
    ferr.print("  > Init vascular network... 1/7                                     \r", true );
    //
    // clear container first
    this->node_vi.clear();
    // insert all network nodes in a list to monitor their state
    for (MeshBase::const_node_iterator n_citer=node_begin; n_citer!=node_end; n_citer++) {
        const Node* node = (*n_citer);
        //
        VascularInfo vasc_info;
        //
        // locate the three-dimensional element that containts this vascular node
        vasc_info.elem3d_id = DofObject::invalid_id;
        for (MeshBase::const_element_iterator e_citer=elem3d_begin; e_citer!=elem3d_end; e_citer++) {
            const Elem* elem_3d = *e_citer;
            //
            if ( elem_3d->contains_point(*node) ) {
                vasc_info.elem3d_id = elem_3d->id();
                break;
            }
        }
        // validity check
        ASSERT_(DofObject::invalid_id!=vasc_info.elem3d_id, "fatal internal error occurred");
        //
        vasc_info.status = VascularInfo::NORMAL;
        vasc_info.flow = VascularInfo::INVALID_FLOW;
        vasc_info.type = VascularInfo::INVALID_TYPE;
        //
        // insert default vascular info for this node
        this->node_vi.insert( std::make_pair(node->id(), vasc_info) );
        // end of vascular nodes loop...
    }
    //
    flog.print("  > Init vascular network... 2/7                                     \r", false);
    ferr.print("  > Init vascular network... 2/7                                     \r", true );
    //
    std::set<dof_id_type> nodes_scanned;
    // set the flow type for each vascular node as well as the corresponding connecting
    // vascular segments to each node
    for (MeshBase::const_element_iterator e_citer=elem_begin; e_citer!=elem_end; e_citer++) {
        const Elem* elem = *e_citer;
        ASSERT_(libMesh::EDGE2==elem->type(), "fatal internal error occurred");
        const VascularInfo::Flow f = static_cast<VascularInfo::Flow>(elem->subdomain_id());
        //
        for (unsigned int l=0; l<2; l++) {
            const dof_id_type node_id = elem->node_id(l);
            //
            ASSERT_(this->node_vi.end()!=this->node_vi.find(node_id), "fatal internal error occurred");
            VascularInfo* vi = &(this->node_vi.find(node_id)->second);
            ASSERT_(NULL!=vi, "fatal internal error occurred");
            //
            // by default enter this vascular segment into the list
            vi->vsc_elem.insert( elem->id() );
            //
            if ( nodes_scanned.end() != nodes_scanned.find(node_id) ) continue;
            nodes_scanned.insert(node_id);
            //
            // set the flow type based on the vessel (one-dimensional) element subdomain ID
            vi->flow = f;
        }
        // end of vascular elements (segments) loop...
    }
    //
    flog.print("  > Init vascular network... 3/7                                     \r", false);
    ferr.print("  > Init vascular network... 3/7                                     \r", true );
    //
    // enforce cleaning containers first
    this->node_inlet.clear();
    this->node_outlet.clear();
    // initialize containers of boundary nodes of vascular network
    WARNING_("ensure that physical IDs of vascular network for inlet/outlet boundary nodes are correct");
    do {
        std::vector<dof_id_type> node_id_list;
        std::vector<boundary_id_type> bc_id_list;
        msh_vn.boundary_info->build_node_list(node_id_list, bc_id_list);
        //
        const unsigned int n_node = node_id_list.size();
        ASSERT_(n_node==bc_id_list.size(), "fatal internal error occurred");
        for (unsigned int i=0; i<n_node; i++) {
            const dof_id_type node_id = node_id_list[i];
            const VascularInfo::Type bc_id = static_cast<VascularInfo::Type>(bc_id_list[i]);
            //
            ASSERT_(this->node_vi.end()!=this->node_vi.find(node_id), "fatal internal error occurred");
            VascularInfo* vi = &(this->node_vi.find(node_id)->second);
            ASSERT_(NULL!=vi, "fatal internal error occurred");
            // check for ID of boundary node of vascular network and then insert into containers
            if        ( VascularInfo::TIP    == bc_id ) {
                vi->type = VascularInfo::TIP;
            } else if ( VascularInfo::INLET  == bc_id ) {
                vi->type = VascularInfo::INLET;
                //
                this->node_inlet.insert( node_id );
            } else if ( VascularInfo::OUTLET == bc_id ) {
                vi->type = VascularInfo::OUTLET;
                //
                this->node_outlet.insert( node_id );
            } else ERROR_("error in boundary ID of vascular network");
        }
        // end of do-while loop...
    } while ( false );
    //
    flog.print("  > Init vascular network... 4/7                                     \r", false);
    ferr.print("  > Init vascular network... 4/7                                     \r", true );
    //
    WARNING_("ensure that never a vascular network trunk shares more than one inlets");
    WARNING_("ensure that vascular inlets do not go straight to a branch");
    WARNING_("ensure that vascular outlets do not come straight from a branch");
    // construct the vascular network tree starting from the inlet(s) towards the outlet(s):
    // find the previous and following nodes to each node
    std::map< dof_id_type, std::set<dof_id_type> > node__elem1d;
    //
    for (MeshBase::const_node_iterator n_citer=node_begin; n_citer!=node_end; n_citer++) {
        const Node* node = *n_citer;
        //
        node__elem1d.insert( std::make_pair(node->id(), std::set<dof_id_type>()) );
        // end of vascular nodes loop...
    }
    //
    for (MeshBase::const_element_iterator e_citer=elem_begin; e_citer!=elem_end; e_citer++) {
        const Elem* elem = *e_citer;
        ASSERT_(libMesh::EDGE2==elem->type(), "fatal internal error occurred");
        //
        for (unsigned int l=0; l<2; l++) {
            node__elem1d.find(elem->node_id(l))->second.insert( elem->id() );
        }
        // end of active vascular elements loop...
    }
    //
    for (MeshBase::const_node_iterator n_citer=node_begin; n_citer!=node_end; n_citer++) {
        const Node* node = *n_citer;
        //
        ASSERT_(this->node_vi.end()!=this->node_vi.find(node->id()), "fatal internal error occurred");
        VascularInfo* vi = &(this->node_vi.find(node->id())->second);
        ASSERT_(NULL!=vi, "fatal internal error occurred");
        //
        const std::set<dof_id_type> elem1d_ls = node__elem1d.find(node->id())->second;
        //
        if ( elem1d_ls.empty() ) {
            ASSERT_(VascularInfo::INVALID_TYPE==vi->type, "vascular information not properly initialised");
            vi->status = VascularInfo::COLLAPSED;
        } else if ( elem1d_ls.size() > 3 ) {
            ERROR_("vascular network node is associated with more than three elements");
        } else if ( elem1d_ls.size() == 3 ) {
            ASSERT_(VascularInfo::INVALID_TYPE==vi->type, "vascular information not properly initialised");
            vi->type = VascularInfo::BRANCH;
        } else if ( elem1d_ls.size() == 2 ) {
            // ASSERT_(VascularInfo::INVALID_TYPE==vi->type, "vascular information not properly initialised");
            vi->type = VascularInfo::SPROUT;
        } else if ( elem1d_ls.size() == 1 ) {
            if ( VascularInfo::TIP == vi->type ) {
                ; // nothing to do here...
            } else if ( this->node_inlet.end()  != this->node_inlet.find(node->id())  ) {
                ASSERT_(VascularInfo::INLET==vi->type, "vascular information not properly initialised");
            } else if ( this->node_outlet.end() != this->node_outlet.find(node->id()) ) {
                ASSERT_(VascularInfo::OUTLET==vi->type, "vascular information not properly initialised");
            } else {
                const int F = uniform_distro(-1, +1);
                if ( F>0 ) {
                    this->node_inlet.insert(node->id());
                    vi->type = VascularInfo::INLET;
                } else {
                    this->node_outlet.insert(node->id());
                    vi->type = VascularInfo::OUTLET;
                }
            }
        } else {
            ERROR_("vascular node not associated with any 3D elements; check your input");
        }
        // end of vascular nodes loop...
    }
    //
    flog.print("  > Init vascular network... 5/7                                     \r", false);
    ferr.print("  > Init vascular network... 5/7                                     \r", true );
    //
    WARNING_("ensure you have read vascular network total number of branches");
    WARNING_("ensure you have read vascular network level for each branch");
    // read the information of the vascular network
    do {
        GetPot info(par.get<std::string>("input vascular network information file").c_str());

        // set the default value first
        VascularInfo::Type blood_vessel_default_type;
        {
            const std::string str = get_par<std::string>(info, "blood/vascular_type/default");
            if      ( str == "MATURE" ) blood_vessel_default_type = VascularInfo::MATURE;
            else if ( str == "SPROUT" ) blood_vessel_default_type = VascularInfo::SPROUT;
            else ERROR_("wrong value for default value of vascular network information inserted");
        }
        VascularInfo::Type lymph_vessel_default_type;
        {
            const std::string str = get_par<std::string>(info, "lymph/vascular_type/default");
            if      ( str == "MATURE" ) lymph_vessel_default_type = VascularInfo::MATURE;
            else if ( str == "SPROUT" ) lymph_vessel_default_type = VascularInfo::SPROUT;
            else ERROR_("wrong value for default value of vascular network information inserted");
        }
        for ( std::map<dof_id_type, VascularInfo>::iterator
              iter=this->node_vi.begin(); iter!=this->node_vi.end(); iter++ ) {
            VascularInfo* vi = &(iter->second);
            //
            if ( VascularInfo::COLLAPSED == vi->status ) continue;
            if ( VascularInfo::INVALID_TYPE == vi->type ) libmesh_error();
            //
            if ( VascularInfo::INLET  == vi->type ||
                 VascularInfo::OUTLET == vi->type ||
                 VascularInfo::BRANCH == vi->type ||
                 VascularInfo::TIP    == vi->type ) continue;
            //
            if      ( VascularInfo::BLOOD == vi->flow ) vi->type = blood_vessel_default_type;
            else if ( VascularInfo::LYMPH == vi->flow ) vi->type = lymph_vessel_default_type;
            else ERROR_("fatal internal error occurred");
        }
        //
        // checks
        ASSERT_(2==get_par_size<Real>(info, "blood/vascular_radius/default"), "fatal error with input occurred");
        ASSERT_(2==get_par_size<Real>(info, "blood/vascular_thickness/default"), "fatal error with input occurred");
        ASSERT_(2==get_par_size<Real>(info, "blood/vascular_pore_radius/default"), "fatal error with input occurred");
        ASSERT_(2==get_par_size<Real>(info, "blood/vascular_pores_fraction/default"), "fatal error with input occurred");
        ASSERT_(2==get_par_size<Real>(info, "blood/vascular_critical_pressure_ratio/default"), "fatal error with input occurred");
        ASSERT_(2==get_par_size<Real>(info, "lymph/vascular_radius/default"), "fatal error with input occurred");
        ASSERT_(2==get_par_size<Real>(info, "lymph/vascular_thickness/default"), "fatal error with input occurred");
        ASSERT_(2==get_par_size<Real>(info, "lymph/vascular_pore_radius/default"), "fatal error with input occurred");
        ASSERT_(2==get_par_size<Real>(info, "lymph/vascular_pores_fraction/default"), "fatal error with input occurred");
        ASSERT_(2==get_par_size<Real>(info, "lymph/vascular_critical_pressure_ratio/default"), "fatal error with input occurred");
        // set the default values for parent blood vessels (mean and standard deviation)
        const Real blood_mean_radius         = get_par_comp<Real>(info, "blood/vascular_radius/default", 0);
        const Real blood_mean_thickness      = get_par_comp<Real>(info, "blood/vascular_thickness/default", 0);
        const Real blood_mean_pore_radius    = get_par_comp<Real>(info, "blood/vascular_pore_radius/default", 0);
        const Real blood_mean_pores_fraction = get_par_comp<Real>(info, "blood/vascular_pores_fraction/default", 0);
        const Real blood_mean_press_ratio_cr = get_par_comp<Real>(info, "blood/vascular_critical_pressure_ratio/default", 0);
        const Real blood_dev_radius         = get_par_comp<Real>(info, "blood/vascular_radius/default", 1);
        const Real blood_dev_thickness      = get_par_comp<Real>(info, "blood/vascular_thickness/default", 1);
        const Real blood_dev_pore_radius    = get_par_comp<Real>(info, "blood/vascular_pore_radius/default", 1);
        const Real blood_dev_pores_fraction = get_par_comp<Real>(info, "blood/vascular_pores_fraction/default", 1);
        const Real blood_dev_press_ratio_cr = get_par_comp<Real>(info, "blood/vascular_critical_pressure_ratio/default", 1);
        // set the default values for parent lymph vessels (mean and standard deviation)
        const Real lymph_mean_radius         = get_par_comp<Real>(info, "lymph/vascular_radius/default", 0);
        const Real lymph_mean_thickness      = get_par_comp<Real>(info, "lymph/vascular_thickness/default", 0);
        const Real lymph_mean_pore_radius    = get_par_comp<Real>(info, "lymph/vascular_pore_radius/default", 0);
        const Real lymph_mean_pores_fraction = get_par_comp<Real>(info, "lymph/vascular_pores_fraction/default", 0);
        const Real lymph_mean_press_ratio_cr = get_par_comp<Real>(info, "lymph/vascular_critical_pressure_ratio/default", 0);
        const Real lymph_dev_radius         = get_par_comp<Real>(info, "lymph/vascular_radius/default", 1);
        const Real lymph_dev_thickness      = get_par_comp<Real>(info, "lymph/vascular_thickness/default", 1);
        const Real lymph_dev_pore_radius    = get_par_comp<Real>(info, "lymph/vascular_pore_radius/default", 1);
        const Real lymph_dev_pores_fraction = get_par_comp<Real>(info, "lymph/vascular_pores_fraction/default", 1);
        const Real lymph_dev_press_ratio_cr = get_par_comp<Real>(info, "lymph/vascular_critical_pressure_ratio/default", 1);
        //
        {
            Parameters* par_ptr = &(this->_equation_systems_3d.parameters);
            //
            par_ptr->set<std::string>("all: input vascular network domain data file") =
                get_par<std::string>(info, "all/vascular_network_domain_data_file", "NA");
            if ( "NA" != par.get<std::string>("input vascular network domain data file") )
                ASSERT_("NA"==par.get<std::string>("all: input vascular network domain data file"), "fatal error with input occurred");
            //
            if ( "NA" != par.get<std::string>("all: input vascular network domain data file") ) {
                ASSERT_(5==get_par_size<Real>(info, "all/vascular_network_domain_data_file/list"), "fatal error with input occurred");
                for (int i=0; i<5; i++) {
                    const std::string str =
                        get_par_comp<std::string>(info, "all/vascular_network_domain_data_file/list", i);
                    ASSERT_(str=="default" || str=="custom", "fatal internal error occurred");
                }
                //
                par_ptr->set<std::string>("all: input vascular network domain data file; radius")                  = get_par_comp<std::string>(info, "all/vascular_network_domain_data_file/list", 0);
                par_ptr->set<std::string>("all: input vascular network domain data file; thickness")               = get_par_comp<std::string>(info, "all/vascular_network_domain_data_file/list", 1);
                par_ptr->set<std::string>("all: input vascular network domain data file; pore radius")             = get_par_comp<std::string>(info, "all/vascular_network_domain_data_file/list", 2);
                par_ptr->set<std::string>("all: input vascular network domain data file; pores fraction")          = get_par_comp<std::string>(info, "all/vascular_network_domain_data_file/list", 3);
                par_ptr->set<std::string>("all: input vascular network domain data file; critical pressure ratio") = get_par_comp<std::string>(info, "all/vascular_network_domain_data_file/list", 4);
                //
                // would be silly to have all values being default in the extra file...
                if ( "default" == par.get<std::string>("all: input vascular network domain data file; radius")                  &&
                     "default" == par.get<std::string>("all: input vascular network domain data file; thickness")               &&
                     "default" == par.get<std::string>("all: input vascular network domain data file; pore radius")             &&
                     "default" == par.get<std::string>("all: input vascular network domain data file; pores fraction")          &&
                     "default" == par.get<std::string>("all: input vascular network domain data file; critical pressure ratio") )
                    ERROR_("fatal internal error occurred");
            }
        }
        //
        // iterate for all vascular nodes in the network
        for ( std::map<dof_id_type, VascularInfo>::iterator
              iter=this->node_vi.begin(); iter!=this->node_vi.end(); iter++ )
        {
            VascularInfo* vi = &(iter->second);
            //
            if ( VascularInfo::BLOOD == vi->flow ) {
                // check about vascular node status
                if ( VascularInfo::COLLAPSED == vi->status ) {
                    // set the default (minimum) values for a collapsed vascular node
                    vi->radius         = blood_mean_radius/1000.0;
                    vi->thickness      = blood_mean_thickness/1000.0;
                    vi->pore_radius    = blood_mean_pore_radius/1000.0;
                    vi->pores_fraction = blood_mean_pores_fraction/1000.0;
                    vi->press_ratio_cr = blood_mean_press_ratio_cr;
                    vi->compression_scale = 1000.0;
                } else {
                    // set the default (random) values for a normal/compressed vascular node
                    vi->radius         = normal_distro(blood_mean_radius        , blood_dev_radius);
                    vi->thickness      = normal_distro(blood_mean_thickness     , blood_dev_thickness);
                    vi->pore_radius    = normal_distro(blood_mean_pore_radius   , blood_dev_pore_radius);
                    vi->pores_fraction = normal_distro(blood_mean_pores_fraction, blood_dev_pores_fraction);
                    vi->press_ratio_cr = normal_distro(blood_mean_press_ratio_cr, blood_dev_press_ratio_cr);
                    vi->compression_scale = 1.0;
                }
            } else if ( VascularInfo::LYMPH == vi->flow ) {
                // check about vascular node status
                if ( VascularInfo::COLLAPSED == vi->status ) {
                    // set the default (minimum) values for a collapsed vascular node
                    vi->radius         = lymph_mean_radius/1000.0;
                    vi->thickness      = lymph_mean_thickness/1000.0;
                    vi->pore_radius    = lymph_mean_pore_radius/1000.0;
                    vi->pores_fraction = lymph_mean_pores_fraction/1000.0;
                    vi->press_ratio_cr = lymph_mean_press_ratio_cr;
                    vi->compression_scale = 1000.0;
                } else {
                    // set the default (random) values for a normal/compressed vascular node
                    vi->radius         = normal_distro(lymph_mean_radius        , lymph_dev_radius);
                    vi->thickness      = normal_distro(lymph_mean_thickness     , lymph_dev_thickness);
                    vi->pore_radius    = normal_distro(lymph_mean_pore_radius   , lymph_dev_pore_radius);
                    vi->pores_fraction = normal_distro(lymph_mean_pores_fraction, lymph_dev_pores_fraction);
                    vi->press_ratio_cr = normal_distro(lymph_mean_press_ratio_cr, lymph_dev_press_ratio_cr);
                    vi->compression_scale = 1.0;
                }
            } else ERROR_("fatal internal error occurred");
            // end of vascular nodes loop...
        }
        //
        // check if to load pre-defined info data for blood/lymph vessel nodes
        if ( "NA" != par.get<std::string>("all: input vascular network domain data file") ) {
            std::ifstream fin(par.get<std::string>("all: input vascular network domain data file"));
            ASSERT_(fin.good(), "fatal internal error occurred");
            //
            unsigned int n_nodes;
            fin >> n_nodes;
            ASSERT_(msh_vn.n_nodes()==n_nodes, "fatal internal error occurred");
            for (unsigned int I=0; I<n_nodes; I++) {
                unsigned int inode;
                fin >> inode;
                --inode;
                ASSERT_(inode==I, "fatal internal error occurred");
                //
                ASSERT_(nodal_id2ptr_vn__map.end()!=nodal_id2ptr_vn__map.find(inode), "fatal internal error occurred");
                const dof_id_type node_id = nodal_id2ptr_vn__map.find(inode)->second->id();
                //
                ASSERT_(this->node_vi.end()!=this->node_vi.find(node_id), "fatal internal error occurred");
                VascularInfo* vi = &(this->node_vi.find(node_id)->second);
                //
                if ( "custom" == par.get<std::string>("all: input vascular network domain data file; radius") ) {
                    Real custom_value = 0.0;
                    fin >> custom_value;
                    vi->radius         = custom_value;
                }
                if ( "custom" == par.get<std::string>("all: input vascular network domain data file; thickness") ) {
                    Real custom_value = 0.0;
                    fin >> custom_value;
                    vi->thickness      = custom_value;
                }
                if ( "custom" == par.get<std::string>("all: input vascular network domain data file; pore radius") ) {
                    Real custom_value = 0.0;
                    fin >> custom_value;
                    vi->pore_radius    = custom_value;
                }
                if ( "custom" == par.get<std::string>("all: input vascular network domain data file; pores fraction") ) {
                    Real custom_value = 0.0;
                    fin >> custom_value;
                    vi->pores_fraction = custom_value;
                }
                if ( "custom" == par.get<std::string>("all: input vascular network domain data file; critical pressure ratio") ) {
                    Real custom_value = 0.0;
                    fin >> custom_value;
                    vi->press_ratio_cr = custom_value;
                }
            }
            // end of if case...
        }
        // end of do-while loop...
    } while ( false );
    //
    flog.print("  > Init vascular network... 6/7                                     \r", false);
    ferr.print("  > Init vascular network... 6/7                                     \r", true );
    //
    // clear container first
    this->branch__node.clear();
    // set this parameter of the equations systems for the vascular network
    this->get_equation_systems().parameters.set<unsigned int>("total number of branches") = 0;
    // check if to calculate the branching IDs
    if ( ! par.get<bool>("input vascular network domain data: skip branching") ) {
        // loop for all inlets and start building the vascular tree from these nodes
        std::set<dof_id_type> scanned_vascular_nodes;
        //
        for ( std::set<dof_id_type>::const_iterator
              n_ci=this->node_inlet.begin(); n_ci!=this->node_inlet.end(); n_ci++ ) {
            const Node* node = msh_vn.node_ptr(*n_ci);
            //
            ASSERT_(this->node_vi.end()!=this->node_vi.find(node->id()), "fatal internal error occurred");
            const VascularInfo* vi = &(this->node_vi.find(node->id())->second);
            ASSERT_(NULL!=vi, "fatal internal error occurred");
            //
            if ( VascularInfo::INVALID_FLOW == vi->flow ) libmesh_error();
            if ( VascularInfo::INVALID_TYPE == vi->type ) libmesh_error();
            if ( VascularInfo::COLLAPSED == vi->status ) continue;
            //
            recursive_vascular_node_find( this->get_mesh(), node->id(), this->node_vi, node__elem1d, this->branch__node,
                                          scanned_vascular_nodes,
                                          this->get_equation_systems().parameters.set<unsigned int>("total number of branches") );
            //
            // increment the total number of branches
            this->get_equation_systems().parameters.set<unsigned int>("total number of branches") += 1;
            // end of inlet nodes loop...
        }
    } else {
        //
        const unsigned int branch_id = this->get_equation_systems().parameters.get<unsigned int>("total number of branches");
        //
        {
            std::set<dof_id_type> node_list;
            //
            this->branch__node.insert( std::make_pair(branch_id, node_list) );
        }
        //
        for ( std::set<dof_id_type>::const_iterator
              n_ci=this->node_inlet.begin(); n_ci!=this->node_inlet.end(); n_ci++ ) {
            const Node* node = msh_vn.node_ptr(*n_ci);
            //
            ASSERT_(this->node_vi.end()!=this->node_vi.find(node->id()), "fatal internal error occurred");
            VascularInfo* vi = &(this->node_vi.find(node->id())->second);
            ASSERT_(NULL!=vi, "fatal internal error occurred");
            //
            if ( VascularInfo::INVALID_FLOW == vi->flow ) libmesh_error();
            if ( VascularInfo::INVALID_TYPE == vi->type ) libmesh_error();
            if ( VascularInfo::COLLAPSED == vi->status ) continue;
            //
            std::map< dof_id_type, std::set<dof_id_type> >::iterator iter = this->branch__node.find(branch_id);
            if ( this->branch__node.end() == iter ) {
                ERROR_("fatal internal error occurred");
            } else {
                iter->second.insert( node->id() );
            }
            //
            vi->branch_id = branch_id;
        }
        //
        // increment the total number of branches
        this->get_equation_systems().parameters.set<unsigned int>("total number of branches") += 1;
    }
    //
    flog.print("  > Init vascular network... 7/7                                     \r", false);
    ferr.print("  > Init vascular network... 7/7                                     \r", true );
    //
    // clear container first
    this->elem3d__node.clear();
    //
    for (MeshBase::const_element_iterator e_citer=elem3d_begin; e_citer!=elem3d_end; e_citer++) {
        const Elem* elem3d = *e_citer;
        //
        this->elem3d__node.insert( std::make_pair(elem3d->id(), std::set<dof_id_type>()) );
    }
    //
    for ( std::map<dof_id_type, VascularInfo>::const_iterator
          ci=this->node_vi.begin(); ci!=this->node_vi.end(); ci++ ) {
        const dof_id_type& node_id = ci->first;
        const dof_id_type& elem3d_id = ci->second.elem3d_id;
        //
        this->elem3d__node.find(elem3d_id)->second.insert(node_id);
    }
    //
    WARNING_("ensure to set previous/following nodes in the initial vascular network tree");
}
//=================================================================================================
void DomainData_VascularNetwork::_init_extern () {
    const Parameters& par = this->_equation_systems_3d.parameters;
    //
    this->load_all(par.get<std::string>("input vascular network domain data file").c_str());
}
//=================================================================================================
void DomainData_VascularNetwork::save_mesh (const char* fname) const {
    std::ofstream fout(fname);
    //
    std::set<dof_id_type> node_tip;
    for ( std::map<dof_id_type, VascularInfo>::const_iterator
          ci=this->node_vi.begin(); ci!=this->node_vi.end(); ci++ ) {
        const VascularInfo* vi = &(ci->second);
        if ( VascularInfo::COLLAPSED == vi->status ||
             VascularInfo::TIP       == vi->type   )
            node_tip.insert(ci->first);
        if ( VascularInfo::INVALID_TYPE == vi->type )
            ERROR_("fatal internal error occurred");
    }
    //
    fout << "$MeshFormat" << std::endl;
    fout << 2.2 << ' ' << 0 << ' ' << 8 << std::endl;
    fout << "$EndMeshFormat" << std::endl;
    //
    fout << "$Nodes" << std::endl;
    fout << this->get_mesh().n_nodes() << std::endl;
    for ( MeshBase::const_node_iterator
          ci=this->get_mesh().nodes_begin(); ci!=this->get_mesh().nodes_end(); ci++ ) {
        const dof_id_type node_id = 1+(*ci)->id();
        const Point& node_coord = this->get_mesh().point((*ci)->id());
        fout << node_id
             << ' ' << node_coord(0) << ' ' << node_coord(1) << ' ' << node_coord(2)
             << std::endl;
    }
    fout << "$EndNodes" << std::endl;
    //
    fout << "$Elements" << std::endl;
    int elem_id = this->node_inlet.size() + this->node_outlet.size() + node_tip.size();
    for ( MeshBase::const_element_iterator
          ci=this->get_mesh().elements_begin(); ci!=this->get_mesh().elements_end(); ci++ ) {
        ++elem_id;
    }
    fout << elem_id << std::endl;
    elem_id = 1;
    for ( MeshBase::const_element_iterator
          ci=this->get_mesh().elements_begin(); ci!=this->get_mesh().elements_end(); ci++ ) {
        const Elem* elem = (*ci);
        const dof_id_type node_0_id = 1+elem->node_id(0);
        const dof_id_type node_1_id = 1+elem->node_id(1);
        fout << elem_id << ' ' << 1 << ' ' << 2 << ' ' << elem->subdomain_id() << ' ' << 0
             << ' ' << node_0_id << ' ' << node_1_id
             << std::endl;
         if ( 1+elem->id() != elem_id ) libmesh_error();
        //
        ++elem_id;
    }
    for ( std::set<dof_id_type>::const_iterator
          ci=this->node_inlet.begin(); ci!=this->node_inlet.end(); ci++ ) {
        const dof_id_type node_id = 1+(*ci);
        fout << elem_id << ' ' << 15 << ' ' << 2 << ' ' << VascularInfo::INLET << ' ' << 0
             << ' ' << node_id
             << std::endl;
        //
        ++elem_id;
    }
    for ( std::set<dof_id_type>::const_iterator
          ci=this->node_outlet.begin(); ci!=this->node_outlet.end(); ci++ ) {
        const dof_id_type node_id = 1+(*ci);
        fout << elem_id << ' ' << 15 << ' ' << 2 << ' ' << VascularInfo::OUTLET << ' ' << 0
             << ' ' << node_id
             << std::endl;
        //
        ++elem_id;
    }
    for ( std::set<dof_id_type>::const_iterator
          ci=node_tip.begin(); ci!=node_tip.end(); ci++ ) {
        const dof_id_type node_id = 1+(*ci);
        fout << elem_id << ' ' << 15 << ' ' << 2 << ' ' << VascularInfo::TIP << ' ' << 0
             << ' ' << node_id
             << std::endl;
        //
        ++elem_id;
    }
    fout << "$EndElements" << std::endl;
}
//=================================================================================================
//
//
//
//=================================================================================================
void DomainData_SolidCoupledWithVascularNetwork::prepare_mesh (const DomainData_VascularNetwork& dmn_vn) {
    flog.print(" >> Prepare coupled-problem mesh for use\n", false);
    // access the coupled 3D mesh (extracellular matrix & vascular network)
    MeshBase& msh_coupled = this->_equation_systems.get_mesh();
    const Parameters& par = this->_equation_systems_3d.parameters;
    //
    // clear containers first
    this->strm_node__cm_strm_node.clear();
    this->cm_strm_node__strm_node.clear();
    this->vasc_node__cm_vasc_node.clear();
    this->cm_vasc_node__vasc_node.clear();
    this->cm_strm_node__boundary.clear();
    this->cm_strm_node__cm_strm_node_segm.clear();
    this->cm_strm_node__cm_vasc_node_segm.clear();
    this->cm_strm_elem__subdmnid.clear();
    this->cm_vasc_elem__vasc_elem.clear();
    this->vasc_elem__cm_vasc_elem.clear();
    this->cm_vasc_node__cm_intf_elem.clear();
    //
    WARNING_("make sure that the mesh is either TET- or HEX-dominant, no support for mixed meshes");
    //
    // construct the network arising from the 3D interstitial elements & nodes, and
    // from the initial vascular network of elements & nodes
    const MeshBase& msh_3d = this->_equation_systems_3d.get_mesh();
    const MeshBase& msh_vn = dmn_vn.get_equation_systems().get_mesh();
    //
    // insert the nodes of the 3D tissue mesh: stromal points
    flog.print("  > Prepare coupled-problem mesh (1/6)          \r", false);
    do {
        // iterate for all nodes of the extracellular matrix
        for ( MeshBase::const_node_iterator
              citer_nd=msh_3d.nodes_begin(); citer_nd!=msh_3d.nodes_end(); citer_nd++ ) {
            const Node* node = *citer_nd;
            const dof_id_type& new_node_id = node->id();
            //
            msh_coupled.add_point(*node, new_node_id);
            //
            // insert the pair of the stroma node between the coupled-mesh and the 3D mesh ...and vice versa
            this->strm_node__cm_strm_node.insert( std::make_pair(node->id(), new_node_id) );
            this->cm_strm_node__strm_node.insert( std::make_pair(new_node_id, node->id()) );
            // end of nodes loop...
        }
        // find the boundary nodes to the extracellular matrix
        for ( MeshBase::const_element_iterator
              citer_el=msh_3d.active_elements_begin(); citer_el!=msh_3d.active_elements_end(); citer_el++ ) {
            const Elem* elem = *citer_el;
            //
            for (unsigned int s=0; s<elem->n_sides(); s++) {
                if ( elem->neighbor_ptr(s) != NULL ) continue;
                //
                UniquePtr<const Elem> side = elem->build_side_ptr(s);
                for (unsigned int l=0; l<side->n_nodes(); l++) {
                    this->cm_strm_node__boundary.insert( this->strm_node__cm_strm_node.find(side->node_id(l))->second );
                    // end of side nodes loop...
                }
                // end of element sides loop...
            }
            // end of active elements loop...
        }
    } while ( false );
    //
    // insert the segments of the 3D tissue mesh: stromal elements
    flog.print("  > Prepare coupled-problem mesh (2/6)          \r", false);
    do {
        std::set<std::string> inserted_pair_of_nodes;
        // iterate for all active 3D elements of the extracellular matrix
        for ( MeshBase::const_element_iterator
              citer_el=msh_3d.active_elements_begin(); citer_el!=msh_3d.active_elements_end(); citer_el++ ) {
            const Elem* elem = *citer_el;
            //
            for (unsigned int i=0; i<elem->n_edges(); i++) {
                UniquePtr<const Elem> edge = elem->build_edge_ptr(i);
                const Node* node_0 = edge->node_ptr(0);
                const Node* node_1 = edge->node_ptr(1);
                // skip following if both ECM nodes have been already processed
                char ns[buffer_len];
                sprintf(ns, "%014d%014d", std::min(node_0->id(), node_1->id()), std::max(node_0->id(), node_1->id()));
                if ( inserted_pair_of_nodes.end() != inserted_pair_of_nodes.find(std::string(ns)) ) continue;
                //
                Elem* tissue = Elem::build(EDGE2).release();
                tissue->set_id( msh_coupled.n_elem() );
                tissue->set_node(0) = msh_coupled.node_ptr( this->strm_node__cm_strm_node.find(node_0->id())->second );
                tissue->set_node(1) = msh_coupled.node_ptr( this->strm_node__cm_strm_node.find(node_1->id())->second );
                tissue->subdomain_id() = static_cast<subdomain_id_type>(DomainData_SolidCoupledWithVascularNetwork::STROMA); // NOTE
                //
                msh_coupled.add_elem(tissue);
                //
                // insert the stroma (segment) element in the list
                this->cm_strm_elem__subdmnid.insert( std::make_pair(tissue->id(), elem->subdomain_id()) );
                //
                inserted_pair_of_nodes.insert(std::string(ns));
                // end of element edges loop...
            }
            // end of active elements loop...
        }
    } while ( false );
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    // insert the nodes of the vascular network mesh: vascular nodes
    flog.print("  > Prepare coupled-problem mesh (3/6)          \r", false);
    do {
        // iterate for all nodes of the vascular network
        for ( MeshBase::const_node_iterator
              citer_nd=msh_vn.nodes_begin(); citer_nd!=msh_vn.nodes_end(); citer_nd++ ) {
            const Node* node = *citer_nd;
            const dof_id_type& cm_vascular_node_id = static_cast<dof_id_type>(node->id() + msh_3d.n_nodes());
            //
            msh_coupled.add_point(*node, cm_vascular_node_id);
            //
            // insert the pair of the vascular node between the coupled-mesh and the vascular network ...and vice versa
            this->vasc_node__cm_vasc_node.insert( std::make_pair(node->id(), cm_vascular_node_id) );
            this->cm_vasc_node__vasc_node.insert( std::make_pair(cm_vascular_node_id, node->id()) );
            // end of vascular nodes loop...
        }
    } while ( false );
    //
    // insert the segments of the vascular network mesh: vascular elements & interface elements
    flog.print("  > Prepare coupled-problem mesh (4/6)          \r", false);
    do {
        // iterate for all active 1D elements of the vascular network
        for ( MeshBase::const_element_iterator
              citer_el=msh_vn.active_elements_begin(); citer_el!=msh_vn.active_elements_end(); citer_el++ ) {
            const Elem* elem = *citer_el;
            const dof_id_type& cm_vascular_node_0_id = this->vasc_node__cm_vasc_node.find(elem->node_id(0))->second;
            const dof_id_type& cm_vascular_node_1_id = this->vasc_node__cm_vasc_node.find(elem->node_id(1))->second;
            //
            Elem* vessel = Elem::build(EDGE2).release();
            vessel->set_id( msh_coupled.n_elem() );
            vessel->set_node(0) = msh_coupled.node_ptr( cm_vascular_node_0_id );
            vessel->set_node(1) = msh_coupled.node_ptr( cm_vascular_node_1_id );
            vessel->subdomain_id() = static_cast<subdomain_id_type>(DomainData_SolidCoupledWithVascularNetwork::VESSEL); // NOTE
            //
            msh_coupled.add_elem(vessel);
            //
            // insert the pair of the vascular (segment) element between the coupled-mesh and the vascular network ...and vice versa
            this->cm_vasc_elem__vasc_elem.insert( std::make_pair(vessel->id(), elem->id()) );
            this->vasc_elem__cm_vasc_elem.insert( std::make_pair(elem->id(), vessel->id()) );
            // end of active elements loop...
        }
        // construct the network that the interfaces the initial vascular nodes and the tissue nodes
        for ( MeshBase::const_node_iterator
              citer_nd=msh_vn.nodes_begin(); citer_nd!=msh_vn.nodes_end(); citer_nd++ ) {
            const Node* node_0 = *citer_nd;
            const dof_id_type& cm_vascular_node_id = this->vasc_node__cm_vasc_node.find(node_0->id())->second;
            //
            std::map<dof_id_type, VascularInfo>::const_iterator ci = dmn_vn.node_vi.find(node_0->id());
            if ( ci == dmn_vn.node_vi.end() ) libmesh_error();
            //
            std::set<dof_id_type> cm_elem_interface;
            //
            const Elem* elem3d = msh_3d.elem_ptr(ci->second.elem3d_id);
            ASSERT_(elem3d, "three-dimensional finite element not found in the stored mesh");
            for (unsigned int i=0; i<elem3d->n_nodes(); i++) {
                const Node* node_1 = elem3d->node_ptr(i);
                const dof_id_type& cm_stromal_node_id = this->strm_node__cm_strm_node.find(node_1->id())->second;
                //
                Elem* interface = Elem::build(EDGE2).release();
                interface->set_id( msh_coupled.n_elem() );
                interface->set_node(0) = msh_coupled.node_ptr( cm_vascular_node_id );
                interface->set_node(1) = msh_coupled.node_ptr( cm_stromal_node_id  );
                interface->subdomain_id() = static_cast<subdomain_id_type>(DomainData_SolidCoupledWithVascularNetwork::INTERFACE); // NOTE
                //
                msh_coupled.add_elem(interface);
                // insert the interface (segment) element in the list
                cm_elem_interface.insert(interface->id());
                //
                std::map<dof_id_type, std::set<pair_dof_id_type> >::iterator iter =
                    this->cm_strm_node__cm_vasc_node_segm.find( cm_stromal_node_id );
                if ( this->cm_strm_node__cm_vasc_node_segm.end() == iter ) {
                    std::set<pair_dof_id_type> ini_set;
                    ini_set.insert( pair_dof_id_type(cm_vascular_node_id, interface->id()) );
                    //
                    this->cm_strm_node__cm_vasc_node_segm.insert( std::make_pair(cm_stromal_node_id, ini_set) );
                } else {
                    iter->second.insert( pair_dof_id_type(cm_vascular_node_id, interface->id()) );
                }
            }
            //
            // insert for this vascular node the list of associated interface elements
            this->cm_vasc_node__cm_intf_elem.insert( std::make_pair(cm_vascular_node_id, cm_elem_interface) );
            // end of vascular nodes loop...
        }
    } while ( false );
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    flog.print("  > Prepare coupled-problem mesh (5/6)          \r", false);
    // prepare the coupled mesh for use; enforce to skip node renumbering
    msh_coupled.prepare_for_use(true);
    //
    flog.print("  > Prepare coupled-problem mesh (6/6)          \r", false);
    do {
        const MeshBase::const_element_iterator begin_el = msh_coupled.active_elements_begin();
        const MeshBase::const_element_iterator end_el = msh_coupled.active_elements_end();
        //
        for ( MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
            const Elem* elem = *citer_el;
            //
            switch ( static_cast<DomainData_SolidCoupledWithVascularNetwork::Type>(elem->subdomain_id()) ) {
            case DomainData_SolidCoupledWithVascularNetwork::STROMA :
                {
                    const dof_id_type& cm_stromal_node_0_id = elem->node_id(0);
                    const dof_id_type& cm_stromal_node_1_id = elem->node_id(1);
                    std::map<dof_id_type, std::set<pair_dof_id_type> >::iterator iter =
                        this->cm_strm_node__cm_strm_node_segm.find( cm_stromal_node_0_id );
                    if ( this->cm_strm_node__cm_strm_node_segm.end() == iter ) {
                        std::set<pair_dof_id_type> ini_set;
                        ini_set.insert( pair_dof_id_type(cm_stromal_node_1_id, elem->id()) );
                        //
                        this->cm_strm_node__cm_strm_node_segm.insert( std::make_pair(cm_stromal_node_0_id, ini_set) );
                    } else {
                        iter->second.insert( pair_dof_id_type(cm_stromal_node_1_id, elem->id()) );
                    }
                }
                {
                    const dof_id_type& cm_stromal_node_0_id = elem->node_id(0);
                    const dof_id_type& cm_stromal_node_1_id = elem->node_id(1);
                    std::map<dof_id_type, std::set<pair_dof_id_type> >::iterator iter =
                        this->cm_strm_node__cm_strm_node_segm.find( cm_stromal_node_1_id );
                    if ( this->cm_strm_node__cm_strm_node_segm.end() == iter ) {
                        std::set<pair_dof_id_type> ini_set;
                        ini_set.insert( pair_dof_id_type(cm_stromal_node_0_id, elem->id()) );
                        //
                        this->cm_strm_node__cm_strm_node_segm.insert( std::make_pair(cm_stromal_node_1_id, ini_set) );
                    } else {
                        iter->second.insert( pair_dof_id_type(cm_stromal_node_0_id, elem->id()) );
                    }
                }
                break;
            case DomainData_SolidCoupledWithVascularNetwork::VESSEL :
            case DomainData_SolidCoupledWithVascularNetwork::INTERFACE :
                break;
            default :
                libmesh_error();
            }
            // end of active (coupled-mesh) elements loop...
        }
    } while ( false );
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // end of coupled FE mesh manipulation
    flog.print("                                                \n", false);
    msh_coupled.print_info();
    msh_coupled.print_info(flog.stream());
    // msh_coupled.print_info(ferr.stream());
    flog.print("                                                \n", false);
    //
    // store a copy of the input Gmsh formatted coupled mesh file
    if ( libMesh::global_processor_id() == 0 )
    {
        char cstr[buffer_len];
        const std::string results_dir  = par.get<std::string>("results directory"),
                          output_fname = par.get<std::string>("output file name"),
                          sim_title    = par.get<std::string>("simulation title");
        sprintf( cstr, "%s/temporary/%s-%s-0-coupled.msh",
                 results_dir.c_str(), output_fname.c_str(), sim_title.c_str() );
        GmshIO(msh_coupled).write(cstr);
    }
}
//=================================================================================================
void DomainData_SolidCoupledWithVascularNetwork::init (const DomainData_VascularNetwork& dmn_vn) {
    flog.print(" >> Init domain data for coupled-problem\n", false);
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
}
//=================================================================================================
