/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/drug_delivery_solver.h"
#include "libmesh/linear_implicit_system.h"
#include "libmesh/petsc_linear_solver.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/quadrature_gauss.h"
#include "libmesh/dof_map.h"
#include "feb3/fem-context.h"
#include "feb3/tumour_angiogenesis_constitutive.h"
#include "feb3/ida.h"
//=================================================================================================
// local variables
// ...description of each variable is provided in the code below!
static
std::vector< std::vector<Real> >  blv_sf_dp0, blv_sf_dp1, blv_P_dp0, blv_P_dp1,
                                  lyv_sf_dp0, lyv_sf_dp1, lyv_P_dp0, lyv_P_dp1;
static
std::vector< std::vector<Real> >  blv_Peff, blv_Kp, blv_S,
                                  lyv_Peff, lyv_Kp, lyv_S;
static
std::vector< std::vector<int> >  n_permeable_blood_vessels, n_permeable_lymph_vessels;
static
std::vector< std::vector<Real> >  intravasc_drug;
static
std::vector< int >  n_perfused_blood_vessels, n_perfused_lymph_vessels;
// external variables
extern
OFStream flog, ferr;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
extern
TumourAngiogenesisConstitutive ta_constitutive;
// local variables
Real (*drug0_injection_) (const std::vector<Real>&, const Real);
Real (*drug1_injection_) (const std::vector<Real>&, const Real);
//=================================================================================================
#include "feb3/domain_data_inl.h"
#include "feb3/solution_io_inl.h"
//=================================================================================================
void DrugDeliverySolver::_injection_498 () {
    const Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    EquationSystems& es_vn = this->_dmn_vn.get_equation_systems();
    const System& vn_flow_sys = es_vn.get_system<System>(par.get<std::string>("tumour angiogenesis solver: label")+"::flow_solution");
    System&   g_sys = es_vn.get_system<System>(this->_label+"::drug_injection::solution");
    System& g_P_sys = es_vn.get_system<System>(this->_label+"::drug_injection::previous_solution");
    //
    // sanity checks
    ASSERT_(8==vn_flow_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(this->par_list.get<unsigned int>("number of drugs")==2, "fatal internal error occurred");
    ASSERT_(2==g_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(2==g_P_sys.n_vars(), "fatal internal error occurred");
    //
    const MeshBase::const_element_iterator begin_el = es_vn.get_mesh().active_local_elements_begin();
    const MeshBase::const_element_iterator end_el   = es_vn.get_mesh().active_local_elements_end();
    const MeshBase::const_node_iterator begin_nd = es_vn.get_mesh().local_nodes_begin();
    const MeshBase::const_node_iterator end_nd   = es_vn.get_mesh().local_nodes_end();
    //
    libmesh_assert( this->_dmn.mater.end() !=
                    this->_dmn.mater.find(par.get<int>("tumour angiogenesis solver: vascular remodelling material ID")) );
    const MaterialInfo* mi__vnr = &(this->_dmn.mater.find(par.get<int>("tumour angiogenesis solver: vascular remodelling material ID"))->second);
    libmesh_assert( NULL != mi__vnr );
    //
    // obtain the flow-field data of the vascular (blood|lymph) network
    std::vector<Number> vn_flow;
    vn_flow_sys.update_global_solution(vn_flow);
    //
    if      (  0 == static_cast<int>(mi__vnr->get_biochem_param(30)) ) drug0_injection_ = &flat_function;
    else if ( 10 == static_cast<int>(mi__vnr->get_biochem_param(30)) ) drug0_injection_ = &step_function;
    else if ( 11 == static_cast<int>(mi__vnr->get_biochem_param(30)) ) drug0_injection_ = &step_function__periodic;
    else if ( 20 == static_cast<int>(mi__vnr->get_biochem_param(30)) ) drug0_injection_ = &ramp_function;
    else if ( 21 == static_cast<int>(mi__vnr->get_biochem_param(30)) ) drug0_injection_ = &ramp_function__periodic;
    else if ( 30 == static_cast<int>(mi__vnr->get_biochem_param(30)) ) drug0_injection_ = &exp_function;
    else if ( 31 == static_cast<int>(mi__vnr->get_biochem_param(30)) ) drug0_injection_ = &exp_function__periodic;
    else if ( 40 == static_cast<int>(mi__vnr->get_biochem_param(30)) ) drug0_injection_ = &logistic_function;
    else if ( 41 == static_cast<int>(mi__vnr->get_biochem_param(30)) ) drug0_injection_ = &logistic_function__periodic;
    else ERROR_("fatal internal error occurred");
    std::vector<Real> drug0_injection_params;
    for (unsigned int j=31; j<=39; j++)
        drug0_injection_params.push_back( mi__vnr->get_biochem_param(j) );
    //
    if      (  0 == static_cast<int>(mi__vnr->get_biochem_param(40)) ) drug1_injection_ = &flat_function;
    else if ( 10 == static_cast<int>(mi__vnr->get_biochem_param(40)) ) drug1_injection_ = &step_function;
    else if ( 11 == static_cast<int>(mi__vnr->get_biochem_param(40)) ) drug1_injection_ = &step_function__periodic;
    else if ( 20 == static_cast<int>(mi__vnr->get_biochem_param(40)) ) drug1_injection_ = &ramp_function;
    else if ( 21 == static_cast<int>(mi__vnr->get_biochem_param(40)) ) drug1_injection_ = &ramp_function__periodic;
    else if ( 30 == static_cast<int>(mi__vnr->get_biochem_param(40)) ) drug1_injection_ = &exp_function;
    else if ( 31 == static_cast<int>(mi__vnr->get_biochem_param(40)) ) drug1_injection_ = &exp_function__periodic;
    else if ( 40 == static_cast<int>(mi__vnr->get_biochem_param(40)) ) drug1_injection_ = &logistic_function;
    else if ( 41 == static_cast<int>(mi__vnr->get_biochem_param(40)) ) drug1_injection_ = &logistic_function__periodic;
    else ERROR_("fatal internal error occurred");
    std::vector<Real> drug1_injection_params;
    for (unsigned int j=41; j<=49; j++)
        drug1_injection_params.push_back( mi__vnr->get_biochem_param(j) );
    //
    const Real injected_drug_concentration[] = { drug0_injection_(drug0_injection_params, par.get<Real>("current time")) ,
                                                 drug1_injection_(drug1_injection_params, par.get<Real>("current time")) },
               zero_drug_concentration     = 0.0,
               low_bound_transport_velocity = mi__vnr->get_biochem_param(201);
    if ( low_bound_transport_velocity <= 0.0 ) WARNING_("check your input for this model parameter");
    //
    // copy the current solution to the previous one
    g_P_sys.solution->operator=(*g_sys.solution);
    g_P_sys.solution->close();
    g_P_sys.update();
    //
    // zero-out the current solution first
    g_sys.solution->zero();
    //
    // evaluate the distribution of the drug in the blood vessels' network
    // loop over all local (vascular) nodes
    for (MeshBase::const_node_iterator citer_nd=begin_nd; citer_nd!=end_nd; citer_nd++) {
        const Node* node = (*citer_nd);
        //
        libmesh_assert( this->_dmn_vn.node_vi.end() != this->_dmn_vn.node_vi.find(node->id()) );
        const VascularInfo* vi = &(this->_dmn_vn.node_vi.find(node->id())->second);
        ASSERT_(NULL!=vi, "fatal internal error occurred");
        //
        const dof_id_type idof[] = { node->dof_number(g_sys.number(), 0, 0) ,
                                     node->dof_number(g_sys.number(), 1, 0) };
        //
        if ( VascularInfo::BLOOD == vi->flow ) {
            //
            const Real Velc = get_system_solution_nodal_value<6>(vn_flow_sys, node, vn_flow);
            if ( VascularInfo::COLLAPSED != vi->status && Velc >= low_bound_transport_velocity ) {
                g_sys.solution->set(idof[0], injected_drug_concentration[0]);
                g_sys.solution->set(idof[1], injected_drug_concentration[1]);
            } else {
                g_sys.solution->set(idof[0], zero_drug_concentration);
                g_sys.solution->set(idof[1], zero_drug_concentration);
            }
            //
        } else if ( VascularInfo::LYMPH == vi->flow ) {
            //
            g_sys.solution->set(idof[0], zero_drug_concentration);
            g_sys.solution->set(idof[1], zero_drug_concentration);
            //
        } else libmesh_error();
        // end of local (vascular) nodes loop...
    }
    //
    // close the solution vector and update the system
    g_sys.solution->close();
    g_sys.update();
}
//=================================================================================================
void DrugDeliverySolver::_delivery_498 () {
    const bool use_forward_Euler = true;//false;//true;

    const Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    const EquationSystems& es_vn = this->_dmn_vn.get_equation_systems();
    const System& vn_flow_sys = es_vn.get_system<System>(par.get<std::string>("tumour angiogenesis solver: label")+"::flow_solution");
    const System& vn_drug_sys = es_vn.get_system<System>(this->_label+"::drug_injection::solution");
    //
    EquationSystems& es_3d = this->_dmn.get_equation_systems();
    LinearImplicitSystem& g_sys = es_3d.get_system<LinearImplicitSystem>(this->_label+"::solution");
    const System&       g_P_sys = es_3d.get_system<System>(this->_label+"::previous_solution");
    const System&      DgDt_sys = es_3d.get_system<System>(this->_label+"::solution_time_derivative");
    const System& s3d_flow_sys  = es_3d.get_system<System>(par.get<std::string>("Navier-Cauchy solver: label")+"::"
                                                          +par.get<std::string>("tumour angiogenesis solver: label")+"::coupled_solution");
    //
    libmesh_assert( this->_dmn.mater.end() !=
                    this->_dmn.mater.find(par.get<int>("tumour angiogenesis solver: vascular remodelling material ID")) );
    const MaterialInfo* mi__vnr = &(this->_dmn.mater.find(par.get<int>("tumour angiogenesis solver: vascular remodelling material ID"))->second);
    libmesh_assert( NULL != mi__vnr );
    //
    // obtain the flow-field data of the vascular (blood|lymph) network
    std::vector<Number> vn_flow;
    vn_flow_sys.update_global_solution(vn_flow);
    //
    // obtain the current solution of the injected (intravenus) drug concentration
    std::vector<Number> vn_drug;
    vn_drug_sys.update_global_solution(vn_drug);
    //
    // obtain the previous solution of the drug concentration in the 3D domain
    std::vector<Number> s3d_drug;
    g_P_sys.update_global_solution(s3d_drug);
    std::vector<Number> s3d_drug_rate;
    if ( ! use_forward_Euler ) {
        DgDt_sys.update_global_solution(s3d_drug_rate);
    }
    //
    // obtain the flow-field data of the 3D (stromal+parenchymal) domain
    std::vector<Number> s3d_flow;
    s3d_flow_sys.update_global_solution(s3d_flow);
    //
    //
    const FEM_Context g_ctx(g_sys);
    const MeshBase::const_element_iterator begin_el = es_3d.get_mesh().active_local_elements_begin();
    const MeshBase::const_element_iterator end_el   = es_3d.get_mesh().active_local_elements_end();
    //
    const unsigned int nvars = this->par_list.get<unsigned int>("dimension");
    std::vector<unsigned short int> var;
    g_ctx.get_var_list(var);
    //
    FEType fe_g_type = g_sys.variable_type(var[0]);
    //
    const bool update_vasc_params = par.get<bool>("update vascular parameters for drug delivery solver");
    //
    const unsigned int ninvars_NCS = par.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = par.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = par.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = par.get<unsigned int>("reaction-diffusion solver: auxiliary variables"),
                       ndofs_DRUG  = this->par_list.get<unsigned int>("number of compartments per drug");
    //
    const Real current_time = par.get<Real>("current time"),
               time_step = par.get<Real>("tumour angiogenesis solver: time step")
                         * par.get<int>("tumour angiogenesis solver: drug biochemics increment");
    const Real low_bound_transport_velocity    = mi__vnr->get_biochem_param(201),
               low_bound_interstitial_velocity = mi__vnr->get_biochem_param(202),
               max_bound_interstitial_velocity = mi__vnr->get_biochem_param(203);
    //
    const bool use_lumped_matrix = true;
    //
    // sanity checks
    ASSERT_(this->par_list.get<unsigned int>("dimension")==nvars_AUX, "fatal internal error occurred");
    ASSERT_(nvars==8, "fatal internal error occurred");
    ASSERT_(nvars==g_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(nvars==g_P_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(ninvars_NCS==3, "fatal internal error occurred");
    //
    // check if to update the vascular (blood & lymph) network parameters
    if ( update_vasc_params ) {
        // make sure memory is clean
        blv_sf_dp0.clear(); lyv_sf_dp0.clear();
        blv_sf_dp1.clear(); lyv_sf_dp1.clear();
        blv_Peff.clear();  lyv_Peff.clear();
        blv_Kp.clear();    lyv_Kp.clear();
        blv_S.clear();     lyv_S.clear();
        blv_P_dp0.clear();  lyv_P_dp0.clear();
        blv_P_dp1.clear();  lyv_P_dp1.clear();
        n_perfused_blood_vessels.clear();  n_perfused_lymph_vessels.clear();
        n_permeable_blood_vessels.clear(); n_permeable_lymph_vessels.clear();
        intravasc_drug.clear();
        // ...and now initialise them properly
        const unsigned int n_elem = es_3d.get_mesh().n_elem();
        std::vector<Real> Z2(2, 0.0);
        std::vector<int> I2(2, 0);
        // drug (phase-1) reflection coefficient wrt the blood|lymph vessel wall
        blv_sf_dp0.assign(n_elem, Z2); lyv_sf_dp0.assign(n_elem, Z2);
        // drug (phase-1) reflection coefficient wrt the blood|lymph vessel wall
        blv_sf_dp1.assign(n_elem, Z2); lyv_sf_dp1.assign(n_elem, Z2);
        // effective pressure = microvascular (blood|lymph) pressure - plasma osmotic pressure
        blv_Peff.assign(n_elem, Z2);   lyv_Peff.assign(n_elem, Z2);
        // blood|lymph vessel hydraulic permeability
        blv_Kp.assign(n_elem, Z2);     lyv_Kp.assign(n_elem, Z2);
        // blood|lymph vessel surface area
        blv_S.assign(n_elem, Z2);      lyv_S.assign(n_elem, Z2);
        // drug (phase-1) diffusive permeability through the blood|lymph vessel wall
        blv_P_dp0.assign(n_elem, Z2);  lyv_P_dp0.assign(n_elem, Z2);
        // drug (phase-2) diffusive permeability through the blood|lymph vessel wall
        blv_P_dp1.assign(n_elem, Z2);  lyv_P_dp1.assign(n_elem, Z2);
        //
        n_perfused_blood_vessels.assign(n_elem, 0);   n_perfused_lymph_vessels.assign(n_elem, 0);
        n_permeable_blood_vessels.assign(n_elem, I2); n_permeable_lymph_vessels.assign(n_elem, I2);
        // average intravascular drug (phase-1 only!!!) concentration
        intravasc_drug.assign(n_elem, Z2);
    }
    //
    // initialise the system matrix or/and rhs-vector, subsequently, add the
    // contribution from the previous (time) step
    if ( ! use_forward_Euler ) {
        // zero out the system matrix and rhs-vector first
        g_sys.matrix->zero();
        g_sys.rhs->zero();
    } else {
        // zero out the system rhs-vector first
        g_sys.rhs->zero();
        //
        if ( ! use_lumped_matrix )
            g_sys.rhs->add_vector(*g_P_sys.solution, g_sys.get_matrix("capacity matrix"));
    }
    //
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        const unsigned int elem_ID = elem->id();
        //
        QGauss quad_volu(SP3D, par.get<Order>("Gauss quadrature order"));
        UniquePtr<FEBase> fe_g_volu = FEBase::build(SP3D, fe_g_type);
        fe_g_volu->attach_quadrature_rule(&quad_volu);
        //
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type>[nvars];
        g_ctx.get_dof_indices(elem, dof_indices, dof_indices_var);
        //
        std::vector<unsigned int> var_dofs;
        calc_var_dofs(nvars, dof_indices_var, var_dofs);
        //
        DenseMatrix<Number> Me(dof_indices.size(), dof_indices.size());
        std::vector< DenseSubMatrix<Number> > Mesub(nvars*nvars, Me);
        DenseVector<Number> Ye(dof_indices.size());
        std::vector< DenseSubVector<Number> > Yesub(nvars, Ye);
        std::vector<unsigned int> offset;
        calc_offset(nvars, var_dofs, offset);
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            for (unsigned int jvar=0; jvar<nvars; jvar++) {
                Mesub[ivar*nvars+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
            }
            Yesub[ivar].reposition(offset[ivar], var_dofs[ivar]);
        }
        // get access to the material region
        libmesh_assert( this->_dmn.regn.end() != this->_dmn.regn.find(elem->subdomain_id()) );
        const RegionInfo* ri_ptr = &(this->_dmn.regn.find(elem->subdomain_id())->second);
        const MaterialInfo* mi__dmn = ri_ptr->mat;
        // get access to the element quadrature points data
        libmesh_assert( this->_dmn.elem_qi_local.end() != this->_dmn.elem_qi_local.find(elem->id()) );
        QuadratureInfo* qpi_ptr = &(this->_dmn.elem_qi_local.find(elem->id())->second);
        const unsigned int n_quad_pnt = qpi_ptr->XYZ.size();
        // get access to the vascular nodes contained in this element
        libmesh_assert( this->_dmn_vn.elem3d__node.end() != this->_dmn_vn.elem3d__node.find(elem->id()) );
        const std::set<dof_id_type>& vasc_nds = this->_dmn_vn.elem3d__node.find(elem->id())->second;
        //
        const std::vector< Real >& JxW_ = fe_g_volu->get_JxW();
        const std::vector< std::vector<Real> >&          Phi = fe_g_volu->get_phi();
        const std::vector< std::vector<RealGradient> >& dPhi_dX = fe_g_volu->get_dphi();
        fe_g_volu->reinit(elem);
        // sanity check
        ASSERT_(n_quad_pnt==quad_volu.n_points(), "fatal internal error occurred");
        //
        std::vector<int> drug_props[2];
        digitise_me(static_cast<int>(mi__dmn->get_biochem_param(120)), drug_props[0]);
        digitise_me(static_cast<int>(mi__dmn->get_biochem_param(130)), drug_props[1]);
        // sanity checks
        ASSERT_(5==drug_props[0].size(), "fatal external data error occurred");
        ASSERT_(5==drug_props[1].size(), "fatal external data error occurred");
        //
        const int props_dp0[] = { drug_props[0][3] ,
                                  drug_props[1][3] },
                  props_dp1[] = { drug_props[0][4] ,
                                  drug_props[1][4] };
        //
        const Real lymph_Kp_S = mi__dmn->get_biochem_param(200); // specific rate of lymphatic drainage of the free drug [1/(Pa*s)]
        const bool use_circulation_system_explicitly = true;
        const bool use_lymphatic_system_explicitly = lymph_Kp_S > 0.0 ? false : true;
        //
        //                             @@@@@@@@@@@@@@@@@@@@@
        if ( update_vasc_params ) { // @@@@@@@ START @@@@@@@
        //                             @@@@@@@@@@@@@@@@@@@@@
        //
        for (std::set<dof_id_type>::const_iterator sci=vasc_nds.begin(); sci!=vasc_nds.end(); sci++) {
            const Node* vasc_node = this->_dmn_vn.get_mesh().node_ptr(*sci);
            libmesh_assert( NULL != vasc_node );
            const Point vasc_node_U(0.0, 0.0, 0.0);
            // get access to the vascular info
            ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(vasc_node->id()), "fatal internal error occurred");
            const VascularInfo* vi = &(this->_dmn_vn.node_vi.find(vasc_node->id())->second);
            ASSERT_(NULL!=vi, "fatal internal error occurred");
            // sanity check
            ASSERT_(elem_ID==vi->elem3d_id, "fatal internal error occurred");
            //
            // ignore the following computations if vessel has collapsed
            if ( VascularInfo::COLLAPSED == vi->status ) continue;
            //
            FieldInfo fld_vasc(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
            fld_vasc.current_time = current_time;
            //
            const Node* conn_vasc_node[] = { NULL, NULL, NULL };
            Point conn_vasc_node_U[] = { Point(0.0,0.0,0.0), Point(0.0,0.0,0.0), Point(0.0,0.0,0.0) };
            //
            unsigned int I = 0;
            for ( std::set<dof_id_type>::const_iterator
                  citer=vi->vsc_elem.begin(); citer!=vi->vsc_elem.end(); citer++ )
            {
                const Elem& vasc_elem = this->_dmn_vn.get_mesh().elem_ref(*citer);
                const dof_id_type conn_node_id = edge2_other_node(vasc_elem, vasc_node->id());
                //
                ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(conn_node_id), "fatal internal error occurred");
                const VascularInfo& conn_vi = this->_dmn_vn.node_vi.find(conn_node_id)->second;
                if ( VascularInfo::COLLAPSED == conn_vi.status ) continue;
                //
                conn_vasc_node[I] = this->_dmn_vn.get_mesh().node_ptr(conn_node_id);
                ASSERT_(NULL!=conn_vasc_node[I], "fatal internal error occurred");
                conn_vasc_node_U[I] = Point(0.0, 0.0, 0.0);
                ASSERT_(3>I, "fatal internal error occurred");
                ++I;
            }
            //
            const Real Velc = get_system_solution_nodal_value<6>(vn_flow_sys, vasc_node, vn_flow);
            //
            // ignore the following computations if biofluid velocity is below a certain threshold
            if ( low_bound_transport_velocity > Velc ) continue;
            //
            const Real Peff = get_system_solution_nodal_value<7>(vn_flow_sys, vasc_node, vn_flow);
            const Real Kp = ta_constitutive.blood_vessel_hydraulic_permeability(vi, mi__vnr, mi__dmn, &fld_vasc);
            const Real S = vascular_circumference(vi)
                         * vascular_length(vasc_node, vasc_node_U, conn_vasc_node, conn_vasc_node_U);
            const Real sigma_dp0[] = { ta_constitutive.vascular_reflection_coefficient_2_drug(vi, mi__vnr, mi__dmn, &fld_vasc, 10) ,
                                       ta_constitutive.vascular_reflection_coefficient_2_drug(vi, mi__vnr, mi__dmn, &fld_vasc, 12) };
            const Real sigma_dp1[] = { ta_constitutive.vascular_reflection_coefficient_2_drug(vi, mi__vnr, mi__dmn, &fld_vasc, 11) ,
                                       ta_constitutive.vascular_reflection_coefficient_2_drug(vi, mi__vnr, mi__dmn, &fld_vasc, 13) };
            const Real P_dp0[] = { ta_constitutive.vascular_hydraulic_permeability_2_drug(vi, mi__vnr, mi__dmn, &fld_vasc, 10) ,
                                   ta_constitutive.vascular_hydraulic_permeability_2_drug(vi, mi__vnr, mi__dmn, &fld_vasc, 12) };
            const Real P_dp1[] = { ta_constitutive.vascular_hydraulic_permeability_2_drug(vi, mi__vnr, mi__dmn, &fld_vasc, 11) ,
                                   ta_constitutive.vascular_hydraulic_permeability_2_drug(vi, mi__vnr, mi__dmn, &fld_vasc, 13) };
            const Real vasc_drug[] = { get_system_solution_nodal_value<0>(vn_drug_sys, vasc_node, vn_drug) ,
                                       get_system_solution_nodal_value<1>(vn_drug_sys, vasc_node, vn_drug) };
            // some sanity checks
            for (int D=0; D<2; D++) {
                if (sigma_dp0[D]>1.0) ERROR_("fatal internal error occurred");
                if (sigma_dp1[D]>1.0) ERROR_("fatal internal error occurred");
            }
            //
            if ( VascularInfo::BLOOD == vi->flow ) {
                //
                for (int D=0; D<2; D++) {
                    if ( vasc_drug[D]<=0.0 ) continue;
                    if ( sigma_dp0[D]>0.99 || sigma_dp1[D]>0.99 ) continue;
                    //
                    blv_sf_dp0[elem_ID][D] += sigma_dp0[D];
                    blv_sf_dp1[elem_ID][D] += sigma_dp1[D];
                    blv_P_dp0[elem_ID][D]  += P_dp0[D];
                    blv_P_dp1[elem_ID][D]  += P_dp1[D];
                    blv_Peff[elem_ID][D]   += Peff;
                    blv_Kp[elem_ID][D]     += Kp;
                    blv_S[elem_ID][D]      += S;
                    //
                    intravasc_drug[elem_ID][D] += vasc_drug[D];
                    //
                    // increment this index...
                    ++n_permeable_blood_vessels[elem_ID][D];
                }
                // increment this index...
                ++n_perfused_blood_vessels[elem_ID];
                //
            } else if ( VascularInfo::LYMPH == vi->flow ) {
                //
                ERROR_("this part of code is still under development");
                //
                // increment this index...
                ++n_perfused_lymph_vessels[elem_ID];
                //
            } else libmesh_error();
            // end of vascular nodes to this 3D element loop...
        }
        // get the average value of the following quantities
        if ( n_perfused_blood_vessels[elem_ID] ) {
            for (int D=0; D<2; D++) {
                if ( ! n_permeable_blood_vessels[elem_ID][D] ) continue;
                //
                blv_sf_dp0[elem_ID][D] /= n_permeable_blood_vessels[elem_ID][D];
                blv_sf_dp1[elem_ID][D] /= n_permeable_blood_vessels[elem_ID][D];
                blv_P_dp0[elem_ID][D]  /= n_permeable_blood_vessels[elem_ID][D];
                blv_P_dp1[elem_ID][D]  /= n_permeable_blood_vessels[elem_ID][D];
                blv_Peff[elem_ID][D]   /= n_permeable_blood_vessels[elem_ID][D];
                blv_Kp[elem_ID][D]     /= n_permeable_blood_vessels[elem_ID][D];
                //
                intravasc_drug[elem_ID][D] /= n_permeable_blood_vessels[elem_ID][D];
            }
        }
        if ( n_perfused_lymph_vessels[elem_ID] ) {
            ERROR_("this part of code is still under development");
        }
        // sanity checks...
        ASSERT_(n_perfused_blood_vessels[elem_ID]>=0, "fatal internal error occurred");
        ASSERT_(n_perfused_lymph_vessels[elem_ID]>=0, "fatal internal error occurred");
        for (int D=0; D<2; D++) {
            if ( 0.0 > intravasc_drug[elem_ID][D] )
                ERROR_("fatal internal error occurred");
                // intravasc_drug[elem_ID][D] = 0.0;
        }
        //
        //   @@@@@@@@@@@@@@@@@@@
        } // @@@@@@@ END @@@@@@@
        //   @@@@@@@@@@@@@@@@@@@
        //
        std::vector<Real> nodal_drug[nvars]; // drug concentration (previous solution)
        std::vector<Real> nodal_IFP; // interstitial fluid pressure
        std::vector<Real> nodal_drug_rate[nvars]; // drug concentration rate (previous solution)
        // initialise the containers
        for (unsigned int I=0; I<nvars; I++) nodal_drug[I].assign(var_dofs[0], 0.0);
        nodal_IFP.assign(var_dofs[0], 0.0);
        for (unsigned int I=0; I<nvars; I++) nodal_drug_rate[I].assign(var_dofs[0], 0.0);
        // now set the corresponding nodal values...
        for (unsigned int l=0; l<var_dofs[0]; l++) {
            nodal_drug[0][l] = get_system_solution_nodal_value<0>(g_P_sys, elem->node_ptr(l), s3d_drug);
            nodal_drug[1][l] = get_system_solution_nodal_value<1>(g_P_sys, elem->node_ptr(l), s3d_drug);
            nodal_drug[2][l] = get_system_solution_nodal_value<2>(g_P_sys, elem->node_ptr(l), s3d_drug);
            nodal_drug[3][l] = get_system_solution_nodal_value<3>(g_P_sys, elem->node_ptr(l), s3d_drug);
            nodal_drug[4][l] = get_system_solution_nodal_value<4>(g_P_sys, elem->node_ptr(l), s3d_drug);
            nodal_drug[5][l] = get_system_solution_nodal_value<5>(g_P_sys, elem->node_ptr(l), s3d_drug);
            nodal_drug[6][l] = get_system_solution_nodal_value<6>(g_P_sys, elem->node_ptr(l), s3d_drug);
            nodal_drug[7][l] = get_system_solution_nodal_value<7>(g_P_sys, elem->node_ptr(l), s3d_drug);
            nodal_IFP[l] = get_system_solution_nodal_value<0>(s3d_flow_sys, elem->node_ptr(l), s3d_flow);
        }
        if ( ! use_forward_Euler )
        for (unsigned int l=0; l<var_dofs[0]; l++) {
            nodal_drug_rate[0][l] = get_system_solution_nodal_value<0>(DgDt_sys, elem->node_ptr(l), s3d_drug_rate);
            nodal_drug_rate[1][l] = get_system_solution_nodal_value<1>(DgDt_sys, elem->node_ptr(l), s3d_drug_rate);
            nodal_drug_rate[2][l] = get_system_solution_nodal_value<2>(DgDt_sys, elem->node_ptr(l), s3d_drug_rate);
            nodal_drug_rate[3][l] = get_system_solution_nodal_value<3>(DgDt_sys, elem->node_ptr(l), s3d_drug_rate);
            nodal_drug_rate[4][l] = get_system_solution_nodal_value<4>(DgDt_sys, elem->node_ptr(l), s3d_drug_rate);
            nodal_drug_rate[5][l] = get_system_solution_nodal_value<5>(DgDt_sys, elem->node_ptr(l), s3d_drug_rate);
            nodal_drug_rate[6][l] = get_system_solution_nodal_value<6>(DgDt_sys, elem->node_ptr(l), s3d_drug_rate);
            nodal_drug_rate[7][l] = get_system_solution_nodal_value<7>(DgDt_sys, elem->node_ptr(l), s3d_drug_rate);
        }
        //
        //
        Real P_int = 0.0,
             strm_V = 0.0;
        for (unsigned int qp=0; qp<quad_volu.n_points(); qp++) {
            // interstitial fluid pressure at the parenchymal tissue
            for (unsigned int l=0; l<var_dofs[0]; l++)
                P_int += Phi[l][qp] * nodal_IFP[l];
            // compute the (current configuration) volume of the element
            strm_V += elem->volume()
                    * determinant(qpi_ptr->deform_grad[qp]);
        }
        P_int /= quad_volu.n_points(); // take the average IFP...
        strm_V /= quad_volu.n_points(); // take the average volume of the tissue (domain)...
        //
        //
        Real phi_dp0[] = { 0.0, 0.0 };
        Real theta_dp0[] = { 0.0, 0.0 };
        Real theta_dp1[] = { 0.0, 0.0 };
        if ( ! use_circulation_system_explicitly ) {
            // there is seriously something wrong in here...
            ERROR_("check the code above, there is an internal error");
        } else if ( n_perfused_blood_vessels[elem_ID] ) {
            for (int D=0; D<2; D++) {
                const Real P_eff = blv_Peff[elem_ID][D],
                           Kp    = blv_Kp[elem_ID][D],
                           S__V  = blv_S[elem_ID][D] / strm_V,
                           Pd0   = blv_P_dp0[elem_ID][D],
                           Pd1   = blv_P_dp1[elem_ID][D],
                           sigma0 = blv_sf_dp0[elem_ID][D],
                           sigma1 = blv_sf_dp1[elem_ID][D];
                ASSERT_(Kp>=0.0, "fatal internal error occurred");
                ASSERT_(Pd0>=0.0, "fatal internal error occurred");
                ASSERT_(Pd1>=0.0, "fatal internal error occurred");
                ASSERT_(S__V>=0.0, "fatal internal error occurred");
                ASSERT_(sigma0>=0.0&&sigma0<=1.0, "fatal internal error occurred");
                ASSERT_(sigma1>=0.0&&sigma1<=1.0, "fatal internal error occurred");
                //
                phi_dp0[D] += Pd0 * S__V; // diffusion rate from blood vessel
                if ( P_eff < P_int ) {
                    theta_dp0[D] += Kp  * S__V * (P_int-P_eff) * (1.0-sigma0); // convection rate to blood vessel
                    theta_dp1[D] += Kp  * S__V * (P_int-P_eff) * (1.0-sigma1); // convection rate to blood vessel
                } else {
                    phi_dp0[D]   += Kp  * S__V * (P_eff-P_int) * (1.0-sigma0); // convection rate from blood vessel
                }
                theta_dp0[D] += Pd0 * S__V; // diffusion rate to blood vessel
                theta_dp1[D] += Pd1 * S__V; // diffusion rate to blood vessel
            }
        }
        if ( ! use_lymphatic_system_explicitly ) {
            if (lymph_Kp_S<0.0) ERROR_("fatal internal error occurred");
            for (int D=0; D<2; D++) {
                theta_dp0[D] += lymph_Kp_S * P_int; // convection rate to lymph vessel (accounted for no reflection!)
                theta_dp1[D] += lymph_Kp_S * P_int; // convection rate to lymph vessel (accounted for no reflection!)
            }
        } else if ( n_perfused_lymph_vessels[elem_ID] ) {
            for (int D=0; D<2; D++) {
                const Real P_eff = lyv_Peff[elem_ID][D];
                if ( P_eff >= P_int ) {
                    ERROR_("lymphatics cannot drain properly; check model input parameters");
                } else {
                    ERROR_("this part of code is still under development");
                }
            }
        }
        // sanity checks...
        for (int D=0; D<2; D++) ASSERT_(phi_dp0[D]>=0.0, "fatal internal error occurred");
        for (int D=0; D<2; D++) ASSERT_(theta_dp0[D]>=0.0, "fatal internal error occurred");
        for (int D=0; D<2; D++) ASSERT_(theta_dp1[D]>=0.0, "fatal internal error occurred");
        //
        //         rate of drug supply from vascular network
        const Real lambda[] = { phi_dp0[0]*intravasc_drug[elem_ID][0] ,
                                phi_dp0[1]*intravasc_drug[elem_ID][1] },
        //         dissassociation rate coefficient: phase-1
                   lamda_dp0[] = { mi__dmn->get_biochem_param(121)*normal_distro(1., mi__dmn->get_biochem_param(145)) ,
                                   mi__dmn->get_biochem_param(131)*normal_distro(1., mi__dmn->get_biochem_param(155)) },
        //         dissassociation rate coefficient: phase-2
                   lamda_dp1[] = { mi__dmn->get_biochem_param(122)*normal_distro(1., mi__dmn->get_biochem_param(145)) ,
                                   mi__dmn->get_biochem_param(132)*normal_distro(1., mi__dmn->get_biochem_param(155)) },
        //         internalisation rate coefficient
                   kappa_int[] = { mi__dmn->get_biochem_param(123)*normal_distro(1., mi__dmn->get_biochem_param(145)) ,
                                   mi__dmn->get_biochem_param(133)*normal_distro(1., mi__dmn->get_biochem_param(155)) },
        //         transformation rate coefficient
                   kappa_trn[] = { mi__dmn->get_biochem_param(124)*normal_distro(1., mi__dmn->get_biochem_param(146)) ,
                                   mi__dmn->get_biochem_param(134)*normal_distro(1., mi__dmn->get_biochem_param(156)) },
        //         drug decay rate coefficient: phase-1
                   delta_dp0[] = { theta_dp0[0]+mi__dmn->get_biochem_param(125)*normal_distro(1., mi__dmn->get_biochem_param(144)) ,
                                   theta_dp0[1]+mi__dmn->get_biochem_param(135)*normal_distro(1., mi__dmn->get_biochem_param(154)) },
        //         drug decay rate coefficient: phase-2
                   delta_dp1[] = { theta_dp1[0]+mi__dmn->get_biochem_param(126)*normal_distro(1., mi__dmn->get_biochem_param(144)) ,
                                   theta_dp1[1]+mi__dmn->get_biochem_param(136)*normal_distro(1., mi__dmn->get_biochem_param(154)) },
        //         association rate coefficient: phase-1
                   kappa_dp0[] = { mi__dmn->get_biochem_param(127)*normal_distro(1., mi__dmn->get_biochem_param(145)) ,
                                   mi__dmn->get_biochem_param(137)*normal_distro(1., mi__dmn->get_biochem_param(155)) },
        //         association rate coefficient: phase-2
                   kappa_dp1[] = { mi__dmn->get_biochem_param(128)*normal_distro(1., mi__dmn->get_biochem_param(145)) ,
                                   mi__dmn->get_biochem_param(138)*normal_distro(1., mi__dmn->get_biochem_param(155)) },
        //         drug decay rate coefficient: internalised
                   delta_int[] = { mi__dmn->get_biochem_param(129)*normal_distro(1., mi__dmn->get_biochem_param(144)) ,
                                   mi__dmn->get_biochem_param(139)*normal_distro(1., mi__dmn->get_biochem_param(154)) };
        //
        // loop over volume integration quadrature points
        for (unsigned int qp=0; qp<quad_volu.n_points(); qp++) {
            const Real JxW   = JxW_[qp];
            const Real JxWoDT = JxW_[qp] / time_step;
            const Real JxWxA = JxW_[qp] * time_step;
            //
            Real drug[] = { 0.0, 0.0, 0.0, 0.0 , // drug concentration, its gradient, and the rate of the drug concentration
                            0.0, 0.0, 0.0, 0.0 };
            RealGradient drug_grad[] = { RealGradient(), RealGradient(), RealGradient(), RealGradient() ,
                                         RealGradient(), RealGradient(), RealGradient(), RealGradient() };
            Real drug_rate[] = { 0.0, 0.0, 0.0, 0.0 ,
                                 0.0, 0.0, 0.0, 0.0 };
            for (unsigned int ivar=0; ivar<nvars; ivar++) {
                for (unsigned int l=0; l<var_dofs[0]; l++) {
                    drug[ivar]      += Phi[l][qp]     * nodal_drug[ivar][l];
                    drug_grad[ivar] += dPhi_dX[l][qp] * nodal_drug[ivar][l];
                    drug_rate[ivar] += Phi[l][qp]     * nodal_drug_rate[ivar][l];
                }
                if ( drug[ivar] < 0.0 ) drug[ivar] = 0.0;
                //
                // qpi_ptr->auxiliary[qp][ivar]      = drug[ivar];
                // qpi_ptr->grad_auxiliary[qp][ivar] = drug_grad[ivar];
                // qpi_ptr->auxiliary_rate[qp][ivar] = drug_rate[ivar];
            }
            //
            libmesh_assert(4==nvars_RDS);
            const Real ECM = apply_bounds(0.01, qpi_ptr->biochem[qp][3], 1.0),
                       F_ECM[] = { pow(ECM, mi__dmn->get_biochem_param(147)) ,
                                   pow(ECM, mi__dmn->get_biochem_param(157)) };
            //
            Point IFV;
            {
                const Point ifvel( qpi_ptr->fluid_velocity[qp](0) ,
                                   qpi_ptr->fluid_velocity[qp](1) ,
                                   qpi_ptr->fluid_velocity[qp](2) );
                const Real ifvel_mag( max_bound_interstitial_velocity>0.0
                                    ? apply_ubound(ifvel.norm(), max_bound_interstitial_velocity)
                                    : ifvel.norm() );
                // interstitial fluid velocity
                IFV = ifvel_mag > low_bound_interstitial_velocity
                    ? ifvel_mag * ifvel.unit()
                    : Point(0.0, 0.0, 0.0);
            }
            //
            FieldInfo fld_strm(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
            for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                fld_strm.biochem[ivar]      = qpi_ptr->biochem[qp][ivar];
                fld_strm.grad_biochem[ivar] = qpi_ptr->grad_biochem[qp][ivar];
                fld_strm.biochem_rate[ivar] = qpi_ptr->biochem_rate[qp][ivar];
            }
            for (unsigned int ivar=nvars_RDS; ivar<nvars_RDS+ndcvars_RDS; ivar++) {
                fld_strm.biochem[ivar] = qpi_ptr->biochem[qp][ivar];
            }
            // for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
            //     fld_strm.auxiliary[ivar]      = qpi_ptr->auxiliary[qp][ivar];
            //     fld_strm.grad_auxiliary[ivar] = qpi_ptr->grad_auxiliary[qp][ivar];
            //     fld_strm.auxiliary_rate[ivar] = qpi_ptr->auxiliary_rate[qp][ivar];
            // }
            fld_strm.strain = qpi_ptr->strain[qp];
            for (unsigned int ivar=0; ivar<ninvars_NCS; ivar++) {
                fld_strm.biomech[ivar] = qpi_ptr->biomech[qp][ivar];
            }
            //
            // convection velocity of a drug molecule or a nanoparticle due to interstitial fluid flow
            const Point Veloc_dp0[] = { ( 1==props_dp0[0] || 3==props_dp0[0] ? IFV : Point(0.,0.,0.) ) ,
                                        ( 1==props_dp0[1] || 3==props_dp0[1] ? IFV : Point(0.,0.,0.) ) };
            const Point Veloc_dp1[] = { ( 1==props_dp1[0] || 3==props_dp1[0] ? IFV : Point(0.,0.,0.) ) ,
                                        ( 1==props_dp1[1] || 3==props_dp1[1] ? IFV : Point(0.,0.,0.) ) };
            //
            // diffusion coefficient of a drug molecule or a nanoparticle with respect to the stroma
            const Dyadic3D Diff_dp0[] = { ( 2==props_dp0[0] || 3==props_dp0[0] ? ta_constitutive.stroma_diffusion_coefficient_2_drug(mi__vnr, mi__dmn, &fld_strm, 10) : zero_matrix ) ,
                                          ( 2==props_dp0[1] || 3==props_dp0[1] ? ta_constitutive.stroma_diffusion_coefficient_2_drug(mi__vnr, mi__dmn, &fld_strm, 12) : zero_matrix ) };
            const Dyadic3D Diff_dp1[] = { ( 2==props_dp1[0] || 3==props_dp1[0] ? ta_constitutive.stroma_diffusion_coefficient_2_drug(mi__vnr, mi__dmn, &fld_strm, 11) : zero_matrix ) ,
                                          ( 2==props_dp1[1] || 3==props_dp1[1] ? ta_constitutive.stroma_diffusion_coefficient_2_drug(mi__vnr, mi__dmn, &fld_strm, 13) : zero_matrix ) };
            //
            //
            //
            // /*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\
            // \*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/
            // /*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\
            // \*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/
            if ( use_forward_Euler )
            // ...execute the following in case we have the
            // forward Euler (explicit solver) in action!!!
            for (unsigned int i=0; i<var_dofs[0]; i++) {
                //
                for (int D=0; D<2; D++) {
                    const unsigned int N_D = ndofs_DRUG * D;
                    unsigned int a = 0;
                    //
                    // vascular supply contribution
                    a = N_D+0; Yesub[a](i) += JxW * (Phi[i][qp]*lambda[D]);
                    // convection contribution
                    a = N_D+0; Yesub[a](i) -= JxW * (Phi[i][qp]*Veloc_dp0[D]*drug_grad[N_D+0]);
                    // diffusion contribution
                    a = N_D+0; Yesub[a](i) -= JxW * (dPhi_dX[i][qp]*Diff_dp0[D]*drug_grad[N_D+0]);
                    // transformation contribution
                    a = N_D+0; Yesub[a](i) += JxW * (-Phi[i][qp]*kappa_trn[D]*drug[N_D+0]);
                    // association contribution
                    a = N_D+0; Yesub[a](i) += JxW * (-Phi[i][qp]*kappa_dp0[D]*drug[N_D+0]*F_ECM[D]);
                    // dissassociation contribution
                    a = N_D+0; Yesub[a](i) += JxW * ( Phi[i][qp]*lamda_dp0[D]*drug[N_D+2]);
                    // lymphatic drainage & decay contribution
                    a = N_D+0; Yesub[a](i) += JxW * (-Phi[i][qp]*delta_dp0[D]*drug[N_D+0]);
                    //
                    // convection contribution
                    a = N_D+1; Yesub[a](i) -= JxW * (Phi[i][qp]*Veloc_dp1[D]*drug_grad[N_D+1]);
                    // diffusion contribution
                    a = N_D+1; Yesub[a](i) -= JxW * (dPhi_dX[i][qp]*Diff_dp1[D]*drug_grad[N_D+1]);
                    // transformation contribution
                    a = N_D+1; Yesub[a](i) += JxW * ( Phi[i][qp]*kappa_trn[D]*drug[N_D+0]);
                    // association contribution
                    a = N_D+1; Yesub[a](i) += JxW * (-Phi[i][qp]*kappa_dp1[D]*drug[N_D+1]*F_ECM[D]);
                    // dissassociation contribution
                    a = N_D+1; Yesub[a](i) += JxW * ( Phi[i][qp]*lamda_dp1[D]*drug[N_D+2]);
                    // lymphatic drainage & decay contribution
                    a = N_D+1; Yesub[a](i) += JxW * (-Phi[i][qp]*delta_dp1[D]*drug[N_D+1]);
                    //
                    // association contribution
                    a = N_D+2; Yesub[a](i) += JxW * ( Phi[i][qp]*kappa_dp0[D]*drug[N_D+0]*F_ECM[D]);
                    // association contribution
                    a = N_D+2; Yesub[a](i) += JxW * ( Phi[i][qp]*kappa_dp1[D]*drug[N_D+1]*F_ECM[D]);
                    // dissassociation contribution
                    a = N_D+2; Yesub[a](i) += JxW * (-Phi[i][qp]*lamda_dp0[D]*drug[N_D+2]);
                    // dissassociation contribution
                    a = N_D+2; Yesub[a](i) += JxW * (-Phi[i][qp]*lamda_dp1[D]*drug[N_D+2]);
                    // internalisation contribution
                    a = N_D+2; Yesub[a](i) += JxW * (-Phi[i][qp]*kappa_int[D]*drug[N_D+2]);
                    //
                    // internalisation contribution
                    a = N_D+3; Yesub[a](i) += JxW * ( Phi[i][qp]*kappa_int[D]*drug[N_D+2]);
                    // decay contribution
                    a = N_D+3; Yesub[a](i) += JxW * (-Phi[i][qp]*delta_int[D]*drug[N_D+3]);
                    //
                }
                // end of DoFs loop...
            }
            // /*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\
            // \*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/
            // /*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\
            // \*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/
            //
            //
            //
            // /*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\
            // \*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/
            // /*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\
            // \*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/
            if ( ! use_forward_Euler )
            // ...execute the following in case we have the
            // backward Euler (implicit solver) in action!!!

            for (unsigned int i=0; i<var_dofs[0]; i++) {
                //
                for (int D=0; D<2; D++) {
                    const unsigned int N_D = ndofs_DRUG * D;
                    unsigned int a = 0;
                    //
                    // capacity contribution
                    a = N_D+0; Yesub[a](i) += JxW * (Phi[i][qp]*drug[a]);
                    //
                    // capacity contribution
                    a = N_D+1; Yesub[a](i) += JxW * (Phi[i][qp]*drug[a]);
                    //
                    // capacity contribution
                    a = N_D+2; Yesub[a](i) += JxW * (Phi[i][qp]*drug[a]);
                    //
                    // capacity contribution
                    a = N_D+3; Yesub[a](i) += JxW * (Phi[i][qp]*drug[a]);
                    //
                    // vascular supply contribution
                    a = N_D+0; Yesub[a](i) += JxWxA * (Phi[i][qp]*lambda[D]);
                    //
                }
                //
                for (unsigned int j=0; j<var_dofs[0]; j++) {
                    //
                    for (int D=0; D<2; D++) {
                        const unsigned int N_D = ndofs_DRUG * D;
                        unsigned int a = 0, b = 0;
                        //
                        // capacity contribution
                        a = N_D+0; b = N_D+0; Mesub[a*nvars+b](i,j) += JxW * (Phi[i][qp]*Phi[j][qp]);
                        //
                        // capacity contribution
                        a = N_D+1; b = N_D+1; Mesub[a*nvars+b](i,j) += JxW * (Phi[i][qp]*Phi[j][qp]);
                        //
                        // capacity contribution
                        a = N_D+2; b = N_D+2; Mesub[a*nvars+b](i,j) += JxW * (Phi[i][qp]*Phi[j][qp]);
                        //
                        // capacity contribution
                        a = N_D+3; b = N_D+3; Mesub[a*nvars+b](i,j) += JxW * (Phi[i][qp]*Phi[j][qp]);
                        //
                        // convection contribution
                        a = N_D+0; b = N_D+0; Mesub[a*nvars+b](i,j) += JxWxA * (Phi[i][qp]*Veloc_dp0[D]*dPhi_dX[j][qp]);
                        // diffusion contribution
                        a = N_D+0; b = N_D+0; Mesub[a*nvars+b](i,j) += JxWxA * (dPhi_dX[i][qp]*Diff_dp0[D]*dPhi_dX[j][qp]);
                        // transformation contribution
                        a = N_D+0; b = N_D+0; Mesub[a*nvars+b](i,j) -= JxWxA * (-Phi[i][qp]*kappa_trn[D]*Phi[j][qp]);
                        // association contribution
                        a = N_D+0; b = N_D+0; Mesub[a*nvars+b](i,j) -= JxWxA * (-Phi[i][qp]*kappa_dp0[D]*Phi[j][qp]*F_ECM[D]);
                        // dissassociation contribution
                        a = N_D+0; b = N_D+2; Mesub[a*nvars+b](i,j) -= JxWxA * ( Phi[i][qp]*lamda_dp0[D]*Phi[j][qp]);
                        // lymphatic drainage & decay contribution
                        a = N_D+0; b = N_D+0; Mesub[a*nvars+b](i,j) -= JxWxA * (-Phi[i][qp]*delta_dp0[D]*Phi[j][qp]);
                        //
                        // convection contribution
                        a = N_D+1; b = N_D+1; Mesub[a*nvars+b](i,j) += JxWxA * (Phi[i][qp]*Veloc_dp1[D]*dPhi_dX[j][qp]);
                        // diffusion contribution
                        a = N_D+1; b = N_D+1; Mesub[a*nvars+b](i,j) += JxWxA * (dPhi_dX[i][qp]*Diff_dp1[D]*dPhi_dX[j][qp]);
                        // transformation contribution
                        a = N_D+1; b = N_D+0; Mesub[a*nvars+b](i,j) -= JxWxA * ( Phi[i][qp]*kappa_trn[D]*Phi[j][qp]);
                        // association contribution
                        a = N_D+1; b = N_D+1; Mesub[a*nvars+b](i,j) -= JxWxA * (-Phi[i][qp]*kappa_dp1[D]*Phi[j][qp]*F_ECM[D]);
                        // dissassociation contribution
                        a = N_D+1; b = N_D+2; Mesub[a*nvars+b](i,j) -= JxWxA * ( Phi[i][qp]*lamda_dp1[D]*Phi[j][qp]);
                        // lymphatic drainage & decay contribution
                        a = N_D+1; b = N_D+1; Mesub[a*nvars+b](i,j) -= JxWxA * (-Phi[i][qp]*delta_dp1[D]*Phi[j][qp]);
                        //
                        // association contribution
                        a = N_D+2; b = N_D+0; Mesub[a*nvars+b](i,j) -= JxWxA * ( Phi[i][qp]*kappa_dp0[D]*Phi[j][qp]*F_ECM[D]);
                        // association contribution
                        a = N_D+2; b = N_D+1; Mesub[a*nvars+b](i,j) -= JxWxA * ( Phi[i][qp]*kappa_dp1[D]*Phi[j][qp]*F_ECM[D]);
                        // dissassociation contribution
                        a = N_D+2; b = N_D+2; Mesub[a*nvars+b](i,j) -= JxWxA * (-Phi[i][qp]*lamda_dp0[D]*Phi[j][qp]);
                        // dissassociation contribution
                        a = N_D+2; b = N_D+2; Mesub[a*nvars+b](i,j) -= JxWxA * (-Phi[i][qp]*lamda_dp1[D]*Phi[j][qp]);
                        // internalisation contribution
                        a = N_D+2; b = N_D+2; Mesub[a*nvars+b](i,j) -= JxWxA * (-Phi[i][qp]*kappa_int[D]*Phi[j][qp]);
                        //
                        // internalisation contribution
                        a = N_D+3; b = N_D+2; Mesub[a*nvars+b](i,j) -= JxWxA * ( Phi[i][qp]*kappa_int[D]*Phi[j][qp]);
                        // decay contribution
                        a = N_D+3; b = N_D+3; Mesub[a*nvars+b](i,j) -= JxWxA * (-Phi[i][qp]*delta_int[D]*Phi[j][qp]);
                        //
                    }
                    // end of DoFs loop...
                }
                // end of DoFs loop...
            }

            /*

            for (unsigned int i=0; i<var_dofs[0]; i++) {
                //
                for (int D=0; D<2; D++) {
                    const unsigned int N_D = ndofs_DRUG * D;
                    unsigned int a = 0;
                    //
                    // capacity contribution
                    a = N_D+0; Yesub[a](i) -= JxW * (Phi[i][qp]*drug_rate[a]);
                    //
                    // capacity contribution
                    a = N_D+1; Yesub[a](i) -= JxW * (Phi[i][qp]*drug_rate[a]);
                    //
                    // capacity contribution
                    a = N_D+2; Yesub[a](i) -= JxW * (Phi[i][qp]*drug_rate[a]);
                    //
                    // capacity contribution
                    a = N_D+3; Yesub[a](i) -= JxW * (Phi[i][qp]*drug_rate[a]);
                    //
                    // vascular supply contribution
                    a = N_D+0; Yesub[a](i) += JxW * (Phi[i][qp]*lambda[D]);
                    // convection contribution
                    a = N_D+0; Yesub[a](i) -= JxW * (Phi[i][qp]*Veloc_dp0[D]*drug_grad[N_D+0]);
                    // diffusion contribution
                    a = N_D+0; Yesub[a](i) -= JxW * (dPhi_dX[i][qp]*Diff_dp0[D]*drug_grad[N_D+0]);
                    // transformation contribution
                    a = N_D+0; Yesub[a](i) += JxW * (-Phi[i][qp]*kappa_trn[D]*drug[N_D+0]);
                    // association contribution
                    a = N_D+0; Yesub[a](i) += JxW * (-Phi[i][qp]*kappa_dp0[D]*drug[N_D+0]*F_ECM[D]);
                    // dissassociation contribution
                    a = N_D+0; Yesub[a](i) += JxW * ( Phi[i][qp]*lamda_dp0[D]*drug[N_D+2]);
                    // lymphatic drainage & decay contribution
                    a = N_D+0; Yesub[a](i) += JxW * (-Phi[i][qp]*delta_dp0[D]*drug[N_D+0]);
                    //
                    // convection contribution
                    a = N_D+1; Yesub[a](i) -= JxW * (Phi[i][qp]*Veloc_dp1[D]*drug_grad[N_D+1]);
                    // diffusion contribution
                    a = N_D+1; Yesub[a](i) -= JxW * (dPhi_dX[i][qp]*Diff_dp1[D]*drug_grad[N_D+1]);
                    // transformation contribution
                    a = N_D+1; Yesub[a](i) += JxW * ( Phi[i][qp]*kappa_trn[D]*drug[N_D+0]);
                    // association contribution
                    a = N_D+1; Yesub[a](i) += JxW * (-Phi[i][qp]*kappa_dp1[D]*drug[N_D+1]*F_ECM[D]);
                    // dissassociation contribution
                    a = N_D+1; Yesub[a](i) += JxW * ( Phi[i][qp]*lamda_dp1[D]*drug[N_D+2]);
                    // lymphatic drainage & decay contribution
                    a = N_D+1; Yesub[a](i) += JxW * (-Phi[i][qp]*delta_dp1[D]*drug[N_D+1]);
                    //
                    // association contribution
                    a = N_D+2; Yesub[a](i) += JxW * ( Phi[i][qp]*kappa_dp0[D]*drug[N_D+0]*F_ECM[D]);
                    // association contribution
                    a = N_D+2; Yesub[a](i) += JxW * ( Phi[i][qp]*kappa_dp1[D]*drug[N_D+1]*F_ECM[D]);
                    // dissassociation contribution
                    a = N_D+2; Yesub[a](i) += JxW * (-Phi[i][qp]*lamda_dp0[D]*drug[N_D+2]);
                    // dissassociation contribution
                    a = N_D+2; Yesub[a](i) += JxW * (-Phi[i][qp]*lamda_dp1[D]*drug[N_D+2]);
                    // internalisation contribution
                    a = N_D+2; Yesub[a](i) += JxW * (-Phi[i][qp]*kappa_int[D]*drug[N_D+2]);
                    //
                    // internalisation contribution
                    a = N_D+3; Yesub[a](i) += JxW * ( Phi[i][qp]*kappa_int[D]*drug[N_D+2]);
                    // decay contribution
                    a = N_D+3; Yesub[a](i) += JxW * (-Phi[i][qp]*delta_int[D]*drug[N_D+3]);
                    //
                }
                //
                for (unsigned int j=0; j<var_dofs[0]; j++) {
                    //
                    for (int D=0; D<2; D++) {
                        const unsigned int N_D = ndofs_DRUG * D;
                        unsigned int a = 0, b = 0;
                        //
                        // capacity contribution
                        a = N_D+0; b = N_D+0; Mesub[a*nvars+b](i,j) += JxWoDT * (Phi[i][qp]*Phi[j][qp]);
                        //
                        // capacity contribution
                        a = N_D+1; b = N_D+1; Mesub[a*nvars+b](i,j) += JxWoDT * (Phi[i][qp]*Phi[j][qp]);
                        //
                        // capacity contribution
                        a = N_D+2; b = N_D+2; Mesub[a*nvars+b](i,j) += JxWoDT * (Phi[i][qp]*Phi[j][qp]);
                        //
                        // capacity contribution
                        a = N_D+3; b = N_D+3; Mesub[a*nvars+b](i,j) += JxWoDT * (Phi[i][qp]*Phi[j][qp]);
                        //
                        // convection contribution
                        a = N_D+0; b = N_D+0; Mesub[a*nvars+b](i,j) += JxW * (Phi[i][qp]*Veloc_dp0[D]*dPhi_dX[j][qp]);
                        // diffusion contribution
                        a = N_D+0; b = N_D+0; Mesub[a*nvars+b](i,j) += JxW * (dPhi_dX[i][qp]*Diff_dp0[D]*dPhi_dX[j][qp]);
                        // transformation contribution
                        a = N_D+0; b = N_D+0; Mesub[a*nvars+b](i,j) -= JxW * (-Phi[i][qp]*kappa_trn[D]*Phi[j][qp]);
                        // association contribution
                        a = N_D+0; b = N_D+0; Mesub[a*nvars+b](i,j) -= JxW * (-Phi[i][qp]*kappa_dp0[D]*Phi[j][qp]*F_ECM[D]);
                        // dissassociation contribution
                        a = N_D+0; b = N_D+2; Mesub[a*nvars+b](i,j) -= JxW * ( Phi[i][qp]*lamda_dp0[D]*Phi[j][qp]);
                        // lymphatic drainage & decay contribution
                        a = N_D+0; b = N_D+0; Mesub[a*nvars+b](i,j) -= JxW * (-Phi[i][qp]*delta_dp0[D]*Phi[j][qp]);
                        //
                        // convection contribution
                        a = N_D+1; b = N_D+1; Mesub[a*nvars+b](i,j) += JxW * (Phi[i][qp]*Veloc_dp1[D]*dPhi_dX[j][qp]);
                        // diffusion contribution
                        a = N_D+1; b = N_D+1; Mesub[a*nvars+b](i,j) += JxW * (dPhi_dX[i][qp]*Diff_dp1[D]*dPhi_dX[j][qp]);
                        // transformation contribution
                        a = N_D+1; b = N_D+0; Mesub[a*nvars+b](i,j) -= JxW * ( Phi[i][qp]*kappa_trn[D]*Phi[j][qp]);
                        // association contribution
                        a = N_D+1; b = N_D+1; Mesub[a*nvars+b](i,j) -= JxW * (-Phi[i][qp]*kappa_dp1[D]*Phi[j][qp]*F_ECM[D]);
                        // dissassociation contribution
                        a = N_D+1; b = N_D+2; Mesub[a*nvars+b](i,j) -= JxW * ( Phi[i][qp]*lamda_dp1[D]*Phi[j][qp]);
                        // lymphatic drainage & decay contribution
                        a = N_D+1; b = N_D+1; Mesub[a*nvars+b](i,j) -= JxW * (-Phi[i][qp]*delta_dp1[D]*Phi[j][qp]);
                        //
                        // association contribution
                        a = N_D+2; b = N_D+0; Mesub[a*nvars+b](i,j) -= JxW * ( Phi[i][qp]*kappa_dp0[D]*Phi[j][qp]*F_ECM[D]);
                        // association contribution
                        a = N_D+2; b = N_D+1; Mesub[a*nvars+b](i,j) -= JxW * ( Phi[i][qp]*kappa_dp1[D]*Phi[j][qp]*F_ECM[D]);
                        // dissassociation contribution
                        a = N_D+2; b = N_D+2; Mesub[a*nvars+b](i,j) -= JxW * (-Phi[i][qp]*lamda_dp0[D]*Phi[j][qp]);
                        // dissassociation contribution
                        a = N_D+2; b = N_D+2; Mesub[a*nvars+b](i,j) -= JxW * (-Phi[i][qp]*lamda_dp1[D]*Phi[j][qp]);
                        // internalisation contribution
                        a = N_D+2; b = N_D+2; Mesub[a*nvars+b](i,j) -= JxW * (-Phi[i][qp]*kappa_int[D]*Phi[j][qp]);
                        //
                        // internalisation contribution
                        a = N_D+3; b = N_D+2; Mesub[a*nvars+b](i,j) -= JxW * ( Phi[i][qp]*kappa_int[D]*Phi[j][qp]);
                        // decay contribution
                        a = N_D+3; b = N_D+3; Mesub[a*nvars+b](i,j) -= JxW * (-Phi[i][qp]*delta_int[D]*Phi[j][qp]);
                        //
                    }
                    // end of DoFs loop...
                }
                // end of DoFs loop...
            }

            */

            // /*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\
            // \*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/
            // /*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\
            // \*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/*\*/
            // end of volume integration quadrature loop...
        }
        // update system rhs vector
        g_sys.rhs->add_vector(Ye, dof_indices);
        // update the system matrix
        if ( ! use_forward_Euler ) g_sys.matrix->add_matrix(Me, dof_indices);
        // free memo
        delete [] dof_indices_var;
        // end of active local elements loop...
    }
    // close the system matrix or/and rhs-vector
    if ( ! use_forward_Euler ) g_sys.matrix->close();
    g_sys.rhs->close();
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    // calculate the current solution
    if ( use_forward_Euler && ! use_lumped_matrix ) {
        // now solve the (transient) linear system
        g_sys.solve();
        // set this important flag
        g_sys.linear_solver->reuse_preconditioner(true);
    } else if ( use_forward_Euler && use_lumped_matrix ) {
        // calculate the effective rhs vector
        g_sys.get_vector("vector: dg").pointwise_mult(*g_sys.rhs, g_sys.get_vector("capacity matrix: inverted diagonal"));
        g_sys.get_vector("vector: dg").close();
        // update the current solution
        g_sys.solution->add(g_sys.get_vector("vector: dg"));
        g_sys.solution->close();
        g_sys.update();
    } else if ( ! use_forward_Euler ) {
        // now solve the (transient) linearised system
        g_sys.solve();
        // set this important flag
        g_sys.linear_solver->reuse_preconditioner(false);
        // update the current solution
        libmesh_error();
        /*

        g_sys.solution->add(*g_P_sys.solution);
        g_sys.solution->close();
        g_sys.update();

        */

    } else libmesh_error();
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        //
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type>[nvars];
        g_ctx.get_dof_indices(elem, dof_indices, dof_indices_var);
        //
        std::vector<unsigned int> var_dofs;
        calc_var_dofs(nvars, dof_indices_var, var_dofs);
        //
        for (unsigned int l=0; l<var_dofs[0]; l++) {
            const Node* node = elem->node_ptr(l);
            // check all nodal solutions if they have negative values
            const dof_id_type comp = 0;
            for (dof_id_type var=0; var<nvars; var++) {
                const dof_id_type dof = node->dof_number(g_sys.number(), var, comp);
                if ( dof< g_sys.solution->first_local_index() ||
                     dof>=g_sys.solution->last_local_index()  ) continue;
                //
                const Real drug = g_sys.solution->el(dof);
                // enforce to zero out negative values in the solution
                if ( drug < 0.0 ) g_sys.solution->set(dof, 0.0);
            }
            // end of element nodes loop...
        }
        // free memo
        delete [] dof_indices_var;
        // end of active local elements loop...
    }
    // now update the system again
    g_sys.solution->close();
    g_sys.update();
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    if ( update_vasc_params ) {
        // since all above local variables have been initialised then set parameter to "false".
        // the parameter will switch back to "true" when the vascular network has been updated!
        this->_dmn.get_equation_systems().parameters.set<bool>("update vascular parameters for drug delivery solver") = false;
    }
}
//=================================================================================================
