/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/drug_delivery_solver.h"
#include "libmesh/linear_implicit_system.h"
#include "libmesh/petsc_linear_solver.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/quadrature_gauss.h"
#include "libmesh/dof_map.h"
#include "feb3/fem-context.h"
#include "feb3/tumour_angiogenesis_constitutive.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
extern // list of all systems used in the analysis
std::vector< std::pair<EquationSystems*, std::string> > systems_ls;
//=================================================================================================
#include "feb3/domain_data_inl.h"
#include "feb3/solution_io_inl.h"
//=================================================================================================
DrugDeliverySolver::DrugDeliverySolver ( DomainData& dmn,
                                         DomainData_VascularNetwork& dmn_vn ) :
    _log("DrugDeliverySolver"),
    _dmn(dmn), _dmn_vn(dmn_vn) {
    // set the default label
    this->_label = "DrugDeliverySolver";
}
//=================================================================================================
void DrugDeliverySolver::init () {
    this->_log.push("init");

    Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    EquationSystems& es_3d = this->_dmn.get_equation_systems();
    EquationSystems& es_vn = this->_dmn_vn.get_equation_systems();
    // set some system parameters
    this->_copy_params();
    // dimension of the reaction-diffusion solver
    const unsigned int ndrugs = this->par_list.get<unsigned int>("number of drugs"),
                       ndcomp = this->par_list.get<unsigned int>("number of compartments per drug");
    // Lagrange basis order
    const Order lbasis = par.get<Order>("Lagrange basis order");
    //
    char var_name[buffer_len];
    int& hvar = par.set<int>("number of variables to hide in Paraview");
    // add the basic system
    do {
        System& sys = es_vn.add_system<System>(this->_label+"::drug_injection::solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int idrug=0; idrug<ndrugs; idrug++) {
            sprintf(var_name, "%s__g%d", this->_label.c_str(), idrug);
            sys.add_variable(var_name, FIRST, LAGRANGE);
        }
    } while ( false );
    // add another system
    do {
        System& sys = es_vn.add_system<System>(this->_label+"::drug_injection::previous_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int idrug=0; idrug<ndrugs; idrug++) {
            sprintf(var_name, "%s__g%d_PREV", this->_label.c_str(), idrug);
            sys.add_variable(var_name, FIRST, LAGRANGE);
        }
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    } while ( false );
    //
    // add another basic system
    do {
        LinearImplicitSystem& sys = es_3d.add_system<LinearImplicitSystem>(this->_label+"::solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int idrug=0; idrug<ndrugs; idrug++) {
            for (unsigned int icomp=0; icomp<ndcomp; icomp++) {
                sprintf(var_name, "%s__g%d%d", this->_label.c_str(), idrug, icomp);
                sys.add_variable(var_name, lbasis, LAGRANGE);
            }
        }
        sys.add_matrix("capacity matrix");
        sys.add_matrix("capacity matrix: diagonalised");
        sys.add_vector("capacity matrix: inverted diagonal");
        sys.add_vector("vector: dg");
        sys.assemble_before_solve = false;
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    } while ( false );
    // add another system
    do {
        System& sys = es_3d.add_system<System>(this->_label+"::previous_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int idrug=0; idrug<ndrugs; idrug++) {
            for (unsigned int icomp=0; icomp<ndcomp; icomp++) {
                sprintf(var_name, "%s__g%d%d_PREV", this->_label.c_str(), idrug, icomp);
                sys.add_variable(var_name, lbasis, LAGRANGE);
            }
        }
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    } while ( false );
    // add another system
    do {
        System& sys = es_3d.add_system<System>(this->_label+"::solution_time_derivative");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int idrug=0; idrug<ndrugs; idrug++) {
            for (unsigned int icomp=0; icomp<ndcomp; icomp++) {
                sprintf(var_name, "%s__Dg%d%d_Dt", this->_label.c_str(), idrug, icomp);
                sys.add_variable(var_name, lbasis, LAGRANGE);
            }
        }
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    } while ( false );

    this->_log.pop("init");
}
//=================================================================================================
//
//
//
//=================================================================================================
void DrugDeliverySolver::_copy_params () {
    this->_log.push("_copy_params");

    Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    const unsigned int ninvars_NCS = par.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = par.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = par.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = par.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    //
    // sanity check
    ASSERT_(ninvars_NCS==3, "fatal internal error occurred");
    //
    // set the label for the solver
    this->_label = par.get<std::string>("drug delivery solver: label");
    //
    if ( par.get<unsigned int>("input restart from file") )
        this->par_list.set<bool>("enforce to initialize solution") = false;
    //
    // check which implementation to operate
    if ( 4==nvars_RDS && 9==ndcvars_RDS && 8==nvars_AUX ) {
        this->par_list.set<unsigned int>("number of drugs") = 2;
        this->par_list.set<unsigned int>("number of compartments per drug") = 4;
    } else libmesh_error();
    //
    this->par_list.set<unsigned int>("dimension") = this->par_list.get<unsigned int>("number of drugs")
                                                  * this->par_list.get<unsigned int>("number of compartments per drug");
    //
    if ( par.get<bool>("membrane elements present") )
        WARNING_("drug delivery solver hasn't been designed to work with membrane elements");
    //
    for (unsigned int ivar=0; ivar<this->par_list.get<unsigned int>("dimension"); ivar++) {
        char field_name[buffer_len];
        sprintf(field_name, "healthy_domain_initial_field_%d", ivar);
        this->par_list.set<Real>(field_name) = 0.0;
        sprintf(field_name, "healthy_domain_initial_field_%d_deviation", ivar);
        this->par_list.set<Real>(field_name) = 0.0;
        sprintf(field_name, "damaged_domain_initial_field_%d", ivar);
        this->par_list.set<Real>(field_name) = 0.0;
        sprintf(field_name, "damaged_domain_initial_field_%d_deviation", ivar);
        this->par_list.set<Real>(field_name) = 0.0;
        sprintf(field_name, "healthy_to_damaged_domain_initial_field_%d_interface", ivar);
        this->par_list.set<bool>(field_name) = 0.0;
    }
    //
    // ensure this parameter is set "true" first
    par.set<bool>("update vascular parameters for drug delivery solver") = true;

    this->_log.pop("_copy_params");
}
//=================================================================================================
void DrugDeliverySolver::_capacity_matrix () {
    this->_log.push("_capacity_matrix");

    const Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    EquationSystems& es_3d = this->_dmn.get_equation_systems();
    LinearImplicitSystem& g_sys = es_3d.get_system<LinearImplicitSystem>(this->_label+"::solution");
    //
    const FEM_Context g_ctx(g_sys);
    const MeshBase::const_element_iterator begin_el = es_3d.get_mesh().active_local_elements_begin();
    const MeshBase::const_element_iterator end_el   = es_3d.get_mesh().active_local_elements_end();
    //
    const unsigned int nvars = this->par_list.get<unsigned int>("dimension");
    std::vector<unsigned short int> var;
    g_ctx.get_var_list(var);
    //
    FEType fe_g_type = g_sys.variable_type(var[0]);
    //
    const Real time_step = par.get<Real>("tumour angiogenesis solver: time step")
                         * par.get<int>("tumour angiogenesis solver: drug biochemics increment");
    //
    // zero-out the matrix container first
    g_sys.matrix->zero();
    g_sys.get_matrix("capacity matrix").zero();
    g_sys.get_matrix("capacity matrix: diagonalised").zero();
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        //
        QGauss quad_volu(SP3D, par.get<Order>("Gauss quadrature order"));
        UniquePtr<FEBase> fe_g_volu = FEBase::build(SP3D, fe_g_type);
        fe_g_volu->attach_quadrature_rule(&quad_volu);
        //
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type>[nvars];
        g_ctx.get_dof_indices(elem, dof_indices, dof_indices_var);
        //
        std::vector<unsigned int> var_dofs;
        calc_var_dofs(nvars, dof_indices_var, var_dofs);
        //
        const unsigned int ndofs = dof_indices.size();
        DenseMatrix<Number> Me(ndofs, ndofs);
        std::vector< DenseSubMatrix<Number> > Mesub(nvars*nvars, Me);
        std::vector<unsigned int> offset;
        calc_offset(nvars, var_dofs, offset);
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            for (unsigned int jvar=0; jvar<nvars; jvar++) {
                Mesub[ivar*nvars+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
            }
        }
        //
        const std::vector< Real >& JxW_ = fe_g_volu->get_JxW();
        const std::vector< std::vector<Real> >& Phi = fe_g_volu->get_phi();
        fe_g_volu->reinit(elem);
        //
        // loop over volume integration quadrature points
        for (unsigned int qp=0; qp<quad_volu.n_points(); qp++) {
            const Real JxWoDT = JxW_[qp] / time_step;
            // capacity contribution
            for (unsigned int a=0; a<var_dofs[0]; a++)
                for (unsigned int b=0; b<var_dofs[0]; b++)
                    for (unsigned int ivar=0; ivar<nvars; ivar++)
                        Mesub[ivar*nvars+ivar](a,b) += JxWoDT * (Phi[a][qp]*Phi[b][qp]);
            // end of volume integration quadrature loop...
        }
        // update system matrix
        g_sys.matrix->add_matrix(Me, dof_indices);
        g_sys.get_matrix("capacity matrix").add_matrix(Me, dof_indices);
        // row-sum method for lumping the capacity matrix
        std::vector<Real> dia(ndofs, 0.0);
        for (unsigned int a=0; a<ndofs; a++)
            for (unsigned int b=0; b<ndofs; b++)
                dia[a] += Me(b,a);
        Me.zero();
        for (unsigned int a=0; a<ndofs; a++)
            Me(a,a) = dia[a];
        // update (diagonal) system matrix
        g_sys.get_matrix("capacity matrix: diagonalised").add_matrix(Me, dof_indices);
        // free memo
        delete [] dof_indices_var;
        // end of active local elements loop...
    }
    // close the vector and matrix containers
    g_sys.matrix->close();
    g_sys.get_matrix("capacity matrix").close();
    g_sys.get_matrix("capacity matrix: diagonalised").close();
    // evaluate the reciprocal vector of the (diagonal) capacity matrix
    NumericVector<Number>& v = g_sys.get_vector("capacity matrix: inverted diagonal");
    v.zero();
    g_sys.get_matrix("capacity matrix: diagonalised").get_diagonal(v);
    v.reciprocal();
    v.close();
    // ...now set this important flag!
    g_sys.linear_solver->reuse_preconditioner(false);

    this->_log.pop("_capacity_matrix");
}
//=================================================================================================
//
//
//
//=================================================================================================
void DrugDeliverySolver::set_initial () {
    this->_log.push("set_initial");

    EquationSystems& es_vn = this->_dmn_vn.get_equation_systems();
    System& g_sys   = es_vn.get_system<System>(this->_label+"::drug_injection::solution");
    System& g_P_sys = es_vn.get_system<System>(this->_label+"::drug_injection::previous_solution");
    //
    EquationSystems& es = this->_dmn.get_equation_systems();
    System& v_sys   = es.get_system<System>(this->_label+"::solution");
    System& v_P_sys = es.get_system<System>(this->_label+"::previous_solution");
    //
    // by default we need to zero-out the current solution for the drug injection
    g_sys.solution->zero();
    g_sys.solution->close();
    g_sys.update();
    //
    // by default we need to zero-out the previous solution for the drug injection
    g_P_sys.solution->zero();
    g_P_sys.solution->close();
    g_P_sys.update();
    //
    // zero-out the previous solution by default
    v_P_sys.solution->zero();
    // get the initial values for the corresponding fields
    const unsigned int nvars = this->par_list.get<unsigned int>("dimension");
    std::vector<Real> initial_field_value__H(nvars),
                      initial_field_value__D(nvars);
    std::vector<Real> initial_field_devia__H(nvars),
                      initial_field_devia__D(nvars);
    std::vector<bool> initial_field_H_D_interface(nvars);
    for (unsigned int ivar=0; ivar<nvars; ivar++) {
        char field_name[buffer_len];
        sprintf(field_name, "healthy_domain_initial_field_%d", ivar);
        initial_field_value__H[ivar] = this->par_list.get<Real>(field_name);
        sprintf(field_name, "damaged_domain_initial_field_%d", ivar);
        initial_field_value__D[ivar] = this->par_list.get<Real>(field_name);
        sprintf(field_name, "healthy_domain_initial_field_%d_deviation", ivar);
        initial_field_devia__H[ivar] = this->par_list.get<Real>(field_name);
        sprintf(field_name, "damaged_domain_initial_field_%d_deviation", ivar);
        initial_field_devia__D[ivar] = this->par_list.get<Real>(field_name);
        sprintf(field_name, "healthy_to_damaged_domain_initial_field_%d_interface", ivar);
        initial_field_H_D_interface[ivar] = this->par_list.get<bool>(field_name);
    }
    // loop over nodes (in damaged regions)
    for ( std::set<unsigned int>::const_iterator
          citer_nd=this->_dmn.damaged_node.begin(); citer_nd!=this->_dmn.damaged_node.end(); citer_nd++ ) {
        const Node* node = es.get_mesh().node_ptr(*citer_nd);
        libmesh_assert( nvars == node->n_vars(v_P_sys.number()) );
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            const dof_id_type idof = node->dof_number(v_P_sys.number(), ivar, 0);
            libmesh_assert( node->n_comp(v_P_sys.number(), ivar) == 1 );
            // set the corresponding initial value
            const Real val = normal_distro(initial_field_value__D[ivar], initial_field_devia__D[ivar]);
            v_P_sys.solution->set(idof, val);
        }
        // end of nodes loop...
    }
    // loop over nodes (in the interface => healthy/damaged region)
    for ( std::set<unsigned int>::const_iterator
          citer_nd=this->_dmn.interface_node.begin(); citer_nd!=this->_dmn.interface_node.end(); citer_nd++ ) {
        const Node* node = es.get_mesh().node_ptr(*citer_nd);
        libmesh_assert( nvars == node->n_vars(v_P_sys.number()) );
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            const dof_id_type idof = node->dof_number(v_P_sys.number(), ivar, 0);
            libmesh_assert( node->n_comp(v_P_sys.number(), ivar) == 1 );
            if ( initial_field_H_D_interface[ivar] ) {
                // set the corresponding initial value
                const Real val = normal_distro(initial_field_value__H[ivar], initial_field_devia__H[ivar]);
                v_P_sys.solution->set(idof, val);
            } else {
                // set the corresponding initial value
                const Real val = normal_distro(initial_field_value__D[ivar], initial_field_devia__D[ivar]);
                v_P_sys.solution->set(idof, val);
            }
            // end of variables loop..
        }
        // end of nodes loop...
    }
    // loop over nodes (in healthy regions)
    for ( std::set<unsigned int>::const_iterator
          citer_nd=this->_dmn.healthy_node.begin(); citer_nd!=this->_dmn.healthy_node.end(); citer_nd++ ) {
        const Node* node = es.get_mesh().node_ptr(*citer_nd);
        libmesh_assert( nvars == node->n_vars(v_P_sys.number()) );
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            const dof_id_type idof = node->dof_number(v_P_sys.number(), ivar, 0);
            libmesh_assert( node->n_comp(v_P_sys.number(), ivar) == 1 );
            // set the corresponding initial value
            const Real val = normal_distro(initial_field_value__H[ivar], initial_field_devia__H[ivar]);
            v_P_sys.solution->set(idof, val);
        }
        // end of nodes loop...
    }
    // close solution vector and update the system
    v_P_sys.solution->close();
    v_P_sys.update();
    //
    v_sys.solution->operator=(*v_P_sys.solution);
    v_sys.solution->close();
    v_sys.update();
    //
    // make sure you have initialised the capacity (mass) matrix of the system
    DrugDeliverySolver::_capacity_matrix();
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("set_initial");
}
//=================================================================================================
void DrugDeliverySolver::injection () {
    this->_log.push("injection");

    EquationSystems& es_3d = this->_dmn.get_equation_systems();
    const Parameters& par = es_3d.parameters;
    //
    const unsigned int ninvars_NCS = par.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = par.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = par.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = par.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    //
    // sanity check
    ASSERT_(ninvars_NCS==3, "fatal internal error occurred");
    //
    // check which implementation to operate
    if ( 4==nvars_RDS && 9==ndcvars_RDS && 8==nvars_AUX ) this->_injection_498();
    else libmesh_error();
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("injection");
}
//=================================================================================================
void DrugDeliverySolver::delivery () {
    this->_log.push("delivery");

    EquationSystems& es_3d = this->_dmn.get_equation_systems();
    const Parameters& par = es_3d.parameters;
    //
    const unsigned int ninvars_NCS = par.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = par.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = par.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = par.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    //
    const Real time_step = par.get<Real>("tumour angiogenesis solver: time step")
                         * par.get<int>("tumour angiogenesis solver: drug biochemics increment");
    //
    // sanity check
    ASSERT_(ninvars_NCS==3, "fatal internal error occurred");
    //
    // check which implementation to operate
    if ( 4==nvars_RDS && 9==ndcvars_RDS && 8==nvars_AUX ) this->_delivery_498();
    else libmesh_error();
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    const System& g_sys = es_3d.get_system<System>(this->_label+"::solution");
    System&  g_P_sys = es_3d.get_system<System>(this->_label+"::previous_solution");
    System& DgDt_sys = es_3d.get_system<System>(this->_label+"::solution_time_derivative");
    // calculate the rate of the solution
    DgDt_sys.solution->operator=(*g_sys.solution);
    DgDt_sys.solution->operator-=(*g_P_sys.solution);
    DgDt_sys.solution->operator*=(1.0/time_step);
    DgDt_sys.solution->close();
    DgDt_sys.update();
    // update the previous solution to the current one
    g_P_sys.solution->operator=(*g_sys.solution);
    g_P_sys.solution->close();
    g_P_sys.update();
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("delivery");
}
//=================================================================================================
