/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/explicit_navier_cauchy_solver_3d.h"
#include "libmesh/mesh.h"
#include "libmesh/equation_systems.h"
#include "libmesh/linear_implicit_system.h"
#include "libmesh/petsc_linear_solver.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/fe.h"
#include "feb3/quadrature.h"
#include "libmesh/dof_map.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/dense_matrix.h"
#include "libmesh/dense_vector.h"
#include "libmesh/dense_submatrix.h"
#include "libmesh/dense_subvector.h"
#include "libmesh/numeric_vector.h"
#include "libmesh/linear_implicit_system.h"
#include "libmesh/boundary_info.h"
#include "feb3/navier_cauchy_solver_3d_aux_inl.h"
#include "feb3/explicit_navier_cauchy_solver_3d_aux_inl.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
//=================================================================================================
void ExplicitNavierCauchySolver3D::_assemble_mass () {
    this->_log.push("_assemble_mass");

    EquationSystems& es = this->_dmn.get_equation_systems();
    SparseMatrix<Number>* system_matrix = es.get_system<LinearImplicitSystem>(this->_label+"::next_solution").matrix;
    SparseMatrix<Number>& mass_d = es.get_system<LinearImplicitSystem>(this->_label+"::next_solution").get_matrix("mass (doubled)");
    SparseMatrix<Number>& mass_n = es.get_system<LinearImplicitSystem>(this->_label+"::next_solution").get_matrix("mass (negative)");
    SparseMatrix<Number>& damping = es.get_system<LinearImplicitSystem>(this->_label+"::next_solution").get_matrix("damping");
    System& u_sys = es.get_system<System>(this->_label+"::solution");
    //
    const FEM_Context u_ctx(u_sys);
    //
    const MeshBase::const_element_iterator begin_el = es.get_mesh().active_local_elements_begin();
    const MeshBase::const_element_iterator end_el = es.get_mesh().active_local_elements_end();
    //
    libmesh_assert( SP3D == es.get_mesh().mesh_dimension() );
    libmesh_assert( SP3D == u_sys.n_vars() );
    //
    FEType fe_u_type = u_sys.variable_type(0);
    //
    const bool membrane_present  = this->par_list.get<bool>("membrane elements present"),
               // apply_body_force  = this->par_list.get<bool>("apply body force"),
               apply_nodal_force = this->par_list.get<bool>("apply nodal force");
    if ( membrane_present ) libmesh_error();
    if ( apply_nodal_force ) libmesh_error();
    const bool apply_wound_contraction = this->par_list.get<bool>("apply wound contraction"),
               apply_tumour_growth     = this->par_list.get<bool>("apply tumour growth");
    if ( apply_wound_contraction && apply_tumour_growth ) libmesh_error();
    const unsigned int ninvars_NCS = es.parameters.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = es.parameters.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = es.parameters.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = es.parameters.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    const Real current_time = this->par_list.get<Real>("current time");
    const Real M_coeff = 1.0 / pow2(this->par_list.get<Real>("time step")),
               C_coeff = 0.5 / this->par_list.get<Real>("time step")
                       * this->par_list.get<Real>("mass-proportional damping coefficient");
    //
    Real system_matrix_max_local  = 0.0,
         system_matrix_max_global = 0.0;
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // zero-out system matrix and mass/damping matrices
    system_matrix->zero();
    mass_d.zero();
    mass_n.zero();
    damping.zero();
    //
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        //
        QGauss quad_volu(SP3D, es.parameters.get<Order>("Gauss quadrature order"));
        UniquePtr<FEBase> fe_u_volu = FEBase::build(SP3D, fe_u_type);
        fe_u_volu->attach_quadrature_rule(&quad_volu);
        //
        const unsigned int nvars = SP3D;
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
        std::vector<dof_id_type> vars;
        for (unsigned int i=0; i<nvars; i++) vars.push_back( i );
        u_ctx.get_dof_indices(elem, dof_indices, dof_indices_var, &vars);
        //
        std::vector<unsigned int> var_dofs;
        calc_var_dofs(nvars, dof_indices_var, var_dofs);
        //
        DenseMatrix<Number> Me(dof_indices.size(), dof_indices.size());
        std::vector< DenseSubMatrix<Number> > Mesub(nvars*nvars, Me);
        DenseMatrix<Number> Ce(dof_indices.size(), dof_indices.size());
        std::vector< DenseSubMatrix<Number> > Cesub(nvars*nvars, Ce);
        DenseMatrix<Number> MCe(dof_indices.size(), dof_indices.size());
        std::vector< DenseSubMatrix<Number> > MCesub(nvars*nvars, MCe);
        std::vector<unsigned int> offset;
        calc_offset(nvars, var_dofs, offset);
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            for (unsigned int jvar=0; jvar<nvars; jvar++) {
                Mesub[ivar*nvars+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
                Cesub[ivar*nvars+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
                MCesub[ivar*nvars+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
            }
        }
        // get access to the material region
        libmesh_assert( this->_dmn.regn.end() != this->_dmn.regn.find(elem->subdomain_id()) );
        const RegionInfo* ri_ptr = &(this->_dmn.regn.find(elem->subdomain_id())->second);
        // initialize and get access to the FE stuff
        const std::vector< Real >&               JxW = fe_u_volu->get_JxW();
        const std::vector< std::vector<Real> >&  Phi = fe_u_volu->get_phi();
        fe_u_volu->reinit(elem);
        // get access to the element quadrature points data
        libmesh_assert( this->_dmn.elem_qi_local.end() != this->_dmn.elem_qi_local.find(elem->id()) );
        const QuadratureInfo* qpi_ptr = &(this->_dmn.elem_qi_local.find(elem->id())->second);
        // loop over volume integration quadrature points
        for (unsigned int qp=0; qp<quad_volu.n_points(); qp++) {
            //
            FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
            fld.current_time = current_time;
            //
            fld.biomech[0] = ri_ptr->damage_factor_const;
            ri_ptr->calc_damage(fld);
            fld.biomech[1] = ri_ptr->mat->get_biomech_param(0) * (1.0-fld.biomech[0]);
            fld.biomech[2] = qpi_ptr->pressure[qp];
            //
            const Real rho = ri_ptr->mat->get_biomech_param(0) * (1.0-fld.biomech[0]);
            //
            for (unsigned int a=0; a<var_dofs[0]; a++) {
                for (unsigned int b=0; b<var_dofs[0]; b++) {
                    const Real Mm = Phi[a][qp]*rho*Phi[b][qp]*M_coeff,
                               Cm = Phi[a][qp]*rho*Phi[b][qp]*C_coeff;
                    //
                    Mesub[0*nvars+0](a,b) += JxW[qp] * Mm;
                    Mesub[1*nvars+1](a,b) += JxW[qp] * Mm;
                    Mesub[2*nvars+2](a,b) += JxW[qp] * Mm;
                    //
                    Cesub[0*nvars+0](a,b) += JxW[qp] * Cm;
                    Cesub[1*nvars+1](a,b) += JxW[qp] * Cm;
                    Cesub[2*nvars+2](a,b) += JxW[qp] * Cm;
                    //
                    MCesub[0*nvars+0](a,b) += JxW[qp] * (Mm+Cm);
                    MCesub[1*nvars+1](a,b) += JxW[qp] * (Mm+Cm);
                    MCesub[2*nvars+2](a,b) += JxW[qp] * (Mm+Cm);
                }
            }
            // end of volume integration quadrature loop...
        }
        system_matrix_max_local = std::max(system_matrix_max_local, MCe.max());
        // update element system matrix and mass/damping matrices
        system_matrix->add_matrix(MCe, dof_indices);
        Me.scale(+2.0); mass_d.add_matrix(Me, dof_indices);
        Me.scale(-0.5); mass_n.add_matrix(Me, dof_indices);
        damping.add_matrix(Ce, dof_indices);
        // free memo
        delete [] dof_indices_var;
        // end of active local elements loop...
    }
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // evaluate the penalty parameter for the mass matrix
    MPI_Allreduce(&system_matrix_max_local, &system_matrix_max_global, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
    // set the penalty parameter
    const Real penalty = system_matrix_max_global * es.parameters.get<Real>("penalty parameter scale");
    this->par_list.set<Real>("penalty parameter") = penalty;
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        //
        // insert essential BCs for "boundary" side elements
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            // check if element has this side on boundary and get the boundary ID for side of element
            if ( elem->neighbor_ptr(s) != NULL ) continue;
            // create a side element
            UniquePtr<const Elem> side = elem->build_side_ptr(s);
            //
            const unsigned int nvars = SP3D;
            std::vector<dof_id_type> dof_indices;
            std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
            std::vector<dof_id_type> vars;
            for (unsigned int i=0; i<nvars; i++) vars.push_back( i );
            u_ctx.get_dof_indices(side.get(), dof_indices, dof_indices_var, &vars);
            //
            std::vector<unsigned int> var_dofs;
            calc_var_dofs(nvars, dof_indices_var, var_dofs);
            //
            DenseMatrix<Number> Me(dof_indices.size(), dof_indices.size());
            std::vector< DenseSubMatrix<Number> > Mesub(nvars*nvars, Me);
            std::vector<unsigned int> offset;
            calc_offset(nvars, var_dofs, offset);
            for (unsigned int ivar=0; ivar<nvars; ivar++) {
                for (unsigned int jvar=0; jvar<nvars; jvar++) {
                    Mesub[ivar*nvars+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
                }
            }
            //
            // get the ID of the boundary
            const boundary_id_type bc_id = es.get_mesh().boundary_info->boundary_id(elem, s);
            // access the boundary condition info data
            const BoundaryConditionInfo* bci_ptr = &(this->_dmn.bcs.find(bc_id)->second);
            // distinguish among types of boundary conditions
            if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_PR == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED == bci_ptr->type    ) {
                for (unsigned int a=0; a<var_dofs[0]; a++) {
                    Mesub[0*nvars+0](a,a) += penalty;
                    Mesub[1*nvars+1](a,a) += penalty;
                    Mesub[2*nvars+2](a,a) += penalty;
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_X == bci_ptr->type ) {
                for (unsigned int a=0; a<var_dofs[0]; a++) {
                    Mesub[0*nvars+0](a,a) += penalty;
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Y == bci_ptr->type ) {
                for (unsigned int a=0; a<var_dofs[0]; a++) {
                    Mesub[1*nvars+1](a,a) += penalty;
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Z == bci_ptr->type ) {
                for (unsigned int a=0; a<var_dofs[0]; a++) {
                    Mesub[2*nvars+2](a,a) += penalty;
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XY == bci_ptr->type ) {
                for (unsigned int a=0; a<var_dofs[0]; a++) {
                    Mesub[0*nvars+0](a,a) += penalty;
                    Mesub[1*nvars+1](a,a) += penalty;
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_YZ == bci_ptr->type ) {
                for (unsigned int a=0; a<var_dofs[0]; a++) {
                    Mesub[1*nvars+1](a,a) += penalty;
                    Mesub[2*nvars+2](a,a) += penalty;
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XZ == bci_ptr->type ) {
                for (unsigned int a=0; a<var_dofs[0]; a++) {
                    Mesub[0*nvars+0](a,a) += penalty;
                    Mesub[2*nvars+2](a,a) += penalty;
                }
            } else if ( BoundaryConditionInfo::NEUMANN_VECTOR == bci_ptr->type ) {
                ; // do nothing here
            } else if ( BoundaryConditionInfo::NEUMANN_NORMAL == bci_ptr->type ) {
                ; // do nothing here
            } else libmesh_error();
            // update element system matrix
            system_matrix->add_matrix(Me, dof_indices);
            // free memo
            delete [] dof_indices_var;
            // end of element sides loop...
        }
        // end of active local elements loop...
    }
    //
    // system matrix and close mass/damping matrices
    system_matrix->close();
    mass_d.close();
    mass_n.close();
    damping.close();
    // enforce to update this flag
    this->par_list.set<bool>("system matrix is factorized") = false;
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("_assemble_mass");
}
//=================================================================================================
void ExplicitNavierCauchySolver3D::_assemble_rhs () {
    this->_log.push("_assemble_rhs");

    EquationSystems& es = this->_dmn.get_equation_systems();
    NumericVector<Number>* system_rhs = es.get_system<LinearImplicitSystem>(this->_label+"::next_solution").rhs;
    NumericVector<Number>& residual = es.get_system<LinearImplicitSystem>(this->_label+"::next_solution").get_vector("residual");
    System& u_sys = es.get_system<System>(this->_label+"::solution");
    //
    std::map<std::string, EquationSystems>& es_rve =
        *(es.parameters.set<std::map<std::string, EquationSystems>*>("pointer to RVE equation system(s)"));
    //
    const FEM_Context u_ctx(u_sys);
    //
    const MeshBase::const_element_iterator begin_el = es.get_mesh().active_local_elements_begin();
    const MeshBase::const_element_iterator end_el = es.get_mesh().active_local_elements_end();
    //
    libmesh_assert( SP3D == es.get_mesh().mesh_dimension() );
    libmesh_assert( SP3D == u_sys.n_vars() );
    //
    FEType fe_u_type = u_sys.variable_type(0);
    //
    const bool membrane_present  = this->par_list.get<bool>("membrane elements present"),
               apply_body_force  = this->par_list.get<bool>("apply body force"),
               apply_nodal_force = this->par_list.get<bool>("apply nodal force");
    if ( membrane_present ) libmesh_error();
    if ( apply_nodal_force ) libmesh_error();
    const bool apply_wound_contraction = this->par_list.get<bool>("apply wound contraction"),
               apply_tumour_growth     = this->par_list.get<bool>("apply tumour growth");
    if ( apply_wound_contraction && apply_tumour_growth ) libmesh_error();
    const unsigned int ninvars_NCS = es.parameters.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = es.parameters.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = es.parameters.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = es.parameters.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    const Real current_time = this->par_list.get<Real>("current time");
    //
    // zero-out system RHS vector and residual vector
    system_rhs->zero();
    residual.zero();
    //
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        //
        QGauss quad_volu(SP3D, es.parameters.get<Order>("Gauss quadrature order"));
        UniquePtr<FEBase> fe_u_volu = FEBase::build(SP3D, fe_u_type);
        fe_u_volu->attach_quadrature_rule(&quad_volu);
        UniquePtr<FEBase> fe_u_volu_m = FEBase::build(SP3D, fe_u_type);
        fe_u_volu_m->attach_quadrature_rule(&quad_volu);
        //
        const unsigned int nvars = SP3D;
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
        std::vector<dof_id_type> vars;
        for (unsigned int i=0; i<nvars; i++) vars.push_back( i );
        u_ctx.get_dof_indices(elem, dof_indices, dof_indices_var, &vars);
        //
        std::vector<unsigned int> var_dofs;
        calc_var_dofs(nvars, dof_indices_var, var_dofs);
        //
        DenseVector<Number> Fe(dof_indices.size());
        std::vector< DenseSubVector<Number> > Fesub(nvars, Fe);
        DenseVector<Number> Re(dof_indices.size());
        std::vector< DenseSubVector<Number> > Resub(nvars, Re);
        std::vector<unsigned int> offset;
        calc_offset(nvars, var_dofs, offset);
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            Fesub[ivar].reposition(offset[ivar], var_dofs[ivar]);
            Resub[ivar].reposition(offset[ivar], var_dofs[ivar]);
        }
        // get access to the material region
        libmesh_assert( this->_dmn.regn.end() != this->_dmn.regn.find(elem->subdomain_id()) );
        const RegionInfo* ri_ptr = &(this->_dmn.regn.find(elem->subdomain_id())->second);
        // initialize and get access to the FE stuff
        const std::vector< Real >&                          JxW = fe_u_volu->get_JxW();
        const std::vector< std::vector<Real> >&             Phi = fe_u_volu->get_phi();
        const std::vector< std::vector<RealGradient> >& dPhi_dX = fe_u_volu->get_dphi();
        fe_u_volu->reinit(elem);
        // parametric coordinates of the quadrature points
        const std::vector<Point>& q_pnt = quad_volu.get_points();
        // get access to the element quadrature points data
        libmesh_assert( this->_dmn.elem_qi_local.end() != this->_dmn.elem_qi_local.find(elem->id()) );
        const QuadratureInfo* qpi_ptr = &(this->_dmn.elem_qi_local.find(elem->id())->second);
        // loop over volume integration quadrature points
        for (unsigned int qp=0; qp<quad_volu.n_points(); qp++) {
            // key to access item from map
            const std::string key = number2string(elem->id()) + "_" + number2string(qp);
            // get access to the equation-systems corresponding to the RVE of this quadrature point
            libmesh_assert( es_rve.end() != es_rve.find(number2string(elem->id())) );
            EquationSystems& es_m = es_rve.find(key)->second;
            //
            FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
            fld.current_time = current_time;
            for (unsigned int ivar=0; ivar<SP3D; ivar++) {
                for (unsigned int l=0; l<var_dofs[ivar]; l++) {
                    const Real ui = u_sys.current_solution(dof_indices_var[ivar][l]);
                    fld.displacement(ivar) += Phi[l][qp] * ui;
                    fld.grad_displacement(ivar) += dPhi_dX[l][qp] * ui;
                }
            }
            // previous-step inelastic deformation gradient tensor
            fld.deform_grad_inel = qpi_ptr->deform_grad_inel[qp];
            // deformation gradient tensor
            fld.deform_gradient = deformation_gradient_3d(fld.grad_displacement);
            for (unsigned int ivar=0; ivar<nvars_RDS+ndcvars_RDS; ivar++) {
                fld.biochem[ivar] = qpi_ptr->biochem[qp][ivar];
            }
            for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                fld.grad_biochem[ivar] = qpi_ptr->grad_biochem[qp][ivar];
                fld.biochem_rate[ivar] = qpi_ptr->biochem_rate[qp][ivar];
            }
            for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                fld.auxiliary[ivar] = qpi_ptr->auxiliary[qp][ivar];
                fld.grad_auxiliary[ivar] = qpi_ptr->grad_auxiliary[qp][ivar];
                fld.auxiliary_rate[ivar] = qpi_ptr->auxiliary_rate[qp][ivar];
            }
            for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                fld.fibre_ref[ispdm] = qpi_ptr->fibre_ref[ispdm][qp];
                fld.fibre_curr[ispdm] = unit(fld.deform_gradient * fld.fibre_ref[ispdm]);
            }
            fld.strain = GreenLagrange_strain(fld.deform_gradient);
            fld.stress = qpi_ptr->stress[qp];
            fld.pressure = qpi_ptr->pressure[qp];
            //
            fld.biomech[0] = ri_ptr->damage_factor_const;
            ri_ptr->calc_damage(fld);
            fld.biomech[1] = ri_ptr->mat->get_biomech_param(0) * (1.0-fld.biomech[0]);
            fld.biomech[2] = qpi_ptr->pressure[qp];
            //
            const Real rho = ri_ptr->mat->get_biomech_param(0) * (1.0-fld.biomech[0]);
            //
            std::vector<FieldInfo> fld_m(8, FieldInfo(0, 0, SP3D));
            calc_fields_RVE_(fe_u_volu_m, fld_m);
            //
            ri_ptr->calc_plastic_deform(fld);
            // second Piola-Kirchhoff stress tensor
            Dyadic3D S2pk;
            Quindic3D Dtsm_uu;
            Real Ph;
            Dyadic3D Dtsm_up, Dtsm_pu;
            Real Dtsm_pp;
            Vector3D Qs;
            // passive contribution of tissue biomechanics
            ri_ptr->calc_biomech_3d(fld, S2pk, Dtsm_uu, Dtsm_up, Ph, Dtsm_pu, Dtsm_pp);
            // wound contraction or tumour growth model
            if ( apply_wound_contraction || apply_tumour_growth )
                ri_ptr->calc_coupled_3d(fld, S2pk, Dtsm_uu, Dtsm_up, Ph, Dtsm_pu, Dtsm_pp);
            // multiscale model contribution
            ri_ptr->calc_multiscale(fld, es_m, fld_m, S2pk, Qs);
            // internal stresses
            for (unsigned int a=0; a<var_dofs[0]; a++) {
                const Vector3D Rv = calc_internal_force_forward(dPhi_dX[a][qp], S2pk, fld.deform_gradient);
                //
                Resub[0](a) -= JxW[qp] * Rv(0);
                Resub[1](a) -= JxW[qp] * Rv(1);
                Resub[2](a) -= JxW[qp] * Rv(2);
            }
            // external loading (body force)
            if ( apply_body_force )
            for (unsigned int a=0; a<var_dofs[0]; a++) {
                const Vector3D Bf = calc_body_force(Phi[a][qp], rho, this->_dmn.bforce, current_time);
                //
                Resub[0](a) += JxW[qp] * Bf(0);
                Resub[1](a) += JxW[qp] * Bf(1);
                Resub[2](a) += JxW[qp] * Bf(2);
            }
            // multiscale model source term
            for (unsigned int a=0; a<var_dofs[0]; a++) {
                Resub[0](a) += JxW[qp] * Qs(0);
                Resub[1](a) += JxW[qp] * Qs(1);
                Resub[2](a) += JxW[qp] * Qs(2);
            }
            // end of volume integration quadrature loop...
        }
        // update element system rhs and residual vector
        system_rhs->add_vector(Re, dof_indices);
        system_rhs->add_vector(Fe, dof_indices);
        residual.add_vector(Re, dof_indices);
        // free memo
        delete [] dof_indices_var;
        // end of active local elements loop...
    }
    //
    const Real penalty = this->par_list.get<Real>("penalty parameter");
    //
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        // insert both Dirichlet and Neumann boundary conditions
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            // check if element has this side on boundary and get the boundary ID for side of element
            if ( elem->neighbor_ptr(s) != NULL ) continue;
            // create a side element
            UniquePtr<const Elem> side = elem->build_side_ptr(s);
            //
            QGauss quad_surf(SP2D, es.parameters.get<Order>("Gauss quadrature order"));
            UniquePtr<FEBase> fe_u_surf = FEBase::build(SP2D, fe_u_type);
            fe_u_surf->attach_quadrature_rule(&quad_surf);
            //
            const unsigned int nvars = SP3D;
            std::vector<dof_id_type> dof_indices;
            std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
            std::vector<dof_id_type> vars;
            for (unsigned int i=0; i<nvars; i++) vars.push_back( i );
            u_ctx.get_dof_indices(side.get(), dof_indices, dof_indices_var, &vars);
            //
            std::vector<unsigned int> var_dofs;
            calc_var_dofs(nvars, dof_indices_var, var_dofs);
            //
            DenseVector<Number> Fe(dof_indices.size());
            std::vector< DenseSubVector<Number> > Fesub(nvars, Fe);
            DenseVector<Number> Re(dof_indices.size());
            std::vector< DenseSubVector<Number> > Resub(nvars, Re);
            std::vector<unsigned int> offset;
            calc_offset(nvars, var_dofs, offset);
            for (unsigned int ivar=0; ivar<nvars; ivar++) {
                Fesub[ivar].reposition(offset[ivar], var_dofs[ivar]);
                Resub[ivar].reposition(offset[ivar], var_dofs[ivar]);
            }
            //
            // get the ID of the boundary
            const boundary_id_type bc_id = es.get_mesh().boundary_info->boundary_id(elem, s);
            // access the boundary condition info data
            const BoundaryConditionInfo* bci_ptr = &(this->_dmn.bcs.find(bc_id)->second);
            // distinguish among types of boundary conditions
            if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_PR == bci_ptr->type ) {
                const Real angle = bci_ptr->value(0)*bci_ptr->calc_time_function(current_time);
                for (unsigned int a=0; a<var_dofs[0]; a++) {
                    const Point ur = rotate_on_plane(side->point(a)-bci_ptr->center, angle, bci_ptr->normal)
                                   - side->point(a) + bci_ptr->center;
                    Fesub[0](a) += penalty * ur(0);
                    Fesub[1](a) += penalty * ur(1);
                    Fesub[2](a) += penalty * ur(2);
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED == bci_ptr->type ) {
                const Real u0 = bci_ptr->value(0)*bci_ptr->calc_time_function(current_time),
                           u1 = bci_ptr->value(1)*bci_ptr->calc_time_function(current_time),
                           u2 = bci_ptr->value(2)*bci_ptr->calc_time_function(current_time);
                for (unsigned int a=0; a<var_dofs[0]; a++) {
                    Fesub[0](a) += penalty * u0;
                    Fesub[1](a) += penalty * u1;
                    Fesub[2](a) += penalty * u2;
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_X == bci_ptr->type ) {
                const Real u0 = bci_ptr->value(0)*bci_ptr->calc_time_function(current_time);
                for (unsigned int a=0; a<var_dofs[0]; a++) {
                    Fesub[0](a) += penalty * u0;
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Y == bci_ptr->type ) {
                const Real u1 = bci_ptr->value(1)*bci_ptr->calc_time_function(current_time);
                for (unsigned int a=0; a<var_dofs[0]; a++) {
                    Fesub[1](a) += penalty * u1;
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Z == bci_ptr->type ) {
                const Real u2 = bci_ptr->value(2)*bci_ptr->calc_time_function(current_time);
                for (unsigned int a=0; a<var_dofs[0]; a++) {
                    Fesub[2](a) += penalty * u2;
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XY == bci_ptr->type ) {
                const Real u0 = bci_ptr->value(0)*bci_ptr->calc_time_function(current_time),
                           u1 = bci_ptr->value(1)*bci_ptr->calc_time_function(current_time);
                for (unsigned int a=0; a<var_dofs[0]; a++) {
                    Fesub[0](a) += penalty * u0;
                    Fesub[1](a) += penalty * u1;
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_YZ == bci_ptr->type ) {
                const Real u1 = bci_ptr->value(1)*bci_ptr->calc_time_function(current_time),
                           u2 = bci_ptr->value(2)*bci_ptr->calc_time_function(current_time);
                for (unsigned int a=0; a<var_dofs[0]; a++) {
                    Fesub[1](a) += penalty * u1;
                    Fesub[2](a) += penalty * u2;
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XZ == bci_ptr->type ) {
                const Real u0 = bci_ptr->value(0)*bci_ptr->calc_time_function(current_time),
                           u2 = bci_ptr->value(2)*bci_ptr->calc_time_function(current_time);
                for (unsigned int a=0; a<var_dofs[0]; a++) {
                    Fesub[0](a) += penalty * u0;
                    Fesub[2](a) += penalty * u2;
                }
            } else if ( BoundaryConditionInfo::NEUMANN_VECTOR == bci_ptr->type ) {
                const Point traction = Point(bci_ptr->value(0), bci_ptr->value(1), bci_ptr->value(2))
                                     * bci_ptr->calc_time_function(current_time);
                // initialize and get access to the FE stuff
                const std::vector<Real>&                JxW = fe_u_surf->get_JxW();
                const std::vector< std::vector<Real> >& Phi = fe_u_surf->get_phi();
                fe_u_surf->reinit(side.get());
                // check if traction magnitude is above some threshold
                if ( traction.norm() > 1.0e-6 )
                // loop over surface integration quadrature points
                for (unsigned int qp=0; qp<quad_surf.n_points(); qp++) {
                    // external loading (traction)
                    for (unsigned int a=0; a<var_dofs[0]; a++) {
                        Vector3D Re;
                        Re.comp = Phi[a][qp]*vec(traction).comp;
                        //
                        Resub[0](a) += JxW[qp] * Re(0);
                        Resub[1](a) += JxW[qp] * Re(1);
                        Resub[2](a) += JxW[qp] * Re(2);
                    }
                    // end of surface integration quadrature loop...
                }
            } else if ( BoundaryConditionInfo::NEUMANN_NORMAL == bci_ptr->type ) {
                const Real pressure = bci_ptr->value(0)*bci_ptr->calc_time_function(current_time);
                // initialize and get access to the FE stuff
                const std::vector<Real>&                 JxW = fe_u_surf->get_JxW();
                const std::vector< std::vector<Real> >&  Phi = fe_u_surf->get_phi();
                const std::vector< std::vector<Real> >& dPhi_dZ0 = fe_u_surf->get_dphidxi();
                const std::vector< std::vector<Real> >& dPhi_dZ1 = fe_u_surf->get_dphideta();
                const std::vector< std::vector<RealGradient> >& dPhi_dX = fe_u_surf->get_dphi();
                fe_u_surf->reinit(side.get());
                // check if pressure magnitude is above some threshold
                if ( fabs(pressure) > 1.0e-6 )
                // loop over surface integration quadrature points
                for (unsigned int qp=0; qp<quad_surf.n_points(); qp++) {
                    //
                    DisplacementGradient du_dX;
                    for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                        for (unsigned int l=0; l<var_dofs[0]; l++) {
                            du_dX(ispdm) += dPhi_dX[l][qp] * u_sys.current_solution(dof_indices_var[ispdm][l]);
                        }
                    }
                    //
                    Point dX_dZ0, dX_dZ1;
                    Point dx_dZ0, dx_dZ1;
                    for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                        for (unsigned int l=0; l<var_dofs[0]; l++) {
                            dX_dZ0(ispdm) += dPhi_dZ0[l][qp] * side->point(l)(ispdm);
                            dX_dZ1(ispdm) += dPhi_dZ1[l][qp] * side->point(l)(ispdm);
                            dx_dZ0(ispdm) += dPhi_dZ0[l][qp] * (side->point(l)(ispdm)+u_sys.current_solution(dof_indices_var[ispdm][l]));
                            dx_dZ1(ispdm) += dPhi_dZ1[l][qp] * (side->point(l)(ispdm)+u_sys.current_solution(dof_indices_var[ispdm][l]));
                        }
                    }
                    // reciprocal of element Jacobian determinant
                    const Real J_recip = 1.0 / dX_dZ0.cross(dX_dZ1).norm();
                    // normal vector to side element (at current step)
                    const Point normal = dx_dZ0.cross(dx_dZ1);
                    // external loading (normal traction)
                    for (unsigned int a=0; a<var_dofs[0]; a++) {
                        Vector3D Re;
                        Re.comp = Phi[a][qp]*pressure*vec(normal).comp*J_recip;
                        //
                        Resub[0](a) += JxW[qp] * Re(0);
                        Resub[1](a) += JxW[qp] * Re(1);
                        Resub[2](a) += JxW[qp] * Re(2);
                    }
                    // end of surface integration quadrature loop...
                }
                // exception caught for boundary condition type
            } else libmesh_error();
            // update element system rhs and residual vector
            system_rhs->add_vector(Re, dof_indices);
            system_rhs->add_vector(Fe, dof_indices);
            residual.add_vector(Re, dof_indices);
            // free memo
            delete [] dof_indices_var;
            // end of element sides loop...
        }
        // end of active local elements loop...
    }
    //
    // close system rhs and residual vector
    system_rhs->close();
    residual.close();
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("_assemble_rhs");
}
//=================================================================================================
void ExplicitNavierCauchySolver3D::_set_quad_info () {
    this->_log.push("_set_quad_info");

    EquationSystems& es = this->_dmn.get_equation_systems();
    System& u_sys = es.get_system<System>(this->_label+"::solution");
    //
    std::map<std::string, EquationSystems>& es_rve =
        *(es.parameters.set<std::map<std::string, EquationSystems>*>("pointer to RVE equation system(s)"));
    //
    const FEM_Context u_ctx(u_sys);
    //
    const MeshBase::const_element_iterator begin_el = es.get_mesh().active_local_elements_begin();
    const MeshBase::const_element_iterator end_el = es.get_mesh().active_local_elements_end();
    //
    libmesh_assert( SP3D == es.get_mesh().mesh_dimension() );
    libmesh_assert( SP3D == u_sys.n_vars() );
    //
    FEType fe_u_type = u_sys.variable_type(0);
    //
    const bool membrane_present  = this->par_list.get<bool>("membrane elements present"),
               // apply_body_force  = this->par_list.get<bool>("apply body force"),
               apply_nodal_force = this->par_list.get<bool>("apply nodal force");
    if ( membrane_present ) libmesh_error();
    if ( apply_nodal_force ) libmesh_error();
    const bool apply_wound_contraction = this->par_list.get<bool>("apply wound contraction"),
               apply_tumour_growth     = this->par_list.get<bool>("apply tumour growth");
    if ( apply_wound_contraction && apply_tumour_growth ) libmesh_error();
    const unsigned int ninvars_NCS = es.parameters.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = es.parameters.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = es.parameters.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = es.parameters.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    const Real current_time = this->par_list.get<Real>("current time");
    //
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        //
        QGauss quad_volu(SP3D, es.parameters.get<Order>("Gauss quadrature order"));
        UniquePtr<FEBase> fe_u_volu = FEBase::build(SP3D, fe_u_type);
        fe_u_volu->attach_quadrature_rule(&quad_volu);
        UniquePtr<FEBase> fe_u_volu_m = FEBase::build(SP3D, fe_u_type);
        fe_u_volu_m->attach_quadrature_rule(&quad_volu);
        //
        const unsigned int nvars = SP3D;
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
        std::vector<dof_id_type> vars;
        for (unsigned int i=0; i<nvars; i++) vars.push_back( i );
        u_ctx.get_dof_indices(elem, dof_indices, dof_indices_var, &vars);
        //
        std::vector<unsigned int> var_dofs;
        calc_var_dofs(nvars, dof_indices_var, var_dofs);
        //
        // get access to the material region
        libmesh_assert( this->_dmn.regn.end() != this->_dmn.regn.find(elem->subdomain_id()) );
        const RegionInfo* ri_ptr = &(this->_dmn.regn.find(elem->subdomain_id())->second);
        // initialize and get access to the FE stuff
        const std::vector< std::vector<Real> >&             Phi = fe_u_volu->get_phi();
        const std::vector< std::vector<RealGradient> >& dPhi_dX = fe_u_volu->get_dphi();
        fe_u_volu->reinit(elem);
        // parametric coordinates of the quadrature points
        const std::vector<Point>& q_pnt = quad_volu.get_points();
        // get access to the element quadrature points data
        libmesh_assert( this->_dmn.elem_qi_local.end() != this->_dmn.elem_qi_local.find(elem->id()) );
        QuadratureInfo* qpi_ptr = &(this->_dmn.elem_qi_local.find(elem->id())->second);
        // loop over volume integration quadrature points
        for (unsigned int qp=0; qp<quad_volu.n_points(); qp++) {
            // key to access item from map
            const std::string key = number2string(elem->id()) + "_" + number2string(qp);
            // get access to the equation-systems corresponding to the RVE of this quadrature point
            libmesh_assert( es_rve.end() != es_rve.find(number2string(elem->id())) );
            EquationSystems& es_m = es_rve.find(key)->second;
            //
            FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
            fld.current_time = current_time;
            for (unsigned int ivar=0; ivar<SP3D; ivar++) {
                for (unsigned int l=0; l<var_dofs[ivar]; l++) {
                    const Real ui = u_sys.current_solution(dof_indices_var[ivar][l]);
                    fld.displacement(ivar) += Phi[l][qp] * ui;
                    fld.grad_displacement(ivar) += dPhi_dX[l][qp] * ui;
                }
            }
            // previous-step inelastic deformation gradient tensor
            fld.deform_grad_inel = qpi_ptr->deform_grad_inel[qp];
            // deformation gradient tensor
            fld.deform_gradient = deformation_gradient_3d(fld.grad_displacement);
            for (unsigned int ivar=0; ivar<nvars_RDS+ndcvars_RDS; ivar++) {
                fld.biochem[ivar] = qpi_ptr->biochem[qp][ivar];
            }
            for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                fld.grad_biochem[ivar] = qpi_ptr->grad_biochem[qp][ivar];
                fld.biochem_rate[ivar] = qpi_ptr->biochem_rate[qp][ivar];
            }
            for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                fld.auxiliary[ivar] = qpi_ptr->auxiliary[qp][ivar];
                fld.grad_auxiliary[ivar] = qpi_ptr->grad_auxiliary[qp][ivar];
                fld.auxiliary_rate[ivar] = qpi_ptr->auxiliary_rate[qp][ivar];
            }
            for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                fld.fibre_ref[ispdm] = qpi_ptr->fibre_ref[ispdm][qp];
                fld.fibre_curr[ispdm] = unit(fld.deform_gradient * fld.fibre_ref[ispdm]);
            }
            fld.strain = GreenLagrange_strain(fld.deform_gradient);
            fld.stress = qpi_ptr->stress[qp];
            fld.pressure = qpi_ptr->pressure[qp];
            //
            fld.biomech[0] = ri_ptr->damage_factor_const;
            ri_ptr->calc_damage(fld);
            fld.biomech[1] = ri_ptr->mat->get_biomech_param(0) * (1.0-fld.biomech[0]);
            fld.biomech[2] = qpi_ptr->pressure[qp];
            //
            std::vector<FieldInfo> fld_m(8, FieldInfo(0, 0, SP3D));
            calc_fields_RVE_(fe_u_volu_m, fld_m);
            //
            ri_ptr->calc_plastic_deform(fld);
            // second Piola-Kirchhoff stress tensor
            Dyadic3D S2pk;
            Quindic3D Dtsm_uu;
            Real Ph;
            Dyadic3D Dtsm_up, Dtsm_pu;
            Real Dtsm_pp;
            Vector3D Qs;
            // passive contribution of tissue biomechanics
            ri_ptr->calc_biomech_3d(fld, S2pk, Dtsm_uu, Dtsm_up, Ph, Dtsm_pu, Dtsm_pp);
            // wound contraction or tumour growth model
            if ( apply_wound_contraction || apply_tumour_growth )
                ri_ptr->calc_coupled_3d(fld, S2pk, Dtsm_uu, Dtsm_up, Ph, Dtsm_pu, Dtsm_pp);
            // multiscale model contribution
            ri_ptr->calc_multiscale(fld, es_m, fld_m, S2pk, Qs);
            //
            // update the quadrature point set data
            qpi_ptr->xyz[qp] = qpi_ptr->XYZ[qp] + fld.displacement;
            qpi_ptr->deform_grad[qp] = fld.deform_gradient;
            qpi_ptr->deform_grad_elas[qp] = fld.deform_grad_elas;
            qpi_ptr->deform_grad_inel[qp] = fld.deform_grad_inel;
            qpi_ptr->strain[qp] = fld.strain;
            qpi_ptr->stress[qp] = S2pk;
            qpi_ptr->pressure[qp] = frac_1o3*trace(qpi_ptr->stress[qp]);
            for (unsigned int i=0; i<SP3D; i++)
                qpi_ptr->fibre_curr[i][qp] = fld.fibre_curr[i];
            for (unsigned int i=0; i<ninvars_NCS; i++)
                qpi_ptr->biomech[qp][i] = fld.biomech[i];
            // end of volume integration quadrature loop...
        }
        // free memo
        delete [] dof_indices_var;
        // end of active local elements loop...
    }
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("_set_quad_info");
}
//=================================================================================================
void ExplicitNavierCauchySolver3D::_reset_step () {
    this->_log.push("_reset_step");

    EquationSystems& es = this->_dmn.get_equation_systems();
    System& d2u_sys = es.get_system<System>(this->_label+"::solution_2nd_time_derivative");
    System& du_sys  = es.get_system<System>(this->_label+"::solution_1st_time_derivative");
    System& u_sys   = es.get_system<System>(this->_label+"::solution");
    System& u_P_sys = es.get_system<System>(this->_label+"::previous_solution");
    const System& u_N_sys = es.get_system<System>(this->_label+"::next_solution");
    //
    const Real time_step = this->par_list.get<Real>("time step");
    //
    d2u_sys.solution->zero();
    d2u_sys.solution->add( 1.0/pow2(time_step), *u_N_sys.solution);
    d2u_sys.solution->add(-2.0/pow2(time_step), *u_sys.solution);
    d2u_sys.solution->add( 1.0/pow2(time_step), *u_P_sys.solution);
    d2u_sys.solution->close();
    d2u_sys.update();
    //
    du_sys.solution->zero();
    du_sys.solution->add( 0.5/time_step, *u_N_sys.solution);
    du_sys.solution->add(-0.5/time_step, *u_P_sys.solution);
    du_sys.solution->close();
    du_sys.update();
    //
    u_P_sys.solution->operator=(*u_sys.solution);
    u_P_sys.solution->close();
    u_P_sys.update();
    //
    u_sys.solution->operator=(*u_N_sys.solution);
    u_sys.solution->close();
    u_sys.update();
    //
    // now update the quadrature points information
    this->_set_quad_info();

    this->_log.pop("_reset_step");
}
//=================================================================================================
