/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/explicit_navier_cauchy_solver_3d.h"
#include "libmesh/linear_implicit_system.h"
#include "libmesh/petsc_linear_solver.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/gmsh_io.h"
#include "libmesh/fe.h"
#include "libmesh/quadrature_gauss.h"
#include "libmesh/dof_map.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/dense_matrix.h"
#include "libmesh/dense_vector.h"
#include "libmesh/dense_submatrix.h"
#include "libmesh/dense_subvector.h"
#include "libmesh/numeric_vector.h"
#include "feb3/iga-context.h"
#include "feb3/fem-context.h"
#include "feb3/paraview_io.h"
#include "feb3/ida.h"
#include "feb3/rve.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
extern // list of all systems used in the analysis
std::vector< std::pair<EquationSystems*, std::string> > systems_ls;
//=================================================================================================
// external routines
extern
void plot_results_rve (EquationSystems& );
//=================================================================================================
#include "feb3/domain_data_inl.h"
#include "feb3/solution_io_inl.h"
//=================================================================================================
ExplicitNavierCauchySolver3D::ExplicitNavierCauchySolver3D (DomainData& dmn) :
    _log("ExplicitNavierCauchySolver3D"), _dmn(dmn) {
    // set the default label
    this->_label = "ExplicitNavierCauchySolver3D";
    // set parameters to a default value
    this->par_list.set<Real>("current time") = 0.0;
    this->par_list.set<Real>("time step") = 1.0;
    this->par_list.set<Real>("mass-proportional damping coefficient") = 0.0;
    this->par_list.set<bool>("membrane elements present") = false;
    this->par_list.set<bool>("apply wound contraction") = false;
    this->par_list.set<bool>("apply tumour growth") = false;
    this->par_list.set<bool>("apply body force") = false;
    this->par_list.set<bool>("apply nodal force") = false;
    this->par_list.set<bool>("calculate gradients") = false;
    this->par_list.set<bool>("enforce to initialize solution") = true;
    this->par_list.set<bool>("system matrix is factorized") = false;
}
//=================================================================================================
void ExplicitNavierCauchySolver3D::init () {
    this->_log.push("init");

    EquationSystems& es = this->_dmn.get_equation_systems();
    Parameters& par = es.parameters;
    // set the system parameters
    this->_copy_params();
    // dimension of the FE grid
    const unsigned int nvars = es.get_mesh().mesh_dimension();
    libmesh_assert(nvars == SP3D);
    // Lagrange basis for displacement DoFs (default), also choose the order
    const Order lbasis = par.get<Order>("Lagrange basis order");
    //
    char var_name[buffer_len];
    int& hvar = par.set<int>("number of variables to hide in Paraview");
    // add the basic (explicit) system
    do {
        LinearImplicitSystem& sys = es.add_system<LinearImplicitSystem>(this->_label+"::next_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__u%d_NEXT", this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis);
        }
        sys.assemble_before_solve = false;
        sys.add_matrix("mass (doubled)");
        sys.add_matrix("mass (negative)");
        sys.add_matrix("damping");
        sys.add_vector("residual");
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    } while ( false );
    // add another system
    do {
        System& sys = es.add_system<System>(this->_label+"::solution_2nd_time_derivative");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__D2u%d_Dt2", this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis);
        }
    } while ( false );
    // add another system
    do {
        System& sys = es.add_system<System>(this->_label+"::solution_1st_time_derivative");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__Du%d_Dt", this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis);
        }
    } while ( false );
    // add another system
    do {
        System& sys = es.add_system<System>(this->_label+"::solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__u%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis);
        }
    } while ( false );
    // add another system
    do {
        System& sys = es.add_system<System>(this->_label+"::previous_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__u%d_PREV", this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis);
        }
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    } while ( false );
    // add another (auxiliary) system
    if ( this->par_list.get<bool>("calculate gradients") ) {
        System& sys = es.add_system<System>(this->_label+"::gradient_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            for (unsigned int jvar=0; jvar<nvars; jvar++) {
                sprintf(var_name, "%s__Fe%d%d", this->_label.c_str(), ivar, jvar);
                sys.add_variable(var_name, CONSTANT, MONOMIAL);
            }
        }
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            for (unsigned int jvar=0; jvar<nvars; jvar++) {
                sprintf(var_name, "%s__Fp%d%d", this->_label.c_str(), ivar, jvar);
                sys.add_variable(var_name, CONSTANT, MONOMIAL);
            }
        }
        for (unsigned int ivar=0; ivar<VOIGT_SP3D; ivar++) {
            const Dual<int> idx = indices_Voigt_3d(ivar);
            sprintf(var_name, "%s__E%d%d", this->_label.c_str(), idx.get_first(), idx.get_second());
            sys.add_variable(var_name, CONSTANT, MONOMIAL);
        }
        for (unsigned int ivar=0; ivar<VOIGT_SP3D; ivar++) {
            const Dual<int> idx = indices_Voigt_3d(ivar);
            sprintf(var_name, "%s__S%d%d", this->_label.c_str(), idx.get_first(), idx.get_second());
            sys.add_variable(var_name, CONSTANT, MONOMIAL);
        }
        for (unsigned int ivar=0; ivar<es.parameters.get<unsigned int>("Navier-Cauchy solver: internal variables"); ivar++) {
            sprintf(var_name, "%s__iv%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name, CONSTANT, MONOMIAL);
        }
        // end of if case...
    }

    this->_log.pop("init");
}
//=================================================================================================
void ExplicitNavierCauchySolver3D::run () {
    this->_log.push("run");

    EquationSystems& es = this->_dmn.get_equation_systems();
    //
    ParaviewIO pvtu(es.get_mesh());
    for (int l=0; l<es.parameters.get<int>("number of variables to hide in Paraview"); l++) {
        char var_name[buffer_len];
        sprintf(var_name, "Paraview variable #%d to hide", l);
        pvtu.hide_variable(es.parameters.get<std::string>(var_name));
    }
    //
    char cstr[buffer_len];
    const std::string results_dir  = es.parameters.get<std::string>("results directory"),
                      output_fname = es.parameters.get<std::string>("output file name"),
                      sim_title    = es.parameters.get<std::string>("simulation title");
    const int ini_incr = this->par_list.get<int>("initial increment");
    int& incr     = this->par_list.set<int>("increment step");
    int& max_incr = this->par_list.set<int>("maximum increments");
    int& soln_step = this->par_list.set<int>("solution increment");
    int& rstr_step = this->par_list.set<int>("restart increment");
    Real& current_time  = this->par_list.set<Real>("current time");
    Real& previous_time = this->par_list.set<Real>("previous time");
    Real& time_step = this->par_list.set<Real>("time step");
    int& increment = es.parameters.set<int>("increment step (solid biomechanics solver)");
    //
    incr = ini_incr;
    increment = ini_incr;
    //
    // set the initial conditions and initialise some member objects
    if ( this->par_list.get<bool>("enforce to initialize solution") ) {
        flog.print(" >> Enforce solution initialization\n", false);
        this->_set_initial();
    // ...else check if to read the restart file(s)
    } else if ( es.parameters.get<unsigned int>("input restart from file") ) {
        flog.print(" >> Load restart file(s)\n", false);
        this->load_restart();
    }
    //
    // write a PVD file to store results
    if ( libMesh::global_processor_id() == 0 ) {
        flog.print(" >> Generate Paraview PVD-file\n", false);
        sprintf( cstr, "%s/%s-%s.pvd",
                 results_dir.c_str(), output_fname.c_str(), sim_title.c_str() );
        std::ofstream fpvd(cstr);
        fpvd.setf(std::ios_base::scientific|std::ios_base::unitbuf);
        //
        fpvd << "<?xml version=\"1.0\"?>" << std::endl;
        fpvd << "<VTKFile type=\"Collection\" version=\"0.1\" byte_order=\"LittleEndian\">" << std::endl;
        fpvd << "<Collection>" << std::endl;
        for (int incr=ini_incr; incr<=max_incr; incr++) {
            const Real vtk_time = (incr+0.0)/max_incr;
            if ( incr == ini_incr || incr == max_incr || incr%soln_step == 0 ) {
                sprintf( cstr, "paraview/%s-%s-%d.vtu",
                         output_fname.c_str(), sim_title.c_str(), incr );
                fpvd << "<DataSet timestep=\"" << vtk_time << "\" group=\"\" part=\"0\" file=\"" << cstr << "\"/>" << std::endl;
            }
            // end of time steps loop...
        }
        fpvd << "</Collection>" << std::endl;
        fpvd << "</VTKFile>" << std::endl;
    }
    //
    // write a PVD file to store results (multiscale model)
    if ( ! es.parameters.get<std::map<std::string, MeshRVE>*>("pointer to RVE parametric mesh(es)")->empty() &&
         ! es.parameters.get<std::map<std::string, EquationSystems>*>("pointer to RVE equation system(s)")->empty() )
    if ( libMesh::global_processor_id() == 0 ) {
        flog.print(" >> Generate Paraview PVD-file\r", false);
        sprintf( cstr, "%s/%s-MS.pvd",
                 results_dir.c_str(), output_fname.c_str() );
        std::ofstream fpvd(cstr);
        fpvd.setf(std::ios_base::scientific|std::ios_base::unitbuf);
        //
        fpvd << "<?xml version=\"1.0\"?>" << std::endl;
        fpvd << "<VTKFile type=\"Collection\" version=\"0.1\" byte_order=\"LittleEndian\">" << std::endl;
        fpvd << "<Collection>" << std::endl;
        for (int incr=ini_incr; incr<=max_incr; incr++) {
            const Real vtk_time = (incr+0.0)/max_incr;
            if ( incr == ini_incr || incr == max_incr || incr%soln_step == 0 ) {
                sprintf( cstr, "paraview/MS/%s-%d.pvtu",
                         output_fname.c_str(), incr );
                fpvd << "<DataSet timestep=\"" << vtk_time << "\" group=\"\" part=\"0\" file=\"" << cstr << "\"/>" << std::endl;
            }
            // end of time steps loop...
        }
        fpvd << "</Collection>" << std::endl;
        fpvd << "</VTKFile>" << std::endl;
    }
    //
    // write initial results
    flog.print("  > Save solution\n", false);
    sprintf( cstr, "%s/paraview/%s-%s-%d.vtu",
             results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), ini_incr );
    pvtu.write_equation_systems(cstr, es);
    //
    flog.print("  > Save (micro-)solution\n", false);
    plot_results_rve(es);
    //
    // assemble the system matrix for the first time
    flog.print("  > Assemble system matrix\n", false);
    this->_assemble_mass();
    //
    // run for all time steps
    do {
        REDO: ;
        // update the increment step and the current/previous time
        incr += 1;
        current_time  += time_step;
        previous_time += time_step;
        //
        increment = incr / static_cast<int>(pow(2.0, this->par_list.get<int>("maximum increments (current scale)")));
        sprintf( cstr, " >> Time step %d out of %d (%.10e)",
                 incr, max_incr, current_time );
        flog.print(cstr, false);
        //
        // invoke the central difference solver
        bool success;
        this->_solve(success);
        sprintf( cstr, " *** residual %.10e                       \r",
                 this->par_list.set<Real>("residual L2-norm") );
        flog.print(cstr, false);
        if ( ! success ) {
            if ( this->par_list.get<bool>("abort non-convergence") ) {
                goto END_LABEL;
            } else if ( this->par_list.get<int>("maximum increments (current scale)") ==
                        this->par_list.get<int>("maximum increments (ceiling scale)") ) {
                goto END_LABEL;
            } else {
                this->par_list.set<int>("maximum increments (current scale)") += 1;
                incr -= 1;
                current_time  -= time_step;
                previous_time -= time_step;
                time_step /= 2.0; // half the time step...
                incr *= 2.0;
                max_incr *= 2.0;
                soln_step *= 2.0;
                rstr_step *= 2.0;
                this->par_list.set<Real>("residual tolerance") *= 4.0; // change the tolerance...
                //
                // re-assemble the system matrix
                flog.print("  > Re-assemble system matrix\r", false);
                this->_assemble_mass();
                // ...now go back and redo this (time) increment
                goto REDO;
            }
        }
        //
        // check if to write results
        if ( incr%soln_step == 0 || incr == max_incr ) {
            clear_line_(std::cout);
            // calculate the gradient solution (if necessary)
            if ( this->par_list.get<bool>("calculate gradients") ) {
                flog.print("  > Calculate gradients\r", false);
                this->_calc_gradients();
            }
            //
            flog.print("  > Save solution\r", false);
            sprintf( cstr, "%s/paraview/%s-%s-%d.vtu",
                     results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), increment );
            pvtu.write_equation_systems(cstr, es);
            //
            flog.print("  > Save (micro-)solution\r", false);
            plot_results_rve(es);
        }
        //
        // check if to write restart file
        if ( incr%rstr_step == 0 || incr == max_incr ) {
            flog.print("  > Save restart\r", false);
            sprintf( cstr, "%s/restart/%s-%s-%d.res-%d",
                     results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), increment, libMesh::global_processor_id() );
            std::ofstream fout(cstr);
            this->save_restart(fout);
        }
        // end of time steps loop...
    } while ( incr < max_incr );
    // end of simulation...
    END_LABEL: ;
    clear_line_(std::cout);
    flog.print(" >> Simulation terminated\n", false);

    this->_log.pop("run");
}
//=================================================================================================
//
//
//
//=================================================================================================
void ExplicitNavierCauchySolver3D::_copy_params () {
    this->_log.push("_copy_params");

    EquationSystems& es = this->_dmn.get_equation_systems();
    // set the label for the solver and the simulation title (if necessary)
    this->_label = es.parameters.get<std::string>("Navier-Cauchy solver: label");
    if ( "_" == es.parameters.get<std::string>("simulation title") )
        es.parameters.set<std::string>("simulation title") = this->_label;
    //
    if ( es.parameters.get<unsigned int>("input restart from file") )
        this->par_list.set<bool>("enforce to initialize solution") = false;
    //
    es.parameters.set<int>("increment step (solid biomechanics solver)") = 0;
    //
    this->par_list.set<int>("initial increment")  = es.parameters.get<int>("Navier-Cauchy solver: initial increment");
    this->par_list.set<int>("maximum increments") = es.parameters.get<int>("Navier-Cauchy solver: maximum increments");
    this->par_list.set<int>("solution increment") = es.parameters.get<int>("Navier-Cauchy solver: solution increment");
    this->par_list.set<int>("restart increment")  = es.parameters.get<int>("Navier-Cauchy solver: restart increment");
    this->par_list.set<Real>("current time")  = (this->par_list.get<int>("initial increment")+0.0)
                                              * es.parameters.get<Real>("Navier-Cauchy solver: window")
                                              / this->par_list.get<int>("maximum increments");
    this->par_list.set<Real>("previous time") = (this->par_list.get<int>("initial increment")-1.0)
                                              * es.parameters.get<Real>("Navier-Cauchy solver: window")
                                              / this->par_list.get<int>("maximum increments");
    this->par_list.set<Real>("time step") = es.parameters.get<Real>("Navier-Cauchy solver: window")
                                          / this->par_list.get<int>("maximum increments");
    this->par_list.set<int>("maximum increments (current scale)") = 0;
    this->par_list.set<int>("maximum increments (ceiling scale)") = es.parameters.get<int>("Navier-Cauchy solver: maximum increments (ceiling scale)");
    this->par_list.set<bool>("abort non-convergence") = es.parameters.get<bool>("Navier-Cauchy solver: abort non-convergence");
    //
    this->par_list.set<Real>("residual tolerance") = es.parameters.get<Real>("Navier-Cauchy solver: residual tolerance");
    this->par_list.set<bool>("calculate gradients") = es.parameters.get<bool>("Navier-Cauchy solver: calculate gradients");
    this->par_list.set<bool>("membrane elements present") = es.parameters.get<bool>("membrane elements present");
    this->par_list.set<bool>("apply wound contraction") = es.parameters.get<bool>("Navier-Cauchy solver: apply wound contraction");
    this->par_list.set<bool>("apply tumour growth") = es.parameters.get<bool>("Navier-Cauchy solver: apply tumour growth");
    this->par_list.set<bool>("apply body force") = es.parameters.get<bool>("Navier-Cauchy solver: apply body force");
    this->par_list.set<bool>("apply nodal force") = es.parameters.get<bool>("Navier-Cauchy solver: apply nodal force");
    //
    this->par_list.set<Real>("mass-proportional damping coefficient") =
        es.parameters.get<Real>("Navier-Cauchy solver: mass-proportional damping coefficient");

    this->_log.pop("_copy_params");
}
//=================================================================================================
void ExplicitNavierCauchySolver3D::_set_initial () {
    this->_log.push("_set_initial");

    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    // update quadrature point set data
    flog.print("  > Update quadrature data\n", false);
    this->_set_quad_info();
    //
    // calculate the gradient solution (if necessary)
    if ( this->par_list.get<bool>("calculate gradients") ) {
        flog.print("  > Calculate gradient solution\n", false);
        this->_calc_gradients();
    }

    this->_log.pop("_set_initial");
}
//=================================================================================================
void ExplicitNavierCauchySolver3D::_solve (bool& success) {
    this->_log.push("_solve");

    EquationSystems& es = this->_dmn.get_equation_systems();
    LinearImplicitSystem& u_N_sys = es.get_system<LinearImplicitSystem>(this->_label+"::next_solution");
    NumericVector<Number>* rhs = u_N_sys.rhs;
    const System& u_sys   = es.get_system<System>(this->_label+"::solution");
    const System& u_P_sys = es.get_system<System>(this->_label+"::previous_solution");
    //
    // assemble the RHS vector and the residual vector first
    this->_assemble_rhs();
    //
    // continue updating the RHS vector
    u_N_sys.get_matrix("mass (doubled)").vector_mult_add(*rhs, *u_sys.solution);
    u_N_sys.get_matrix("mass (negative)").vector_mult_add(*rhs, *u_P_sys.solution);
    if ( this->par_list.get<Real>("mass-proportional damping coefficient") )
        u_N_sys.get_matrix("damping").vector_mult_add(*rhs, *u_P_sys.solution);
    rhs->close();
    //
    // solve the linear system
    const bool reuse_precond = this->par_list.get<bool>("system matrix is factorized")
                             ? true : false;
    u_N_sys.linear_solver->reuse_preconditioner(reuse_precond);
    this->par_list.set<bool>("system matrix is factorized") = true;
    u_N_sys.solve();
    //
    // now update the residual vector
    NumericVector<Number>& Y = u_N_sys.get_vector("residual");
    u_N_sys.get_matrix("mass (doubled)").vector_mult_add(Y, *u_sys.solution);
    u_N_sys.get_matrix("mass (negative)").vector_mult_add(Y, *u_P_sys.solution);
    u_N_sys.get_matrix("mass (negative)").vector_mult_add(Y, *u_N_sys.solution);
    if ( this->par_list.get<Real>("mass-proportional damping coefficient") )
        u_N_sys.get_matrix("damping").vector_mult_add(Y, *u_N_sys.solution);
    Y.close();
    // ...and calculate its L2-norm
    this->par_list.set<Real>("residual L2-norm") = Y.l2_norm();
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    // now check if the explicit solver performs well
    success = false;
    if ( this->par_list.get<Real>("residual L2-norm") <=
         this->par_list.get<Real>("residual tolerance") ) {
        // reset the solution containers
        this->_reset_step();
        // ...and set this flag accordingly
        success = true;
    }

    this->_log.pop("_solve");
}
//=================================================================================================
void ExplicitNavierCauchySolver3D::_calc_gradients () {
    this->_log.push("_calc_gradients");

    EquationSystems& es = this->_dmn.get_equation_systems();
    System& gr_sys = es.get_system<System>(this->_label+"::gradient_solution");
    //
    const MeshBase::const_element_iterator begin_el = es.get_mesh().active_local_elements_begin();
    const MeshBase::const_element_iterator end_el = es.get_mesh().active_local_elements_end();
    //
    FEType fe_gr_type = gr_sys.variable_type(0);
    //
    const unsigned int ninvars_NCS = es.parameters.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = es.parameters.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = es.parameters.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = es.parameters.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    const Real current_time = this->par_list.get<Real>("current time");
    //
    const unsigned int nvars = SP3D*SP3D + SP3D*SP3D
                             + VOIGT_SP3D + VOIGT_SP3D
                             + ninvars_NCS;
    // some sanity checks
    libmesh_assert( SP3D == es.get_mesh().mesh_dimension() );
    libmesh_assert( n_vars == gr_sys.n_vars() );
    //
    unsigned int deformation_gradient_elas_var[SP3D*SP3D];
    for (unsigned int i=0; i<SP3D; i++) {
        for (unsigned int j=0; j<SP3D; j++) {
            const unsigned int l = i*SP3D+j;
            char var_name[buffer_len];
            sprintf(var_name, "%s__Fe%d%d", this->_label.c_str(), i, j);
            deformation_gradient_elas_var[l] = gr_sys.variable_number(var_name);
        }
    }
    unsigned int deformation_gradient_inel_var[SP3D*SP3D];
    for (unsigned int i=0; i<SP3D; i++) {
        for (unsigned int j=0; j<SP3D; j++) {
            const unsigned int l = i*SP3D+j;
            char var_name[buffer_len];
            sprintf(var_name, "%s__Fp%d%d", this->_label.c_str(), i, j);
            deformation_gradient_inel_var[l] = gr_sys.variable_number(var_name);
        }
    }
    unsigned int strain_var[VOIGT_SP3D];
    for (unsigned int i=0; i<VOIGT_SP3D; i++) {
        const Dual<int> idx = indices_Voigt_3d(i);
        char var_name[buffer_len];
        sprintf(var_name, "%s__E%d%d", this->_label.c_str(), idx.get_first(), idx.get_second());
        strain_var[i] = gr_sys.variable_number(var_name);
    }
    unsigned int stress_var[VOIGT_SP3D];
    for (unsigned int i=0; i<VOIGT_SP3D; i++) {
        const Dual<int> idx = indices_Voigt_3d(i);
        char var_name[buffer_len];
        sprintf(var_name, "%s__S%d%d", this->_label.c_str(), idx.get_first(), idx.get_second());
        stress_var[i] = gr_sys.variable_number(var_name);
    }
    unsigned int internv_var[ninvars_NCS];
    for (unsigned int i=0; i<ninvars_NCS; i++) {
        char var_name[buffer_len];
        sprintf(var_name, "%s__iv%d", this->_label.c_str(), i);
        internv_var[i] = gr_sys.variable_number(var_name);
    }
    //
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        //
        QGauss quad_volu(SP3D, es.parameters.get<Order>("Gauss quadrature order"));
        UniquePtr<FEBase> fe_gr_volu = FEBase::build(SP3D, fe_gr_type);
        fe_gr_volu->attach_quadrature_rule(&quad_volu);
        //
        Dyadic3D deformation_gradient_elas,
                 deformation_gradient_inel;
        std::vector<Real> strain,
                          stress;
        std::vector<Real> internal_variables;
        //
        deformation_gradient_elas.comp = 0.0;
        deformation_gradient_inel.comp = 0.0;
        strain.assign(VOIGT_SP3D, 0.0);
        stress.assign(VOIGT_SP3D, 0.0);
        internal_variables.assign(ninvars_NCS, 0.0);
        //
        {
            // initialize and get access to the FE stuff
            const std::vector< Real >& JxW = fe_gr_volu->get_JxW();
            fe_gr_volu->reinit(elem);
            // get access to the element quadrature points data
            const QuadratureInfo* quad_inf = &this->_dmn.elem_qi_local.find(elem->id())->second;
            const unsigned int n_qp = quad_inf->XYZ.size();
            // loop over volume integration quadrature points
            for (unsigned int qp=0; qp<n_qp; qp++) {
                //
                FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
                fld.current_time = current_time;
                //
                deformation_gradient_elas.comp += JxW[qp] * quad_inf->deform_grad_elas[qp].comp;
                deformation_gradient_inel.comp += JxW[qp] * quad_inf->deform_grad_inel[qp].comp;
                for (unsigned int ii=0; ii<VOIGT_SP3D; ii++) {
                    const Dual<int> idx = indices_Voigt_3d(ii);
                    strain[ii] += JxW[qp] * quad_inf->strain[qp](idx.get_first(),idx.get_second());
                    stress[ii] += JxW[qp] * quad_inf->stress[qp](idx.get_first(),idx.get_second());
                }
                for (unsigned int ii=0; ii<ninvars_NCS; ii++) {
                    internal_variables[ii] += JxW[qp] * quad_inf->biomech[qp][ii];
                }
                // end of quadrature points loop...
            }
            //
            const Real eVol_recip = 1.0 / elem->volume();
            // get the average field values by dividing them with the element volume
            deformation_gradient_elas.comp *= eVol_recip;
            deformation_gradient_inel.comp *= eVol_recip;
            for (unsigned int ii=0; ii<VOIGT_SP3D; ii++) {
                strain[ii] *= eVol_recip;
                stress[ii] *= eVol_recip;
            }
            for (unsigned int ii=0; ii<ninvars_NCS; ii++) {
                internal_variables[ii] *= eVol_recip;
            }
            // end of scope...
        }
        // local degrees of freedom for each 'gradient' variable
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
        //
        const unsigned int dof_F = gr_sys.solution->first_local_index(),
                           dof_L = gr_sys.solution->last_local_index();
        // now set the values of the elastic deformation gradient for this element
        for (unsigned int i=0; i<SP3D; i++) {
            for (unsigned int j=0; j<SP3D; j++) {
                const unsigned int k = i*SP3D+j; // local index
                const unsigned int l = k+0; // variable index
                gr_sys.get_dof_map().dof_indices(elem, dof_indices_var[l], deformation_gradient_elas_var[k]);
                const unsigned int idof = dof_indices_var[l][0];
                if ( dof_F <= idof && idof < dof_L )
                    gr_sys.solution->set(idof, deformation_gradient_elas(i,j));
            }
        }
        // now set the values of the plastic (inelastic) deformation gradient for this element
        for (unsigned int i=0; i<SP3D; i++) {
            for (unsigned int j=0; j<SP3D; j++) {
                const unsigned int k = i*SP3D+j; // local index
                const unsigned int l = k+SP3D*SP3D; // variable index
                gr_sys.get_dof_map().dof_indices(elem, dof_indices_var[l], deformation_gradient_inel_var[k]);
                const unsigned int idof = dof_indices_var[l][0];
                if ( dof_F <= idof && idof < dof_L )
                    gr_sys.solution->set(idof, deformation_gradient_inel(i,j));
            }
        }
        // now set the values of the strain for this element
        for (unsigned int i=0; i<VOIGT_SP3D; i++) {
            const unsigned int k = i; // local index
            const unsigned int l = k+2*SP3D*SP3D; // variable index
            gr_sys.get_dof_map().dof_indices(elem, dof_indices_var[l], strain_var[k]);
            const unsigned int idof = dof_indices_var[l][0];
            if ( dof_F <= idof && idof < dof_L )
                gr_sys.solution->set(idof, strain[i]);
        }
        // now set the values of the stress for this element
        for (unsigned int i=0; i<VOIGT_SP3D; i++) {
            const unsigned int k = i; // local index
            const unsigned int l = k+2*SP3D*SP3D+VOIGT_SP3D; // variable index
            gr_sys.get_dof_map().dof_indices(elem, dof_indices_var[l], stress_var[k]);
            const unsigned int idof = dof_indices_var[l][0];
            if ( dof_F <= idof && idof < dof_L )
                gr_sys.solution->set(idof, stress[i]);
        }
        // now set the values of the internal variables for this element
        for (unsigned int i=0; i<ninvars_NCS; i++) {
            const unsigned int k = i; // local index
            const unsigned int l = k+2*SP3D*SP3D+2*VOIGT_SP3D; // variable index
            gr_sys.get_dof_map().dof_indices(elem, dof_indices_var[l], internv_var[k]);
            const unsigned int idof = dof_indices_var[l][0];
            if ( dof_F <= idof && idof < dof_L )
                gr_sys.solution->set(idof, internal_variables[i]);
        }
        // free memo
        delete [] dof_indices_var;
        // end of active local elements loop...
    }
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // close the solution container and update the system
    gr_sys.solution->close();
    gr_sys.update();

    this->_log.pop("_calc_gradients");
}
//=================================================================================================
void ExplicitNavierCauchySolver3D::save_restart (std::ofstream& fout) const {
    this->_log.push("save_restart");

    for (unsigned isys=0; isys<systems_ls.size(); isys++) {
        const EquationSystems* es_ptr = systems_ls[isys].first;
        const System& sys = es_ptr->get_system<System>(systems_ls[isys].second);
        //
        sys.solution->close();
        std::vector<Number> soln;
        sys.solution->localize_to_one(soln);
        //
        if ( libMesh::global_processor_id() ) continue;
        //
        const unsigned int n = soln.size();
        fout.write((char*)&n, sizeof(unsigned int));
        for (unsigned int i=0; i<n; i++) {
            const Number component = soln[i];
            fout.write((char*)&component, sizeof(Number));
        }
    }
    //
    this->_dmn.save_quadrature(fout, QuadratureInfo::ALL, QuadratureInfo::LEVEL_3);
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("save_restart");
}
//=================================================================================================
void ExplicitNavierCauchySolver3D::load_restart () {
    this->_log.push("load_restart");

    const Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    for (unsigned int ifile=0; ifile<par.get<unsigned int>("input restart from file"); ifile++) {
        const std::string str_name = "input restart file name #" + std::to_string(ifile);
        const std::string restart_file_name = par.get<std::string>(str_name.c_str());
        //
        std::ifstream fin(restart_file_name.c_str());
        // read the system solution from the first restart file only
        if ( 0 == ifile )
            for (unsigned isys=0; isys<systems_ls.size(); isys++) {
                EquationSystems* es_ptr = systems_ls[isys].first;
                System& sys = es_ptr->get_system<System>(systems_ls[isys].second);
                //
                unsigned int n = 0;
                fin.read((char*)&n, sizeof(unsigned int));
                for (unsigned int i=0; i<n; i++) {
                    Number component;
                    fin.read((char*)&component, sizeof(Number));
                    sys.solution->set(i, component);
                }
            }
        // read Quadrature data for this processor only
        if ( libMesh::global_processor_id() == ifile )
            this->_dmn.load_quadrature(fin, QuadratureInfo::ALL, QuadratureInfo::LEVEL_3);
        // end of restart files loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    for (unsigned isys=0; isys<systems_ls.size(); isys++) {
        EquationSystems* es_ptr = systems_ls[isys].first;
        System& sys = es_ptr->get_system<System>(systems_ls[isys].second);
        //
        sys.solution->close();
        sys.update();
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("load_restart");
}
//=================================================================================================
