/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/auxiliary.h"
#include "feb3/iga-context.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
//=================================================================================================
void IGA_Context::get_var_list (std::vector<unsigned short int>& vars) const {
    const unsigned int nvars = this->_system->n_vars();
    //
    vars.resize(nvars);
    for (unsigned int ivar=0; ivar<nvars; ivar++) {
        const std::string& name = this->_system->variable_name(ivar);
        unsigned short int vrbl = this->_system->variable_number(name);
        vars[ivar] = vrbl;
    }
}
//=================================================================================================
void IGA_Context::get_dof_indices ( const NURBS* nurbs,
                                    std::vector<dof_id_type>& dofs,
                                    std::vector<dof_id_type>* dofs_var,
                                    std::vector<unsigned int>* vn ) const {
    libmesh_assert( elem );
    libmesh_assert( dofs_var );

    if ( vn ) {
        const unsigned int nvars = vn->size();
        const unsigned int ncpnt = nurbs->size();
        //
        dofs.clear();
        for (unsigned int jvar=0; jvar<nvars; jvar++) {
            dofs_var[jvar].clear();
            //
            const unsigned int ivar = vn->at(jvar);
            for (unsigned int icpnt=0; icpnt<ncpnt; icpnt++) {
                const Node* node = nurbs->get_ctrl_node_indexed(icpnt);
                libmesh_assert( node->n_comp(this->_system->number(), ivar) == 1 );
                const unsigned int idof = node->dof_number(this->_system->number(), ivar, 0);
                dofs_var[jvar].push_back( idof );
            }
            for (unsigned int i=0; i<dofs_var[jvar].size(); i++) {
                dofs.push_back( dofs_var[jvar][i] );
            }
        }
    } else {
        const unsigned int nvars = this->_system->n_vars();
        const unsigned int ncpnt = nurbs->size();
        //
        dofs.clear();
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            dofs_var[ivar].clear();
            //
            for (unsigned int icpnt=0; icpnt<ncpnt; icpnt++) {
                const Node* node = nurbs->get_ctrl_node_indexed(icpnt);
                libmesh_assert( node->n_comp(this->_system->number(), ivar) == 1 );
                const unsigned int idof = node->dof_number(this->_system->number(), ivar, 0);
                dofs_var[ivar].push_back( idof );
            }
            for (unsigned int i=0; i<dofs_var[ivar].size(); i++) {
                dofs.push_back( dofs_var[ivar][i] );
            }
        }
    }
}
//=================================================================================================
//
//
//
//=================================================================================================
void send_list_func (std::vector<dof_id_type>& send_list, void* context) {
    IGA_Context* iga_c = static_cast<IGA_Context*>(context);
    // the system
    const System& system = iga_c->get_system();
    const unsigned int nvars = system.n_vars();
    //const DofMap& dof_map = system.get_dof_map();
    // control mesh
    const MeshBase& msh = system.get_mesh();
    // parametric mesh
    const MeshIGA& msh_iga = iga_c->get_mesh_IGA();
    // free memory
    send_list.clear();

    // local nodes ID list
    std::set<dof_id_type> node_send_list;
    // iterate for all local NUBRS
    for ( std::vector<NURBS>::const_iterator
          np_ci=msh_iga.nurbs_begin(); np_ci!=msh_iga.nurbs_end(); np_ci++ ) {
        const NURBS* nurbs_p = &(*np_ci);
        // check if any control points are remote
        for (unsigned int inode=0; inode<nurbs_p->size(); inode++) {
            const dof_id_type node_id = nurbs_p->get_ctrl_node_indexed(inode)->id();
            //
            if ( msh.node_ptr(node_id)->processor_id() != libMesh::global_processor_id() )
                node_send_list.insert( node_id );
            // end of 'local NURBS' control points loop...
        }
        // check if any neighborhood NURBS contain local control points of this NURBS
        for (unsigned int s=0; s<nurbs_p->n_sides(); s++) {
            const NURBS* nurbs_p_n = nurbs_p->neighbor_ptr(s);
            // ignore following is this NURBS has no neighbors
            if ( NULL == nurbs_p_n ) continue;
            // ignore following is this NURBS is also local
            if ( nurbs_p_n->processor_id() == libMesh::global_processor_id() ) continue;
            //
            bool a_local_node_is_here = false;
            for (unsigned int jnode=0; jnode<nurbs_p_n->size(); jnode++) {
                // check for all local nodes if they are control points to this NURBS
                for (MeshBase::const_node_iterator ci=msh.local_nodes_begin(); ci!=msh.local_nodes_end(); ci++) {
                    if ( (*ci)->id() == nurbs_p_n->get_ctrl_node_indexed(jnode)->id() ) {
                        a_local_node_is_here = true;
                        goto JUMP01;
                    }
                }
                // end of 'remote neighbor NURBS' control points loop...
            }
            //
            JUMP01: ;
            //
            if ( a_local_node_is_here )
            // iterate for 'remote neighbor NURBS' control points
            for (unsigned int jnode=0; jnode<nurbs_p_n->size(); jnode++) {
                const dof_id_type node_id = nurbs_p_n->get_ctrl_node_indexed(jnode)->id();
                //
                if ( msh.node_ptr(node_id)->processor_id() != libMesh::global_processor_id() )
                    node_send_list.insert( node_id );
                // end of 'remote neighbor NURBS' control points loop...
            }
            // end of 'local NURBS' sides loop...
        }
        // end of local NURBS loop...
    }

    for (unsigned int ivar=0; ivar<nvars; ivar++) {
        for ( std::set<dof_id_type>::const_iterator
              sl_ci=node_send_list.begin(); sl_ci!=node_send_list.end(); sl_ci++ ) {
            const Node* node = msh.node_ptr(*sl_ci);
            //
            libmesh_assert( node->n_comp(system.number(), ivar) == 1 );
            send_list.push_back( node->dof_number(system.number(), ivar, 0) );
            // end of nodes to send loop...
        }
        // end of variables loop...
    }
}
//=================================================================================================
void sparsity_func ( SparsityPattern::Graph& sp,
                     std::vector<dof_id_type>& n_nz, std::vector<dof_id_type>& n_oz, void* context )
{
    IGA_Context* iga_c = static_cast<IGA_Context*>(context);
    // the system and its associated DoFs
    const System& system = iga_c->get_system();
    const unsigned int nvars = system.n_vars();
    //const DofMap& dof_map = system.get_dof_map();
    // control-net mesh
    const MeshBase& msh = system.get_mesh();
    MeshBase::const_node_iterator ctrl_nodes_begin = msh.active_nodes_begin();
    MeshBase::const_node_iterator ctrl_nodes_end   = msh.active_nodes_end();
    // IGA mesh
    const MeshIGA& msh_iga = iga_c->get_mesh_IGA();
    const std::map<Node*, std::vector<NURBS*> >& nodes2nurbs = msh_iga.get_nodes_to_nurbs();

    unsigned int n_global_nodes = 0;
    unsigned int n_local_nodes = 0;
    for ( MeshBase::const_node_iterator
          cn_citer=ctrl_nodes_begin; cn_citer!=ctrl_nodes_end; cn_citer++ ) {
        const Node* node = *cn_citer;
        if ( node->processor_id() == libMesh::global_processor_id() ) ++n_local_nodes;
        ++n_global_nodes;
    }
    // free and allocate memory
    n_nz.clear();
    n_oz.clear();
    if ( n_local_nodes == 0 ) return;

    // iterate for all active nodes
    for ( MeshBase::const_node_iterator
          cn_citer=ctrl_nodes_begin; cn_citer!=ctrl_nodes_end; cn_citer++ ) {
        const Node* node = *cn_citer;

        if ( node->processor_id() != libMesh::global_processor_id() ) continue;

        const std::vector<NURBS*>& nurbs_of_this_node = nodes2nurbs.find(*cn_citer)->second;

        std::set<dof_id_type> local_nodes;
        std::set<dof_id_type> remote_nodes;
        // insert itself first
        local_nodes.insert( node->id() );
        //
        for ( std::vector<NURBS*>::const_iterator
              nid_citer=nurbs_of_this_node.begin(); nid_citer!=nurbs_of_this_node.end(); nid_citer++ ) {
            const NURBS* nurbs = *nid_citer;
            //
            for (int A=0; A<nurbs->size(); A++) {
                const Node* ctrl_node = nurbs->get_ctrl_node_indexed(A);

                // check if this control node is local or remote
                if ( ctrl_node->processor_id() == libMesh::global_processor_id() )
                    local_nodes.insert( ctrl_node->id() );
                else
                    remote_nodes.insert( ctrl_node->id() );
                // end of control points loop...
            }
        }

        const dof_id_type n_nz__val = std::max<int>(nvars*local_nodes.size(),  nvars*n_local_nodes);
        const dof_id_type n_oz__val = std::max<int>(nvars*remote_nodes.size(), nvars*n_global_nodes-n_nz__val);
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            n_nz.push_back(n_nz__val);
            n_oz.push_back(n_oz__val);
        }
        // end of control nodes loop...
    }
    // sanity checks
    if ( n_nz.size() != (n_local_nodes*nvars) ) libmesh_error();
    if ( n_oz.size() != (n_local_nodes*nvars) ) libmesh_error();
}
//=================================================================================================
