/*************************************************************************************************

    Copyright (C) 2009-2019    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/utils.h"
#include "libmesh/equation_systems.h"
#include "libmesh/getpot.h"
#include "feb3/domain_data.h"
#include "libmesh/mesh.h"
#include "libmesh/string_to_enum.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
std::string tmp_dir;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
//=================================================================================================
// local variables
static
char err_msg[buffer_len];
static
Parameters* par_ptr;
//=================================================================================================
// list of parameters inserted for the analysis
std::set<std::string> in_pars;
// list of constrained nodes (temporary container/prior mesh preparation)
std::list< my_triple<Point, BoundaryConditionInfo, my_pair<std::string, std::string> > > nodal_bcs__unsorted;
//=================================================================================================
#include "feb3/input_inl.h"
//=================================================================================================
// external functions
#ifdef BIODYNAMO_ENABLED
extern
void read_BDM_parameters ( GetPot& in, std::set<std::string>& in_p,
                           libMesh::Parameters& params );
#endif
//=================================================================================================
static
void input_preamble (MeshBase& msh, GetPot& inp) {
    // get the directory path where files will be stored
    par_ptr->set<std::string>("results directory") = get_par<std::string>(inp, "results_directory");
    ASSERT_("." != par_ptr->get<std::string>("results directory"), "wrong results directory inserted");
    //
    par_ptr->set<std::string>("results directory: paraview") = par_ptr->get<std::string>("results directory") + "/paraview";
    par_ptr->set<std::string>("results directory: quadrature data") = par_ptr->get<std::string>("results directory") + "/quadrature_data";
    par_ptr->set<std::string>("results directory: nodal data") = par_ptr->get<std::string>("results directory") + "/nodal_data";
    par_ptr->set<std::string>("results directory: temporary") = par_ptr->get<std::string>("results directory") + "/temporary";
    par_ptr->set<std::string>("results directory: restart") = par_ptr->get<std::string>("results directory") + "/restart";
    // set-up the temporary directory
    tmp_dir = par_ptr->get<std::string>("results directory: temporary");
    //
    if ( libMesh::global_processor_id() == 0 ) {
    // check if to delete old (existing) results directory or to simply continue as is
    if ( get_par<bool>(inp, "delete_old_results_directory", false) ) {
        // delete the old directory
        const std::string Epath0 = "rm -fr " + par_ptr->get<std::string>("results directory");
        ASSERT_(0==std::system(Epath0.c_str()), "could not execute system command");
        // now create all necessary folders
        const std::string Rpath[] = { "mkdir " + par_ptr->get<std::string>("results directory"),
                                      "mkdir " + par_ptr->get<std::string>("results directory: paraview"),
                                      "mkdir " + par_ptr->get<std::string>("results directory: paraview") + "/MS",
                                      "mkdir " + par_ptr->get<std::string>("results directory: paraview") + "/BDM",
                                      "mkdir " + par_ptr->get<std::string>("results directory: quadrature data"),
                                      "mkdir " + par_ptr->get<std::string>("results directory: nodal data"),
                                      "mkdir " + par_ptr->get<std::string>("results directory: temporary"),
                                      "mkdir " + par_ptr->get<std::string>("results directory: restart") };
        for (int i=0; i<7; i++) {
            ASSERT_(0==std::system(Rpath[i].c_str()), "could not execute system command");
        }
    } else {
        // now create all necessary folders
        const std::string Rpath[] = { "mkdir " + par_ptr->get<std::string>("results directory"),
                                      "mkdir " + par_ptr->get<std::string>("results directory: paraview"),
                                      "mkdir " + par_ptr->get<std::string>("results directory: paraview") + "/MS",
                                      "mkdir " + par_ptr->get<std::string>("results directory: paraview") + "/BDM",
                                      "mkdir " + par_ptr->get<std::string>("results directory: quadrature data"),
                                      "mkdir " + par_ptr->get<std::string>("results directory: nodal data"),
                                      "mkdir " + par_ptr->get<std::string>("results directory: temporary"),
                                      "mkdir " + par_ptr->get<std::string>("results directory: restart") };
        for (int i=0; i<7; i++) {
            if (0!=std::system(Rpath[i].c_str())) WARNING_("a commands wasn't executed successfully");
        }
    }
    // end if check...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    par_ptr->set<std::string>("output file name") =
        get_par<std::string>(inp, "output_file_name");
    const std::string fLog = par_ptr->get<std::string>("results directory") + "/"
                           + par_ptr->get<std::string>("output file name") + ".log";
    const std::string fErr = par_ptr->get<std::string>("results directory") + "/"
                           + par_ptr->get<std::string>("output file name") + ".err";
    flog.open(fLog.c_str(), true);
    ferr.open(fErr.c_str(), false);
    // check if to automatically delete the error file(s)
    par_ptr->set<bool>("automatically delete error file") =
        get_par<bool>(inp, "automatically_delete_error_file", true);
    // now copy the input file into the results directory
    if ( libMesh::global_processor_id() == 0 ) {
        const std::string Input = inp.get(0, "");
        const std::string Input_copy = par_ptr->get<std::string>("results directory") + "/"
                                     + par_ptr->get<std::string>("output file name") + ".dat";
        const std::string cmd = "cp " + Input + ' ' + Input_copy;
        ASSERT_(0==std::system(cmd.c_str()), "could not execute system command");
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    const std::string prbcase_str = get_par<std::string>(inp, "problem_case");
    // check for the case of problem to be solved
    if        ( prbcase_str == "no_solver"            ) {
        par_ptr->set<ProblemCase>("solver") = NO_SOLVER;
        par_ptr->set<bool>("FEM framework") = false;
        par_ptr->set<bool>("IGA framework") = false;
    } else if ( prbcase_str == "biomechanics_solver"  ) {
        par_ptr->set<ProblemCase>("solver") = FEM_BIOMECHANICS_SOLVER;
        par_ptr->set<bool>("FEM framework") = true;
        par_ptr->set<bool>("IGA framework") = false;
    } else if ( prbcase_str == "biomechanics_solver-IGA"  ) {
        par_ptr->set<ProblemCase>("solver") = IGA_BIOMECHANICS_SOLVER;
        par_ptr->set<bool>("FEM framework") = false;
        par_ptr->set<bool>("IGA framework") = true;
    } else if ( prbcase_str == "biochemics_solver"    ) {
        par_ptr->set<ProblemCase>("solver") = FEM_BIOCHEMICS_SOLVER;
        par_ptr->set<bool>("FEM framework") = true;
        par_ptr->set<bool>("IGA framework") = false;
    } else if ( prbcase_str == "wound_healing_solver" ) {
        par_ptr->set<ProblemCase>("solver") = FEM_WOUND_HEALING_SOLVER;
        par_ptr->set<bool>("FEM framework") = true;
        par_ptr->set<bool>("IGA framework") = false;
    } else if ( prbcase_str == "tumour_angiogenesis_solver" ) {
        par_ptr->set<ProblemCase>("solver") = FEM_TUMOUR_ANGIOGENESIS_SOLVER;
        par_ptr->set<bool>("FEM framework") = true;
        par_ptr->set<bool>("IGA framework") = false;
    } else if ( prbcase_str == "invitro_tumoroid_solver" ) {
        par_ptr->set<ProblemCase>("solver") = FEM_INVITRO_TUMOROID_SOLVER;
        par_ptr->set<bool>("FEM framework") = true;
        par_ptr->set<bool>("IGA framework") = false;
    } else if ( prbcase_str == "radiotherapy_tumour_fibrosis_solver" ) {
        par_ptr->set<ProblemCase>("solver") = FEM_RT_TUMOUR_FIBROSIS_SOLVER;
        par_ptr->set<bool>("FEM framework") = true;
        par_ptr->set<bool>("IGA framework") = false;
    } else ERROR_("exception caught for type of solver for analysis");
    //
    // obtain the file name of the finite element mesh for FEM or the control grid for IGA
    par_ptr->set<std::string>("input mesh file") =
        get_par<std::string>(inp, "input_Gmsh_file");
    //
    // check if we run pure FEM analysis
    if ( par_ptr->get<bool>("FEM framework") ) {
        //
        par_ptr->set<Order>("Lagrange basis order") =
            Utility::string_to_enum<Order>( get_par<std::string>(inp, "Lagrange_basis_order", "FIRST") );
        ASSERT_( par_ptr->get<Order>("Lagrange basis order") == FIRST ||
                 par_ptr->get<Order>("Lagrange basis order") == SECOND,
                 "wrong value for Lagrange basis order inserted");
        //
        par_ptr->set<Order>("Gauss quadrature order") =
            Utility::string_to_enum<Order>( get_par<std::string>(inp, "Gauss_quadrature_order", "SECOND") );
        if ( par_ptr->get<Order>("Lagrange basis order") == FIRST ) {
            ASSERT_( par_ptr->get<Order>("Gauss quadrature order") == FIRST  ||
                     par_ptr->get<Order>("Gauss quadrature order") == SECOND ||
                     par_ptr->get<Order>("Gauss quadrature order") == FOURTH,
                     "wrong value for Gauss quadrature order inserted");
        } else if ( par_ptr->get<Order>("Lagrange basis order") == SECOND ) {
            ASSERT_( par_ptr->get<Order>("Gauss quadrature order") == SECOND ||
                     par_ptr->get<Order>("Gauss quadrature order") == FOURTH,
                     "wrong value for Gauss quadrature order inserted");
        } else libmesh_error();
        //
        par_ptr->set<bool>("enforce input mesh order") =
            get_par<bool>(inp, "enforce_input_mesh_order", true);
        if ( par_ptr->get<bool>("enforce input mesh order") && par_ptr->get<Order>("Lagrange basis order") == SECOND ) {
            par_ptr->set<bool>("enforce input mesh 2nd order to full") =
                get_par<bool>(inp, "enforce_input_mesh_order/full_2nd_order", true);
        } else {
            par_ptr->set<bool>("enforce input mesh 2nd order to full") = true;
        }
        //
        par_ptr->set<bool>("skip input mesh renumbering") =
            get_par<bool>(inp, "skip_input_mesh_renumbering", false);
        //
    // check if we run pure IGA analysis
    } else if ( par_ptr->get<bool>("IGA framework") ) {
        par_ptr->set<unsigned int>("number of IGA parametric mesh(es)") =  get_par_size<std::string>(inp, "input_NURBS_info_file");
        //
        std::set<std::string> all_labels;
        //
        const unsigned int n_files = get_par_size<std::string>(inp, "input_NURBS_info_file");
        ASSERT_(n_files >= 1, "wrong number of NURBS information files have been inserted");
        ASSERT_(n_files == get_par_size<std::string>(inp, "IGA_parametric_mesh_label"), "wrong number of IGA parametric mesh labels inserted");
        for (unsigned int i=0; i<n_files; i++) {
            // get the filename for the NURBS info
            par_ptr->set<std::string>("input NURBS info file name #"+std::to_string(i)) =
                get_par_comp<std::string>(inp, "input_NURBS_info_file", i);
            // now copy the input file into the results directory
            if ( libMesh::global_processor_id() == 0 ) {
                const std::string Input = par_ptr->set<std::string>("input NURBS info file name #"+std::to_string(i));
                const std::string Input_copy = par_ptr->get<std::string>("results directory") + "/"
                                             + par_ptr->get<std::string>("output file name") + ".nurbs-" + number2string<int>(i);
                const std::string cmd = "cp " + Input + ' ' + Input_copy;
                ASSERT_(0==std::system(cmd.c_str()), "could not execute system command");
            }
            // get the filename for the NURBS info
            par_ptr->set<std::string>("IGA parametric mesh label #"+std::to_string(i)) =
                get_par_comp<std::string>(inp, "IGA_parametric_mesh_label", i);
            //
            all_labels.insert( par_ptr->get<std::string>("IGA parametric mesh label #"+std::to_string(i)) );
        }
        //
        par_ptr->set<std::string>("IGA parametric mesh to use for analysis") =
            get_par<std::string>(inp, "IGA_parametric_mesh_to_use_for_analysis/label");
        ASSERT_( all_labels.end() != all_labels.find(par_ptr->get<std::string>("IGA parametric mesh to use for analysis")),
                 "label for IGA parametric mesh to utilize not found" );
        //
        par_ptr->set<bool>("adaptive Gauss quadrature order") =
            get_par<bool>(inp, "Gauss_quadrature_order/adaptive", false);
        //
        par_ptr->set<Order>("Gauss quadrature order") =
            Utility::string_to_enum<Order>( get_par<std::string>(inp, "Gauss_quadrature_order") );
        //
        par_ptr->set<int>("sampling number for NURBS") =
            get_par<int>(inp, "sampling_number_for_NURBS", 10);
        //
        par_ptr->set<bool>("enforce input mesh order") = true;
        //
        par_ptr->set<bool>("skip input mesh renumbering") = true;
        //
        // excepetion is caught...
    } else libmesh_error();
    //
    //
    // set this flag to true (by default)
    par_ptr->set<bool>("delete unwanted elements/nodes") = true;
    //
    par_ptr->set<bool>("initialise RVEs") =
        get_par<bool>(inp, "initialise_RVEs", true);
    //
    //
    // check if to read Quadrature information from binary file(s)
    if        ( 0 == get_par_size<std::string>(inp, "input_Quadrature_file") ) {
        par_ptr->set<unsigned int>("input Quadrature from file") = 0;
    } else if ( 1 == get_par_size<std::string>(inp, "input_Quadrature_file") ) {
        par_ptr->set<unsigned int>("input Quadrature from file") = libMesh::global_n_processors();
        ASSERT_(get_par_comp<std::string>(inp, "input_Quadrature_file", 0).find("*")!=std::string::npos, "input Quadrature file not correctly given");
        //
        for (unsigned int i=0; i<libMesh::global_n_processors(); i++) {
            std::string filename = get_par_comp<std::string>(inp, "input_Quadrature_file", 0);
            find_n_replace(filename, "*", number2string(i).c_str());
            // get the filename for the Quadrature data
            par_ptr->set<std::string>("input Quadrature file name #"+std::to_string(i)) = filename;
        }
    } else ERROR_("cannot read Quadrature file properly");
    // check if to read restart information from binary file(s)
    if        ( 0 == get_par_size<std::string>(inp, "input_restart_file") ) {
        par_ptr->set<unsigned int>("input restart from file") = 0;
    } else if ( 1 == get_par_size<std::string>(inp, "input_restart_file") ) {
        par_ptr->set<unsigned int>("input restart from file") = libMesh::global_n_processors();
        ASSERT_(get_par_comp<std::string>(inp, "input_restart_file", 0).find("*")!=std::string::npos, "input restart file not correctly given");
        //
        for (unsigned int i=0; i<libMesh::global_n_processors(); i++) {
            std::string filename = get_par_comp<std::string>(inp, "input_restart_file", 0);
            find_n_replace(filename, "*", number2string(i).c_str());
            // get the filename for the restart data
            par_ptr->set<std::string>("input restart file name #"+std::to_string(i)) = filename;
        }
    } else ERROR_("cannot read restart file properly");
    //
    // ...and some validity checks
    if ( par_ptr->get<unsigned int>("input Quadrature from file") && par_ptr->get<unsigned int>("input restart from file") )
        ERROR_("cannot read Quadrature & restart file(s) at the same time for this analysis");
    //
    //
    //
    if ( par_ptr->get<ProblemCase>("solver") == FEM_TUMOUR_ANGIOGENESIS_SOLVER ) {
        // get the filename for the 1D finite element mesh of the vascular network
        par_ptr->set<std::string>("input vascular network file") =
            get_par<std::string>(inp, "input_vascular_network_file");
        // now copy the input file into the results directory
        if ( libMesh::global_processor_id() == 0 ) {
            const std::string Input = par_ptr->get<std::string>("input vascular network file");
            const std::string Input_copy = par_ptr->get<std::string>("results directory") + "/"
                                         + par_ptr->get<std::string>("output file name") + ".input_vascular_network.msh";
            const std::string cmd = "cp " + Input + ' ' + Input_copy;
            ASSERT_(0==std::system(cmd.c_str()), "could not execute system command");
        }
        // get the filename for the vascular network information
        par_ptr->set<std::string>("input vascular network information file") =
            get_par<std::string>(inp, "input_vascular_network_information_file");
        // now copy the input file into the results directory
        if ( libMesh::global_processor_id() == 0 ) {
            const std::string Input = par_ptr->get<std::string>("input vascular network information file");
            const std::string Input_copy = par_ptr->get<std::string>("results directory") + "/"
                                         + par_ptr->get<std::string>("output file name") + ".input_vascular_network.info";
            const std::string cmd = "cp " + Input + ' ' + Input_copy;
            ASSERT_(0==std::system(cmd.c_str()), "could not execute system command");
        }
        // get the filename for the vascular network information
        par_ptr->set<std::string>("input vascular network domain data file") =
            get_par<std::string>(inp, "input_vascular_network_domain_data_file", "NA");
        // end of solver-specific data...
    }
    //
    //
    par_ptr->set<Real>("penalty parameter scale") =
        get_par<Real>(inp, "penalty_parameter_scale", 1.0e+5);
    ASSERT_( par_ptr->get<Real>("penalty parameter scale") > 9.99999e+04,
             "wrong value (too low) for penalty parameter scale inserted");
    ASSERT_( par_ptr->get<Real>("penalty parameter scale") < 1.0000e+12,
             "wrong value (too high) for penalty parameter scale inserted");
    //
    par_ptr->set<bool>("print Paraview output (quadrature)") =
        get_par<bool>(inp, "print_Paraview_output/quadrature", false);
    //
    par_ptr->set<int>("number of variables to hide in Paraview") = 0;
    //
    par_ptr->set<Real>("current time") = 0.0;
    par_ptr->set<std::string>("simulation title") = get_par<std::string>(inp, "simulation_title", "_");
}
//=================================================================================================
static
bool input_BC_surfaces (DomainData& dmn, GetPot& inp) {
    bool have_follower_forces = false;
    // loop for all boundary condition surfaces
    for (int isurf=0; isurf<get_par<int>(inp, "number_of_BC_surfaces"); isurf++) {
        //
        const boundary_id_type bc_id =
            get_par<int>(inp, "BC_surface_"+std::to_string(isurf)+"/ID");
        //
        BoundaryConditionInfo bc_info;
        //
        const std::string bc_type_str =
            get_par<std::string>(inp, "BC_surface_"+std::to_string(isurf)+"/type");
        //
        if        ( "displacement_prescribed"   == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED;
        } else if ( "displacement_prescribed_x" == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_X;
        } else if ( "displacement_prescribed_y" == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Y;
        } else if ( "displacement_prescribed_z" == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Z;
        } else if ( "displacement_prescribed_radial"   == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_RD;
        } else if ( "displacement_prescribed_plane"    == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_PL;
        } else if ( "displacement_prescribed_cylinder" == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_CY;
        } else if ( "displacement_prescribed_planar_rotation" == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_PR;
        } else if ( "displacement_prescribed_xy" == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XY;
        } else if ( "displacement_prescribed_yz" == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_YZ;
        } else if ( "displacement_prescribed_xz" == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XZ;
        } else if ( "traction_normal" == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::NEUMANN_NORMAL;
        } else if ( "traction_vector" == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::NEUMANN_VECTOR;
        } else {
            sprintf(err_msg, "exception caught for type of BC on boundary surface %d", isurf);
            ERROR_(err_msg);
        }
        //
        if        ( BoundaryConditionInfo::NEUMANN_VECTOR == bc_info.type ) {
            WARNING_("the boundary value for the prescribed traction should be total value");
            // set the direction vector component X
            bc_info.value.comp(0) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_X");
            // set the direction vector component Y
            bc_info.value.comp(1) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_Y");
            // set the direction vector component Z
            bc_info.value.comp(2) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_Z");
        } else if ( BoundaryConditionInfo::NEUMANN_NORMAL == bc_info.type ) {
            WARNING_("the boundary value for the prescribed traction should be total value");
            // set the magnitude of the normal traction vector
            bc_info.value.comp(0) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/magnitude");
            // check if magnitude is not zero, thus expect for follower load to apply
            if ( ! have_follower_forces && fabs(bc_info.value.comp(0)) > 1.0e-6 ) {
                have_follower_forces = true;
            }
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED == bc_info.type ) {
            WARNING_("the boundary value for the prescribed displacement should be total value");
            // set the direction vector components X, Y, Z
            bc_info.value.comp(0) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_X");
            bc_info.value.comp(1) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_Y");
            bc_info.value.comp(2) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_Z");
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_X == bc_info.type ) {
            WARNING_("the boundary value for the prescribed displacement should be total value");
            // set the direction vector component X
            bc_info.value.comp(0) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_X");
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Y == bc_info.type ) {
            WARNING_("the boundary value for the prescribed displacement should be total value");
            // set the direction vector component Y
            bc_info.value.comp(1) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_Y");
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Z == bc_info.type ) {
            WARNING_("the boundary value for the prescribed displacement should be total value");
            // set the direction vector component Z
            bc_info.value.comp(2) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_Z");
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_RD == bc_info.type ) {
            WARNING_("the boundary value for the prescribed displacement should be total value");
            // set the direction vector component R
            bc_info.value.comp(0) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_R");
            // set the center point
            if ( get_par_size<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/center") == SP3D ) {
                bc_info.center(0) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/center", 0);
                bc_info.center(1) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/center", 1);
                bc_info.center(2) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/center", 2);
            } else {
                sprintf(err_msg, "center point components for BC surface %d not equal to 3", isurf);
                ERROR_(err_msg);
            }
            //
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_PR == bc_info.type ) {
            WARNING_("the boundary value for the prescribed displacement should be total value");
            // set the "centre" point of planar-rotation
            if ( get_par_size<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/center") == SP3D ) {
                bc_info.center(0) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/center", 0);
                bc_info.center(1) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/center", 1);
                bc_info.center(2) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/center", 2);
            } else {
                sprintf(err_msg, "planar-rotation point components for BC surface %d not equal to 3", isurf);
                ERROR_(err_msg);
            }
            // set the normal to planar-rotation
            if ( get_par_size<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/normal") == SP3D ) {
                bc_info.normal(0) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/normal", 0);
                bc_info.normal(1) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/normal", 1);
                bc_info.normal(2) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/normal", 2);
                const Real mag = bc_info.normal.norm();
                if ( fabs(mag-1.0) > 1.0e-4 ) {
                    sprintf(err_msg, "normal vector for BC surface %d is not unit (=%f)", isurf, mag);
                    WARNING_(err_msg);
                    //
                    bc_info.normal(0) /= mag;
                    bc_info.normal(1) /= mag;
                    bc_info.normal(2) /= mag;
                }
            } else {
                sprintf(err_msg, "normal to planar-rotation components for BC surface %d not equal to 3", isurf);
                ERROR_(err_msg);
            }
            // set the percentage of displacement in the direction of the normal vector
            bc_info.value.comp(0) =
                degrees_to_radians( get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/angle", 1.0) );
            if ( bc_info.value.comp(0) < 0.0 ) {
                ERROR_("angle inserted for planar-rotation essential BC is negative");
            }
            //
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_PL == bc_info.type ) {
            // check if to define the plane based on a 3-point definition or...
            if ( get_par<bool>(inp, "BC_surface_"+std::to_string(isurf)+"/3_point_plane_definition", false) ) {
                //
                if ( get_par_size<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point_0") == SP3D ) {
                    bc_info.center(0) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point_0", 0);
                    bc_info.center(1) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point_0", 1);
                    bc_info.center(2) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point_0", 2);
                } else {
                    sprintf(err_msg, "point-0 components for BC surface %d not equal to 3", isurf);
                    ERROR_(err_msg);
                }
                //
                Point pnt1;
                if ( get_par_size<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point_1") == SP3D ) {
                    pnt1(0) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point_1", 0);
                    pnt1(1) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point_1", 1);
                    pnt1(2) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point_1", 2);
                } else {
                    sprintf(err_msg, "point-1 components for BC surface %d not equal to 3", isurf);
                    ERROR_(err_msg);
                }
                //
                Point pnt2;
                if ( get_par_size<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point_2") == SP3D ) {
                    pnt2(0) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point_2", 0);
                    pnt2(1) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point_2", 1);
                    pnt2(2) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point_2", 2);
                } else {
                    sprintf(err_msg, "point-2 components for BC surface %d not equal to 3", isurf);
                    ERROR_(err_msg);
                }
                // calculate the normal to projection plane
                const Point v1 = pnt1 - bc_info.center,
                            v2 = pnt2 - bc_info.center,
                            n  = v1.cross(v2);
                bc_info.normal = n / n.norm();
                // ...to insert one point and a normal to the plane...
            } else {
                // set the projection-plane point
                if ( get_par_size<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point") == SP3D ) {
                    bc_info.center(0) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point", 0);
                    bc_info.center(1) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point", 1);
                    bc_info.center(2) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point", 2);
                } else {
                    sprintf(err_msg, "projection-plane point components for BC surface %d not equal to 3", isurf);
                    ERROR_(err_msg);
                }
                // set the normal to projection-plane
                if ( get_par_size<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/normal") == SP3D ) {
                    bc_info.normal(0) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/normal", 0);
                    bc_info.normal(1) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/normal", 1);
                    bc_info.normal(2) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/normal", 2);
                    const Real mag = bc_info.normal.norm();
                    if ( fabs(mag-1.0) > 1.0e-4 ) {
                        sprintf(err_msg, "normal vector for BC surface %d is not unit (=%f)", isurf, mag);
                        WARNING_(err_msg);
                        //
                        bc_info.normal(0) /= mag;
                        bc_info.normal(1) /= mag;
                        bc_info.normal(2) /= mag;
                    }
                } else {
                    sprintf(err_msg, "normal to projection-plane components for BC surface %d not equal to 3", isurf);
                    ERROR_(err_msg);
                }
                // end of if case...
            }
            // set the percentage of displacement in the direction of the normal vector
            bc_info.value.comp(0) =
                get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/magnitude", 1.0);
            if ( bc_info.value.comp(0) < 0.0 || bc_info.value.comp(0) > 1.0 ) {
                ERROR_("wrong displacement magnitude inserted for planar essential BCs");
            }
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_CY == bc_info.type ) {
            // set the projection-cylinder point
            if ( get_par_size<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point") == SP3D ) {
                bc_info.center(0) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point", 0);
                bc_info.center(1) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point", 1);
                bc_info.center(2) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/point", 2);
            } else {
                sprintf(err_msg, "projection-cylinder point components for BC surface %d not equal to 3", isurf);
                ERROR_(err_msg);
            }
            // set the normal to projection-cylinder
            if ( get_par_size<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/normal") == SP3D ) {
                bc_info.normal(0) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/normal", 0);
                bc_info.normal(1) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/normal", 1);
                bc_info.normal(2) = get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/normal", 2);
                const Real mag = bc_info.normal.norm();
                if ( fabs(mag-1.0) > 1.0e-4 ) {
                    sprintf(err_msg, "normal vector for BC surface %d is not unit (=%f)", isurf, mag);
                    WARNING_(err_msg);
                    //
                    bc_info.normal(0) /= mag;
                    bc_info.normal(1) /= mag;
                    bc_info.normal(2) /= mag;
                }
            } else {
                sprintf(err_msg, "normal to projection-cylinder components for BC surface %d not equal to 3", isurf);
                ERROR_(err_msg);
            }
            // set the percentage of displacement in the direction of the normal vector
            bc_info.value.comp(0) =
                get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/magnitude", 1.0);
            if ( bc_info.value.comp(0) < 0.0 || bc_info.value.comp(0) > 1.0 ) {
                ERROR_("wrong displacement magnitude inserted for planar essential BCs");
            }
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XY == bc_info.type ) {
            WARNING_("the boundary value for the prescribed displacement should be total value");
            // set the direction vector components X, Y
            bc_info.value.comp(0) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_X");
            bc_info.value.comp(1) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_Y");
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_YZ == bc_info.type ) {
            WARNING_("the boundary value for the prescribed displacement should be total value");
            // set the direction vector components Y, Z
            bc_info.value.comp(1) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_Y");
            bc_info.value.comp(2) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_Z");
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XZ == bc_info.type ) {
            WARNING_("the boundary value for the prescribed displacement should be total value");
            // set the direction vector components X, Z
            bc_info.value.comp(0) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_X");
            bc_info.value.comp(2) = get_par<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/direction_Z");
            // ...exception is caught here
        } else libmesh_error();
        //
        if ( true ) {
            //
            const std::string ft_str = get_par<std::string>(inp, "BC_surface_"+std::to_string(isurf)+"/time_function");
            const unsigned int n_params = get_par_size<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/time_parameters");
            //
            if        ( "flat_time_function"             == ft_str ) {
                bc_info.set_time_function(&flat_function);
                ASSERT_(n_params == 1, "wrong number of time parameters inserted");
            } else if ( "linear_time_function"           == ft_str ) {
                bc_info.set_time_function(&linear_function);
                ASSERT_(n_params == 1, "wrong number of time parameters inserted");
            } else if ( "linear_time_function__periodic" == ft_str ) {
                bc_info.set_time_function(&linear_function__periodic);
                ASSERT_(n_params == 2, "wrong number of time parameters inserted");
            } else if ( "ramp_time_function"             == ft_str ) {
                bc_info.set_time_function(&ramp_function);
                ASSERT_(n_params == 4, "wrong number of time parameters inserted");
            } else if ( "ramp_time_function__periodic"   == ft_str ) {
                bc_info.set_time_function(&ramp_function__periodic);
                ASSERT_(n_params == 5, "wrong number of time parameters inserted");
            } else if ( "steep_time_function"            == ft_str ) {
                bc_info.set_time_function(&steep_function);
                ASSERT_(n_params == 6, "wrong number of time parameters inserted");
            } else if ( "steep_time_function__periodic"  == ft_str ) {
                bc_info.set_time_function(&steep_function__periodic);
                ASSERT_(n_params == 7, "wrong number of time parameters inserted");
            } else if ( "exp_time_function"              == ft_str ) {
                bc_info.set_time_function(&exp_function);
                ASSERT_(n_params == 5, "wrong number of time parameters inserted");
            } else if ( "exp_time_function__periodic"    == ft_str ) {
                bc_info.set_time_function(&exp_function__periodic);
                ASSERT_(n_params == 6, "wrong number of time parameters inserted");
            } else {
                sprintf(err_msg, "exception is caught for time function of BC surface %d ", isurf);
                ERROR_(err_msg);
            }
            // set the time parameters
            bc_info.time_params.assign(n_params, 0.0);
            for (unsigned int i=0; i<n_params; i++) {
                bc_info.time_params[i] =
                    get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/time_parameters", i);
            }
            // end of if case...
        }
        //
        if ( "displacement_prescribed"   == bc_type_str               ||
             "displacement_prescribed_x" == bc_type_str               ||
             "displacement_prescribed_y" == bc_type_str               ||
             "displacement_prescribed_z" == bc_type_str               ||
             "displacement_prescribed_xy" == bc_type_str              ||
             "displacement_prescribed_yz" == bc_type_str              ||
             "displacement_prescribed_xz" == bc_type_str              ||
             "displacement_prescribed_radial"   == bc_type_str        ||
             "displacement_prescribed_plane"    == bc_type_str        ||
             "displacement_prescribed_cylinder" == bc_type_str        ||
             "displacement_prescribed_planar_rotation" == bc_type_str )
        {
            //
            const std::string ft_str = get_par<std::string>(inp, "BC_surface_"+std::to_string(isurf)+"/active_function");
            const unsigned int n_params = get_par_size<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/active_parameters");
            //
            if        ( "flat_time_function"             == ft_str ) {
                bc_info.set_active_function(&flat_function);
                ASSERT_(n_params == 1, "wrong number of active parameters inserted");
            } else if ( "linear_time_function"           == ft_str ) {
                bc_info.set_active_function(&linear_function);
                ASSERT_(n_params == 1, "wrong number of active parameters inserted");
            } else if ( "linear_time_function__periodic" == ft_str ) {
                bc_info.set_active_function(&linear_function__periodic);
                ASSERT_(n_params == 2, "wrong number of active parameters inserted");
            } else if ( "ramp_time_function"             == ft_str ) {
                bc_info.set_active_function(&ramp_function);
                ASSERT_(n_params == 4, "wrong number of active parameters inserted");
            } else if ( "ramp_time_function__periodic"   == ft_str ) {
                bc_info.set_active_function(&ramp_function__periodic);
                ASSERT_(n_params == 5, "wrong number of active parameters inserted");
            } else if ( "steep_time_function"            == ft_str ) {
                bc_info.set_active_function(&steep_function);
                ASSERT_(n_params == 6, "wrong number of active parameters inserted");
            } else if ( "steep_time_function__periodic"  == ft_str ) {
                bc_info.set_active_function(&steep_function__periodic);
                ASSERT_(n_params == 7, "wrong number of active parameters inserted");
            } else if ( "exp_time_function"              == ft_str ) {
                bc_info.set_active_function(&exp_function);
                ASSERT_(n_params == 5, "wrong number of active parameters inserted");
            } else if ( "exp_time_function__periodic"    == ft_str ) {
                bc_info.set_active_function(&exp_function__periodic);
                ASSERT_(n_params == 6, "wrong number of active parameters inserted");
            } else if ( "step_time_function"             == ft_str ) {
                bc_info.set_active_function(&step_function);
                ASSERT_(n_params == 3, "wrong number of active parameters inserted");
            } else {
                sprintf(err_msg, "exception is caught for active function of BC surface %d ", isurf);
                ERROR_(err_msg);
            }
            // set the active parameters
            bc_info.active_params.assign(n_params, 0.0);
            for (unsigned int i=0; i<n_params; i++) {
                bc_info.active_params[i] =
                    get_par_comp<Real>(inp, "BC_surface_"+std::to_string(isurf)+"/active_parameters", i);
            }
            //
        } else if ( "traction_normal" == bc_type_str ||
                    "traction_vector" == bc_type_str )
        {
            //
            bc_info.set_active_function(&flat_function);
            bc_info.active_params.assign(1, 1.0);
            //
            // ...exception is caught here
        } else {
            sprintf(err_msg, "exception caught for type of BC on boundary surface %d", isurf);
            ERROR_(err_msg);
        }
        //
        // insert data into corresponding container but make this check first
        ASSERT_( dmn.bcs.end() == dmn.bcs.find(bc_id),
                 "boundary surface has already been inserted");
        dmn.bcs.insert( std::make_pair(bc_id, bc_info) );
        // end of boundary condition surfaces loop...
    }
    //
    return have_follower_forces;
}
//=================================================================================================
static
void input_BC_nodes (DomainData& dmn, GetPot& inp) {
    // clear container first
    nodal_bcs__unsorted.clear();
    // set the geometric tolerance for finding the constrained boundary nodes
    if ( get_par<int>(inp, "number_of_BC_nodes") ) {
        par_ptr->set<Real>("constrained boundary nodes geometric tolerance") =
            get_par<Real>(inp, "BC_nodes/geometric_tolerance", 1.0e-3);
    }
    // loop for all nodal boundary condition
    for (int inode=0; inode<get_par<int>(inp, "number_of_BC_nodes"); inode++) {
        // set the point (coordinates)
        Point bc_pnt;
        if ( get_par<bool>(inp, "BC_node_"+std::to_string(inode)+"/explicit_coordinates", true) ) {
            if ( get_par_size<Real>(inp, "BC_node_"+std::to_string(inode)+"/point") == SP3D ) {
                bc_pnt(0) = get_par_comp<Real>(inp, "BC_node_"+std::to_string(inode)+"/point", 0);
                bc_pnt(1) = get_par_comp<Real>(inp, "BC_node_"+std::to_string(inode)+"/point", 1);
                bc_pnt(2) = get_par_comp<Real>(inp, "BC_node_"+std::to_string(inode)+"/point", 2);
            } else {
                sprintf(err_msg, "components for BC node %d not equal to 3", inode);
                ERROR_(err_msg);
            }
        } else {
            bc_pnt(0) = get_par<Real>(inp, "BC_node_"+std::to_string(inode)+"/point_X", BoundaryConditionInfo::generic_coord);
            bc_pnt(1) = get_par<Real>(inp, "BC_node_"+std::to_string(inode)+"/point_Y", BoundaryConditionInfo::generic_coord);
            bc_pnt(2) = get_par<Real>(inp, "BC_node_"+std::to_string(inode)+"/point_Z", BoundaryConditionInfo::generic_coord);
        }
        //
        ASSERT_( BoundaryConditionInfo::generic_coord != bc_pnt(0) ||
                 BoundaryConditionInfo::generic_coord != bc_pnt(1) ||
                 BoundaryConditionInfo::generic_coord != bc_pnt(2),
                 "error occured for boundary node, wrong coordinates");
        //
        BoundaryConditionInfo bc_info;
        //
        const std::string bc_type_str = get_par<std::string>(inp, "BC_node_"+std::to_string(inode)+"/type");
        //
        if        ( "displacement_prescribed"    == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED;
        } else if ( "displacement_prescribed_x"  == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_X;
        } else if ( "displacement_prescribed_y"  == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Y;
        } else if ( "displacement_prescribed_z"  == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Z;
        } else if ( "displacement_prescribed_xy" == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XY;
        } else if ( "displacement_prescribed_yz" == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_YZ;
        } else if ( "displacement_prescribed_xz" == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XZ;
        } else if ( "traction_vector"            == bc_type_str ) {
            bc_info.type = BoundaryConditionInfo::NEUMANN_VECTOR;
        } else {
            sprintf(err_msg, "exception caught for type of BC on boundary node %d", inode);
            ERROR_(err_msg);
        }
        //
        if ( BoundaryConditionInfo::NEUMANN_VECTOR == bc_info.type       ||
             BoundaryConditionInfo::DIRICHLET_PRESCRIBED == bc_info.type ) {
            WARNING_("the boundary value for the prescribed traction|displacement should be total value");
            // set the direction vector components X, Y, Z
            bc_info.value.comp(0) = get_par<Real>(inp, "BC_node_"+std::to_string(inode)+"/direction_X");
            bc_info.value.comp(1) = get_par<Real>(inp, "BC_node_"+std::to_string(inode)+"/direction_Y");
            bc_info.value.comp(2) = get_par<Real>(inp, "BC_node_"+std::to_string(inode)+"/direction_Z");
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_X == bc_info.type ) {
            WARNING_("the boundary value for the prescribed displacement should be total value");
            // set the direction vector component X
            bc_info.value.comp(0) = get_par<Real>(inp, "BC_node_"+std::to_string(inode)+"/direction_X");
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Y == bc_info.type ) {
            WARNING_("the boundary value for the prescribed displacement should be total value");
            // set the direction vector component Y
            bc_info.value.comp(1) = get_par<Real>(inp, "BC_node_"+std::to_string(inode)+"/direction_Y");
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Z == bc_info.type ) {
            WARNING_("the boundary value for the prescribed displacement should be total value");
            // set the direction vector component Z
            bc_info.value.comp(2) = get_par<Real>(inp, "BC_node_"+std::to_string(inode)+"/direction_Z");
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XY == bc_info.type ) {
            WARNING_("the boundary value for the prescribed displacement should be total value");
            // set the direction vector components X, Y
            bc_info.value.comp(0) = get_par<Real>(inp, "BC_node_"+std::to_string(inode)+"/direction_X");
            bc_info.value.comp(1) = get_par<Real>(inp, "BC_node_"+std::to_string(inode)+"/direction_Y");
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_YZ == bc_info.type ) {
            WARNING_("the boundary value for the prescribed displacement should be total value");
            // set the direction vector components Y, Z
            bc_info.value.comp(1) = get_par<Real>(inp, "BC_node_"+std::to_string(inode)+"/direction_Y");
            bc_info.value.comp(2) = get_par<Real>(inp, "BC_node_"+std::to_string(inode)+"/direction_Z");
        } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XZ == bc_info.type ) {
            WARNING_("the boundary value for the prescribed displacement should be total value");
            // set the direction vector components X, Z
            bc_info.value.comp(0) = get_par<Real>(inp, "BC_node_"+std::to_string(inode)+"/direction_X");
            bc_info.value.comp(2) = get_par<Real>(inp, "BC_node_"+std::to_string(inode)+"/direction_Z");
            // ...exception is caught here
        } else libmesh_error();
        //
        std::string time_func__str;
        std::string active_func__str;
        //
        if ( true ) {
            //
            const std::string ft_str = get_par<std::string>(inp, "BC_node_"+std::to_string(inode)+"/time_function");
            const unsigned int n_params = get_par_size<Real>(inp, "BC_node_"+std::to_string(inode)+"/time_parameters");
            //
            if        ( "flat_time_function"             == ft_str ) {
                bc_info.set_time_function(&flat_function);
                ASSERT_(n_params == 1, "wrong number of time parameters inserted");
            } else if ( "linear_time_function"           == ft_str ) {
                bc_info.set_time_function(&linear_function);
                ASSERT_(n_params == 1, "wrong number of time parameters inserted");
            } else if ( "linear_time_function__periodic" == ft_str ) {
                bc_info.set_time_function(&linear_function__periodic);
                ASSERT_(n_params == 2, "wrong number of time parameters inserted");
            } else if ( "ramp_time_function"             == ft_str ) {
                bc_info.set_time_function(&ramp_function);
                ASSERT_(n_params == 4, "wrong number of time parameters inserted");
            } else if ( "ramp_time_function__periodic"   == ft_str ) {
                bc_info.set_time_function(&ramp_function__periodic);
                ASSERT_(n_params == 5, "wrong number of time parameters inserted");
            } else if ( "steep_time_function"            == ft_str ) {
                bc_info.set_time_function(&steep_function);
                ASSERT_(n_params == 6, "wrong number of time parameters inserted");
            } else if ( "steep_time_function__periodic"  == ft_str ) {
                bc_info.set_time_function(&steep_function__periodic);
                ASSERT_(n_params == 7, "wrong number of time parameters inserted");
            } else if ( "exp_time_function"              == ft_str ) {
                bc_info.set_time_function(&exp_function);
                ASSERT_(n_params == 5, "wrong number of time parameters inserted");
            } else if ( "exp_time_function__periodic"    == ft_str ) {
                bc_info.set_time_function(&exp_function__periodic);
                ASSERT_(n_params == 6, "wrong number of time parameters inserted");
            } else {
                sprintf(err_msg, "exception caught for time function of BC node %d", inode);
                ERROR_(err_msg);
            }
            //
            time_func__str = ft_str;
            // set the time parameters
            bc_info.time_params.assign(n_params, 0.0);
            for (unsigned int i=0; i<n_params; i++) {
                bc_info.time_params[i] = get_par_comp<Real>(inp, "BC_node_"+std::to_string(inode)+"/time_parameters", i);
            }
            // end of if case...
        }
        //
        if ( "displacement_prescribed"    == bc_type_str ||
             "displacement_prescribed_x"  == bc_type_str ||
             "displacement_prescribed_y"  == bc_type_str ||
             "displacement_prescribed_z"  == bc_type_str ||
             "displacement_prescribed_xy" == bc_type_str ||
             "displacement_prescribed_yz" == bc_type_str ||
             "displacement_prescribed_xz" == bc_type_str )
        {
            //
            const std::string ft_str = get_par<std::string>(inp, "BC_node_"+std::to_string(inode)+"/active_function");
            const unsigned int n_params = get_par_size<Real>(inp, "BC_node_"+std::to_string(inode)+"/active_parameters");
            //
            if        ( "flat_time_function"             == ft_str ) {
                bc_info.set_active_function(&flat_function);
                ASSERT_(n_params == 1, "wrong number of active parameters inserted");
            } else if ( "linear_time_function"           == ft_str ) {
                bc_info.set_active_function(&linear_function);
                ASSERT_(n_params == 1, "wrong number of active parameters inserted");
            } else if ( "linear_time_function__periodic" == ft_str ) {
                bc_info.set_active_function(&linear_function__periodic);
                ASSERT_(n_params == 2, "wrong number of active parameters inserted");
            } else if ( "ramp_time_function"             == ft_str ) {
                bc_info.set_active_function(&ramp_function);
                ASSERT_(n_params == 4, "wrong number of active parameters inserted");
            } else if ( "ramp_time_function__periodic"   == ft_str ) {
                bc_info.set_active_function(&ramp_function__periodic);
                ASSERT_(n_params == 5, "wrong number of active parameters inserted");
            } else if ( "steep_time_function"            == ft_str ) {
                bc_info.set_active_function(&steep_function);
                ASSERT_(n_params == 6, "wrong number of active parameters inserted");
            } else if ( "steep_time_function__periodic"  == ft_str ) {
                bc_info.set_active_function(&steep_function__periodic);
                ASSERT_(n_params == 7, "wrong number of active parameters inserted");
            } else if ( "exp_time_function"              == ft_str ) {
                bc_info.set_active_function(&exp_function);
                ASSERT_(n_params == 5, "wrong number of active parameters inserted");
            } else if ( "exp_time_function__periodic"    == ft_str ) {
                bc_info.set_active_function(&exp_function__periodic);
                ASSERT_(n_params == 6, "wrong number of active parameters inserted");
            } else if ( "step_time_function"             == ft_str ) {
                bc_info.set_active_function(&step_function);
                ASSERT_(n_params == 3, "wrong number of active parameters inserted");
            } else {
                sprintf(err_msg, "exception caught for active function of BC node %d", inode);
                ERROR_(err_msg);
            }
            //
            active_func__str = ft_str;
            // set the active parameters
            bc_info.active_params.assign(n_params, 0.0);
            for (unsigned int i=0; i<n_params; i++) {
                bc_info.active_params[i] = get_par_comp<Real>(inp, "BC_node_"+std::to_string(inode)+"/active_parameters", i);
            }
            //
        } else if ( "traction_vector" == bc_type_str ) {
            //
            active_func__str = "flat_time_function";
            //
            bc_info.set_active_function(&flat_function);
            bc_info.active_params.assign(1, 1.0);
            //
            // ...exception is caught here
        } else {
            sprintf(err_msg, "exception caught for type of BC on boundary node %d", inode);
            ERROR_(err_msg);
        }
        //
        // insert data into corresponding container
        const my_pair<std::string, std::string> func_str(time_func__str, active_func__str);
        nodal_bcs__unsorted.push_back( my_triple< Point, BoundaryConditionInfo, my_pair<std::string, std::string> >( bc_pnt,
                                                                                                                     bc_info,
                                                                                                                     func_str ) );
        // end of nodal boundary condition loop...
    }
}
//=================================================================================================
static
void input_materials (DomainData& dmn, GetPot& inp) {
    //
    // loop for all biochemical materials
    std::map< int, std::vector<Real> > biochemical_mat;
    for (int imat=0; imat<get_par<int>(inp, "number_of_biochemical_materials"); imat++) {
        //
        const int m_id = get_par<int>(inp, "biochemical_material_"+std::to_string(imat)+"/ID", imat);
        //
        const int n_props = get_par<int>(inp, "biochemical_material_"+std::to_string(imat)+"/number_of_properties");
        //
        if ( par_ptr->get<ProblemCase>("solver") == FEM_WOUND_HEALING_SOLVER ) {
            // sanity check of material parameters inserted
            if        ( 5 == par_ptr->get<unsigned int>("reaction-diffusion solver: dimension") ) {
                ASSERT_( n_props == 22,
                         "number of biochemical material properties to insert is invalid");
            } else if ( 6 == par_ptr->get<unsigned int>("reaction-diffusion solver: dimension") ) {
                ASSERT_( n_props == 24,
                         "number of biochemical material properties to insert is invalid");
            } else {
                ; // do nothing here...
            }
            // end of if case...
        }
        //
        if ( par_ptr->get<ProblemCase>("solver") == FEM_TUMOUR_ANGIOGENESIS_SOLVER ) {
            // sanity check of material parameters inserted
            ASSERT_( n_props == 500,
                     "number of biochemical material properties to insert is invalid");
            // end of if case...
        }
        //
        std::vector<Real> tmp_vec;
        tmp_vec.assign(n_props, 0.0);
        biochemical_mat.insert( std::make_pair(m_id, tmp_vec) );
        //
        for (int iprop=0; iprop<n_props; iprop++) {
            biochemical_mat.find(m_id)->second[iprop] =
                get_par<Real>(inp, "biochemical_material_"+std::to_string(imat)+"/property_"+std::to_string(iprop), 0.0);
            // end of properties loop...
        }
        // end of biochemical materials loop...
    }
    //
    // loop for all biomechanical materials
    std::map< int, std::vector<Real> > biomechanical_mat;
    for (int imat=0; imat<get_par<int>(inp, "number_of_biomechanical_materials"); imat++) {
        //
        const int m_id = get_par<int>(inp, "biomechanical_material_"+std::to_string(imat)+"/ID", imat);
        //
        const int n_props = get_par<int>(inp, "biomechanical_material_"+std::to_string(imat)+"/number_of_properties");
        //
        if ( 3 == par_ptr->get<unsigned int>("Navier-Cauchy solver: dimension") ) {
            ASSERT_( n_props == 200,
                     "number of biomechanical material properties to insert is invalid");
        } else {
            ; // do nothing here...
        }
        //
        std::vector<Real> tmp_vec;
        tmp_vec.assign(n_props, 0.0);
        biomechanical_mat.insert( std::make_pair(m_id, tmp_vec) );
        //
        for (int iprop=0; iprop<n_props; iprop++) {
            biomechanical_mat.find(m_id)->second[iprop] =
                get_par<Real>(inp, "biomechanical_material_"+std::to_string(imat)+"/property_"+std::to_string(iprop), 0.0);
            // end of properties loop...
        }
        // end of biochemical materials loop...
    }
    // loop for material sets
    for (int imat=0; imat<get_par<int>(inp, "number_of_materials"); imat++) {
        //
        const material_id_type m_id = get_par<int>(inp, "material_"+std::to_string(imat)+"/ID");
        //
        MaterialInfo m_info;
        //
        const std::string m_default_title = "a_custom_material_" + std::to_string(m_id);
        //
        const std::string m_title_str = get_par<std::string>(inp, "material_"+std::to_string(imat)+"/title", m_default_title);
        //
        const int biochemical_mat_id =
            get_par<int>(inp, "material_"+std::to_string(imat)+"/biochemical_material_ID");
        ASSERT_( biochemical_mat.end() != biochemical_mat.find(biochemical_mat_id),
                 "biochemical material cannot be found");
        //
        const int biomechanical_mat_id =
            get_par<int>(inp, "material_"+std::to_string(imat)+"/biomechanical_material_ID");
        ASSERT_( biomechanical_mat.end() != biomechanical_mat.find(biomechanical_mat_id),
                 "biomechanical material cannot be found");
        //
        m_info.set_title() = m_title_str;
        m_info.set_biomech_param() = biomechanical_mat.find(biomechanical_mat_id)->second;
        m_info.set_biochem_param() = biochemical_mat.find(biochemical_mat_id)->second;
        //
        // insert data into corresponding container but make this check first
        ASSERT_( dmn.mater.end() == dmn.mater.find(m_id),
                 "material has already been inserted");
        dmn.mater.insert( std::make_pair(m_id, m_info) );
        // end of material sets loop...
    }
}
//=================================================================================================
static
void input_regions (DomainData& dmn, GetPot& inp) {
    // by default there is no skin (membrane regions)
    par_ptr->set<bool>("membrane elements present") = false;
    // loop for all regions
    for (int ireg=0; ireg<get_par<int>(inp, "number_of_regions"); ireg++) {
        //
        const region_id_type r_id = get_par<int>(inp, "region_"+std::to_string(ireg)+"/ID");
        //
        RegionInfo r_info;
        //
        r_info.is_damaged = get_par<bool>(inp, "region_"+std::to_string(ireg)+"/is_damaged");
        //
        r_info.is_membrane = get_par<bool>(inp, "region_"+std::to_string(ireg)+"/is_membrane");
        //
        if ( r_info.is_membrane ) par_ptr->set<bool>("membrane elements present") = true;
        //
        const material_id_type m_id = get_par<int>(inp, "region_"+std::to_string(ireg)+"/material_ID");
        //
        ASSERT_( dmn.mater.end() != dmn.mater.find(m_id),
                 "material has not been inserted in corresponding map container");
        r_info.mat = &(dmn.mater.find(m_id)->second);
        //
        if ( r_info.is_damaged ) {
            if ( get_par<bool>(inp, "region_"+std::to_string(ireg)+"/is_damage_constant_and_uniform", false) )
            {
                // obtain the constant value of the damage factor
                r_info.damage_factor_const =
                    get_par<Real>(inp, "region_"+std::to_string(ireg)+"/damage_value");
                // make this check first
                ASSERT_( r_info.damage_factor_const >= 0.0000 && //r_info.min_damage_factor &&
                         r_info.damage_factor_const <= 0.9999,   //r_info.max_damage_factor,
                         "damaged factor inserted is not valid" );
                //
                r_info.set_damage(damage_model__nothing);
                //
            } else {
                //
                r_info.set_damage(damage_model__max);
                //
            }
            r_info.set_plastic_deform(zero_permanent_strain);
            // end of if case...
        } else {
            r_info.set_non_damage(damage_model__min);
            r_info.set_plastic_deform(zero_permanent_strain);
        }
        //
        // set this default function for the multiscale case
        r_info.set_multiscale(multiscale_0);
        //
        //
        //
        if ( par_ptr->get<ProblemCase>("solver") == FEM_BIOCHEMICS_SOLVER ) {
        // check for the reaction-diffusion solver dimensionality
        if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension")               ==13 &&
             par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 1 &&
             par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables")     == 2 )
        {
            r_info.set_damage(damage_model__nothing);
            r_info.set_non_damage(damage_model__min);
            r_info.set_biochem( r_info.is_damaged
                              ? biochemics_reaction_diffusion_solver::damaged_region_1312
                              : biochemics_reaction_diffusion_solver::healthy_region_1312 );
        } else {
            libmesh_error();
        }
        // ...end this if-check
        }
        //
        //
        //
        if ( par_ptr->get<ProblemCase>("solver") == FEM_BIOMECHANICS_SOLVER ||
             par_ptr->get<ProblemCase>("solver") == IGA_BIOMECHANICS_SOLVER ) {
        // check if analysis is forward or inverse
        if ( par_ptr->get<bool>("Navier-Cauchy solver: inverse formulation") ) {
            if ( r_info.is_membrane ) {
                r_info.set_biomech_2d( r_info.is_damaged
                                     ? damaged::polynomial_2d_inverse
                                     : healthy::polynomial_2d_inverse );
            } else {
                if ( par_ptr->get<bool>("Navier-Cauchy solver: mixed formulation") ) {
                    r_info.set_biomech_3d( r_info.is_damaged
                                         ? damaged::polynomial_3d_inverse_2f
                                         : healthy::polynomial_3d_inverse_2f );
                } else {
                    const bool constitutive_modified =
                        get_par<bool>(inp, "region_"+std::to_string(ireg)+"/constitutive_modified", true);
                    //
                    if ( constitutive_modified )
                        r_info.set_biomech_3d( r_info.is_damaged
                                             ? damaged::polynomial_3d_inverse_mod
                                             : healthy::polynomial_3d_inverse_mod );
                    else
                        r_info.set_biomech_3d( r_info.is_damaged
                                             ? damaged::polynomial_3d_inverse_std
                                             : healthy::polynomial_3d_inverse_std );
                }
            }
        } else {
            if ( r_info.is_membrane ) {
                r_info.set_biomech_2d( r_info.is_damaged
                                     ? damaged::polynomial_2d_forward
                                     : healthy::polynomial_2d_forward );
            } else {
                if ( par_ptr->get<bool>("Navier-Cauchy solver: mixed formulation") ) {
                    r_info.set_biomech_3d( r_info.is_damaged
                                         ? damaged::polynomial_3d_forward_2f
                                         : healthy::polynomial_3d_forward_2f );
                } else {
                    const bool constitutive_modified =
                        get_par<bool>(inp, "region_"+std::to_string(ireg)+"/constitutive_modified", true);
                    //
                    if ( constitutive_modified )
                        r_info.set_biomech_3d( r_info.is_damaged
                                             ? damaged::polynomial_3d_forward_mod
                                             : healthy::polynomial_3d_forward_mod );
                    else
                        r_info.set_biomech_3d( r_info.is_damaged
                                             ? damaged::polynomial_3d_forward_std
                                             : healthy::polynomial_3d_forward_std );
                }
            }
        }
        // ...end this if-check
        }
        //
        //
        //
        if ( par_ptr->get<ProblemCase>("solver") == FEM_BIOMECHANICS_SOLVER &&
             par_ptr->get<bool>("Navier-Cauchy solver: explicit") ) {
        // sanity check
        if ( par_ptr->get<bool>("Navier-Cauchy solver: inverse formulation") ) libmesh_error();
        //
        if ( r_info.is_membrane ) {
            libmesh_error();
        } else {
            if ( par_ptr->get<bool>("Navier-Cauchy solver: mixed formulation") ) {
                r_info.set_biomech_3d( r_info.is_damaged
                                     ? damaged_explicit::polynomial_3d_forward_2f
                                     : healthy_explicit::polynomial_3d_forward_2f );
            } else {
                const bool constitutive_modified =
                    get_par<bool>(inp, "region_"+std::to_string(ireg)+"/constitutive_modified", true);
                //
                if ( constitutive_modified )
                    r_info.set_biomech_3d( r_info.is_damaged
                                         ? damaged_explicit::polynomial_3d_forward_mod
                                         : healthy_explicit::polynomial_3d_forward_mod );
                else
                    r_info.set_biomech_3d( r_info.is_damaged
                                         ? damaged_explicit::polynomial_3d_forward_std
                                         : healthy_explicit::polynomial_3d_forward_std );
            }
        }
        //
        // now distinguish amongst the types of multiscale models available
        const int multiscale_model = r_info.mat->get_biomech_param(100);
        if ( 0 == multiscale_model ) {
            // set the default function for the multiscale model
            r_info.set_multiscale(multiscale_0);
        } else if ( 1 == multiscale_model ) {
            // set this specified function for the multiscale model
            r_info.set_multiscale(collagen_extracellular_matrix::multiscale_3);
            // set the RVE input file
            const std::string fname =
                get_par<std::string>(inp, "region_"+std::to_string(ireg)+"/RVE/input_Gmsh_file");
            //
            par_ptr->set<std::string>("region "+std::to_string(r_id)+": input RVE mesh file") = fname;
            // now copy the input file into the results directory
            if ( libMesh::global_processor_id() == 0 ) {
                const std::string Input = fname;
                const std::string Input_copy = par_ptr->get<std::string>("results directory") + "/"
                                             + par_ptr->get<std::string>("output file name") + ".rve-" + number2string<int>(r_id) + ".msh";
                const std::string cmd = "cp " + Input + ' ' + Input_copy;
                ASSERT_(0==std::system(cmd.c_str()), "could not execute system command");
            }
            // an exception is caught here...
        } else libmesh_error();
        // ...end this if-check
        }
        //
        //
        //
        if ( par_ptr->get<ProblemCase>("solver") == FEM_WOUND_HEALING_SOLVER ) {
        // sanity checks
        if ( par_ptr->get<bool>("Navier-Cauchy solver: inverse formulation") ) libmesh_error();
        if ( par_ptr->get<bool>("Navier-Cauchy solver: explicit") ) libmesh_error();
        // check for the reaction-diffusion solver dimensionality
        if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension")               == 5 &&
             par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 1 &&
             par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables")     == 1 )
        {
            r_info.set_damage(wound_healing::damage_model_511);
            r_info.set_non_damage(damage_model__min);
            if ( par_ptr->get<bool>("reaction-diffusion solver: explicit") ) {
                r_info.set_biochem( r_info.is_damaged
                                  ? wound_healing::damaged_tissue_511
                                  : wound_healing::healthy_tissue_511 );
            } else {
                libmesh_error();
            }
            r_info.is_membrane ? r_info.set_coupled_2d(wound_healing::contraction_membrane_511)
                               : r_info.set_coupled_3d(wound_healing::contraction_solid_511);
            if ( r_info.is_membrane ) {
                r_info.set_biomech_2d( r_info.is_damaged
                                     ? damaged::polynomial_2d_forward
                                     : healthy::polynomial_2d_forward );
            } else {
                if ( par_ptr->get<bool>("Navier-Cauchy solver: mixed formulation") ) {
                    r_info.set_biomech_3d( r_info.is_damaged
                                         ? damaged::polynomial_3d_forward_2f
                                         : healthy::polynomial_3d_forward_2f );
                } else {
                    const bool constitutive_modified =
                        get_par<bool>(inp, "region_"+std::to_string(ireg)+"/constitutive_modified", true);
                    //
                    if ( constitutive_modified )
                        r_info.set_biomech_3d( r_info.is_damaged
                                             ? damaged::polynomial_3d_forward_mod
                                             : healthy::polynomial_3d_forward_mod );
                    else
                        r_info.set_biomech_3d( r_info.is_damaged
                                             ? damaged::polynomial_3d_forward_std
                                             : healthy::polynomial_3d_forward_std );
                }
            }
        } else if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension")               == 6 &&
                    par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 1 &&
                    par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables")     == 1 )
        {
            r_info.set_damage(wound_healing::damage_model_611);
            r_info.set_non_damage(damage_model__min);
            if ( par_ptr->get<bool>("reaction-diffusion solver: explicit") ) {
                r_info.set_biochem( r_info.is_damaged
                                  ? wound_healing::damaged_tissue_611
                                  : wound_healing::healthy_tissue_611 );
            } else {
                libmesh_error();
            }
            r_info.is_membrane ? r_info.set_coupled_2d(wound_healing::contraction_membrane_611)
                               : r_info.set_coupled_3d(wound_healing::contraction_solid_611);
            if ( r_info.is_membrane ) {
                r_info.set_biomech_2d( r_info.is_damaged
                                     ? damaged::polynomial_2d_forward
                                     : healthy::polynomial_2d_forward );
            } else {
                if ( par_ptr->get<bool>("Navier-Cauchy solver: mixed formulation") ) {
                    r_info.set_biomech_3d( r_info.is_damaged
                                         ? damaged::polynomial_3d_forward_2f
                                         : healthy::polynomial_3d_forward_2f );
                } else {
                    const bool constitutive_modified =
                        get_par<bool>(inp, "region_"+std::to_string(ireg)+"/constitutive_modified", true);
                    //
                    if ( constitutive_modified )
                        r_info.set_biomech_3d( r_info.is_damaged
                                             ? damaged::polynomial_3d_forward_mod
                                             : healthy::polynomial_3d_forward_mod );
                    else
                        r_info.set_biomech_3d( r_info.is_damaged
                                             ? damaged::polynomial_3d_forward_std
                                             : healthy::polynomial_3d_forward_std );
                }
            }
        } else {
            libmesh_error();
        }
        // ...end this if-check
        }
        //
        //
        //
        if ( par_ptr->get<ProblemCase>("solver") == FEM_TUMOUR_ANGIOGENESIS_SOLVER ) {
        // sanity checks
        if ( par_ptr->get<bool>("Navier-Cauchy solver: inverse formulation") ) libmesh_error();
        if ( par_ptr->get<bool>("Navier-Cauchy solver: explicit") ) libmesh_error();
        // check for the reaction-diffusion solver dimensionality
        if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension")               == 4 &&
             par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 9 &&
             par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables")     == 8 )
        {
            r_info.set_damage(tumour_angiogenesis::damage_model_498);
            r_info.set_non_damage(tumour_angiogenesis::damage_model_498);
            r_info.set_biochem( r_info.is_damaged
                              ? tumour_angiogenesis::damaged_tissue_498
                              : tumour_angiogenesis::healthy_tissue_498 );
            if ( r_info.is_membrane ) {
                r_info.set_coupled_2d(tumour_angiogenesis::growth_membrane_498);
                r_info.set_plastic_deform( r_info.is_damaged
                                         ? tumour_angiogenesis::inelastic_deform_membrane_498
                                         : zero_permanent_strain );
                r_info.set_biomech_2d( damaged::polynomial_2d_forward );
            } else {
                r_info.set_coupled_3d(tumour_angiogenesis::growth_solid_498);
                r_info.set_plastic_deform( r_info.is_damaged
                                         ? tumour_angiogenesis::inelastic_deform_solid_498
                                         : zero_permanent_strain );
                if ( par_ptr->get<bool>("Navier-Cauchy solver: mixed formulation") ) {
                    r_info.set_biomech_3d( damaged::polynomial_3d_forward_2f );
                } else {
                    const bool constitutive_modified =
                        get_par<bool>(inp, "region_"+std::to_string(ireg)+"/constitutive_modified", true);
                    //
                    if ( constitutive_modified )
                        r_info.set_biomech_3d( damaged::polynomial_3d_forward_mod );
                    else
                        r_info.set_biomech_3d( damaged::polynomial_3d_forward_std );
                }
            }
        } else {
            libmesh_error();
        }
        // ...end this if-check
        }
        //
        //
        //
        if ( par_ptr->get<ProblemCase>("solver") == FEM_INVITRO_TUMOROID_SOLVER ) {
        // check for the reaction-diffusion solver dimensionality
        if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension")               == 7 &&
             par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 1 &&
             par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables")     == 1 )
        {
            r_info.set_damage(damage_model__nothing);
            r_info.set_non_damage(damage_model__min);
            r_info.set_biochem( r_info.is_damaged
                              ? invitro_tumoroid_growth::damaged_region_711
                              : invitro_tumoroid_growth::healthy_region_711 );
        // check for the reaction-diffusion solver dimensionality
        } else if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension")               == 8 &&
                    par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 1 &&
                    par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables")     == 3 )
        {
            r_info.set_damage(damage_model__nothing);
            r_info.set_non_damage(damage_model__min);
            r_info.set_biochem( r_info.is_damaged
                              ? invitro_tumoroid_growth::damaged_region_813
                              : invitro_tumoroid_growth::healthy_region_813 );
        // check for the reaction-diffusion solver dimensionality
        } else if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension")               ==11 &&
                    par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 1 &&
                    par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables")     == 5 )
        {
            r_info.set_damage(damage_model__nothing);
            r_info.set_non_damage(damage_model__min);
            r_info.set_biochem( r_info.is_damaged
                              ? invitro_tumoroid_growth::damaged_region_1115
                              : invitro_tumoroid_growth::healthy_region_1115 );
        // check for the reaction-diffusion solver dimensionality
        } else if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension")               ==45 &&
                    par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 9 &&
                    par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables")     == 6 )
        {
            r_info.set_damage(damage_model__nothing);
            r_info.set_non_damage(damage_model__min);
            r_info.set_biochem(invitro_tumoroid_growth::multi_region_4596);
        } else {
            libmesh_error();
        }
        // ...end this if-check
        }
        //
        //
        //
        if ( par_ptr->get<ProblemCase>("solver") == FEM_RT_TUMOUR_FIBROSIS_SOLVER ) {
        // check for the reaction-diffusion solver dimensionality
        if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension")               == 3 &&
             par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 1 &&
             par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables")     == 1 )
        {
            r_info.set_damage(damage_model__nothing);
            r_info.set_non_damage(damage_model__min);
            r_info.set_biochem( r_info.is_damaged
                              ? radiotherapy_tumour_fibrosis::damaged_region_311
                              : radiotherapy_tumour_fibrosis::healthy_region_311 );
        } else {
            libmesh_error();
        }
        // ...end this if-check
        }
        //
        //
        //
        // NOTE enforce initialisation of pointer-to-function routine for the coupled systems
        // if region is not damaged at all!
        if ( ! r_info.is_damaged ) {
            r_info.is_membrane ? r_info.set_coupled_2d(coupled_membrane_0d)
                               : r_info.set_coupled_3d(coupled_solid_0d);
        }
        //
        // insert data into corresponding container but make this check first
        ASSERT_( dmn.regn.end() == dmn.regn.find(r_id),
                 "region has already been inserted");
        dmn.regn.insert( std::make_pair(r_id, r_info) );
        // end of regions loop...
    }
}
//=================================================================================================
static
void input_body_force (DomainData& dmn, GetPot& inp) {
    if ( par_ptr->get<ProblemCase>("solver") != IGA_BIOMECHANICS_SOLVER        &&
         par_ptr->get<ProblemCase>("solver") != FEM_BIOMECHANICS_SOLVER        &&
         par_ptr->get<ProblemCase>("solver") != FEM_WOUND_HEALING_SOLVER       &&
         par_ptr->get<ProblemCase>("solver") != FEM_TUMOUR_ANGIOGENESIS_SOLVER ) return;
    //
    const bool apply_body_force =
        get_par<bool>(inp, "navier_cauchy_solver/apply_body_force", false);
    //
    par_ptr->set<bool>("Navier-Cauchy solver: apply body force") = apply_body_force;
    if ( ! apply_body_force ) return;
    //
    if ( get_par_size<Real>(inp, "navier_cauchy_solver/body_force/direction") != SP3D )
        ERROR_("vector components of the body force direction loading is not equal to 3");
    // set the direction vector
    dmn.bforce.value.comp =
        get_par_comp<Real>(inp, "navier_cauchy_solver/body_force/direction", 0),
        get_par_comp<Real>(inp, "navier_cauchy_solver/body_force/direction", 1),
        get_par_comp<Real>(inp, "navier_cauchy_solver/body_force/direction", 2);
    //
    if ( par_ptr->get<ProblemCase>("solver") == FEM_WOUND_HEALING_SOLVER ) return;
    //
    const std::string ft_str =
        get_par<std::string>(inp, "navier_cauchy_solver/body_force/time_function");
    const unsigned int n_params =
        get_par_size<Real>(inp, "navier_cauchy_solver/body_force/time_parameters");
    //
    if        ( "flat_time_function"            == ft_str ) {
        dmn.bforce.set_time_function(&flat_function);
        ASSERT_(n_params == 1, "wrong number of time parameters inserted");
    } else if ( "step_time_function"            == ft_str ) {
        dmn.bforce.set_time_function(&step_function);
        ASSERT_(n_params == 3, "wrong number of time parameters inserted");
    } else if ( "step_time_function__periodic"  == ft_str ) {
        dmn.bforce.set_time_function(&step_function__periodic);
        ASSERT_(n_params == 4, "wrong number of time parameters inserted");
    } else if ( "linear_time_function"           == ft_str ) {
        dmn.bforce.set_time_function(&linear_function);
        ASSERT_(n_params == 1, "wrong number of time parameters inserted");
    } else if ( "linear_time_function__periodic" == ft_str ) {
        dmn.bforce.set_time_function(&linear_function__periodic);
        ASSERT_(n_params == 2, "wrong number of time parameters inserted");
    } else if ( "ramp_time_function"             == ft_str ) {
        dmn.bforce.set_time_function(&ramp_function);
        ASSERT_(n_params == 4, "wrong number of time parameters inserted");
    } else if ( "ramp_time_function__periodic"   == ft_str ) {
        dmn.bforce.set_time_function(&ramp_function__periodic);
        ASSERT_(n_params == 5, "wrong number of time parameters inserted");
    } else if ( "steep_time_function"            == ft_str ) {
        dmn.bforce.set_time_function(&steep_function);
        ASSERT_(n_params == 6, "wrong number of time parameters inserted");
    } else if ( "steep_time_function__periodic"  == ft_str ) {
        dmn.bforce.set_time_function(&steep_function__periodic);
        ASSERT_(n_params == 7, "wrong number of time parameters inserted");
    } else if ( "exp_time_function"              == ft_str ) {
        dmn.bforce.set_time_function(&exp_function);
        ASSERT_(n_params == 5, "wrong number of time parameters inserted");
    } else if ( "exp_time_function__periodic"    == ft_str ) {
        dmn.bforce.set_time_function(&exp_function__periodic);
        ASSERT_(n_params == 6, "wrong number of time parameters inserted");
    } else {
        ERROR_("exception is caught for body force time function");
    }
    // set the time parameters
    dmn.bforce.time_params.assign(n_params, 0.0);
    for (unsigned int i=0; i<n_params; i++) {
        dmn.bforce.time_params[i] =
            get_par_comp<Real>(inp, "navier_cauchy_solver/body_force/time_parameters", i);
    }
}
//=================================================================================================
static
void input_nodal_force (DomainData& dmn, GetPot& inp) {
    par_ptr->set<bool>("Navier-Cauchy solver: apply nodal force") = false;
    //
    if ( par_ptr->get<ProblemCase>("solver") != IGA_BIOMECHANICS_SOLVER        &&
         par_ptr->get<ProblemCase>("solver") != FEM_BIOMECHANICS_SOLVER        &&
         par_ptr->get<ProblemCase>("solver") != FEM_WOUND_HEALING_SOLVER       &&
         par_ptr->get<ProblemCase>("solver") != FEM_TUMOUR_ANGIOGENESIS_SOLVER ) return;
    //
    const bool apply_nodal_force =
        get_par<bool>(inp, "navier_cauchy_solver/apply_nodal_force", false);
    //
    par_ptr->set<bool>("Navier-Cauchy solver: apply nodal force") = apply_nodal_force;
    if ( ! apply_nodal_force ) return;
    //
    if ( par_ptr->get<ProblemCase>("solver") == FEM_WOUND_HEALING_SOLVER ) return;
    //
    std::string ft_str =
        get_par<std::string>(inp, "navier_cauchy_solver/employ_nodal_force/time_function");
    unsigned int n_params =
        get_par_size<Real>(inp, "navier_cauchy_solver/employ_nodal_force/time_parameters");
    //
    if        ( "flat_time_function"           == ft_str ) {
        dmn.nforce.set_time_function__employ(&flat_function);
        ASSERT_(n_params == 1, "wrong number of time parameters inserted");
    } else if ( "step_time_function"           == ft_str ) {
        dmn.nforce.set_time_function__employ(&step_function);
        ASSERT_(n_params == 3, "wrong number of time parameters inserted");
    } else if ( "step_time_function__periodic" == ft_str ) {
        dmn.nforce.set_time_function__employ(&step_function__periodic);
        ASSERT_(n_params == 4, "wrong number of time parameters inserted");
    } else {
        ERROR_("exception is caught for nodal force (apply) time function");
    }
    // set the time parameters
    dmn.nforce.time_params__employ.assign(n_params, 0.0);
    for (unsigned int i=0; i<n_params; i++) {
        dmn.nforce.time_params__employ[i] =
            get_par_comp<Real>(inp, "navier_cauchy_solver/employ_nodal_force/time_parameters", i);
    }
    //
    ft_str =
        get_par<std::string>(inp, "navier_cauchy_solver/evaluate_nodal_force/time_function");
    n_params =
        get_par_size<Real>(inp, "navier_cauchy_solver/evaluate_nodal_force/time_parameters");
    //
    if        ( "step_time_function"           == ft_str ) {
        dmn.nforce.set_time_function__evaluate(&step_function);
        ASSERT_(n_params == 3, "wrong number of time parameters inserted");
    } else if ( "step_time_function__periodic" == ft_str ) {
        dmn.nforce.set_time_function__evaluate(&step_function__periodic);
        ASSERT_(n_params == 4, "wrong number of time parameters inserted");
    } else {
        ERROR_("exception is caught for nodal force (evaluate) time function");
    }
    // set the time parameters
    dmn.nforce.time_params__evaluate.assign(n_params, 0.0);
    for (unsigned int i=0; i<n_params; i++) {
        dmn.nforce.time_params__evaluate[i] =
            get_par_comp<Real>(inp, "navier_cauchy_solver/evaluate_nodal_force/time_parameters", i);
    }
    //
    WARNING_("check for no overlapping between the time fuctions used for nodal force");
}
//=================================================================================================
static
void input_nodal_force_info (const DomainData& dmn, GetPot& inp) {
    par_ptr->set<int>("nodal force evaluation: number of external surfaces") = 0;
    par_ptr->set<int>("nodal force evaluation: number of regions") = 0;
    par_ptr->set<int>("nodal force evaluation: number of internal surfaces") = 0;
    //
    if ( par_ptr->get<ProblemCase>("solver") != IGA_BIOMECHANICS_SOLVER        &&
         par_ptr->get<ProblemCase>("solver") != FEM_BIOMECHANICS_SOLVER        &&
         par_ptr->get<ProblemCase>("solver") != FEM_WOUND_HEALING_SOLVER       &&
         par_ptr->get<ProblemCase>("solver") != FEM_TUMOUR_ANGIOGENESIS_SOLVER ) return;
    //
    par_ptr->set<int>("nodal force evaluation: number of external surfaces") =
        get_par<int>(inp, "nodal_force_evaluation/number_of_external_surfaces", 0);
    //
    for (int l=0; l<par_ptr->get<int>("nodal force evaluation: number of external surfaces"); l++) {
        //
        const boundary_id_type bc_id =
            get_par<int>(inp, "nodal_force_evaluation/external_surface_"+std::to_string(l)+"/ID");
        //
        ASSERT_( dmn.bcs.end() != dmn.bcs.find(bc_id),
                 "boundary surface has not been inserted");
        //
        par_ptr->set<boundary_id_type>("nodal force evaluation: external surface "+std::to_string(l)) = bc_id;
    }
    //
    par_ptr->set<int>("nodal force evaluation: number of regions") =
        get_par<int>(inp, "nodal_force_evaluation/number_of_regions", 0);
    //
    if ( (int)dmn.regn.size() == par_ptr->get<int>("nodal force evaluation: number of regions") ) {
        int l = 0;
        for ( std::map<region_id_type, RegionInfo>::const_iterator
              ci=dmn.regn.begin(); ci!=dmn.regn.end(); ci++ ) {
            const region_id_type r_id = ci->first;
            //
            par_ptr->set<region_id_type>("nodal force evaluation: region "+std::to_string(l)) = r_id;
            //
            ++l;
        }
    } else {
        for (int l=0; l<par_ptr->get<int>("nodal force evaluation: number of regions"); l++) {
            //
            const region_id_type r_id =
                get_par<int>(inp, "nodal_force_evaluation/region_"+std::to_string(l)+"/ID");
            //
            ASSERT_( dmn.regn.end() != dmn.regn.find(r_id),
                     "region has not been inserted");
            //
            par_ptr->set<region_id_type>("nodal force evaluation: region "+std::to_string(l)) = r_id;
        }
    }
    //
    par_ptr->set<int>("nodal force evaluation: number of internal surfaces") =
        get_par<int>(inp, "nodal_force_evaluation/number_of_internal_surfaces", 0);
    //
    for (int l=0; l<par_ptr->get<int>("nodal force evaluation: number of internal surfaces"); l++) {
        //
        if ( get_par_size<int>(inp, "nodal_force_evaluation/region_pair_"+std::to_string(l)+"/ID") != 2 ) {
            sprintf(err_msg, "nodal force evaluation for pair of regions (#%d) is not correct", l);
            ERROR_(err_msg);
        } else {
            const region_id_type r0_id = get_par_comp<int>(inp, "nodal_force_evaluation/region_pair_"+std::to_string(l)+"/ID", 0),
                                 r1_id = get_par_comp<int>(inp, "nodal_force_evaluation/region_pair_"+std::to_string(l)+"/ID", 1);
            //
            ASSERT_(dmn.regn.end() != dmn.regn.find(r0_id), "region has not been inserted");
            ASSERT_(dmn.regn.end() != dmn.regn.find(r1_id), "region has not been inserted");
            //
            par_ptr->set<region_id_type>("nodal force evaluation: region-0 "+std::to_string(l)) = r0_id;
            par_ptr->set<region_id_type>("nodal force evaluation: region-1 "+std::to_string(l)) = r1_id;
        }
    }
}
//=================================================================================================
static
void input_navier_cauchy_solver (GetPot& inp) {
    // set these default parameters for these solvers
    if ( par_ptr->get<ProblemCase>("solver") != IGA_BIOMECHANICS_SOLVER        &&
         par_ptr->get<ProblemCase>("solver") != FEM_BIOMECHANICS_SOLVER        &&
         par_ptr->get<ProblemCase>("solver") != FEM_WOUND_HEALING_SOLVER       &&
         par_ptr->get<ProblemCase>("solver") != FEM_TUMOUR_ANGIOGENESIS_SOLVER )
    {
        //
        par_ptr->set<unsigned int>("Navier-Cauchy solver: dimension") = 0;
        par_ptr->set<unsigned int>("Navier-Cauchy solver: internal variables") = 0;
        //
        par_ptr->set<bool>("Navier-Cauchy solver: mixed formulation") = false;
        par_ptr->set<bool>("Navier-Cauchy solver: inverse formulation") = false;
        //
        return;
    }
    //
    const unsigned int nvars     = get_par<int>(inp, "navier_cauchy_solver/dimension");
    const unsigned int nvars_int = get_par<int>(inp, "navier_cauchy_solver/internal_vars", 3);
    par_ptr->set<unsigned int>("Navier-Cauchy solver: dimension") = nvars;
    par_ptr->set<unsigned int>("Navier-Cauchy solver: internal variables") = nvars_int;
    //
    ASSERT_(nvars==SP3D, "Navier-Cauchy solver variables should be exact 3");
    ASSERT_(nvars_int>=3, "Navier-Cauchy solver internal variables should be >=3");
    //
    par_ptr->set<std::string>("Navier-Cauchy solver: label") =
        get_par<std::string>(inp, "navier_cauchy_solver/label", "NavierCauchySolver3D");
    //
    par_ptr->set<bool>("Navier-Cauchy solver: explicit") =
        get_par<bool>(inp, "navier_cauchy_solver/explicit_implementation", false);
    //
    if ( par_ptr->get<bool>("Navier-Cauchy solver: explicit") ) {
        //
        par_ptr->set<bool>("Navier-Cauchy solver: abort non-convergence") =
            get_par<bool>(inp, "navier_cauchy_solver/abort_nonconvergence", false);
        par_ptr->set<int>("Navier-Cauchy solver: maximum increments") =
            get_par<int>(inp, "navier_cauchy_solver/maximum_increments", 1);
        par_ptr->set<int>("Navier-Cauchy solver: maximum increments (ceiling scale)") =
            get_par<int>(inp, "navier_cauchy_solver/maximum_increments/ceiling_scale", 0);
        par_ptr->set<Real>("Navier-Cauchy solver: window") =
            get_par<Real>(inp, "navier_cauchy_solver/window", 1.0);
        //
        par_ptr->set<Real>("Navier-Cauchy solver: residual tolerance") =
            get_par<Real>(inp, "navier_cauchy_solver/residual_tolerance");
        //
        par_ptr->set<Real>("Navier-Cauchy solver: mass-proportional damping coefficient") =
            get_par<Real>(inp, "navier_cauchy_solver/damping_coefficient", 0.0);
    } else {
        //
        par_ptr->set<bool>("Navier-Cauchy solver: abort non-convergence") =
            get_par<bool>(inp, "navier_cauchy_solver/abort_nonconvergence", true);
        par_ptr->set<int>("Navier-Cauchy solver: maximum increments") =
            get_par<int>(inp, "navier_cauchy_solver/maximum_increments", 1);
        par_ptr->set<int>("Navier-Cauchy solver: maximum increments (ceiling scale)") =
            get_par<int>(inp, "navier_cauchy_solver/maximum_increments/ceiling_scale", 0);
        par_ptr->set<Real>("Navier-Cauchy solver: window") =
            get_par<Real>(inp, "navier_cauchy_solver/window", 1.0);
        //
        par_ptr->set<bool>("Navier-Cauchy solver: full Newton-Raphson") =
            get_par<bool>(inp, "navier_cauchy_solver/full_Newton_Raphson", true);
        par_ptr->set<int>("Navier-Cauchy solver: maximum iterations") =
            get_par<int>(inp, "navier_cauchy_solver/maximum_iterations", 10);
        par_ptr->set<int>("Navier-Cauchy solver: minimum iterations") =
            get_par<int>(inp, "navier_cauchy_solver/minimum_iterations", 4);
        //
        par_ptr->set<Real>("Navier-Cauchy solver: solution tolerance") =
            get_par<Real>(inp, "navier_cauchy_solver/solution_tolerance");
        par_ptr->set<Real>("Navier-Cauchy solver: residual tolerance") =
            get_par<Real>(inp, "navier_cauchy_solver/residual_tolerance");
        par_ptr->set<Real>("Navier-Cauchy solver: energy tolerance")   =
            get_par<Real>(inp, "navier_cauchy_solver/energy_tolerance");
        par_ptr->set<bool>("Navier-Cauchy solver: load control")        =
            get_par<bool>(inp, "navier_cauchy_solver/load_control");
        //
    }
    //
    par_ptr->set<int>("Navier-Cauchy solver: solution increment") =
        get_par<int>(inp, "navier_cauchy_solver/solution_increment", 1);
    par_ptr->set<int>("Navier-Cauchy solver: restart increment") =
        get_par<int>(inp, "navier_cauchy_solver/restart_increment", 1);
    //
    par_ptr->set<bool>("Navier-Cauchy solver: calculate gradients") =
        get_par<bool>(inp, "navier_cauchy_solver/calculate_gradients");
    if        ( par_ptr->get<ProblemCase>("solver") ==  IGA_BIOMECHANICS_SOLVER ||
                par_ptr->get<ProblemCase>("solver") ==  FEM_BIOMECHANICS_SOLVER ) {
        par_ptr->set<bool>("Navier-Cauchy solver: save updated mesh") =
            get_par<bool>(inp, "navier_cauchy_solver/save_updated_mesh", true);
    } else if ( par_ptr->get<ProblemCase>("solver") == FEM_WOUND_HEALING_SOLVER       ||
                par_ptr->get<ProblemCase>("solver") == FEM_TUMOUR_ANGIOGENESIS_SOLVER ) {
        par_ptr->set<bool>("Navier-Cauchy solver: save updated mesh") = false;
    } else libmesh_error();
    if ( par_ptr->get<unsigned int>("input restart from file") ) {
        if        ( par_ptr->get<ProblemCase>("solver") ==  IGA_BIOMECHANICS_SOLVER ||
                    par_ptr->get<ProblemCase>("solver") ==  FEM_BIOMECHANICS_SOLVER ) {
            par_ptr->set<int>("Navier-Cauchy solver: initial increment") =
                get_par<int>(inp, "navier_cauchy_solver/initial_increment", 0);
            if ( par_ptr->get<int>("Navier-Cauchy solver: initial increment") < 0 ) libmesh_error();
            if ( par_ptr->get<int>("Navier-Cauchy solver: initial increment") >= par_ptr->get<int>("Navier-Cauchy solver: maximum increments") ) libmesh_error();
        } else if ( par_ptr->get<ProblemCase>("solver") == FEM_WOUND_HEALING_SOLVER       ||
                    par_ptr->get<ProblemCase>("solver") == FEM_TUMOUR_ANGIOGENESIS_SOLVER ) {
            par_ptr->set<int>("Navier-Cauchy solver: initial increment") = 0;
        } else libmesh_error();
    } else {
        par_ptr->set<int>("Navier-Cauchy solver: initial increment") = 0;
    }
    //
    par_ptr->set<bool>("Navier-Cauchy solver: mixed formulation") =
        get_par<bool>(inp, "navier_cauchy_solver/mixed_formulation", false);
    if ( par_ptr->get<bool>("Navier-Cauchy solver: mixed formulation") ) {
        // check if analysis is based on IGA
        if ( par_ptr->get<ProblemCase>("solver") == IGA_BIOMECHANICS_SOLVER ) {
            //
            par_ptr->set<std::string>("Navier-Cauchy solver: mixed formulation IGA parametric mesh for discontinuous pressure") =
                get_par<std::string>(inp, "navier_cauchy_solver/mixed_formulation/IGA_parametric_mesh_label");
            //
        // ...or in FEM
        } else if ( par_ptr->get<ProblemCase>("solver") == FEM_BIOMECHANICS_SOLVER        ||
                    par_ptr->get<ProblemCase>("solver") == FEM_WOUND_HEALING_SOLVER       ||
                    par_ptr->get<ProblemCase>("solver") == FEM_TUMOUR_ANGIOGENESIS_SOLVER ) {
            //
            par_ptr->set<bool>("Navier-Cauchy solver: mixed formulation discontinuous pressure") =
                get_par<bool>(inp, "navier_cauchy_solver/mixed_formulation/discontinuous_pressure", true);
            //
        // ...else an exception is caught!
        } else libmesh_error();
    }
    //
    if        ( par_ptr->get<ProblemCase>("solver") == FEM_WOUND_HEALING_SOLVER ) {
        par_ptr->set<bool>("Navier-Cauchy solver: apply wound contraction") = true;
    } else if ( par_ptr->get<ProblemCase>("solver") == IGA_BIOMECHANICS_SOLVER        ||
                par_ptr->get<ProblemCase>("solver") == FEM_BIOMECHANICS_SOLVER        ||
                par_ptr->get<ProblemCase>("solver") == FEM_TUMOUR_ANGIOGENESIS_SOLVER ) {
        par_ptr->set<bool>("Navier-Cauchy solver: apply wound contraction") = false;
    } else libmesh_error();
    //
    if        ( par_ptr->get<ProblemCase>("solver") == FEM_TUMOUR_ANGIOGENESIS_SOLVER ) {
        par_ptr->set<bool>("Navier-Cauchy solver: apply tumour growth") = true;
    } else if ( par_ptr->get<ProblemCase>("solver") == IGA_BIOMECHANICS_SOLVER  ||
                par_ptr->get<ProblemCase>("solver") == FEM_BIOMECHANICS_SOLVER  ||
                par_ptr->get<ProblemCase>("solver") == FEM_WOUND_HEALING_SOLVER ) {
        par_ptr->set<bool>("Navier-Cauchy solver: apply tumour growth") = false;
    } else libmesh_error();
    //
    if        ( par_ptr->get<ProblemCase>("solver") == IGA_BIOMECHANICS_SOLVER ||
                par_ptr->get<ProblemCase>("solver") == FEM_BIOMECHANICS_SOLVER ) {
        par_ptr->set<bool>("Navier-Cauchy solver: inverse formulation") =
            get_par<bool>(inp, "navier_cauchy_solver/inverse_analysis", false);
    } else if ( par_ptr->get<ProblemCase>("solver") == FEM_WOUND_HEALING_SOLVER       ||
                par_ptr->get<ProblemCase>("solver") == FEM_TUMOUR_ANGIOGENESIS_SOLVER ) {
        par_ptr->set<bool>("Navier-Cauchy solver: inverse formulation") = false;
    } else libmesh_error();
}
//=================================================================================================
static
void input_reaction_diffusion_solver (GetPot& inp) {
    // set these default parameters for these solvers
    if ( par_ptr->get<ProblemCase>("solver") != FEM_BIOCHEMICS_SOLVER             &&
         par_ptr->get<ProblemCase>("solver") != FEM_WOUND_HEALING_SOLVER          &&
         par_ptr->get<ProblemCase>("solver") != FEM_INVITRO_TUMOROID_SOLVER       &&
         par_ptr->get<ProblemCase>("solver") != FEM_TUMOUR_ANGIOGENESIS_SOLVER    &&
         par_ptr->get<ProblemCase>("solver") != FEM_RT_TUMOUR_FIBROSIS_SOLVER     )
    {
        //
        par_ptr->set<unsigned int>("reaction-diffusion solver: dimension") = 0;
        par_ptr->set<unsigned int>("reaction-diffusion solver: discontinuous variables") = 0;
        par_ptr->set<unsigned int>("reaction-diffusion solver: auxiliary variables") = 0;
        //
        return;
    }
    //
    const unsigned int nvars = get_par<int>(inp, "reaction_diffusion_solver/dimension");
    par_ptr->set<unsigned int>("reaction-diffusion solver: dimension") = nvars;
    //
    par_ptr->set<std::string>("reaction-diffusion solver: label") =
        get_par<std::string>(inp, "reaction_diffusion_solver/label", "ReactionDiffusionSolverXD");
    //
    const unsigned int ndcvars = get_par<int>(inp, "reaction_diffusion_solver/discontinuous_variables", 1);
    par_ptr->set<unsigned int>("reaction-diffusion solver: discontinuous variables") = ndcvars;
    //
    const unsigned int nauxvars = get_par<int>(inp, "reaction_diffusion_solver/auxiliary_variables", 0);
    par_ptr->set<unsigned int>("reaction-diffusion solver: auxiliary variables") = nauxvars;
    //
    par_ptr->set<Real>("reaction-diffusion solver: time step") =
        get_par<Real>(inp, "reaction_diffusion_solver/time_step", 1.0);
    par_ptr->set<int>("reaction-diffusion solver: maximum time steps") =
        get_par<int>(inp, "reaction_diffusion_solver/maximum_time_steps", 1);
    par_ptr->set<int>("reaction-diffusion solver: solution increment") =
        get_par<int>(inp, "reaction_diffusion_solver/solution_increment", 1);
    par_ptr->set<int>("reaction-diffusion solver: restart increment") =
        get_par<int>(inp, "reaction_diffusion_solver/restart_increment", 1);
    if ( par_ptr->get<unsigned int>("input restart from file") ) {
        if ( par_ptr->get<ProblemCase>("solver") == FEM_BIOCHEMICS_SOLVER             ||
             par_ptr->get<ProblemCase>("solver") == FEM_INVITRO_TUMOROID_SOLVER       ||
             par_ptr->get<ProblemCase>("solver") == FEM_RT_TUMOUR_FIBROSIS_SOLVER     ) {
            par_ptr->set<int>("reaction-diffusion solver: initial time") =
                get_par<int>(inp, "reaction_diffusion_solver/initial_time", 0);
            par_ptr->set<Real>("current time") =
                par_ptr->get<int>("reaction-diffusion solver: initial time") *
                par_ptr->get<Real>("reaction-diffusion solver: time step");
            // some validation checks
            const int ini_time = par_ptr->get<int>("reaction-diffusion solver: initial time"),
                      max_time = par_ptr->get<int>("reaction-diffusion solver: maximum time steps");
            ASSERT_(ini_time>=0, "erroneous value inserted for the initial time");
            ASSERT_(ini_time<max_time, "erroneous value inserted for the initial time");
        } else if ( par_ptr->get<ProblemCase>("solver") == FEM_TUMOUR_ANGIOGENESIS_SOLVER ) {
            ; // nothing to do here...
        } else if ( par_ptr->get<ProblemCase>("solver") == FEM_WOUND_HEALING_SOLVER ) {
            par_ptr->set<int>("reaction-diffusion solver: initial time") = 0;
            par_ptr->set<Real>("current time") = 0.0;
        } else libmesh_error();
    } else {
        par_ptr->set<int>("reaction-diffusion solver: initial time") = 0;
        par_ptr->set<Real>("current time") = 0.0;
        // some validation checks
        const int ini_time = get_par<int>(inp, "reaction_diffusion_solver/initial_time", 0),
                  max_time = par_ptr->get<int>("reaction-diffusion solver: maximum time steps");
        ASSERT_(ini_time==0, "erroneous value inserted for the initial time");
        ASSERT_(ini_time<max_time, "erroneous value inserted for the initial time");
    }
    //
    //
    ASSERT_(get_par_size<int>(inp, "reaction_diffusion_solver/healthy_domain/initial_field_function") == nvars,
            "number of functions for (healthy domain) initial values for reaction-diffusion solver is not correct");
    for (unsigned int ivar=0; ivar<nvars; ivar++) {
        par_ptr->set<int>("reaction-diffusion solver: healthy domain initial field "+std::to_string(ivar)+" function") =
            get_par_comp<int>(inp, "reaction_diffusion_solver/healthy_domain/initial_field_function", ivar);
    }
    //
    ASSERT_(get_par_size<Real>(inp, "reaction_diffusion_solver/healthy_domain/initial_field") == nvars,
            "number of (healthy domain) initial values for reaction-diffusion solver is not correct");
    for (unsigned int ivar=0; ivar<nvars; ivar++) {
        par_ptr->set<Real>("reaction-diffusion solver: healthy domain initial field "+std::to_string(ivar)) =
            get_par_comp<Real>(inp, "reaction_diffusion_solver/healthy_domain/initial_field", ivar);
    }
    //
    ASSERT_(get_par_size<Real>(inp, "reaction_diffusion_solver/healthy_domain/initial_field_deviation") == nvars,
            "number of (healthy domain) initial values for deviation for reaction-diffusion solver is not correct");
    for (unsigned int ivar=0; ivar<nvars; ivar++) {
        par_ptr->set<Real>("reaction-diffusion solver: healthy domain initial field "+std::to_string(ivar)+" deviation") =
            get_par_comp<Real>(inp, "reaction_diffusion_solver/healthy_domain/initial_field_deviation", ivar);
    }
    //
    ASSERT_(get_par_size<int>(inp, "reaction_diffusion_solver/damaged_domain/initial_field_function") == nvars,
            "number of functions for (damaged domain) initial values for reaction-diffusion solver is not correct");
    for (unsigned int ivar=0; ivar<nvars; ivar++) {
        par_ptr->set<int>("reaction-diffusion solver: damaged domain initial field "+std::to_string(ivar)+" function") =
            get_par_comp<int>(inp, "reaction_diffusion_solver/damaged_domain/initial_field_function", ivar);
    }
    //
    ASSERT_(get_par_size<Real>(inp, "reaction_diffusion_solver/damaged_domain/initial_field") == nvars,
            "number of (damaged domain) initial values for reaction-diffusion solver is not correct");
    for (unsigned int ivar=0; ivar<nvars; ivar++) {
        par_ptr->set<Real>("reaction-diffusion solver: damaged domain initial field "+std::to_string(ivar)) =
            get_par_comp<Real>(inp, "reaction_diffusion_solver/damaged_domain/initial_field", ivar);
    }
    //
    ASSERT_(get_par_size<Real>(inp, "reaction_diffusion_solver/damaged_domain/initial_field_deviation") == nvars,
            "number of (damaged domain) initial values for deviation for reaction-diffusion solver is not correct");
    for (unsigned int ivar=0; ivar<nvars; ivar++) {
        par_ptr->set<Real>("reaction-diffusion solver: damaged domain initial field "+std::to_string(ivar)+" deviation") =
            get_par_comp<Real>(inp, "reaction_diffusion_solver/damaged_domain/initial_field_deviation", ivar);
    }
    //
    ASSERT_(get_par_size<Real>(inp, "reaction_diffusion_solver/interface/healthy_prior_to_damaged_initial_field") == nvars,
            "number of (healthy/damaged domain) initial condition at interface for reaction-diffusion solver is not correct");
    for (unsigned int ivar=0; ivar<nvars; ivar++) {
        par_ptr->set<bool>("reaction-diffusion solver: healthy/damaged domain initial field "+std::to_string(ivar)+" at interface") =
            get_par_comp<int>(inp, "reaction_diffusion_solver/interface/healthy_prior_to_damaged_initial_field", ivar);
    }
    //
    ASSERT_(get_par_size<int>(inp, "reaction_diffusion_solver/constrained_field_type") == nvars,
            "number of constraint-type of the field values for reaction-diffusion solver is not correct");
    for (unsigned int ivar=0; ivar<nvars; ivar++) {
        par_ptr->set<int>("reaction-diffusion solver: constrained field "+std::to_string(ivar)+" type") =
            get_par_comp<int>(inp, "reaction_diffusion_solver/constrained_field_type", ivar);
    }
    //
    // identify if to execute explicit or implicit time-integration for the solver
    par_ptr->set<bool>("reaction-diffusion solver: explicit") =
        get_par<bool>(inp, "reaction_diffusion_solver/explicit", true);
    //
    #ifdef BIODYNAMO_ENABLED
    // identify if to employ BioDynaMo's agent-based modelling platform
    par_ptr->set<bool>("reaction-diffusion solver: BioDynaMo") =
        get_par<bool>(inp, "reaction_diffusion_solver/biodynamo", false);
    // now provide explicitly (if BioDynaMo is used) details about the agent-based model
    if ( par_ptr->get<bool>("reaction-diffusion solver: BioDynaMo") ) {
        // now read the parameters from the input file
        read_BDM_parameters(inp, in_pars, *par_ptr);
    }
    #else
    // definitely we are not using BioDynaMo in FEB3
    par_ptr->set<bool>("reaction-diffusion solver: BioDynaMo") = false;
    #endif
    //
    // some important checks
    if        ( par_ptr->get<ProblemCase>("solver") == FEM_WOUND_HEALING_SOLVER ) {
        //
        if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension") == 5 ) {
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 1,
                     "number of discontinuous variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables") == 0,
                     "number of auxiliary variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<bool>("reaction-diffusion solver: explicit"),
                     "type of time-integration scheme for reaction-diffusion solver is not valid" );
        } else if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension") == 6 ) {
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 1,
                     "number of discontinuous variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables") == 0,
                     "number of auxiliary variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<bool>("reaction-diffusion solver: explicit"),
                     "type of time-integration scheme for reaction-diffusion solver is not valid" );
        } else {
            ERROR_("dimension inserted for reaction-diffusion solver is not valid" );
        }
        //
    } else if ( par_ptr->get<ProblemCase>("solver") == FEM_BIOCHEMICS_SOLVER ) {
        //
        if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension") == 13 ) {
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 1,
                     "number of discontinuous variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables") == 2,
                     "number of auxiliary variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<bool>("reaction-diffusion solver: explicit"),
                     "type of time-integration scheme for reaction-diffusion solver is not valid" );
        } else {
            ERROR_("dimension inserted for reaction-diffusion solver is not valid" );
        }
        //
    } else if ( par_ptr->get<ProblemCase>("solver") == FEM_TUMOUR_ANGIOGENESIS_SOLVER ) {
        //
        if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension") == 4 ) {
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 9,
                     "number of discontinuous variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables") == 8,
                     "number of auxiliary variables inserted for reaction-diffusion solver is not valid" );
        } else {
            ERROR_("dimension inserted for reaction-diffusion solver is not valid" );
        }
        //
    } else if ( par_ptr->get<ProblemCase>("solver") == FEM_INVITRO_TUMOROID_SOLVER ) {
        //
        if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension") == 7 ) {
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 1,
                     "number of discontinuous variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables") == 1,
                     "number of auxiliary variables inserted for reaction-diffusion solver is not valid" );
        } else if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension") == 8 ) {
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 1,
                     "number of discontinuous variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables") == 3,
                     "number of auxiliary variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<bool>("reaction-diffusion solver: explicit"),
                     "type of time-integration scheme for reaction-diffusion solver is not valid" );
        } else if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension") == 10 ) {
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 1,
                     "number of discontinuous variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables") == 5,
                     "number of auxiliary variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<bool>("reaction-diffusion solver: explicit"),
                     "type of time-integration scheme for reaction-diffusion solver is not valid" );
        } else if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension") == 45 ) {
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 9,
                     "number of discontinuous variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables") == 6,
                     "number of auxiliary variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<bool>("reaction-diffusion solver: explicit"),
                     "type of time-integration scheme for reaction-diffusion solver is not valid" );
        } else {
            ERROR_("dimension inserted for reaction-diffusion solver is not valid" );
        }
        //
    } else if ( par_ptr->get<ProblemCase>("solver") == FEM_RT_TUMOUR_FIBROSIS_SOLVER ) {
        //
        if ( par_ptr->get<unsigned int>("reaction-diffusion solver: dimension") == 3 ) {
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: auxiliary variables") == 1,
                     "number of auxiliary variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 1,
                     "number of discontinuous variables inserted for reaction-diffusion solver is not valid" );
            ASSERT_( par_ptr->get<bool>("reaction-diffusion solver: explicit"),
                     "type of time-integration scheme for reaction-diffusion solver is not valid" );
        } else {
            ERROR_("dimension inserted for reaction-diffusion solver is not valid" );
        }
        //
    } else libmesh_error();
    //
    if ( ndcvars ) {
        //
        ASSERT_(get_par_size<Real>(inp, "reaction_diffusion_solver/discontinuous_initial_field") == ndcvars,
                "number of discontinuous (variables) initial values for reaction-diffusion solver is not correct");
        for (unsigned int ivar=0; ivar<ndcvars; ivar++) {
            par_ptr->set<Real>("reaction-diffusion solver: discontinuous initial field "+std::to_string(ivar)) =
                get_par_comp<Real>(inp, "reaction_diffusion_solver/discontinuous_initial_field", ivar);
        }
        //
        ASSERT_(get_par_size<Real>(inp, "reaction_diffusion_solver/discontinuous_initial_field_deviation") == ndcvars,
                "number of discontinuous (variables) initial values for deviation for reaction-diffusion solver is not correct");
        for (unsigned int ivar=0; ivar<ndcvars; ivar++) {
            par_ptr->set<Real>("reaction-diffusion solver: discontinuous initial field "+std::to_string(ivar)+" deviation") =
                get_par_comp<Real>(inp, "reaction_diffusion_solver/discontinuous_initial_field_deviation", ivar);
        }
        //
    }
    //
    // check if to read the initial nodal field values (primary or/and auxiliary) from an ASCII file
    par_ptr->set<std::string>("reaction-diffusion solver: predefined initial nodal field (file)") =
        get_par<std::string>(inp, "reaction_diffusion_solver/predefined_initial_nodal_field/file", "NA");
    // check if to read the initial elemental field values (discontinuous only!) from an ASCII file
    par_ptr->set<std::string>("reaction-diffusion solver: predefined initial elemental field (file)") =
        get_par<std::string>(inp, "reaction_diffusion_solver/predefined_initial_elemental_field/file", "NA");
    //
    // identify if to enforce the solution for each of the state-variables @post-process
    par_ptr->set<bool>("reaction-diffusion solver: enforce solution") =
        get_par<bool>(inp, "reaction_diffusion_solver/enforce_solution", false);
}
//=================================================================================================
static
void input_wound_healing (GetPot& inp) {
    par_ptr->set<bool>("wound healing solver: apply wound closure") =
        (bool) get_par_size<int>(inp, "wound_healing_solver/wound_closure/BC_surfaces");
    //
    par_ptr->set<int>("wound healing solver: wound closure: number of BC surfaces") =
        get_par_size<int>(inp, "wound_healing_solver/wound_closure/BC_surfaces");
    //
    const int nsurf = par_ptr->get<int>("wound healing solver: wound closure: number of BC surfaces");
    for (int isurf=0; isurf<nsurf; isurf++) {
        char field_name[buffer_len];
        sprintf(field_name, "wound healing solver: wound closure: BC surface %d", isurf);
        par_ptr->set<boundary_id_type>(field_name) =
            (boundary_id_type) get_par_comp<int>(inp, "wound_healing_solver/wound_closure/BC_surfaces", isurf);
    }
    //
    par_ptr->set<int>("wound healing solver: solution increment") =
        get_par<int>(inp, "wound_healing_solver/solution_increment");
    par_ptr->set<int>("wound healing solver: solid solver increment") =
        get_par<int>(inp, "wound_healing_solver/solid_solver_increment");
    //
    par_ptr->set<bool>("wound healing solver: apply gravity (upright)") =
        (bool) get_par_size<Real>(inp, "wound_healing_solver/apply_gravity_upright");
    //
    if ( par_ptr->get<bool>("wound healing solver: apply gravity (upright)") ) {
        par_ptr->set<Point>("wound healing solver: apply gravity (upright): direction") =
            Point( get_par_comp<Real>(inp, "wound_healing_solver/apply_gravity_upright", 0),
                   get_par_comp<Real>(inp, "wound_healing_solver/apply_gravity_upright", 1),
                   get_par_comp<Real>(inp, "wound_healing_solver/apply_gravity_upright", 2) );
    }
}
//=================================================================================================
static
void input_tumour_angiogenesis (GetPot& inp) {
    par_ptr->set<std::string>("tumour angiogenesis solver: label") =
        get_par<std::string>(inp, "tumour_angiogenesis_solver/label", "TA3D");
    //
    par_ptr->set<std::string>("drug delivery solver: label") =
        get_par<std::string>(inp, "drug_delivery_simulation/label", "DDS3D");
    //
    par_ptr->set<int>("tumour angiogenesis solver: solution increment") =
        get_par<int>(inp, "tumour_angiogenesis_solver/solution_increment");
    par_ptr->set<int>("tumour angiogenesis solver: biochemics increment") =
        get_par<int>(inp, "tumour_angiogenesis_solver/biochemics_increment");
    par_ptr->set<int>("tumour angiogenesis solver: biomechanics increment") =
        get_par<int>(inp, "tumour_angiogenesis_solver/biomechanics_increment");
    par_ptr->set<int>("tumour angiogenesis solver: angiogenesis increment") =
        get_par<int>(inp, "tumour_angiogenesis_solver/angiogenesis_increment");
    par_ptr->set<int>("tumour angiogenesis solver: restart increment") =
        get_par<int>(inp, "tumour_angiogenesis_solver/restart_increment");
    par_ptr->set<int>("tumour angiogenesis solver: drug delivery increment") =
        get_par<int>(inp, "tumour_angiogenesis_solver/drug_delivery_increment");
    par_ptr->set<int>("tumour angiogenesis solver: drug biochemics increment") =
        get_par<int>(inp, "tumour_angiogenesis_solver/drug_biochemics_increment");
    //
    par_ptr->set<Real>("tumour angiogenesis solver: time step") =
        get_par<Real>(inp, "tumour_angiogenesis_solver/time_step");
    par_ptr->set<int>("tumour angiogenesis solver: maximum time steps") =
        get_par<int>(inp, "tumour_angiogenesis_solver/maximum_time_steps");
    //
    if ( par_ptr->get<unsigned int>("input restart from file") ) {
        par_ptr->set<int>("tumour angiogenesis solver: initial time") =
            get_par<int>(inp, "tumour_angiogenesis_solver/initial_time");
        par_ptr->set<Real>("current time") =
            par_ptr->get<int>("tumour angiogenesis solver: initial time") *
            par_ptr->get<Real>("tumour angiogenesis solver: time step");
        // some validation checks
        const int ini_time = par_ptr->get<int>("tumour angiogenesis solver: initial time"),
                  max_time = par_ptr->get<int>("tumour angiogenesis solver: maximum time steps");
        ASSERT_(ini_time>=0, "erroneous value inserted for the initial time");
        ASSERT_(ini_time<max_time, "erroneous value inserted for the initial time");
    } else {
        par_ptr->set<int>("tumour angiogenesis solver: initial time") = 0;
        par_ptr->set<Real>("current time") = 0.0;
        // some validation checks
        const int ini_time = get_par<int>(inp, "tumour_angiogenesis_solver/initial_time", 0),
                  max_time = par_ptr->get<int>("tumour angiogenesis solver: maximum time steps");
        ASSERT_(ini_time==0, "erroneous value inserted for the initial time");
        ASSERT_(ini_time<max_time, "erroneous value inserted for the initial time");
    }
    //
    par_ptr->set<Real>("reaction-diffusion solver: time step") =
        par_ptr->get<Real>("tumour angiogenesis solver: time step") *
        par_ptr->get<int>("tumour angiogenesis solver: biochemics increment");
    par_ptr->set<int>("reaction-diffusion solver: maximum time steps") =
        par_ptr->get<int>("tumour angiogenesis solver: maximum time steps") /
        par_ptr->get<int>("tumour angiogenesis solver: biochemics increment");
    par_ptr->set<int>("reaction-diffusion solver: initial time") =
        par_ptr->get<int>("tumour angiogenesis solver: initial time");
    //
    par_ptr->set<Real>("drug biochemics solver: time step") =
        par_ptr->get<Real>("tumour angiogenesis solver: time step") *
        par_ptr->get<int>("tumour angiogenesis solver: drug biochemics increment");
    par_ptr->set<int>("drug biochemics solver: maximum time steps") =
        par_ptr->get<int>("tumour angiogenesis solver: maximum time steps") /
        par_ptr->get<int>("tumour angiogenesis solver: drug biochemics increment");
    par_ptr->set<int>("drug biochemics solver: initial time") =
        par_ptr->get<int>("tumour angiogenesis solver: initial time");
    //
    par_ptr->set<Real>("drug delivery solver: time step") =
        par_ptr->get<Real>("tumour angiogenesis solver: time step") *
        par_ptr->get<int>("tumour angiogenesis solver: drug delivery increment");
    par_ptr->set<int>("drug delivery solver: maximum time steps") =
        par_ptr->get<int>("tumour angiogenesis solver: maximum time steps") /
        par_ptr->get<int>("tumour angiogenesis solver: drug delivery increment");
    //
    par_ptr->set<int>("tumour angiogenesis solver: vascular remodelling material ID") =
        get_par<int>(inp, "tumour_angiogenesis_solver/vascular_remodelling/material_ID");
    //
    // check if to skip renumbering the nodes and elements (segments) of the vascular network
    if ( "NA" == par_ptr->get<std::string>("input vascular network domain data file") ) {
        par_ptr->set<bool>("input vascular network domain data: skip renumbering") =
            get_par<bool>(inp, "tumour_angiogenesis_solver/input_vascular_network_domain_data/skip_renumbering", false);
        if ( 0.0==par_ptr->get<Real>("current time") &&
             par_ptr->get<bool>("input vascular network domain data: skip renumbering") )
            WARNING_("check if indeed you need to skip renumbering the vascular mesh");
    } else {
        par_ptr->set<bool>("input vascular network domain data: skip renumbering") = true;
        if ( 0.0!=par_ptr->get<Real>("current time") &&
             ! par_ptr->get<bool>("input vascular network domain data: skip renumbering") )
            ERROR_("check if indeed you need to renumber the vascular mesh");
    }
    //
    // check if to skip evaluating the branching IDs of the vascular network
    par_ptr->set<bool>("input vascular network domain data: skip branching") =
        get_par<bool>(inp, "tumour_angiogenesis_solver/input_vascular_network_domain_data/skip_branching", false);
    //
    // check if to fix the maximum value of the oxygen level in presence of perfused vessels
    par_ptr->set<bool>("fix tissue 02 maximum level") =
        get_par<bool>(inp, "tumour_angiogenesis_solver/tissue_oxygen_maximum_level/fix", true);
    // ...and provide a value for the maximum oxygen level (default =1)
    par_ptr->set<Real>("fix tissue 02 maximum level: value") = 1.0;
    if ( par_ptr->get<bool>("fix tissue 02 maximum level") ) {
        par_ptr->set<Real>("fix tissue 02 maximum level: value") =
            get_par<Real>(inp, "tissue_oxygen_maximum_level/fix/value", 1.0);
    }
}
//=================================================================================================
void input (EquationSystems& es, DomainData& dmn, GetPot& inp) {
    // set this local pointer to the global system parameters list
    par_ptr = &es.parameters;
    // read all the basic important information about the analysis
    input_preamble(es.get_mesh(), inp);
    // input the information about the boundary surfaces & their BCs
    es.parameters.set<bool>("Navier-Cauchy solver: have follower forces") =
        input_BC_surfaces(dmn, inp);
    // ...and now for boundary nodes as well
    input_BC_nodes(dmn, inp);
    // read solver-specific data
    if ( NO_SOLVER != par_ptr->get<ProblemCase>("solver") ) {
        input_navier_cauchy_solver(inp);
        input_reaction_diffusion_solver(inp);
        // ...and problem-specific data
        if ( FEM_WOUND_HEALING_SOLVER == par_ptr->get<ProblemCase>("solver") )
            input_wound_healing(inp);
        if ( FEM_TUMOUR_ANGIOGENESIS_SOLVER == par_ptr->get<ProblemCase>("solver") )
            input_tumour_angiogenesis(inp);
    }
    // read the material parameters
    input_materials(dmn, inp);
    // read the region data and assign material type
    input_regions(dmn, inp);
    // read if to apply body or nodal force
    input_body_force(dmn, inp);
    input_nodal_force(dmn, inp);
    input_nodal_force_info(dmn, inp);
    // print on screen all unidentified variables provided in the input file
    std::vector<std::string> nonvars = inp.unidentified_variables(in_pars);
    for (unsigned int i=0; i<nonvars.size(); i++) {
        std::stringstream msg;
        msg << "unidentified variable \"" << nonvars[i] << "\" in your input file" << std::flush;
        WARNING_(msg.str());
    }
    // now print into the log file(s) the arguments contained into the input data file
    flog.print(" Input File arguments: \n", true);
    for (unsigned int i=0; i<inp.size(); i++) {
        char inp_arg[buffer_len];
        sprintf(inp_arg, "  %s\n", inp.get(i, " "));
        flog.print(inp_arg, true);
    }
    flog.print("\n\n", true);
    flog.print(" >> Input file read complete\n", false);
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // nullify this pointer now
    par_ptr = 0;
}
//=================================================================================================
