/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "libmesh/petsc_macro.h"
#include "petsc.h"
#include "libmesh/getpot.h"
#include "libmesh/libmesh_logging.h"
#include "feb3/invitro_tumoroid_solver.h"
#include "feb3/reaction_diffusion_solver_Xd.h"
#include "feb3/navier_cauchy_solver_3d.h"
#include "feb3/navier_cauchy_iga_solver_3d.h"
#include "feb3/explicit_navier_cauchy_solver_3d.h"
#include "feb3/wound_healing_solver.h"
#include "feb3/tumour_angiogenesis_solver.h"
#include "feb3/rt_tumour_fibrosis_solver.h"
#include "libmesh/mesh.h"
#include "feb3/mesh_iga.h"
#include "feb3/rve.h"
#include "feb3/paraview_io.h"
//=================================================================================================
// global variables
OFStream flog, ferr; // log and error files to keep track of things...
std::string tmp_dir;
GlobalTolerance tol;
// perfomance log for BioDynaMo only
PerfLog bdm_log("Coupling_with_BioDynaMo");
// communicators between processes for MPI to work
libMesh::Parallel::Communicator* gcomm_ptr;
libMesh::Parallel::Communicator* lcomm_ptr;
// list of all systems used in the analysis
std::vector< std::pair<EquationSystems*, std::string> > systems_ls;
//=================================================================================================
// external routines
extern
void input (EquationSystems& , DomainData& , GetPot& );
extern
void run_a_test (const char* arg);
extern
void run_a_utility (const char* arg);
//=================================================================================================
static
void helper () {
    std::cout << std::endl;
    std::cout << "\tHelper: " << std::endl;
    std::cout << "\t\tNeed to write this up soon ... \n" << std::endl;
    std::cout << std::endl;
}
//=================================================================================================
int main (int argc, char* argv[]) {
    LibMeshInit init(argc, argv);
    // arguments and input file parser objects
    GetPot cmdl(argc, argv);
    // declare a local communicator (for parallel computing)
    Parallel::Communicator local_comm(MPI_COMM_SELF);
    // set the pointer to the global & local communicator
    gcomm_ptr = &init.comm();
    lcomm_ptr = &local_comm;
    // check if we plan to display the help message, run a test problem or run a utility function
    std::string input_file = "input.dat";
    if ( cmdl.search(1, "--help") ) {
        helper();
        return 0;
    } else if ( cmdl.search(1, "-t") ) {
        run_a_test(cmdl.next("0"));
        return 0;
    } else if ( cmdl.search(1, "-u") ) {
        run_a_utility(cmdl.next("0"));
        return 0;
    } else if ( cmdl.search(1, "-i") ) {
        input_file = cmdl.next("input.dat");
    } else libmesh_error();
    //
    GetPot inp(input_file);
    // mesh, systems, domain and Paraview data: Macroscopic domain
    Mesh msh(init.comm(), SP3D);
    EquationSystems es(msh);
    DomainData dmn(es);
    // create a list of the IGA parametric mesh(es)
    std::map<std::string, MeshIGA> msh_iga;
    es.parameters.set<std::map<std::string, MeshIGA>*>("pointer to IGA parametric mesh(es)") = &msh_iga;
    // create a list of the RVE parametric mesh(es)
    std::map<std::string, MeshRVE> msh_rve;
    es.parameters.set<std::map<std::string, MeshRVE>*>("pointer to RVE parametric mesh(es)") = &msh_rve;
    // create a list of the RVE equation system(s)
    std::map<std::string, EquationSystems> es_rve;
    es.parameters.set<std::map<std::string, EquationSystems>*>("pointer to RVE equation system(s)") = &es_rve;
    // mesh, systems, domain and Paraview data: Vascular Network
    Mesh msh_vn(init.comm(), SP1D);
    EquationSystems es_vn(msh_vn);
    DomainData_VascularNetwork dmn_vn(es_vn, es);
    // mesh, systems, domain: Coupled Problem
    Mesh msh_coupled(init.comm(), SP1D);
    EquationSystems es_coupled(msh_coupled);
    DomainData_SolidCoupledWithVascularNetwork dmn_coupled(es_coupled, es);
    // read all necessary input data
    input(es, dmn, inp);
    if ( libMesh::global_processor_id() == 0 ) {
        std::string cmdl_str = " >> Command line:";
        for (int i=0; i<argc; i++)
            cmdl_str += " " + std::string(argv[i]);
        flog.print(cmdl_str+"\n", false);
    }
    // enforce PETSc solver type
    #ifndef LIBMESH_HAVE_PETSC
    #   error "FEB3 requires the PETSc library to operate!"
    #else
    #   if PETSC_VERSION_LESS_THAN(3,7,0)
    PetscOptionsSetValue("-ksp_type", "preonly");
    PetscOptionsSetValue("-pc_type", "lu");
    PetscOptionsSetValue("-pc_factor_mat_solver_package", "superlu_dist");
    // PetscOptionsSetValue("-pc_factor_mat_solver_package", "mumps");
    #   else
    PetscOptionsSetValue(NULL, "-ksp_type", "preonly");
    PetscOptionsSetValue(NULL, "-pc_type", "lu");
    PetscOptionsSetValue(NULL, "-pc_factor_mat_solver_package", "superlu_dist");
    // PetscOptionsSetValue(NULL, "-pc_factor_mat_solver_package", "mumps");
    #   endif
    #endif
    // manipulate the finite element 3D grid information
    dmn.prepare_mesh();
    // obtain access to the parameters list
    const Parameters& par = es.parameters;
    // distinguish for various solver cases
    if ( FEM_BIOMECHANICS_SOLVER == par.get<ProblemCase>("solver") ) {
        // check if we run implicit or explicit implementation
        if ( par.get<bool>("Navier-Cauchy solver: explicit") ) {
            // declare the solver
            ExplicitNavierCauchySolver3D encs(dmn);
            // initialise the systems and the domain data
            encs.init();
            es.init();
            es.print_info();
            es.print_info(flog.stream());
            dmn.init();
            // now run the solver
            encs.run();
            // store info into log
            flog.print(encs.get_log().c_str(), true);
        } else {
            // declare the solver
            NavierCauchySolver3D ncs(dmn);
            // initialise the systems and the domain data
            ncs.init();
            es.init();
            es.print_info();
            es.print_info(flog.stream());
            dmn.init();
            // now run the solver
            ncs.run();
            // store info into log
            flog.print(ncs.get_log().c_str(), true);
        }
    } else if ( IGA_BIOMECHANICS_SOLVER == par.get<ProblemCase>("solver") ) {
        const unsigned int ns = par.get<int>("sampling number for NURBS");
        // label to the IGA parametric mesh
        const std::string label =
            par.get<std::string>("IGA parametric mesh to use for analysis");
        // IGA mesh used for the present analysis
        MeshIGA& msh_iga_a = msh_iga.find(label)->second;
        // a ghost mesh (not for analysis) to plot out the IGA results
        Mesh msh_post(init.comm(), SP3D);
        EquationSystems es_post(msh_post);
        msh_iga_a.set_post_data(&es_post, ns, ns, ns);
        // check if we run implicit or explicit implementation
        if ( par.get<bool>("Navier-Cauchy solver: explicit") ) {
            libmesh_error();
        } else {
            // declare the solver
            NavierCauchyIGASolver3D ncs_iga(dmn);
            // initialise the systems and the domain data
            ncs_iga.init();
            es.init();
            es.print_info();
            es.print_info(flog.stream());
            dmn.init();
            // now run the solver
            ncs_iga.run();
            // store info into log
            flog.print(ncs_iga.get_log().c_str(), true);
        }
    } else if ( FEM_BIOCHEMICS_SOLVER == par.get<ProblemCase>("solver") ) {
        // declare the solver
        ReactionDiffusionSolverXD rds(dmn);
        // initialise the systems and the domain data
        rds.init();
        es.init();
        es.print_info();
        es.print_info(flog.stream());
        dmn.init();
        // now run the solver
        rds.run();
        // store info into log
        flog.print(rds.get_log().c_str(), true);
    } else if ( FEM_WOUND_HEALING_SOLVER == par.get<ProblemCase>("solver") ) {
        // declare the solver
        WoundHealingSolver whs(dmn);
        // initialise the systems and the domain data
        whs.init();
        es.init();
        es.print_info();
        es.print_info(flog.stream());
        dmn.init();
        // now run the solver
        whs.run();
        // store info into log
        flog.print(whs.get_log().c_str(), true);
    } else if ( FEM_TUMOUR_ANGIOGENESIS_SOLVER == par.get<ProblemCase>("solver") ) {
        // declare the solver
        TumourAngiogenesisSolver tas(dmn, dmn_vn, dmn_coupled);
        // manipulate the finite element 3D grid information
        dmn_vn.prepare_mesh();
        // initialise the systems and the domain data
        tas.init();
        es.init();
        es.print_info();
        es.print_info(flog.stream());
        es_vn.init();
        es_vn.print_info();
        es_vn.print_info(flog.stream());
        dmn.init();
        dmn_vn.init();
        // manipulate the finite element 3D grid information
        dmn_coupled.prepare_mesh(dmn_vn);
        // initialise the coupled system and the domain data
        es_coupled.init();
        es_coupled.print_info();
        es_coupled.print_info(flog.stream());
        dmn_coupled.init(dmn_vn);
        // now run the solver
        tas.run();
        // store info into log
        flog.print(tas.get_log().c_str(), true);
    } else if ( FEM_INVITRO_TUMOROID_SOLVER == par.get<ProblemCase>("solver") ) {
        // declare the solver
        InVitroTumoroidSolver ivts(dmn);
        // initialise the systems and the domain data
        ivts.init();
        es.init();
        es.print_info();
        es.print_info(flog.stream());
        dmn.init();
        // now run the solver
        ivts.run();
        // store info into log
        flog.print(ivts.get_log().c_str(), true);
    } else if ( FEM_RT_TUMOUR_FIBROSIS_SOLVER == par.get<ProblemCase>("solver") ) {
        // declare the solver
        RtTumourFibrosisSolver rttfs(dmn);
        // initialise the systems and the domain data
        rttfs.init();
        es.init();
        es.print_info();
        es.print_info(flog.stream());
        dmn.init();
        // now run the solver
        rttfs.run();
        // store info into log
        flog.print(rttfs.get_log().c_str(), true);
    } else {
        char foutVTU[buffer_len];
        sprintf( foutVTU, "%s/paraview/%s.vtu",
                 par.get<std::string>("results directory").c_str(),
                 par.get<std::string>("output file name").c_str() );
        ParaviewIO(dmn.get_mesh()).write_equation_systems(foutVTU, es);
    }
    // store info into log
    flog.print(bdm_log.get_log().c_str(), true);
    flog.print(libMesh::perflog.get_log().c_str(), true);
    if ( par.get<bool>("automatically delete error file") ) ferr.deleteme();
    // ...and clear some memory too
    systems_ls.clear();
    gcomm_ptr = lcomm_ptr = 0;
    // exit program normally
    return 0;
}
//=================================================================================================
