/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/domain_data.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
//=================================================================================================
namespace healthy_explicit {
//=================================================================================================
#ifdef _LOCAL_ARGUMENTS_3D_
    #error "an unexpected error occurred!"
#else
    #define _LOCAL_ARGUMENTS_3D_ \
        const MaterialInfo& mat, \
        const FieldInfo& fld,    \
        Dyadic3D& S2pk,          \
        Quindic3D& Dtsm_uu,      \
        Dyadic3D& Dtsm_up,       \
        Real& Ph,                \
        Dyadic3D& Dtsm_pu,       \
        Real& Dtsm_pp
#endif
//=================================================================================================
#define J1          ibi.J1
#define J2          ibi.J2
#define J3          ibi.J3
#define J4          ibi.J4
#define dJ1_dC      ibi.dJ1_dC
#define dJ2_dC      ibi.dJ2_dC
#define dJ3_dC      ibi.dJ3_dC
#define dJ4_dC      ibi.dJ4_dC
//=================================================================================================
// W = alpha_1 * (I1-3) + beta_1 * (I1-3)^2 ...
//   + alpha_2 * (I2-3) + beta_2 * (I2-3)^2 ...
//   + gamma_1 * (I1-3) * (I2-3) ...
//   + kappa_0 * (J3-1)^2 + kappa_1 * (J3^2-1) - lambda_1 * log(J3) + lambda_2 * (log(J3))^2 ...
//   + ksi4_0 * (J4-3) + ksi4_1 * (J4-3)^2 ...
//=================================================================================================
void polynomial_3d_forward_std ( _LOCAL_ARGUMENTS_3D_ ) {
    const Real& alpha_1 = mat.get_biomech_param(2), beta_1 = mat.get_biomech_param(3),
                alpha_2 = mat.get_biomech_param(4), beta_2 = mat.get_biomech_param(5),
                gamma_1 = mat.get_biomech_param(6),
                kappa_0  = mat.get_biomech_param(7), kappa_1  = mat.get_biomech_param(8),
                lambda_1 = mat.get_biomech_param(9), lambda_2 = mat.get_biomech_param(10);
    const Real& ksi4_0   = mat.get_biomech_param(11), ksi4_1  = mat.get_biomech_param(12);
    //
    const InvariantsBasicInfo ibi(right_CauchyGreen_deformation(fld.deform_grad_elas), fld.fibre_ref);
    //
    const Real dW_dJ1 = alpha_1*pow(J3,frac_2o3)+2.0*beta_1*pow(J3,frac_4o3)*J1-6.0*beta_1*pow(J3,frac_2o3)+gamma_1*J3*J3*J2-3.0*gamma_1*pow(J3,frac_2o3);
    const Real dW_dJ2 = alpha_2*pow(J3,frac_4o3)+2.0*beta_2*pow(J3,frac_8o3)*J2-6.0*beta_2*pow(J3,frac_4o3)+gamma_1*J3*J3*J1-3.0*gamma_1*pow(J3,frac_4o3);
    const Real dW_dJ3 = 2.0*kappa_1*J3+(2.0*alpha_1*J1*J3+4.0*beta_1*J1*J1*pow(J3,frac_5o3)-12.0*beta_1*J1*J3+4.0*alpha_2*J2*pow(J3,frac_5o3)+8.0*beta_2*J2*J2*J3*J3*J3-24.0*beta_2*J2*pow(J3,frac_5o3)+6.0*gamma_1*J1*pow(J3,frac_7o3)*J2-6.0*gamma_1*J1*J3-12.0*gamma_1*J2*pow(J3,frac_5o3)+6.0*kappa_0*pow(J3,frac_7o3)-6.0*kappa_0*pow(J3,frac_4o3)-3.0*lambda_1*pow(J3,frac_1o3)+6.0*lambda_2*log(J3)*pow(J3,frac_1o3))/pow(J3,frac_4o3)*frac_1o3;
    const Real dW_dJ4 = ksi4_0+2.0*ksi4_1*(J4-3.0);
    //
    blitz::firstIndex  i;
    blitz::secondIndex j;
    //
    S2pk.comp += (2.0*dW_dJ1)*dJ1_dC.comp(i,j)
               + (2.0*dW_dJ2)*dJ2_dC.comp(i,j)
               + (2.0*dW_dJ3)*dJ3_dC.comp(i,j)
               + (2.0*dW_dJ4)*dJ4_dC.comp(i,j);
}
//=================================================================================================
// W = alpha_1 * (J1-3) + beta_1 * (J1-3)^2 ...
//   + alpha_2 * (J2-3) + beta_2 * (J2-3)^2 ...
//   + gamma_1 * (J1-3) * (J2-3) ...
//   + kappa_0 * (J3-1)^2 + kappa_1 * (J3^2-1) - lambda_1 * log(J3) + lambda_2 * (log(J3))^2 ...
//   + ksi4_0 * (J4-3) + ksi4_1 * (J4-3)^2 ...
//=================================================================================================
void polynomial_3d_forward_mod ( _LOCAL_ARGUMENTS_3D_ ) {
    const Real& alpha_1 = mat.get_biomech_param(2), beta_1 = mat.get_biomech_param(3),
                alpha_2 = mat.get_biomech_param(4), beta_2 = mat.get_biomech_param(5),
                gamma_1 = mat.get_biomech_param(6),
                kappa_0  = mat.get_biomech_param(7), kappa_1  = mat.get_biomech_param(8),
                lambda_1 = mat.get_biomech_param(9), lambda_2 = mat.get_biomech_param(10);
    const Real& ksi4_0   = mat.get_biomech_param(11), ksi4_1  = mat.get_biomech_param(12);
    //
    const InvariantsBasicInfo ibi(right_CauchyGreen_deformation(fld.deform_grad_elas), fld.fibre_ref);
    //
    const Real dW_dJ1 = alpha_1+2.0*beta_1*J1-6.0*beta_1+gamma_1*J2-3.0*gamma_1;
    const Real dW_dJ2 = alpha_2+2.0*beta_2*J2-6.0*beta_2+gamma_1*J1-3.0*gamma_1;
    const Real dW_dJ3 = (2.0*kappa_0*J3-2.0*kappa_0+2.0*kappa_1*J3)-(lambda_1-2.0*lambda_2*log(J3))/J3;
    const Real dW_dJ4 = ksi4_0+2.0*ksi4_1*(J4-3.0);
    //
    blitz::firstIndex  i;
    blitz::secondIndex j;
    //
    S2pk.comp += (2.0*dW_dJ1)*dJ1_dC.comp(i,j)
               + (2.0*dW_dJ2)*dJ2_dC.comp(i,j)
               + (2.0*dW_dJ3)*dJ3_dC.comp(i,j)
               + (2.0*dW_dJ4)*dJ4_dC.comp(i,j);
}
//=================================================================================================
// W = alpha_1 * (J1-3) + beta_1 * (J1-3)^2 ...
//   + alpha_2 * (J2-3) + beta_2 * (J2-3)^2 ...
//   + gamma_1 * (J1-3) * (J2-3) ...
//   + kappa_0 * (J3-1)^2 + kappa_1 * (J3^2-1) - lambda_1 * log(J3) + lambda_2 * (log(J3))^2 ...
//   + ksi4_0 * (J4-3) + ksi4_1 * (J4-3)^2 ...
//=================================================================================================
void polynomial_3d_forward_2f ( _LOCAL_ARGUMENTS_3D_ ) {
    libmesh_error();
}
//=================================================================================================
#undef J1
#undef J2
#undef J3
#undef J4
#undef dJ1_dC
#undef dJ2_dC
#undef dJ3_dC
#undef dJ4_dC
//=================================================================================================
#undef _LOCAL_ARGUMENTS_3D_
//=================================================================================================
} // end of namespace healthy_explicit...
//=================================================================================================
