/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/domain_data.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
//=================================================================================================
namespace healthy {
//=================================================================================================
#ifdef _LOCAL_ARGUMENTS_2D_
    #error "an unexpected error occurred!"
#else
    #define _LOCAL_ARGUMENTS_2D_ \
        const MaterialInfo& mat, \
        const FieldInfo& fld,    \
        Dyadic2D& Sc,            \
        Quindic2D& Dtsm
#endif
#ifdef _LOCAL_ARGUMENTS_3D_
    #error "an unexpected error occurred!"
#else
    #define _LOCAL_ARGUMENTS_3D_ \
        const MaterialInfo& mat, \
        const FieldInfo& fld,    \
        Dyadic3D& Sc,            \
        Quindic3D& Dtsm_uu,      \
        Dyadic3D& Dtsm_up,       \
        Real& Ph,                \
        Dyadic3D& Dtsm_pu,       \
        Real& Dtsm_pp
#endif
//=================================================================================================
// strain-energy function defined in healthy::polynomial_2d_forward() routine
//=================================================================================================
void polynomial_2d_inverse ( _LOCAL_ARGUMENTS_2D_ ) {
    const Real& alpha_1 = mat.get_biomech_param(2), beta_1 = mat.get_biomech_param(3),
                alpha_2 = mat.get_biomech_param(4), beta_2 = mat.get_biomech_param(5),
                gamma_1 = mat.get_biomech_param(6),
                alpha_3 = mat.get_biomech_param(7), beta_3 = mat.get_biomech_param(8);
    //
    Vector3D G_[SP2D];
    covariant_basis(fld.grad_space_reference, G_);
    Vector3D g_[SP2D];
    covariant_basis(fld.grad_space_current, g_);
    //
    const Dyadic2D G_cova = metric_tensor(G_), G_cont = inverse(G_cova),
                   g_cova = metric_tensor(g_), g_cont = inverse(g_cova);
    // invariants
    const Real I1 = outer_product(g_cova, G_cont),
               I2 = determinant(g_cova)/determinant(G_cova);
    const Real J = sqrt(I2);
    //
    const Real dW_dI1 = alpha_1+2.0*beta_1*I1-6.0*beta_1+gamma_1*I2-3.0*gamma_1
                      + alpha_3*beta_3*exp(beta_3*(I1-3));
    const Real dW_dI2 = alpha_2+2.0*beta_2*I2-6.0*beta_2+gamma_1*I1-3.0*gamma_1;
    const Real d2W_dI1dI1 = 2.0*beta_1
                          + alpha_3*pow2(beta_3)*exp(beta_3*(I1-3));
    const Real d2W_dI1dI2 = gamma_1;
    const Real d2W_dI2dI1 = gamma_1;
    const Real d2W_dI2dI2 = 2.0*beta_2;
    //
    blitz::firstIndex  i;
    blitz::secondIndex j;
    blitz::thirdIndex  k;
    blitz::fourthIndex l;
    //
    Dyadic2D dI1_dG,
             dI2_dG;
    dI1_dG.comp = -0.5*mul_mi_mn_nj(G_cont, g_cova, G_cont).comp(i,j)
                -  0.5*mul_mj_mn_ni(G_cont, g_cova, G_cont).comp(i,j);
    dI2_dG.comp = -I2*G_cont.comp(i,j);
    //
    Dyadic2D dI1_dg,
             dI2_dg;
    dI1_dg.comp = G_cont.comp(i,j);
    dI2_dg.comp = I2*g_cont.comp(i,j);
    //
    Quindic2D d2I1_dgdG,
              d2I2_dgdG;
    d2I1_dgdG.comp = -0.5*G_cont.comp(i,k)*G_cont.comp(j,l)
                   -  0.5*G_cont.comp(i,l)*G_cont.comp(j,k);
    d2I2_dgdG.comp = g_cont.comp(i,j)*dI2_dG.comp(k,l);
    //
    Dyadic2D d2W_dI1dG,
             d2W_dI2dG;
    d2W_dI1dG.comp = d2W_dI1dI1*dI1_dG.comp(i,j) + d2W_dI1dI2*dI2_dG.comp(i,j);
    d2W_dI2dG.comp = d2W_dI2dI1*dI1_dG.comp(i,j) + d2W_dI2dI2*dI2_dG.comp(i,j);
    //
    Sc.comp += (2.0/J)*dW_dI1*dI1_dg.comp(i,j)
             + (2.0/J)*dW_dI2*dI2_dg.comp(i,j);
    //
    Dtsm.comp = (4.0/J)*d2W_dI1dG.comp(k,l)*dI1_dg.comp(i,j) + (4.0/J)*dW_dI1*d2I1_dgdG.comp(i,j,k,l)
              + (4.0/J)*d2W_dI2dG.comp(k,l)*dI2_dg.comp(i,j) + (4.0/J)*dW_dI2*d2I2_dgdG.comp(i,j,k,l)
              + Sc.comp(i,j)*G_cont.comp(k,l);
}
//=================================================================================================
#define I    identity_matrix
#define I_s  identity_tensor
#define c    inv_inf.C
#define ci   inv_inf.Ci
#define c2   inv_inf.C2
#define ci_s inv_inf.Ci_s
#define n0_n0  inv_inf.N_N[0]
#define n1_n1  inv_inf.N_N[1]
#define n2_n2  inv_inf.N_N[2]
#define i1   inv_inf.I1
#define i2   inv_inf.I2
#define i3   inv_inf.I3
#define i4   inv_inf.I4
#define i6   inv_inf.I6
#define i8   inv_inf.I8
#define di1_dc   inv_inf.dI1_dC
#define di2_dc   inv_inf.dI2_dC
#define di3_dc   inv_inf.dI3_dC
#define di4_dc   inv_inf.dI4_dC
#define di6_dc   inv_inf.dI6_dC
#define di8_dc   inv_inf.dI8_dC
#define J1   inv_inf_FWD.J1
#define J2   inv_inf_FWD.J2
#define J3   inv_inf_FWD.J3
#define J4   inv_inf_FWD.J4
#define J6   inv_inf_FWD.J6
#define J8   inv_inf_FWD.J8
//=================================================================================================
// strain-energy function defined in healthy::polynomial_3d_forward_std() routine
//=================================================================================================
void polynomial_3d_inverse_std ( _LOCAL_ARGUMENTS_3D_ ) {
    const Real& alpha_1 = mat.get_biomech_param(2), beta_1 = mat.get_biomech_param(3),
                alpha_2 = mat.get_biomech_param(4), beta_2 = mat.get_biomech_param(5),
                gamma_1 = mat.get_biomech_param(6),
                kappa_0  = mat.get_biomech_param(7), kappa_1  = mat.get_biomech_param(8),
                lambda_1 = mat.get_biomech_param(9), lambda_2 = mat.get_biomech_param(10);
    //
    const InvariantsInfo inv_inf(right_CauchyGreen_deformation(fld.deform_grad_elas), fld.fibre_ref);
    const InvariantsInfo inv_inf_FWD(inv_inf.Ci, fld.fibre_curr);
    //
    const Real dW_dJ1 = alpha_1*pow(J3,frac_2o3)+2.0*beta_1*pow(J3,frac_4o3)*J1-6.0*beta_1*pow(J3,frac_2o3)+gamma_1*J3*J3*J2-3.0*gamma_1*pow(J3,frac_2o3);
    const Real dW_dJ2 = alpha_2*pow(J3,frac_4o3)+2.0*beta_2*pow(J3,frac_8o3)*J2-6.0*beta_2*pow(J3,frac_4o3)+gamma_1*J3*J3*J1-3.0*gamma_1*pow(J3,frac_4o3);
    const Real dW_dJ3 = 2.0*kappa_1*J3+(2.0*alpha_1*J1*J3+4.0*beta_1*J1*J1*pow(J3,frac_5o3)-12.0*beta_1*J1*J3+4.0*alpha_2*J2*pow(J3,frac_5o3)+8.0*beta_2*J2*J2*J3*J3*J3-24.0*beta_2*J2*pow(J3,frac_5o3)+6.0*gamma_1*J1*pow(J3,frac_7o3)*J2-6.0*gamma_1*J1*J3-12.0*gamma_1*J2*pow(J3,frac_5o3)+6.0*kappa_0*pow(J3,frac_7o3)-6.0*kappa_0*pow(J3,frac_4o3)-3.0*lambda_1*pow(J3,frac_1o3)+6.0*lambda_2*log(J3)*pow(J3,frac_1o3))/pow(J3,frac_4o3)*frac_1o3;
    const Real d2W_dJ1dJ1 = 2.0*beta_1*pow(J3,frac_4o3);
    const Real d2W_dJ1dJ2 = gamma_1*J3*J3;
    const Real d2W_dJ1dJ3 = frac_2o3*(alpha_1+4.0*beta_1*J1*pow(J3,frac_2o3)-6.0*beta_1+3.0*gamma_1*J2*pow(J3,frac_4o3)-3.0*gamma_1)/pow(J3,frac_1o3);
    const Real d2W_dJ2dJ1 = gamma_1*J3*J3;
    const Real d2W_dJ2dJ2 = 2.0*beta_2*pow(J3,frac_8o3);
    const Real d2W_dJ2dJ3 = frac_4o3*alpha_2*pow(J3,frac_1o3)+frac_16o3*beta_2*J2*pow(J3,frac_5o3)-8.0*beta_2*pow(J3,frac_1o3)+2.0*gamma_1*J1*J3-4.0*gamma_1*pow(J3,frac_1o3);
    const Real d2W_dJ3dJ1 = frac_2o3*(alpha_1+4.0*beta_1*J1*pow(J3,frac_2o3)-6.0*beta_1+3.0*gamma_1*J2*pow(J3,frac_4o3)-3.0*gamma_1)/pow(J3,frac_1o3);
    const Real d2W_dJ3dJ2 = frac_4o3*alpha_2*pow(J3,frac_1o3)+frac_16o3*beta_2*J2*pow(J3,frac_5o3)-8.0*beta_2*pow(J3,frac_1o3)+2.0*gamma_1*J1*J3-4.0*gamma_1*pow(J3,frac_1o3);
    const Real d2W_dJ3dJ3 = 2.0*kappa_1-1.0/pow(J3,frac_7o3)*(2.0*alpha_1*J1*J3-4.0*beta_1*J1*J1*pow(J3,frac_5o3)-12.0*beta_1*J1*J3-4.0*alpha_2*J2*pow(J3,frac_5o3)-40.0*beta_2*J2*J2*J3*J3*J3+24.0*beta_2*J2*pow(J3,frac_5o3)-18.0*gamma_1*J1*pow(J3,frac_7o3)*J2-6.0*gamma_1*J1*J3+12.0*gamma_1*J2*pow(J3,frac_5o3)-18.0*kappa_0*pow(J3,frac_7o3)-9.0*lambda_1*pow(J3,frac_1o3)-18.0*lambda_2*pow(J3,frac_1o3)+18.0*lambda_2*log(J3)*pow(J3,frac_1o3))*frac_1o9;
    //
    blitz::firstIndex  i;
    blitz::secondIndex j;
    blitz::thirdIndex  k;
    blitz::fourthIndex l;
    //
    Dyadic3D dI1_dc, dJ1_dc,
             dI2_dc, dJ2_dc,
             dI3_dc, dJ3_dc;
    dI1_dc.comp = (1.0/i3)*di2_dc.comp(i,j) - (i2/pow2(i3))*di3_dc.comp(i,j);
    dI2_dc.comp = (1.0/i3)*di1_dc.comp(i,j) - (i1/pow2(i3))*di3_dc.comp(i,j);
    dI3_dc.comp = (-1.0/pow2(i3))*di3_dc.comp(i,j);
    dJ1_dc.comp = pow(i3, frac_1o3) * dI1_dc.comp - frac_1o3*i2*pow(i3, frac_1o3)*dI3_dc.comp(i,j);
    dJ2_dc.comp = pow(i3, frac_2o3) * dI2_dc.comp - frac_2o3*i1*pow(i3, frac_2o3)*dI3_dc.comp(i,j);
    dJ3_dc.comp = (frac_1o2*pow(i3, frac_1o2))*dI3_dc.comp(i,j);
    //
    Dyadic3D BJ3_dJ1_dB, // (1/J3) * B(i,s) dJ1_dB(s,j)
             BJ3_dJ2_dB, // (1/J3) * B(i,s) dJ2_dB(s,j)
             BJ3_dJ3_dB; // (1/J3) * B(i,s) dJ3_dB(s,j)
    BJ3_dJ1_dB.comp = pow(i3, frac_5o6)*ci.comp(i,j)
                    - (frac_1o3*i2*pow(i3, -frac_1o6))*I.comp(i,j);
    BJ3_dJ2_dB.comp = (i1*pow(i3, frac_1o6))*I.comp(i,j)
                    - pow(i3, frac_1o6)*c.comp(i,j)
                    - (frac_2o3*i1*pow(i3, frac_1o6))*I.comp(i,j);
    BJ3_dJ3_dB.comp = frac_1o2*I.comp(i,j);
    //
    Quindic3D BJ3_dJ1_dBdc, // d[(1/J3) * B(i,s) dJ1_dB(s,j)]_dc(k,l)
              BJ3_dJ2_dBdc, // d[(1/J3) * B(i,s) dJ2_dB(s,j)]_dc(k,l)
              BJ3_dJ3_dBdc; // d[(1/J3) * B(i,s) dJ1_dB(s,j)]_dc(k,l)
    BJ3_dJ1_dBdc.comp = (frac_5o6*pow(i3, -frac_1o6))*ci.comp(i,j)*di3_dc.comp(k,l)
                      + pow(i3, frac_5o6)*ci_s.comp(i,j,k,l)
                      - (frac_1o3*pow(i3, -frac_1o6))*I.comp(i,j)*di2_dc.comp(k,l)
                      + (frac_1o18*i2*pow(i3, -frac_7o6))*I.comp(i,j)*di3_dc.comp(k,l);
    BJ3_dJ2_dBdc.comp = pow(i3, frac_1o6)*I.comp(i,j)*di1_dc.comp(k,l)
                      + (frac_1o6*i1*pow(i3, -frac_5o6))*I.comp(i,j)*di3_dc.comp(k,l)
                      - (frac_1o6*pow(i3, -frac_5o6))*c.comp(i,j)*di3_dc.comp(k,l)
                      - pow(i3, frac_1o6)*I_s.comp(i,j,k,l)
                      - (frac_2o3*pow(i3, frac_1o6))*I.comp(i,j)*di1_dc.comp(k,l)
                      - (frac_2o18*i1*pow(i3, -frac_5o6))*I.comp(i,j)*di3_dc.comp(k,l);
    BJ3_dJ3_dBdc.comp = 0.0;
    //
    Dyadic3D d2W_dJ1dc,
             d2W_dJ2dc,
             d2W_dJ3dc;
    d2W_dJ1dc.comp = d2W_dJ1dJ1*dJ1_dc.comp(i,j) + d2W_dJ1dJ2*dJ2_dc.comp(i,j) + d2W_dJ1dJ3*dJ3_dc.comp(i,j);
    d2W_dJ2dc.comp = d2W_dJ2dJ1*dJ1_dc.comp(i,j) + d2W_dJ2dJ2*dJ2_dc.comp(i,j) + d2W_dJ2dJ3*dJ3_dc.comp(i,j);
    d2W_dJ3dc.comp = d2W_dJ3dJ1*dJ1_dc.comp(i,j) + d2W_dJ3dJ2*dJ2_dc.comp(i,j) + d2W_dJ3dJ3*dJ3_dc.comp(i,j);
    //
    Sc.comp += (2.0*dW_dJ1)*BJ3_dJ1_dB.comp(i,j)
             + (2.0*dW_dJ2)*BJ3_dJ2_dB.comp(i,j)
             + (2.0*dW_dJ3)*BJ3_dJ3_dB.comp(i,j);
    //
    Dtsm_uu.comp = 4.0*d2W_dJ1dc.comp(k,l)*BJ3_dJ1_dB.comp(i,j) + (4.0*dW_dJ1)*BJ3_dJ1_dBdc.comp(i,j,k,l)
                 + 4.0*d2W_dJ2dc.comp(k,l)*BJ3_dJ2_dB.comp(i,j) + (4.0*dW_dJ2)*BJ3_dJ2_dBdc.comp(i,j,k,l)
                 + 4.0*d2W_dJ3dc.comp(k,l)*BJ3_dJ3_dB.comp(i,j) + (4.0*dW_dJ3)*BJ3_dJ3_dBdc.comp(i,j,k,l);
    //
    if ( ! are_equal(fld.deform_grad_inel, identity_matrix) ) {
        libmesh_error();
    }
}
//=================================================================================================
// strain-energy function defined in healthy::polynomial_3d_forward_mod() routine
//=================================================================================================
void polynomial_3d_inverse_mod ( _LOCAL_ARGUMENTS_3D_ ) {
    const Real& alpha_1 = mat.get_biomech_param(2), beta_1 = mat.get_biomech_param(3),
                alpha_2 = mat.get_biomech_param(4), beta_2 = mat.get_biomech_param(5),
                gamma_1 = mat.get_biomech_param(6),
                kappa_0  = mat.get_biomech_param(7), kappa_1  = mat.get_biomech_param(8),
                lambda_1 = mat.get_biomech_param(9), lambda_2 = mat.get_biomech_param(10);
    const Real& ksi4_0   = mat.get_biomech_param(11), ksi4_1  = mat.get_biomech_param(12),
                ksi6_0   = mat.get_biomech_param(13), ksi6_1  = mat.get_biomech_param(14),
                ksi8_0   = mat.get_biomech_param(15), ksi8_1  = mat.get_biomech_param(16),
                delta1_0 = mat.get_biomech_param(17), delta1_1 = mat.get_biomech_param(18),
                delta2_0 = mat.get_biomech_param(19), delta2_1 = mat.get_biomech_param(20),
                epsilon4_0 = mat.get_biomech_param(21), epsilon4_1 = mat.get_biomech_param(22),
                epsilon6_0 = mat.get_biomech_param(23), epsilon6_1 = mat.get_biomech_param(24),
                epsilon8_0 = mat.get_biomech_param(25), epsilon8_1 = mat.get_biomech_param(26);
    //
    const InvariantsInfo inv_inf(right_CauchyGreen_deformation(fld.deform_grad_elas), fld.fibre_ref);
    const InvariantsInfo inv_inf_FWD(inv_inf.Ci, fld.fibre_curr);
    //
    const Real dW_dJ1 = alpha_1+2.0*beta_1*J1-6.0*beta_1+gamma_1*J2-3.0*gamma_1+delta1_0*delta1_1*exp(delta1_1*(J1-3.0));
    const Real dW_dJ2 = alpha_2+2.0*beta_2*J2-6.0*beta_2+gamma_1*J1-3.0*gamma_1+delta2_0*delta2_1*exp(delta2_1*(J2-3.0));
    const Real dW_dJ3 = (2.0*kappa_0*J3-2.0*kappa_0+2.0*kappa_1*J3)-(lambda_1-2.0*lambda_2*log(J3))/J3;
    const Real d2W_dJ1dJ1 = 2.0*beta_1+delta1_0*delta1_1*delta1_1*exp(delta1_1*(J1-3.0));
    const Real d2W_dJ1dJ2 = gamma_1;
    const Real d2W_dJ2dJ1 = gamma_1;
    const Real d2W_dJ2dJ2 = 2.0*beta_2+delta2_0*delta2_1*delta2_1*exp(delta2_1*(J2-3.0));
    const Real d2W_dJ3dJ3 = (2.0*kappa_0+2.0*kappa_1)+(lambda_1+2.0*lambda_2-2.0*lambda_2*log(J3))/(J3*J3);
    const Real dW_dJ4 = ksi4_0+2.0*ksi4_1*(J4-3.0)+2.0*epsilon4_0*epsilon4_1*(J4-1.0)*exp(epsilon4_1*pow2(J4-1.0));
    const Real dW_dJ6 = ksi6_0+2.0*ksi6_1*(J6-3.0)+2.0*epsilon6_0*epsilon6_1*(J6-1.0)*exp(epsilon6_1*pow2(J6-1.0));
    const Real dW_dJ8 = ksi8_0+2.0*ksi8_1*(J8-3.0)+2.0*epsilon8_0*epsilon8_1*(J8-1.0)*exp(epsilon8_1*pow2(J8-1.0));
    const Real d2W_dJ4dJ4 = 2.0*ksi4_1+2.0*epsilon4_0*epsilon4_1*exp(epsilon4_1*pow2(J4-1.0))+4.0*epsilon4_0*pow2(epsilon4_1*(J4-1.0))*exp(epsilon4_1*pow2(J4-1.0));
    const Real d2W_dJ6dJ6 = 2.0*ksi6_1+2.0*epsilon6_0*epsilon6_1*exp(epsilon6_1*pow2(J6-1.0))+4.0*epsilon6_0*pow2(epsilon6_1*(J6-1.0))*exp(epsilon6_1*pow2(J6-1.0));
    const Real d2W_dJ8dJ8 = 2.0*ksi8_1+2.0*epsilon8_0*epsilon8_1*exp(epsilon8_1*pow2(J8-1.0))+4.0*epsilon8_0*pow2(epsilon8_1*(J8-1.0))*exp(epsilon8_1*pow2(J8-1.0));
    //
    blitz::firstIndex  i;
    blitz::secondIndex j;
    blitz::thirdIndex  k;
    blitz::fourthIndex l;
    //
    // I1 = i2/i3; I2 = i1/i3; I3 = 1/i3; I4 = 1/i4; I5 = i5/i4; I6 = 1/i6; I7 = i7/i6; I8 = 1/i8; I9 = i9/i8
    Dyadic3D dI1_dc, dJ1_dc,
             dI2_dc, dJ2_dc,
             dI3_dc, dJ3_dc;
    dI1_dc.comp = (1.0/i3)*di2_dc.comp(i,j) - (i2/pow2(i3))*di3_dc.comp(i,j);
    dI2_dc.comp = (1.0/i3)*di1_dc.comp(i,j) - (i1/pow2(i3))*di3_dc.comp(i,j);
    dI3_dc.comp = (-1.0/pow2(i3))*di3_dc.comp(i,j);
    dJ1_dc.comp = pow(i3, frac_1o3) * dI1_dc.comp - frac_1o3*i2*pow(i3, frac_1o3)*dI3_dc.comp(i,j);
    dJ2_dc.comp = pow(i3, frac_2o3) * dI2_dc.comp - frac_2o3*i1*pow(i3, frac_2o3)*dI3_dc.comp(i,j);
    dJ3_dc.comp = (frac_1o2*pow(i3, frac_1o2))*dI3_dc.comp(i,j);
    Dyadic3D dI4_dc, dJ4_dc,
             dI6_dc, dJ6_dc,
             dI8_dc, dJ8_dc;
    dI4_dc.comp = (-1.0/pow2(i4))*di4_dc.comp(i,j);
    dI6_dc.comp = (-1.0/pow2(i6))*di6_dc.comp(i,j);
    dI8_dc.comp = (-1.0/pow2(i8))*di8_dc.comp(i,j);
    dJ4_dc.comp = (frac_1o3/pow(i3, frac_2o3)/i4)*di3_dc.comp(i,j) - (pow(i3, frac_1o3)/pow2(i4))*di4_dc.comp(i,j);
    dJ6_dc.comp = (frac_1o3/pow(i3, frac_2o3)/i6)*di3_dc.comp(i,j) - (pow(i3, frac_1o3)/pow2(i6))*di6_dc.comp(i,j);
    dJ8_dc.comp = (frac_1o3/pow(i3, frac_2o3)/i8)*di3_dc.comp(i,j) - (pow(i3, frac_1o3)/pow2(i8))*di8_dc.comp(i,j);
    //
    Dyadic3D BJ3_dJ1_dB, // (1/J3) * B(i,s) dJ1_dB(s,j)
             BJ3_dJ2_dB, // (1/J3) * B(i,s) dJ2_dB(s,j)
             BJ3_dJ3_dB; // (1/J3) * B(i,s) dJ3_dB(s,j)
    BJ3_dJ1_dB.comp = pow(i3, frac_5o6)*ci.comp(i,j)
                    - (frac_1o3*i2*pow(i3, -frac_1o6))*I.comp(i,j);
    BJ3_dJ2_dB.comp = (i1*pow(i3, frac_1o6))*I.comp(i,j)
                    - pow(i3, frac_1o6)*c.comp(i,j)
                    - (frac_2o3*i1*pow(i3, frac_1o6))*I.comp(i,j);
    BJ3_dJ3_dB.comp = frac_1o2*I.comp(i,j);
    Dyadic3D BJ3_dJ4_dB, // (1/J3) * B(i,s) dJ4_dB(s,j)
             BJ3_dJ6_dB, // (1/J3) * B(i,s) dJ6_dB(s,j)
             BJ3_dJ8_dB; // (1/J3) * B(i,s) dJ8_dB(s,j)
    BJ3_dJ4_dB.comp = (pow(i3, frac_5o6)/i4)*n0_n0.comp(i,j);
    BJ3_dJ6_dB.comp = (pow(i3, frac_5o6)/i6)*n1_n1.comp(i,j);
    BJ3_dJ8_dB.comp = (pow(i3, frac_5o6)/i8)*n2_n2.comp(i,j);
    //
    Quindic3D BJ3_dJ1_dBdc, // d[(1/J3) * B(i,s) dJ1_dB(s,j)]_dc(k,l)
              BJ3_dJ2_dBdc, // d[(1/J3) * B(i,s) dJ2_dB(s,j)]_dc(k,l)
              BJ3_dJ3_dBdc; // d[(1/J3) * B(i,s) dJ3_dB(s,j)]_dc(k,l)
    BJ3_dJ1_dBdc.comp = (frac_5o6*pow(i3, -frac_1o6))*ci.comp(i,j)*di3_dc.comp(k,l)
                      + pow(i3, frac_5o6)*ci_s.comp(i,j,k,l)
                      - (frac_1o3*pow(i3, -frac_1o6))*I.comp(i,j)*di2_dc.comp(k,l)
                      + (frac_1o18*i2*pow(i3, -frac_7o6))*I.comp(i,j)*di3_dc.comp(k,l);
    BJ3_dJ2_dBdc.comp = pow(i3, frac_1o6)*I.comp(i,j)*di1_dc.comp(k,l)
                      + (frac_1o6*i1*pow(i3, -frac_5o6))*I.comp(i,j)*di3_dc.comp(k,l)
                      - (frac_1o6*pow(i3, -frac_5o6))*c.comp(i,j)*di3_dc.comp(k,l)
                      - pow(i3, frac_1o6)*I_s.comp(i,j,k,l)
                      - (frac_2o3*pow(i3, frac_1o6))*I.comp(i,j)*di1_dc.comp(k,l)
                      - (frac_2o18*i1*pow(i3, -frac_5o6))*I.comp(i,j)*di3_dc.comp(k,l);
    BJ3_dJ3_dBdc.comp = 0.0;
    Quindic3D BJ3_dJ4_dBdc, // d[(1/J3) * B(i,s) dJ4_dB(s,j)]_dc(k,l)
              BJ3_dJ6_dBdc, // d[(1/J3) * B(i,s) dJ6_dB(s,j)]_dc(k,l)
              BJ3_dJ8_dBdc; // d[(1/J3) * B(i,s) dJ8_dB(s,j)]_dc(k,l)
    BJ3_dJ4_dBdc.comp = (frac_5o6/pow(i3, frac_1o6)/i4)*n0_n0.comp(i,j)*di3_dc.comp(k,l)
                      - (pow(i3, frac_5o6)/pow2(i4))*n0_n0.comp(i,j)*di4_dc.comp(k,l);
    BJ3_dJ6_dBdc.comp = (frac_5o6/pow(i3, frac_1o6)/i6)*n1_n1.comp(i,j)*di3_dc.comp(k,l)
                      - (pow(i3, frac_5o6)/pow2(i6))*n1_n1.comp(i,j)*di6_dc.comp(k,l);
    BJ3_dJ8_dBdc.comp = (frac_5o6/pow(i3, frac_1o6)/i8)*n2_n2.comp(i,j)*di3_dc.comp(k,l)
                      - (pow(i3, frac_5o6)/pow2(i8))*n2_n2.comp(i,j)*di8_dc.comp(k,l);
    //
    Dyadic3D d2W_dJ1dc,
             d2W_dJ2dc,
             d2W_dJ3dc;
    d2W_dJ1dc.comp = d2W_dJ1dJ1*dJ1_dc.comp(i,j) + d2W_dJ1dJ2*dJ2_dc.comp(i,j);
    d2W_dJ2dc.comp = d2W_dJ2dJ1*dJ1_dc.comp(i,j) + d2W_dJ2dJ2*dJ2_dc.comp(i,j);
    d2W_dJ3dc.comp = d2W_dJ3dJ3*dJ3_dc.comp(i,j);
    Dyadic3D d2W_dJ4dc,
             d2W_dJ6dc,
             d2W_dJ8dc;
    d2W_dJ4dc.comp = d2W_dJ4dJ4*dJ4_dc.comp(i,j);
    d2W_dJ6dc.comp = d2W_dJ6dJ6*dJ6_dc.comp(i,j);
    d2W_dJ8dc.comp = d2W_dJ8dJ8*dJ8_dc.comp(i,j);
    //
    Sc.comp += (2.0*dW_dJ1)*BJ3_dJ1_dB.comp(i,j)
             + (2.0*dW_dJ2)*BJ3_dJ2_dB.comp(i,j)
             + (2.0*dW_dJ3)*BJ3_dJ3_dB.comp(i,j)
             + (2.0*dW_dJ4)*BJ3_dJ4_dB.comp(i,j)
             + (2.0*dW_dJ6)*BJ3_dJ6_dB.comp(i,j)
             + (2.0*dW_dJ8)*BJ3_dJ8_dB.comp(i,j);
    //
    Dtsm_uu.comp = 4.0*d2W_dJ1dc.comp(k,l)*BJ3_dJ1_dB.comp(i,j) + (4.0*dW_dJ1)*BJ3_dJ1_dBdc.comp(i,j,k,l)
                 + 4.0*d2W_dJ2dc.comp(k,l)*BJ3_dJ2_dB.comp(i,j) + (4.0*dW_dJ2)*BJ3_dJ2_dBdc.comp(i,j,k,l)
                 + 4.0*d2W_dJ3dc.comp(k,l)*BJ3_dJ3_dB.comp(i,j) + (4.0*dW_dJ3)*BJ3_dJ3_dBdc.comp(i,j,k,l)
                 + 4.0*d2W_dJ4dc.comp(k,l)*BJ3_dJ4_dB.comp(i,j) + (4.0*dW_dJ4)*BJ3_dJ4_dBdc.comp(i,j,k,l)
                 + 4.0*d2W_dJ6dc.comp(k,l)*BJ3_dJ6_dB.comp(i,j) + (4.0*dW_dJ6)*BJ3_dJ6_dBdc.comp(i,j,k,l)
                 + 4.0*d2W_dJ8dc.comp(k,l)*BJ3_dJ8_dB.comp(i,j) + (4.0*dW_dJ8)*BJ3_dJ8_dBdc.comp(i,j,k,l);
    //
    if ( ! are_equal(fld.deform_grad_inel, identity_matrix) ) {
        libmesh_error();
    }
}
//=================================================================================================
// strain-energy function defined in healthy::polynomial_3d_forward_2f() routine
//=================================================================================================
void polynomial_3d_inverse_2f ( _LOCAL_ARGUMENTS_3D_ ) {
    const Real& alpha_1 = mat.get_biomech_param(2), beta_1 = mat.get_biomech_param(3),
                alpha_2 = mat.get_biomech_param(4), beta_2 = mat.get_biomech_param(5),
                gamma_1 = mat.get_biomech_param(6),
                kappa_0  = mat.get_biomech_param(7), kappa_1  = mat.get_biomech_param(8),
                lambda_1 = mat.get_biomech_param(9), lambda_2 = mat.get_biomech_param(10);
    const Real& ksi4_0   = mat.get_biomech_param(11), ksi4_1  = mat.get_biomech_param(12),
                ksi6_0   = mat.get_biomech_param(13), ksi6_1  = mat.get_biomech_param(14),
                ksi8_0   = mat.get_biomech_param(15), ksi8_1  = mat.get_biomech_param(16),
                delta1_0 = mat.get_biomech_param(17), delta1_1 = mat.get_biomech_param(18),
                delta2_0 = mat.get_biomech_param(19), delta2_1 = mat.get_biomech_param(20),
                epsilon4_0 = mat.get_biomech_param(21), epsilon4_1 = mat.get_biomech_param(22),
                epsilon6_0 = mat.get_biomech_param(23), epsilon6_1 = mat.get_biomech_param(24),
                epsilon8_0 = mat.get_biomech_param(25), epsilon8_1 = mat.get_biomech_param(26);
    //
    const InvariantsInfo inv_inf(right_CauchyGreen_deformation(fld.deform_grad_elas), fld.fibre_ref);
    const InvariantsInfo inv_inf_FWD(inv_inf.Ci, fld.fibre_curr);
    //
    const Real& p_t = fld.pressure;
    //
    const Real dW_dJ1 = alpha_1+2.0*beta_1*J1-6.0*beta_1+gamma_1*J2-3.0*gamma_1+delta1_0*delta1_1*exp(delta1_1*(J1-3.0));
    const Real dW_dJ2 = alpha_2+2.0*beta_2*J2-6.0*beta_2+gamma_1*J1-3.0*gamma_1+delta2_0*delta2_1*exp(delta2_1*(J2-3.0));
    const Real dW_dJ3 = (2.0*kappa_0*J3-2.0*kappa_0+2.0*kappa_1*J3)-(lambda_1-2.0*lambda_2*log(J3))/J3;
    const Real d2W_dJ1dJ1 = 2.0*beta_1+delta1_0*delta1_1*delta1_1*exp(delta1_1*(J1-3.0));
    const Real d2W_dJ1dJ2 = gamma_1;
    const Real d2W_dJ2dJ1 = gamma_1;
    const Real d2W_dJ2dJ2 = 2.0*beta_2+delta2_0*delta2_1*delta2_1*exp(delta2_1*(J2-3.0));
    const Real d2W_dJ3dJ3 = (2.0*kappa_0+2.0*kappa_1)+(lambda_1+2.0*lambda_2-2.0*lambda_2*log(J3))/(J3*J3);
    const Real dW_dJ4 = ksi4_0+2.0*ksi4_1*(J4-3.0)+2.0*epsilon4_0*epsilon4_1*(J4-1.0)*exp(epsilon4_1*pow2(J4-1.0));
    const Real dW_dJ6 = ksi6_0+2.0*ksi6_1*(J6-3.0)+2.0*epsilon6_0*epsilon6_1*(J6-1.0)*exp(epsilon6_1*pow2(J6-1.0));
    const Real dW_dJ8 = ksi8_0+2.0*ksi8_1*(J8-3.0)+2.0*epsilon8_0*epsilon8_1*(J8-1.0)*exp(epsilon8_1*pow2(J8-1.0));
    const Real d2W_dJ4dJ4 = 2.0*ksi4_1+2.0*epsilon4_0*epsilon4_1*exp(epsilon4_1*pow2(J4-1.0))+4.0*epsilon4_0*pow2(epsilon4_1*(J4-1.0))*exp(epsilon4_1*pow2(J4-1.0));
    const Real d2W_dJ6dJ6 = 2.0*ksi6_1+2.0*epsilon6_0*epsilon6_1*exp(epsilon6_1*pow2(J6-1.0))+4.0*epsilon6_0*pow2(epsilon6_1*(J6-1.0))*exp(epsilon6_1*pow2(J6-1.0));
    const Real d2W_dJ8dJ8 = 2.0*ksi8_1+2.0*epsilon8_0*epsilon8_1*exp(epsilon8_1*pow2(J8-1.0))+4.0*epsilon8_0*pow2(epsilon8_1*(J8-1.0))*exp(epsilon8_1*pow2(J8-1.0));
    //
    const Real d3W_dJ3dJ3dJ3 = -2.0*(lambda_1+3.0*lambda_2-2.0*lambda_2*log(J3))/(J3*J3*J3);
    const Real d4W_dJ3dJ3dJ3dJ3 = 2.0*(3.0*lambda_1+11.0*lambda_2-6.0*lambda_2*log(J3))/(J3*J3*J3*J3);
    //
    blitz::firstIndex  i;
    blitz::secondIndex j;
    blitz::thirdIndex  k;
    blitz::fourthIndex l;
    //
    // I1 = i2/i3; I2 = i1/i3; I3 = 1/i3; I4 = 1/i4; I5 = i5/i4; I6 = 1/i6; I7 = i7/i6; I8 = 1/i8; I9 = i9/i8
    Dyadic3D dI1_dc, dJ1_dc,
             dI2_dc, dJ2_dc,
             dI3_dc, dJ3_dc;
    dI1_dc.comp = (1.0/i3)*di2_dc.comp(i,j) - (i2/pow2(i3))*di3_dc.comp(i,j);
    dI2_dc.comp = (1.0/i3)*di1_dc.comp(i,j) - (i1/pow2(i3))*di3_dc.comp(i,j);
    dI3_dc.comp = (-1.0/pow2(i3))*di3_dc.comp(i,j);
    dJ1_dc.comp = pow(i3, frac_1o3) * dI1_dc.comp - frac_1o3*i2*pow(i3, frac_1o3)*dI3_dc.comp(i,j);
    dJ2_dc.comp = pow(i3, frac_2o3) * dI2_dc.comp - frac_2o3*i1*pow(i3, frac_2o3)*dI3_dc.comp(i,j);
    dJ3_dc.comp = (frac_1o2*pow(i3, frac_1o2))*dI3_dc.comp(i,j);
    Dyadic3D dI4_dc, dJ4_dc,
             dI6_dc, dJ6_dc,
             dI8_dc, dJ8_dc;
    dI4_dc.comp = (-1.0/pow2(i4))*di4_dc.comp(i,j);
    dI6_dc.comp = (-1.0/pow2(i6))*di6_dc.comp(i,j);
    dI8_dc.comp = (-1.0/pow2(i8))*di8_dc.comp(i,j);
    dJ4_dc.comp = (frac_1o3/pow(i3, frac_2o3)/i4)*di3_dc.comp(i,j) - (pow(i3, frac_1o3)/pow2(i4))*di4_dc.comp(i,j);
    dJ6_dc.comp = (frac_1o3/pow(i3, frac_2o3)/i6)*di3_dc.comp(i,j) - (pow(i3, frac_1o3)/pow2(i6))*di6_dc.comp(i,j);
    dJ8_dc.comp = (frac_1o3/pow(i3, frac_2o3)/i8)*di3_dc.comp(i,j) - (pow(i3, frac_1o3)/pow2(i8))*di8_dc.comp(i,j);
    //
    Dyadic3D BJ3_dJ1_dB, // (1/J3) * B(i,s) dJ1_dB(s,j)
             BJ3_dJ2_dB, // (1/J3) * B(i,s) dJ2_dB(s,j)
             BJ3_dJ3_dB; // (1/J3) * B(i,s) dJ3_dB(s,j)
    BJ3_dJ1_dB.comp = pow(i3, frac_5o6)*ci.comp(i,j)
                    - (frac_1o3*i2*pow(i3, -frac_1o6))*I.comp(i,j);
    BJ3_dJ2_dB.comp = (i1*pow(i3, frac_1o6))*I.comp(i,j)
                    - pow(i3, frac_1o6)*c.comp(i,j)
                    - (frac_2o3*i1*pow(i3, frac_1o6))*I.comp(i,j);
    BJ3_dJ3_dB.comp = frac_1o2*I.comp(i,j);
    Dyadic3D BJ3_dJ4_dB, // (1/J3) * B(i,s) dJ4_dB(s,j)
             BJ3_dJ6_dB, // (1/J3) * B(i,s) dJ6_dB(s,j)
             BJ3_dJ8_dB; // (1/J3) * B(i,s) dJ8_dB(s,j)
    BJ3_dJ4_dB.comp = (pow(i3, frac_5o6)/i4)*n0_n0.comp(i,j);
    BJ3_dJ6_dB.comp = (pow(i3, frac_5o6)/i6)*n1_n1.comp(i,j);
    BJ3_dJ8_dB.comp = (pow(i3, frac_5o6)/i8)*n2_n2.comp(i,j);
    //
    Quindic3D BJ3_dJ1_dBdc, // d[(1/J3) * B(i,s) dJ1_dB(s,j)]_dc(k,l)
              BJ3_dJ2_dBdc, // d[(1/J3) * B(i,s) dJ2_dB(s,j)]_dc(k,l)
              BJ3_dJ3_dBdc; // d[(1/J3) * B(i,s) dJ3_dB(s,j)]_dc(k,l)
    BJ3_dJ1_dBdc.comp = (frac_5o6*pow(i3, -frac_1o6))*ci.comp(i,j)*di3_dc.comp(k,l)
                      + pow(i3, frac_5o6)*ci_s.comp(i,j,k,l)
                      - (frac_1o3*pow(i3, -frac_1o6))*I.comp(i,j)*di2_dc.comp(k,l)
                      + (frac_1o18*i2*pow(i3, -frac_7o6))*I.comp(i,j)*di3_dc.comp(k,l);
    BJ3_dJ2_dBdc.comp = pow(i3, frac_1o6)*I.comp(i,j)*di1_dc.comp(k,l)
                      + (frac_1o6*i1*pow(i3, -frac_5o6))*I.comp(i,j)*di3_dc.comp(k,l)
                      - (frac_1o6*pow(i3, -frac_5o6))*c.comp(i,j)*di3_dc.comp(k,l)
                      - pow(i3, frac_1o6)*I_s.comp(i,j,k,l)
                      - (frac_2o3*pow(i3, frac_1o6))*I.comp(i,j)*di1_dc.comp(k,l)
                      - (frac_2o18*i1*pow(i3, -frac_5o6))*I.comp(i,j)*di3_dc.comp(k,l);
    BJ3_dJ3_dBdc.comp = 0.0;
    Quindic3D BJ3_dJ4_dBdc, // d[(1/J3) * B(i,s) dJ4_dB(s,j)]_dc(k,l)
              BJ3_dJ6_dBdc, // d[(1/J3) * B(i,s) dJ6_dB(s,j)]_dc(k,l)
              BJ3_dJ8_dBdc; // d[(1/J3) * B(i,s) dJ8_dB(s,j)]_dc(k,l)
    BJ3_dJ4_dBdc.comp = (frac_5o6/pow(i3, frac_1o6)/i4)*n0_n0.comp(i,j)*di3_dc.comp(k,l)
                      - (pow(i3, frac_5o6)/pow2(i4))*n0_n0.comp(i,j)*di4_dc.comp(k,l);
    BJ3_dJ6_dBdc.comp = (frac_5o6/pow(i3, frac_1o6)/i6)*n1_n1.comp(i,j)*di3_dc.comp(k,l)
                      - (pow(i3, frac_5o6)/pow2(i6))*n1_n1.comp(i,j)*di6_dc.comp(k,l);
    BJ3_dJ8_dBdc.comp = (frac_5o6/pow(i3, frac_1o6)/i8)*n2_n2.comp(i,j)*di3_dc.comp(k,l)
                      - (pow(i3, frac_5o6)/pow2(i8))*n2_n2.comp(i,j)*di8_dc.comp(k,l);
    //
    Dyadic3D d2W_dJ1dc,
             d2W_dJ2dc,
             d2W_dJ3dc;
    d2W_dJ1dc.comp = d2W_dJ1dJ1*dJ1_dc.comp(i,j) + d2W_dJ1dJ2*dJ2_dc.comp(i,j);
    d2W_dJ2dc.comp = d2W_dJ2dJ1*dJ1_dc.comp(i,j) + d2W_dJ2dJ2*dJ2_dc.comp(i,j);
    d2W_dJ3dc.comp = d2W_dJ3dJ3*dJ3_dc.comp(i,j);
    Dyadic3D d2W_dJ4dc,
             d2W_dJ6dc,
             d2W_dJ8dc;
    d2W_dJ4dc.comp = d2W_dJ4dJ4*dJ4_dc.comp(i,j);
    d2W_dJ6dc.comp = d2W_dJ6dJ6*dJ6_dc.comp(i,j);
    d2W_dJ8dc.comp = d2W_dJ8dJ8*dJ8_dc.comp(i,j);
    //
    const Real s_p = -1.0,
               s_Q = +1.0;
    const Real p = s_p*dW_dJ3,
               Q = s_Q*d2W_dJ3dJ3; // dp_dJ
    const Real dp_dJ3 = s_p*d2W_dJ3dJ3,
               dQ_dJ3 = s_Q*d3W_dJ3dJ3dJ3;
    const Real d2p_dJ3dJ3 = s_p*d3W_dJ3dJ3dJ3,
               d2Q_dJ3dJ3 = s_Q*d4W_dJ3dJ3dJ3dJ3;
    //
    Dyadic3D dp_dc,
             dQ_dc;
    dp_dc.comp = dp_dJ3*dJ3_dc.comp(i,j);
    dQ_dc.comp = dQ_dJ3*dJ3_dc.comp(i,j); //###
    //
    Dyadic3D BJ3_dp_dB, // (1/J3) * B(i,s) dp_dB(s,j)
             BJ3_dQ_dB; // (1/J3) * B(i,s) dQ_dB(s,j)
    BJ3_dp_dB.comp = (frac_1o2*dp_dJ3)*I.comp(i,j);
    BJ3_dQ_dB.comp = (frac_1o2*dQ_dJ3)*I.comp(i,j); //###
    //
    Quindic3D BJ3_dp_dBdc, // d[(1/J3) * B(i,s) dp_dB(s,j)]_dc(k,l)
              BJ3_dQ_dBdc; // d[(1/J3) * B(i,s) dQ_dB(s,j)]_dc(k,l)
    BJ3_dp_dBdc.comp = (frac_1o2*d2p_dJ3dJ3)*I.comp(i,j)*dJ3_dc.comp(k,l); //###
    BJ3_dQ_dBdc.comp = (frac_1o2*d2Q_dJ3dJ3)*I.comp(i,j)*dJ3_dc.comp(k,l); //###
    //
    Sc.comp += (2.0*dW_dJ1)*BJ3_dJ1_dB.comp(i,j)
             + (2.0*dW_dJ2)*BJ3_dJ2_dB.comp(i,j)
             + (2.0*dW_dJ3)*BJ3_dJ3_dB.comp(i,j)
             + (2.0*dW_dJ4)*BJ3_dJ4_dB.comp(i,j)
             + (2.0*dW_dJ6)*BJ3_dJ6_dB.comp(i,j)
             + (2.0*dW_dJ8)*BJ3_dJ8_dB.comp(i,j)
             - (2.0*(p-p_t)/Q)*BJ3_dp_dB.comp(i,j)
             + (pow2(p-p_t)/pow2(Q))*BJ3_dQ_dB.comp(i,j);//###
    //
    Ph = (1.0/Q)*(p-p_t);
    //
    Dtsm_uu.comp = 4.0*d2W_dJ1dc.comp(k,l)*BJ3_dJ1_dB.comp(i,j) + (4.0*dW_dJ1)*BJ3_dJ1_dBdc.comp(i,j,k,l)
                 + 4.0*d2W_dJ2dc.comp(k,l)*BJ3_dJ2_dB.comp(i,j) + (4.0*dW_dJ2)*BJ3_dJ2_dBdc.comp(i,j,k,l)
                 + 4.0*d2W_dJ3dc.comp(k,l)*BJ3_dJ3_dB.comp(i,j) + (4.0*dW_dJ3)*BJ3_dJ3_dBdc.comp(i,j,k,l)
                 + 4.0*d2W_dJ4dc.comp(k,l)*BJ3_dJ4_dB.comp(i,j) + (4.0*dW_dJ4)*BJ3_dJ4_dBdc.comp(i,j,k,l)
                 + 4.0*d2W_dJ6dc.comp(k,l)*BJ3_dJ6_dB.comp(i,j) + (4.0*dW_dJ6)*BJ3_dJ6_dBdc.comp(i,j,k,l)
                 + 4.0*d2W_dJ8dc.comp(k,l)*BJ3_dJ8_dB.comp(i,j) + (4.0*dW_dJ8)*BJ3_dJ8_dBdc.comp(i,j,k,l)
                 - (4.0/Q)*BJ3_dp_dB.comp(i,j)*dp_dc.comp(k,l)
                 + (4.0*(p-p_t)/pow2(Q))*(BJ3_dp_dB.comp(i,j)*dQ_dc.comp(k,l)+BJ3_dQ_dB.comp(i,j)*dp_dc.comp(k,l)) //### 1+2
                 - (4.0*(p-p_t)/Q)*BJ3_dp_dBdc.comp(i,j,k,l)//###
                 - (4.0*pow2(p-p_t)/pow3(Q))*BJ3_dQ_dB.comp(i,j)*dQ_dc.comp(k,l)//###
                 + (2.0*pow2(p-p_t)/pow2(Q))*BJ3_dQ_dBdc.comp(i,j,k,l);//###
    //
    Dtsm_up.comp = (2.0/Q)*BJ3_dp_dB.comp(i,j)
                 - (2.0*(p-p_t)/pow2(Q))*BJ3_dQ_dB.comp(i,j);//###
    //
    Dtsm_pu.comp = (2.0/Q)*dp_dc.comp(i,j)
                 - (2.0/pow2(Q))*(p-p_t)*dQ_dc.comp(i,j);//###
    //
    Dtsm_pp = (-1.0/Q);
    //
    if ( ! are_equal(fld.deform_grad_inel, identity_matrix) ) {
        libmesh_error();
    }
}
//=================================================================================================
#undef I
#undef I_s
#undef c
#undef ci
#undef c2
#undef ci_s
#undef n0_n0
#undef n1_n1
#undef n2_n2
#undef i1
#undef i2
#undef i3
#undef i4
#undef i6
#undef i8
#undef di1_dc
#undef di2_dc
#undef di3_dc
#undef di4_dc
#undef di6_dc
#undef di8_dc
#undef J1
#undef J2
#undef J3
#undef J4
#undef J6
#undef J8
//=================================================================================================
#undef _LOCAL_ARGUMENTS_2D_
#undef _LOCAL_ARGUMENTS_3D_
//=================================================================================================
} // end of namespace healthy...
//=================================================================================================
