/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/domain_data.h"
#include "feb3/rve.h"
//=================================================================================================
namespace collagen_extracellular_matrix {
//=================================================================================================
void multiscale_3 ( const MaterialInfo& mat, const FieldInfo& fld,
                    EquationSystems& es_m, std::vector<FieldInfo>& fld_m,
                    Dyadic3D& S2pk, Vector3D& Qs )
{
    const RVE& rve = *(es_m.parameters.get<RVE*>("pointer to RVE mesh"));
    const std::vector<Real>& par = *(es_m.parameters.get<const std::vector<Real>*>("pointer to RVE parameters"));
    // if region does not employ a multiscale implementation then simply exit this routine
    if ( ! par[100] ) return;
    //
    // Jacobian determinant: transformation from element parametric space [Zm,ZM] to RVE space [0,L]
    const Real ZM_Zm___L = par[101] / rve.length();
    for (unsigned int v=0; v<8; v++) {
        // transform the displacement gradient for all RVE vertices
        for (unsigned int ivar=0; ivar<SP3D; ivar++) {
            fld_m[v].grad_displacement(ivar) = transpose(inverse(fld_m[v].jacobian))
                                             * fld_m[v].grad_displacement(ivar)
                                             * ZM_Zm___L;
        }
    }
    //
    // pass the macro-micro field information to the equation systems
    es_m.parameters.set<std::vector<FieldInfo>*>("pointer to macro-micro field information") = &fld_m;
    // solve the solid mechanics problem at the microscale
    collagen_extracellular_matrix::multiscale_3___solve(es_m);
    //
    // update the macroscopic stress tensor
    VectorVoigt3D S2pk_V;
    for (int i=0; i<VOIGT_SP3D; i++) {
        std::stringstream par_name;
        par_name << "macroscopic stress component S" << i << std::flush;
        S2pk_V.comp(i) = es_m.parameters.get<Real>(par_name.str());
    }
    S2pk.comp += uncompress_stress(S2pk_V).comp;
    // update the source term stemming from the multiscale effects
    for (int i=0; i<SP3D; i++) {
        std::stringstream par_name;
        par_name << "multiscale source-term component Q" << i << std::flush;
        Qs.comp(i) = es_m.parameters.get<Real>(par_name.str());
    }
}
//=================================================================================================
} // end of namespace collagen_extracellular_matrix...
//=================================================================================================
//
//
//
//=================================================================================================
#include "libmesh/linear_implicit_system.h"
#include "libmesh/fe.h"
#include "libmesh/elem.h"
#include "libmesh/dof_map.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/numeric_vector.h"
#include "libmesh/dense_matrix.h"
#include "libmesh/dense_vector.h"
#include "libmesh/type_tensor.h"
#include "feb3/ida.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
extern // list of all systems used in the analysis
std::vector< std::pair<EquationSystems*, std::string> > systems_ls;
//=================================================================================================
#ifdef BOOST_COMPUTATIONS
#error "macro parameter \"BOOST_COMPUTATIONS\" has been defined elsewhere!"
#endif
#define BOOST_COMPUTATIONS
// #undef  BOOST_COMPUTATIONS
//=================================================================================================
namespace collagen_extracellular_matrix {
//=================================================================================================
void multiscale_3___solve (EquationSystems& es) {
    #ifndef BOOST_COMPUTATIONS
    //
    const RVE& rve = *(es.parameters.get<RVE*>("pointer to RVE mesh"));
    libmesh_assert( SP1D == rve.mesh_dimension() );
    const std::string sname = es.parameters.get<std::string>("system name");
    //
    // access the basic linear system
    LinearImplicitSystem& sys = es.get_system<LinearImplicitSystem>(sname);
    libmesh_assert( SP3D == sys.n_vars() );
    //
    // access the parameters of the RVE
    const std::vector<Real>& par = *(es.parameters.get<const std::vector<Real>*>("pointer to RVE parameters"));
    // access the RVE field information
    const std::vector<FieldInfo>& fld = *(es.parameters.get<std::vector<FieldInfo>*>("pointer to macro-micro field information"));
    //
    // perform non-affine computations
    if ( 0 == par[102] ) {
        //
        // solve the linear system of equations for the RVE
        sys.solve();
        //
    // perform affine computations
    } else if ( 1 == par[102] ) {
        //
        InverseDistanceAlgorithm ida(es.parameters.get<Real>("IDA exponent"));
        ida.init(rve.vertices());
        // access the displacement solution for all RVE vectices
        std::vector<Real> U0, U1, U2;
        for (unsigned int v=0; v<8; v++) {
            const Displacement& U = fld[v].displacement;
            U0.push_back(U(0));
            U1.push_back(U(1));
            U2.push_back(U(2));
        }
        // loop for all RVE nodes and set the displacement solution accordingly
        for ( MeshBase::const_node_iterator
              ci_node=rve.nodes_begin(); ci_node!=rve.nodes_end(); ci_node++)
        {
            const Node* node = (*ci_node);
            sys.solution->set(node->dof_number(sys.number(), 0, 0), ida.interp<Real>(*node, U0));
            sys.solution->set(node->dof_number(sys.number(), 1, 0), ida.interp<Real>(*node, U1));
            sys.solution->set(node->dof_number(sys.number(), 2, 0), ida.interp<Real>(*node, U2));
        }
        // close the solution container and update the system
        sys.solution->close();
        sys.update();
        //
    // ...an exception is caught here
    } else libmesh_error();
    // now perform the post-processing step
    collagen_extracellular_matrix::multiscale_3___post_process(es);
    //
    #else
    //
    // perform the post-processing step only!
    collagen_extracellular_matrix::multiscale_3___post_process(es);
    //
    #endif
}
//=================================================================================================
void multiscale_3___setup_system (EquationSystems& es) {
    const std::string sname = "RVE_Balance";
    // Lagrange basis for displacement DoFs (default), also choose the order
    const Order lbasis = FIRST;
    // add the main linear system of the multiscale solver
    #ifndef BOOST_COMPUTATIONS
    do {
        LinearImplicitSystem& sys = es.add_system<LinearImplicitSystem>(sname);
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        // add the system variables
        for (unsigned int ivar=0; ivar<SP3D; ivar++)
            sys.add_variable("MS__u"+number2string(ivar), lbasis);
        // attach the assembly routine
        sys.attach_assemble_function(collagen_extracellular_matrix::multiscale_3___system_assembly);
    } while ( false );
    #endif
    // add another system
    do {
        System& sys = es.add_system<System>(sname+"::auxiliary");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        // add the system variables
        for (unsigned int ivar=0; ivar<SP3D; ivar++)
            sys.add_variable("MS__C"+number2string(ivar), lbasis);
        for (unsigned int ivar=0; ivar<SP3D; ivar++)
            sys.add_variable("MS__iHat"+number2string(ivar), lbasis);
        for (unsigned int ivar=0; ivar<SP3D; ivar++)
            sys.add_variable("MS__jHat"+number2string(ivar), lbasis);
        for (unsigned int ivar=0; ivar<SP3D; ivar++)
            sys.add_variable("MS__kHat"+number2string(ivar), lbasis);
        sys.add_variable("MS__L", lbasis);
        sys.add_variable("MS__V", lbasis);
    } while ( false );
    // add another system
    #ifndef BOOST_COMPUTATIONS
    do {
        System& sys = es.add_system<System>(sname+"::stress_strain");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        // add the system variables
        sys.add_variable("MS__stress", CONSTANT, MONOMIAL);
        sys.add_variable("MS__strain", CONSTANT, MONOMIAL);
    } while ( false );
    #endif
    // add another system
    #ifndef BOOST_COMPUTATIONS
    for (unsigned int ivar=0; ivar<SP3D; ivar++) {
        System& sys = es.add_system<System>(sname+"::omega"+number2string(ivar));
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        // add the system variable
        sys.add_variable("MS__w"+number2string(ivar), CONSTANT, MONOMIAL);
    }
    #endif
    // and finally initialise the equation systems
    es.init();
    // now set some default parameters, the rest come from the biomechanical material parameters
    // of the region, also add the end results as IO-parameters
    es.parameters.set<std::string>("system name") = sname;
    es.parameters.set<Real>("system matrix penalty parameter") = 1.0e+5;
    es.parameters.set<Real>("IDA exponent") = 2.5;
    es.parameters.set<Real>("macroscopic stress component S0") = 0.0;
    es.parameters.set<Real>("macroscopic stress component S1") = 0.0;
    es.parameters.set<Real>("macroscopic stress component S2") = 0.0;
    es.parameters.set<Real>("macroscopic stress component S3") = 0.0;
    es.parameters.set<Real>("macroscopic stress component S4") = 0.0;
    es.parameters.set<Real>("macroscopic stress component S5") = 0.0;
    es.parameters.set<Real>("multiscale source-term component Q0") = 0.0;
    es.parameters.set<Real>("multiscale source-term component Q1") = 0.0;
    es.parameters.set<Real>("multiscale source-term component Q2") = 0.0;
    // access the parameters of the RVE
    const std::vector<Real>& par = *(es.parameters.get<const std::vector<Real>*>("pointer to RVE parameters"));
    // access the 1D-mesh of the RVE
    const RVE& rve = *(es.parameters.get<RVE*>("pointer to RVE mesh"));
    //
    // make these checks first
    if ( par.at(100)!=1 ) libmesh_error();
    if ( par.at(101)<=0.000 || par.at(101)>=0.845 ) libmesh_error();
    if ( par.at(102)!=0 && par.at(102)!=1 ) libmesh_error();
    if ( par.at(103)<=0.0 ) libmesh_error();
    if ( par.at(104)<=0.0 ) libmesh_error();
    if ( par.at(105)<=0.0 || par.at(106)<=0.0 ) libmesh_error();
    if ( par.at(107)<=0.0 || par.at(107)>=1.0 ) libmesh_error();
    //
    const Real //length_downscaling_factor = par[103],
               fibril_cross_section_area = par[104],
               fibril_stiffness__tension     = par[105],
               fibril_stiffness__compression = par[106];
               //fibre_volume_fraction = par[107];
    //
    Real K_max = 0.0;
    const MeshBase::const_element_iterator begin_el = rve.active_elements_begin();
    const MeshBase::const_element_iterator end_el = rve.active_elements_end();
    for (MeshBase::const_element_iterator ci_elem=begin_el; ci_elem!=end_el; ci_elem++) {
        const Elem* elem = (*ci_elem);
        libmesh_assert( libMesh::EDGE2==elem->type() );
        //
        const Real L = (elem->point(1)-elem->point(0)).norm(),
                   AE_L = fibril_cross_section_area
                        * std::max(fibril_stiffness__tension, fibril_stiffness__compression)
                        / L;
        K_max = std::max(K_max, AE_L);
    }
    es.parameters.set<Real>("system matrix penalty parameter") *= K_max;
    //
    es.parameters.set<Real>("RVE total fibre length (reference)") = 0.0;
    //
    const Point origin = es.parameters.get<Point>("RVE: origin point space vector");
    const Point orient[] = { es.parameters.get<Point>("RVE: orientation vector #0") ,
                             es.parameters.get<Point>("RVE: orientation vector #1") ,
                             es.parameters.get<Point>("RVE: orientation vector #2") };
    //
    System& aux_sys = es.get_system<System>(sname+"::auxiliary");
    // loop for all RVE nodes and set the solution accordingly
    for ( MeshBase::const_node_iterator
          ci_node=rve.nodes_begin(); ci_node!=rve.nodes_end(); ci_node++)
    {
        const Node* node = (*ci_node);
        //
        aux_sys.solution->set(node->dof_number(aux_sys.number(), 0, 0), origin(0));
        aux_sys.solution->set(node->dof_number(aux_sys.number(), 1, 0), origin(1));
        aux_sys.solution->set(node->dof_number(aux_sys.number(), 2, 0), origin(2));
        aux_sys.solution->set(node->dof_number(aux_sys.number(), 3, 0), orient[0](0));
        aux_sys.solution->set(node->dof_number(aux_sys.number(), 4, 0), orient[0](1));
        aux_sys.solution->set(node->dof_number(aux_sys.number(), 5, 0), orient[0](2));
        aux_sys.solution->set(node->dof_number(aux_sys.number(), 6, 0), orient[1](0));
        aux_sys.solution->set(node->dof_number(aux_sys.number(), 7, 0), orient[1](1));
        aux_sys.solution->set(node->dof_number(aux_sys.number(), 8, 0), orient[1](2));
        aux_sys.solution->set(node->dof_number(aux_sys.number(), 9, 0), orient[2](0));
        aux_sys.solution->set(node->dof_number(aux_sys.number(),10, 0), orient[2](1));
        aux_sys.solution->set(node->dof_number(aux_sys.number(),11, 0), orient[2](2));
        aux_sys.solution->set(node->dof_number(aux_sys.number(),12, 0), rve.length());
        aux_sys.solution->set(node->dof_number(aux_sys.number(),13, 0), rve.volume());
    }
    // close the solution container and update the system
    aux_sys.solution->close();
    aux_sys.update();
    //
    // generate some dummy field information
    std::vector<FieldInfo> fld_m(8, FieldInfo(0, 0, SP3D));
    // pass the macro-micro field information to the equation systems
    es.parameters.set<std::vector<FieldInfo>*>("pointer to macro-micro field information") = &fld_m;
    // now perform the post-processing step
    collagen_extracellular_matrix::multiscale_3___post_process(es);
}
//=================================================================================================
void multiscale_3___system_assembly (EquationSystems& es, const std::string& sname) {
    libmesh_assert( sname == es.parameters.get<std::string>("system name") );
    LinearImplicitSystem& sys = es.get_system<LinearImplicitSystem>(sname);
    libmesh_assert( SP3D == sys.n_vars() );
    //
    // access the 1D-mesh of the RVE
    const RVE& rve = *(es.parameters.get<RVE*>("pointer to RVE mesh"));
    // access the parameters of the RVE
    const std::vector<Real>& par = *(es.parameters.get<const std::vector<Real>*>("pointer to RVE parameters"));
    // access the RVE field information
    const std::vector<FieldInfo>& fld = *(es.parameters.get<std::vector<FieldInfo>*>("pointer to macro-micro field information"));
    //
    const MeshBase::const_element_iterator begin_el = rve.active_elements_begin();
    const MeshBase::const_element_iterator end_el = rve.active_elements_end();
    //
    const unsigned int u_var[] = { sys.variable_number("MS__u"+number2string(0)) ,
                                   sys.variable_number("MS__u"+number2string(1)) ,
                                   sys.variable_number("MS__u"+number2string(2)) };
    //
    const Real //length_downscaling_factor = par[103],
               fibril_cross_section_area = par[104],
               fibril_stiffness__tension     = par[105],
               fibril_stiffness__compression = par[106];
               //fibre_volume_fraction = par[107];
    // penalty parameter to enforce Dirichlet BCs in the system
    const Real penalty = es.parameters.get<Real>("system matrix penalty parameter"),
    // inverse distance algorithm exponent parameter
               ida_exp = es.parameters.get<Real>("IDA exponent");
    // unit direction vectors
    const Point e[] = { Point(1., 0., 0.) , Point(0., 1., 0.) , Point(0., 0., 1.) };
    //
    // zero-out the system matrix and RHS vector
    sys.matrix->zero();
    sys.rhs->zero();
    // loop over all active 1-D finite elements
    for (MeshBase::const_element_iterator ci_elem=begin_el; ci_elem!=end_el; ci_elem++) {
        const Elem* elem = (*ci_elem);
        //
        std::vector<unsigned int> dof_indices;
        sys.get_dof_map().dof_indices(elem, dof_indices);
        //
        std::vector<unsigned int> dof_indices_u[SP3D];
        for (unsigned int ivar=0; ivar<SP3D; ivar++)
            sys.get_dof_map().dof_indices(elem, dof_indices_u[ivar], u_var[ivar]);
        //
        const unsigned long var_dofs[] = { dof_indices_u[0].size() ,
                                           dof_indices_u[1].size() ,
                                           dof_indices_u[2].size() };
        const unsigned long offset[] = { 0                                   ,
                                         var_dofs[0]                         ,
                                         var_dofs[0]+var_dofs[1]             ,
                                         var_dofs[0]+var_dofs[1]+var_dofs[2] };
        //
        DenseMatrix<Number> Ke(dof_indices.size(), dof_indices.size());
        std::vector< DenseSubMatrix<Number> > Kesub(SP3D*SP3D, Ke);
        DenseVector<Number> Fe(dof_indices.size());
        std::vector< DenseSubVector<Number> > Fesub(SP3D, Fe);
        for (unsigned int ivar=0; ivar<SP3D; ivar++) {
            for (unsigned int jvar=0; jvar<SP3D; jvar++)
                Kesub[ivar*SP3D+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
            Fesub[ivar].reposition(offset[ivar], var_dofs[ivar]);
        }
        //
        const Point u[] = { Point( sys.solution->el(elem->node_ref(0).dof_number(sys.number(), 0, 0)) ,
                                   sys.solution->el(elem->node_ref(0).dof_number(sys.number(), 1, 0)) ,
                                   sys.solution->el(elem->node_ref(0).dof_number(sys.number(), 2, 0)) ) ,
                            Point( sys.solution->el(elem->node_ref(1).dof_number(sys.number(), 0, 0)) ,
                                   sys.solution->el(elem->node_ref(1).dof_number(sys.number(), 1, 0)) ,
                                   sys.solution->el(elem->node_ref(1).dof_number(sys.number(), 2, 0)) ) },
                    fibre0 = (elem->point(1)     -elem->point(0)     ),
                    fibreT = (elem->point(1)+u[1]-elem->point(0)-u[0]);
        const Real L0 = fibre0.norm(),
                   LT = fibreT.norm();
        //
        const Real s = LT / L0, // stretch ratio
                   AE_L = fibril_cross_section_area / L0
                        * ( s >= 1.0 ? fibril_stiffness__tension : fibril_stiffness__compression );
        const Real C[] = { (e[0]*fibre0.unit()),  (e[1]*fibre0.unit()),  (e[2]*fibre0.unit()) };
        //
        for (unsigned int a=0; a<var_dofs[0]; a++) {
            for (unsigned int b=0; b<var_dofs[0]; b++) {
                const Real coef = AE_L*( a == b ? +1 : -1 );
                Kesub[0*SP3D+0](a,b) += coef*C[0]*C[0];
                Kesub[0*SP3D+1](a,b) += coef*C[0]*C[1];
                Kesub[0*SP3D+2](a,b) += coef*C[0]*C[2];
                Kesub[1*SP3D+0](a,b) += coef*C[1]*C[0];
                Kesub[1*SP3D+1](a,b) += coef*C[1]*C[1];
                Kesub[1*SP3D+2](a,b) += coef*C[1]*C[2];
                Kesub[2*SP3D+0](a,b) += coef*C[2]*C[0];
                Kesub[2*SP3D+1](a,b) += coef*C[2]*C[1];
                Kesub[2*SP3D+2](a,b) += coef*C[2]*C[2];
            }
        }
        //
        int face = -1;
        for (unsigned int ne=0; ne<elem->n_nodes(); ne++)
        if ( ! elem->neighbor_ptr(ne) ||
             rve.node_is_boundary(elem->node_id(ne), face) ) {
            //
            std::vector<Real> U0, U1, U2;
            for (unsigned int v=0; v<8; v++) {
                const Displacement& U = fld[v].displacement;
                U0.push_back(U(0));
                U1.push_back(U(1));
                U2.push_back(U(2));
            }
            //
            InverseDistanceAlgorithm ida(ida_exp);
            ida.init(rve.vertices());
            //
            Kesub[0*SP3D+0](ne,ne) += penalty;
            Kesub[1*SP3D+1](ne,ne) += penalty;
            Kesub[2*SP3D+2](ne,ne) += penalty;
            Fesub[0](ne) += penalty * ida.interp<Real>(elem->point(ne), U0);
            Fesub[1](ne) += penalty * ida.interp<Real>(elem->point(ne), U1);
            Fesub[2](ne) += penalty * ida.interp<Real>(elem->point(ne), U2);
            // end of element nodes loop...
        }
        // update the system stiffness matrix and RHS vector
        sys.matrix->add_matrix(Ke, dof_indices);
        sys.rhs->add_vector(Fe, dof_indices);
        // end of active 1-D finite elements loop...
    }
    // close the system matrix and RHS vector
    sys.matrix->close();
    sys.rhs->close();
}
//=================================================================================================
void multiscale_3___post_process (EquationSystems& es) {
    const std::string sname = es.parameters.get<std::string>("system name");
    #ifndef BOOST_COMPUTATIONS
    System& ss_sys = es.get_system<System>(sname+"::stress_strain");
    libmesh_assert( 2 == ss_sys.n_vars() );
    #endif
    //
    // access the 1D-mesh of the RVE
    const RVE& rve = *(es.parameters.get<RVE*>("pointer to RVE mesh"));
    // access the parameters of the RVE
    const std::vector<Real>& par = *(es.parameters.get<const std::vector<Real>*>("pointer to RVE parameters"));
    // access the RVE field information
    const std::vector<FieldInfo>& fld = *(es.parameters.get<std::vector<FieldInfo>*>("pointer to macro-micro field information"));
    //
    const MeshBase::const_element_iterator begin_el = rve.active_elements_begin();
    const MeshBase::const_element_iterator end_el = rve.active_elements_end();
    //
    const Real length_downscaling_factor = par[103],
               fibril_cross_section_area = par[104],
               fibril_stiffness__tension     = par[105],
               fibril_stiffness__compression = par[106],
               fibre_volume_fraction = par[107];
    // inverse distance algorithm exponent parameter
    const Real ida_exp = es.parameters.get<Real>("IDA exponent");
    // macroscopic stress tensor
    RealTensorValue S2pk(0., 0., 0.,
                         0., 0., 0.,
                         0., 0., 0.);
    // micro/macro-interaction source term
    RealVectorValue Q_source(0., 0., 0.);
    //
    std::vector<Point> xyz[6]; // NOTE: RVE has 6 faces by definition
    std::vector<RealTensorValue> sigma[6]; // NOTE: RVE has 6 faces by definition
    //
    #ifndef BOOST_COMPUTATIONS
    //
    const unsigned int stress_var = ss_sys.variable_number("MS__stress"),
                       strain_var = ss_sys.variable_number("MS__strain");
    //
    // total length of fibres in the RVE
    Real fibres_length = 0.0;
    // fibres orientation vector parameter
    RealVectorValue Omega(0., 0., 0.);
    //
    // loop over all active 1-D finite elements
    for (MeshBase::const_element_iterator ci_elem=begin_el; ci_elem!=end_el; ci_elem++) {
        const Elem* elem = (*ci_elem);
        //
        InverseDistanceAlgorithm ida(ida_exp);
        ida.init(rve.vertices());
        //
        Dyadic3D F_0, F_1;
        for (int i=0; i<SP3D; i++) {
            for (int j=0; j<SP3D; j++) {
                std::vector<Real> F_(8);
                for (int v=0; v<8; v++)
                    F_[v] = fld[v].deform_gradient(i,j);
                F_0.comp(i,j) = ida.interp<Real>(elem->point(0), F_);
                F_1.comp(i,j) = ida.interp<Real>(elem->point(1), F_);
            }
        }
        //
        const Point fibre0 = elem->point(1) - elem->point(0),
                    fibreT = (F_1*elem->point(1)) - (F_0*elem->point(0));
        const Real fibre0_length = fibre0.norm(),
                   fibreT_length = fibreT.norm();
        //
        fibres_length += fibreT_length;
        //
        Omega(0) += pow2(fibreT(0)) / fibreT_length;
        Omega(1) += pow2(fibreT(1)) / fibreT_length;
        Omega(2) += pow2(fibreT(2)) / fibreT_length;
        //
        const Real stretch = fibreT_length / fibre0_length,
                   strain = stretch - 1.0,
                   stress = ( stretch >= 1.0
                            ? fibril_stiffness__tension*(stretch-1.0) : fibril_stiffness__compression*(1.0-stretch) ),
                   force = stress * fibril_cross_section_area;
        //
        // local degrees of freedom for each 'gradient' variable
        std::vector<dof_id_type>* dof_indices_ss_var = new std::vector<dof_id_type> [2];
        // local indices valid range
        const unsigned int dof_F = ss_sys.solution->first_local_index(),
                           dof_L = ss_sys.solution->last_local_index();
        // now set the (discontinuous) stress & strain for this element
        {
            const unsigned int l = 0; // variable index
            ss_sys.get_dof_map().dof_indices(elem, dof_indices_ss_var[l], stress_var);
            const unsigned int idof = dof_indices_ss_var[l][0];
            if ( dof_F <= idof && idof < dof_L )
                ss_sys.solution->set(idof, stress/pow2(length_downscaling_factor));
        }
        {
            const unsigned int l = 1; // variable index
            ss_sys.get_dof_map().dof_indices(elem, dof_indices_ss_var[l], strain_var);
            const unsigned int idof = dof_indices_ss_var[l][0];
            if ( dof_F <= idof && idof < dof_L )
                ss_sys.solution->set(idof, strain);
        }
        // free memo
        delete [] dof_indices_ss_var;
        // end of active 1-D finite elements loop...
    }
    //
    // close the solution container and update the system
    ss_sys.solution->close();
    ss_sys.update();
    //
    #endif
    //
    InverseDistanceAlgorithm ida(ida_exp);
    ida.init(rve.vertices());
    //
    std::vector<Real> F_v[SP3D][SP3D];
    std::vector<Real> U_v[SP3D];
    for (int v=0; v<8; v++) {
        for (int i=0; i<SP3D; i++) {
            U_v[i].push_back( fld[v].displacement(i) );
            for (int j=0; j<SP3D; j++) {
                F_v[i][j].push_back( fld[v].deform_gradient(i,j) );
            }
        }
    }
    //
    for (int face=0; face<6; face++)
    // iterate for all boundary nodes of this face of the RVE
    for ( std::map<dof_id_type, std::set<dof_id_type>>::const_iterator
          ci_m=rve.bnode_neighbor_begin(face); ci_m!=rve.bnode_neighbor_end(face); ci_m++ )
    {
        const Node* node = rve.node_ptr(ci_m->first);
        //
        Dyadic3D F;
        for (int i=0; i<SP3D; i++)
            for (int j=0; j<SP3D; j++)
                F.comp(i,j) = ida.interp<Real>(*node, F_v[i][j]);
        //
        const Point eta_B = rve.normal(face);
        //
        for ( std::set<dof_id_type>::const_iterator
              ci_s=ci_m->second.begin(); ci_s!=ci_m->second.end(); ci_s++ )
        {
            const Node* node_adj = rve.node_ptr(*ci_s);
            //
            Dyadic3D F_adj;
            for (int i=0; i<SP3D; i++)
                for (int j=0; j<SP3D; j++)
                    F_adj.comp(i,j) = ida.interp<Real>(*node_adj, F_v[i][j]);
            //
            const Point fibre0 = (*node) - (*node_adj),
                        fibreT = (F*(*node)) - (F_adj*(*node_adj));
            const Real fibre0_length = fibre0.norm(),
                       fibreT_length = fibreT.norm();
            //
            const Real stretch = fibreT_length / fibre0_length,
                       stress = (stretch >= 1.0 ? fibril_stiffness__tension : fibril_stiffness__compression)
                              * (stretch-1.0),
                       force = stress * fibril_cross_section_area;
            //
            const Point pnt = (*node) - rve.centre(),
                        eta_F = fibre0.unit();
            //
            RealTensorValue sm_ij;
            for (unsigned int i=0; i<SP3D; i++) {
                for (unsigned int j=0; j<SP3D; j++) {
                    sm_ij(i,j) = eta_F(i)*stress*eta_B(j);
                    //
                    S2pk(i,j) += pnt(i)*force*eta_F(j)
                               / rve.volume();
                }
            }
            //
            xyz[face].push_back(*node);
            sigma[face].push_back(sm_ij);
        }
        // end of boundary nodes (of this face) loop...
    }
    //
    // NOTE: The code below is taken from the previous version of the source code, and it
    // has an "small" bug! Here we iterate for all 1D elements, which happens to allow
    // computation of the stress tensors for the boundary nodes twice, hence, give erroneous
    // values for Sij and Qj in the end. Note the differences with the update code above!
    //
    /* *** *** *** *** *** *** *** *** *** *** *** *** *** *
    for (MeshBase::const_element_iterator ci_elem=begin_el; ci_elem!=end_el; ci_elem++) {
        const Elem* elem = (*ci_elem);
        //
        int face = -1;
        for (unsigned int ne=0; ne<elem->n_nodes(); ne++)
        if ( rve.node_is_boundary(elem->node_id(ne), face) ) {
            //
            InverseDistanceAlgorithm ida(ida_exp);
            ida.init(rve.vertices());
            //
            Dyadic3D F_0, F_1;
            for (int i=0; i<SP3D; i++) {
                for (int j=0; j<SP3D; j++) {
                    std::vector<Real> F_(8);
                    for (int v=0; v<8; v++)
                        F_[v] = fld[v].deform_gradient(i,j);
                    F_0.comp(i,j) = ida.interp<Real>(elem->point(0), F_);
                    F_1.comp(i,j) = ida.interp<Real>(elem->point(1), F_);
                }
            }
            //
            const Point fibre0 = elem->point(1) - elem->point(0),
                        fibreT = (F_1*elem->point(1)) - (F_0*elem->point(0));
            const Real fibre0_length = fibre0.norm(),
                       fibreT_length = fibreT.norm();
            //
            const Real stretch = fibreT_length / fibre0_length,
                       strain = stretch - 1.0,
                       stress = ( stretch >= 1.0
                                ? fibril_stiffness__tension*(stretch-1.0) : fibril_stiffness__compression*(1.0-stretch) ),
                       force = stress * fibril_cross_section_area;
            //
            const Point eta_B = rve.normal(face);
            Point eta_F;
            if      ( 0 == ne && stretch >= 1.0 ) eta_F = - fibre0.unit();
            else if ( 0 == ne && stretch  < 1.0 ) eta_F =   fibre0.unit();
            else if ( 1 == ne && stretch >= 1.0 ) eta_F =   fibre0.unit();
            else if ( 1 == ne && stretch  < 1.0 ) eta_F = - fibre0.unit();
            else libmesh_error();
            //
            const Point pnt = elem->point(ne) - rve.centre();
            //
            for (unsigned int i=0; i<SP3D; i++)
                for (unsigned int j=0; j<SP3D; j++)
                    S2pk(i,j) += pnt(i)*force*eta_F(j)
                               / rve.volume();
            //
            RealTensorValue sm_ij;
            for (unsigned int i=0; i<SP3D; i++)
                for (unsigned int j=0; j<SP3D; j++)
                    sm_ij(i,j) = eta_F(i)*stress*eta_B(j);
            //
            xyz[face].push_back(elem->point(ne));
            sigma[face].push_back(sm_ij);
            // end of element nodes loop...
        }
        // end of active 1-D finite elements loop...
    }
    * *** *** *** *** *** *** *** *** *** *** *** *** *** */
    //
    // loop over all 6 faces of the RVE cube
    for (unsigned int face=0; face<6; face++) {
        const Point& eta_B = rve.normal(face);
        const std::vector<int> id = vertices_of_HEX_face(face);
        //
        std::vector<Point> vpnt;
        std::vector<RealVectorValue> dUdy0, dUdy1, dUdy2;
        for (unsigned int i=0; i<4; i++) {
            const int v = id[i];
            vpnt.push_back( rve.vertex(v) );
            const DisplacementGradient& dU = fld[v].grad_displacement;
            dUdy0.push_back( RealVectorValue(dU(0,0), dU(1,0), dU(2,0)) );
            dUdy1.push_back( RealVectorValue(dU(0,1), dU(1,1), dU(2,1)) );
            dUdy2.push_back( RealVectorValue(dU(0,2), dU(1,2), dU(2,2)) );
        }
        //
        InverseDistanceAlgorithm ida_1(ida_exp);
        ida_1.init(xyz[face]);
        //
        InverseDistanceAlgorithm ida_2(ida_exp);
        ida_2.init(vpnt);
        //
        for (unsigned int qp=0; qp<4; qp++) {
            const Point& qpnt = rve.quadrature(face)[qp];
            // interpolate to obtain the microscopic stresses at the quadrature point
            const RealTensorValue sm_ = ida_1.interp<RealTensorValue>(qpnt, sigma[face]);
            // interpolate to obtain the displacement gradients at the quadrature point
            const RealVectorValue dUdy0_ = ida_2.interp<RealVectorValue>(qpnt, dUdy0);
            const RealVectorValue dUdy1_ = ida_2.interp<RealVectorValue>(qpnt, dUdy1);
            const RealVectorValue dUdy2_ = ida_2.interp<RealVectorValue>(qpnt, dUdy2);
            //
            const RealVectorValue n__dU(eta_B*dUdy0_, eta_B*dUdy1_, eta_B*dUdy2_);
            const RealTensorValue D = (sm_ - S2pk).transpose();
            //
            // now compute the "Q" (vector) source term
            Q_source += rve.jacobian()
                      * (1.0/rve.volume())
                      * (D * n__dU);
            // end of quadrature points loop...
        }
        // end of RVE faces loop...
    }
    //
    // transformation jacobian from micro- (RVE) to macro- domain
    const Real Jac_S = (fibre_volume_fraction/fibril_cross_section_area)
                     * (rve.volume()/rve.fibres_length())
                     / pow2(length_downscaling_factor);
    const Real Jac_Q = Jac_S
                     / length_downscaling_factor;
    // upscale the (symmetric) stress tensor
    es.parameters.set<Real>("macroscopic stress component S0") = Jac_S*S2pk(0,0);
    es.parameters.set<Real>("macroscopic stress component S1") = Jac_S*S2pk(1,1);
    es.parameters.set<Real>("macroscopic stress component S2") = Jac_S*S2pk(2,2);
    es.parameters.set<Real>("macroscopic stress component S3") = Jac_S*(S2pk(1,2)+S2pk(2,1))*0.5;
    es.parameters.set<Real>("macroscopic stress component S4") = Jac_S*(S2pk(0,2)+S2pk(2,0))*0.5;
    es.parameters.set<Real>("macroscopic stress component S5") = Jac_S*(S2pk(0,1)+S2pk(1,0))*0.5;
    // upscale the source-term vector
    es.parameters.set<Real>("multiscale source-term component Q0") = Jac_Q*Q_source(0);
    es.parameters.set<Real>("multiscale source-term component Q1") = Jac_Q*Q_source(1);
    es.parameters.set<Real>("multiscale source-term component Q2") = Jac_Q*Q_source(2);
    //
    #ifndef BOOST_COMPUTATIONS
    //
    // set these parameters first regarding the total length of fibres in the RVE
    if ( 0.0 == es.parameters.get<Real>("RVE total fibre length (reference)") ) {
        es.parameters.set<Real>("RVE total fibre length (reference)") = fibres_length;
        es.parameters.set<Real>("RVE total fibre length (current)") = fibres_length;
    } else {
        es.parameters.set<Real>("RVE total fibre length (current)") = fibres_length;
    }
    fibres_length = es.parameters.get<Real>("RVE total fibre length (reference)");
    //
    // set up the averaged RVE orientation vector
    for (unsigned int ivar=0; ivar<SP3D; ivar++) {
        System& w_sys = es.get_system<System>(sname+"::omega"+number2string(ivar));
        libmesh_assert( 1 == w_sys.n_vars() );
        //
        // set the solution vector
        w_sys.solution->zero();
        w_sys.solution->add(Omega(ivar)/fibres_length);
        // close the solution container and update the system
        w_sys.solution->close();
        w_sys.update();
    }
    //
    #endif
}
//=================================================================================================
} // end of namespace collagen_extracellular_matrix...
//=================================================================================================
#undef BOOST_COMPUTATIONS
//=================================================================================================
