/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/utils.h"
#include "time.h"
//=================================================================================================
// local variable
static
const Dyadic3D& I = identity_matrix;
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
//=================================================================================================
namespace tumour_angiogenesis {
//=================================================================================================
void damage_model_498 (const MaterialInfo& mat, FieldInfo& fld) {
    const unsigned int nvars = 4;
    //
    const Real ecm = apply_bounds(0.0, fld.biochem[3], 1.0);
    const Real temp_load = fld.biochem[nvars+8];
    //
    Real& zeta = fld.biomech[0];
    //
    const int damage_model = mat.get_biochem_param(420);
    //
    if ( damage_model == 0 ) {
        zeta = 1.0 - ecm;
    } else if ( damage_model == 1 ) {
        const Real a_w = mat.get_biochem_param(421);
        zeta = 1.0 - pow(ecm, a_w);
    } else libmesh_error();
    //
    if ( zeta > material__max_damage_factor ) zeta = material__max_damage_factor;
}
//=================================================================================================
} // end of namespace tumour_angiogenesis...
//=================================================================================================
//
//
//
//=================================================================================================
namespace tumour_angiogenesis {
//=================================================================================================
void inelastic_deform_membrane_498 (const MaterialInfo& mat, const FieldInfo& fld) {
    libmesh_error();
}
//=================================================================================================
void inelastic_deform_solid_498 (const MaterialInfo& mat, const FieldInfo& fld) {
    const unsigned int nvars = 4;
    //
    const Real oxy = apply_lbound(0.0, fld.biochem[1]),
               ecm = apply_bounds(0.0, fld.biochem[3], 1.0);
    const RealGradient Grad_oxy = fld.grad_biochem[1],
                       Grad_ecm = fld.grad_biochem[3];
    //const Real temp_load = fld.biochem[nvars+8];
    //
    // mean solid stress (hydrostatic pressure)
    //const Real p = fld.biomech[2];
    //
    const int growth_model = mat.get_biochem_param(400);
    //
    // calculate the inelastic (plastic) deformation gradient tensor
    if ( 1 == growth_model ) {
        // tumour structural integrity function
        const Real tsi_f = pow(ecm, mat.get_biochem_param(406));
        const Real alpha_iso = mat.get_biochem_param(401)*tsi_f; // -
        const Real alpha_oxy = mat.get_biochem_param(402)*tsi_f; // -
        const Real alpha_ecm = mat.get_biochem_param(403)*tsi_f; // -
        const Real beta  = mat.get_biochem_param(404); // -
        const Real gamma = mat.get_biochem_param(405); // -
        //
        Real epsilon = 0.0; // Green-Lagrange strain
        Real lambda[] = { 1.0, 1.0, 1.0 }; // stretch ratio
        //
        if ( alpha_iso > 0.0 ) {
            epsilon = alpha_iso*(exp(-beta*exp(-gamma*oxy))-exp(-beta));
            lambda[0] = sqrt(2.0*epsilon+1.0);
        }
        fld.deform_grad_inel.comp = lambda[0] * I.comp;
        if ( alpha_oxy > 0.0 ) {
            epsilon = alpha_oxy*(exp(-beta*exp(-gamma*oxy))-exp(-beta))*Grad_oxy.norm();
            lambda[1] = sqrt(2.0*epsilon+1.0) - lambda[0];
            if ( Grad_oxy.norm() > 1.0e-6 )
                fld.deform_grad_inel.comp += lambda[1] * tensor(vec(Grad_oxy.unit())).comp;
        }
        if ( alpha_ecm > 0.0 ) {
            epsilon = alpha_ecm*(exp(-beta*exp(-gamma*oxy))-exp(-beta))*Grad_ecm.norm();
            lambda[2] = sqrt(2.0*epsilon+1.0) - lambda[0];
            if ( Grad_ecm.norm() > 1.0e-6 )
                fld.deform_grad_inel.comp += lambda[2] * tensor(vec(Grad_ecm.unit())).comp;
        }
        //
    } else libmesh_error();
    //
    // calculate the elastic (recoverable) deformation gradient tensor
    fld.deform_grad_elas = mul_im_mj(fld.deform_gradient, inverse(fld.deform_grad_inel));
}
//=================================================================================================
} // end of namespace tumour_angiogenesis...
//=================================================================================================
//
//
//
//=================================================================================================
namespace tumour_angiogenesis {
//=================================================================================================
#ifdef _LOCAL_ARGUMENTS_2D_
    #error "an unexpected error occurred!"
#else
    #define _LOCAL_ARGUMENTS_2D_ \
        const MaterialInfo& mat, \
        const FieldInfo& fld,    \
        Dyadic2D& S2pk,          \
        Quindic2D& Dtsm
#endif
#ifdef _LOCAL_ARGUMENTS_3D_
    #error "an unexpected error occurred!"
#else
    #define _LOCAL_ARGUMENTS_3D_ \
        const MaterialInfo& mat, \
        const FieldInfo& fld,    \
        Dyadic3D& S2pk,          \
        Quindic3D& Dtsm_uu,      \
        Dyadic3D& Dtsm_up,       \
        Real& Ph,                \
        Dyadic3D& Dtsm_pu,       \
        Real& Dtsm_pp
#endif
//=================================================================================================
void growth_membrane_498 ( _LOCAL_ARGUMENTS_2D_ ) {
    //const unsigned int nvars = 4;
    //
    libmesh_error();
}
//=================================================================================================
void growth_solid_498 ( _LOCAL_ARGUMENTS_3D_ ) {
    //const unsigned int nvars = 4;
    //
    ; // nothing to do here...
}
//=================================================================================================
#undef _LOCAL_ARGUMENTS_2D_
#undef _LOCAL_ARGUMENTS_3D_
//=================================================================================================
} // end of namespace tumour_angiogenesis...
//=================================================================================================
//
//
//
//=================================================================================================
namespace tumour_angiogenesis {
//=================================================================================================
void growth_vascular_498 (const MaterialInfo& mat, const FieldInfo& fld, const VascularInfo& vsc, Displacement& increment) {
    const unsigned int nvars = 4;
    //
    const Real taf = fld.biochem[0];
    const RealGradient Grad_taf = fld.grad_biochem[0];
    const RealGradient Grad_ecm = fld.grad_biochem[3];
    //
    std::vector<Real> eigval_s;
    std::vector<Vector3D> eigvec_s;
    eigen_solve(fld.stress, eigval_s, eigvec_s);
    //
    const RealGradient sigma = RealGradient(eigvec_s[2](0), eigvec_s[2](1), eigvec_s[2](2) );
    //
    const Real taf_star = ( fld.div_velocity ? mat.get_biochem_param(99) : mat.get_biochem_param(98) );
    //
    // space vector of the vascular point under consideration
    const Point vpnt = increment;
    //
    // calculate the magnitude of the tip/sprout extension
    Real displ = 0.0;
    if ( taf >= taf_star ) {
        const Real V0 = mat.get_biochem_param(28),
                   V1 = mat.get_biochem_param(29);
        const Real R_bar = mat.get_biochem_param(30);
        const Real V_max = mat.get_biochem_param(31),
                   V_min = mat.get_biochem_param(32);
        //
        Real veloc = V0 + V1 * exp(-vsc.radius/R_bar);
        if      ( V_max > 0.0 && veloc > V_max ) veloc = V_max;
        else if ( V_min > 0.0 && veloc < V_min ) veloc = V_min;
        //
        displ = fld.time_step * veloc;
    }
    //
    // calculate the unit direction vector of the tip/sprout extension
    Point eta(0.0, 0.0, 0.0);
    if ( taf >= taf_star ) {
        const Real kc = uniform_distro(mat.get_biochem_param(21), mat.get_biochem_param(22));
        const Real kh = uniform_distro(mat.get_biochem_param(23), mat.get_biochem_param(24));
        const Real km = uniform_distro(mat.get_biochem_param(25), mat.get_biochem_param(26));
        //
        eta = kc * Grad_taf // chemotactic contribution
            + kh * Grad_ecm // haptotactic contribution
            - km * sigma;   // mechanotactic contribution
    }
    //
    // update the unit direction vector of the tip/sprout extension; incorporate some random perturbation
    if ( eta.norm() > 1.0e-12 ) {
        do {
            const Real u = uniform_distro(-1.0, 1.0);
            const Real v = uniform_distro(-1.0, 1.0);
            if ( (pow2(u)+pow2(v)) >= 1.0 ) continue;
            //
            const Point n( 2.0*u*sqrt(1.0-pow2(u)-pow2(v)),
                           2.0*v*sqrt(1.0-pow2(u)-pow2(v)),
                           1.0-2.0*(pow2(u)+pow2(v)) );
            const Real e_dot_n = eta.unit() * n;
            if ( e_dot_n <= 0.0 ) continue;
            const Real angle = acos(e_dot_n);
            if ( angle <= degrees_to_radians(45.0) ) {
                eta = n;
                break;
            }
        } while (true);
        // const Point nu = eta.unit();
        // //
        // const Real angl = m2PI*uniform_distro(0.0, 1.0);
        // const Point pert = (rotate_about_axis(nu, angl, vpnt) - vpnt).unit();
        // // set finally the unit direction vector
        // eta = (nu + pert).unit();
    } else {
        increment = Point();
        // else exit with zero vector
        return;
    }
    //
    // return the incremental space vector of the tip/sprout extension
    increment = displ * eta;
}
//=================================================================================================
} // end of namespace tumour_angiogenesis...
//=================================================================================================
//
//
//
//=================================================================================================
#ifdef UPDATE_MIN_MAX_VASC_
#error "macro function \"UPDATE_MIN_MAX_VASC_\" has been defined elsewhere!"
#endif
#define UPDATE_MIN_MAX_VASC_(_MIN_,_MAX_,_STD_) \
    do { \
        const Real StDev = (_STD_), Delta = (_MAX_-_MIN_), Ratio = (Delta/_MAX_); \
        const Real _LOW_ = _MIN_*uniform_distro(1.0-StDev, 1.0+StDev); \
        const Real _UP_  = _MAX_*uniform_distro(1.0-StDev, 1.0+StDev); \
        if (_MIN_ > _MAX_) { \
            ERROR_("error with the bounds inserted"); \
        } else if (Ratio<=1.0e-3 && StDev<=1.0e-3) { \
            char msg[1000]; sprintf(msg, "standard deviation of geometric value is extremely low: %f~%f; %f; %f", _MIN_,_MAX_,Ratio,StDev); \
            ERROR_(msg); \
        } else if (Ratio<=1.0e-3 && StDev> 0.0 && StDev< 1.0) { \
            _MIN_ = std::min(_LOW_, _UP_); _MAX_ = std::max(_LOW_, _UP_); \
        } else if (Ratio >1.0e-3 && StDev>-1.0 && StDev< 0.0) { \
            _MIN_ = _LOW_; _MAX_ = _LOW_+Delta; ASSERT_(_MIN_>0.0, "minimum cannot be <=0; case 1"); \
        } else if (Ratio >1.0e-3 && StDev> 0.0 && StDev< 1.0) { \
            _MAX_ = _UP_;  _MIN_ = _UP_ -Delta; ASSERT_(_MIN_>0.0, "minimum cannot be <=0; case 2"); \
        } \
    } while ( false )
//=================================================================================================
namespace tumour_angiogenesis {
//=================================================================================================
void remodel_vascular_XYZ (const MaterialInfo& mat, VascularInfo& vn_info, const Real* features) {
    libmesh_assert( features != NULL );
    // exit the routine normally for boundary (in/outlet) vascular node
    if ( VascularInfo::INLET  == vn_info.type ||
         VascularInfo::OUTLET == vn_info.type ) return;
    // exit the routine normally for mature (uncompressed) vascular node
    if ( VascularInfo::MATURE == vn_info.type ) {
        if ( vn_info.press_ratio <= 1.0 ) return;
    }
    // exit the routine normally for original (uncompressed) vascular node
    /*
    if ( 0.0 == vn_info.time_generate ) {
        // if ( VascularInfo::NORMAL == vn_info.status ) return;
        if ( vn_info.press_ratio <= 1.0 ) return;
    }
    */
    //
    const Real T = features[0] - vn_info.time_generate,
               //vascular_press = features[1],
               blood_viscosity = features[2],
               wall_shear_stress = features[3],
               mean_velocity = features[4],
               temperature = features[5];
    const Real drug0_p[] = { features[6], features[7] },
               drug1_p[] = { features[8], features[9] };
    //
    // NOTE that currently the lymphatic vessel remodelling time function uses the same time-parameters as
    // as the blood vessel remodelling time function
    //
    const Real mean_velocity__min = mat.get_biochem_param(10);
    const Real T_m_min = mat.get_biochem_param(12),
               T_m_max = mat.get_biochem_param(11);
    const Real Eps_max = mat.get_biochem_param(13);
    const Real P_bar = vn_info.press_ratio;
    Real&       P_c   = vn_info.press_ratio_cr;
    libmesh_assert( Eps_max > 0.0 && Eps_max < 1.0 );
    //
    Real T_m = T_m_max;
    if ( VascularInfo::COLLAPSED != vn_info.status && 1000.0 != vn_info.compression_scale ) {
        const Real R_max = VascularInfo::BLOOD == vn_info.flow ? mat.get_biochem_param(53) : mat.get_biochem_param(103);
        const Real WSS_bar = 4.0*mean_velocity__min*blood_viscosity/R_max;
        const Real WSS_ = wall_shear_stress / WSS_bar;
        //
        T_m = std::min( vn_info.time_mature, ( WSS_ >= 1.0 ? T_m_min : T_m_min+(T_m_max-T_m_min)*exp(1.0-1.0/(1.0-pow2(WSS_))) ) );
        // linear decrease of blood vessel maturation time:                 T_m_max-(T_m_max-T_m_min)*WSS_
        // smooth decrease (bump function) of blood vessel maturation time: T_m_min+(T_m_max-T_m_min)*exp(1-1/(1-pow2(WSS_)))
    }
    libmesh_assert( T_m >= T_m_min );
    //
    const Real R_def = ( VascularInfo::BLOOD==vn_info.flow
                       ? normal_distro(mat.get_biochem_param( 50), mat.get_biochem_param( 51)) : normal_distro(mat.get_biochem_param(100), mat.get_biochem_param(101)) );
    Real R_min = ( VascularInfo::BLOOD==vn_info.flow ? mat.get_biochem_param( 52) : mat.get_biochem_param(102) );
    Real R_max = ( VascularInfo::BLOOD==vn_info.flow ? mat.get_biochem_param( 53) : mat.get_biochem_param(103) );
    //
    const Real H_def = ( VascularInfo::BLOOD==vn_info.flow
                       ? normal_distro(mat.get_biochem_param( 54), mat.get_biochem_param( 55)) : normal_distro(mat.get_biochem_param(104), mat.get_biochem_param(105)) );
    Real H_min = ( VascularInfo::BLOOD==vn_info.flow ? mat.get_biochem_param( 56) : mat.get_biochem_param(106) );
    Real H_max = ( VascularInfo::BLOOD==vn_info.flow ? mat.get_biochem_param( 57) : mat.get_biochem_param(107) );
    //
    const Real G_def = ( VascularInfo::BLOOD==vn_info.flow
                       ? normal_distro(mat.get_biochem_param( 58), mat.get_biochem_param( 59)) : normal_distro(mat.get_biochem_param(108), mat.get_biochem_param(109)) );
    Real G_min = ( VascularInfo::BLOOD==vn_info.flow ? mat.get_biochem_param( 60) : mat.get_biochem_param(110) );
    Real G_max = ( VascularInfo::BLOOD==vn_info.flow ? mat.get_biochem_param( 61) : mat.get_biochem_param(111) );
    //
    const Real F_def = ( VascularInfo::BLOOD==vn_info.flow
                       ? normal_distro(mat.get_biochem_param( 62), mat.get_biochem_param( 63)) : normal_distro(mat.get_biochem_param(112), mat.get_biochem_param(113)) );
    Real F_min = ( VascularInfo::BLOOD==vn_info.flow ? mat.get_biochem_param( 64) : mat.get_biochem_param(114) );
    Real F_max = ( VascularInfo::BLOOD==vn_info.flow ? mat.get_biochem_param( 65) : mat.get_biochem_param(115) );
    //
    const Real P_c_def = ( VascularInfo::BLOOD==vn_info.flow
                         ? normal_distro(mat.get_biochem_param( 66), mat.get_biochem_param( 67)) : normal_distro(mat.get_biochem_param(116), mat.get_biochem_param(117)) );
    Real P_c_min = ( VascularInfo::BLOOD==vn_info.flow ? mat.get_biochem_param( 68) : mat.get_biochem_param(118) );
    Real P_c_max = ( VascularInfo::BLOOD==vn_info.flow ? mat.get_biochem_param( 69) : mat.get_biochem_param(119) );
    //
    UPDATE_MIN_MAX_VASC_(R_min,   R_max,   (VascularInfo::BLOOD==vn_info.flow ? mat.get_biochem_param(150) : mat.get_biochem_param(170)));
    UPDATE_MIN_MAX_VASC_(H_min,   H_max,   (VascularInfo::BLOOD==vn_info.flow ? mat.get_biochem_param(151) : mat.get_biochem_param(171)));
    UPDATE_MIN_MAX_VASC_(G_min,   G_max,   (VascularInfo::BLOOD==vn_info.flow ? mat.get_biochem_param(152) : mat.get_biochem_param(172)));
    UPDATE_MIN_MAX_VASC_(F_min,   F_max,   (VascularInfo::BLOOD==vn_info.flow ? mat.get_biochem_param(153) : mat.get_biochem_param(173)));
    UPDATE_MIN_MAX_VASC_(P_c_min, P_c_max, (VascularInfo::BLOOD==vn_info.flow ? mat.get_biochem_param(154) : mat.get_biochem_param(174)));
    //
    const Real drug0_max = ( VascularInfo::BLOOD==vn_info.flow
                           ? mat.get_biochem_param(300) : mat.get_biochem_param(350) );
    const Real R_d0s = ( VascularInfo::BLOOD==vn_info.flow
                       ? (mat.get_biochem_param(301)>=0.0 ? mat.get_biochem_param(301) : 1.0) : (mat.get_biochem_param(351)>=0.0 ? mat.get_biochem_param(351) : 1.0) );
    const Real H_d0s = ( VascularInfo::BLOOD==vn_info.flow
                       ? (mat.get_biochem_param(302)>=0.0 ? mat.get_biochem_param(302) : 1.0) : (mat.get_biochem_param(352)>=0.0 ? mat.get_biochem_param(352) : 1.0) );
    const Real G_d0s = ( VascularInfo::BLOOD==vn_info.flow
                       ? (mat.get_biochem_param(303)> 0.0 ? mat.get_biochem_param(303) : 1.0) : (mat.get_biochem_param(353)> 0.0 ? mat.get_biochem_param(353) : 1.0) );
    const Real F_d0s = ( VascularInfo::BLOOD==vn_info.flow
                       ? (mat.get_biochem_param(304)> 0.0 ? mat.get_biochem_param(304) : 1.0) : (mat.get_biochem_param(354)> 0.0 ? mat.get_biochem_param(354) : 1.0) );
    //
    const Real drug1_max = ( VascularInfo::BLOOD==vn_info.flow
                           ? mat.get_biochem_param(310) : mat.get_biochem_param(360) );
    const Real R_d1s = ( VascularInfo::BLOOD==vn_info.flow
                       ? (mat.get_biochem_param(311)>=0.0 ? mat.get_biochem_param(311) : 1.0) : (mat.get_biochem_param(361)>=0.0 ? mat.get_biochem_param(361) : 1.0) );
    const Real H_d1s = ( VascularInfo::BLOOD==vn_info.flow
                       ? (mat.get_biochem_param(312)>=0.0 ? mat.get_biochem_param(312) : 1.0) : (mat.get_biochem_param(362)>=0.0 ? mat.get_biochem_param(362) : 1.0) );
    const Real G_d1s = ( VascularInfo::BLOOD==vn_info.flow
                       ? (mat.get_biochem_param(313)> 0.0 ? mat.get_biochem_param(313) : 1.0) : (mat.get_biochem_param(363)> 0.0 ? mat.get_biochem_param(363) : 1.0) );
    const Real F_d1s = ( VascularInfo::BLOOD==vn_info.flow
                       ? (mat.get_biochem_param(314)> 0.0 ? mat.get_biochem_param(314) : 1.0) : (mat.get_biochem_param(364)> 0.0 ? mat.get_biochem_param(364) : 1.0) );
    //
    if ( drug0_max && ( drug0_p[0]>drug0_max || drug0_p[1]>drug0_max ) ) {
        R_min *= R_d0s; R_max *= R_d0s;
        H_min *= H_d0s; H_max *= H_d0s;
        G_min *= G_d0s; G_max *= G_d0s;
        F_min *= F_d0s; F_max *= F_d0s;
    } else if ( drug0_max && ( drug0_p[0]>0.0 || drug0_p[1]>0.0 ) ) {
        const Real drug_ratio = std::max(drug0_p[0], drug0_p[1]) / drug0_max;
        R_min *= R_d0s*drug_ratio; R_max *= R_d0s*drug_ratio;
        H_min *= H_d0s*drug_ratio; H_max *= H_d0s*drug_ratio;
        G_min *= G_d0s*drug_ratio; G_max *= G_d0s*drug_ratio;
        F_min *= F_d0s*drug_ratio; F_max *= F_d0s*drug_ratio;
    }
    //
    if ( drug1_max && ( drug1_p[0]>drug1_max || drug1_p[1]>drug1_max ) ) {
        R_min *= R_d1s; R_max *= R_d1s;
        H_min *= H_d1s; H_max *= H_d1s;
        G_min *= G_d1s; G_max *= G_d1s;
        F_min *= F_d1s; F_max *= F_d1s;
    } else if ( drug1_max && ( drug1_p[0]>0.0 || drug1_p[1]>0.0 ) ) {
        const Real drug_ratio = std::max(drug1_p[0], drug1_p[1]) / drug1_max;
        R_min *= R_d1s*drug_ratio; R_max *= R_d1s*drug_ratio;
        H_min *= H_d1s*drug_ratio; H_max *= H_d1s*drug_ratio;
        G_min *= G_d1s*drug_ratio; G_max *= G_d1s*drug_ratio;
        F_min *= F_d1s*drug_ratio; F_max *= F_d1s*drug_ratio;
    }
    //
    if ( ( 0.0==R_min && 0.0==R_max ) ||
         ( 0.0==H_min && 0.0==H_max ) ) {
        vn_info.status = VascularInfo::COLLAPSED;
        vn_info.compression_scale = 1000.0;
    }
    //
    const Real DR   = R_max - R_min,
               DH   = H_max - H_min,
               DG   = G_max - G_min,
               DF   = F_max - F_min,
               DP_c = P_c_max - P_c_min,
               DL   = (R_max+0.5*H_max) - (R_min+0.5*H_min);
    //
    if ( VascularInfo::COLLAPSED == vn_info.status && 1000.0 == vn_info.compression_scale ) {
        //
        vn_info.time_mature = 1000.0 * T_m_max;
        //
    } else {
        //
        if ( 0.0 != vn_info.time_generate ) {
            if ( T >= T_m && VascularInfo::SPROUT == vn_info.type )
                vn_info.type = VascularInfo::MATURE;
            //
            vn_info.time_mature = apply_bounds(T_m_min, T_m, T_m_max);
        }
        //
    }
    //
    if ( 0.0 == vn_info.time_generate ) {
        //
        if ( VascularInfo::MATURE == vn_info.type ) {
            //
            vn_info.radius = R_def;
            vn_info.thickness = H_def;
            vn_info.pore_radius = G_def;
            vn_info.pores_fraction = F_def;
            //
            P_c = P_c_def;
            //
        }
        //
    } else {
        //
        const Real T_ = T/T_m;
        //const Real R_t = ( VascularInfo::TIP == vn_info.type ? R_min : ( T >= T_m ? R_max : ((-2.0*DR)*pow3(T_)+(3.0*DR)*pow2(T_)+R_min) ) );
        const Real R_t = (DR*exp(-11.0*exp(-4.4*T_))+R_min);
        const Real H_t = ( T_ >= 1.0 ? H_max : (DH*T_+H_min) );
        const Real G_t = ( T_ >= 1.0 ? G_min : ((2.0*DG)*pow3(T_)+(-3.0*DG)*pow2(T_)+G_max) );
        const Real F_t = ( T_ >= 1.0 ? F_min : (F_max-DF*T_) );
        //
        vn_info.radius = R_t;
        vn_info.thickness = H_t;
        vn_info.pore_radius = G_t;
        vn_info.pores_fraction = F_t;
        //
        const Real P_c_tang = DP_c / DL;
        //
        P_c = P_c_min + P_c_tang*(vn_info.radius+0.5*vn_info.thickness-R_min-0.5*H_min);
        //
    }
    libmesh_assert( P_c >= 1.0 );
    //
    if ( VascularInfo::COLLAPSED == vn_info.status && 1000.0 == vn_info.compression_scale ) {
        //
        vn_info.radius = R_min;
        vn_info.thickness = H_min/100.0;
        vn_info.pore_radius = G_min/1000.0;
        vn_info.pores_fraction = F_min/1000.0;
        //
        vn_info.status = VascularInfo::COLLAPSED;
        //
        vn_info.compression_scale = 1000.0;
        //
    } else if ( VascularInfo::COLLAPSED == vn_info.status && 1000.0 != vn_info.compression_scale ) {
        //
        libmesh_error();
        //
    } else {
        //
        if ( P_bar <= 1.0 ) {
            //
            vn_info.status = VascularInfo::NORMAL;
            //
            vn_info.compression_scale = 1.0;
            //
        } else if ( P_bar < P_c ) {
            //
            vn_info.status = VascularInfo::COMPRESSED;
            //
            Real strain = 0.0; // (radial) engineering strain
            {
                const Real a_ = (1.0-P_c)/pow2(Eps_max),
                           b_ = 2.0*(P_c-1.0)/Eps_max,
                           c_ = 1.0-P_bar,
                           r0_ = (-sqrt(pow2(b_)-4.0*a_*c_)-b_)/(2.0*a_),
                           r1_ = (+sqrt(pow2(b_)-4.0*a_*c_)-b_)/(2.0*a_);
                if      ( r0_ < Eps_max ) strain = r0_;
                else if ( r1_ < Eps_max ) strain = r1_;
                else libmesh_error();
            }
            vn_info.compression_scale = (strain+1.0);
            //
        } else {
            //
            vn_info.status = VascularInfo::COLLAPSED;
            //
            Real strain = Eps_max; // (radial) engineering strain
            vn_info.compression_scale = (strain+1.0);
            //
        }
        //
    }
    //
    if ( VascularInfo::COLLAPSED == vn_info.status ) {
        ; // nothing to do here...
    } else {
        //
        if ( VascularInfo::TIP == vn_info.type ) {
            //
            vn_info.radius = R_min;
            vn_info.thickness = H_min;
            vn_info.pore_radius = G_max;
            vn_info.pores_fraction = F_max;
            //
            vn_info.compression_scale = 1.0;
            //
        }
        //
    }
}
//=================================================================================================
} // end of namespace tumour_angiogenesis...
//=================================================================================================
#undef UPDATE_MIN_MAX_VASC_
//=================================================================================================
