/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include <map>
#include "libmesh/libmesh_common.h"
#include "libmesh/gmsh_io.h"
#include "feb3/mesh_iga.h"
#include <iomanip>
#include "libmesh/parallel.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
std::string tmp_dir;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
// local parameter
static
const Real gtol = 1.0e-4;
//=================================================================================================
static
bool in_segment (const libMesh::Point& s, const libMesh::Point& e, const libMesh::Point& in) {
    if ( in(0) >= s(0)-gtol && in(0) <= e(0)+gtol ) return true;
    return false;
}
//=================================================================================================
static
bool in_quadrangle (const libMesh::Point& s, const libMesh::Point& e, const libMesh::Point& in) {
    if ( in(0) >= s(0)-gtol && in(0) <= e(0)+gtol &&
         in(1) >= s(1)-gtol && in(1) <= e(1)+gtol ) return true;
    return false;
}
//=================================================================================================
static
bool in_cube (const libMesh::Point& s, const libMesh::Point& e, const libMesh::Point& in) {
    if ( in(0) >= s(0)-gtol && in(0) <= e(0)+gtol &&
         in(1) >= s(1)-gtol && in(1) <= e(1)+gtol &&
         in(2) >= s(2)-gtol && in(2) <= e(2)+gtol ) return true;
    return false;
}
//=================================================================================================
static
void init_parametric_mesh ();
static
void find_neighbors ();
static
void set_local_nodes2nurbs ();
static
void set_boundary_info ();
//=================================================================================================
// local variables
std::vector<NURBS>* nurbs__ptr;
std::map<Node*, std::vector<NURBS*> >* nodes_to_nurbs__ptr;
Mesh* parametric_mesh__ptr;
MeshBase* control_mesh__ptr;
static
std::string mesh_label;
//=================================================================================================
#include "feb3/mesh_iga.h"
//=================================================================================================
MeshIGA::MeshIGA (MeshBase& mesh, std::string label) :
    ParallelObject(mesh.comm()),
    _ctrl_msh(mesh),
    _pphys_msh(mesh.comm(), mesh.mesh_dimension()) {
    this->set_label(label);
}
//=================================================================================================
void MeshIGA::print_info (std::ostream& os) const {
    os << std::endl;
    this->_pphys_msh.print_info(os);
    this->_pphys_msh.boundary_info->print_info(os);
    os << std::endl;
    // loop for all NURBS of the IGA mesh
    for ( std::vector<NURBS>::const_iterator
          n_ci=MeshIGA::nurbs_begin(); n_ci!=MeshIGA::nurbs_end(); n_ci++ ) {
        const NURBS* nurbs = &(*n_ci);
        nurbs->print_info(os);
        for (int s=0; s<nurbs->n_sides(); s++) {
            if ( ! nurbs->neighbor_ptr(s) ) {
                char txt[buffer_len];
                sprintf( txt, " > NURBS (ID:%d) has no neighbors (Boundary ID:%d) on side %d \n",
                         nurbs->id(), nurbs->boundary_id(s), s );
                os << txt;
            } else {
                char txt[buffer_len];
                sprintf( txt, " > NURBS (ID:%d) neighbors with NURBS (ID:%d) on side %d \n",
                         nurbs->id(), nurbs->neighbor_ptr(s)->id(), s );
                os << txt;
            }
        }
    }
    os << std::endl;
    //
    for ( std::map<Node*, std::vector<NURBS*> >::const_iterator
         map_ci=MeshIGA::nodes_to_nurbs_begin(); map_ci!=MeshIGA::nodes_to_nurbs_end(); map_ci++ ) {
        const std::vector<NURBS*>& nurbs_list = map_ci->second;
        os << " Node " << map_ci->first->id() << " is control point of " << nurbs_list.size() << " NURBS [";
        for (unsigned int i=0; i<nurbs_list.size(); i++) {
            os << nurbs_list[i]->id();
            if ( i < nurbs_list.size()-1 ) os << ',';
        }
        os << "]\n";
    }
    os << std::endl;
}
//=================================================================================================
void MeshIGA::set_nurbs (const std::string& fname) {
    std::ifstream fin(fname.c_str());
    libmesh_assert( fin.good() );

    // clean the memory
    this->_nurbs.clear();
    this->_nodes_to_nurbs.clear();
    //
    do {
        MeshBase::node_iterator cni_begin = this->_ctrl_msh.active_nodes_begin();
        MeshBase::node_iterator cni_end   = this->_ctrl_msh.active_nodes_end();
        // loop over all active nodes of the control-net
        for (MeshBase::node_iterator cni=cni_begin; cni!=cni_end; cni++ ) {
            Node* node = *cni;
            std::vector<NURBS*> a_temporary_vector;
            this->_nodes_to_nurbs.insert( std::make_pair(node, a_temporary_vector) );
        }
    } while ( false );
    // obtain the number of NURBS to read
    unsigned int n_nurbs;
    fin >> n_nurbs;
    // allocate memory
    this->_nurbs.clear();
    this->_nurbs.resize(n_nurbs);
    // now read all the NURBS
    for (unsigned int icell=0; icell<n_nurbs; icell++) {
        // set up the default values for the NURBS
        int ncp[SP3D] = {1, 1, 1};
        Order ord[SP3D] {CONSTANT, CONSTANT, CONSTANT};
        // dimension of the NURBS
        int dim;
        fin >> dim;
        ASSERT_(SP3D == dim, "NURBS element inserted in \"MeshIGA\" is not three-dimensional");
        // subdomain ID of the NURBS
        int sbd_id;
        fin >> sbd_id;
        // order of the NURBS (for each direction)
        for (int ispdm=0; ispdm<dim; ispdm++) {
            int o;
            fin >> o;
            ord[ispdm] = static_cast<Order>(o);
        }
        // number of control points of the NURBS (for each direction)
        for (int ispdm=0; ispdm<dim; ispdm++) {
            fin >> ncp[ispdm];
        }
        // initialise the NURBS object
        this->_nurbs[icell].init(ncp, ord, icell);
        this->_nurbs[icell].set_subdomain_id() = static_cast<subdomain_id_type>(sbd_id);
        //
        // read the connectivity of the NURBS (enumeration starts from 1)
        const int all_ncp = ncp[0]*ncp[1]*ncp[2];
        Blitz_VectorInt cp_table(all_ncp);
        for (int inode=0; inode<all_ncp; inode++) {
            int node_id;
            fin >> node_id;
            --node_id; // reduce the ID of the node by '1'
            //
            cp_table(inode) = node_id;
            //
            Node* node = this->_ctrl_msh.node_ptr(node_id);
            if ( NULL == node ) libmesh_error();
            //
            this->_nodes_to_nurbs.find(node)->second.push_back( &this->_nurbs[icell] );
        }
        // check if NURBS is rational
        int is_rational = 0;
        fin >> is_rational;
        if ( is_rational ) {
            this->_nurbs[icell].set_to_rational();
        } else {
            this->_nurbs[icell].set_to_nonrational();
        }
        // read for the corresponding control points the weights
        Blitz_VectorReal wght_table(all_ncp);
        if ( is_rational ) {
            for (int inode=0; inode<all_ncp; inode++) {
                fin >> wght_table(inode);
            }
        }
        // now assign the control points to this NURBS
        if ( SP1D == dim ) {
            int I = 0;
            for (int n=0; n<all_ncp; n++) {
                this->_nurbs[icell].set_ctrl_node(I) = &( this->_ctrl_msh.node_ref(cp_table(n)) );
                if ( is_rational )
                    this->_nurbs[icell].set_weight(I) = wght_table(n);
                ++I;
            }
        } else if ( SP2D == dim ) {
            int I = 0, J = 0;
            for (int n=0; n<all_ncp; n++) {
                this->_nurbs[icell].set_ctrl_node(I,J) = &( this->_ctrl_msh.node_ref(cp_table(n)) );
                if ( is_rational )
                    this->_nurbs[icell].set_weight(I,J) = wght_table(n);
                ++I;
                if ( I % this->_nurbs[icell].n_ctrl_pnt(0) == 0 ) {
                    I = 0;
                    ++J;
                }
            }
        } else if ( SP3D == dim ) {
            int I = 0, J = 0, K = 0;
            for (int n=0; n<all_ncp; n++) {
                this->_nurbs[icell].set_ctrl_node(I,J,K) = &( this->_ctrl_msh.node_ref(cp_table(n)) );
                if ( is_rational )
                    this->_nurbs[icell].set_weight(I,J,K) = wght_table(n);
                ++I;
                if ( I % this->_nurbs[icell].n_ctrl_pnt(0) == 0 ) {
                    I = 0;
                    ++J;
                    if ( J % this->_nurbs[icell].n_ctrl_pnt(1) == 0 ) {
                        J = 0;
                        ++K;
                    }
                }
            }
        } else libmesh_error();
        // check if to re-set the knot vectors
        int change_kvs = 0;
        fin >> change_kvs;
        if ( change_kvs ) {
            // loop for all dimensions of the NURBS
            for (int d=0; d<dim; d++) {
                int kv_size;
                fin >> kv_size;
                if ( kv_size == 0 ) continue;
                // ignore following ...if zero size provided
                Blitz_VectorReal kv(kv_size);
                for (int i=0; i<kv_size; i++) fin >> kv(i);
                //
                this->_nurbs[icell].set_knot_vec(d, kv);
            }
            // end of if case...
        }
        // end of NURBS loop...
    }
    // close the input file stream
    fin.close();

    // sanity checks
    for ( std::map<Node*, std::vector<NURBS*> >::const_iterator
          map_ci=MeshIGA::nodes_to_nurbs_begin(); map_ci!=MeshIGA::nodes_to_nurbs_end(); map_ci++ ) {
        const std::vector<NURBS*>& nurbs_list = map_ci->second;
        if ( nurbs_list.empty() ) {
            char msg[buffer_len];
            sprintf(msg, " node %d belongs in no NURBS patch", map_ci->first->id());
            ERROR_(msg);
        }
    }
}
//=================================================================================================
void MeshIGA::set_parametric_mesh () {
    const int DIM = this->_ctrl_msh.mesh_dimension();
    // sanity checks
    for (unsigned int l=0; l<this->_nurbs.size(); l++)
        if ( DIM != this->_nurbs[l].dimension() ) libmesh_error();
    // copy local pointers
    nurbs__ptr = &this->_nurbs;
    nodes_to_nurbs__ptr = &this->_nodes_to_nurbs;
    parametric_mesh__ptr = &this->_pphys_msh;
    control_mesh__ptr = &this->_ctrl_msh;
    mesh_label = this->get_label();
    //
    set_boundary_info();
    find_neighbors();
    init_parametric_mesh();
    set_local_nodes2nurbs();
    set_boundary_info();
}
//=================================================================================================
void MeshIGA::set_post_data ( EquationSystems* es,
                              const unsigned int nsample_i,
                              const unsigned int nsample_j,
                              const unsigned int nsample_k )
{
    if (NULL==es) libmesh_error();
    const MeshBase::const_element_iterator begin_el = this->get_parametric_mesh().active_elements_begin();
    const MeshBase::const_element_iterator end_el   = this->get_parametric_mesh().active_elements_end();
    //
    // initialise these
    if ( true ) {
        // force to clear memo
        this->post_data.nurbs_pelem_cells.clear();
        this->post_data.cells_vertices.clear();
        //
        unsigned int cell_id = 0;
        //
        this->post_data.nurbs_pelem_cells.resize(this->get_nurbs().size());
        for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
            const Elem* elem_prm = *citer_el;
            const NURBS* elem = &this->get_nurbs().at(elem_prm->subdomain_id());
            //
            std::set<dof_id_type> tmp_set;
            this->post_data.nurbs_pelem_cells[elem->id()].insert( std::make_pair(elem_prm->id(), tmp_set) );
            //
            std::vector<Point> tmp_vec;
            tmp_vec.assign(8, Point());
            for (unsigned int i=1; i<nsample_i; i++) {
                for (unsigned int j=1; j<nsample_j; j++) {
                    for (unsigned int k=1; k<nsample_k; k++) {
                        this->post_data.cells_vertices.insert( std::make_pair(cell_id, tmp_vec) );
                        ++cell_id;
                    }
                }
            }
        }
        if (cell_id!=(this->get_parametric_mesh().n_elem()*(nsample_i-1)*(nsample_j-1)*(nsample_k-1))) libmesh_error();
        //
        this->post_data.es = es;
        // end of if case...
    }
    //
    unsigned int n_points = 0;
    unsigned int n_cells  = 0;
    //
    std::map<unsigned int, Point> points_map;
    std::map<unsigned int, Point> points_prm_coord_map;
    std::map<unsigned int, std::vector<unsigned int> > cells_conn_map;
    //
    // loop over all active (parametric) elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem_prm = *citer_el;
        // access the NURBS
        const NURBS* elem = &this->get_nurbs().at(elem_prm->subdomain_id());
        libmesh_assert( elem->dimension() == SP3D );
        //
        Point alpha, beta;
        iga_transf_3d(elem_prm, alpha, beta);
        //
        blitz::Array<unsigned int, 3> pnt_array(nsample_i, nsample_j, nsample_k);
        //
        for (unsigned int k=0; k<nsample_k; k++) {
            const Real zeta = -1.0 + k * (2.0/(nsample_k-1.0));
            for (unsigned int j=0; j<nsample_j; j++) {
                const Real eta = -1.0 + j * (2.0/(nsample_j-1.0));
                for (unsigned int i=0; i<nsample_i; i++) {
                    const Real xi = -1.0 + i * (2.0/(nsample_i-1.0));
                    //
                    const Point Zp = Point(xi, eta, zeta);
                    const Point Z = self_scale(alpha, Zp) + beta;
                    // initialize and get access to the IGA stuff
                    const Point& P = elem->get_xyz();
                    elem->reinit(Z);
                    //
                    points_map.insert( std::make_pair(n_points, P) );
                    points_prm_coord_map.insert( std::make_pair(n_points, Z) );
                    //
                    pnt_array((int)i,(int)j,(int)k) = n_points;
                    //
                    // increment this index
                    ++n_points;
                }
            }
        }
        //
        for (int k=0; k<nsample_k-1; k++) {
            for (int j=0; j<nsample_j-1; j++) {
                for (int i=0; i<nsample_i-1; i++) {
                    std::vector<unsigned int> conn(8);
                    //
                    conn[0] = pnt_array(i+0,j+0,k+0);
                    conn[1] = pnt_array(i+1,j+0,k+0);
                    conn[2] = pnt_array(i+1,j+1,k+0);
                    conn[3] = pnt_array(i+0,j+1,k+0);
                    conn[4] = pnt_array(i+0,j+0,k+1);
                    conn[5] = pnt_array(i+1,j+0,k+1);
                    conn[6] = pnt_array(i+1,j+1,k+1);
                    conn[7] = pnt_array(i+0,j+1,k+1);
                    //
                    cells_conn_map.insert( std::make_pair(n_cells, conn) );
                    //
                    std::vector<Point> prm_coords;
                    for (int n=0; n<8; n++)
                        prm_coords.push_back(points_prm_coord_map[conn[n]]);
                    //
                    this->post_data.cells_vertices[n_cells] = prm_coords;
                    //
                    this->post_data.nurbs_pelem_cells[elem->id()].find(elem_prm->id())->second.insert(n_cells);
                    // increment this index
                    ++n_cells;
                }
            }
        }
        // end of active (parametric) elements loop...
    }
    //
    for ( std::map<unsigned int, Point>::const_iterator
          citer_map=points_map.begin(); citer_map!=points_map.end(); citer_map++ ) {
        const unsigned int& id = citer_map->first;
        const Point& xyz = citer_map->second;
        //
        this->post_data.es->get_mesh().add_point(xyz, id);
    }
    //
    for ( std::map<unsigned int, std::vector<unsigned int> >::const_iterator
          citer_map=cells_conn_map.begin(); citer_map!=cells_conn_map.end(); citer_map++ ) {
        const unsigned int& id = citer_map->first;
        const std::vector<unsigned int>& conn = citer_map->second;
        //
        Elem* elem = Elem::build(libMesh::HEX8).release();
        elem->set_id(id);
        for (unsigned int i=0; i<8; i++) {
            elem->set_node(i) = this->post_data.es->get_mesh().node_ptr(conn[i]);
        }
        //
        this->post_data.es->get_mesh().add_elem(elem);
    }
    //
    this->post_data.es->get_mesh().prepare_for_use(true);
    // this->post_data.es->get_mesh().print_info();
}
//=================================================================================================
//
//
//
//=================================================================================================
void init_parametric_mesh () {
    const int DIM = control_mesh__ptr->mesh_dimension();
    // define and save (temporarily: Gmsh) the parametric (knot span) mesh in Cartesian space
    std::string fname;
    // only zero-processor will run calculations (avoid redundancy)
    if ( libMesh::global_processor_id() == 0 ) {
        //
        std::map< int, Point > nodes_map;
        std::map< int, Blitz_VectorInt > cells_conn_map;
        std::map< int, int > cells_nurbs_map;
        // local indices
        int inode = 0;
        int ielem = 0;
        // check for NURBS dimension
        if ( DIM == SP1D ) {
            // iterate for all NUBRS in the list
            for ( std::vector<NURBS>::const_iterator
                  n_ci=nurbs__ptr->begin(); n_ci!=nurbs__ptr->end(); n_ci++ ) {
                //
                blitz::Array<int, SP1D> net(n_ci->n_knots(0));
                //
                for (int I=0; I<n_ci->n_knots(0); I++) {
                    const Point& Z = n_ci->nonrepeated_knot(I, 0, 0);
                    const Point& p = n_ci->get_xyz();
                    n_ci->reinit(Z);
                    const int knode = node_present(nodes_map, p);
                    if ( knode == -1 ) {
                        //
                        nodes_map.insert( std::make_pair(inode, p) );
                        net(I) = inode;
                        //
                        ++inode;
                    } else {
                        //
                        net(I) = knode;
                        //
                    }
                    // end of knots loop...
                }
                //
                int I = 0;
                for (int icell=1; icell<=n_ci->span(0); icell++) {
                    Blitz_VectorInt conn(2);
                    conn = net(I+0),
                           net(I+1);
                    ++I;
                    //
                    cells_conn_map.insert( std::make_pair(ielem, conn) );
                    cells_nurbs_map.insert( std::make_pair(ielem, n_ci->id()) );
                    //
                    ++ielem;
                }
                //  end of NURBS loop...
            }
        } else if ( DIM == SP2D ) {
            // iterate for all NUBRS in the list
            for ( std::vector<NURBS>::const_iterator
                  n_ci=nurbs__ptr->begin(); n_ci!=nurbs__ptr->end(); n_ci++ ) {
                //
                blitz::Array<int, SP2D> net(n_ci->n_knots(0), n_ci->n_knots(1));
                //
                for (int J=0; J<n_ci->n_knots(1); J++)
                for (int I=0; I<n_ci->n_knots(0); I++) {
                    const Point& Z = n_ci->nonrepeated_knot(I, J, 0);
                    const Point& p = n_ci->get_xyz();
                    n_ci->reinit(Z);
                    const int knode = node_present(nodes_map, p);
                    if ( knode == -1 ) {
                        //
                        nodes_map.insert( std::make_pair(inode, p) );
                        net(I,J) = inode;
                        //
                        ++inode;
                    } else {
                        //
                        net(I,J) = knode;
                        //
                    }
                    // end of knots loop...
                }
                //
                int I = 0;
                int J = 0;
                for (int icell=1; icell<=n_ci->span(0)*n_ci->span(1); icell++) {
                    Blitz_VectorInt conn(4);
                    conn = net(I+0,J+0), net(I+1,J+0),
                           net(I+1,J+1), net(I+0,J+1);
                    ++I;
                    if ( icell%n_ci->span(0) == 0 ) {
                        I = 0;
                        ++J;
                    }
                    //
                    cells_conn_map.insert( std::make_pair(ielem, conn) );
                    cells_nurbs_map.insert( std::make_pair(ielem, n_ci->id()) );
                    //
                    ++ielem;
                }
                //  end of NURBS loop...
            }
        } else if ( DIM == SP3D ) {
            // iterate for all NUBRS in the list
            for ( std::vector<NURBS>::const_iterator
                  n_ci=nurbs__ptr->begin(); n_ci!=nurbs__ptr->end(); n_ci++ ) {
                //
                blitz::Array<int, SP3D> net(n_ci->n_knots(0), n_ci->n_knots(1), n_ci->n_knots(2));
                //
                for (int K=0; K<n_ci->n_knots(2); K++)
                for (int J=0; J<n_ci->n_knots(1); J++)
                for (int I=0; I<n_ci->n_knots(0); I++) {
                    const Point& Z = n_ci->nonrepeated_knot(I, J, K);
                    const Point& p = n_ci->get_xyz();
                    n_ci->reinit(Z);
                    const int knode = node_present(nodes_map, p);
                    if ( knode == -1 ) {
                        //
                        nodes_map.insert( std::make_pair(inode, p) );
                        net(I,J,K) = inode;
                        //
                        ++inode;
                    } else {
                        //
                        net(I,J,K) = knode;
                        //
                    }
                    // end of knots loop...
                }
                //
                int I = 0;
                int J = 0;
                int K = 0;
                for (int icell=1; icell<=n_ci->span(0)*n_ci->span(1)*n_ci->span(2); icell++) {
                    Blitz_VectorInt conn(8);
                    conn = net(I+0,J+0,K+0), net(I+1,J+0,K+0),
                           net(I+1,J+1,K+0), net(I+0,J+1,K+0),
                           net(I+0,J+0,K+1), net(I+1,J+0,K+1),
                           net(I+1,J+1,K+1), net(I+0,J+1,K+1);
                    ++I;
                    if ( icell%n_ci->span(0) == 0 ) {
                        I = 0;
                        ++J;
                        if ( icell%(n_ci->span(0)*n_ci->span(1)) == 0 ) {
                            J = 0;
                            ++K;
                        }
                    }
                    //
                    cells_conn_map.insert( std::make_pair(ielem, conn) );
                    cells_nurbs_map.insert( std::make_pair(ielem, n_ci->id()) );
                    //
                    ++ielem;
                }
                //  end of NURBS loop...
            }
            // ...else exception is caught...
        } else libmesh_error();
        //
        {
            time_t rawtime;
            time(&rawtime);
            //
            struct tm* timeinfo;
            timeinfo = localtime(&rawtime);
            //
            char buffer[buffer_len];
            strftime(buffer, buffer_len, "%d-%m-%Y-%I-%M-%S", timeinfo);
            std::string ts(buffer);
            //
            char name[2*buffer_len];
            sprintf(name, "%s/NURBS_knots.%s.msh", tmp_dir.c_str(), ts.c_str());
            fname = name;
        }
        // print-out all knot spans into a single GMSH file
        std::ofstream fout(fname.c_str());
        fout << "$MeshFormat\n" << std::flush;
        fout << "2.0 0 8\n" << std::flush;
        fout << "$EndMeshFormat\n" << std::flush;
        fout << "$Nodes\n" << std::flush;
        fout << nodes_map.size() << '\n' << std::flush;
        for (std::map< int, Point >::const_iterator ci=nodes_map.begin(); ci!=nodes_map.end(); ci++) {
            fout << (ci->first+1) << ' ';
            for (int dim=0; dim<SP3D; dim++)
                fout << ' ' << ci->second(dim);
            fout << std::endl;
        }
        fout << "$EndNodes\n" << std::flush;
        fout << "$Elements\n" << std::flush;
        fout << cells_conn_map.size() << '\n' << std::flush;
        std::map< int, int >::const_iterator cj=cells_nurbs_map.begin();
        for (std::map< int, Blitz_VectorInt >::const_iterator ci=cells_conn_map.begin(); ci!=cells_conn_map.end(); ci++) {
            fout << (ci->first+1) << ' ';
            if      ( ci->second.extent(0) == 2 ) fout << " 2 ";
            else if ( ci->second.extent(0) == 4 ) fout << " 3 ";
            else if ( ci->second.extent(0) == 8 ) fout << " 5 ";
            else libmesh_error();
            fout << " 3 " << cj->second << " 0 " << cj->second << " ";
            for (int l=0; l<ci->second.extent(0); l++)
                fout << ' ' << (ci->second(l)+1);
            fout << std::endl;
            cj++;
        }
        fout << "$EndElements\n" << std::flush;
        fout.close();
        //
        if ( libMesh::global_n_processors() > 1) {
            for (unsigned int iproc=1; iproc<libMesh::global_n_processors(); iproc++) {
                gcomm_ptr->send(iproc, fname, Parallel::MessageTag(111));
            }
        }
        // check for other processes...
    } else {
        //
        gcomm_ptr->receive(0, fname, Parallel::MessageTag(111));
        // end of if case...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // parametric grid (temporary object) in Cartesian space for METIS to decompose
    Mesh p_msh(control_mesh__ptr->comm(), DIM);
    // now read the mesh of the knot-spans
    GmshIO(p_msh).read(fname.c_str());
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    if ( libMesh::global_processor_id() == 0 ) {
        char cmd[3*buffer_len];
        sprintf(cmd, "rm %s", fname.c_str());
        if ( system(cmd) ) libmesh_error();
    }
    // now prepare the mesh (renumber nodes, partition elements, etc)
    p_msh.prepare_for_use(true);
    // now start to set the physical mesh in parametric space
    parametric_mesh__ptr->clear();
    //
    std::map< int, Point > nodes_map;
    std::map< int, Blitz_VectorInt > cells_conn_map;
    std::map< int, int > cells_nurbs_map;
    std::vector< Blitz_VectorInt > nurbs_cells_ls(nurbs__ptr->size());
    // local indices
    int inode = 0;
    int ielem = 0;
    // check for NURBS dimension
    if ( DIM == SP1D ) {
        int inurbs = 0;
        // iterate for all NUBRS in the list
        for ( std::vector<NURBS>::const_iterator
              n_ci=nurbs__ptr->begin(); n_ci!=nurbs__ptr->end(); n_ci++ ) {
            //
            blitz::Array<int, SP1D> net(n_ci->n_knots(0));
            //
            for (int I=0; I<n_ci->n_knots(0); I++) {
                const Point& Z = n_ci->nonrepeated_knot(I, 0, 0);
                const int knode = node_present(nodes_map, Z);
                if ( knode == -1 ) {
                    nodes_map.insert( std::make_pair(inode, Z) );
                    net(I) = inode;
                    //
                    ++inode;
                } else net(I) = knode;
                // end of knots loop...
            }
            //
            Blitz_VectorInt s_cells(n_ci->span(0));
            int I = 0;
            for (int icell=1; icell<=n_ci->span(0); icell++) {
                Blitz_VectorInt conn(2);
                conn = net(I+0),
                       net(I+1);
                ++I;
                //
                cells_conn_map.insert( std::make_pair(ielem, conn) );
                cells_nurbs_map.insert( std::make_pair(ielem, n_ci->id()) );
                s_cells(icell-1) = ielem;
                //
                ++ielem;
            }
            //
            nurbs_cells_ls[inurbs].resize(s_cells.size());
            nurbs_cells_ls[inurbs] = s_cells;
            //
            ++inurbs;
            //  end of NURBS loop...
        }
    } else if ( DIM == SP2D ) {
        int inurbs = 0;
        // iterate for all NUBRS in the list
        for ( std::vector<NURBS>::const_iterator
              n_ci=nurbs__ptr->begin(); n_ci!=nurbs__ptr->end(); n_ci++ ) {
            //
            blitz::Array<int, SP2D> net(n_ci->n_knots(0), n_ci->n_knots(1));
            //
            for (int J=0; J<n_ci->n_knots(1); J++)
            for (int I=0; I<n_ci->n_knots(0); I++) {
                const Point& Z = n_ci->nonrepeated_knot(I, J, 0);
                const int knode = node_present(nodes_map, Z);
                if ( knode == -1 ) {
                    nodes_map.insert( std::make_pair(inode, Z) );
                    net(I,J) = inode;
                    //
                    ++inode;
                } else net(I,J) = knode;
                // end of knots loop...
            }
            //
            Blitz_VectorInt s_cells(n_ci->span(0)*n_ci->span(1));
            int I = 0;
            int J = 0;
            for (int icell=1; icell<=n_ci->span(0)*n_ci->span(1); icell++) {
                Blitz_VectorInt conn(4);
                conn = net(I+0,J+0), net(I+1,J+0),
                       net(I+1,J+1), net(I+0,J+1);
                ++I;
                if ( icell%n_ci->span(0) == 0 ) {
                    I = 0;
                    ++J;
                }
                //
                cells_conn_map.insert( std::make_pair(ielem, conn) );
                cells_nurbs_map.insert( std::make_pair(ielem, n_ci->id()) );
                s_cells(icell-1) = ielem;
                //
                ++ielem;
            }
            //
            nurbs_cells_ls[inurbs].resize(s_cells.size());
            nurbs_cells_ls[inurbs] = s_cells;
            //
            ++inurbs;
            //  end of NURBS loop...
        }
    } else if ( DIM == SP3D ) {
        int inurbs = 0;
        // iterate for all NUBRS in the list
        for ( std::vector<NURBS>::const_iterator
              n_ci=nurbs__ptr->begin(); n_ci!=nurbs__ptr->end(); n_ci++ ) {
            //
            blitz::Array<int, SP3D> net(n_ci->n_knots(0), n_ci->n_knots(1), n_ci->n_knots(2));
            //
            for (int K=0; K<n_ci->n_knots(2); K++)
            for (int J=0; J<n_ci->n_knots(1); J++)
            for (int I=0; I<n_ci->n_knots(0); I++) {
                const Point& Z = n_ci->nonrepeated_knot(I, J, K);
                const int knode = node_present(nodes_map, Z);
                if ( knode == -1 ) {
                    nodes_map.insert( std::make_pair(inode, Z) );
                    net(I,J,K) = inode;
                    //
                    ++inode;
                } else net(I,J,K) = knode;
                // end of knots loop...
            }
            //
            Blitz_VectorInt s_cells(n_ci->span(0)*n_ci->span(1)*n_ci->span(2));
            int I = 0;
            int J = 0;
            int K = 0;
            for (int icell=1; icell<=n_ci->span(0)*n_ci->span(1)*n_ci->span(2); icell++) {
                Blitz_VectorInt conn(8);
                conn = net(I+0,J+0,K+0), net(I+1,J+0,K+0),
                       net(I+1,J+1,K+0), net(I+0,J+1,K+0),
                       net(I+0,J+0,K+1), net(I+1,J+0,K+1),
                       net(I+1,J+1,K+1), net(I+0,J+1,K+1);
                ++I;
                if ( icell%n_ci->span(0) == 0 ) {
                    I = 0;
                    ++J;
                    if ( icell%(n_ci->span(0)*n_ci->span(1)) == 0 ) {
                        J = 0;
                        ++K;
                    }
                }
                //
                cells_conn_map.insert( std::make_pair(ielem, conn) );
                cells_nurbs_map.insert( std::make_pair(ielem, n_ci->id()) );
                s_cells(icell-1) = ielem;
                //
                ++ielem;
            }
            //
            nurbs_cells_ls[inurbs].resize(s_cells.size());
            nurbs_cells_ls[inurbs] = s_cells;
            //
            ++inurbs;
            //  end of NURBS loop...
        }
        // ...else exception is caught...
    } else libmesh_error();
    //
    libmesh_assert( inode == nodes_map.size() );
    libmesh_assert( p_msh.n_elem() == cells_conn_map.size() );
    libmesh_assert( p_msh.n_elem() == cells_nurbs_map.size() );
    //
    parametric_mesh__ptr->reserve_nodes(nodes_map.size());
    parametric_mesh__ptr->reserve_elem(p_msh.n_elem());
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    for ( std::map< int, Point >::const_iterator
          ci_n=nodes_map.begin(); ci_n!=nodes_map.end(); ci_n++ ) {
        // add the node into the list
        parametric_mesh__ptr->add_point(ci_n->second, ci_n->first);
        // end of physical-mesh parametric points insertion loop...
    }
    //
    ElemType etype;
    if      ( DIM == SP1D ) etype = EDGE2;
    else if ( DIM == SP2D ) etype = QUAD4;
    else if ( DIM == SP3D ) etype = HEX8;
    else libmesh_error();
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    std::map< int, int >::const_iterator              ci_cnm = cells_nurbs_map.begin();
    std::map< int, Blitz_VectorInt >::const_iterator  ci_ecm = cells_conn_map.begin();
    for ( ; ci_cnm!=cells_nurbs_map.end() && ci_ecm!=cells_conn_map.end() ; ) {
        const Blitz_VectorInt& conn = ci_ecm->second;
        //
        Elem* elem = Elem::build(etype).release();
        elem->set_id(ci_ecm->first);
        for (int i=0; i<conn.extent(0); i++)
            elem->set_node(i) = parametric_mesh__ptr->node_ptr(conn(i));
        elem->subdomain_id() = ci_cnm->second;
        // add the element into the list
        parametric_mesh__ptr->add_elem(elem);
        // increment indices too
        ++ci_cnm;
        ++ci_ecm;
        // end of elements loop...
    }
    //
    parametric_mesh__ptr->prepare_for_use(true);
    parametric_mesh__ptr->partition(1);
    //
    MeshBase::const_element_iterator ci_e = p_msh.elements_begin();
    MeshBase::const_element_iterator ci_e_end = p_msh.elements_end();
    for ( ; ci_e!=ci_e_end; ci_e++ ) {
        const Elem* elem_p = *ci_e;
        // explicitly set the processor ID for the parametric element
        parametric_mesh__ptr->elem_ptr(elem_p->id())->processor_id() = elem_p->processor_id();
        // ...and its neighborhood elements as well
        for (unsigned int iside=0; iside<elem_p->n_sides(); iside++) {
            Elem* elem_neigh = (Elem*) elem_p->neighbor_ptr(iside);
            //
            // parametric element in parametric space
            Elem* p_elem = parametric_mesh__ptr->elem_ptr(elem_p->id());
            // validity check
            if ( p_elem == NULL ) libmesh_error();
            //
            // parametric neighboring/interface element found
            if ( elem_neigh ) {
                // parametric element in parametric space
                Elem* p_elem_neigh = parametric_mesh__ptr->elem_ptr(elem_neigh->id());
                // validity check
                if ( p_elem_neigh == NULL ) libmesh_error();
                //
                p_elem->set_neighbor(iside, p_elem_neigh);
            } else {
                const boundary_id_type b_id = nurbs__ptr->at(p_elem->subdomain_id()).boundary_id(iside);
                //
                p_elem->set_neighbor(iside, NULL);
                //
                parametric_mesh__ptr->boundary_info->add_side(p_elem, iside, b_id);
            }
        }
        // end of elements loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // now localise the control mesh nodes
    if ( DIM == SP1D ) {
        // iterate for all NUBRS in the list
        for ( std::vector<NURBS>::iterator
              n_i=nurbs__ptr->begin(); n_i!=nurbs__ptr->end(); n_i++ ) {
            //
            const Blitz_VectorInt& s_cells = nurbs_cells_ls[n_i->id()];
            const int n_cells = s_cells.size();
            //
            const Real DZ_0 = 1.0 / (n_i->n_ctrl_pnt(0)-1);
            for (int I=0; I<n_i->n_ctrl_pnt(0); I++) {
                Node* node = n_i->set_ctrl_node(I, 0, 0);
                //
                const Point Z(DZ_0*I, 0.0, 0.0);
                for (int icell=0; icell<n_cells; icell++) {
                    const Elem* cell = parametric_mesh__ptr->elem_ptr( s_cells(icell) );
                    libmesh_assert( n_i->id() == cell->subdomain_id() );
                    if ( in_segment( cell->point(0),
                                     cell->point(1), Z ) ) {
                        node->processor_id() = cell->processor_id();
                        break;
                    }
                    // end of cells in this NURBS surface
                }
                // end of knots loop...
            }
            //  end of NURBS loop...
        }
        //
    } else if ( DIM == SP2D ) {
        // iterate for all NUBRS in the list
        for ( std::vector<NURBS>::iterator
              n_i=nurbs__ptr->begin(); n_i!=nurbs__ptr->end(); n_i++ ) {
            //
            const Blitz_VectorInt& s_cells = nurbs_cells_ls[n_i->id()];
            const int n_cells = s_cells.size();
            //
            const Real DZ_1 = 1.0 / (n_i->n_ctrl_pnt(1)-1);
            const Real DZ_0 = 1.0 / (n_i->n_ctrl_pnt(0)-1);
            for (int J=0; J<n_i->n_ctrl_pnt(1); J++)
            for (int I=0; I<n_i->n_ctrl_pnt(0); I++) {
                Node* node = n_i->set_ctrl_node(I, J, 0);
                //
                const Point Z(DZ_0*I, DZ_1*J, 0.0);
                for (int icell=0; icell<n_cells; icell++) {
                    const Elem* cell = parametric_mesh__ptr->elem_ptr( s_cells(icell) );
                    //
                    if ( n_i->id() != cell->subdomain_id() ) libmesh_error();
                    libmesh_assert( n_i->id() == cell->subdomain_id() );
                    if ( in_quadrangle( cell->point(0),
                                        cell->point(2), Z ) ) {
                        node->processor_id() = cell->processor_id();
                        break;
                    }
                    // end of cells in this NURBS surface
                }
                // end of knots loop...
            }
            //  end of NURBS loop...
        }
        //
    } else if ( DIM == SP3D ) {
        // iterate for all NUBRS in the list
        for ( std::vector<NURBS>::iterator
              n_i=nurbs__ptr->begin(); n_i!=nurbs__ptr->end(); n_i++ ) {
            //
            const Blitz_VectorInt& s_cells = nurbs_cells_ls[n_i->id()];
            const int n_cells = s_cells.size();
            //
            const Real DZ_2 = 1.0 / (n_i->n_ctrl_pnt(2)-1);
            const Real DZ_1 = 1.0 / (n_i->n_ctrl_pnt(1)-1);
            const Real DZ_0 = 1.0 / (n_i->n_ctrl_pnt(0)-1);
            for (int K=0; K<n_i->n_ctrl_pnt(2); K++)
            for (int J=0; J<n_i->n_ctrl_pnt(1); J++)
            for (int I=0; I<n_i->n_ctrl_pnt(0); I++) {
                Node* node = n_i->set_ctrl_node(I, J, K);
                //
                const Point Z(DZ_0*I, DZ_1*J, DZ_2*K);
                for (int icell=0; icell<n_cells; icell++) {
                    const Elem* cell = parametric_mesh__ptr->elem_ptr( s_cells(icell) );
                    //
                    if ( n_i->id() != cell->subdomain_id() ) libmesh_error();
                    libmesh_assert( n_i->id() == cell->subdomain_id() );
                    if ( in_cube( cell->point(0),
                                  cell->point(6), Z ) ) {
                        node->processor_id() = cell->processor_id();
                        break;
                    }
                    // end of cells in this NURBS surface
                }
                // end of knots loop...
            }
            //  end of NURBS loop...
        }
        // ...else exception is caught...
    } else libmesh_error();
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    if ( libMesh::global_processor_id() == 0 ) {
        char name[buffer_len];
        sprintf(name, "%s/NURBS.%s.knots-Cartesian.msh", tmp_dir.c_str(), mesh_label.c_str());
        //
        GmshIO(p_msh).write(name);
    }
    if ( libMesh::global_processor_id() == 0 ) {
        char name[buffer_len];
        sprintf(name, "%s/NURBS.%s.knots-parametric.msh", tmp_dir.c_str(), mesh_label.c_str());
        //
        GmshIO(*parametric_mesh__ptr).write(name);
    }
}
//=================================================================================================
void find_neighbors () {
    const int DIM = control_mesh__ptr->mesh_dimension();
    // sanity checks
    if ( nurbs__ptr->empty() ) libmesh_error();
    for (unsigned int l=0; l<nurbs__ptr->size(); l++)
        if ( DIM != nurbs__ptr->at(l).dimension() ) libmesh_error();
    //
    std::map< dof_id_type, Point > nodes_map;
    std::map< dof_id_type, Blitz_VectorInt > cells_conn_map;
    // check for NURBS dimension
    if ( DIM == SP1D ) {
        //
        for ( std::vector<NURBS>::const_iterator
              n_ci=nurbs__ptr->begin(); n_ci!=nurbs__ptr->end(); n_ci++ ) {
            const int n_nodes = 2;
            Blitz_VectorInt conn(n_nodes);
            conn = n_ci->get_ctrl_node(0,                     0, 0)->id(),
                   n_ci->get_ctrl_node(n_ci->n_ctrl_pnt(0)-1, 0, 0)->id();
            cells_conn_map.insert( std::make_pair(n_ci->id(), conn) );
            for (int i=0; i<n_nodes; i++) {
                nodes_map.insert( std::make_pair(conn(i), control_mesh__ptr->point(conn(i))) );
            }
        }
        //
    } else if ( DIM == SP2D ) {
        //
        for ( std::vector<NURBS>::const_iterator
              n_ci=nurbs__ptr->begin(); n_ci!=nurbs__ptr->end(); n_ci++ ) {
            const int n_nodes = 4;
            Blitz_VectorInt conn(n_nodes);
            conn = n_ci->get_ctrl_node(0,                     0,                     0)->id(),
                   n_ci->get_ctrl_node(n_ci->n_ctrl_pnt(0)-1, 0,                     0)->id(),
                   n_ci->get_ctrl_node(n_ci->n_ctrl_pnt(0)-1, n_ci->n_ctrl_pnt(1)-1, 0)->id(),
                   n_ci->get_ctrl_node(0,                     n_ci->n_ctrl_pnt(1)-1, 0)->id();
            cells_conn_map.insert( std::make_pair(n_ci->id(), conn) );
            for (int i=0; i<n_nodes; i++) {
                nodes_map.insert( std::make_pair(conn(i), control_mesh__ptr->point(conn(i))) );
            }
        }
        //
    } else if ( DIM == SP3D ) {
        //
        for ( std::vector<NURBS>::const_iterator
              n_ci=nurbs__ptr->begin(); n_ci!=nurbs__ptr->end(); n_ci++ ) {
            const int n_nodes = 8;
            Blitz_VectorInt conn(n_nodes);
            conn = n_ci->get_ctrl_node(0,                     0,                     0)->id(),
                   n_ci->get_ctrl_node(n_ci->n_ctrl_pnt(0)-1, 0,                     0)->id(),
                   n_ci->get_ctrl_node(n_ci->n_ctrl_pnt(0)-1, n_ci->n_ctrl_pnt(1)-1, 0)->id(),
                   n_ci->get_ctrl_node(0,                     n_ci->n_ctrl_pnt(1)-1, 0)->id(),
                   n_ci->get_ctrl_node(0,                     0,                     n_ci->n_ctrl_pnt(2)-1)->id(),
                   n_ci->get_ctrl_node(n_ci->n_ctrl_pnt(0)-1, 0,                     n_ci->n_ctrl_pnt(2)-1)->id(),
                   n_ci->get_ctrl_node(n_ci->n_ctrl_pnt(0)-1, n_ci->n_ctrl_pnt(1)-1, n_ci->n_ctrl_pnt(2)-1)->id(),
                   n_ci->get_ctrl_node(0,                     n_ci->n_ctrl_pnt(1)-1, n_ci->n_ctrl_pnt(2)-1)->id();
            cells_conn_map.insert( std::make_pair(n_ci->id(), conn) );
            for (int i=0; i<n_nodes; i++) {
                nodes_map.insert( std::make_pair(conn(i), control_mesh__ptr->point(conn(i))) );
            }
        }
        // ...else exception is caught...
    } else libmesh_error();
    // sanity check
    if ( cells_conn_map.size() != nurbs__ptr->size() ) libmesh_error();
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // control grid (temporary object) in Cartesian space to find neighbors
    Mesh c_msh(control_mesh__ptr->comm(), DIM);
    //
    c_msh.reserve_nodes(nodes_map.size());
    c_msh.reserve_elem(cells_conn_map.size());
    // add the points
    for ( std::map< dof_id_type, Point >::const_iterator
          ci_n=nodes_map.begin(); ci_n!=nodes_map.end(); ci_n++ ) {
        // add the node into the list
        c_msh.add_point(ci_n->second, ci_n->first);
        // end of physical-mesh parametric points insertion loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    ElemType etype;
    if      ( DIM == SP1D ) etype = EDGE2;
    else if ( DIM == SP2D ) etype = QUAD4;
    else if ( DIM == SP3D ) etype = HEX8;
    else libmesh_error();
    // add the cells
    for ( std::map< dof_id_type, Blitz_VectorInt >::const_iterator
          ci_ccm=cells_conn_map.begin(); ci_ccm!=cells_conn_map.end(); ci_ccm++ ) {
        const Blitz_VectorInt& conn = ci_ccm->second;
        //
        Elem* elem = Elem::build(etype).release();
        elem->set_id(ci_ccm->first);
        for (int i=0; i<conn.size(); i++)
            elem->set_node(i) = control_mesh__ptr->node_ptr(conn(i));
        // add the element into the list
        c_msh.add_elem(elem);
        // end of elements loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    c_msh.prepare_for_use(true);
    c_msh.partition(1);
    //
    MeshBase::const_element_iterator e_ci = c_msh.elements_begin();
    for ( std::vector<NURBS>::iterator
          n_i=nurbs__ptr->begin(); n_i!=nurbs__ptr->end(); n_i++ ) {
        NURBS* nurbs = &(*n_i);
        //
        Elem* elem = (*e_ci);
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            Elem* neighbor_elem = elem->neighbor_ptr(s);
            // check if there are any neighbors
            if ( neighbor_elem != NULL ) {
                const dof_id_type L = neighbor_elem->id();
                nurbs->set_neighbor(s, &(nurbs__ptr->at(L)));
                // NURBS has not neighbors
            } else {
                //
                std::set<dof_id_type> boundary_control_nodes;
                // NURBS curve
                if        ( DIM == SP1D && 0 == s ) {
                    int I = 0;
                    boundary_control_nodes.insert( nurbs->get_ctrl_node(I)->id() );
                } else if ( DIM == SP1D && 1 == s ) {
                    int I = nurbs->n_ctrl_pnt(0) - 1;
                    boundary_control_nodes.insert( nurbs->get_ctrl_node(I)->id() );
                // NURBS surface
                } else if ( DIM == SP2D && 0 == s ) {
                    for (int I=0; I<nurbs->n_ctrl_pnt(0); I++) {
                        int J = 0;
                        boundary_control_nodes.insert( nurbs->get_ctrl_node(I, J)->id() );
                    }
                } else if ( DIM == SP2D && 1 == s ) {
                    for (int J=0; J<nurbs->n_ctrl_pnt(1); J++) {
                        int I = nurbs->n_ctrl_pnt(0) - 1;
                        boundary_control_nodes.insert( nurbs->get_ctrl_node(I, J)->id() );
                    }
                } else if ( DIM == SP2D && 2 == s ) {
                    for (int I=0; I<nurbs->n_ctrl_pnt(0); I++) {
                        int J = nurbs->n_ctrl_pnt(1) - 1;
                        boundary_control_nodes.insert( nurbs->get_ctrl_node(I, J)->id() );
                    }
                } else if ( DIM == SP2D && 3 == s ) {
                    for (int J=0; J<nurbs->n_ctrl_pnt(1); J++) {
                        int I = 0;
                        boundary_control_nodes.insert( nurbs->get_ctrl_node(I, J)->id() );
                    }
                // NURBS solid
                } else if ( DIM == SP3D && 0 == s ) {
                    for (int I=0; I<nurbs->n_ctrl_pnt(0); I++) {
                        for (int J=0; J<nurbs->n_ctrl_pnt(1); J++) {
                            int K = 0;
                            boundary_control_nodes.insert( nurbs->get_ctrl_node(I, J, K)->id() );
                        }
                    }
                } else if ( DIM == SP3D && 1 == s ) {
                    for (int I=0; I<nurbs->n_ctrl_pnt(0); I++) {
                        for (int K=0; K<nurbs->n_ctrl_pnt(2); K++) {
                            int J = 0;
                            boundary_control_nodes.insert( nurbs->get_ctrl_node(I, J, K)->id() );
                        }
                    }
                } else if ( DIM == SP3D && 2 == s ) {
                    for (int J=0; J<nurbs->n_ctrl_pnt(1); J++) {
                        for (int K=0; K<nurbs->n_ctrl_pnt(2); K++) {
                            int I = nurbs->n_ctrl_pnt(0) - 1;
                            boundary_control_nodes.insert( nurbs->get_ctrl_node(I, J, K)->id() );
                        }
                    }
                } else if ( DIM == SP3D && 3 == s ) {
                    for (int I=0; I<nurbs->n_ctrl_pnt(0); I++) {
                        for (int K=0; K<nurbs->n_ctrl_pnt(2); K++) {
                            int J = nurbs->n_ctrl_pnt(1) - 1;
                            boundary_control_nodes.insert( nurbs->get_ctrl_node(I, J, K)->id() );
                        }
                    }
                } else if ( DIM == SP3D && 4 == s ) {
                    for (int J=0; J<nurbs->n_ctrl_pnt(1); J++) {
                        for (int K=0; K<nurbs->n_ctrl_pnt(2); K++) {
                            int I = 0;
                            boundary_control_nodes.insert( nurbs->get_ctrl_node(I, J, K)->id() );
                        }
                    }
                } else if ( DIM == SP3D && 5 == s ) {
                    for (int I=0; I<nurbs->n_ctrl_pnt(0); I++) {
                        for (int J=0; J<nurbs->n_ctrl_pnt(1); J++) {
                            int K = nurbs->n_ctrl_pnt(2) - 1;
                            boundary_control_nodes.insert( nurbs->get_ctrl_node(I, J, K)->id() );
                        }
                    }
                } else libmesh_error();
                //
                std::set<boundary_id_type> bc_ids;
                // loop over active elements of the control grid
                MeshBase::const_element_iterator citer_el_begin = control_mesh__ptr->active_elements_begin();
                MeshBase::const_element_iterator citer_el_end = control_mesh__ptr->active_elements_end();
                for ( MeshBase::const_element_iterator
                      citer_el=citer_el_begin; citer_el!=citer_el_end; citer_el++ ) {
                    const Elem* elem = *citer_el;
                    //
                    for (unsigned int iside=0; iside<elem->n_sides(); iside++) {
                        if ( elem->neighbor_ptr(iside) != NULL ) continue;
                        //
                        const boundary_id_type bc_id = control_mesh__ptr->boundary_info->boundary_id(elem, iside);
                        UniquePtr<const Elem> side = elem->build_side_ptr(iside);
                        //
                        bool all_found = true;
                        for (unsigned int inode=0; inode<side->n_nodes(); inode++) {
                            if ( boundary_control_nodes.end()                     ==
                                 boundary_control_nodes.find(side->node_id(inode)) )
                                all_found = false;
                        }
                        //
                        if ( all_found ) bc_ids.insert( bc_id );
                        // end of sides loop...
                    }
                    // end of active elements loop...
                }
                // verification checks
                if      ( bc_ids.size() == 0 ) libmesh_error();
                else if ( bc_ids.size() >= 2 ) libmesh_error();
                //
                nurbs->set_boundary_id(s, *bc_ids.begin());
                // end of if case...
            }
            // end of element sides...
        }
        // increment this index
        ++e_ci;
        // end of NURBS loop...
    }
}
//=================================================================================================
void set_local_nodes2nurbs () {
    std::map<dof_id_type, std::set<dof_id_type> > local_nodes2nurbs;
    // sanity checks
    if ( nurbs__ptr->empty() ) libmesh_error();

    MeshBase::const_node_iterator lnode_begin = control_mesh__ptr->local_nodes_begin();
    MeshBase::const_node_iterator lnode_end   = control_mesh__ptr->local_nodes_end();
    // allocate memory for the local nodes
    for ( MeshBase::const_node_iterator
          lnode_ci=lnode_begin; lnode_ci!=lnode_end; lnode_ci++ ) {
        const Node* node = *lnode_ci;
        // insert empty set by default
        std::set<dof_id_type>  tmp_set;
        local_nodes2nurbs.insert( std::make_pair(node->id(), tmp_set) );
        //  end of local nodes loop...
    }
    //
    // loop the local NURBS
    for ( std::vector<NURBS>::const_iterator
          nurbs_ci=nurbs__ptr->begin(); nurbs_ci!=nurbs__ptr->end(); nurbs_ci++ ) {
        const NURBS& nurbs = *nurbs_ci;
        // check all control points of this local NURBS
        for (int A=0; A<nurbs.size(); A++) {
            const dof_id_type node_id = nurbs.get_ctrl_node_indexed(A)->id();
            //
            std::map<dof_id_type, std::set<dof_id_type> >::iterator map_iter = local_nodes2nurbs.find(node_id);
            if ( map_iter == local_nodes2nurbs.end() ) continue;
            //
            map_iter->second.insert( nurbs.id() );
            // end of control points loop...
        }
        // end of local NURBS loop.
    }
    //
    ferr.stream() << " Local nodes (" << local_nodes2nurbs.size() << ") to NURBS " << std::endl;
    for ( std::map<dof_id_type, std::set<dof_id_type> >::const_iterator
          ci=local_nodes2nurbs.begin(); ci!=local_nodes2nurbs.end(); ci++ ) {
        const std::set<dof_id_type>& nurbs_set = ci->second;
        ferr.stream() << ' ' << ci->first;
        for (std::set<dof_id_type>::const_iterator ni=nurbs_set.begin(); ni!=nurbs_set.end(); ni++) {
            ferr.stream() << ' ' << *ni;
        }
        ferr.stream() << ' ' << std::endl;
    }
    ferr.stream() << std::endl;
}
//=================================================================================================
void set_boundary_info () {
    const int DIM = control_mesh__ptr->mesh_dimension();
    // sanity checks
    if ( DIM != SP3D && DIM != SP2D ) libmesh_error();
    if ( DIM == SP1D ) libmesh_error();

    std::map< boundary_id_type, std::set<dof_id_type> > boundary_ctrl_nodes;

    const MeshBase::const_element_iterator begin_el = control_mesh__ptr->active_elements_begin();
    const MeshBase::const_element_iterator end_el = control_mesh__ptr->active_elements_end();
    //
    for (MeshBase::const_element_iterator e_ci=begin_el; e_ci!=end_el; e_ci++) {
        const Elem* elem = *e_ci;
        // check for all sides of the control grid
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            if ( elem->neighbor_ptr(s) != NULL ) continue;
            // create a side element
            UniquePtr<const Elem> side = elem->build_side_ptr(s);
            // get the ID of the boundary
            const boundary_id_type b_id = control_mesh__ptr->boundary_info->boundary_id(elem, s);

            std::map< boundary_id_type, std::set<dof_id_type> >::iterator map_iter = boundary_ctrl_nodes.find(b_id);
            if ( boundary_ctrl_nodes.end() == map_iter ) {
                std::set<dof_id_type> bnodes;
                for (unsigned int i=0; i<side->n_nodes(); i++) {
                    bnodes.insert( side->node_id(i) );
                }
                //
                boundary_ctrl_nodes.insert( std::make_pair(b_id, bnodes) );
            } else {
                for (unsigned int i=0; i<side->n_nodes(); i++)
                    map_iter->second.insert( side->node_id(i) );
            }

            // end of sides loop...
        }
        // end of active elements of the control grid loop...
    }

    const int n_sides = ( DIM == SP3D ? 6 : ( DIM == SP2D ? 4 : 2 ) );

    for (std::vector<NURBS>::iterator n_i=nurbs__ptr->begin(); n_i!=nurbs__ptr->end(); n_i++) {
        NURBS* ielem = &(*n_i);
        //
        for (unsigned int i=0; i<n_sides; i++) {
            ielem->set_neighbor(i, NULL);
        }
    }

    for (std::vector<NURBS>::iterator n_i=nurbs__ptr->begin(); n_i!=nurbs__ptr->end(); n_i++) {
        NURBS* ielem = &(*n_i);
        // check for all sides of the NURBS
        for (unsigned int i=0; i<n_sides; i++) {
            // create a side NURBS
            UniquePtr<const NURBS> iside = ielem->build_side_ptr(i);

            std::set<dof_id_type> iside_nodes;
            for (unsigned int A=0; A<iside->size(); A++) {
                iside_nodes.insert( iside->get_ctrl_node_indexed(A)->id() );
            }

            for (std::vector<NURBS>::iterator n_j=nurbs__ptr->begin(); n_j!=nurbs__ptr->end(); n_j++) {
                NURBS* jelem = &(*n_j);
                //
                if ( ielem->id() == jelem->id() ) continue;
                //
                for (unsigned int j=0; j<n_sides; j++) {
                    // create a side NURBS
                    UniquePtr<const NURBS> jside = jelem->build_side_ptr(j);

                    std::set<dof_id_type> jside_nodes;
                    for (unsigned int B=0; B<jside->size(); B++) {
                        jside_nodes.insert( jside->get_ctrl_node_indexed(B)->id() );
                    }

                    // compare the nodes of the side NURBS
                    if ( iside_nodes == jside_nodes ) {
                        ielem->set_neighbor(i, jelem);
                        jelem->set_neighbor(j, ielem);
                    }
                    // end of NURBS sides loop...
                }
                // end of NURBS loop...
            }

            // find the boundary ID for the NURBS side then...
            if ( ielem->neighbor_ptr(i) == NULL ) {
                // loop for all boundaries
                for ( std::map< boundary_id_type, std::set<dof_id_type> >::const_iterator
                      map_ci=boundary_ctrl_nodes.begin(); map_ci!=boundary_ctrl_nodes.end(); map_ci++ ) {
                    const boundary_id_type b_id = map_ci->first;
                    const std::set<dof_id_type>& bnodes = map_ci->second;
                    //
                    bool all_found = true;
                    for (std::set<dof_id_type>::const_iterator n=iside_nodes.begin(); n!=iside_nodes.end(); n++) {
                        if ( bnodes.end() == bnodes.find(*n) ) {
                            all_found = false;
                            break;
                        }
                    }
                    //
                    if ( all_found ) {
                        ielem->set_boundary_id(i, b_id);
                        break;
                    }
                }
                // end of if case...
            }
            // end of NURBS sides loop...
        }
        // end of NURBS loop...
    }
}
//=================================================================================================
