/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/navier_cauchy_iga_solver_3d.h"
#include "libmesh/mesh.h"
#include "libmesh/equation_systems.h"
#include "libmesh/linear_implicit_system.h"
#include "libmesh/petsc_linear_solver.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/fe.h"
#include "feb3/quadrature.h"
#include "libmesh/dof_map.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/dense_matrix.h"
#include "libmesh/dense_vector.h"
#include "libmesh/dense_submatrix.h"
#include "libmesh/dense_subvector.h"
#include "libmesh/numeric_vector.h"
#include "libmesh/linear_implicit_system.h"
#include "libmesh/boundary_info.h"
#include "feb3/navier_cauchy_solver_3d_aux_inl.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
//=================================================================================================
void NavierCauchyIGASolver3D::_assemble_system__FWD_U () {
    this->_log.push("_assemble_system__FWD_U");

    EquationSystems& es = this->_dmn.get_equation_systems();
    //const MeshBase& msh = es.get_mesh(); // IGA control mesh
    SparseMatrix<Number>*  system_matrix = es.get_system<LinearImplicitSystem>(this->_label+"::iterative_solution").matrix;
    NumericVector<Number>* system_rhs    = es.get_system<LinearImplicitSystem>(this->_label+"::iterative_solution").rhs;
    NumericVector<Number>& residual      = es.get_system<LinearImplicitSystem>(this->_label+"::iterative_solution").get_vector("residual");
    System& u_sys = es.get_system<System>(this->_label+"::solution");
    const System& DU_sys = es.get_system<System>(this->_label+"::incremental_solution");
    //
    const IGA_Context& u_ctx = this->par_list.get<IGA_Context>(this->_label+"::solution");
    const std::vector<NURBS>& nurbs_ls = u_ctx.get_mesh_IGA().get_nurbs();
    const MeshBase::const_element_iterator begin_el = u_ctx.get_mesh_IGA().get_parametric_mesh().active_local_elements_begin();
    const MeshBase::const_element_iterator end_el = u_ctx.get_mesh_IGA().get_parametric_mesh().active_local_elements_end();
    //
    libmesh_assert( SP3D == msh.mesh_dimension() );
    libmesh_assert( SP3D == u_sys.n_vars() );
    //
    const int //increment_step = this->par_list.get<int>("increment step"),
              iteration_step = this->par_list.get<int>("iteration step");
    const bool set_system_matrix = ( ! this->par_list.get<bool>("full Newton-Raphson") && iteration_step != 0 ) ? false : true;
    const bool membrane_present  = this->par_list.get<bool>("membrane elements present"),
               apply_body_force  = this->par_list.get<bool>("apply body force"),
               apply_nodal_force = this->par_list.get<bool>("apply nodal force");
    const bool adaptive_integration = es.parameters.get<bool>("adaptive Gauss quadrature order");
    const bool apply_wound_contraction = this->par_list.get<bool>("apply wound contraction"),
               apply_tumour_growth     = this->par_list.get<bool>("apply tumour growth");
    if ( apply_wound_contraction && apply_tumour_growth ) libmesh_error();
    const unsigned int ninvars_NCS = es.parameters.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = es.parameters.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = es.parameters.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = es.parameters.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    const Real pseudo_time = this->par_list.get<Real>("current pseudo-time");
    const Real current_time = es.parameters.get<Real>("current time");
    Real system_matrix_max_local  = 0.0,
         system_matrix_max_global = 0.0;
    //
    if ( set_system_matrix ) system_matrix->zero();
    system_rhs->zero();
    residual.zero();
    //
    std::set<dof_id_type> local_constrained_nodes_set;
    //
    // loop over all active local (parametric) elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const NURBS* elem = &nurbs_ls[(*citer_el)->subdomain_id()];
        //
        NonUniform_QGauss quad_volu( SP3D,
                                     static_cast<Order>((adaptive_integration ? elem->order(0) : 0)+(int)es.parameters.get<Order>("Gauss quadrature order")),
                                     static_cast<Order>((adaptive_integration ? elem->order(1) : 0)+(int)es.parameters.get<Order>("Gauss quadrature order")),
                                     static_cast<Order>((adaptive_integration ? elem->order(2) : 0)+(int)es.parameters.get<Order>("Gauss quadrature order")) );
        // access the quadrature
        const std::vector< Point >& Qp = quad_volu.get_points();
        const std::vector< Real >& W = quad_volu.get_weights();
        //
        const unsigned int nvars = SP3D;
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
        std::vector<dof_id_type> vars;
        for (unsigned int i=0; i<nvars; i++) vars.push_back( i );
        u_ctx.get_dof_indices(elem, dof_indices, dof_indices_var, &vars);
        //
        std::vector<unsigned int> var_dofs;
        calc_var_dofs(nvars, dof_indices_var, var_dofs);
        // tangent system matrix and residual vector
        DenseMatrix<Number> Ke(dof_indices.size(), dof_indices.size());
        std::vector< DenseSubMatrix<Number> > Kesub(nvars*nvars, Ke);
        DenseVector<Number> Fe(dof_indices.size());
        std::vector< DenseSubVector<Number> > Fesub(nvars, Fe);
        std::vector<unsigned int> offset;
        calc_offset(nvars, var_dofs, offset);
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            for (unsigned int jvar=0; jvar<nvars; jvar++) {
                Kesub[ivar*nvars+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
            }
            Fesub[ivar].reposition(offset[ivar], var_dofs[ivar]);
        }
        // get access to the material region
        libmesh_assert( this->_dmn.regn.end() != this->_dmn.regn.find(elem->subdomain_id()) );
        const RegionInfo* ri_ptr = &(this->_dmn.regn.find(elem->subdomain_id())->second);
        // calculate the parametric element jacobian
        Point alpha, beta;
        const Real Jac_t = iga_transf_3d(*citer_el, alpha, beta);
        // parametric coordinates of the quadrature points
        //const std::vector<Point>& q_pnt = quad_volu.get_points();
        // get access to the element quadrature points data
        libmesh_assert( this->_dmn.elem_qi_local.end() != this->_dmn.elem_qi_local.find((*citer_el)->id()) );
        const QuadratureInfo* qpi_ptr = &(this->_dmn.elem_qi_local.find((*citer_el)->id())->second);
        // loop over volume integration quadrature points
        for (unsigned int qp=0; qp<quad_volu.n_points(); qp++) {
            const Point Z = self_scale(alpha, Qp[qp]) + beta;
            // initialize and get access to the IGA stuff
            const Real& Jac_e = elem->get_jacobian();
            const std::vector< Real >& Phi = elem->get_phi();
            const std::vector< Vector3D >& dPhi_dX = elem->get_dphi();
            elem->reinit(Z);
            const Real JxW = Jac_e*Jac_t*W[qp];
            //
            FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
            fld.current_time = current_time;
            for (unsigned int ivar=0; ivar<SP3D; ivar++) {
                for (unsigned int l=0; l<var_dofs[ivar]; l++) {
                    fld.displacement(ivar) += Phi[l] * u_sys.current_solution(dof_indices_var[ivar][l]);
                    fld.grad_displacement(ivar) += vec(dPhi_dX[l]) * u_sys.current_solution(dof_indices_var[ivar][l]);
                }
            }
            DisplacementGradient grad_incremental_displacement;
            for (unsigned int ivar=0; ivar<SP3D; ivar++) {
                for (unsigned int l=0; l<var_dofs[ivar]; l++) {
                    grad_incremental_displacement(ivar) += vec(dPhi_dX[l]) * DU_sys.current_solution(dof_indices_var[ivar][l]);
                }
            }
            // previous-step inelastic deformation gradient tensor
            fld.deform_grad_inel = qpi_ptr->deform_grad_inel[qp];
            // deformation gradient tensor
            fld.deform_gradient = deformation_gradient_3d(fld.grad_displacement);
            for (unsigned int ivar=0; ivar<nvars_RDS+ndcvars_RDS; ivar++) {
                fld.biochem[ivar] = qpi_ptr->biochem[qp][ivar];
            }
            for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                fld.grad_biochem[ivar] = qpi_ptr->grad_biochem[qp][ivar];
                fld.biochem_rate[ivar] = qpi_ptr->biochem_rate[qp][ivar];
            }
            for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                fld.auxiliary[ivar] = qpi_ptr->auxiliary[qp][ivar];
                fld.grad_auxiliary[ivar] = qpi_ptr->grad_auxiliary[qp][ivar];
                fld.auxiliary_rate[ivar] = qpi_ptr->auxiliary_rate[qp][ivar];
            }
            for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                fld.fibre_ref[ispdm] = qpi_ptr->fibre_ref[ispdm][qp];
                fld.fibre_curr[ispdm] = unit(fld.deform_gradient * fld.fibre_ref[ispdm]);
            }
            fld.strain = GreenLagrange_strain(fld.deform_gradient);
            fld.strain_incr = GreenLagrange_strain(deformation_gradient_3d(grad_incremental_displacement));
            fld.stress = qpi_ptr->stress[qp];
            fld.pressure = qpi_ptr->pressure[qp];
            //
            fld.biomech[0] = ri_ptr->damage_factor_const;
            ri_ptr->calc_damage(fld);
            fld.biomech[1] = ri_ptr->mat->get_biomech_param(0) * (1.0-fld.biomech[0]);
            fld.biomech[2] = qpi_ptr->pressure[qp];
            //
            const Real rho = ri_ptr->mat->get_biomech_param(0) * (1.0-fld.biomech[0]);
            //
            ri_ptr->calc_plastic_deform(fld);
            // second Piola-Kirchhoff stress tensor and tangent stiffness tensor
            Dyadic3D S2pk;
            Quindic3D Dtsm_uu;
            Real Ph;
            Dyadic3D Dtsm_up, Dtsm_pu;
            Real Dtsm_pp;
            // passive contribution of tissue biomechanics
            ri_ptr->calc_biomech_3d(fld, S2pk, Dtsm_uu, Dtsm_up, Ph, Dtsm_pu, Dtsm_pp);
            // wound contraction or tumour growth model
            if ( apply_wound_contraction || apply_tumour_growth )
                ri_ptr->calc_coupled_3d(fld, S2pk, Dtsm_uu, Dtsm_up, Ph, Dtsm_pu, Dtsm_pp);
            // tangent stiffness matrix
            if ( set_system_matrix )
            for (unsigned int a=0; a<var_dofs[0]; a++) {
                for (unsigned int b=0; b<var_dofs[0]; b++) {
                    const Dyadic3D Km = calc_stiffness_forward(dPhi_dX[a], Dtsm_uu, S2pk, dPhi_dX[b], fld.deform_gradient);
                    //
                    Kesub[0*nvars+0](a,b) += JxW * Km(0,0);
                    Kesub[0*nvars+1](a,b) += JxW * Km(0,1);
                    Kesub[0*nvars+2](a,b) += JxW * Km(0,2);
                    Kesub[1*nvars+0](a,b) += JxW * Km(1,0);
                    Kesub[1*nvars+1](a,b) += JxW * Km(1,1);
                    Kesub[1*nvars+2](a,b) += JxW * Km(1,2);
                    Kesub[2*nvars+0](a,b) += JxW * Km(2,0);
                    Kesub[2*nvars+1](a,b) += JxW * Km(2,1);
                    Kesub[2*nvars+2](a,b) += JxW * Km(2,2);
                }
            }
            // internal stresses
            for (unsigned int a=0; a<var_dofs[0]; a++) {
                const Vector3D Rv = calc_internal_force_forward(dPhi_dX[a], S2pk, fld.deform_gradient);
                //
                Fesub[0](a) -= JxW * Rv(0);
                Fesub[1](a) -= JxW * Rv(1);
                Fesub[2](a) -= JxW * Rv(2);
            }
            // external loading (body force)
            if ( apply_body_force )
            for (unsigned int a=0; a<var_dofs[0]; a++) {
                const Vector3D Bf = calc_body_force(Phi[a], rho, this->_dmn.bforce, pseudo_time);
                //
                Fesub[0](a) += JxW * Bf(0);
                Fesub[1](a) += JxW * Bf(1);
                Fesub[2](a) += JxW * Bf(2);
            }
            // end of volume integration quadrature loop...
        }
        // check if there are any constrained boundary nodes
        if ( ! this->_dmn.nodal_bcs.empty() )
        // loop over all NURBS nodes for any constraints
        for (unsigned int ne=0; ne<elem->size(); ne++) {
            const Node* node = elem->get_ctrl_node_indexed(ne);
            // check if NURBS node is constrained
            std::map<unsigned int, BoundaryConditionInfo> ::const_iterator map_citer = this->_dmn.nodal_bcs.find( node->id() );
            if ( this->_dmn.nodal_bcs.end() == map_citer ) continue;
            // check if constrained node is local and if has been already processed
            if ( node->processor_id() != libMesh::global_processor_id() ) continue;
            if ( local_constrained_nodes_set.end() != local_constrained_nodes_set.find(node->id()) ) continue;
            local_constrained_nodes_set.insert(node->id());
            //
            const BoundaryConditionInfo* bci_ptr = &(map_citer->second);
            //
            // distinguish among types of boundary conditions
            if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED    == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_X  == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Y  == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Z  == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XY == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_YZ == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XZ == bci_ptr->type ) {
                ; // do nothing here...
            } else if ( BoundaryConditionInfo::NEUMANN_VECTOR == bci_ptr->type ) {
                const Point traction = Point(bci_ptr->value(0), bci_ptr->value(1), bci_ptr->value(2))
                                     * bci_ptr->calc_time_function(pseudo_time);
                // external loading (traction vector)
                Fesub[0](ne) += traction(0);
                Fesub[1](ne) += traction(1);
                Fesub[2](ne) += traction(2);
                // exception caught for boundary condition type
            } else libmesh_error();
            // end of NURBS nodes loop...
        }
        // find the maximum value of the stiffness matrix (if necessary)
        if ( set_system_matrix )
            system_matrix_max_local = std::max(system_matrix_max_local, Ke.max());
        // update element stiffness matrix and rhs vector
        if ( set_system_matrix )
            system_matrix->add_matrix(Ke, dof_indices);
        system_rhs->add_vector(Fe, dof_indices);
        residual.add_vector(Fe, dof_indices);
        // free memo
        delete [] dof_indices_var;
        // end of active local (parametric) elements loop...
    }
    //
    // loop over all active local (parametric) elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const NURBS* elem = &nurbs_ls[(*citer_el)->subdomain_id()];
        // insert membrane contribution and Neumann boundary conditions
        for (unsigned int s=0; s<(*citer_el)->n_sides(); s++) {
            // check if element has this side on boundary and get the boundary ID for side of element
            if ( (*citer_el)->neighbor_ptr(s) != NULL ) continue;
            // create a side element
            UniquePtr<const NURBS> side = elem->build_side_ptr(s);
            //
            NonUniform_QGauss quad_surf( SP2D,
                                         static_cast<Order>((adaptive_integration ? side->order(0) : 0)+(int)es.parameters.get<Order>("Gauss quadrature order")),
                                         static_cast<Order>((adaptive_integration ? side->order(1) : 0)+(int)es.parameters.get<Order>("Gauss quadrature order")) );
            // access the quadrature
            const std::vector< Point >& Qp = quad_surf.get_points();
            const std::vector< Real >& W = quad_surf.get_weights();
            //
            const unsigned int nvars = SP3D;
            std::vector<dof_id_type> dof_indices;
            std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
            std::vector<dof_id_type> vars;
            for (unsigned int i=0; i<nvars; i++) vars.push_back( i );
            u_ctx.get_dof_indices(side.get(), dof_indices, dof_indices_var, &vars);
            //
            std::vector<unsigned int> var_dofs;
            calc_var_dofs(nvars, dof_indices_var, var_dofs);
            // tangent system matrix and residual vector
            DenseMatrix<Number> Ke(dof_indices.size(), dof_indices.size());
            std::vector< DenseSubMatrix<Number> > Kesub(nvars*nvars, Ke);
            DenseVector<Number> Fe(dof_indices.size());
            std::vector< DenseSubVector<Number> > Fesub(nvars, Fe);
            std::vector<unsigned int> offset;
            calc_offset(nvars, var_dofs, offset);
            for (unsigned int ivar=0; ivar<nvars; ivar++) {
                for (unsigned int jvar=0; jvar<nvars; jvar++) {
                    Kesub[ivar*nvars+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
                }
                Fesub[ivar].reposition(offset[ivar], var_dofs[ivar]);
            }
            // get the ID of the boundary
            const boundary_id_type bc_id = elem->boundary_id(s);
            // check if membrane material region exists
            std::map<region_id_type, RegionInfo>::const_iterator citer_regn = this->_dmn.regn.find(bc_id);
            if ( this->_dmn.regn.end() != citer_regn && membrane_present ) {
                // get access to the membrane material region
                const RegionInfo* ri_ptr = &(citer_regn->second);
                libmesh_assert( ri_ptr->is_membrane );
                // get access to the element quadrature points data
                libmesh_assert( this->_dmn.elem_side_qi_local.end() != this->_dmn.elem_side_qi_local.find((*citer_el)->id()) );
                libmesh_assert( this->_dmn.elem_side_qi_local.find((*citer_el)->id())->second.end() != this->_dmn.elem_side_qi_local.find((*citer_el)->id())->second.find(s) );
                const QuadratureInfo* qpi_ptr = &(this->_dmn.elem_side_qi_local.find((*citer_el)->id())->second.find(s)->second);
                // calculate the parametric element jacobian
                Point alpha, beta;
                const Real Jac_t = iga_transf_3d(*citer_el, alpha, beta, s);
                // loop over surface integration quadrature points
                for (unsigned int qp=0; qp<quad_surf.n_points(); qp++) {
                    const Point Z = self_scale(alpha, Qp[qp]) + beta;
                    // initialize and get access to the IGA stuff
                    const Real& Jac_e  = side->get_jacobian();
                    const std::vector< Real >& Phi = side->get_phi();
                    const std::vector<Real>& dPhi_dZ0 = side->get_dphidz<0>();
                    const std::vector<Real>& dPhi_dZ1 = side->get_dphidz<1>();
                    side->reinit(Z);
                    const Real JxW = Jac_e*Jac_t*W[qp];
                    //
                    FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
                    fld.current_time = current_time;
                    for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                        for (unsigned int l=0; l<var_dofs[ispdm]; l++) {
                            fld.displacement(ispdm) += Phi[l] * u_sys.current_solution(dof_indices_var[ispdm][l]);
                            const RealGradient dPhi_dZ = RealGradient(dPhi_dZ0[l], dPhi_dZ1[l], 0.0);
                            fld.grad_space_reference(ispdm) += dPhi_dZ * side->ctrl_pnt_indexed(l)(ispdm);
                            fld.grad_space_current(ispdm)   += dPhi_dZ * (side->ctrl_pnt_indexed(l)(ispdm)+u_sys.current_solution(dof_indices_var[ispdm][l]));
                        }
                    }
                    Vector3D G_[SP2D];
                    covariant_basis(fld.grad_space_reference, G_);
                    Vector3D g_[SP2D];
                    covariant_basis(fld.grad_space_current, g_);
                    // previous-step inelastic deformation gradient tensor
                    fld.deform_grad_inel = qpi_ptr->deform_grad_inel[qp];
                    // deformation gradient tensor
                    fld.deform_gradient = deformation_gradient_3d(fld.grad_space_reference, fld.grad_space_current);
                    for (unsigned int ivar=0; ivar<nvars_RDS+ndcvars_RDS; ivar++) {
                        fld.biochem[ivar] = qpi_ptr->biochem[qp][ivar];
                    }
                    for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                        fld.grad_biochem[ivar] = qpi_ptr->grad_biochem[qp][ivar];
                        fld.biochem_rate[ivar] = qpi_ptr->biochem_rate[qp][ivar];
                    }
                    for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                        fld.auxiliary[ivar] = qpi_ptr->auxiliary[qp][ivar];
                        fld.grad_auxiliary[ivar] = qpi_ptr->grad_auxiliary[qp][ivar];
                        fld.auxiliary_rate[ivar] = qpi_ptr->auxiliary_rate[qp][ivar];
                    }
                    for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                        fld.fibre_ref[ispdm] = qpi_ptr->fibre_ref[ispdm][qp];
                        fld.fibre_curr[ispdm] = unit(fld.deform_gradient * fld.fibre_ref[ispdm]);
                    }
                    fld.strain = GreenLagrange_strain(fld.deform_gradient);
                    fld.pressure = qpi_ptr->pressure[qp];
                    //
                    fld.biomech[0] = ri_ptr->damage_factor_const;
                    ri_ptr->calc_damage(fld);
                    fld.biomech[1] = ri_ptr->mat->get_biomech_param(0) * (1.0-fld.biomech[0]);
                    fld.biomech[2] = qpi_ptr->pressure[qp];
                    //
                    const Real rho = ri_ptr->mat->get_biomech_param(0) * (1.0-fld.biomech[0]);
                    const Real thick = ri_ptr->mat->get_biomech_param(1) * (1.0-fld.biomech[0]);
                    //
                    ri_ptr->calc_plastic_deform(fld);
                    // second Piola-Kirchhoff stress tensor and tangent stiffness tensor (in local/intrisic coordinates)
                    Dyadic2D S2pk;
                    Quindic2D Dtsm;
                    // passive contribution of tissue biomechanics
                    ri_ptr->calc_biomech_2d(fld, S2pk, Dtsm);
                    // wound contraction or tumour growth model
                    if ( apply_wound_contraction || apply_tumour_growth )
                        ri_ptr->calc_coupled_2d(fld, S2pk, Dtsm);
                    // tangent stiffness matrix
                    if ( set_system_matrix )
                    for (unsigned int a=0; a<var_dofs[0]; a++) {
                        for (unsigned int b=0; b<var_dofs[0]; b++) {
                            const Dyadic3D Km = calc_stiffness_forward(dPhi_dZ0[a], dPhi_dZ1[a], Dtsm, S2pk, dPhi_dZ0[b], dPhi_dZ1[b], G_, g_, thick);
                            //
                            Kesub[0*nvars+0](a,b) += JxW * Km(0,0);
                            Kesub[0*nvars+1](a,b) += JxW * Km(0,1);
                            Kesub[0*nvars+2](a,b) += JxW * Km(0,2);
                            Kesub[1*nvars+0](a,b) += JxW * Km(1,0);
                            Kesub[1*nvars+1](a,b) += JxW * Km(1,1);
                            Kesub[1*nvars+2](a,b) += JxW * Km(1,2);
                            Kesub[2*nvars+0](a,b) += JxW * Km(2,0);
                            Kesub[2*nvars+1](a,b) += JxW * Km(2,1);
                            Kesub[2*nvars+2](a,b) += JxW * Km(2,2);
                        }
                    }
                    // internal stresses
                    for (unsigned int a=0; a<var_dofs[0]; a++) {
                        const Vector3D Rv = calc_internal_force_forward(dPhi_dZ0[a], dPhi_dZ1[a], S2pk, g_, thick);
                        //
                        Fesub[0](a) -= JxW * Rv(0);
                        Fesub[1](a) -= JxW * Rv(1);
                        Fesub[2](a) -= JxW * Rv(2);
                    }
                    // external loading (body force)
                    if ( apply_body_force )
                    for (unsigned int a=0; a<var_dofs[0]; a++) {
                        const Vector3D Bf = calc_body_force(Phi[a], rho*thick, this->_dmn.bforce, pseudo_time);
                        //
                        Fesub[0](a) += JxW * Bf(0);
                        Fesub[1](a) += JxW * Bf(1);
                        Fesub[2](a) += JxW * Bf(2);
                    }
                    // end of surface integration quadrature loop...
                }
                // end of if case...
            }
            // access the boundary condition info data
            const BoundaryConditionInfo* bci_ptr = &(this->_dmn.bcs.find(bc_id)->second);
            // distinguish among types of boundary conditions
            if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED    == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_X  == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Y  == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Z  == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XY == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_YZ == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XZ == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_RD == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_PL == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_PR == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_CY == bci_ptr->type ) {
                ; // do nothing here...
            } else if ( BoundaryConditionInfo::NEUMANN_VECTOR == bci_ptr->type ) {
                const Point traction = Point(bci_ptr->value(0), bci_ptr->value(1), bci_ptr->value(2))
                                     * bci_ptr->calc_time_function(pseudo_time);
                // calculate the parametric element jacobian
                Point alpha, beta;
                const Real Jac_t = iga_transf_3d(*citer_el, alpha, beta, s);
                // check if traction magnitude is above some threshold
                if ( traction.norm() > 1.0e-6 )
                // loop over surface integration quadrature points
                for (unsigned int qp=0; qp<quad_surf.n_points(); qp++) {
                    const Point Z = self_scale(alpha, Qp[qp]) + beta;
                    // initialize and get access to the IGA stuff
                    const Real& Jac_e  = side->get_jacobian();
                    const std::vector< Real >& Phi = side->get_phi();
                    side->reinit(Z);
                    const Real JxW = Jac_e*Jac_t*W[qp];
                    // external loading (traction)
                    for (unsigned int a=0; a<var_dofs[0]; a++) {
                        Vector3D Re;
                        Re.comp = Phi[a]*vec(traction).comp;
                        //
                        Fesub[0](a) += JxW * Re(0);
                        Fesub[1](a) += JxW * Re(1);
                        Fesub[2](a) += JxW * Re(2);
                    }
                    // end of surface integration quadrature loop...
                }
            } else if ( BoundaryConditionInfo::NEUMANN_NORMAL == bci_ptr->type ) {
                const Real pressure = bci_ptr->value(0)*bci_ptr->calc_time_function(pseudo_time);
                // calculate the parametric element jacobian
                Point alpha, beta;
                const Real Jac_t = iga_transf_3d(*citer_el, alpha, beta, s);
                // check if pressure magnitude is above some threshold
                if ( fabs(pressure) > 1.0e-6 )
                // loop over surface integration quadrature points
                for (unsigned int qp=0; qp<quad_surf.n_points(); qp++) {
                    const Point Z = self_scale(alpha, Qp[qp]) + beta;
                    // initialize and get access to the IGA stuff
                    const Real& Jac_e  = side->get_jacobian();
                    const std::vector< Real >& Phi = side->get_phi();
                    const std::vector<Real>&  dPhi_dZ0 = side->get_dphidz<0>();
                    const std::vector<Real>&  dPhi_dZ1 = side->get_dphidz<1>();
                    const std::vector< Vector3D >& dPhi_dX = side->get_dphi();
                    side->reinit(Z);
                    const Real JxW = Jac_e*Jac_t*W[qp];
                    //
                    DisplacementGradient du_dX;
                    for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                        for (unsigned int l=0; l<var_dofs[ispdm]; l++) {
                            du_dX(ispdm) += vec(dPhi_dX[l]) * u_sys.current_solution(dof_indices_var[ispdm][l]);
                        }
                    }
                    //
                    Point dX_dZ0, dX_dZ1;
                    Point dx_dZ0, dx_dZ1;
                    for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                        for (unsigned int l=0; l<var_dofs[ispdm]; l++) {
                            dX_dZ0(ispdm) += dPhi_dZ0[l] * side->ctrl_pnt_indexed(l)(ispdm);
                            dX_dZ1(ispdm) += dPhi_dZ1[l] * side->ctrl_pnt_indexed(l)(ispdm);
                            dx_dZ0(ispdm) += dPhi_dZ0[l] * (side->ctrl_pnt_indexed(l)(ispdm)+u_sys.current_solution(dof_indices_var[ispdm][l]));
                            dx_dZ1(ispdm) += dPhi_dZ1[l] * (side->ctrl_pnt_indexed(l)(ispdm)+u_sys.current_solution(dof_indices_var[ispdm][l]));
                        }
                    }
                    // deformation gradient tensor
                    const Dyadic3D F = deformation_gradient_3d(du_dX);
                    // reciprocal of element Jacobian determinant
                    const Real J_recip = 1.0 / dX_dZ0.cross(dX_dZ1).norm();
                    // normal vector to side element (at current step)
                    const Point normal = dx_dZ0.cross(dx_dZ1);
                    // tangent stiffness matrix
                    if ( set_system_matrix )
                    for (unsigned int a=0; a<var_dofs[0]; a++) {
                        for (unsigned int b=0; b<var_dofs[0]; b++) {
                            Dyadic3D Km_0;
                            Km_0.comp =  0.0,       -dx_dZ1(2),  dx_dZ1(1),
                                         dx_dZ1(2),  0.0,       -dx_dZ1(0),
                                        -dx_dZ1(1),  dx_dZ1(0),  0.0;
                            Km_0.comp *= dPhi_dZ0[b];
                            //
                            Dyadic3D Km_1;
                            Km_1.comp =  0.0,       -dx_dZ0(2),  dx_dZ0(1),
                                         dx_dZ0(2),  0.0,       -dx_dZ0(0),
                                        -dx_dZ0(1),  dx_dZ0(0),  0.0;
                            Km_1.comp *= dPhi_dZ1[b];
                            //
                            Dyadic3D Km;
                            Km.comp = Phi[a]*pressure*(Km_0.comp-Km_1.comp)*J_recip;
                            //
                            Kesub[0*nvars+0](a,b) += JxW * Km(0,0);
                            Kesub[0*nvars+1](a,b) += JxW * Km(0,1);
                            Kesub[0*nvars+2](a,b) += JxW * Km(0,2);
                            Kesub[1*nvars+0](a,b) += JxW * Km(1,0);
                            Kesub[1*nvars+1](a,b) += JxW * Km(1,1);
                            Kesub[1*nvars+2](a,b) += JxW * Km(1,2);
                            Kesub[2*nvars+0](a,b) += JxW * Km(2,0);
                            Kesub[2*nvars+1](a,b) += JxW * Km(2,1);
                            Kesub[2*nvars+2](a,b) += JxW * Km(2,2);
                        }
                    }
                    // external loading (normal traction)
                    for (unsigned int a=0; a<var_dofs[0]; a++) {
                        Vector3D Re;
                        Re.comp = Phi[a]*pressure*vec(normal).comp*J_recip;
                        //
                        Fesub[0](a) += JxW * Re(0);
                        Fesub[1](a) += JxW * Re(1);
                        Fesub[2](a) += JxW * Re(2);
                    }
                    // end of surface integration quadrature loop...
                }
                // exception caught for boundary condition type
            } else libmesh_error();
            // find the maximum value of the stiffness matrix (if necessary)
            if ( set_system_matrix )
                system_matrix_max_local = std::max(system_matrix_max_local, Ke.max());
            // update element stiffness matrix and rhs vector
            if ( set_system_matrix )
                system_matrix->add_matrix(Ke, dof_indices);
            system_rhs->add_vector(Fe, dof_indices);
            residual.add_vector(Fe, dof_indices);
            // free memo
            delete [] dof_indices_var;
            // end of element sides loop...
        }
        // end of active local (parametric) elements loop...
    }
    //
    // check if to add nodal forces contribution
    if ( apply_nodal_force ) {
        const Real fnodal_scale_ = this->_dmn.nforce.calc_time_function__employ(pseudo_time);
        libmesh_assert(fnodal_scale_ >= 0.0);
        if ( fnodal_scale_ > 0.0 ) {
            const NumericVector<Number>& fnodal_part_ = u_sys.get_vector("nodal load (part)");
            //
            system_rhs->add(fnodal_scale_, fnodal_part_);
            residual.add(fnodal_scale_, fnodal_part_);
        }
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // evaluate the penalty parameter for the stiffness matrix
    if ( set_system_matrix ) {
        MPI_Allreduce(&system_matrix_max_local, &system_matrix_max_global, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
        this->par_list.set<Real>("system matrix maximum value") = system_matrix_max_global;
    }
    // incorporate the essential (Dirichlet) boundary conditions
    this->_assemble_system_essential_bcs();
    // mark this flag!
    this->par_list.set<bool>("system matrix updated") = set_system_matrix ? true : false;

    this->_log.pop("_assemble_system__FWD_U");
}
//=================================================================================================
void NavierCauchyIGASolver3D::_set_quad_info__FWD_U () {
    this->_log.push("_set_quad_info__FWD_U");

    EquationSystems& es = this->_dmn.get_equation_systems();
    const MeshBase& msh = es.get_mesh();
    System& u_sys = es.get_system<System>(this->_label+"::solution");
    const System& DU_sys = es.get_system<System>(this->_label+"::incremental_solution");
    //
    const IGA_Context& u_ctx = this->par_list.get<IGA_Context>(this->_label+"::solution");
    const std::vector<NURBS>& nurbs_ls = u_ctx.get_mesh_IGA().get_nurbs();
    const MeshBase::const_element_iterator begin_el = u_ctx.get_mesh_IGA().get_parametric_mesh().active_local_elements_begin();
    const MeshBase::const_element_iterator end_el = u_ctx.get_mesh_IGA().get_parametric_mesh().active_local_elements_end();
    //
    libmesh_assert( SP3D == msh.mesh_dimension() );
    libmesh_assert( SP3D == u_sys.n_vars() );
    //
    const bool membrane_present  = this->par_list.get<bool>("membrane elements present"),
               apply_body_force  = this->par_list.get<bool>("apply body force"),
               apply_nodal_force = this->par_list.get<bool>("apply nodal force");
    const bool adaptive_integration = es.parameters.get<bool>("adaptive Gauss quadrature order");
    const bool apply_wound_contraction = this->par_list.get<bool>("apply wound contraction"),
               apply_tumour_growth     = this->par_list.get<bool>("apply tumour growth");
    if ( apply_wound_contraction && apply_tumour_growth ) libmesh_error();
    const unsigned int ninvars_NCS = es.parameters.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = es.parameters.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = es.parameters.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = es.parameters.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    const Real pseudo_time = this->par_list.get<Real>("current pseudo-time");
    const Real current_time = es.parameters.get<Real>("current time");
    //
    NumericVector<Number>& fnodal = u_sys.get_vector("nodal load");
    if ( apply_nodal_force ) fnodal.zero();
    //
    std::set<dof_id_type> local_constrained_nodes_set;
    //
    // loop over all active local (parametric) elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const NURBS* elem = &nurbs_ls[(*citer_el)->subdomain_id()];
        //
        NonUniform_QGauss quad_volu( SP3D,
                                     static_cast<Order>((adaptive_integration ? elem->order(0) : 0)+(int)es.parameters.get<Order>("Gauss quadrature order")),
                                     static_cast<Order>((adaptive_integration ? elem->order(1) : 0)+(int)es.parameters.get<Order>("Gauss quadrature order")),
                                     static_cast<Order>((adaptive_integration ? elem->order(2) : 0)+(int)es.parameters.get<Order>("Gauss quadrature order")) );
        // access the quadrature
        const std::vector< Point >& Qp = quad_volu.get_points();
        const std::vector< Real >& W = quad_volu.get_weights();
        //
        const unsigned int nvars = SP3D;
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
        std::vector<dof_id_type> vars;
        for (unsigned int i=0; i<nvars; i++) vars.push_back( i );
        u_ctx.get_dof_indices(elem, dof_indices, dof_indices_var, &vars);
        //
        std::vector<unsigned int> var_dofs;
        calc_var_dofs(nvars, dof_indices_var, var_dofs);
        // force vector
        DenseVector<Number> Fe(dof_indices.size());
        std::vector< DenseSubVector<Number> > Fesub(nvars, Fe);
        std::vector<unsigned int> offset;
        calc_offset(nvars, var_dofs, offset);
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            Fesub[ivar].reposition(offset[ivar], var_dofs[ivar]);
        }
        // get access to the material region
        libmesh_assert( this->_dmn.regn.end() != this->_dmn.regn.find(elem->subdomain_id()) );
        const RegionInfo* ri_ptr = &(this->_dmn.regn.find(elem->subdomain_id())->second);
        // calculate the parametric element jacobian
        Point alpha, beta;
        const Real Jac_t = iga_transf_3d(*citer_el, alpha, beta);
        // parametric coordinates of the quadrature points
        //const std::vector<Point>& q_pnt = quad_volu.get_points();
        // get access to the element quadrature points data
        libmesh_assert( this->_dmn.elem_qi_local.end() != this->_dmn.elem_qi_local.find((*citer_el)->id()) );
        QuadratureInfo* qpi_ptr = &(this->_dmn.elem_qi_local.find((*citer_el)->id())->second);
        // loop over volume integration quadrature points
        for (unsigned int qp=0; qp<quad_volu.n_points(); qp++) {
            const Point Z = self_scale(alpha, Qp[qp]) + beta;
            // initialize and get access to the IGA stuff
            const Real& Jac_e = elem->get_jacobian();
            const std::vector< Real >& Phi = elem->get_phi();
            const std::vector< Vector3D >& dPhi_dX = elem->get_dphi();
            elem->reinit(Z);
            const Real JxW = Jac_e*Jac_t*W[qp];
            //
            FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
            fld.current_time = current_time;
            for (unsigned int ivar=0; ivar<SP3D; ivar++) {
                for (unsigned int l=0; l<var_dofs[ivar]; l++) {
                    fld.displacement(ivar) += Phi[l] * u_sys.current_solution(dof_indices_var[ivar][l]);
                    fld.grad_displacement(ivar) += vec(dPhi_dX[l]) * u_sys.current_solution(dof_indices_var[ivar][l]);
                }
            }
            DisplacementGradient grad_incremental_displacement;
            for (unsigned int ivar=0; ivar<SP3D; ivar++) {
                for (unsigned int l=0; l<var_dofs[ivar]; l++) {
                    grad_incremental_displacement(ivar) += vec(dPhi_dX[l]) * DU_sys.current_solution(dof_indices_var[ivar][l]);
                }
            }
            // previous-step inelastic deformation gradient tensor
            fld.deform_grad_inel = qpi_ptr->deform_grad_inel[qp];
            // deformation gradient tensor
            fld.deform_gradient = deformation_gradient_3d(fld.grad_displacement);
            for (unsigned int ivar=0; ivar<nvars_RDS+ndcvars_RDS; ivar++) {
                fld.biochem[ivar] = qpi_ptr->biochem[qp][ivar];
            }
            for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                fld.grad_biochem[ivar] = qpi_ptr->grad_biochem[qp][ivar];
                fld.biochem_rate[ivar] = qpi_ptr->biochem_rate[qp][ivar];
            }
            for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                fld.auxiliary[ivar] = qpi_ptr->auxiliary[qp][ivar];
                fld.grad_auxiliary[ivar] = qpi_ptr->grad_auxiliary[qp][ivar];
                fld.auxiliary_rate[ivar] = qpi_ptr->auxiliary_rate[qp][ivar];
            }
            for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                fld.fibre_ref[ispdm] = qpi_ptr->fibre_ref[ispdm][qp];
                fld.fibre_curr[ispdm] = unit(fld.deform_gradient * fld.fibre_ref[ispdm]);
            }
            fld.strain = GreenLagrange_strain(fld.deform_gradient);
            fld.strain_incr = GreenLagrange_strain(deformation_gradient_3d(grad_incremental_displacement));
            fld.stress = qpi_ptr->stress[qp];
            fld.pressure = qpi_ptr->pressure[qp];
            //
            fld.biomech[0] = ri_ptr->damage_factor_const;
            ri_ptr->calc_damage(fld);
            fld.biomech[1] = ri_ptr->mat->get_biomech_param(0) * (1.0-fld.biomech[0]);
            fld.biomech[2] = qpi_ptr->pressure[qp];
            //
            const Real rho = ri_ptr->mat->get_biomech_param(0) * (1.0-fld.biomech[0]);
            //
            ri_ptr->calc_plastic_deform(fld);
            // second Piola-Kirchhoff stress tensor
            Dyadic3D S2pk;
            Quindic3D Dtsm_uu;
            Real Ph;
            Dyadic3D Dtsm_up, Dtsm_pu;
            Real Dtsm_pp;
            // passive contribution of tissue biomechanics
            ri_ptr->calc_biomech_3d(fld, S2pk, Dtsm_uu, Dtsm_up, Ph, Dtsm_pu, Dtsm_pp);
            // wound contraction or tumour growth model
            if ( apply_wound_contraction || apply_tumour_growth )
                ri_ptr->calc_coupled_3d(fld, S2pk, Dtsm_uu, Dtsm_up, Ph, Dtsm_pu, Dtsm_pp);
            // internal stresses
            if ( apply_nodal_force )
            for (unsigned int a=0; a<var_dofs[0]; a++) {
                const Vector3D Rv = calc_internal_force_forward(dPhi_dX[a], S2pk, fld.deform_gradient);
                //
                Fesub[0](a) += JxW * Rv(0);
                Fesub[1](a) += JxW * Rv(1);
                Fesub[2](a) += JxW * Rv(2);
            }
            // external loading (body force)
            if ( apply_nodal_force )
            if ( apply_body_force )
            for (unsigned int a=0; a<var_dofs[0]; a++) {
                const Vector3D Bf = calc_body_force(Phi[a], rho, this->_dmn.bforce, pseudo_time);
                //
                Fesub[0](a) -= JxW * Bf(0);
                Fesub[1](a) -= JxW * Bf(1);
                Fesub[2](a) -= JxW * Bf(2);
            }
            // update the quadrature point set data
            qpi_ptr->xyz[qp] = qpi_ptr->XYZ[qp] + fld.displacement;
            qpi_ptr->deform_grad[qp] = fld.deform_gradient;
            qpi_ptr->deform_grad_elas[qp] = fld.deform_grad_elas;
            qpi_ptr->deform_grad_inel[qp] = fld.deform_grad_inel;
            qpi_ptr->strain[qp] = fld.strain;
            qpi_ptr->stress[qp] = S2pk;
            qpi_ptr->pressure[qp] = frac_1o3*trace(qpi_ptr->stress[qp]);
            for (unsigned int i=0; i<SP3D; i++)
                qpi_ptr->fibre_curr[i][qp] = fld.fibre_curr[i];
            for (unsigned int i=0; i<ninvars_NCS; i++)
                qpi_ptr->biomech[qp][i] = fld.biomech[i];
            // end of volume integration quadrature loop...
        }
        // check if there are any constrained nodes
        if ( apply_nodal_force )
        if ( ! this->_dmn.nodal_bcs.empty() )
        // loop over all NURBS nodes for any constraints
        for (unsigned int ne=0; ne<elem->size(); ne++) {
            const Node* node = elem->get_ctrl_node_indexed(ne);
            // check if NURBS node is constrained
            std::map<unsigned int, BoundaryConditionInfo> ::const_iterator map_citer = this->_dmn.nodal_bcs.find( node->id() );
            if ( this->_dmn.nodal_bcs.end() == map_citer ) continue;
            // check if constrained node is local and if has been already processed
            if ( node->processor_id() != libMesh::global_processor_id() ) continue;
            if ( local_constrained_nodes_set.end() != local_constrained_nodes_set.find(node->id()) ) continue;
            local_constrained_nodes_set.insert(node->id());
            //
            const BoundaryConditionInfo* bci_ptr = &(map_citer->second);
            //
            // distinguish among types of boundary conditions
            if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED    == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_X  == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Y  == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Z  == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XY == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_YZ == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XZ == bci_ptr->type ) {
                ; // do nothing here...
            } else if ( BoundaryConditionInfo::NEUMANN_VECTOR == bci_ptr->type ) {
                const Point traction = Point(bci_ptr->value(0), bci_ptr->value(1), bci_ptr->value(2))
                                     * bci_ptr->calc_time_function(pseudo_time);
                // external loading (traction vector)
                Fesub[0](ne) -= traction(0);
                Fesub[1](ne) -= traction(1);
                Fesub[2](ne) -= traction(2);
                // exception caught for boundary condition type
            } else libmesh_error();
            // end of NURBS nodes loop...
        }
        // update element nodal force vector
        if ( apply_nodal_force ) fnodal.add_vector(Fe, dof_indices);
        // free memo
        delete [] dof_indices_var;
        // end of active local (parametric) elements loop...
    }
    //
    // loop over all active local (parametric) elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const NURBS* elem = &nurbs_ls[(*citer_el)->subdomain_id()];
        // insert membrane contribution and Neumann boundary conditions
        for (unsigned int s=0; s<(*citer_el)->n_sides(); s++) {
            // check if element has this side on boundary and get the boundary ID for side of element
            if ( (*citer_el)->neighbor_ptr(s) != NULL ) continue;
            // create a side element
            UniquePtr<const NURBS> side = elem->build_side_ptr(s);
            //
            NonUniform_QGauss quad_surf( SP2D,
                                         static_cast<Order>((adaptive_integration ? side->order(0) : 0)+(int)es.parameters.get<Order>("Gauss quadrature order")),
                                         static_cast<Order>((adaptive_integration ? side->order(1) : 0)+(int)es.parameters.get<Order>("Gauss quadrature order")) );
            // access the quadrature
            const std::vector< Point >& Qp = quad_surf.get_points();
            const std::vector< Real >& W = quad_surf.get_weights();
            //
            const unsigned int nvars = SP3D;
            std::vector<dof_id_type> dof_indices;
            std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
            std::vector<dof_id_type> vars;
            for (unsigned int i=0; i<nvars; i++) vars.push_back( i );
            u_ctx.get_dof_indices(side.get(), dof_indices, dof_indices_var, &vars);
            //
            std::vector<unsigned int> var_dofs;
            calc_var_dofs(nvars, dof_indices_var, var_dofs);
            // force vector
            DenseVector<Number> Fe(dof_indices.size());
            std::vector< DenseSubVector<Number> > Fesub(nvars, Fe);
            std::vector<unsigned int> offset;
            calc_offset(nvars, var_dofs, offset);
            for (unsigned int ivar=0; ivar<nvars; ivar++) {
                Fesub[ivar].reposition(offset[ivar], var_dofs[ivar]);
            }
            // get the ID of the boundary
            const boundary_id_type bc_id = elem->boundary_id(s);
            // check if membrane material region exists
            std::map<region_id_type, RegionInfo>::const_iterator citer_regn = this->_dmn.regn.find(bc_id);
            if ( this->_dmn.regn.end() != citer_regn && membrane_present ) {
                // get access to the membrane material region
                const RegionInfo* ri_ptr = &(citer_regn->second);
                libmesh_assert( ri_ptr->is_membrane );
                // get access to the element quadrature points data
                libmesh_assert( this->_dmn.elem_side_qi_local.end() != this->_dmn.elem_side_qi_local.find((*citer_el)->id()) );
                libmesh_assert( this->_dmn.elem_side_qi_local.find((*citer_el)->id())->second.end() != this->_dmn.elem_side_qi_local.find((*citer_el)->id())->second.find(s) );
                QuadratureInfo* qpi_ptr = &(this->_dmn.elem_side_qi_local.find((*citer_el)->id())->second.find(s)->second);
                // calculate the parametric element jacobian
                Point alpha, beta;
                const Real Jac_t = iga_transf_3d(*citer_el, alpha, beta, s);
                // loop over surface integration quadrature points
                for (unsigned int qp=0; qp<quad_surf.n_points(); qp++) {
                    const Point Z = self_scale(alpha, Qp[qp]) + beta;
                    // initialize and get access to the IGA stuff
                    const Real& Jac_e  = side->get_jacobian();
                    const std::vector< Real >& Phi = side->get_phi();
                    const std::vector<Real>& dPhi_dZ0 = side->get_dphidz<0>();
                    const std::vector<Real>& dPhi_dZ1 = side->get_dphidz<1>();
                    side->reinit(Z);
                    const Real JxW = Jac_e*Jac_t*W[qp];
                    //
                    FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
                    fld.current_time = current_time;
                    for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                        for (unsigned int l=0; l<var_dofs[ispdm]; l++) {
                            fld.displacement(ispdm) += Phi[l] * u_sys.current_solution(dof_indices_var[ispdm][l]);
                            const RealGradient dPhi_dZ = RealGradient(dPhi_dZ0[l], dPhi_dZ1[l], 0.0);
                            fld.grad_space_reference(ispdm) += dPhi_dZ * side->ctrl_pnt_indexed(l)(ispdm);
                            fld.grad_space_current(ispdm)   += dPhi_dZ * (side->ctrl_pnt_indexed(l)(ispdm)+u_sys.current_solution(dof_indices_var[ispdm][l]));
                        }
                    }
                    Vector3D G_[SP2D];
                    covariant_basis(fld.grad_space_reference, G_);
                    Vector3D g_[SP2D];
                    covariant_basis(fld.grad_space_current, g_);
                    // previous-step inelastic deformation gradient tensor
                    fld.deform_grad_inel = qpi_ptr->deform_grad_inel[qp];
                    // deformation gradient tensor
                    fld.deform_gradient = deformation_gradient_3d(fld.grad_space_reference, fld.grad_space_current);
                    for (unsigned int ivar=0; ivar<nvars_RDS+ndcvars_RDS; ivar++) {
                        fld.biochem[ivar] = qpi_ptr->biochem[qp][ivar];
                    }
                    for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                        fld.grad_biochem[ivar] = qpi_ptr->grad_biochem[qp][ivar];
                        fld.biochem_rate[ivar] = qpi_ptr->biochem_rate[qp][ivar];
                    }
                    for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                        fld.auxiliary[ivar] = qpi_ptr->auxiliary[qp][ivar];
                        fld.grad_auxiliary[ivar] = qpi_ptr->grad_auxiliary[qp][ivar];
                        fld.auxiliary_rate[ivar] = qpi_ptr->auxiliary_rate[qp][ivar];
                    }
                    for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                        fld.fibre_ref[ispdm] = qpi_ptr->fibre_ref[ispdm][qp];
                        fld.fibre_curr[ispdm] = unit(fld.deform_gradient * fld.fibre_ref[ispdm]);
                    }
                    fld.strain = GreenLagrange_strain(fld.deform_gradient);
                    fld.pressure = qpi_ptr->pressure[qp];
                    //
                    fld.biomech[0] = ri_ptr->damage_factor_const;
                    ri_ptr->calc_damage(fld);
                    fld.biomech[1] = ri_ptr->mat->get_biomech_param(0) * (1.0-fld.biomech[0]);
                    fld.biomech[2] = qpi_ptr->pressure[qp];
                    //
                    const Real rho = ri_ptr->mat->get_biomech_param(0) * (1.0-fld.biomech[0]);
                    const Real thick = ri_ptr->mat->get_biomech_param(1) * (1.0-fld.biomech[0]);
                    //
                    ri_ptr->calc_plastic_deform(fld);
                    // second Piola-Kirchhoff stress tensor (in local/intrisic coordinates)
                    Dyadic2D S2pk;
                    Quindic2D Dtsm;
                    // passive contribution of tissue biomechanics
                    ri_ptr->calc_biomech_2d(fld, S2pk, Dtsm);
                    // wound contraction or tumour growth model
                    if ( apply_wound_contraction || apply_tumour_growth )
                        ri_ptr->calc_coupled_2d(fld, S2pk, Dtsm);
                    // internal stresses
                    if ( apply_nodal_force )
                    for (unsigned int a=0; a<var_dofs[0]; a++) {
                        const Vector3D Rv = calc_internal_force_forward(dPhi_dZ0[a], dPhi_dZ1[a], S2pk, g_, thick);
                        //
                        Fesub[0](a) += JxW * Rv(0);
                        Fesub[1](a) += JxW * Rv(1);
                        Fesub[2](a) += JxW * Rv(2);
                    }
                    // external loading (body force)
                    if ( apply_nodal_force )
                    if ( apply_body_force )
                    for (unsigned int a=0; a<var_dofs[0]; a++) {
                        const Vector3D Bf = calc_body_force(Phi[a], rho*thick, this->_dmn.bforce, pseudo_time);
                        //
                        Fesub[0](a) -= JxW * Bf(0);
                        Fesub[1](a) -= JxW * Bf(1);
                        Fesub[2](a) -= JxW * Bf(2);
                    }
                    // update the quadrature point set data
                    qpi_ptr->xyz[qp] = qpi_ptr->XYZ[qp] + fld.displacement;
                    qpi_ptr->deform_grad[qp] = fld.deform_gradient;
                    qpi_ptr->deform_grad_elas[qp] = fld.deform_grad_elas;
                    qpi_ptr->deform_grad_inel[qp] = fld.deform_grad_inel;
                    qpi_ptr->strain[qp] = fld.strain;
                    qpi_ptr->stress[qp] = pull_backward(inverse(qpi_ptr->deform_grad[qp]), push_forward(g_, S2pk));
                    qpi_ptr->pressure[qp] = frac_1o3*trace(qpi_ptr->stress[qp]);
                    for (unsigned int i=0; i<SP3D; i++)
                        qpi_ptr->fibre_curr[i][qp] = fld.fibre_curr[i];
                    for (unsigned int i=0; i<ninvars_NCS; i++)
                        qpi_ptr->biomech[qp][i] = fld.biomech[i];
                    // end of surface integration quadrature loop...
                }
                // end of if case...
            }
            // access the boundary condition info data
            const BoundaryConditionInfo* bci_ptr = &(this->_dmn.bcs.find(bc_id)->second);
            // external loading (traction vector or normal)
            if ( apply_nodal_force ) {
            // distinguish among types of boundary conditions
            if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED    == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_X  == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Y  == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Z  == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XY == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_YZ == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XZ == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_RD == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_PL == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_PR == bci_ptr->type ||
                 BoundaryConditionInfo::DIRICHLET_PRESCRIBED_CY == bci_ptr->type ) {
                ; // do nothing here...
            } else if ( BoundaryConditionInfo::NEUMANN_VECTOR == bci_ptr->type ) {
                const Point traction = Point(bci_ptr->value(0), bci_ptr->value(1), bci_ptr->value(2))
                                     * bci_ptr->calc_time_function(pseudo_time);
                // calculate the parametric element jacobian
                Point alpha, beta;
                const Real Jac_t = iga_transf_3d(*citer_el, alpha, beta, s);
                // check if traction magnitude is above some threshold
                if ( traction.norm() > 1.0e-6 )
                // loop over surface integration quadrature points
                for (unsigned int qp=0; qp<quad_surf.n_points(); qp++) {
                    const Point Z = self_scale(alpha, Qp[qp]) + beta;
                    // initialize and get access to the IGA stuff
                    const Real& Jac_e  = side->get_jacobian();
                    const std::vector< Real >& Phi = side->get_phi();
                    side->reinit(Z);
                    const Real JxW = Jac_e*Jac_t*W[qp];
                    // external loading (traction)
                    for (unsigned int a=0; a<var_dofs[0]; a++) {
                        Vector3D Re;
                        Re.comp = Phi[a]*vec(traction).comp;
                        //
                        Fesub[0](a) -= JxW * Re(0);
                        Fesub[1](a) -= JxW * Re(1);
                        Fesub[2](a) -= JxW * Re(2);
                    }
                    // end of surface integration quadrature loop...
                }
            } else if ( BoundaryConditionInfo::NEUMANN_NORMAL == bci_ptr->type ) {
                const Real pressure = bci_ptr->value(0)*bci_ptr->calc_time_function(pseudo_time);
                // calculate the parametric element jacobian
                Point alpha, beta;
                const Real Jac_t = iga_transf_3d(*citer_el, alpha, beta, s);
                // check if pressure magnitude is above some threshold
                if ( fabs(pressure) > 1.0e-6 )
                // loop over surface integration quadrature points
                for (unsigned int qp=0; qp<quad_surf.n_points(); qp++) {
                    const Point Z = self_scale(alpha, Qp[qp]) + beta;
                    // initialize and get access to the IGA stuff
                    const Real& Jac_e  = side->get_jacobian();
                    const std::vector< Real >& Phi = side->get_phi();
                    const std::vector<Real>&  dPhi_dZ0 = side->get_dphidz<0>();
                    const std::vector<Real>&  dPhi_dZ1 = side->get_dphidz<1>();
                    const std::vector< Vector3D >& dPhi_dX = side->get_dphi();
                    side->reinit(Z);
                    const Real JxW = Jac_e*Jac_t*W[qp];
                    //
                    DisplacementGradient du_dX;
                    for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                        for (unsigned int l=0; l<var_dofs[ispdm]; l++) {
                            du_dX(ispdm) += vec(dPhi_dX[l]) * u_sys.current_solution(dof_indices_var[ispdm][l]);
                        }
                    }
                    //
                    Point dX_dZ0, dX_dZ1;
                    Point dx_dZ0, dx_dZ1;
                    for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                        for (unsigned int l=0; l<var_dofs[ispdm]; l++) {
                            dX_dZ0(ispdm) += dPhi_dZ0[l] * side->ctrl_pnt_indexed(l)(ispdm);
                            dX_dZ1(ispdm) += dPhi_dZ1[l] * side->ctrl_pnt_indexed(l)(ispdm);
                            dx_dZ0(ispdm) += dPhi_dZ0[l] * (side->ctrl_pnt_indexed(l)(ispdm)+u_sys.current_solution(dof_indices_var[ispdm][l]));
                            dx_dZ1(ispdm) += dPhi_dZ1[l] * (side->ctrl_pnt_indexed(l)(ispdm)+u_sys.current_solution(dof_indices_var[ispdm][l]));
                        }
                    }
                    // deformation gradient tensor
                    const Dyadic3D F = deformation_gradient_3d(du_dX);
                    // reciprocal of element Jacobian determinant
                    const Real J_recip = 1.0 / dX_dZ0.cross(dX_dZ1).norm();
                    // normal vector to side element (at current step)
                    const Point normal = dx_dZ0.cross(dx_dZ1);
                    // external loading (normal traction)
                    for (unsigned int a=0; a<var_dofs[0]; a++) {
                        Vector3D Re;
                        Re.comp = Phi[a]*pressure*vec(normal).comp*J_recip;
                        //
                        Fesub[0](a) -= JxW * Re(0);
                        Fesub[1](a) -= JxW * Re(1);
                        Fesub[2](a) -= JxW * Re(2);
                    }
                    // end of surface integration quadrature loop...
                }
                // exception caught for boundary condition type
            } else libmesh_error();
            // end if ...external loading (traction vector or normal)
            }
            // update element nodal force vector
            if ( apply_nodal_force ) fnodal.add_vector(Fe, dof_indices);
            // free memo
            delete [] dof_indices_var;
            // end of element sides loop...
        }
        // end of active local (parametric) elements loop...
    }
    //
    // close nodal forces vector container if necessary
    if ( apply_nodal_force ) {
        fnodal.close();
        u_sys.update();
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // check if to calculate nodal forces contribution
    if ( apply_nodal_force )
    if ( this->_dmn.nforce.calc_time_function__evaluate(pseudo_time) ) {
        NumericVector<Number>& fnodal_part_ = u_sys.get_vector("nodal load (part)");
        //
        fnodal_part_.close();
        fnodal_part_.zero();
        // check if to the forces for a specified set of nodes or all of them
        if ( this->_dmn.nforce.node_local.empty() ) {
            fnodal_part_.add(fnodal);
        } else {
            for ( std::set<unsigned int>::const_iterator
                  ci=this->_dmn.nforce.node_local.begin(); ci!=this->_dmn.nforce.node_local.end(); ci++ ) {
                const Node& node = msh.node_ref( *ci );
                const dof_id_type idof[] = { node.dof_number(u_sys.number(), 0, 0),
                                             node.dof_number(u_sys.number(), 1, 0),
                                             node.dof_number(u_sys.number(), 2, 0) };
                const Number fnod[] = { fnodal.el(idof[0]),
                                        fnodal.el(idof[1]),
                                        fnodal.el(idof[2]) };
                for (unsigned int ivar=0; ivar<SP3D; ivar++)
                    fnodal_part_.add(idof[ivar], fnod[ivar]);
            }
        }
        fnodal_part_.close();
        u_sys.update();
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        // end of if case...
    }

    this->_log.pop("_set_quad_info__FWD_U");
}
//=================================================================================================
