/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/navier_cauchy_iga_solver_3d.h"
#include "libmesh/linear_implicit_system.h"
#include "libmesh/petsc_linear_solver.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/gmsh_io.h"
#include "libmesh/fe.h"
#include "libmesh/quadrature_gauss.h"
#include "libmesh/dof_map.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/dense_matrix.h"
#include "libmesh/dense_vector.h"
#include "libmesh/dense_submatrix.h"
#include "libmesh/dense_subvector.h"
#include "libmesh/numeric_vector.h"
#include "feb3/iga-context.h"
#include "feb3/fem-context.h"
#include "feb3/paraview_io.h"
#include "feb3/ida.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
extern // list of all systems used in the analysis
std::vector< std::pair<EquationSystems*, std::string> > systems_ls;
// external routines
extern
void update_mesh_lagrange (DomainData&, std::vector< std::vector<Real> >&, const IndexPair& );
//=================================================================================================
#include "feb3/domain_data_inl.h"
#include "feb3/solution_io_inl.h"
//=================================================================================================
NavierCauchyIGASolver3D::NavierCauchyIGASolver3D (DomainData& dmn) :
    _log("NavierCauchyIGASolver3D"), _dmn(dmn) {
    // set the default label
    this->_label = "NavierCauchyIGASolver3D";
    // set parameters to a default value
    this->par_list.set<Real>("current pseudo-time") = 0.0;
    this->par_list.set<bool>("inverse formulation") = false;
    this->par_list.set<bool>("mixed formulation") = false;
    this->par_list.set<bool>("load control") = true;
    this->par_list.set<Real>("relaxing parameter") = 1.0;
    this->par_list.set<bool>("have follower forces") = true;
    this->par_list.set<bool>("apply wound contraction") = false;
    this->par_list.set<bool>("apply tumour growth") = false;
    this->par_list.set<bool>("apply body force") = false;
    this->par_list.set<bool>("apply nodal force") = false;
    this->par_list.set<bool>("save intermediate solution") = true;
    this->par_list.set<bool>("calculate gradients") = false;
    this->par_list.set<bool>("save updated mesh") = false;
    this->par_list.set<bool>("enforce to initialize solution") = true;
}
//=================================================================================================
void NavierCauchyIGASolver3D::init () {
    this->_log.push("init");

    EquationSystems& es = this->_dmn.get_equation_systems();
    Parameters& par = es.parameters;
    std::map<std::string, MeshIGA>* all_msh_iga = par.get<std::map<std::string, MeshIGA>*>("pointer to IGA parametric mesh(es)");
    // get access to the MeshIGA of the state-variables' space
    MeshIGA* msh_iga = &(all_msh_iga->find(par.get<std::string>("IGA parametric mesh to use for analysis"))->second);
    libmesh_assert( msh_iga );
    // set the system parameters
    this->_copy_params();
    // dimension of the FE grid
    const unsigned int nvars = es.get_mesh().mesh_dimension();
    libmesh_assert(nvars == SP3D);
    //
    char var_name[buffer_len];
    int& hvar = par.set<int>("number of variables to hide in Paraview");
    // add the basic system
    do {
        LinearImplicitSystem& sys = es.add_system<LinearImplicitSystem>(this->_label+"::iterative_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__du%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name);
        }
        if ( this->par_list.get<bool>("mixed formulation") ) {
            sprintf(var_name, "%s__dp", this->_label.c_str());
            sys.add_variable(var_name);
        }
        sys.assemble_before_solve = false;
        sys.add_vector("residual");
        IGA_Context* ctx_p = &this->par_list.set<IGA_Context>(this->_label+"::iterative_solution");
        ctx_p->init(sys, msh_iga);
        sys.get_dof_map().attach_extra_send_list_function(send_list_func, ctx_p);
        sys.get_dof_map().attach_extra_sparsity_function(sparsity_func, ctx_p);
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    } while ( false );
    // add another system
    do {
        System& sys = es.add_system<System>(this->_label+"::solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__u%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name);
        }
        if ( this->par_list.get<bool>("mixed formulation") ) {
            sprintf(var_name, "%s__p", this->_label.c_str());
            sys.add_variable(var_name);
        }
        sys.add_vector("nodal load");
        sys.add_vector("nodal load (part)");
        IGA_Context* ctx_p = &this->par_list.set<IGA_Context>(this->_label+"::solution");
        ctx_p->init(sys, msh_iga);
        sys.get_dof_map().attach_extra_send_list_function(send_list_func, ctx_p);
        sys.get_dof_map().attach_extra_sparsity_function(sparsity_func, ctx_p);
        // get access to the solution vector
        this->_dmn.nforce.load_vec = &sys.get_vector("nodal load (part)");
    } while ( false );
    // add another system
    do {
        System& sys = es.add_system<System>(this->_label+"::previous_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__u%d_PREV", this->_label.c_str(), ivar);
            sys.add_variable(var_name);
        }
        if ( this->par_list.get<bool>("mixed formulation") ) {
            sprintf(var_name, "%s__p_PREV", this->_label.c_str());
            sys.add_variable(var_name);
        }
        IGA_Context* ctx_p = &this->par_list.set<IGA_Context>(this->_label+"::previous_solution");
        ctx_p->init(sys, msh_iga);
        sys.get_dof_map().attach_extra_send_list_function(send_list_func, ctx_p);
        sys.get_dof_map().attach_extra_sparsity_function(sparsity_func, ctx_p);
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    } while ( false );
    // add another system
    do {
        System& sys = es.add_system<System>(this->_label+"::incremental_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__DU%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name);
        }
        if ( this->par_list.get<bool>("mixed formulation") ) {
            sprintf(var_name, "%s__DP", this->_label.c_str());
            sys.add_variable(var_name);
        }
        IGA_Context* ctx_p = &this->par_list.set<IGA_Context>(this->_label+"::incremental_solution");
        ctx_p->init(sys, msh_iga);
        sys.get_dof_map().attach_extra_send_list_function(send_list_func, ctx_p);
        sys.get_dof_map().attach_extra_sparsity_function(sparsity_func, ctx_p);
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    } while ( false );
    // add another system
    if ( this->par_list.get<bool>("save intermediate solution") ) {
        System& sys = es.add_system<System>(this->_label+"::intermediate_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__u%d_INTER", this->_label.c_str(), ivar);
            sys.add_variable(var_name);
        }
        if ( this->par_list.get<bool>("mixed formulation") ) {
            sprintf(var_name, "%s__p_INTER", this->_label.c_str());
            sys.add_variable(var_name);
        }
        IGA_Context* ctx_p = &this->par_list.set<IGA_Context>(this->_label+"::intermediate_solution");
        ctx_p->init(sys, msh_iga);
        sys.get_dof_map().attach_extra_send_list_function(send_list_func, ctx_p);
        sys.get_dof_map().attach_extra_sparsity_function(sparsity_func, ctx_p);
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    }
    //
    EquationSystems& es_res = *(msh_iga->post_data.es);
    // add an auxiliary system
    do {
        System& sys = es_res.add_system<System>(this->_label+"::displacement_solution");
        //
        for (unsigned int ivar=0; ivar<SP3D; ivar++) {
            sprintf(var_name, "%s__u%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name, FIRST, LAGRANGE);
        }
        // end of if case...
    } while ( false );
    // add another (auxiliary) system
    if ( this->par_list.get<bool>("calculate gradients") ) {
        System& sys = es_res.add_system<System>(this->_label+"::gradient_solution");
        //
        for (unsigned int ivar=0; ivar<SP3D; ivar++) {
            for (unsigned int jvar=0; jvar<SP3D; jvar++) {
                sprintf(var_name, "%s__Fe%d%d", this->_label.c_str(), ivar, jvar);
                sys.add_variable(var_name, FIRST, LAGRANGE);
            }
        }
        for (unsigned int ivar=0; ivar<SP3D; ivar++) {
            for (unsigned int jvar=0; jvar<SP3D; jvar++) {
                sprintf(var_name, "%s__Fp%d%d", this->_label.c_str(), ivar, jvar);
                sys.add_variable(var_name, FIRST, LAGRANGE);
            }
        }
        for (unsigned int ivar=0; ivar<VOIGT_SP3D; ivar++) {
            const Dual<int> idx = indices_Voigt_3d(ivar);
            sprintf(var_name, "%s__E%d%d", this->_label.c_str(), idx.get_first(), idx.get_second());
            sys.add_variable(var_name, FIRST, LAGRANGE);
        }
        for (unsigned int ivar=0; ivar<VOIGT_SP3D; ivar++) {
            const Dual<int> idx = indices_Voigt_3d(ivar);
            sprintf(var_name, "%s__S%d%d", this->_label.c_str(), idx.get_first(), idx.get_second());
            sys.add_variable(var_name, FIRST, LAGRANGE);
        }
        for (unsigned int ivar=0; ivar<es.parameters.get<unsigned int>("Navier-Cauchy solver: internal variables"); ivar++) {
            sprintf(var_name, "%s__iv%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name, FIRST, LAGRANGE);
        }
        // end of if case...
    }

    this->_log.pop("init");
}
//=================================================================================================
void NavierCauchyIGASolver3D::run () {
    this->_log.push("run");

    EquationSystems& es = this->_dmn.get_equation_systems();
    Parameters& par = es.parameters;
    std::map<std::string, MeshIGA>* all_msh_iga = par.get<std::map<std::string, MeshIGA>*>("pointer to IGA parametric mesh(es)");
    MeshIGA& msh_iga = all_msh_iga->find(es.parameters.get<std::string>("IGA parametric mesh to use for analysis"))->second;
    //
    ParaviewIO pvtu_c(es.get_mesh());
    for (int l=0; l<es.parameters.get<int>("number of variables to hide in Paraview"); l++) {
        char var_name[buffer_len];
        sprintf(var_name, "Paraview variable #%d to hide", l);
        pvtu_c.hide_variable(es.parameters.get<std::string>(var_name));
    }
    //
    ParaviewIO pvtu(msh_iga.post_data.es->get_mesh());
    // initialise the post-processing data for the "ghost" IGA mesh
    msh_iga.post_data.es->init();
    msh_iga.post_data.es->print_info();
    //
    char cstr[buffer_len];
    const std::string results_dir  = es.parameters.get<std::string>("results directory"),
                      output_fname = es.parameters.get<std::string>("output file name"),
                      sim_title    = es.parameters.get<std::string>("simulation title");
    const Real current_time = es.parameters.get<Real>("current time");
    Real& current_pseudo_time  = this->par_list.set<Real>("current pseudo-time");
    Real& previous_pseudo_time = this->par_list.set<Real>("previous pseudo-time");
    Real& pseudo_time_step = this->par_list.set<Real>("pseudo-time step");
    const int ini_incr = this->par_list.get<int>("initial increment");
    int& max_incr = this->par_list.set<int>("maximum increments");
    int& soln_step = this->par_list.set<int>("solution increment");
    int& rstr_step = this->par_list.set<int>("restart increment");
    //
    // set the initial conditions and initialise some member objects
    if ( this->par_list.get<bool>("enforce to initialize solution") ) {
        flog.print(" >> Enforce solution initialization\n", false);
        this->_set_initial();
    // ...else check if to read the restart file(s)
    } else if ( es.parameters.get<unsigned int>("input restart from file") ) {
        flog.print(" >> Load restart file(s)\n", false);
        this->load_restart();
    }
    //
    // write a PVD file to store results
    if ( libMesh::global_processor_id() == 0 ) {
        flog.print(" >> Generate Paraview PVD-file\n", false);
        sprintf( cstr, "%s/%s-%s-%s.pvd",
                 results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), stringfy_number(current_time).c_str() );
        std::ofstream fpvd(cstr);
        fpvd.setf(std::ios_base::scientific|std::ios_base::unitbuf);
        //
        fpvd << "<?xml version=\"1.0\"?>" << std::endl;
        fpvd << "<VTKFile type=\"Collection\" version=\"0.1\" byte_order=\"LittleEndian\">" << std::endl;
        fpvd << "<Collection>" << std::endl;
        for (int incr=ini_incr; incr<=max_incr; incr++) {
            const Real vtk_time = (incr+0.0)/max_incr;
            if ( incr == ini_incr || incr == max_incr || incr%soln_step == 0 ) {
                sprintf( cstr, "paraview/%s-%s-%s-%d.vtu",
                         output_fname.c_str(), sim_title.c_str(), stringfy_number(current_time).c_str(), incr );
                fpvd << "<DataSet timestep=\"" << vtk_time << "\" group=\"\" part=\"0\" file=\"" << cstr << "\"/>" << std::endl;
            }
            // end of time steps loop...
        }
        fpvd << "</Collection>" << std::endl;
        fpvd << "</VTKFile>" << std::endl;
    }
    //
    // write initial results
    flog.print("  > Save solution\n", false);
    sprintf( cstr, "%s/paraview/%s-%s-%s-%d.vtu",
             results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), stringfy_number(current_time).c_str(), ini_incr );
    pvtu.write_equation_systems(cstr, *msh_iga.post_data.es);
    //
    // run for all pseudo-time steps
    this->par_list.set<int>("increment step") = ini_incr;
    do {
        REDO: ;
        current_pseudo_time  += pseudo_time_step;
        previous_pseudo_time += pseudo_time_step;
        //
        const int incr = this->par_list.get<int>("increment step") + 1;
        const int increment = incr / static_cast<int>(pow(2.0, this->par_list.get<int>("maximum increments (current scale)")));
        if ( this->par_list.get<bool>("apply nodal force") ) {
            sprintf( cstr, " >> Pseudo-time %d out of %d (%.10f)\t Fn(Eval):%d\t Fn(Appl):%d\n",
                     incr, max_incr, current_pseudo_time, this->_dmn.nforce.calc_time_function__evaluate(current_pseudo_time), (bool)this->_dmn.nforce.calc_time_function__employ(current_pseudo_time) );
        } else {
            sprintf( cstr, " >> Pseudo-time %d out of %d (%.10f)\n",
                     incr, max_incr, current_pseudo_time );
        }
        flog.print(cstr, false);
        //
        // run for all iteration steps
        const int niters = this->par_list.get<int>("maximum iterations");
        for (int iter=0; iter<niters; iter++) {
            this->par_list.set<int>("iteration step") = iter;
            //
            // assemble the linearized system
            if ( this->par_list.get<bool>("inverse formulation") ) {
                this->par_list.get<bool>("mixed formulation") ?
                this->_assemble_system__INV_UP() : this->_assemble_system__INV_U();
            } else {
                this->par_list.get<bool>("mixed formulation") ?
                this->_assemble_system__FWD_UP() : this->_assemble_system__FWD_U();
            }
            //
            // solve and update solution via Newton-Raphson
            this->_solve();
            //
            // calculate convergence criteria
            const int status = this->_check_convergence();
            // check if to exit loop, abort or continue
            if      ( status == 1 ) break;
            else if ( status <  0 ) {
                //
                // reset solution after failure
                this->_reset_increment(true);
                //
                // check if to resume analysis with increased incrementation
                if ( this->par_list.get<int>("maximum increments (ceiling scale)") >  0                                                            &&
                     this->par_list.get<int>("maximum increments (ceiling scale)") >= this->par_list.get<int>("maximum increments (current scale)") )
                    goto REDO;
                flog.print(" >> Solver is giving up!\n", false);
                //
                // calculate the displacement solution
                this->_calc_displacements();
                // calculate the gradient solution (if necessary)
                if ( this->par_list.get<bool>("calculate gradients") ) {
                    flog.print("  > Calculate gradient solution\n", false);
                    this->_calc_gradients();
                }
                //
                flog.print("  > Save solution\n", false);
                sprintf( cstr, "%s/paraview/%s-%s-%s-%d.vtu",
                         results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), stringfy_number(current_time).c_str(), increment );
                pvtu.write_equation_systems(cstr, *msh_iga.post_data.es);
                // now jump code...
                goto END_LABEL;
            }
            // end of iterations loop...
        }
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        //
        // update quadrature point set data
        flog.print("  > Update quadrature data\n", false);
        if ( this->par_list.get<bool>("inverse formulation") ) {
            this->par_list.get<bool>("mixed formulation") ?
            this->_set_quad_info__INV_UP() : this->_set_quad_info__INV_U();
        } else {
            this->par_list.get<bool>("mixed formulation") ?
            this->_set_quad_info__FWD_UP() : this->_set_quad_info__FWD_U();
        }
        //
        // calculate the displacement solution
        this->_calc_displacements();
        // calculate the gradient solution (if necessary)
        if ( this->par_list.get<bool>("calculate gradients") ) {
            flog.print("  > Calculate gradient solution\n", false);
            this->_calc_gradients();
        }
        //
        // check if to write results
        if ( incr%soln_step == 0 || incr == max_incr ) {
            //
            flog.print("  > Save solution\n", false);
            sprintf( cstr, "%s/paraview/%s-%s-%s-%d.vtu",
                     results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), stringfy_number(current_time).c_str(), increment );
            pvtu.write_equation_systems(cstr, *msh_iga.post_data.es);
        }
        //
        // check if to write restart file
        if ( incr%rstr_step == 0 || incr == max_incr ) {
            flog.print("  > Save restart\n", false);
            sprintf( cstr, "%s/restart/%s-%s-%s-%d.res-%d",
                     results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), stringfy_number(current_time).c_str(), increment, libMesh::global_processor_id() );
            std::ofstream fout(cstr);
            this->save_restart(fout);
        }
        //
        // reset (zero-out) incremental solution for subsequent increment step
        this->_reset_increment();
        // end of pseudo-time steps loop...
    } while ( this->par_list.get<int>("increment step") < max_incr );
    //
    // store a copy of the input Gmsh formatted mesh file
    if ( this->par_list.get<bool>("save updated mesh") ) {
        flog.print(" >> Update Lagrangian mesh\n", false);
        {
            std::vector< std::vector<Real> > solu_u;
            IndexPair solu_idx;
            this->get_displacements("::solution", solu_u, solu_idx);
            update_mesh_lagrange(this->_dmn, solu_u, solu_idx);
        }
        // save the updated (deformed) finite element grid
        if ( libMesh::global_processor_id() == 0 ) {
            flog.print(" >> Save mesh\n", false);
            sprintf( cstr, "%s/%s-%s.msh",
                     results_dir.c_str(), output_fname.c_str(), stringfy_number(current_time).c_str() );
            GmshIO(es.get_mesh()).write(cstr);
            sprintf( cstr, "%s/paraview/%s-%s-boundary.vtu",
                     results_dir.c_str(), output_fname.c_str(), stringfy_number(current_time).c_str() );
            ParaviewIO(es.get_mesh()).write_boundary(cstr);
        }
        // end of if case...
    }
    //
    // check if to update the intermediate solution vector
    if ( this->par_list.get<bool>("save intermediate solution") ) {
        const System& u_sys = es.get_system<System>(this->_label+"::solution");
        System& u_T_sys = es.get_system<System>(this->_label+"::intermediate_solution");
        //
        NumericVector<Number>&  u   = *u_sys.solution;
        NumericVector<Number>&  u_T = *u_T_sys.solution;
        //
        u.close();
        //
        u_T.close();
        u_T.zero();
        u_T.add(u);
        u_T.close();
        u_T_sys.update();
    }
    // end of simulation...
    END_LABEL: ;
    clear_line_(std::cout);
    flog.print(" >> Simulation terminated\n", false);

    this->_log.pop("run");
}
//=================================================================================================
//
//
//
//=================================================================================================
void NavierCauchyIGASolver3D::_copy_params () {
    this->_log.push("_copy_params");

    EquationSystems& es = this->_dmn.get_equation_systems();
    // set the label for the solver and the simulation title (if necessary)
    this->_label = es.parameters.get<std::string>("Navier-Cauchy solver: label");
    if ( "_" == es.parameters.get<std::string>("simulation title") )
        es.parameters.set<std::string>("simulation title") = this->_label;
    //
    if ( es.parameters.get<unsigned int>("input restart from file") )
        this->par_list.set<bool>("enforce to initialize solution") = false;
    //
    this->par_list.set<bool>("full Newton-Raphson") = es.parameters.get<bool>("Navier-Cauchy solver: full Newton-Raphson");
    this->par_list.set<int>("initial increment")  = es.parameters.get<int>("Navier-Cauchy solver: initial increment");
    this->par_list.set<int>("maximum increments") = es.parameters.get<int>("Navier-Cauchy solver: maximum increments");
    this->par_list.set<int>("maximum iterations") = es.parameters.get<int>("Navier-Cauchy solver: maximum iterations");
    this->par_list.set<int>("minimum iterations") = es.parameters.get<int>("Navier-Cauchy solver: minimum iterations");
    this->par_list.set<int>("solution increment") = es.parameters.get<int>("Navier-Cauchy solver: solution increment");
    this->par_list.set<int>("restart increment") = es.parameters.get<int>("Navier-Cauchy solver: restart increment");
    this->par_list.set<Real>("current pseudo-time")  = (this->par_list.get<int>("initial increment")+0.0)
                                                     * es.parameters.get<Real>("Navier-Cauchy solver: window")
                                                     / this->par_list.get<int>("maximum increments");
    this->par_list.set<Real>("previous pseudo-time") = (this->par_list.get<int>("initial increment")-1.0)
                                                     * es.parameters.get<Real>("Navier-Cauchy solver: window")
                                                     / this->par_list.get<int>("maximum increments");
    this->par_list.set<Real>("pseudo-time step") = es.parameters.get<Real>("Navier-Cauchy solver: window")
                                                 / this->par_list.get<int>("maximum increments");
    if ( es.parameters.get<Real>("Navier-Cauchy solver: window")!=1.0 ) libmesh_error();
    this->par_list.set<int>("maximum increments (current scale)") = 0;
    this->par_list.set<int>("maximum increments (ceiling scale)") = es.parameters.get<int>("Navier-Cauchy solver: maximum increments (ceiling scale)");
    this->par_list.set<bool>("abort non-convergence") = es.parameters.get<bool>("Navier-Cauchy solver: abort non-convergence");
    //
    this->par_list.set<Real>("solution tolerance") = es.parameters.get<Real>("Navier-Cauchy solver: solution tolerance");
    this->par_list.set<Real>("residual tolerance") = es.parameters.get<Real>("Navier-Cauchy solver: residual tolerance");
    this->par_list.set<Real>("energy tolerance")   = es.parameters.get<Real>("Navier-Cauchy solver: energy tolerance");
    this->par_list.set<bool>("load control")        = es.parameters.get<bool>("Navier-Cauchy solver: load control");
    this->par_list.set<bool>("calculate gradients") = es.parameters.get<bool>("Navier-Cauchy solver: calculate gradients");
    this->par_list.set<bool>("mixed formulation")   = es.parameters.get<bool>("Navier-Cauchy solver: mixed formulation");
    this->par_list.set<bool>("inverse formulation") = es.parameters.get<bool>("Navier-Cauchy solver: inverse formulation");
    this->par_list.set<bool>("have follower forces") = es.parameters.get<bool>("Navier-Cauchy solver: have follower forces");
    this->par_list.set<bool>("membrane elements present") = es.parameters.get<bool>("membrane elements present");
    this->par_list.set<bool>("apply wound contraction") = es.parameters.get<bool>("Navier-Cauchy solver: apply wound contraction");
    this->par_list.set<bool>("apply tumour growth") = es.parameters.get<bool>("Navier-Cauchy solver: apply tumour growth");
    this->par_list.set<bool>("apply body force") = es.parameters.get<bool>("Navier-Cauchy solver: apply body force");
    this->par_list.set<bool>("apply nodal force") = es.parameters.get<bool>("Navier-Cauchy solver: apply nodal force");
    this->par_list.set<bool>("save updated mesh") = es.parameters.get<bool>("Navier-Cauchy solver: save updated mesh");
    //
    this->par_list.set<unsigned int>("sampling number for NURBS") = es.parameters.get<int>("sampling number for NURBS");

    this->_log.pop("_copy_params");
}
//=================================================================================================
void NavierCauchyIGASolver3D::_set_initial () {
    this->_log.push("_set_initial");

    EquationSystems& es = this->_dmn.get_equation_systems();
    System& du_sys = es.get_system<System>(this->_label+"::iterative_solution");
    System& DU_sys = es.get_system<System>(this->_label+"::incremental_solution");
    System& u_sys   = es.get_system<System>(this->_label+"::solution");
    System& u_P_sys = es.get_system<System>(this->_label+"::previous_solution");
    //
    {
        NumericVector<Number>& du = *du_sys.solution;
        //
        du.zero();
        du.close();
        du_sys.update();
    }
    //
    {
        NumericVector<Number>& DU = *DU_sys.solution;
        //
        DU.zero();
        DU.close();
        DU_sys.update();
    }
    //
    {
        NumericVector<Number>& u = *u_sys.solution;
        NumericVector<Number>& f = u_sys.get_vector("nodal load");
        NumericVector<Number>& f_p = u_sys.get_vector("nodal load (part)");
        //
        u.zero();
        u.close();
        f.zero();
        f.close();
        f_p.zero();
        f_p.close();
        u_sys.update();
    }
    //
    {
        NumericVector<Number>& u = *u_P_sys.solution;
        //
        u.zero();
        u.close();
        u_P_sys.update();
    }
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    // update quadrature point set data
    flog.print("  > Update quadrature data\n", false);
    if ( this->par_list.get<bool>("inverse formulation") ) {
        this->par_list.get<bool>("mixed formulation") ?
        this->_set_quad_info__INV_UP() : this->_set_quad_info__INV_U();
    } else {
        this->par_list.get<bool>("mixed formulation") ?
        this->_set_quad_info__FWD_UP() : this->_set_quad_info__FWD_U();
    }
    //
    // calculate the displacement solution
    this->_calc_displacements();
    // calculate the gradient solution (if necessary)
    if ( this->par_list.get<bool>("calculate gradients") ) {
        flog.print("  > Calculate gradient solution\n", false);
        this->_calc_gradients();
    }

    this->_log.pop("_set_initial");
}
//=================================================================================================
void NavierCauchyIGASolver3D::_assemble_system_essential_bcs () {
    this->_log.push("_assemble_system_essential_bcs");

    EquationSystems& es = this->_dmn.get_equation_systems();
    //const MeshBase& msh = es.get_mesh(); // IGA control mesh
    SparseMatrix<Number>*  system_matrix = es.get_system<LinearImplicitSystem>(this->_label+"::iterative_solution").matrix;
    NumericVector<Number>* system_rhs    = es.get_system<LinearImplicitSystem>(this->_label+"::iterative_solution").rhs;
    NumericVector<Number>& residual      = es.get_system<LinearImplicitSystem>(this->_label+"::iterative_solution").get_vector("residual");
    const System& DU_sys = es.get_system<System>(this->_label+"::incremental_solution");
    // const System& u_T_sys = this->par_list.get<bool>("save intermediate solution")
    //                       ? es.get_system<System>(this->_label+"::intermediate_solution") : es.get_system<System>(this->_label+"::solution");
    //
    const IGA_Context& DU_ctx = this->par_list.get<IGA_Context>(this->_label+"::incremental_solution");
    const std::vector<NURBS>& nurbs_ls = DU_ctx.get_mesh_IGA().get_nurbs();
    const MeshBase::const_element_iterator begin_el = DU_ctx.get_mesh_IGA().get_parametric_mesh().active_local_elements_begin();
    const MeshBase::const_element_iterator end_el = DU_ctx.get_mesh_IGA().get_parametric_mesh().active_local_elements_end();
    //
    libmesh_assert( SP3D == msh.mesh_dimension() );
    libmesh_assert( (this->par_list.get<bool>("mixed formulation") ? SP3D+1 : SP3D) == DU_sys.n_vars() );
    std::vector<unsigned short int> var;
    DU_ctx.get_var_list(var);
    //
    const int //increment_step = this->par_list.get<int>("increment step"),
              iteration_step = this->par_list.get<int>("iteration step");
    const bool set_system_matrix = ( ! this->par_list.get<bool>("full Newton-Raphson") && iteration_step != 0 ) ? false : true;
    const Real previous_time = this->par_list.get<Real>("previous pseudo-time"),
               current_time  = this->par_list.get<Real>("current pseudo-time");
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // set the penalty parameter
    const Real penalty = this->par_list.get<Real>("system matrix maximum value")
                       * es.parameters.get<Real>("penalty parameter scale");
    //
    // loop over all active local (parametric) elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const NURBS* elem = &nurbs_ls[(*citer_el)->subdomain_id()];
        //
        const unsigned int nvars = this->par_list.get<bool>("mixed formulation") ? SP3D+1 : SP3D;
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
        std::vector<dof_id_type> vars;
        for (unsigned int i=0; i<nvars; i++) vars.push_back( i );
        DU_ctx.get_dof_indices(elem, dof_indices, dof_indices_var, &vars);
        //
        std::vector<unsigned int> var_dofs;
        calc_var_dofs(nvars, dof_indices_var, var_dofs);
        //
        DenseMatrix<Number> Ke(dof_indices.size(), dof_indices.size());
        std::vector< DenseSubMatrix<Number> > Kesub(nvars*nvars, Ke);
        DenseVector<Number> Fe(dof_indices.size());
        std::vector< DenseSubVector<Number> > Fesub(nvars, Fe);
        std::vector<unsigned int> offset;
        calc_offset(nvars, var_dofs, offset);
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            for (unsigned int jvar=0; jvar<nvars; jvar++) {
                Kesub[ivar*nvars+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
            }
            Fesub[ivar].reposition(offset[ivar], var_dofs[ivar]);
        }
        //
        // loop over all NURBS nodes for any essential constraints
        for (unsigned int ne=0; ne<elem->size(); ne++) {
            const Node* node = elem->get_ctrl_node_indexed(ne);
            // check if NURBS node is constrained
            std::map<unsigned int, BoundaryConditionInfo> ::const_iterator map_citer = this->_dmn.nodal_bcs.find( node->id() );
            if ( this->_dmn.nodal_bcs.end() == map_citer ) continue;
            if ( node->processor_id() != libMesh::global_processor_id() ) continue;
            //
            const BoundaryConditionInfo* bci_ptr = &(map_citer->second);
            // check for active boundary condition
            if ( bci_ptr->calc_active_function(current_time) ) {
            // distinguish among types of boundary conditions
            if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment,
                           u1_increment = bci_ptr->value(1)*U_increment,
                           u2_increment = bci_ptr->value(2)*U_increment;
                if ( set_system_matrix ) {
                    Kesub[0*nvars+0](ne,ne) += penalty;
                    Kesub[1*nvars+1](ne,ne) += penalty;
                    Kesub[2*nvars+2](ne,ne) += penalty;
                }
                if ( iteration_step == 0 ) {
                    Fesub[0](ne) += penalty * u0_increment;
                    Fesub[1](ne) += penalty * u1_increment;
                    Fesub[2](ne) += penalty * u2_increment;
                } else {
                    const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                               DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                               DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                    Fesub[0](ne) += penalty * (u0_increment-DU0);
                    Fesub[1](ne) += penalty * (u1_increment-DU1);
                    Fesub[2](ne) += penalty * (u2_increment-DU2);
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_X == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment;
                if ( set_system_matrix ) {
                    Kesub[0*nvars+0](ne,ne) += penalty;
                }
                if ( iteration_step == 0 ) {
                    Fesub[0](ne) += penalty * u0_increment;
                } else {
                    const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) );
                    Fesub[0](ne) += penalty * (u0_increment-DU0);
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Y == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u1_increment = bci_ptr->value(1)*U_increment;
                if ( set_system_matrix ) {
                    Kesub[1*nvars+1](ne,ne) += penalty;
                }
                if ( iteration_step == 0 ) {
                    Fesub[1](ne) += penalty * u1_increment;
                } else {
                    const Real DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) );
                    Fesub[1](ne) += penalty * (u1_increment-DU1);
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Z == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u2_increment = bci_ptr->value(2)*U_increment;
                if ( set_system_matrix ) {
                    Kesub[2*nvars+2](ne,ne) += penalty;
                }
                if ( iteration_step == 0 ) {
                    Fesub[2](ne) += penalty * u2_increment;
                } else {
                    const Real DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                    Fesub[2](ne) += penalty * (u2_increment-DU2);
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XY == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment,
                           u1_increment = bci_ptr->value(1)*U_increment;
                if ( set_system_matrix ) {
                    Kesub[0*nvars+0](ne,ne) += penalty;
                    Kesub[1*nvars+1](ne,ne) += penalty;
                }
                if ( iteration_step == 0 ) {
                    Fesub[0](ne) += penalty * u0_increment;
                    Fesub[1](ne) += penalty * u1_increment;
                } else {
                    const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                               DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) );
                    Fesub[0](ne) += penalty * (u0_increment-DU0);
                    Fesub[1](ne) += penalty * (u1_increment-DU1);
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_YZ == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u1_increment = bci_ptr->value(1)*U_increment,
                           u2_increment = bci_ptr->value(2)*U_increment;
                if ( set_system_matrix ) {
                    Kesub[1*nvars+1](ne,ne) += penalty;
                    Kesub[2*nvars+2](ne,ne) += penalty;
                }
                if ( iteration_step == 0 ) {
                    Fesub[1](ne) += penalty * u1_increment;
                    Fesub[2](ne) += penalty * u2_increment;
                } else {
                    const Real DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                               DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                    Fesub[1](ne) += penalty * (u1_increment-DU1);
                    Fesub[2](ne) += penalty * (u2_increment-DU2);
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XZ == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment,
                           u2_increment = bci_ptr->value(2)*U_increment;
                if ( set_system_matrix ) {
                    Kesub[0*nvars+0](ne,ne) += penalty;
                    Kesub[2*nvars+2](ne,ne) += penalty;
                }
                if ( iteration_step == 0 ) {
                    Fesub[0](ne) += penalty * u0_increment;
                    Fesub[2](ne) += penalty * u2_increment;
                } else {
                    const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                               DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                    Fesub[0](ne) += penalty * (u0_increment-DU0);
                    Fesub[2](ne) += penalty * (u2_increment-DU2);
                }
            } else if ( BoundaryConditionInfo::NEUMANN_VECTOR == bci_ptr->type ) {
                ; // do nothing here
            } else libmesh_error();
            // end if ...check for active boundary condition
            }
            // end of NURBS nodes loop...
        }
        // update element stiffness matrix and rhs vector
        if ( set_system_matrix )
            system_matrix->add_matrix(Ke, dof_indices);
        system_rhs->add_vector(Fe, dof_indices);
        residual.add_vector(Fe, dof_indices);
        // free memo
        delete [] dof_indices_var;
        // end of active local (parametric) elements loop...
    }
    //
    // loop over all active local (parametric) elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const NURBS* elem = &nurbs_ls[(*citer_el)->subdomain_id()];
        //
        // insert essential BCs for "boundary" side elements
        for (unsigned int s=0; s<(*citer_el)->n_sides(); s++) {
            // check if element has this side on boundary and get the boundary ID for side of element
            if ( (*citer_el)->neighbor_ptr(s) != NULL ) continue;
            //
            const boundary_id_type bc_id = elem->boundary_id(s);
            const BoundaryConditionInfo* bci_ptr = &(this->_dmn.bcs.find(bc_id)->second);
            // create a side element
            UniquePtr<const NURBS> side = elem->build_side_ptr(s);
            //
            const unsigned int nvars = SP3D;
            std::vector<dof_id_type> dof_indices;
            std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
            std::vector<dof_id_type> vars;
            for (unsigned int i=0; i<nvars; i++) vars.push_back( i );
            DU_ctx.get_dof_indices(side.get(), dof_indices, dof_indices_var, &vars);
            //
            std::vector<unsigned int> var_dofs;
            calc_var_dofs(nvars, dof_indices_var, var_dofs);
            //
            DenseMatrix<Number> Ke(dof_indices.size(), dof_indices.size());
            std::vector< DenseSubMatrix<Number> > Kesub(nvars*nvars, Ke);
            DenseVector<Number> Fe(dof_indices.size());
            std::vector< DenseSubVector<Number> > Fesub(nvars, Fe);
            std::vector<unsigned int> offset;
            calc_offset(nvars, var_dofs, offset);
            for (unsigned int ivar=0; ivar<nvars; ivar++) {
                for (unsigned int jvar=0; jvar<nvars; jvar++) {
                    Kesub[ivar*nvars+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
                }
                Fesub[ivar].reposition(offset[ivar], var_dofs[ivar]);
            }
            // check for active boundary condition
            if ( bci_ptr->calc_active_function(current_time) ) {
            // distinguish among types of boundary conditions
            if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_RD == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real ur_increment = bci_ptr->value(0)*U_increment;
                if ( fabs(U_increment) > 1.0e-6 ) // NOTE: this check is valid only for this type of BC
                for (unsigned int ne=0; ne<side->size(); ne++) {
                    const Node* node = side->get_ctrl_node_indexed(ne);
                    Point radial = *node - bci_ptr->center;
                    if ( radial.norm() < 1.0e-6 ) radial = 0.0;
                    else radial = radial.unit();
                    const Real u0_increment = ur_increment*radial(0),
                               u1_increment = ur_increment*radial(1),
                               u2_increment = ur_increment*radial(2);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                        Kesub[1*nvars+1](ne,ne) += penalty;
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                        Fesub[1](ne) += penalty * u1_increment;
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                                   DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                                   DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_CY == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real ur_increment = bci_ptr->value(0)*U_increment;
                if ( fabs(U_increment) > 1.0e-6 ) // NOTE: this check is valid only for this type of BC
                for (unsigned int ne=0; ne<side->size(); ne++) {
                    const Node* node = side->get_ctrl_node_indexed(ne);
                    const Point node_proj = bci_ptr->center + ((*node-bci_ptr->center)*bci_ptr->normal) * bci_ptr->normal - *node;
                    const Real u0_increment = ur_increment*node_proj(0),
                               u1_increment = ur_increment*node_proj(1),
                               u2_increment = ur_increment*node_proj(2);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                        Kesub[1*nvars+1](ne,ne) += penalty;
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                        Fesub[1](ne) += penalty * u1_increment;
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                                   DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                                   DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_PL == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real un_increment = bci_ptr->value(0)*U_increment;
                if ( fabs(U_increment) > 1.0e-6 ) // NOTE: this check is valid only for this type of BC
                for (unsigned int ne=0; ne<side->size(); ne++) {
                    const Node* node = side->get_ctrl_node_indexed(ne);
                    const Point node_proj = project_to_plane(*node, bci_ptr->center, bci_ptr->normal) - *node;
                    const Real u0_increment = un_increment*node_proj(0),
                               u1_increment = un_increment*node_proj(1),
                               u2_increment = un_increment*node_proj(2);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                        Kesub[1*nvars+1](ne,ne) += penalty;
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                        Fesub[1](ne) += penalty * u1_increment;
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                                   DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                                   DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_PR == bci_ptr->type ) {
                const Real angle_1 = bci_ptr->value(0)*bci_ptr->calc_time_function(current_time);
                const Real angle_0 = bci_ptr->value(0)*bci_ptr->calc_time_function(previous_time);
                for (unsigned int ne=0; ne<side->size(); ne++) {
                    const Node* node = side->get_ctrl_node_indexed(ne);
                    const Point ur = rotate_on_plane(*node, angle_1, bci_ptr->normal)
                                   - rotate_on_plane(*node, angle_0, bci_ptr->normal);
                    const Real u0_increment = ur(0),
                               u1_increment = ur(1),
                               u2_increment = ur(2);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                        Kesub[1*nvars+1](ne,ne) += penalty;
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                        Fesub[1](ne) += penalty * u1_increment;
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                                   DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                                   DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment,
                           u1_increment = bci_ptr->value(1)*U_increment,
                           u2_increment = bci_ptr->value(2)*U_increment;
                for (unsigned int ne=0; ne<side->size(); ne++) {
                    const Node* node = side->get_ctrl_node_indexed(ne);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                        Kesub[1*nvars+1](ne,ne) += penalty;
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                        Fesub[1](ne) += penalty * u1_increment;
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                                   DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                                   DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_X == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment;
                for (unsigned int ne=0; ne<side->size(); ne++) {
                    const Node* node = side->get_ctrl_node_indexed(ne);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Y == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u1_increment = bci_ptr->value(1)*U_increment;
                for (unsigned int ne=0; ne<side->size(); ne++) {
                    const Node* node = side->get_ctrl_node_indexed(ne);
                    if ( set_system_matrix ) {
                        Kesub[1*nvars+1](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[1](ne) += penalty * u1_increment;
                    } else {
                        const Real DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) );
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Z == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u2_increment = bci_ptr->value(2)*U_increment;
                for (unsigned int ne=0; ne<side->size(); ne++) {
                    const Node* node = side->get_ctrl_node_indexed(ne);
                    if ( set_system_matrix ) {
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XY == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment,
                           u1_increment = bci_ptr->value(1)*U_increment;
                for (unsigned int ne=0; ne<side->size(); ne++) {
                    const Node* node = side->get_ctrl_node_indexed(ne);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                        Kesub[1*nvars+1](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                        Fesub[1](ne) += penalty * u1_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                                   DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_YZ == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u1_increment = bci_ptr->value(1)*U_increment,
                           u2_increment = bci_ptr->value(2)*U_increment;
                for (unsigned int ne=0; ne<side->size(); ne++) {
                    const Node* node = side->get_ctrl_node_indexed(ne);
                    if ( set_system_matrix ) {
                        Kesub[1*nvars+1](ne,ne) += penalty;
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[1](ne) += penalty * u1_increment;
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                                   DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XZ == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment,
                           u2_increment = bci_ptr->value(2)*U_increment;
                for (unsigned int ne=0; ne<side->size(); ne++) {
                    const Node* node = side->get_ctrl_node_indexed(ne);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                                   DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::NEUMANN_VECTOR == bci_ptr->type ) {
                ; // do nothing here
            } else if ( BoundaryConditionInfo::NEUMANN_NORMAL == bci_ptr->type ) {
                ; // do nothing here
            } else libmesh_error();
            // check for active boundary condition
            }
            // update element stiffness matrix and rhs vector
            if ( set_system_matrix )
                system_matrix->add_matrix(Ke, dof_indices);
            system_rhs->add_vector(Fe, dof_indices);
            residual.add_vector(Fe, dof_indices);
            // free memo
            delete [] dof_indices_var;
            // end of (parametric) element sides loop...
        }
        // end of active local (parametric) elements loop...
    }
    // close system matrix, rhs and residual vector
    if ( set_system_matrix ) system_matrix->close();
    system_rhs->close();
    residual.close();
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("_assemble_system_essential_bcs");
}
//=================================================================================================
void NavierCauchyIGASolver3D::_solve () {
    this->_log.push("_solve");

    EquationSystems& es = this->_dmn.get_equation_systems();
    LinearImplicitSystem& du_sys = es.get_system<LinearImplicitSystem>(this->_label+"::iterative_solution");
    System&               DU_sys = es.get_system<System>(this->_label+"::incremental_solution");
    System&               u_sys  = es.get_system<System>(this->_label+"::solution");
    //
    NumericVector<Number>&  du = *du_sys.solution;
    NumericVector<Number>&  DU = *DU_sys.solution;
    NumericVector<Number>&  u  = *u_sys.solution;
    NumericVector<Number>&  Y = du_sys.get_vector("residual");
    // solve the linear system
    du_sys.linear_solver->reuse_preconditioner(this->par_list.get<bool>("system matrix updated") ? false : true);
    du_sys.solve();
    //
    Y.close();
    //
    DU.add(this->par_list.get<Real>("relaxing parameter"), du);
    DU.close();
    DU_sys.update();
    //
    u.add(this->par_list.get<Real>("relaxing parameter"), du);
    u.close();
    u_sys.update();
    //
    this->par_list.set<Real>("iterative solution norm (i)")   = du.l2_norm();
    this->par_list.set<Real>("incremental solution norm (i)") = DU.l2_norm();
    this->par_list.set<Real>("residual norm (i)") = Y.l2_norm();
    this->par_list.set<Real>("energy norm (i)") = fabs(Y.dot(du));
    if ( this->par_list.get<int>("iteration step") == 0 ) {
        this->par_list.set<Real>("iterative solution norm (0)")   = this->par_list.get<Real>("iterative solution norm (i)");
        this->par_list.set<Real>("incremental solution norm (0)") = this->par_list.get<Real>("incremental solution norm (i)");
        this->par_list.set<Real>("residual norm (0)") = this->par_list.get<Real>("residual norm (i)");
        this->par_list.set<Real>("energy norm (0)")   = this->par_list.get<Real>("energy norm (i)");
    } else {
        this->par_list.set<Real>("previous solution ratio") = this->par_list.get<Real>("solution ratio");
    }
    this->par_list.set<Real>("solution ratio") = this->par_list.get<Real>("iterative solution norm (i)")
                                               / this->par_list.set<Real>("iterative solution norm (0)");
    this->par_list.set<Real>("residual ratio") = this->par_list.get<Real>("residual norm (i)")
                                               / this->par_list.get<Real>("residual norm (0)");
    this->par_list.set<Real>("energy ratio") = this->par_list.get<Real>("energy norm (i)")
                                             / this->par_list.get<Real>("energy norm (0)");

    this->_log.pop("_solve");
}
//=================================================================================================
int NavierCauchyIGASolver3D::_check_convergence () const {
    this->_log.push("_check_convergence");

    int status = 0;
    const int iter_step = this->par_list.get<int>("iteration step") + 1;
    if ( this->par_list.get<bool>("load control") ) {
        const char end_of_line[] = "           \n";
        char txt_norms[buffer_len];
        sprintf( txt_norms, " |du| = %.8e; |du|/|du0| = %.8e; |Y|/|Y0| = %.8e; |du*Y|/|du0*Y0| = %.8e ",
                 this->par_list.get<Real>("iterative solution norm (i)"),
                 this->par_list.get<Real>("solution ratio"),
                 this->par_list.get<Real>("residual ratio"),
                 this->par_list.get<Real>("energy ratio")  );
        char txt[buffer_len];
        sprintf( txt, "  * Iteration %d;%s%s",
                 iter_step, txt_norms, end_of_line );
        flog.print(txt, false);
        // check for minimum number of iterations is achieved
        if ( iter_step > this->par_list.get<int>("minimum iterations") ) {
        // check if Newton-Raphson solver has converged
        if ( this->par_list.get<Real>("solution ratio") <= this->par_list.get<Real>("solution tolerance") &&
             this->par_list.get<Real>("residual ratio") <= this->par_list.get<Real>("residual tolerance") &&
             this->par_list.get<Real>("energy ratio")   <= this->par_list.get<Real>("energy tolerance")   ) {
            status = 1;
        } else if ( this->par_list.get<Real>("iterative solution norm (i)") <= 1.0e-9 ) {
            status = 1;
        } else if ( this->par_list.get<Real>("solution ratio") >= 100.0 ) {
            flog.print("    Solver diverged & Abort!!!\n", false);
            status = -1;
        } else if ( iter_step == this->par_list.get<int>("maximum iterations") &&
                    this->par_list.get<bool>("abort non-convergence")          ) {
            flog.print("    Abort non-converging solver!!!\n", false);
            status = -1;
        }
        // end if ...check for minimum number of iterations is achieved
        }
    } else {
        const char end_of_line[] = "           \n";
        char txt_norms[buffer_len];
        sprintf( txt_norms, " |du| = %.8e; |du|/|du0| = %.8e ",
                 this->par_list.get<Real>("iterative solution norm (i)"),
                 this->par_list.get<Real>("solution ratio") );
        char txt[buffer_len];
        sprintf( txt, "  * Iteration %d;%s%s",
                 iter_step, txt_norms, end_of_line );
        flog.print(txt, false);
        // check for minimum number of iterations is achieved
        if ( iter_step > this->par_list.get<int>("minimum iterations") ) {
        // check if Newton-Raphson solver has converged
        if ( this->par_list.get<Real>("solution ratio") <= this->par_list.get<Real>("solution tolerance") ) {
            status = 1;
        } else if ( this->par_list.get<Real>("iterative solution norm (i)") <= 1.0e-12 ) {
            status = 1;
        } else if ( this->par_list.get<Real>("solution ratio") >= 100.0 ) {
            flog.print("    Solver diverged & Abort!!!\n", false);
            status = -1;
        } else if ( iter_step == this->par_list.get<int>("maximum iterations") &&
                    this->par_list.get<bool>("abort non-convergence")          ) {
            flog.print("    Abort non-converging solver!!!\n", false);
            status = -1;
        }
        // end if ...check for minimum number of iterations is achieved
        }
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("_check_convergence");
    return status;
}
//=================================================================================================
void NavierCauchyIGASolver3D::_reset_increment (bool failed) {
    this->_log.push("_reset_increment");

    EquationSystems& es = this->_dmn.get_equation_systems();
    LinearImplicitSystem& du_sys = es.get_system<LinearImplicitSystem>(this->_label+"::iterative_solution");
    System&               DU_sys = es.get_system<System>(this->_label+"::incremental_solution");
    System&               u_sys  = es.get_system<System>(this->_label+"::solution");
    System&             u_P_sys  = es.get_system<System>(this->_label+"::previous_solution");
    //
    NumericVector<Number>&  du = *du_sys.solution;
    NumericVector<Number>&  DU = *DU_sys.solution;
    NumericVector<Number>&  u  = *u_sys.solution;
    NumericVector<Number>&  u_P = *u_P_sys.solution;
    //
    if ( failed ) {
        du.close();
        DU.close();
        u.close();
        //
        du.zero();
        du.close();
        du_sys.update();
        //
        u.add(-1.0, DU);
        u.close();
        u_sys.update();
        //
        DU.zero();
        DU.close();
        DU_sys.update();
        //
        this->par_list.set<int>("maximum increments (current scale)") += 1;
        this->par_list.set<Real>("current pseudo-time")  -= this->par_list.get<Real>("pseudo-time step");
        this->par_list.set<Real>("previous pseudo-time") -= this->par_list.get<Real>("pseudo-time step");
        this->par_list.set<Real>("pseudo-time step") /= 2.0;
        this->par_list.set<int>("increment step") *= 2.0;
        this->par_list.set<int>("maximum increments") *= 2.0;
        this->par_list.set<int>("solution increment") *= 2.0;
        this->par_list.set<int>("restart increment") *= 2.0;
        // else has not failed...
    } else {
        du.close();
        DU.close();
        u.close();
        u_P.close();
        //
        du.zero();
        du.close();
        du_sys.update();
        //
        DU.zero();
        DU.close();
        DU_sys.update();
        //
        u_P.operator=(u);
        u_P.close();
        u_P_sys.update();
        //
        this->par_list.set<int>("increment step") += 1;
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("_reset_increment");
}
//=================================================================================================
void NavierCauchyIGASolver3D::get_displacements ( const std::string& solu_type,
                                                  std::vector< std::vector<Real> >& solu_u,
                                                  IndexPair& solu_idx ) const
{
    this->_log.push("get_displacements");

    const EquationSystems& es = this->_dmn.get_equation_systems();
    const System& sys = es.get_system<System>(this->_label+solu_type);
    const MeshBase& msh = es.get_mesh();
    const std::string res_dir = es.parameters.get<std::string>("results directory");
    // clear memory
    solu_idx.clear();
    solu_u.clear();
    // close the solution vector
    sys.solution->close();
    // write in a temporary binary file the solution of all local nodes
    if ( true ) {
        // open the temporary binary file
        char foutBIN[buffer_len];
        sprintf(foutBIN, "%s/.nodal_solution.%d", res_dir.c_str(), libMesh::global_processor_id());
        std::ofstream fout(foutBIN, std::ios::binary);
        // write the number of local nodes
        const unsigned int n_local_nodes = msh.n_local_nodes();
        fout.write((char*)&n_local_nodes, sizeof(unsigned int));
        // loop over local nodes list
        for ( MeshBase::const_node_iterator
              citer_nd=msh.local_nodes_begin(); citer_nd!=msh.local_nodes_end(); citer_nd++ ) {
            const unsigned int node_id = (*citer_nd)->id();
            const Real         node_u0 = sys.current_solution((*citer_nd)->dof_number(sys.number(), 0, 0)),
                               node_u1 = sys.current_solution((*citer_nd)->dof_number(sys.number(), 1, 0)),
                               node_u2 = sys.current_solution((*citer_nd)->dof_number(sys.number(), 2, 0));
            fout.write((char*)&node_id, sizeof(unsigned int));
            fout.write((char*)&node_u0, sizeof(Real));
            fout.write((char*)&node_u1, sizeof(Real));
            fout.write((char*)&node_u2, sizeof(Real));
            // end of local nodes list loop...
        }
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // loop over all processes
    for (unsigned int iproc=0; iproc<libMesh::global_n_processors(); iproc++) {
        // open the temporary binary file
        char finBIN[buffer_len];
        sprintf(finBIN, "%s/.nodal_solution.%d", res_dir.c_str(), iproc);
        std::ifstream fin(finBIN, std::ios::binary);
        // read the number of local nodes
        unsigned int n_l_nodes = 0;
        fin.read((char*)&n_l_nodes, sizeof(unsigned int));
        // loop over local nodes list
        for (unsigned int inode=0; inode<n_l_nodes; inode++) {
            unsigned int node_id = 0;
            Real         node_u0 = 0.0,
                         node_u1 = 0.0,
                         node_u2 = 0.0;
            fin.read((char*)&node_id, sizeof(unsigned int));
            fin.read((char*)&node_u0, sizeof(Real));
            fin.read((char*)&node_u1, sizeof(Real));
            fin.read((char*)&node_u2, sizeof(Real));
            //
            std::vector<Real> node_u(SP3D);
            node_u[0] = node_u0; node_u[1] = node_u1; node_u[2] = node_u2;
            // check if node has already been inserted in the IndexPair container
            if ( solu_idx.confirm_id(node_id) ) continue;
            // update the corresponding containers
            solu_idx.insert(node_id);
            solu_u.push_back(node_u);
            // end of local nodes list loop...
        }
        // end of all processes loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // make sure to delete those temporary files
    if ( libMesh::global_processor_id() == 0 ) {
        for (unsigned int iproc=0; iproc<libMesh::global_n_processors(); iproc++) {
            char finBIN[buffer_len];
            sprintf(finBIN, "%s/.nodal_solution.%d", res_dir.c_str(), iproc);
            remove(finBIN);
        }
    }

    this->_log.pop("get_displacements");
}
//=================================================================================================
void NavierCauchyIGASolver3D::get_pressure ( const std::string& solu_type,
                                             std::vector< std::vector<Real> >& solu_p,
                                             IndexPair& solu_idx ) const
{
    this->_log.push("get_pressure");

    libmesh_error();

    this->_log.pop("get_pressure");
}
//=================================================================================================
void NavierCauchyIGASolver3D::_calc_displacements () {
    this->_log.push("_calc_displacements");

    EquationSystems& es = this->_dmn.get_equation_systems();
    const System& u_sys = es.get_system<System>(this->_label+"::solution");
    libmesh_assert( SP3D == u_sys.n_vars() );
    //
    std::vector<Number> displacement;
    u_sys.update_global_solution(displacement);
    //
    const IGA_Context& u_ctx = this->par_list.get<IGA_Context>(this->_label+"::solution");
    const std::vector<NURBS>& nurbs_ls = u_ctx.get_mesh_IGA().get_nurbs();
    const MeshBase::const_element_iterator begin_el = u_ctx.get_mesh_IGA().get_parametric_mesh().active_local_elements_begin();
    const MeshBase::const_element_iterator end_el   = u_ctx.get_mesh_IGA().get_parametric_mesh().active_local_elements_end();
    //
    const MeshIGA::IGA_PostData& iga_post_data = u_ctx.get_mesh_IGA().post_data;
    const MeshBase& msh_post = iga_post_data.es->get_mesh();
    //
    System& d_sys = iga_post_data.es->get_system<System>(this->_label+"::displacement_solution");
    libmesh_assert( SP3D == d_sys.n_vars() );
    //
    if ( this->par_list.get<bool>("membrane elements present") ) libmesh_error();
    //
    // loop over all active local (parametric) elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* p_elem = (*citer_el);
        const NURBS* elem = &nurbs_ls[p_elem->subdomain_id()];
        //
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [SP3D];
        std::vector<dof_id_type> vars;
        for (unsigned int i=0; i<SP3D; i++) vars.push_back( i );
        u_ctx.get_dof_indices(elem, dof_indices, dof_indices_var, &vars);
        //
        std::vector<unsigned int> var_dofs;
        calc_var_dofs(SP3D, dof_indices_var, var_dofs);
        //
        const std::set<dof_id_type>& cell_set =
            iga_post_data.nurbs_pelem_cells[elem->id()].find(p_elem->id())->second;
        //
        // iterated for all cells of the parametric element
        for ( std::set<dof_id_type>::const_iterator
              ci_s=cell_set.begin(); ci_s!=cell_set.end(); ci_s++ ) {
            const Elem* cell = msh_post.elem_ptr(*ci_s);
            //
            const std::vector<Point>& prm_coords =
                iga_post_data.cells_vertices.find(cell->id())->second;
            //
            for (unsigned int inode=0; inode<8; inode++) {
                const Node* node = cell->node_ptr(inode);
                //
                const Point Z = prm_coords[inode];
                const std::vector< Real >& Phi = elem->get_phi();
                elem->reinit(Z);
                //
                Point disp;
                for (unsigned int ivar=0; ivar<SP3D; ivar++) {
                    for (unsigned int l=0; l<var_dofs[ivar]; l++) {
                        disp(ivar) += Phi[l] * displacement[dof_indices_var[ivar][l]];
                    }
                }
                //
                set_system_solution_nodal_value<0>(d_sys, node, disp(0));
                set_system_solution_nodal_value<1>(d_sys, node, disp(1));
                set_system_solution_nodal_value<2>(d_sys, node, disp(2));
            }
            // end of individual cells loop...
        }
        // end of active local (parametric) elements loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // now update the system
    d_sys.solution->close();
    d_sys.update();

    this->_log.pop("_calc_displacements");
}
//=================================================================================================
void NavierCauchyIGASolver3D::_calc_gradients () {
    this->_log.push("_calc_gradients");

    EquationSystems& es = this->_dmn.get_equation_systems();
    //
    const IGA_Context& u_ctx = this->par_list.get<IGA_Context>(this->_label+"::solution");
    const std::vector<NURBS>& nurbs_ls = u_ctx.get_mesh_IGA().get_nurbs();
    const MeshBase::const_element_iterator begin_el = u_ctx.get_mesh_IGA().get_parametric_mesh().active_local_elements_begin();
    const MeshBase::const_element_iterator end_el   = u_ctx.get_mesh_IGA().get_parametric_mesh().active_local_elements_end();
    //
    const MeshIGA::IGA_PostData& iga_post_data = u_ctx.get_mesh_IGA().post_data;
    const MeshBase& msh_post = iga_post_data.es->get_mesh();
    //
    if ( this->par_list.get<bool>("membrane elements present") ) libmesh_error();
    const bool apply_wound_contraction = this->par_list.get<bool>("apply wound contraction"),
               apply_tumour_growth     = this->par_list.get<bool>("apply tumour growth");
    const unsigned int ninvars_NCS = es.parameters.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = es.parameters.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = es.parameters.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = es.parameters.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    const Real current_pseudo_time = this->par_list.get<Real>("current pseudo-time");
    //
    const System& u_sys = es.get_system<System>(this->_label+"::solution");
    const System& DU_sys = es.get_system<System>(this->_label+"::incremental_solution");
    System& gr_sys = iga_post_data.es->get_system<System>(this->_label+"::gradient_solution");
    // some sanity checks
    if ( 3!=u_sys.n_vars() ) libmesh_error();
    if ( 3!=DU_sys.n_vars() ) libmesh_error();
    if ( 33!=gr_sys.n_vars() ) libmesh_error();
    //
    std::vector<Number> displacement;
    u_sys.update_global_solution(displacement);
    std::vector<Number> displacement_incr;
    DU_sys.update_global_solution(displacement_incr);
    std::vector<Number> gradient_previous;
    gr_sys.update_global_solution(gradient_previous);
    //
    std::map<dof_id_type, QuadratureInfo>* elem_qpi = ( libMesh::global_n_processors() == 1 )
                                                    ? &this->_dmn.elem_qi_local
                                                    : &this->_dmn.elem_qi;
    std::map<dof_id_type, std::map<unsigned int, QuadratureInfo> >* elem_side_qpi = ( libMesh::global_n_processors() == 1 )
                                                                                  ? &this->_dmn.elem_side_qi_local
                                                                                  : &this->_dmn.elem_side_qi;
    //
    // only in parallel mode, and only if the "global" quadrature data structure is empty
    if ( libMesh::global_n_processors() > 1 )
    if ( elem_qpi->empty() || elem_side_qpi->empty() ) {
        char fname[buffer_len];
        //
        sprintf( fname, "%s/quadrature_data/%s.qpi.%d",
                 es.parameters.get<std::string>("results directory").c_str(),
                 es.parameters.get<std::string>("output file name").c_str(),
                 libMesh::global_processor_id() );
        this->_dmn.save_quadrature(fname, QuadratureInfo::ALL, QuadratureInfo::LEVEL_3);
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        //
        std::vector<std::string> fname_ls(libMesh::global_n_processors());
        for (unsigned int iproc=0; iproc<libMesh::global_n_processors(); iproc++) {
            sprintf( fname, "%s/quadrature_data/%s.qpi.%d",
                     es.parameters.get<std::string>("results directory").c_str(),
                     es.parameters.get<std::string>("output file name").c_str(),
                     iproc );
            fname_ls[iproc] = std::string(fname);
        }
        this->_dmn.gather_quadrature(fname_ls, QuadratureInfo::ALL, QuadratureInfo::LEVEL_3);
    }
    //
    // loop over all active local (parametric) elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* p_elem = (*citer_el);
        const NURBS* elem = &nurbs_ls[p_elem->subdomain_id()];
        //
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [SP3D];
        std::vector<dof_id_type> vars;
        for (unsigned int i=0; i<SP3D; i++) vars.push_back( i );
        u_ctx.get_dof_indices(elem, dof_indices, dof_indices_var, &vars);
        //
        std::vector<unsigned int> var_dofs;
        calc_var_dofs(SP3D, dof_indices_var, var_dofs);
        //
        const std::set<dof_id_type>& cell_set =
            iga_post_data.nurbs_pelem_cells[elem->id()].find(p_elem->id())->second;
        //
        const RegionInfo* ri_ptr = &(this->_dmn.regn.find(elem->subdomain_id())->second);
        //
        const QuadratureInfo* qpi_ptr = &(elem_qpi->find(p_elem->id())->second);
        //
        InverseDistanceAlgorithm ida(3.3);
        ida.init(qpi_ptr->XYZ);
        //
        // iterated for all cells of the parametric element
        for ( std::set<dof_id_type>::const_iterator
              ci_s=cell_set.begin(); ci_s!=cell_set.end(); ci_s++ ) {
            const Elem* cell = msh_post.elem_ptr(*ci_s);
            //
            const std::vector<Point>& prm_coords =
                iga_post_data.cells_vertices.find(cell->id())->second;
            //
            std::vector<Real> weight;
            ida.calc(cell->centroid(), weight);
            //
            for (unsigned int inode=0; inode<8; inode++) {
                const Node* node = cell->node_ptr(inode);
                //
                const Point Z = prm_coords[inode];
                const std::vector< Real >& Phi = elem->get_phi();
                const std::vector< Vector3D >& dPhi_dX = elem->get_dphi();
                elem->reinit(Z);
                //
                FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
                fld.current_time = current_pseudo_time;
                //
                for (unsigned int ivar=0; ivar<SP3D; ivar++) {
                    for (unsigned int l=0; l<var_dofs[ivar]; l++) {
                        fld.displacement(ivar) += Phi[l] * displacement[dof_indices_var[ivar][l]];
                        fld.grad_displacement(ivar) += vec(dPhi_dX[l]) * displacement[dof_indices_var[ivar][l]];
                    }
                }
                DisplacementGradient grad_incremental_displacement;
                for (unsigned int ivar=0; ivar<SP3D; ivar++) {
                    for (unsigned int l=0; l<var_dofs[ivar]; l++) {
                        grad_incremental_displacement(ivar) += vec(dPhi_dX[l]) * displacement_incr[dof_indices_var[ivar][l]];
                    }
                }
                {
                    const Real Fp00 = gradient_previous[node->dof_number(gr_sys.number(),  9, 0)],
                               Fp01 = gradient_previous[node->dof_number(gr_sys.number(), 10, 0)],
                               Fp02 = gradient_previous[node->dof_number(gr_sys.number(), 11, 0)],
                               Fp10 = gradient_previous[node->dof_number(gr_sys.number(), 12, 0)],
                               Fp11 = gradient_previous[node->dof_number(gr_sys.number(), 13, 0)],
                               Fp12 = gradient_previous[node->dof_number(gr_sys.number(), 14, 0)],
                               Fp20 = gradient_previous[node->dof_number(gr_sys.number(), 15, 0)],
                               Fp21 = gradient_previous[node->dof_number(gr_sys.number(), 16, 0)],
                               Fp22 = gradient_previous[node->dof_number(gr_sys.number(), 17, 0)];
                    fld.deform_grad_inel.comp = Fp00, Fp01, Fp02,
                                                Fp10, Fp11, Fp12,
                                                Fp20, Fp21, Fp22;
                }
                fld.deform_gradient = deformation_gradient_3d(fld.grad_displacement);
                for (unsigned int ivar=0; ivar<nvars_RDS+ndcvars_RDS; ivar++) {
                    fld.biochem[ivar] = 0.0;
                    for (unsigned int l=0; l<weight.size(); l++)
                        fld.biochem[ivar] += weight[l] * qpi_ptr->biochem[l][ivar];
                }
                for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                    fld.grad_biochem[ivar] = RealGradient(0.0, 0.0, 0.0);
                    for (unsigned int l=0; l<weight.size(); l++)
                        fld.grad_biochem[ivar] += weight[l] * qpi_ptr->grad_biochem[l][ivar];
                    fld.biochem_rate[ivar] = 0.0;
                    for (unsigned int l=0; l<weight.size(); l++)
                        fld.biochem_rate[ivar] += weight[l] * qpi_ptr->biochem_rate[l][ivar];
                }
                for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                    fld.auxiliary[ivar] = 0.0;
                    for (unsigned int l=0; l<weight.size(); l++)
                        fld.auxiliary[ivar] += weight[l] * qpi_ptr->auxiliary[l][ivar];
                    fld.grad_auxiliary[ivar] = RealGradient(0.0, 0.0, 0.0);
                    for (unsigned int l=0; l<weight.size(); l++)
                        fld.grad_auxiliary[ivar] += weight[l] * qpi_ptr->grad_auxiliary[l][ivar];
                    fld.auxiliary_rate[ivar] = 0.0;
                    for (unsigned int l=0; l<weight.size(); l++)
                        fld.auxiliary_rate[ivar] += weight[l] * qpi_ptr->auxiliary_rate[l][ivar];
                }
                for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                    fld.fibre_ref[ispdm].comp = 0.0;
                    for (unsigned int l=0; l<weight.size(); l++)
                        fld.fibre_ref[ispdm].comp += weight[l] * qpi_ptr->fibre_ref[ispdm][l].comp;
                    fld.fibre_curr[ispdm] = unit(fld.deform_gradient * fld.fibre_ref[ispdm]);
                }
                fld.strain = GreenLagrange_strain(fld.deform_gradient);
                fld.strain_incr = GreenLagrange_strain(deformation_gradient_3d(grad_incremental_displacement));
                {
                    const Real S00 = gradient_previous[node->dof_number(gr_sys.number(), 24, 0)],
                               S11 = gradient_previous[node->dof_number(gr_sys.number(), 25, 0)],
                               S22 = gradient_previous[node->dof_number(gr_sys.number(), 26, 0)],
                               S12 = gradient_previous[node->dof_number(gr_sys.number(), 27, 0)],
                               S02 = gradient_previous[node->dof_number(gr_sys.number(), 28, 0)],
                               S01 = gradient_previous[node->dof_number(gr_sys.number(), 29, 0)];
                    fld.stress.comp = S00, S01, S02,
                                      S01, S11, S12,
                                      S02, S12, S22;
                }
                fld.pressure = frac_1o3*trace(fld.stress);
                //
                fld.biomech[0] = ri_ptr->damage_factor_const;
                ri_ptr->calc_damage(fld);
                fld.biomech[1] = ri_ptr->mat->get_biomech_param(0) * (1.0-fld.biomech[0]);
                fld.biomech[2] = fld.pressure;
                //
                ri_ptr->calc_plastic_deform(fld);
                //
                Dyadic3D S2pk;
                Quindic3D Dtsm_uu;
                Real Ph;
                Dyadic3D Dtsm_up, Dtsm_pu;
                Real Dtsm_pp;
                // passive contribution of tissue biomechanics
                ri_ptr->calc_biomech_3d(fld, S2pk, Dtsm_uu, Dtsm_up, Ph, Dtsm_pu, Dtsm_pp);
                // wound contraction or tumour growth model
                if ( apply_wound_contraction || apply_tumour_growth )
                    ri_ptr->calc_coupled_3d(fld, S2pk, Dtsm_uu, Dtsm_up, Ph, Dtsm_pu, Dtsm_pp);
                //
                fld.stress = S2pk;
                //
                set_system_solution_nodal_value< 0>(gr_sys, node, fld.deform_grad_elas(0,0));
                set_system_solution_nodal_value< 1>(gr_sys, node, fld.deform_grad_elas(0,1));
                set_system_solution_nodal_value< 2>(gr_sys, node, fld.deform_grad_elas(0,2));
                set_system_solution_nodal_value< 3>(gr_sys, node, fld.deform_grad_elas(1,0));
                set_system_solution_nodal_value< 4>(gr_sys, node, fld.deform_grad_elas(1,1));
                set_system_solution_nodal_value< 5>(gr_sys, node, fld.deform_grad_elas(1,2));
                set_system_solution_nodal_value< 6>(gr_sys, node, fld.deform_grad_elas(2,0));
                set_system_solution_nodal_value< 7>(gr_sys, node, fld.deform_grad_elas(2,1));
                set_system_solution_nodal_value< 8>(gr_sys, node, fld.deform_grad_elas(2,2));
                set_system_solution_nodal_value< 9>(gr_sys, node, fld.deform_grad_inel(0,0));
                set_system_solution_nodal_value<10>(gr_sys, node, fld.deform_grad_inel(0,1));
                set_system_solution_nodal_value<11>(gr_sys, node, fld.deform_grad_inel(0,2));
                set_system_solution_nodal_value<12>(gr_sys, node, fld.deform_grad_inel(1,0));
                set_system_solution_nodal_value<13>(gr_sys, node, fld.deform_grad_inel(1,1));
                set_system_solution_nodal_value<14>(gr_sys, node, fld.deform_grad_inel(1,2));
                set_system_solution_nodal_value<15>(gr_sys, node, fld.deform_grad_inel(2,0));
                set_system_solution_nodal_value<16>(gr_sys, node, fld.deform_grad_inel(2,1));
                set_system_solution_nodal_value<17>(gr_sys, node, fld.deform_grad_inel(2,2));
                set_system_solution_nodal_value<18>(gr_sys, node, fld.strain(0,0));
                set_system_solution_nodal_value<19>(gr_sys, node, fld.strain(1,1));
                set_system_solution_nodal_value<20>(gr_sys, node, fld.strain(2,2));
                set_system_solution_nodal_value<21>(gr_sys, node, fld.strain(1,2));
                set_system_solution_nodal_value<22>(gr_sys, node, fld.strain(0,2));
                set_system_solution_nodal_value<23>(gr_sys, node, fld.strain(0,1));
                set_system_solution_nodal_value<24>(gr_sys, node, fld.stress(0,0));
                set_system_solution_nodal_value<25>(gr_sys, node, fld.stress(1,1));
                set_system_solution_nodal_value<26>(gr_sys, node, fld.stress(2,2));
                set_system_solution_nodal_value<27>(gr_sys, node, fld.stress(1,2));
                set_system_solution_nodal_value<28>(gr_sys, node, fld.stress(0,2));
                set_system_solution_nodal_value<29>(gr_sys, node, fld.stress(0,1));
                set_system_solution_nodal_value<30>(gr_sys, node, fld.biomech[0]);
                set_system_solution_nodal_value<31>(gr_sys, node, fld.biomech[1]);
                set_system_solution_nodal_value<32>(gr_sys, node, fld.biomech[2]);
            }
            // end of individual cells loop...
        }
        // end of active local (parametric) elements loop...
    }
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // close the solution container and update the system
    gr_sys.solution->close();
    gr_sys.update();

    this->_log.pop("_calc_gradients");
}
//=================================================================================================
void NavierCauchyIGASolver3D::save_restart (std::ofstream& fout) const {
    this->_log.push("save_restart");

    for (unsigned isys=0; isys<systems_ls.size(); isys++) {
        const EquationSystems* es_ptr = systems_ls[isys].first;
        const System& sys = es_ptr->get_system<System>(systems_ls[isys].second);
        //
        sys.solution->close();
        std::vector<Number> soln;
        sys.solution->localize_to_one(soln);
        //
        if ( libMesh::global_processor_id() ) continue;
        //
        const unsigned int n = soln.size();
        fout.write((char*)&n, sizeof(unsigned int));
        for (unsigned int i=0; i<n; i++) {
            const Number component = soln[i];
            fout.write((char*)&component, sizeof(Number));
        }
    }
    //
    this->_dmn.save_quadrature(fout, QuadratureInfo::ALL, QuadratureInfo::LEVEL_3);
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("save_restart");
}
//=================================================================================================
void NavierCauchyIGASolver3D::load_restart () {
    this->_log.push("load_restart");

    const Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    for (unsigned int ifile=0; ifile<par.get<unsigned int>("input restart from file"); ifile++) {
        const std::string str_name = "input restart file name #" + std::to_string(ifile);
        const std::string restart_file_name = par.get<std::string>(str_name.c_str());
        //
        std::ifstream fin(restart_file_name.c_str());
        // read the system solution from the first restart file only
        if ( 0 == ifile )
            for (unsigned isys=0; isys<systems_ls.size(); isys++) {
                EquationSystems* es_ptr = systems_ls[isys].first;
                System& sys = es_ptr->get_system<System>(systems_ls[isys].second);
                //
                unsigned int n = 0;
                fin.read((char*)&n, sizeof(unsigned int));
                for (unsigned int i=0; i<n; i++) {
                    Number component;
                    fin.read((char*)&component, sizeof(Number));
                    sys.solution->set(i, component);
                }
            }
        // read Quadrature data for this processor only
        if ( libMesh::global_processor_id() == ifile )
            this->_dmn.load_quadrature(fin, QuadratureInfo::ALL, QuadratureInfo::LEVEL_3);
        // end of restart files loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    for (unsigned isys=0; isys<systems_ls.size(); isys++) {
        EquationSystems* es_ptr = systems_ls[isys].first;
        System& sys = es_ptr->get_system<System>(systems_ls[isys].second);
        //
        sys.solution->close();
        sys.update();
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("load_restart");
}
//=================================================================================================
