/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/navier_cauchy_solver_3d.h"
#include "libmesh/linear_implicit_system.h"
#include "libmesh/petsc_linear_solver.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/gmsh_io.h"
#include "libmesh/fe.h"
#include "libmesh/quadrature_gauss.h"
#include "libmesh/dof_map.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/dense_matrix.h"
#include "libmesh/dense_vector.h"
#include "libmesh/dense_submatrix.h"
#include "libmesh/dense_subvector.h"
#include "libmesh/numeric_vector.h"
#include "feb3/iga-context.h"
#include "feb3/fem-context.h"
#include "feb3/paraview_io.h"
#include "feb3/ida.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
extern // list of all systems used in the analysis
std::vector< std::pair<EquationSystems*, std::string> > systems_ls;
// external routines
extern
void update_mesh_lagrange ( DomainData&,
                            std::vector< std::vector<Real> >&, const IndexPair& );
//=================================================================================================
#include "feb3/domain_data_inl.h"
#include "feb3/solution_io_inl.h"
//=================================================================================================
NavierCauchySolver3D::NavierCauchySolver3D (DomainData& dmn) :
    _log("NavierCauchySolver3D"), _dmn(dmn) {
    // set the default label
    this->_label = "NavierCauchySolver3D";
    // set parameters to a default value
    this->par_list.set<Real>("current pseudo-time") = 0.0;
    this->par_list.set<bool>("inverse formulation") = false;
    this->par_list.set<bool>("mixed formulation") = false;
    this->par_list.set<bool>("load control") = true;
    this->par_list.set<Real>("relaxing parameter") = 1.0;
    this->par_list.set<bool>("have follower forces") = true;
    this->par_list.set<bool>("apply wound contraction") = false;
    this->par_list.set<bool>("apply tumour growth") = false;
    this->par_list.set<bool>("apply body force") = false;
    this->par_list.set<bool>("apply nodal force") = false;
    this->par_list.set<bool>("save intermediate solution") = true;
    this->par_list.set<bool>("calculate gradients") = false;
    this->par_list.set<bool>("save updated mesh") = false;
    this->par_list.set<bool>("enforce to initialize solution") = true;
}
//=================================================================================================
void NavierCauchySolver3D::init () {
    this->_log.push("init");

    EquationSystems& es = this->_dmn.get_equation_systems();
    Parameters& par = es.parameters;
    // set the system parameters
    this->_copy_params();
    // dimension of the FE grid
    const unsigned int nvars = es.get_mesh().mesh_dimension();
    libmesh_assert(nvars == SP3D);
    // Lagrange basis for displacement DoFs (default), also choose the order
    const Order lbasis = par.get<Order>("Lagrange basis order");
    // Lagrange or Monomial basis for pressure DoF, also choose the order
    Order pbasis;
    FEFamily ptype;
    if ( this->par_list.get<bool>("mixed formulation") ) {
        pbasis = this->par_list.get<bool>("mixed formulation discontinuous pressure")
               ? this->par_list.get<Order>("Monomial (P) basis order") : this->par_list.get<Order>("Lagrange (P) basis order");
        ptype = this->par_list.get<bool>("mixed formulation discontinuous pressure")
              ? MONOMIAL : LAGRANGE;
    }
    //
    char var_name[buffer_len];
    int& hvar = par.set<int>("number of variables to hide in Paraview");
    // add the basic system
    do {
        LinearImplicitSystem& sys = es.add_system<LinearImplicitSystem>(this->_label+"::iterative_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__du%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis);
        }
        if ( this->par_list.get<bool>("mixed formulation") ) {
            sprintf(var_name, "%s__dp", this->_label.c_str());
            sys.add_variable(var_name, pbasis, ptype);
        }
        sys.assemble_before_solve = false;
        sys.add_vector("residual");
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    } while ( false );
    // add another system
    do {
        System& sys = es.add_system<System>(this->_label+"::solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__u%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis);
        }
        if ( this->par_list.get<bool>("mixed formulation") ) {
            sprintf(var_name, "%s__p", this->_label.c_str());
            sys.add_variable(var_name, pbasis, ptype);
        }
        sys.add_vector("nodal load");
        sys.add_vector("nodal load (part)");
        // get access to the solution vector
        this->_dmn.nforce.load_vec = &sys.get_vector("nodal load (part)");
    } while ( false );
    // add another system
    do {
        System& sys = es.add_system<System>(this->_label+"::previous_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__u%d_PREV", this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis);
        }
        if ( this->par_list.get<bool>("mixed formulation") ) {
            sprintf(var_name, "%s__p_PREV", this->_label.c_str());
            sys.add_variable(var_name, pbasis, ptype);
        }
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    } while ( false );
    // add another system
    do {
        System& sys = es.add_system<System>(this->_label+"::incremental_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__DU%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis);
        }
        if ( this->par_list.get<bool>("mixed formulation") ) {
            sprintf(var_name, "%s__DP", this->_label.c_str());
            sys.add_variable(var_name, pbasis, ptype);
        }
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    } while ( false );
    // add another system
    if ( this->par_list.get<bool>("save intermediate solution") ) {
        System& sys = es.add_system<System>(this->_label+"::intermediate_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__u%d_INTER", this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis);
        }
        if ( this->par_list.get<bool>("mixed formulation") ) {
            sprintf(var_name, "%s__p_INTER", this->_label.c_str());
            sys.add_variable(var_name, pbasis, ptype);
        }
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    }
    // add another (auxiliary) system
    if ( this->par_list.get<bool>("calculate gradients") ) {
        System& sys = es.add_system<System>(this->_label+"::gradient_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            for (unsigned int jvar=0; jvar<nvars; jvar++) {
                sprintf(var_name, "%s__Fe%d%d", this->_label.c_str(), ivar, jvar);
                sys.add_variable(var_name, CONSTANT, MONOMIAL);
            }
        }
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            for (unsigned int jvar=0; jvar<nvars; jvar++) {
                sprintf(var_name, "%s__Fp%d%d", this->_label.c_str(), ivar, jvar);
                sys.add_variable(var_name, CONSTANT, MONOMIAL);
            }
        }
        for (unsigned int ivar=0; ivar<VOIGT_SP3D; ivar++) {
            const Dual<int> idx = indices_Voigt_3d(ivar);
            sprintf(var_name, "%s__E%d%d", this->_label.c_str(), idx.get_first(), idx.get_second());
            sys.add_variable(var_name, CONSTANT, MONOMIAL);
        }
        for (unsigned int ivar=0; ivar<VOIGT_SP3D; ivar++) {
            const Dual<int> idx = indices_Voigt_3d(ivar);
            sprintf(var_name, "%s__S%d%d", this->_label.c_str(), idx.get_first(), idx.get_second());
            sys.add_variable(var_name, CONSTANT, MONOMIAL);
        }
        for (unsigned int ivar=0; ivar<es.parameters.get<unsigned int>("Navier-Cauchy solver: internal variables"); ivar++) {
            sprintf(var_name, "%s__iv%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name, CONSTANT, MONOMIAL);
        }
        // end of if case...
    }

    this->_log.pop("init");
}
//=================================================================================================
void NavierCauchySolver3D::run () {
    this->_log.push("run");

    EquationSystems& es = this->_dmn.get_equation_systems();
    //
    ParaviewIO pvtu(es.get_mesh());
    for (int l=0; l<es.parameters.get<int>("number of variables to hide in Paraview"); l++) {
        char var_name[buffer_len];
        sprintf(var_name, "Paraview variable #%d to hide", l);
        pvtu.hide_variable(es.parameters.get<std::string>(var_name));
    }
    //
    char cstr[buffer_len];
    const std::string results_dir  = es.parameters.get<std::string>("results directory"),
                      output_fname = es.parameters.get<std::string>("output file name"),
                      sim_title    = es.parameters.get<std::string>("simulation title");
    const Real current_time = es.parameters.get<Real>("current time");
    Real& current_pseudo_time  = this->par_list.set<Real>("current pseudo-time");
    Real& previous_pseudo_time = this->par_list.set<Real>("previous pseudo-time");
    Real& pseudo_time_step = this->par_list.set<Real>("pseudo-time step");
    const int ini_incr = this->par_list.get<int>("initial increment");
    int& max_incr = this->par_list.set<int>("maximum increments");
    int& soln_step = this->par_list.set<int>("solution increment");
    int& rstr_step = this->par_list.set<int>("restart increment");
    //
    // set the initial conditions and initialise some member objects
    if ( this->par_list.get<bool>("enforce to initialize solution") ) {
        flog.print(" >> Enforce solution initialization\n", false);
        this->_set_initial();
    // ...else check if to read the restart file(s)
    } else if ( es.parameters.get<unsigned int>("input restart from file") ) {
        flog.print(" >> Load restart file(s)\n", false);
        this->load_restart();
    }
    //
    // write a PVD file to store results
    if ( libMesh::global_processor_id() == 0 ) {
        flog.print(" >> Generate Paraview PVD-file\n", false);
        sprintf( cstr, "%s/%s-%s-%s.pvd",
                 results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), stringfy_number(current_time).c_str() );
        std::ofstream fpvd(cstr);
        fpvd.setf(std::ios_base::scientific|std::ios_base::unitbuf);
        //
        fpvd << "<?xml version=\"1.0\"?>" << std::endl;
        fpvd << "<VTKFile type=\"Collection\" version=\"0.1\" byte_order=\"LittleEndian\">" << std::endl;
        fpvd << "<Collection>" << std::endl;
        for (int incr=ini_incr; incr<=max_incr; incr++) {
            const Real vtk_time = (incr+0.0)/max_incr;
            if ( incr == ini_incr || incr == max_incr || incr%soln_step == 0 ) {
                sprintf( cstr, "paraview/%s-%s-%s-%d.vtu",
                         output_fname.c_str(), sim_title.c_str(), stringfy_number(current_time).c_str(), incr );
                fpvd << "<DataSet timestep=\"" << vtk_time << "\" group=\"\" part=\"0\" file=\"" << cstr << "\"/>" << std::endl;
            }
            // end of time steps loop...
        }
        fpvd << "</Collection>" << std::endl;
        fpvd << "</VTKFile>" << std::endl;
    }
    //
    // write initial results
    flog.print("  > Save solution\n", false);
    sprintf( cstr, "%s/paraview/%s-%s-%s-%d.vtu",
             results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), stringfy_number(current_time).c_str(), ini_incr );
    pvtu.write_equation_systems(cstr, es);
    //
    // run for all pseudo-time steps
    this->par_list.set<int>("increment step") = ini_incr;
    do {
        REDO: ;
        current_pseudo_time  += pseudo_time_step;
        previous_pseudo_time += pseudo_time_step;
        //
        const int incr = this->par_list.get<int>("increment step") + 1;
        const int increment = incr / static_cast<int>(pow(2.0, this->par_list.get<int>("maximum increments (current scale)")));
        if ( this->par_list.get<bool>("apply nodal force") ) {
            sprintf( cstr, " >> Pseudo-time %d out of %d (%.10f)\t Fn(Eval):%d\t Fn(Appl):%d\n",
                     incr, max_incr, current_pseudo_time, this->_dmn.nforce.calc_time_function__evaluate(current_pseudo_time), (bool)this->_dmn.nforce.calc_time_function__employ(current_pseudo_time) );
        } else {
            sprintf( cstr, " >> Pseudo-time %d out of %d (%.10f)\n",
                     incr, max_incr, current_pseudo_time );
        }
        flog.print(cstr, false);
        //
        // run for all iteration steps
        const int niters = this->par_list.get<int>("maximum iterations");
        for (int iter=0; iter<niters; iter++) {
            this->par_list.set<int>("iteration step") = iter;
            //
            // assemble the linearized system
            if ( this->par_list.get<bool>("inverse formulation") ) {
                this->par_list.get<bool>("mixed formulation") ?
                this->_assemble_system__INV_UP() : this->_assemble_system__INV_U();
            } else {
                this->par_list.get<bool>("mixed formulation") ?
                this->_assemble_system__FWD_UP() : this->_assemble_system__FWD_U();
            }
            //
            // solve and update solution via Newton-Raphson
            this->_solve();
            //
            // calculate convergence criteria
            const int status = this->_check_convergence();
            // check if to exit loop, abort or continue
            if      ( status == 1 ) break;
            else if ( status <  0 ) {
                //
                // reset solution after failure
                this->_reset_increment(true);
                //
                // check if to resume analysis with increased incrementation
                if ( this->par_list.get<int>("maximum increments (ceiling scale)") >  0                                                            &&
                     this->par_list.get<int>("maximum increments (ceiling scale)") >= this->par_list.get<int>("maximum increments (current scale)") )
                    goto REDO;
                flog.print(" >> Solver is giving up!\n", false);
                //
                // calculate the gradient solution (if necessary)
                if ( this->par_list.get<bool>("calculate gradients") ) {
                    flog.print("  > Calculate gradient solution\n", false);
                    this->_calc_gradients();
                }
                //
                flog.print("  > Save solution\n", false);
                sprintf( cstr, "%s/paraview/%s-%s-%s-%d.vtu",
                         results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), stringfy_number(current_time).c_str(), increment );
                pvtu.write_equation_systems(cstr, es);
                // now jump code...
                goto END_LABEL;
            }
            // end of iterations loop...
        }
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        //
        // update quadrature point set data
        flog.print("  > Update quadrature data\n", false);
        if ( this->par_list.get<bool>("inverse formulation") ) {
            this->par_list.get<bool>("mixed formulation") ?
            this->_set_quad_info__INV_UP() : this->_set_quad_info__INV_U();
        } else {
            this->par_list.get<bool>("mixed formulation") ?
            this->_set_quad_info__FWD_UP() : this->_set_quad_info__FWD_U();
        }
        //
        // calculate the gradient solution (if necessary)
        if ( this->par_list.get<bool>("calculate gradients") ) {
            flog.print("  > Calculate gradient solution\n", false);
            this->_calc_gradients();
        }
        //
        // check if to write results
        if ( incr%soln_step == 0 || incr == max_incr ) {
            //
            flog.print("  > Save solution\n", false);
            sprintf( cstr, "%s/paraview/%s-%s-%s-%d.vtu",
                     results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), stringfy_number(current_time).c_str(), increment );
            pvtu.write_equation_systems(cstr, es);
        }
        //
        // check if to write restart file
        if ( incr%rstr_step == 0 || incr == max_incr ) {
            flog.print("  > Save restart\n", false);
            sprintf( cstr, "%s/restart/%s-%s-%s-%d.res-%d",
                     results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), stringfy_number(current_time).c_str(), increment, libMesh::global_processor_id() );
            std::ofstream fout(cstr);
            this->save_restart(fout);
        }
        //
        // reset (zero-out) incremental solution for subsequent increment step
        this->_reset_increment();
        // end of pseudo-time steps loop...
    } while ( this->par_list.get<int>("increment step") < max_incr );
    //
    // store a copy of the input Gmsh formatted mesh file
    if ( this->par_list.get<bool>("save updated mesh") ) {
        flog.print(" >> Update Lagrangian mesh\n", false);
        {
            std::vector< std::vector<Real> > solu_u;
            IndexPair solu_idx;
            this->get_displacements("::solution", solu_u, solu_idx);
            update_mesh_lagrange(this->_dmn, solu_u, solu_idx);
        }
        // save the updated (deformed) finite element grid
        if ( libMesh::global_processor_id() == 0 ) {
            flog.print(" >> Save mesh\n", false);
            sprintf( cstr, "%s/%s-%s.msh",
                     results_dir.c_str(), output_fname.c_str(), stringfy_number(current_time).c_str() );
            GmshIO(es.get_mesh()).write(cstr);
            sprintf( cstr, "%s/paraview/%s-%s-boundary.vtu",
                     results_dir.c_str(), output_fname.c_str(), stringfy_number(current_time).c_str() );
            ParaviewIO(es.get_mesh()).write_boundary(cstr);
        }
        // end of if case...
    }
    //
    // check if to update the intermediate solution vector
    if ( this->par_list.get<bool>("save intermediate solution") ) {
        const System& u_sys = es.get_system<System>(this->_label+"::solution");
        System& u_T_sys = es.get_system<System>(this->_label+"::intermediate_solution");
        //
        NumericVector<Number>&  u   = *u_sys.solution;
        NumericVector<Number>&  u_T = *u_T_sys.solution;
        //
        u.close();
        //
        u_T.close();
        u_T.zero();
        u_T.add(u);
        u_T.close();
        u_T_sys.update();
    }
    // end of simulation...
    END_LABEL: ;
    clear_line_(std::cout);
    flog.print(" >> Simulation terminated\n", false);

    this->_log.pop("run");
}
//=================================================================================================
//
//
//
//=================================================================================================
void NavierCauchySolver3D::_copy_params () {
    this->_log.push("_copy_params");

    EquationSystems& es = this->_dmn.get_equation_systems();
    // set the label for the solver and the simulation title (if necessary)
    this->_label = es.parameters.get<std::string>("Navier-Cauchy solver: label");
    if ( "_" == es.parameters.get<std::string>("simulation title") )
        es.parameters.set<std::string>("simulation title") = this->_label;
    //
    if ( es.parameters.get<unsigned int>("input restart from file") )
        this->par_list.set<bool>("enforce to initialize solution") = false;
    //
    this->par_list.set<bool>("full Newton-Raphson") = es.parameters.get<bool>("Navier-Cauchy solver: full Newton-Raphson");
    this->par_list.set<int>("initial increment")  = es.parameters.get<int>("Navier-Cauchy solver: initial increment");
    this->par_list.set<int>("maximum increments") = es.parameters.get<int>("Navier-Cauchy solver: maximum increments");
    this->par_list.set<int>("maximum iterations") = es.parameters.get<int>("Navier-Cauchy solver: maximum iterations");
    this->par_list.set<int>("minimum iterations") = es.parameters.get<int>("Navier-Cauchy solver: minimum iterations");
    this->par_list.set<int>("solution increment") = es.parameters.get<int>("Navier-Cauchy solver: solution increment");
    this->par_list.set<int>("restart increment") = es.parameters.get<int>("Navier-Cauchy solver: restart increment");
    this->par_list.set<Real>("current pseudo-time")  = (this->par_list.get<int>("initial increment")+0.0)
                                                     * es.parameters.get<Real>("Navier-Cauchy solver: window")
                                                     / this->par_list.get<int>("maximum increments");
    this->par_list.set<Real>("previous pseudo-time") = (this->par_list.get<int>("initial increment")-1.0)
                                                     * es.parameters.get<Real>("Navier-Cauchy solver: window")
                                                     / this->par_list.get<int>("maximum increments");
    this->par_list.set<Real>("pseudo-time step") = es.parameters.get<Real>("Navier-Cauchy solver: window")
                                                 / this->par_list.get<int>("maximum increments");
    if ( es.parameters.get<Real>("Navier-Cauchy solver: window")!=1.0 ) libmesh_error();
    this->par_list.set<int>("maximum increments (current scale)") = 0;
    this->par_list.set<int>("maximum increments (ceiling scale)") = es.parameters.get<int>("Navier-Cauchy solver: maximum increments (ceiling scale)");
    this->par_list.set<bool>("abort non-convergence") = es.parameters.get<bool>("Navier-Cauchy solver: abort non-convergence");
    //
    this->par_list.set<Real>("solution tolerance") = es.parameters.get<Real>("Navier-Cauchy solver: solution tolerance");
    this->par_list.set<Real>("residual tolerance") = es.parameters.get<Real>("Navier-Cauchy solver: residual tolerance");
    this->par_list.set<Real>("energy tolerance")   = es.parameters.get<Real>("Navier-Cauchy solver: energy tolerance");
    this->par_list.set<bool>("load control")        = es.parameters.get<bool>("Navier-Cauchy solver: load control");
    this->par_list.set<bool>("calculate gradients") = es.parameters.get<bool>("Navier-Cauchy solver: calculate gradients");
    this->par_list.set<bool>("mixed formulation")   = es.parameters.get<bool>("Navier-Cauchy solver: mixed formulation");
    if ( this->par_list.get<bool>("mixed formulation") ) {
        this->par_list.set<bool>("mixed formulation discontinuous pressure") =
            es.parameters.get<bool>("Navier-Cauchy solver: mixed formulation discontinuous pressure");
        //
        if ( this->par_list.get<bool>("mixed formulation discontinuous pressure") ) {
            if      ( FIRST  == es.parameters.get<Order>("Lagrange basis order") )
                this->par_list.set<Order>("Monomial (P) basis order") = CONSTANT;
            else if ( SECOND == es.parameters.get<Order>("Lagrange basis order") )
                // this->par_list.set<Order>("Monomial (P) basis order") = CONSTANT;
                this->par_list.set<Order>("Monomial (P) basis order") = FIRST;
            else
                ERROR_("mixed formulation with discontinuous pressure should have 1st- or 2nd-order Lagrange basis only");
        } else {
            if ( SECOND == es.parameters.get<Order>("Lagrange basis order") )
                this->par_list.set<Order>("Lagrange (P) basis order") = FIRST;
            else
                ERROR_("mixed formulation with continuous pressure should have 2nd-order Lagrange basis");
        }
    }
    this->par_list.set<bool>("inverse formulation") = es.parameters.get<bool>("Navier-Cauchy solver: inverse formulation");
    this->par_list.set<bool>("have follower forces") = es.parameters.get<bool>("Navier-Cauchy solver: have follower forces");
    this->par_list.set<bool>("membrane elements present") = es.parameters.get<bool>("membrane elements present");
    this->par_list.set<bool>("apply wound contraction") = es.parameters.get<bool>("Navier-Cauchy solver: apply wound contraction");
    this->par_list.set<bool>("apply tumour growth") = es.parameters.get<bool>("Navier-Cauchy solver: apply tumour growth");
    this->par_list.set<bool>("apply body force") = es.parameters.get<bool>("Navier-Cauchy solver: apply body force");
    this->par_list.set<bool>("apply nodal force") = es.parameters.get<bool>("Navier-Cauchy solver: apply nodal force");
    this->par_list.set<bool>("save updated mesh") = es.parameters.get<bool>("Navier-Cauchy solver: save updated mesh");

    this->_log.pop("_copy_params");
}
//=================================================================================================
void NavierCauchySolver3D::_set_initial () {
    this->_log.push("_set_initial");

    EquationSystems& es = this->_dmn.get_equation_systems();
    System& du_sys = es.get_system<System>(this->_label+"::iterative_solution");
    System& DU_sys = es.get_system<System>(this->_label+"::incremental_solution");
    System& u_sys   = es.get_system<System>(this->_label+"::solution");
    System& u_P_sys = es.get_system<System>(this->_label+"::previous_solution");
    //
    {
        NumericVector<Number>& du = *du_sys.solution;
        //
        du.zero();
        du.close();
        du_sys.update();
    }
    //
    {
        NumericVector<Number>& DU = *DU_sys.solution;
        //
        DU.zero();
        DU.close();
        DU_sys.update();
    }
    //
    {
        NumericVector<Number>& u = *u_sys.solution;
        NumericVector<Number>& f = u_sys.get_vector("nodal load");
        NumericVector<Number>& f_p = u_sys.get_vector("nodal load (part)");
        //
        u.zero();
        u.close();
        f.zero();
        f.close();
        f_p.zero();
        f_p.close();
        u_sys.update();
    }
    //
    {
        NumericVector<Number>& u = *u_P_sys.solution;
        //
        u.zero();
        u.close();
        u_P_sys.update();
    }
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    // update quadrature point set data
    flog.print("  > Update quadrature data\n", false);
    if ( this->par_list.get<bool>("inverse formulation") ) {
        this->par_list.get<bool>("mixed formulation") ?
        this->_set_quad_info__INV_UP() : this->_set_quad_info__INV_U();
    } else {
        this->par_list.get<bool>("mixed formulation") ?
        this->_set_quad_info__FWD_UP() : this->_set_quad_info__FWD_U();
    }
    //
    // calculate the gradient solution (if necessary)
    if ( this->par_list.get<bool>("calculate gradients") ) {
        flog.print("  > Calculate gradient solution\n", false);
        this->_calc_gradients();
    }

    this->_log.pop("_set_initial");
}
//=================================================================================================
void NavierCauchySolver3D::_assemble_system_essential_bcs () {
    this->_log.push("_assemble_system_essential_bcs");

    EquationSystems& es = this->_dmn.get_equation_systems();
    const MeshBase& msh = es.get_mesh();
    SparseMatrix<Number>*  system_matrix = es.get_system<LinearImplicitSystem>(this->_label+"::iterative_solution").matrix;
    NumericVector<Number>* system_rhs    = es.get_system<LinearImplicitSystem>(this->_label+"::iterative_solution").rhs;
    NumericVector<Number>& residual      = es.get_system<LinearImplicitSystem>(this->_label+"::iterative_solution").get_vector("residual");
    const System& DU_sys = es.get_system<System>(this->_label+"::incremental_solution");
    // const System& u_T_sys = this->par_list.get<bool>("save intermediate solution")
    //                       ? es.get_system<System>(this->_label+"::intermediate_solution") : es.get_system<System>(this->_label+"::solution");
    //
    const FEM_Context DU_ctx(es.get_system<System>(this->_label+"::incremental_solution"));
    const MeshBase::const_element_iterator begin_el = msh.active_local_elements_begin();
    const MeshBase::const_element_iterator end_el = msh.active_local_elements_end();
    //
    libmesh_assert( SP3D == msh.mesh_dimension() );
    libmesh_assert( (this->par_list.get<bool>("mixed formulation") ? SP3D+1 : SP3D) == DU_sys.n_vars() );
    std::vector<unsigned short int> var;
    DU_ctx.get_var_list(var);
    //
    const int //increment_step = this->par_list.get<int>("increment step"),
              iteration_step = this->par_list.get<int>("iteration step");
    const bool set_system_matrix = ( ! this->par_list.get<bool>("full Newton-Raphson") && iteration_step != 0 ) ? false : true;
    const Real previous_time = this->par_list.get<Real>("previous pseudo-time"),
               current_time  = this->par_list.get<Real>("current pseudo-time");
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // set the penalty parameter
    const Real penalty = this->par_list.get<Real>("system matrix maximum value")
                       * es.parameters.get<Real>("penalty parameter scale");
    //
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        //
        const unsigned int nvars = this->par_list.get<bool>("mixed formulation") ? SP3D+1 : SP3D;
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
        std::vector<dof_id_type> vars;
        for (unsigned int i=0; i<nvars; i++) vars.push_back( i );
        DU_ctx.get_dof_indices(elem, dof_indices, dof_indices_var, &vars);
        //
        std::vector<unsigned int> var_dofs;
        calc_var_dofs(nvars, dof_indices_var, var_dofs);
        //
        DenseMatrix<Number> Ke(dof_indices.size(), dof_indices.size());
        std::vector< DenseSubMatrix<Number> > Kesub(nvars*nvars, Ke);
        DenseVector<Number> Fe(dof_indices.size());
        std::vector< DenseSubVector<Number> > Fesub(nvars, Fe);
        std::vector<unsigned int> offset;
        calc_offset(nvars, var_dofs, offset);
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            for (unsigned int jvar=0; jvar<nvars; jvar++) {
                Kesub[ivar*nvars+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
            }
            Fesub[ivar].reposition(offset[ivar], var_dofs[ivar]);
        }
        //
        // loop over all FE nodes for any essential constraints
        for (unsigned int ne=0; ne<elem->n_nodes(); ne++) {
            const Node* node = elem->node_ptr(ne);
            // check if element node is constrained
            std::map<unsigned int, BoundaryConditionInfo> ::const_iterator map_citer = this->_dmn.nodal_bcs.find( node->id() );
            if ( this->_dmn.nodal_bcs.end() == map_citer ) continue;
            if ( node->processor_id() != libMesh::global_processor_id() ) continue;
            //
            const BoundaryConditionInfo* bci_ptr = &(map_citer->second);
            // check for active boundary condition
            if ( bci_ptr->calc_active_function(current_time) ) {
            // distinguish among types of boundary conditions
            if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment,
                           u1_increment = bci_ptr->value(1)*U_increment,
                           u2_increment = bci_ptr->value(2)*U_increment;
                if ( set_system_matrix ) {
                    Kesub[0*nvars+0](ne,ne) += penalty;
                    Kesub[1*nvars+1](ne,ne) += penalty;
                    Kesub[2*nvars+2](ne,ne) += penalty;
                }
                if ( iteration_step == 0 ) {
                    Fesub[0](ne) += penalty * u0_increment;
                    Fesub[1](ne) += penalty * u1_increment;
                    Fesub[2](ne) += penalty * u2_increment;
                } else {
                    const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                               DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                               DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                    Fesub[0](ne) += penalty * (u0_increment-DU0);
                    Fesub[1](ne) += penalty * (u1_increment-DU1);
                    Fesub[2](ne) += penalty * (u2_increment-DU2);
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_X == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment;
                if ( set_system_matrix ) {
                    Kesub[0*nvars+0](ne,ne) += penalty;
                }
                if ( iteration_step == 0 ) {
                    Fesub[0](ne) += penalty * u0_increment;
                } else {
                    const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) );
                    Fesub[0](ne) += penalty * (u0_increment-DU0);
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Y == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u1_increment = bci_ptr->value(1)*U_increment;
                if ( set_system_matrix ) {
                    Kesub[1*nvars+1](ne,ne) += penalty;
                }
                if ( iteration_step == 0 ) {
                    Fesub[1](ne) += penalty * u1_increment;
                } else {
                    const Real DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) );
                    Fesub[1](ne) += penalty * (u1_increment-DU1);
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Z == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u2_increment = bci_ptr->value(2)*U_increment;
                if ( set_system_matrix ) {
                    Kesub[2*nvars+2](ne,ne) += penalty;
                }
                if ( iteration_step == 0 ) {
                    Fesub[2](ne) += penalty * u2_increment;
                } else {
                    const Real DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                    Fesub[2](ne) += penalty * (u2_increment-DU2);
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XY == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment,
                           u1_increment = bci_ptr->value(1)*U_increment;
                if ( set_system_matrix ) {
                    Kesub[0*nvars+0](ne,ne) += penalty;
                    Kesub[1*nvars+1](ne,ne) += penalty;
                }
                if ( iteration_step == 0 ) {
                    Fesub[0](ne) += penalty * u0_increment;
                    Fesub[1](ne) += penalty * u1_increment;
                } else {
                    const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                               DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) );
                    Fesub[0](ne) += penalty * (u0_increment-DU0);
                    Fesub[1](ne) += penalty * (u1_increment-DU1);
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_YZ == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u1_increment = bci_ptr->value(1)*U_increment,
                           u2_increment = bci_ptr->value(2)*U_increment;
                if ( set_system_matrix ) {
                    Kesub[1*nvars+1](ne,ne) += penalty;
                    Kesub[2*nvars+2](ne,ne) += penalty;
                }
                if ( iteration_step == 0 ) {
                    Fesub[1](ne) += penalty * u1_increment;
                    Fesub[2](ne) += penalty * u2_increment;
                } else {
                    const Real DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                               DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                    Fesub[1](ne) += penalty * (u1_increment-DU1);
                    Fesub[2](ne) += penalty * (u2_increment-DU2);
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XZ == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment,
                           u2_increment = bci_ptr->value(2)*U_increment;
                if ( set_system_matrix ) {
                    Kesub[0*nvars+0](ne,ne) += penalty;
                    Kesub[2*nvars+2](ne,ne) += penalty;
                }
                if ( iteration_step == 0 ) {
                    Fesub[0](ne) += penalty * u0_increment;
                    Fesub[2](ne) += penalty * u2_increment;
                } else {
                    const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                               DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                    Fesub[0](ne) += penalty * (u0_increment-DU0);
                    Fesub[2](ne) += penalty * (u2_increment-DU2);
                }
            } else if ( BoundaryConditionInfo::NEUMANN_VECTOR == bci_ptr->type ) {
                ; // do nothing here
            } else libmesh_error();
            // end if ...check for active boundary condition
            }
            // end of FE nodes loop...
        }
        // update element stiffness matrix and rhs vector
        if ( set_system_matrix )
            system_matrix->add_matrix(Ke, dof_indices);
        system_rhs->add_vector(Fe, dof_indices);
        residual.add_vector(Fe, dof_indices);
        // free memo
        delete [] dof_indices_var;
        // end of active local elements loop...
    }
    //
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        //
        // insert essential BCs for "boundary" side elements
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            // check if element has this side on boundary and get the boundary ID for side of element
            if ( elem->neighbor_ptr(s) != NULL ) continue;
            //
            const boundary_id_type bc_id = msh.boundary_info->boundary_id(elem, s);
            const BoundaryConditionInfo* bci_ptr = &(this->_dmn.bcs.find(bc_id)->second);
            // create a side element
            UniquePtr<const Elem> side = elem->build_side_ptr(s);
            //
            const unsigned int nvars = SP3D;
            std::vector<dof_id_type> dof_indices;
            std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
            std::vector<dof_id_type> vars;
            for (unsigned int i=0; i<nvars; i++) vars.push_back( i );
            DU_ctx.get_dof_indices(side.get(), dof_indices, dof_indices_var, &vars);
            //
            std::vector<unsigned int> var_dofs;
            calc_var_dofs(nvars, dof_indices_var, var_dofs);
            //
            DenseMatrix<Number> Ke(dof_indices.size(), dof_indices.size());
            std::vector< DenseSubMatrix<Number> > Kesub(nvars*nvars, Ke);
            DenseVector<Number> Fe(dof_indices.size());
            std::vector< DenseSubVector<Number> > Fesub(nvars, Fe);
            std::vector<unsigned int> offset;
            calc_offset(nvars, var_dofs, offset);
            for (unsigned int ivar=0; ivar<nvars; ivar++) {
                for (unsigned int jvar=0; jvar<nvars; jvar++) {
                    Kesub[ivar*nvars+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
                }
                Fesub[ivar].reposition(offset[ivar], var_dofs[ivar]);
            }
            // check for active boundary condition
            if ( bci_ptr->calc_active_function(current_time) ) {
            // distinguish among types of boundary conditions
            if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_RD == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real ur_increment = bci_ptr->value(0)*U_increment;
                if ( fabs(U_increment) > 1.0e-6 ) // NOTE: this check is valid only for this type of BC
                for (unsigned int ne=0; ne<side->n_nodes(); ne++) {
                    const Node* node = side->node_ptr(ne);
                    Point radial = *node - bci_ptr->center;
                    if ( radial.norm() < 1.0e-6 ) radial = 0.0;
                    else radial = radial.unit();
                    const Real u0_increment = ur_increment*radial(0),
                               u1_increment = ur_increment*radial(1),
                               u2_increment = ur_increment*radial(2);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                        Kesub[1*nvars+1](ne,ne) += penalty;
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                        Fesub[1](ne) += penalty * u1_increment;
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                                   DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                                   DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_CY == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real ur_increment = bci_ptr->value(0)*U_increment;
                if ( fabs(U_increment) > 1.0e-6 ) // NOTE: this check is valid only for this type of BC
                for (unsigned int ne=0; ne<side->n_nodes(); ne++) {
                    const Node* node = side->node_ptr(ne);
                    const Point node_proj = bci_ptr->center + ((*node-bci_ptr->center)*bci_ptr->normal) * bci_ptr->normal - *node;
                    const Real u0_increment = ur_increment*node_proj(0),
                               u1_increment = ur_increment*node_proj(1),
                               u2_increment = ur_increment*node_proj(2);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                        Kesub[1*nvars+1](ne,ne) += penalty;
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                        Fesub[1](ne) += penalty * u1_increment;
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                                   DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                                   DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_PL == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real un_increment = bci_ptr->value(0)*U_increment;
                if ( fabs(U_increment) > 1.0e-6 ) // NOTE: this check is valid only for this type of BC
                for (unsigned int ne=0; ne<side->n_nodes(); ne++) {
                    const Node* node = side->node_ptr(ne);
                    const Point node_proj = project_to_plane(*node, bci_ptr->center, bci_ptr->normal) - *node;
                    const Real u0_increment = un_increment*node_proj(0),
                               u1_increment = un_increment*node_proj(1),
                               u2_increment = un_increment*node_proj(2);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                        Kesub[1*nvars+1](ne,ne) += penalty;
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                        Fesub[1](ne) += penalty * u1_increment;
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                                   DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                                   DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_PR == bci_ptr->type ) {
                const Real angle_1 = bci_ptr->value(0)*bci_ptr->calc_time_function(current_time);
                const Real angle_0 = bci_ptr->value(0)*bci_ptr->calc_time_function(previous_time);
                for (unsigned int ne=0; ne<side->n_nodes(); ne++) {
                    const Node* node = side->node_ptr(ne);
                    const Point ur = rotate_on_plane(*node, angle_1, bci_ptr->normal)
                                   - rotate_on_plane(*node, angle_0, bci_ptr->normal);
                    const Real u0_increment = ur(0),
                               u1_increment = ur(1),
                               u2_increment = ur(2);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                        Kesub[1*nvars+1](ne,ne) += penalty;
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                        Fesub[1](ne) += penalty * u1_increment;
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                                   DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                                   DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment,
                           u1_increment = bci_ptr->value(1)*U_increment,
                           u2_increment = bci_ptr->value(2)*U_increment;
                for (unsigned int ne=0; ne<side->n_nodes(); ne++) {
                    const Node* node = side->node_ptr(ne);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                        Kesub[1*nvars+1](ne,ne) += penalty;
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                        Fesub[1](ne) += penalty * u1_increment;
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                                   DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                                   DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_X == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment;
                for (unsigned int ne=0; ne<side->n_nodes(); ne++) {
                    const Node* node = side->node_ptr(ne);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Y == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u1_increment = bci_ptr->value(1)*U_increment;
                for (unsigned int ne=0; ne<side->n_nodes(); ne++) {
                    const Node* node = side->node_ptr(ne);
                    if ( set_system_matrix ) {
                        Kesub[1*nvars+1](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[1](ne) += penalty * u1_increment;
                    } else {
                        const Real DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) );
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_Z == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u2_increment = bci_ptr->value(2)*U_increment;
                for (unsigned int ne=0; ne<side->n_nodes(); ne++) {
                    const Node* node = side->node_ptr(ne);
                    if ( set_system_matrix ) {
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XY == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment,
                           u1_increment = bci_ptr->value(1)*U_increment;
                for (unsigned int ne=0; ne<side->n_nodes(); ne++) {
                    const Node* node = side->node_ptr(ne);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                        Kesub[1*nvars+1](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                        Fesub[1](ne) += penalty * u1_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                                   DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_YZ == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u1_increment = bci_ptr->value(1)*U_increment,
                           u2_increment = bci_ptr->value(2)*U_increment;
                for (unsigned int ne=0; ne<side->n_nodes(); ne++) {
                    const Node* node = side->node_ptr(ne);
                    if ( set_system_matrix ) {
                        Kesub[1*nvars+1](ne,ne) += penalty;
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[1](ne) += penalty * u1_increment;
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU1 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 1, 0) ),
                                   DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[1](ne) += penalty * (u1_increment-DU1);
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::DIRICHLET_PRESCRIBED_XZ == bci_ptr->type ) {
                const Real U_increment = bci_ptr->calc_time_function(current_time)
                                       - bci_ptr->calc_time_function(previous_time);
                const Real u0_increment = bci_ptr->value(0)*U_increment,
                           u2_increment = bci_ptr->value(2)*U_increment;
                for (unsigned int ne=0; ne<side->n_nodes(); ne++) {
                    const Node* node = side->node_ptr(ne);
                    if ( set_system_matrix ) {
                        Kesub[0*nvars+0](ne,ne) += penalty;
                        Kesub[2*nvars+2](ne,ne) += penalty;
                    }
                    if ( iteration_step == 0 ) {
                        Fesub[0](ne) += penalty * u0_increment;
                        Fesub[2](ne) += penalty * u2_increment;
                    } else {
                        const Real DU0 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 0, 0) ),
                                   DU2 = DU_sys.current_solution( node->dof_number(DU_sys.number(), 2, 0) );
                        Fesub[0](ne) += penalty * (u0_increment-DU0);
                        Fesub[2](ne) += penalty * (u2_increment-DU2);
                    }
                }
            } else if ( BoundaryConditionInfo::NEUMANN_VECTOR == bci_ptr->type ) {
                ; // do nothing here
            } else if ( BoundaryConditionInfo::NEUMANN_NORMAL == bci_ptr->type ) {
                ; // do nothing here
            } else libmesh_error();
            // check for active boundary condition
            }
            // update element stiffness matrix and rhs vector
            if ( set_system_matrix )
                system_matrix->add_matrix(Ke, dof_indices);
            system_rhs->add_vector(Fe, dof_indices);
            residual.add_vector(Fe, dof_indices);
            // free memo
            delete [] dof_indices_var;
            // end of element sides loop...
        }
        // end of active local elements loop...
    }
    // close system matrix, rhs and residual vector
    if ( set_system_matrix ) system_matrix->close();
    system_rhs->close();
    residual.close();
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("_assemble_system_essential_bcs");
}
//=================================================================================================
void NavierCauchySolver3D::_solve () {
    this->_log.push("_solve");

    EquationSystems& es = this->_dmn.get_equation_systems();
    LinearImplicitSystem& du_sys = es.get_system<LinearImplicitSystem>(this->_label+"::iterative_solution");
    System&               DU_sys = es.get_system<System>(this->_label+"::incremental_solution");
    System&               u_sys  = es.get_system<System>(this->_label+"::solution");
    //
    NumericVector<Number>&  du = *du_sys.solution;
    NumericVector<Number>&  DU = *DU_sys.solution;
    NumericVector<Number>&  u  = *u_sys.solution;
    NumericVector<Number>&  Y = du_sys.get_vector("residual");
    // solve the linear system
    du_sys.linear_solver->reuse_preconditioner(this->par_list.get<bool>("system matrix updated") ? false : true);
    du_sys.solve();
    //
    Y.close();
    //
    DU.add(this->par_list.get<Real>("relaxing parameter"), du);
    DU.close();
    DU_sys.update();
    //
    u.add(this->par_list.get<Real>("relaxing parameter"), du);
    u.close();
    u_sys.update();
    //
    this->par_list.set<Real>("iterative solution norm (i)")   = du.l2_norm();
    this->par_list.set<Real>("incremental solution norm (i)") = DU.l2_norm();
    this->par_list.set<Real>("residual norm (i)") = Y.l2_norm();
    this->par_list.set<Real>("energy norm (i)") = fabs(Y.dot(du));
    if ( this->par_list.get<int>("iteration step") == 0 ) {
        this->par_list.set<Real>("iterative solution norm (0)")   = this->par_list.get<Real>("iterative solution norm (i)");
        this->par_list.set<Real>("incremental solution norm (0)") = this->par_list.get<Real>("incremental solution norm (i)");
        this->par_list.set<Real>("residual norm (0)") = this->par_list.get<Real>("residual norm (i)");
        this->par_list.set<Real>("energy norm (0)")   = this->par_list.get<Real>("energy norm (i)");
    } else {
        this->par_list.set<Real>("previous solution ratio") = this->par_list.get<Real>("solution ratio");
    }
    this->par_list.set<Real>("solution ratio") = this->par_list.get<Real>("iterative solution norm (i)")
                                               / this->par_list.set<Real>("iterative solution norm (0)");
    this->par_list.set<Real>("residual ratio") = this->par_list.get<Real>("residual norm (i)")
                                               / this->par_list.get<Real>("residual norm (0)");
    this->par_list.set<Real>("energy ratio") = this->par_list.get<Real>("energy norm (i)")
                                             / this->par_list.get<Real>("energy norm (0)");

    this->_log.pop("_solve");
}
//=================================================================================================
int NavierCauchySolver3D::_check_convergence () const {
    this->_log.push("_check_convergence");

    int status = 0;
    const int iter_step = this->par_list.get<int>("iteration step") + 1;
    if ( this->par_list.get<bool>("load control") ) {
        const char end_of_line[] = "           \n";
        char txt_norms[buffer_len];
        sprintf( txt_norms, " |du| = %.8e; |du|/|du0| = %.8e; |Y|/|Y0| = %.8e; |du*Y|/|du0*Y0| = %.8e ",
                 this->par_list.get<Real>("iterative solution norm (i)"),
                 this->par_list.get<Real>("solution ratio"),
                 this->par_list.get<Real>("residual ratio"),
                 this->par_list.get<Real>("energy ratio")  );
        char txt[buffer_len];
        sprintf( txt, "  * Iteration %d;%s%s",
                 iter_step, txt_norms, end_of_line );
        flog.print(txt, false);
        // check for minimum number of iterations is achieved
        if ( iter_step > this->par_list.get<int>("minimum iterations") ) {
        // check if Newton-Raphson solver has converged
        if ( this->par_list.get<Real>("solution ratio") <= this->par_list.get<Real>("solution tolerance") &&
             this->par_list.get<Real>("residual ratio") <= this->par_list.get<Real>("residual tolerance") &&
             this->par_list.get<Real>("energy ratio")   <= this->par_list.get<Real>("energy tolerance")   ) {
            status = 1;
        } else if ( this->par_list.get<Real>("iterative solution norm (i)") <= 1.0e-9 ) {
            status = 1;
        } else if ( this->par_list.get<Real>("solution ratio") >= 100.0 ) {
            flog.print("    Solver diverged & Abort!!!\n", false);
            status = -1;
        } else if ( iter_step == this->par_list.get<int>("maximum iterations") &&
                    this->par_list.get<bool>("abort non-convergence")          ) {
            flog.print("    Abort non-converging solver!!!\n", false);
            status = -1;
        }
        // end if ...check for minimum number of iterations is achieved
        }
    } else {
        const char end_of_line[] = "           \n";
        char txt_norms[buffer_len];
        sprintf( txt_norms, " |du| = %.8e; |du|/|du0| = %.8e ",
                 this->par_list.get<Real>("iterative solution norm (i)"),
                 this->par_list.get<Real>("solution ratio") );
        char txt[buffer_len];
        sprintf( txt, "  * Iteration %d;%s%s",
                 iter_step, txt_norms, end_of_line );
        flog.print(txt, false);
        // check for minimum number of iterations is achieved
        if ( iter_step > this->par_list.get<int>("minimum iterations") ) {
        // check if Newton-Raphson solver has converged
        if ( this->par_list.get<Real>("solution ratio") <= this->par_list.get<Real>("solution tolerance") ) {
            status = 1;
        } else if ( this->par_list.get<Real>("iterative solution norm (i)") <= 1.0e-12 ) {
            status = 1;
        } else if ( this->par_list.get<Real>("solution ratio") >= 100.0 ) {
            flog.print("    Solver diverged & Abort!!!\n", false);
            status = -1;
        } else if ( iter_step == this->par_list.get<int>("maximum iterations") &&
                    this->par_list.get<bool>("abort non-convergence")          ) {
            flog.print("    Abort non-converging solver!!!\n", false);
            status = -1;
        }
        // end if ...check for minimum number of iterations is achieved
        }
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("_check_convergence");
    return status;
}
//=================================================================================================
void NavierCauchySolver3D::_reset_increment (bool failed) {
    this->_log.push("_reset_increment");

    EquationSystems& es = this->_dmn.get_equation_systems();
    LinearImplicitSystem& du_sys = es.get_system<LinearImplicitSystem>(this->_label+"::iterative_solution");
    System&               DU_sys = es.get_system<System>(this->_label+"::incremental_solution");
    System&               u_sys  = es.get_system<System>(this->_label+"::solution");
    System&             u_P_sys  = es.get_system<System>(this->_label+"::previous_solution");
    //
    NumericVector<Number>&  du = *du_sys.solution;
    NumericVector<Number>&  DU = *DU_sys.solution;
    NumericVector<Number>&  u  = *u_sys.solution;
    NumericVector<Number>&  u_P = *u_P_sys.solution;
    //
    if ( failed ) {
        du.close();
        DU.close();
        u.close();
        //
        du.zero();
        du.close();
        du_sys.update();
        //
        u.add(-1.0, DU);
        u.close();
        u_sys.update();
        //
        DU.zero();
        DU.close();
        DU_sys.update();
        //
        this->par_list.set<int>("maximum increments (current scale)") += 1;
        this->par_list.set<Real>("current pseudo-time")  -= this->par_list.get<Real>("pseudo-time step");
        this->par_list.set<Real>("previous pseudo-time") -= this->par_list.get<Real>("pseudo-time step");
        this->par_list.set<Real>("pseudo-time step") /= 2.0;
        this->par_list.set<int>("increment step") *= 2.0;
        this->par_list.set<int>("maximum increments") *= 2.0;
        this->par_list.set<int>("solution increment") *= 2.0;
        this->par_list.set<int>("restart increment") *= 2.0;
        // else has not failed...
    } else {
        du.close();
        DU.close();
        u.close();
        u_P.close();
        //
        du.zero();
        du.close();
        du_sys.update();
        //
        DU.zero();
        DU.close();
        DU_sys.update();
        //
        u_P.operator=(u);
        u_P.close();
        u_P_sys.update();
        //
        this->par_list.set<int>("increment step") += 1;
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("_reset_increment");
}
//=================================================================================================
void NavierCauchySolver3D::get_displacements ( const std::string& solu_type,
                                               std::vector< std::vector<Real> >& solu_u,
                                               IndexPair& solu_idx ) const
{
    this->_log.push("get_displacements");

    const EquationSystems& es = this->_dmn.get_equation_systems();
    const System& sys = es.get_system<System>(this->_label+solu_type);
    const MeshBase& msh = es.get_mesh();
    const std::string res_dir = es.parameters.get<std::string>("results directory");
    // clear memory
    solu_idx.clear();
    solu_u.clear();
    // close the solution vector
    sys.solution->close();
    // write in a temporary binary file the solution of all local nodes
    if ( true ) {
        // open the temporary binary file
        char foutBIN[buffer_len];
        sprintf(foutBIN, "%s/.nodal_solution.%d", res_dir.c_str(), libMesh::global_processor_id());
        std::ofstream fout(foutBIN, std::ios::binary);
        // write the number of local nodes
        const unsigned int n_local_nodes = msh.n_local_nodes();
        fout.write((char*)&n_local_nodes, sizeof(unsigned int));
        // loop over local nodes list
        for ( MeshBase::const_node_iterator
              citer_nd=msh.local_nodes_begin(); citer_nd!=msh.local_nodes_end(); citer_nd++ ) {
            const unsigned int node_id = (*citer_nd)->id();
            const Real         node_u0 = sys.current_solution((*citer_nd)->dof_number(sys.number(), 0, 0)),
                               node_u1 = sys.current_solution((*citer_nd)->dof_number(sys.number(), 1, 0)),
                               node_u2 = sys.current_solution((*citer_nd)->dof_number(sys.number(), 2, 0));
            fout.write((char*)&node_id, sizeof(unsigned int));
            fout.write((char*)&node_u0, sizeof(Real));
            fout.write((char*)&node_u1, sizeof(Real));
            fout.write((char*)&node_u2, sizeof(Real));
            // end of local nodes list loop...
        }
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // loop over all processes
    for (unsigned int iproc=0; iproc<libMesh::global_n_processors(); iproc++) {
        // open the temporary binary file
        char finBIN[buffer_len];
        sprintf(finBIN, "%s/.nodal_solution.%d", res_dir.c_str(), iproc);
        std::ifstream fin(finBIN, std::ios::binary);
        // read the number of local nodes
        unsigned int n_l_nodes = 0;
        fin.read((char*)&n_l_nodes, sizeof(unsigned int));
        // loop over local nodes list
        for (unsigned int inode=0; inode<n_l_nodes; inode++) {
            unsigned int node_id = 0;
            Real         node_u0 = 0.0,
                         node_u1 = 0.0,
                         node_u2 = 0.0;
            fin.read((char*)&node_id, sizeof(unsigned int));
            fin.read((char*)&node_u0, sizeof(Real));
            fin.read((char*)&node_u1, sizeof(Real));
            fin.read((char*)&node_u2, sizeof(Real));
            //
            std::vector<Real> node_u(SP3D);
            node_u[0] = node_u0; node_u[1] = node_u1; node_u[2] = node_u2;
            // check if node has already been inserted in the IndexPair container
            if ( solu_idx.confirm_id(node_id) ) continue;
            // update the corresponding containers
            solu_idx.insert(node_id);
            solu_u.push_back(node_u);
            // end of local nodes list loop...
        }
        // end of all processes loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // make sure to delete those temporary files
    if ( libMesh::global_processor_id() == 0 ) {
        for (unsigned int iproc=0; iproc<libMesh::global_n_processors(); iproc++) {
            char finBIN[buffer_len];
            sprintf(finBIN, "%s/.nodal_solution.%d", res_dir.c_str(), iproc);
            remove(finBIN);
        }
    }

    this->_log.pop("get_displacements");
}
//=================================================================================================
void NavierCauchySolver3D::get_pressure ( const std::string& solu_type,
                                          std::vector< std::vector<Real> >& solu_p,
                                          IndexPair& solu_idx ) const
{
    this->_log.push("get_pressure");

    const EquationSystems& es = this->_dmn.get_equation_systems();
    const System& sys = es.get_system<System>(this->_label+solu_type);
    const MeshBase& msh = es.get_mesh();
    const std::string res_dir = es.parameters.get<std::string>("results directory");
    // make this check first
    libmesh_assert( this->par_list.get<bool>("mixed formulation") );
    // clear memory
    solu_idx.clear();
    solu_p.clear();
    // close the solution vector
    sys.solution->close();
    // check if pressure variable is continuous across elements (nodal solution) or discontinuous
    if ( ! this->par_list.get<bool>("mixed formulation discontinuous pressure") ) {
        // find local nodes for which pressure is a variable, i.e. only vertex ones
        std::set<unsigned int> active_local_nodes;
        for ( MeshBase::const_element_iterator
            citer_el=msh.active_local_elements_begin(); citer_el!=msh.active_local_elements_end(); citer_el++ ) {
            const Elem* elem = *citer_el;
            for (unsigned int i=0; i<elem->n_nodes(); i++) {
                if ( elem->is_vertex(i) ) active_local_nodes.insert( elem->node_id(i) );
            }
            // end of active local elements list loop...
        }
        // write in a temporary binary file the solution of all local (vertex only) nodes
        if ( true ) {
            // open the temporary binary file
            char foutBIN[buffer_len];
            sprintf(foutBIN, "%s/.nodal_solution.%d", res_dir.c_str(), libMesh::global_processor_id());
            std::ofstream fout(foutBIN, std::ios::binary);
            // write the number of local nodes
            const unsigned int n_local_nodes = active_local_nodes.size();
            fout.write((char*)&n_local_nodes, sizeof(unsigned int));
            // loop over local nodes list
            for ( std::set<unsigned int>::const_iterator
                  citer_nd=active_local_nodes.begin(); citer_nd!=active_local_nodes.end(); citer_nd++ ) {
                const unsigned int node_id = (*citer_nd);
                const Node*        node_ptr = msh.node_ptr(node_id);
                const Real         node_p  = sys.current_solution(node_ptr->dof_number(sys.number(), 3, 0));
                fout.write((char*)&node_id, sizeof(unsigned int));
                fout.write((char*)&node_p, sizeof(Real));
                // end of local nodes list loop...
            }
        }
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        // loop over all processes
        for (unsigned int iproc=0; iproc<libMesh::global_n_processors(); iproc++) {
            // open the temporary binary file
            char finBIN[buffer_len];
            sprintf(finBIN, "%s/.nodal_solution.%d", res_dir.c_str(), iproc);
            std::ifstream fin(finBIN, std::ios::binary);
            // read the number of local nodes
            unsigned int n_l_nodes = 0;
            fin.read((char*)&n_l_nodes, sizeof(unsigned int));
            // loop over local nodes list
            for (unsigned int inode=0; inode<n_l_nodes; inode++) {
                unsigned int node_id = 0;
                Real         node_hp = 0.0;
                fin.read((char*)&node_id, sizeof(unsigned int));
                fin.read((char*)&node_hp, sizeof(Real));
                //
                std::vector<Real> node_p(1);
                node_p[0] = node_hp;
                // check if node has already been inserted in the IndexPair container
                if ( solu_idx.confirm_id(node_id) ) continue;
                // update the corresponding containers
                solu_idx.insert(node_id);
                solu_p.push_back(node_p);
                // end of local nodes list loop...
            }
            // end of all processes loop...
        }
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        // make sure to delete those temporary files
        if ( libMesh::global_processor_id() == 0 ) {
            for (unsigned int iproc=0; iproc<libMesh::global_n_processors(); iproc++) {
                char finBIN[buffer_len];
                sprintf(finBIN, "%s/.nodal_solution.%d", res_dir.c_str(), iproc);
                remove(finBIN);
            }
        }
        // ...else pressure DoFs are of modal basis (discontinuous across elements)...
    } else {
        // number of variables for UP mixed formulation
        const unsigned int nvars = SP3D+1;
        // write in a temporary binary file the solution of all local elements
        if ( true ) {
            // open the temporary binary file
            char foutBIN[buffer_len];
            sprintf(foutBIN, "%s/.elemental_solution.%d", res_dir.c_str(), libMesh::global_processor_id());
            std::ofstream fout(foutBIN, std::ios::binary);
            // write the number of active local elements
            const unsigned int n_active_local_elems = msh.n_active_local_elem();
            fout.write((char*)&n_active_local_elems, sizeof(unsigned int));
            // get the list of system variables
            std::vector<unsigned short int> var;
            FEM_Context((System&)sys).get_var_list(var);
            // loop over active local elements
            for ( MeshBase::const_element_iterator
                  citer_el=msh.active_local_elements_begin(); citer_el!=msh.active_local_elements_end(); citer_el++ ) {
                const Elem* elem = *citer_el;
                libmesh_assert( sys.n_vars() == nvars );
                //
                std::vector<dof_id_type> dof_indices;
                std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
                FEM_Context((System&)sys).get_dof_indices(elem, dof_indices, dof_indices_var);
                //
                std::vector<unsigned int> var_dofs;
                calc_var_dofs(nvars, dof_indices_var, var_dofs);
                //
                std::vector<Real> elem_p;
                for (unsigned int l=0; l<var_dofs[SP3D]; l++) {
                    elem_p.push_back( sys.current_solution(dof_indices_var[SP3D][l]) );
                }
                //
                const unsigned int elem_id = elem->id();
                const unsigned int elem_n_p = elem_p.size();
                fout.write((char*)&elem_id, sizeof(unsigned int));
                fout.write((char*)&elem_n_p, sizeof(unsigned int));
                for (unsigned int i=0; i<elem_n_p; i++) {
                    fout.write((char*)&elem_p[i], sizeof(Real));
                }
                // free memo
                delete [] dof_indices_var;
                // end of active local elements list loop...
            }
        }
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        // loop over all processes
        for (unsigned int iproc=0; iproc<libMesh::global_n_processors(); iproc++) {
            // open the temporary binary file
            char finBIN[buffer_len];
            sprintf(finBIN, "%s/.elemental_solution.%d", res_dir.c_str(), iproc);
            std::ifstream fin(finBIN, std::ios::binary);
            // read the number of active local elements
            unsigned int n_a_l_elems = 0;
            fin.read((char*)&n_a_l_elems, sizeof(unsigned int));
            // loop over active local elements list
            for (unsigned int ielem=0; ielem<n_a_l_elems; ielem++) {
                unsigned int elem_id = 0;
                unsigned int elem_n_p = 0;
                fin.read((char*)&elem_id, sizeof(unsigned int));
                fin.read((char*)&elem_n_p, sizeof(unsigned int));
                std::vector<Real> elem_hp(elem_n_p);
                for (unsigned int i=0; i<elem_n_p; i++) {
                    fin.read((char*)&elem_hp[i], sizeof(Real));
                }
                //
                std::vector<Real> elem_p;
                elem_p = elem_hp;
                // check if node has already been inserted in the IndexPair container
                if ( solu_idx.confirm_id(elem_id) ) continue;
                // update the corresponding containers
                solu_idx.insert(elem_id);
                solu_p.push_back(elem_p);
                // end of active local elements list loop...
            }
            // end of all processes loop...
        }
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        // make sure to delete those temporary files
        if ( libMesh::global_processor_id() == 0 ) {
            for (unsigned int iproc=0; iproc<libMesh::global_n_processors(); iproc++) {
                char finBIN[buffer_len];
                sprintf(finBIN, "%s/.elemental_solution.%d", res_dir.c_str(), iproc);
                remove(finBIN);
            }
        }
        // end if case...
    }

    this->_log.pop("get_pressure");
}
//=================================================================================================
void NavierCauchySolver3D::_calc_gradients () {
    this->_log.push("_calc_gradients");

    EquationSystems& es = this->_dmn.get_equation_systems();
    System& gr_sys = es.get_system<System>(this->_label+"::gradient_solution");
    //
    const MeshBase::const_element_iterator begin_el = es.get_mesh().active_local_elements_begin();
    const MeshBase::const_element_iterator end_el = es.get_mesh().active_local_elements_end();
    //
    FEType fe_gr_type = gr_sys.variable_type(0);
    //
    const unsigned int ninvars_NCS = es.parameters.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = es.parameters.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = es.parameters.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = es.parameters.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    const Real current_pseudo_time = this->par_list.get<Real>("current pseudo-time");
    //
    const unsigned int nvars = SP3D*SP3D + SP3D*SP3D
                             + VOIGT_SP3D + VOIGT_SP3D
                             + ninvars_NCS;
    // some sanity checks
    libmesh_assert( SP3D == es.get_mesh().mesh_dimension() );
    libmesh_assert( n_vars == gr_sys.n_vars() );
    //
    unsigned int deformation_gradient_elas_var[SP3D*SP3D];
    for (unsigned int i=0; i<SP3D; i++) {
        for (unsigned int j=0; j<SP3D; j++) {
            const unsigned int l = i*SP3D+j;
            char var_name[buffer_len];
            sprintf(var_name, "%s__Fe%d%d", this->_label.c_str(), i, j);
            deformation_gradient_elas_var[l] = gr_sys.variable_number(var_name);
        }
    }
    unsigned int deformation_gradient_inel_var[SP3D*SP3D];
    for (unsigned int i=0; i<SP3D; i++) {
        for (unsigned int j=0; j<SP3D; j++) {
            const unsigned int l = i*SP3D+j;
            char var_name[buffer_len];
            sprintf(var_name, "%s__Fp%d%d", this->_label.c_str(), i, j);
            deformation_gradient_inel_var[l] = gr_sys.variable_number(var_name);
        }
    }
    unsigned int strain_var[VOIGT_SP3D];
    for (unsigned int i=0; i<VOIGT_SP3D; i++) {
        const Dual<int> idx = indices_Voigt_3d(i);
        char var_name[buffer_len];
        sprintf(var_name, "%s__E%d%d", this->_label.c_str(), idx.get_first(), idx.get_second());
        strain_var[i] = gr_sys.variable_number(var_name);
    }
    unsigned int stress_var[VOIGT_SP3D];
    for (unsigned int i=0; i<VOIGT_SP3D; i++) {
        const Dual<int> idx = indices_Voigt_3d(i);
        char var_name[buffer_len];
        sprintf(var_name, "%s__S%d%d", this->_label.c_str(), idx.get_first(), idx.get_second());
        stress_var[i] = gr_sys.variable_number(var_name);
    }
    unsigned int internv_var[ninvars_NCS];
    for (unsigned int i=0; i<ninvars_NCS; i++) {
        char var_name[buffer_len];
        sprintf(var_name, "%s__iv%d", this->_label.c_str(), i);
        internv_var[i] = gr_sys.variable_number(var_name);
    }
    //
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        //
        QGauss quad_volu(SP3D, es.parameters.get<Order>("Gauss quadrature order"));
        UniquePtr<FEBase> fe_gr_volu = FEBase::build(SP3D, fe_gr_type);
        fe_gr_volu->attach_quadrature_rule(&quad_volu);
        //
        Dyadic3D deformation_gradient_elas,
                 deformation_gradient_inel;
        std::vector<Real> strain,
                          stress;
        std::vector<Real> internal_variables;
        //
        deformation_gradient_elas.comp = 0.0;
        deformation_gradient_inel.comp = 0.0;
        strain.assign(VOIGT_SP3D, 0.0);
        stress.assign(VOIGT_SP3D, 0.0);
        internal_variables.assign(ninvars_NCS, 0.0);
        //
        {
            // initialize and get access to the FE stuff
            const std::vector< Real >& JxW = fe_gr_volu->get_JxW();
            fe_gr_volu->reinit(elem);
            // get access to the element quadrature points data
            const QuadratureInfo* quad_inf = &this->_dmn.elem_qi_local.find(elem->id())->second;
            const unsigned int n_qp = quad_inf->XYZ.size();
            // loop over volume integration quadrature points
            for (unsigned int qp=0; qp<n_qp; qp++) {
                //
                FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
                fld.current_time = current_pseudo_time;
                //
                deformation_gradient_elas.comp += JxW[qp] * quad_inf->deform_grad_elas[qp].comp;
                deformation_gradient_inel.comp += JxW[qp] * quad_inf->deform_grad_inel[qp].comp;
                for (unsigned int ii=0; ii<VOIGT_SP3D; ii++) {
                    const Dual<int> idx = indices_Voigt_3d(ii);
                    strain[ii] += JxW[qp] * quad_inf->strain[qp](idx.get_first(),idx.get_second());
                    stress[ii] += JxW[qp] * quad_inf->stress[qp](idx.get_first(),idx.get_second());
                }
                for (unsigned int ii=0; ii<ninvars_NCS; ii++) {
                    internal_variables[ii] += JxW[qp] * quad_inf->biomech[qp][ii];
                }
                // end of quadrature points loop...
            }
            //
            const Real eVol_recip = 1.0 / elem->volume();
            // get the average field values by dividing them with the element volume
            deformation_gradient_elas.comp *= eVol_recip;
            deformation_gradient_inel.comp *= eVol_recip;
            for (unsigned int ii=0; ii<VOIGT_SP3D; ii++) {
                strain[ii] *= eVol_recip;
                stress[ii] *= eVol_recip;
            }
            for (unsigned int ii=0; ii<ninvars_NCS; ii++) {
                internal_variables[ii] *= eVol_recip;
            }
            // end of scope...
        }
        // local degrees of freedom for each 'gradient' variable
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars];
        //
        const unsigned int dof_F = gr_sys.solution->first_local_index(),
                           dof_L = gr_sys.solution->last_local_index();
        // now set the values of the elastic deformation gradient for this element
        for (unsigned int i=0; i<SP3D; i++) {
            for (unsigned int j=0; j<SP3D; j++) {
                const unsigned int k = i*SP3D+j; // local index
                const unsigned int l = k+0; // variable index
                gr_sys.get_dof_map().dof_indices(elem, dof_indices_var[l], deformation_gradient_elas_var[k]);
                const unsigned int idof = dof_indices_var[l][0];
                if ( dof_F <= idof && idof < dof_L )
                    gr_sys.solution->set(idof, deformation_gradient_elas(i,j));
            }
        }
        // now set the values of the plastic (inelastic) deformation gradient for this element
        for (unsigned int i=0; i<SP3D; i++) {
            for (unsigned int j=0; j<SP3D; j++) {
                const unsigned int k = i*SP3D+j; // local index
                const unsigned int l = k+SP3D*SP3D; // variable index
                gr_sys.get_dof_map().dof_indices(elem, dof_indices_var[l], deformation_gradient_inel_var[k]);
                const unsigned int idof = dof_indices_var[l][0];
                if ( dof_F <= idof && idof < dof_L )
                    gr_sys.solution->set(idof, deformation_gradient_inel(i,j));
            }
        }
        // now set the values of the strain for this element
        for (unsigned int i=0; i<VOIGT_SP3D; i++) {
            const unsigned int k = i; // local index
            const unsigned int l = k+2*SP3D*SP3D; // variable index
            gr_sys.get_dof_map().dof_indices(elem, dof_indices_var[l], strain_var[k]);
            const unsigned int idof = dof_indices_var[l][0];
            if ( dof_F <= idof && idof < dof_L )
                gr_sys.solution->set(idof, strain[i]);
        }
        // now set the values of the stress for this element
        for (unsigned int i=0; i<VOIGT_SP3D; i++) {
            const unsigned int k = i; // local index
            const unsigned int l = k+2*SP3D*SP3D+VOIGT_SP3D; // variable index
            gr_sys.get_dof_map().dof_indices(elem, dof_indices_var[l], stress_var[k]);
            const unsigned int idof = dof_indices_var[l][0];
            if ( dof_F <= idof && idof < dof_L )
                gr_sys.solution->set(idof, stress[i]);
        }
        // now set the values of the internal variables for this element
        for (unsigned int i=0; i<ninvars_NCS; i++) {
            const unsigned int k = i; // local index
            const unsigned int l = k+2*SP3D*SP3D+2*VOIGT_SP3D; // variable index
            gr_sys.get_dof_map().dof_indices(elem, dof_indices_var[l], internv_var[k]);
            const unsigned int idof = dof_indices_var[l][0];
            if ( dof_F <= idof && idof < dof_L )
                gr_sys.solution->set(idof, internal_variables[i]);
        }
        // free memo
        delete [] dof_indices_var;
        // end of active local elements loop...
    }
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // close the solution container and update the system
    gr_sys.solution->close();
    gr_sys.update();

    this->_log.pop("_calc_gradients");
}
//=================================================================================================
void NavierCauchySolver3D::save_restart (std::ofstream& fout) const {
    this->_log.push("save_restart");

    for (unsigned isys=0; isys<systems_ls.size(); isys++) {
        const EquationSystems* es_ptr = systems_ls[isys].first;
        const System& sys = es_ptr->get_system<System>(systems_ls[isys].second);
        //
        sys.solution->close();
        std::vector<Number> soln;
        sys.solution->localize_to_one(soln);
        //
        if ( libMesh::global_processor_id() ) continue;
        //
        const unsigned int n = soln.size();
        fout.write((char*)&n, sizeof(unsigned int));
        for (unsigned int i=0; i<n; i++) {
            const Number component = soln[i];
            fout.write((char*)&component, sizeof(Number));
        }
    }
    //
    this->_dmn.save_quadrature(fout, QuadratureInfo::ALL, QuadratureInfo::LEVEL_3);
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("save_restart");
}
//=================================================================================================
void NavierCauchySolver3D::load_restart () {
    this->_log.push("load_restart");

    const Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    for (unsigned int ifile=0; ifile<par.get<unsigned int>("input restart from file"); ifile++) {
        const std::string str_name = "input restart file name #" + std::to_string(ifile);
        const std::string restart_file_name = par.get<std::string>(str_name.c_str());
        //
        std::ifstream fin(restart_file_name.c_str());
        // read the system solution from the first restart file only
        if ( 0 == ifile )
            for (unsigned isys=0; isys<systems_ls.size(); isys++) {
                EquationSystems* es_ptr = systems_ls[isys].first;
                System& sys = es_ptr->get_system<System>(systems_ls[isys].second);
                //
                unsigned int n = 0;
                fin.read((char*)&n, sizeof(unsigned int));
                for (unsigned int i=0; i<n; i++) {
                    Number component;
                    fin.read((char*)&component, sizeof(Number));
                    sys.solution->set(i, component);
                }
            }
        // read Quadrature data for this processor only
        if ( libMesh::global_processor_id() == ifile )
            this->_dmn.load_quadrature(fin, QuadratureInfo::ALL, QuadratureInfo::LEVEL_3);
        // end of restart files loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    for (unsigned isys=0; isys<systems_ls.size(); isys++) {
        EquationSystems* es_ptr = systems_ls[isys].first;
        System& sys = es_ptr->get_system<System>(systems_ls[isys].second);
        //
        sys.solution->close();
        sys.update();
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("load_restart");
}
//=================================================================================================
