/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/nurbs.h"
//=================================================================================================
#define dz_dX    this->_inv_map_dzdX
#define dN_dz    this->_dphidz
//=================================================================================================
void NURBS::calc_dphi (const Point& crd) const {
    if ( ! this->_calculate_dphi ) return;
    // check for NURBS space
    if ( SP1D == this->dimension() ) {
        //
        for (int a=0; a<this->size(); a++) {
            this->_dphi[a].comp(0) = dz_dX(0,0)*dN_dz[0][a];
            this->_dphi[a].comp(1) = dz_dX(0,1)*dN_dz[0][a];
            this->_dphi[a].comp(2) = dz_dX(0,2)*dN_dz[0][a];
        }
        //
    } else if ( SP2D == this->dimension() ) {
        //
        for (int a=0; a<this->size(); a++) {
            this->_dphi[a].comp(0) = dz_dX(0,0)*dN_dz[0][a]
                                   + dz_dX(1,0)*dN_dz[1][a];
            this->_dphi[a].comp(1) = dz_dX(0,1)*dN_dz[0][a]
                                   + dz_dX(1,1)*dN_dz[1][a];
            this->_dphi[a].comp(2) = dz_dX(0,2)*dN_dz[0][a]
                                   + dz_dX(1,2)*dN_dz[1][a];
        }
        //
    } else if ( SP3D == this->dimension() ) {
        //
        for (int a=0; a<this->size(); a++) {
            this->_dphi[a].comp(0) = dz_dX(0,0)*dN_dz[0][a]
                                   + dz_dX(1,0)*dN_dz[1][a]
                                   + dz_dX(2,0)*dN_dz[2][a];
            this->_dphi[a].comp(1) = dz_dX(0,1)*dN_dz[0][a]
                                   + dz_dX(1,1)*dN_dz[1][a]
                                   + dz_dX(2,1)*dN_dz[2][a];
            this->_dphi[a].comp(2) = dz_dX(0,2)*dN_dz[0][a]
                                   + dz_dX(1,2)*dN_dz[1][a]
                                   + dz_dX(2,2)*dN_dz[2][a];
        }
        //
    } else libmesh_error();
}
//=================================================================================================
#undef dz_dX
#undef dN_dz
//=================================================================================================
void NURBS::calc_dphidxi (const Point& crd) const {
    const int L = 0;
    if ( ! this->_calculate_dphidz[L] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a);
            this->_dphidz[L][a] = this->_basef_1st_deriv<0>(crd(0), i);
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a);
            this->_dphidz[L][a] = this->_basef_1st_deriv<0>(crd(0), i)
                                * this->_basef_0th_deriv<1>(crd(1), j);
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_dphidz[L][a] = this->_basef_1st_deriv<0>(crd(0), i)
                                * this->_basef_0th_deriv<1>(crd(1), j)
                                * this->_basef_0th_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_dphi(L);
}
//=================================================================================================
void NURBS::calc_dphideta (const Point& crd) const {
    const int L = 1;
    if ( ! this->_calculate_dphidz[L] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_dphidz[L][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a);
            this->_dphidz[L][a] = this->_basef_0th_deriv<0>(crd(0), i)
                                * this->_basef_1st_deriv<1>(crd(1), j);
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_dphidz[L][a] = this->_basef_0th_deriv<0>(crd(0), i)
                                * this->_basef_1st_deriv<1>(crd(1), j)
                                * this->_basef_0th_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_dphi(L);
}
//=================================================================================================
void NURBS::calc_dphidzeta (const Point& crd) const {
    const int L = 2;
    if ( ! this->_calculate_dphidz[L] ) return;
    //
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a);
            this->_dphidz[L][a] = 0.0;
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            //const int i = this->index<0>(a),
            //          j = this->index<1>(a);
            this->_dphidz[L][a] = 0.0;
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_dphidz[L][a] = this->_basef_0th_deriv<0>(crd(0), i)
                                * this->_basef_0th_deriv<1>(crd(1), j)
                                * this->_basef_1st_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_dphi(L);
}
//=================================================================================================
