/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/nurbs.h"
//=================================================================================================
NURBS::NURBS (const dof_id_type id) : _id(id), _sbd_id(0), _processor_id(-1), _dim(0) {
    for (int d=0; d<SP3D; d++) {
        this->_open_begin[d] = this->_open_end[d] = true;
        this->_order[d] = static_cast<int>(libMeshEnums::CONSTANT);
    }
}
//=================================================================================================
NURBS::NURBS (const int ncp[], const Order ord[], const dof_id_type id) {
    this->init(ncp, ord, id);
}
//=================================================================================================
NURBS::~NURBS () {
    delete [] this->_neighbor;
    delete [] this->_boundary_id;
}
//=================================================================================================
void NURBS::init (const int ncp[], const Order ord[], const dof_id_type id) {
    // setup the ID of the NURBS object
    this->_id = id;
    // make some checks and set dimension of the NURBS
    if ( ncp[0] < 1 || ncp[1] < 1 || ncp[2] < 1 ) libmesh_error();
    else if ( ncp[2] == 1 && ncp[1] == 1 && ncp[0] >= 2 ) this->_dim = 1;
    else if ( ncp[2] == 1 && ncp[1] >= 2 && ncp[0] >= 2 ) this->_dim = 2;
    else if ( ncp[2] >= 2 && ncp[1] >= 2 && ncp[0] >= 2 ) this->_dim = 3;
    else libmesh_error();
    // set the order of the polynomial basis
    for (int d=0; d<SP3D; d++) {
        this->_order[d] = static_cast<int>(ord[d]);
        if ( this->_order[d] >= ncp[d] ) libmesh_error();
    }
    // allocate memory for the control points and their weights
    this->_ctrl_pnt.resize(ncp[0], ncp[1], ncp[2]);
    this->_weight.resize(ncp[0], ncp[1], ncp[2]);
    // set up the indices lookup table
    for (int i=0; i<ncp[0]; i++) {
        for (int j=0; j<ncp[1]; j++) {
            for (int k=0; k<ncp[2]; k++) {
                Blitz_VectorInt idxs(SP3D);
                idxs = i, j, k;
                this->_idx_table.push_back(idxs);
            }
        }
    }
    // set the default value for the weights
    this->_weight = 1.0;
    this->_is_rational = false;
    // allocate memory the knot vectors and setup the default values
    for (int d=0; d<SP3D; d++) {
        this->_open_begin[d] = this->_open_end[d] = true;
        // allocate memo
        this->_knot_vec[d].resize(this->n_ctrl_pnt(d)+this->order(d)+1);
        this->_knot_vec[d] = 0.0;
        for (int i=this->n_ctrl_pnt(d); i<this->n_ctrl_pnt(d)+this->order(d)+1; i++) {
            this->_knot_vec[d](i) = 1.0;
        }
        const Real D_crd = 1.0 / (this->n_ctrl_pnt(d)-this->order(d));
        for (int i=this->order(d)+1; i<this->n_ctrl_pnt(d); i++) {
            this->_knot_vec[d](i) = this->_knot_vec[d](i-1) + D_crd;
        }
        //
        if ( this->dimension() < d+1 ) this->_knot_vec[d] = 0.0;
        // set the non-repeated knot vector container
        std::set<Real> knot_v;
        for (int i=0; i<this->_knot_vec[d].extent(0); i++) {
            knot_v.insert( this->_knot_vec[d](i) );
        }
        this->_nonrepeated_knot_vec[d].resize( knot_v.size() );
        int j = 0;
        for (std::set<Real>::const_iterator ci=knot_v.begin(); ci!=knot_v.end(); ci++) {
            this->_nonrepeated_knot_vec[d](j) = *ci;
            ++j;
        }
    }
    //
    const unsigned int Ncp = this->size();
    // allocate memory for the neighborhood NURBS (if necessary)
    switch ( this->dimension() ) {
    case SP1D :
        this->_neighbor = new NURBS* [2];
        this->_boundary_id = new boundary_id_type [2];
        for (int l=0; l<2; l++) {
            this->_neighbor[l] = NULL;
            this->_boundary_id[l] = -1;
        }
        break;
    case SP2D :
        this->_neighbor = new NURBS* [4];
        this->_boundary_id = new boundary_id_type [4];
        for (int l=0; l<4; l++) {
            this->_neighbor[l] = NULL;
            this->_boundary_id[l] = -1;
        }
        break;
    case SP3D :
        this->_neighbor = new NURBS* [6];
        this->_boundary_id = new boundary_id_type [6];
        for (int l=0; l<6; l++) {
            this->_neighbor[l] = NULL;
            this->_boundary_id[l] = -1;
        }
        break;
    default :
        libmesh_error();
    }
    // allocate memory for the shape functions and their derivatives
    this->_phi.resize(Ncp);
    for (int i=0; i<SP3D; i++) {
        this->_dphidz[i].resize(Ncp);
        for (int j=0; j<SP3D; j++) {
            this->_d2phidz2[i][j].resize(Ncp);
            for (int k=0; k<SP3D; k++) {
                this->_d3phidz3[i][j][k].resize(Ncp);
            }
        }
    }
    this->_dphi.resize(Ncp);
    this->_d2phi.resize(Ncp);
    this->_d3phi.resize(Ncp);
    // set all flags to default
    this->_calculate_coord = false;
    for (int i=0; i<SP3D; i++) {
        this->_calculate_dphidz[i] = false;
        for (int j=0; j<SP3D; j++) {
            this->_calculate_d2phidz2[i][j] = false;
            for (int k=0; k<SP3D; k++) {
                this->_calculate_d3phidz3[i][j][k] = false;
            }
        }
    }
    this->_calculate_jacobian   =
    this->_calculate_jacobian_2 =
    this->_calculate_jacobian_3 = false;
    this->_calculate_dphi  =
    this->_calculate_d2phi =
    this->_calculate_d3phi = false;
}
//=================================================================================================
void NURBS::print_info (std::ostream& os, bool coord_out, bool reduced_out) const {
    const int DIM = this->dimension();
    //
    std::ostringstream oss;
    oss << "\n ******************************** \n";
    oss << " *** NURBS object information *** \n";
    oss << " ******************************** \n";
    oss << '\n';
    oss << "     ID: " << this->id() << '\n';
    oss << "     dimension: " << DIM << '\n';
    oss << "     rational: " << (this->_is_rational ? "true" : "false") << '\n';
    oss << "     subdomain ID: " << this->subdomain_id() << '\n';
    for (int d=0; d<DIM; d++)
        oss << "     polynomial order dim#" << (d+1) << ": " << this->order(d) << '\n';
    for (int d=0; d<DIM; d++)
        oss << "     control points dim#" << (d+1) << ": " << this->n_ctrl_pnt(d) << '\n';
    oss << "     control nodes (total=" << this->size() << "): \n";
    if ( ! reduced_out )
    for (int a=0; a<this->_idx_table.size(); a++) {
        oss << " A : " << a << "  - ";
        const int i = this->index<0>(a),
                  j = this->index<1>(a),
                  k = this->index<2>(a);
        oss <<     " I : " << i
            << "  -  J : " << j
            << "  -  K : " << k;
        if ( this->_is_rational )
            oss << "  -  Wgt : " << this->weight(i,j,k);
        oss << "  -  Node : " << this->get_ctrl_node(i,j,k)->id();
        if ( coord_out ) {
            oss << "  -  (";
            for (int d=0; d<SP3D; d++) {
                if ( d ) oss << ',';
                oss << this->ctrl_pnt(i,j,k)(d);
            }
            oss << ")";
        }
        oss << '\n';
    }
    oss << "     knots (total=" << this->all_knots() << "): \n";
    if ( ! reduced_out )
    for (int d=0; d<DIM; d++) {
        const unsigned int n = this->all_knots(d);
        oss << "     knot vector dim#" << (d+1)
             << " (size=" << n << "): \n";
        for (int i=0; i<n; i++)
            oss << ' ' << this->_knot_vec[d](i);
        oss << '\n';
    }
    int n_b;
    if      ( this->dimension() == SP1D ) n_b = 2;
    else if ( this->dimension() == SP2D ) n_b = 4;
    else if ( this->dimension() == SP3D ) n_b = 6;
    else libmesh_error();
    if ( ! reduced_out )
    oss << "     bounds (total=" << n_b << "): \n";
    for (int l=0; l<n_b; l++) {
        if ( this->_neighbor[l] == NULL ) {
            oss << " B : " << this->_boundary_id[l] << " \n";
        } else {
            oss << " I : " << this->_neighbor[l]->id() << " \n";
        }
    }
    //
    oss << std::flush;
    os << oss.str();
}
//=================================================================================================
void NURBS::reinit (const Point& crd) const {
    libmesh_assert( crd(0) >= 0 && crd(0) <= 1.0 );
    libmesh_assert( crd(1) >= 0 && crd(1) <= 1.0 );
    libmesh_assert( crd(2) >= 0 && crd(2) <= 1.0 );
    //
    this->_jac_det = 0.0;
    this->_inv_map_dzdX.free();
    this->_inv_map_d2zdX2.free();
    this->_inv_map_d3zdX3.free();
    this->_map_dXdz.free();
    this->_map_d2Xdz2.free();
    this->_map_d3Xdz3.free();
    //
    this->calc_phi(crd);
    this->calc_xyz(crd);
    //
    this->calc_dphidxi(crd);
    this->calc_dphideta(crd);
    this->calc_dphidzeta(crd);
    this->calc_jacobian(crd);
    this->calc_dphi(crd);
    //
    this->calc_d2phidxidxi(crd);   this->calc_d2phidxideta(crd);   this->calc_d2phidxidzeta(crd);
    this->calc_d2phidetadxi(crd);  this->calc_d2phidetadeta(crd);  this->calc_d2phidetadzeta(crd);
    this->calc_d2phidzetadxi(crd); this->calc_d2phidzetadeta(crd); this->calc_d2phidzetadzeta(crd);
    this->calc_jacobian_2(crd);
    this->calc_d2phi(crd);
    //
    this->calc_d3phidxidxidxi(crd);     this->calc_d3phidxidxideta(crd);     this->calc_d3phidxidxidzeta(crd);
    this->calc_d3phidxidetadxi(crd);    this->calc_d3phidxidetadeta(crd);    this->calc_d3phidxidetadzeta(crd);
    this->calc_d3phidxidzetadxi(crd);   this->calc_d3phidxidzetadeta(crd);   this->calc_d3phidxidzetadzeta(crd);
    this->calc_d3phidetadxidxi(crd);    this->calc_d3phidetadxideta(crd);    this->calc_d3phidetadxidzeta(crd);
    this->calc_d3phidetadetadxi(crd);   this->calc_d3phidetadetadeta(crd);   this->calc_d3phidetadetadzeta(crd);
    this->calc_d3phidetadzetadxi(crd);  this->calc_d3phidetadzetadeta(crd);  this->calc_d3phidetadzetadzeta(crd);
    this->calc_d3phidzetadxidxi(crd);   this->calc_d3phidzetadxideta(crd);   this->calc_d3phidzetadxidzeta(crd);
    this->calc_d3phidzetadetadxi(crd);  this->calc_d3phidzetadetadeta(crd);  this->calc_d3phidzetadetadzeta(crd);
    this->calc_d3phidzetadzetadxi(crd); this->calc_d3phidzetadzetadeta(crd); this->calc_d3phidzetadzetadzeta(crd);
    this->calc_jacobian_3(crd);
    this->calc_d3phi(crd);
    //
    // set all flags to default
    for (int i=0; i<SP3D; i++) {
        this->_calculate_dphidz[i] = false;
        for (int j=0; j<SP3D; j++) {
            this->_calculate_d2phidz2[i][j] = false;
            for (int k=0; k<SP3D; k++) {
                this->_calculate_d3phidz3[i][j][k] = false;
            }
        }
    }
    this->_calculate_jacobian   =
    this->_calculate_jacobian_2 =
    this->_calculate_jacobian_3 = false;
    this->_calculate_dphi  =
    this->_calculate_d2phi =
    this->_calculate_d3phi = false;
}
//=================================================================================================
void NURBS::calc_xyz (const Point& crd) const {
    if ( ! this->_calculate_coord ) return;
    //
    this->_xyz = 0.0;
    for (int a=0; a<this->size(); a++) {
        this->_xyz += this->_phi[a]
                    * this->ctrl_pnt_indexed(a);
    }
}
//=================================================================================================
void NURBS::calc_phi (const Point& crd) const {
    // check for NRUBS space
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a);
            this->_phi[a] = this->_basef_0th_deriv<0>(crd(0), i);
        }
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a);
            this->_phi[a] = this->_basef_0th_deriv<0>(crd(0), i)
                          * this->_basef_0th_deriv<1>(crd(1), j);
        }
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const int i = this->index<0>(a),
                      j = this->index<1>(a),
                      k = this->index<2>(a);
            this->_phi[a] = this->_basef_0th_deriv<0>(crd(0), i)
                          * this->_basef_0th_deriv<1>(crd(1), j)
                          * this->_basef_0th_deriv<2>(crd(2), k);
        }
    }
    //
    if ( this->_is_rational ) this->_rationalize_phi();
}
//=================================================================================================
#define dz_dX    this->_inv_map_dzdX
#define d2z_dX2  this->_inv_map_d2zdX2
#define d3z_dX3  this->_inv_map_d3zdX3
//=================================================================================================
#define dX_dz   this->_map_dXdz
//=================================================================================================
void NURBS::calc_jacobian (const Point& crd) const {
    if ( ! this->_calculate_jacobian ) return;
    // \partial{Cartesian coordinates} / \partial{Parametric coordinates}
    dX_dz.resize(SP3D, this->dimension());
    dX_dz = 0.0;
    // \partial{Parametric coordinates} / \partial{Cartesian coordinates}
    dz_dX.resize(this->dimension(), SP3D);
    dz_dX = 0.0;
    // check for NURBS space
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const Point& X_a = this->ctrl_pnt_indexed(a);
            for (int d=0; d<SP3D; d++) {
                dX_dz(d,0) += this->_dphidz[0][a] * X_a(d);
            }
        }
        //
        this->_jac_det = sqrt(pow2(dX_dz(0,0))+pow2(dX_dz(1,0))+pow2(dX_dz(2,0)));
        const Real recip_J = 1.0 / pow2(this->_jac_det);
        //
        dz_dX(0,0) = dX_dz(0,0) * recip_J;
        dz_dX(0,1) = dX_dz(1,0) * recip_J;
        dz_dX(0,2) = dX_dz(2,0) * recip_J;
        //
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const Point& X_a = this->ctrl_pnt_indexed(a);
            for (int d=0; d<SP3D; d++) {
                dX_dz(d,0) += this->_dphidz[0][a] * X_a(d);
                dX_dz(d,1) += this->_dphidz[1][a] * X_a(d);
            }
        }
        //
        Dyadic3D dX_dz_2;
        dX_dz_2.comp = dX_dz(0,0)*dX_dz(0,0)+dX_dz(0,1)*dX_dz(0,1), dX_dz(0,0)*dX_dz(1,0)+dX_dz(0,1)*dX_dz(1,1), dX_dz(0,0)*dX_dz(2,0)+dX_dz(0,1)*dX_dz(2,1),
                       dX_dz(0,0)*dX_dz(1,0)+dX_dz(0,1)*dX_dz(1,1), dX_dz(1,0)*dX_dz(1,0)+dX_dz(1,1)*dX_dz(1,1), dX_dz(1,0)*dX_dz(2,0)+dX_dz(1,1)*dX_dz(2,1),
                       dX_dz(0,0)*dX_dz(2,0)+dX_dz(0,1)*dX_dz(2,1), dX_dz(1,0)*dX_dz(2,0)+dX_dz(1,1)*dX_dz(2,1), dX_dz(2,0)*dX_dz(2,0)+dX_dz(2,1)*dX_dz(2,1);
        //
        const Dyadic3D dX_dz_inv = inverse(dX_dz_2);
        //
        const Point e1(dX_dz(0,0), dX_dz(1,0), dX_dz(2,0));
        const Point e2(dX_dz(0,1), dX_dz(1,1), dX_dz(2,1));
        this->_jac_det = e1.cross(e2).size();
        //
        dz_dX(0,0) = dX_dz(0,0)*dX_dz_inv(0,0)+dX_dz(1,0)*dX_dz_inv(1,0)+dX_dz(2,0)*dX_dz_inv(2,0);
        dz_dX(0,1) = dX_dz(0,0)*dX_dz_inv(0,1)+dX_dz(1,0)*dX_dz_inv(1,1)+dX_dz(2,0)*dX_dz_inv(2,1);
        dz_dX(0,2) = dX_dz(0,0)*dX_dz_inv(0,2)+dX_dz(1,0)*dX_dz_inv(1,2)+dX_dz(2,0)*dX_dz_inv(2,2);
        dz_dX(1,0) = dX_dz(0,1)*dX_dz_inv(0,0)+dX_dz(1,1)*dX_dz_inv(1,0)+dX_dz(2,1)*dX_dz_inv(2,0);
        dz_dX(1,1) = dX_dz(0,1)*dX_dz_inv(0,1)+dX_dz(1,1)*dX_dz_inv(1,1)+dX_dz(2,1)*dX_dz_inv(2,1);
        dz_dX(1,2) = dX_dz(0,1)*dX_dz_inv(0,2)+dX_dz(1,1)*dX_dz_inv(1,2)+dX_dz(2,1)*dX_dz_inv(2,2);
        //
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const Point& X_a = this->ctrl_pnt_indexed(a);
            for (int d=0; d<SP3D; d++) {
                dX_dz(d,0) += this->_dphidz[0][a] * X_a(d);
                dX_dz(d,1) += this->_dphidz[1][a] * X_a(d);
                dX_dz(d,2) += this->_dphidz[2][a] * X_a(d);
            }
        }
        //
        Dyadic3D jac;
        jac.comp =  dX_dz(0,0), dX_dz(0,1), dX_dz(0,2),
                    dX_dz(1,0), dX_dz(1,1), dX_dz(1,2),
                    dX_dz(2,0), dX_dz(2,1), dX_dz(2,2);
        //
        this->_jac_det = determinant(jac);
        const Real recip_J = 1.0 / this->_jac_det;
        //
        dz_dX(0,0) =  (jac(1,1)*jac(2,2)-jac(2,1)*jac(1,2)) * recip_J;
        dz_dX(0,1) = -(jac(0,1)*jac(2,2)-jac(2,1)*jac(0,2)) * recip_J;
        dz_dX(0,2) =  (jac(0,1)*jac(1,2)-jac(1,1)*jac(0,2)) * recip_J;
        dz_dX(1,0) = -(jac(1,0)*jac(2,2)-jac(2,0)*jac(1,2)) * recip_J;
        dz_dX(1,1) =  (jac(0,0)*jac(2,2)-jac(2,0)*jac(0,2)) * recip_J;
        dz_dX(1,2) = -(jac(0,0)*jac(1,2)-jac(1,0)*jac(0,2)) * recip_J;
        dz_dX(2,0) =  (jac(1,0)*jac(2,1)-jac(2,0)*jac(1,1)) * recip_J;
        dz_dX(2,1) = -(jac(0,0)*jac(2,1)-jac(2,0)*jac(0,1)) * recip_J;
        dz_dX(2,2) =  (jac(0,0)*jac(1,1)-jac(1,0)*jac(0,1)) * recip_J;
        //
    } else libmesh_error();
}
//=================================================================================================
#undef dX_dz
//=================================================================================================
#define d2X_dz2   this->_map_d2Xdz2
//=================================================================================================
void NURBS::calc_jacobian_2 (const Point& crd) const {
    if ( ! this->_calculate_jacobian_2 ) return;
    // \partial{Cartesian coordinates} / \partial{Parametric coordinates}
    d2X_dz2.resize(SP3D, this->dimension(), this->dimension());
    d2X_dz2 = 0.0;
    // check for NURBS space
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const Point& X_a = this->ctrl_pnt_indexed(a);
            for (int d=0; d<SP3D; d++) {
                d2X_dz2(d,0,0) += this->_d2phidz2[0][0][a] * X_a(d);
            }
        }
        //
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const Point& X_a = this->ctrl_pnt_indexed(a);
            for (int d=0; d<SP3D; d++) {
                d2X_dz2(d,0,0) += this->_d2phidz2[0][0][a] * X_a(d);
                d2X_dz2(d,0,1) += this->_d2phidz2[0][1][a] * X_a(d);
                d2X_dz2(d,1,0) += this->_d2phidz2[1][0][a] * X_a(d);
                d2X_dz2(d,1,1) += this->_d2phidz2[1][1][a] * X_a(d);
            }
        }
        //
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const Point& X_a = this->ctrl_pnt_indexed(a);
            for (int d=0; d<SP3D; d++) {
                d2X_dz2(d,0,0) += this->_d2phidz2[0][0][a] * X_a(d);
                d2X_dz2(d,0,1) += this->_d2phidz2[0][1][a] * X_a(d);
                d2X_dz2(d,0,2) += this->_d2phidz2[0][2][a] * X_a(d);
                d2X_dz2(d,1,0) += this->_d2phidz2[1][0][a] * X_a(d);
                d2X_dz2(d,1,1) += this->_d2phidz2[1][1][a] * X_a(d);
                d2X_dz2(d,1,2) += this->_d2phidz2[1][2][a] * X_a(d);
                d2X_dz2(d,2,0) += this->_d2phidz2[2][0][a] * X_a(d);
                d2X_dz2(d,2,1) += this->_d2phidz2[2][1][a] * X_a(d);
                d2X_dz2(d,2,2) += this->_d2phidz2[2][2][a] * X_a(d);
            }
        }
        //
    } else libmesh_error();
    //
    blitz::Array<Real, 3> _d2zdX2_0(this->dimension(), SP3D, SP3D);
    {
        //
        blitz::firstIndex  i;
        blitz::secondIndex j;
        blitz::thirdIndex  k;
        blitz::fourthIndex l;
        //
        blitz::Array<Real, 3> _d2zdX2_1A(SP3D, SP3D, this->dimension());
        _d2zdX2_1A = sum(-d2X_dz2(i,l,k)*dz_dX(l,j), l); // NOTE: the minus sign!!!
        //
        blitz::Array<Real, 3> _d2zdX2_1B(SP3D, SP3D, SP3D);
        _d2zdX2_1B = sum(_d2zdX2_1A(i,j,l)*dz_dX(l,k), l);
        //
        _d2zdX2_0 = sum(dz_dX(i,l)*_d2zdX2_1B(l,j,k), l);
        //
    }
    // \partial{Parametric coordinates} / \partial{Cartesian coordinates}
    d2z_dX2.resize(this->dimension(), SP3D, SP3D);
    d2z_dX2 = _d2zdX2_0;
}
//=================================================================================================
#undef d2X_dz2
//=================================================================================================
#define d2X_dz2   this->_map_d2Xdz2
#define d3X_dz3   this->_map_d3Xdz3
//=================================================================================================
void NURBS::calc_jacobian_3 (const Point& crd) const {
    if ( ! this->_calculate_jacobian_3 ) return;

    // \partial{Cartesian coordinates} / \partial{Parametric coordinates}
    d3X_dz3.resize(SP3D, this->dimension(), this->dimension(), this->dimension());
    d3X_dz3 = 0.0;
    // check for NURBS space
    if        ( SP1D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const Point& X_a = this->ctrl_pnt_indexed(a);
            for (int d=0; d<SP3D; d++) {
                d3X_dz3(d,0,0,0) += this->_d3phidz3[0][0][0][a] * X_a(d);
            }
        }
        //
    } else if ( SP2D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const Point& X_a = this->ctrl_pnt_indexed(a);
            for (int d=0; d<SP3D; d++) {
                d3X_dz3(d,0,0,0) += this->_d3phidz3[0][0][0][a] * X_a(d);
                d3X_dz3(d,0,0,1) += this->_d3phidz3[0][0][1][a] * X_a(d);
                d3X_dz3(d,0,1,0) += this->_d3phidz3[0][1][0][a] * X_a(d);
                d3X_dz3(d,0,1,1) += this->_d3phidz3[0][1][1][a] * X_a(d);
                d3X_dz3(d,1,0,0) += this->_d3phidz3[1][0][0][a] * X_a(d);
                d3X_dz3(d,1,0,1) += this->_d3phidz3[1][0][1][a] * X_a(d);
                d3X_dz3(d,1,1,0) += this->_d3phidz3[1][1][0][a] * X_a(d);
                d3X_dz3(d,1,1,1) += this->_d3phidz3[1][1][1][a] * X_a(d);
            }
        }
        //
    } else if ( SP3D == this->dimension() ) {
        for (int a=0; a<this->size(); a++) {
            const Point& X_a = this->ctrl_pnt_indexed(a);
            for (int d=0; d<SP3D; d++) {
                d3X_dz3(d,0,0,0) += this->_d3phidz3[0][0][0][a] * X_a(d);
                d3X_dz3(d,0,0,1) += this->_d3phidz3[0][0][1][a] * X_a(d);
                d3X_dz3(d,0,0,2) += this->_d3phidz3[0][0][2][a] * X_a(d);
                d3X_dz3(d,0,1,0) += this->_d3phidz3[0][1][0][a] * X_a(d);
                d3X_dz3(d,0,1,1) += this->_d3phidz3[0][1][1][a] * X_a(d);
                d3X_dz3(d,0,1,2) += this->_d3phidz3[0][1][2][a] * X_a(d);
                d3X_dz3(d,0,2,0) += this->_d3phidz3[0][2][0][a] * X_a(d);
                d3X_dz3(d,0,2,1) += this->_d3phidz3[0][2][1][a] * X_a(d);
                d3X_dz3(d,0,2,2) += this->_d3phidz3[0][2][2][a] * X_a(d);
                d3X_dz3(d,1,0,0) += this->_d3phidz3[1][0][0][a] * X_a(d);
                d3X_dz3(d,1,0,1) += this->_d3phidz3[1][0][1][a] * X_a(d);
                d3X_dz3(d,1,0,2) += this->_d3phidz3[1][0][2][a] * X_a(d);
                d3X_dz3(d,1,1,0) += this->_d3phidz3[1][1][0][a] * X_a(d);
                d3X_dz3(d,1,1,1) += this->_d3phidz3[1][1][1][a] * X_a(d);
                d3X_dz3(d,1,1,2) += this->_d3phidz3[1][1][2][a] * X_a(d);
                d3X_dz3(d,1,2,0) += this->_d3phidz3[1][2][0][a] * X_a(d);
                d3X_dz3(d,1,2,1) += this->_d3phidz3[1][2][1][a] * X_a(d);
                d3X_dz3(d,1,2,2) += this->_d3phidz3[1][2][2][a] * X_a(d);
                d3X_dz3(d,2,0,0) += this->_d3phidz3[2][0][0][a] * X_a(d);
                d3X_dz3(d,2,0,1) += this->_d3phidz3[2][0][1][a] * X_a(d);
                d3X_dz3(d,2,0,2) += this->_d3phidz3[2][0][2][a] * X_a(d);
                d3X_dz3(d,2,1,0) += this->_d3phidz3[2][1][0][a] * X_a(d);
                d3X_dz3(d,2,1,1) += this->_d3phidz3[2][1][1][a] * X_a(d);
                d3X_dz3(d,2,1,2) += this->_d3phidz3[2][1][2][a] * X_a(d);
                d3X_dz3(d,2,2,0) += this->_d3phidz3[2][2][0][a] * X_a(d);
                d3X_dz3(d,2,2,1) += this->_d3phidz3[2][2][1][a] * X_a(d);
                d3X_dz3(d,2,2,2) += this->_d3phidz3[2][2][2][a] * X_a(d);
            }
        }
        //
    }
    //
    blitz::Array<Real, 4> _d3zdX3_0(this->dimension(), SP3D, SP3D, SP3D);
    {
        //
        blitz::firstIndex  i;
        blitz::secondIndex j;
        blitz::thirdIndex  k;
        blitz::fourthIndex l;
        blitz::fifthIndex  m;
        //
        blitz::Array<Real, 4> _d3zdX3_0A(SP3D, SP3D, this->dimension(), this->dimension());
        _d3zdX3_0A = sum(-d3X_dz3(i,m,k,l)*dz_dX(m,j), m); // NOTE: the minus sign!!!
        //
        blitz::Array<Real, 4> _d3zdX3_0B(SP3D, SP3D, SP3D, this->dimension());
        _d3zdX3_0B = sum(_d3zdX3_0A(i,j,m,l)*dz_dX(m,k), m);
        //
        blitz::Array<Real, 4> _d3zdX3_0C(SP3D, SP3D, SP3D, SP3D);
        _d3zdX3_0C = sum(_d3zdX3_0B(i,j,k,m)*dz_dX(m,l), m);
        //
        _d3zdX3_0 = sum(_d3zdX3_0C(m,j,k,m)*dz_dX(i,m), m);
        //
    }
    //
    blitz::Array<Real, 4> _d3zdX3_1(this->dimension(), SP3D, SP3D, SP3D);
    {
        //
        blitz::firstIndex  i;
        blitz::secondIndex j;
        blitz::thirdIndex  k;
        blitz::fourthIndex l;
        blitz::fifthIndex  m;
        //
        blitz::Array<Real, 3> _d3zdX3_1A(SP3D, this->dimension(), SP3D);
        _d3zdX3_1A = sum(-d2X_dz2(i,j,l)*dz_dX(l,k), l); // NOTE: the minus sign!!!
        //
        blitz::Array<Real, 3> _d3zdX3_1B(this->dimension(), this->dimension(), SP3D);
        _d3zdX3_1B = sum(_d3zdX3_1A(l,j,k)*dz_dX(i,l), l );
        //
        _d3zdX3_1 = sum(_d3zdX3_1B(i,m,k)*d2z_dX2(m,j,l), m);
        //
    }
    //
    blitz::Array<Real, 4> _d3zdX3_2(this->dimension(), SP3D, SP3D, SP3D);
    {
        //
        blitz::firstIndex  i;
        blitz::secondIndex j;
        blitz::thirdIndex  k;
        blitz::fourthIndex l;
        blitz::fifthIndex  m;
        //
        blitz::Array<Real, 3> _d3zdX3_2A(SP3D, SP3D, this->dimension());
        _d3zdX3_2A = sum(-d2X_dz2(i,l,k)*dz_dX(l,j), l); // NOTE: the minus sign!!!
        //
        blitz::Array<Real, 3> _d3zdX3_2B(this->dimension(), SP3D, this->dimension());
        _d3zdX3_2B = sum(_d3zdX3_2A(l,j,k)*dz_dX(i,l), l);
        //
        _d3zdX3_2 = sum(_d3zdX3_2B(i,j,m)*d2z_dX2(m,k,l), m);
        //
    }
    //
    blitz::Array<Real, 4> _d3zdX3_3(this->dimension(), SP3D, SP3D, SP3D);
    {
        //
        blitz::firstIndex  i;
        blitz::secondIndex j;
        blitz::thirdIndex  k;
        blitz::fourthIndex l;
        blitz::fifthIndex  m;
        //
        blitz::Array<Real, 3> _d3zdX3_3A(SP3D, SP3D, this->dimension());
        _d3zdX3_3A = sum(-d2X_dz2(i,l,k)*dz_dX(l,j), l); // NOTE: the minus sign!!!
        //
        blitz::Array<Real, 3> _d3zdX3_3B(SP3D, SP3D, SP3D);
        _d3zdX3_3B = sum(_d3zdX3_3A(i,j,l)*dz_dX(l,k), l);
        //
        _d3zdX3_3 = sum(_d3zdX3_3B(m,j,k)*d2z_dX2(i,m,l), m);
        //
    }
    // \partial{Parametric coordinates} / \partial{Cartesian coordinates}
    d3z_dX3.resize(this->dimension(), SP3D, SP3D, SP3D);
    d3z_dX3 = _d3zdX3_0 + _d3zdX3_1 + _d3zdX3_2 + _d3zdX3_3;
}
//=================================================================================================
#undef d2X_dz2
#undef d3X_dz3
//=================================================================================================
#undef dz_dX
#undef d2z_dX2
#undef d3z_dX3
//=================================================================================================
UniquePtr<const NURBS> NURBS::build_side_ptr (const unsigned int side) const {
    UniquePtr<NURBS> face(new NURBS());
    //
    int ncp[SP3D] = { 1, 1, 1 };
    Order ord[SP3D] = { CONSTANT, CONSTANT, CONSTANT };
    int idim, jdim;
    Blitz_VectorReal kv[SP2D];
    // distinguish for NURBS dimension
    if ( this->dimension() == SP1D ) {
        libmesh_error();
    } else if ( this->dimension() == SP2D ) {
        switch ( side ) {
        case 0 :
            idim = 0;
            ncp[0] = this->n_ctrl_pnt(idim);
            ord[0] = (Order)this->order(idim);
            face->init(ncp, ord, DofObject::invalid_id);
            if ( this->_is_rational )
                face->set_to_rational();
            for (int i=0; i<this->n_ctrl_pnt(idim); i++) {
                int I = i;
                int J = 0;
                face->set_ctrl_node(i) = this->get_ctrl_node(I, J);
                if ( this->_is_rational )
                    face->set_weight(i) = this->weight(I, J);
            }
            kv[0].resize(this->all_knots(idim));
            kv[0] = this->get_knot_vec(idim);
            break;
        case 1 :
            idim = 1;
            ncp[0] = this->n_ctrl_pnt(idim);
            ord[0] = (Order)this->order(idim);
            face->init(ncp, ord, DofObject::invalid_id);
            if ( this->_is_rational )
                face->set_to_rational();
            for (int i=0; i<this->n_ctrl_pnt(idim); i++) {
                int I = this->n_ctrl_pnt(0)-1;
                int J = i;
                face->set_ctrl_node(i) = this->get_ctrl_node(I, J);
                if ( this->_is_rational )
                    face->set_weight(i) = this->weight(I, J);
            }
            kv[0].resize(this->all_knots(idim));
            kv[0] = this->get_knot_vec(idim);
            break;
        case 2 :
            idim = 0; // NOTE: opposite direction
            ncp[0] = this->n_ctrl_pnt(idim);
            ord[0] = (Order)this->order(idim);
            face->init(ncp, ord, DofObject::invalid_id);
            if ( this->_is_rational )
                face->set_to_rational();
            for (int i=0; i<this->n_ctrl_pnt(idim); i++) {
                int I = this->n_ctrl_pnt(idim)-1-i;
                int J = this->n_ctrl_pnt(1)-1;
                face->set_ctrl_node(i) = this->get_ctrl_node(I, J);
                if ( this->_is_rational )
                    face->set_weight(i) = this->weight(I, J);
            }
            kv[0].resize(this->all_knots(idim));
            kv[0] = 1.0-this->get_knot_vec(idim); kv[0].reverseSelf(0);
            break;
        case 3 :
            idim = 1; // NOTE: opposite direction
            ncp[0] = this->n_ctrl_pnt(idim);
            ord[0] = (Order)this->order(idim);
            face->init(ncp, ord, DofObject::invalid_id);
            if ( this->_is_rational )
                face->set_to_rational();
            for (int i=0; i<this->n_ctrl_pnt(idim); i++) {
                int I = 0;
                int J = this->n_ctrl_pnt(idim)-1-i;
                face->set_ctrl_node(i) = this->get_ctrl_node(I, J);
                if ( this->_is_rational )
                    face->set_weight(i) = this->weight(I, J);
            }
            kv[0].resize(this->all_knots(idim));
            kv[0] = 1.0-this->get_knot_vec(idim); kv[0].reverseSelf(0);
            break;
        default :
            libmesh_error();
        }
        //
        face->set_knot_vec(0, kv[0]);
        if ( this->_is_rational )
            face->set_to_rational();
    } else if ( this->dimension() == SP3D ) {
        switch ( side ) {
        case 0 :
            idim = 0;
            jdim = 1; // NOTE: opposite direction
            ncp[0] = this->n_ctrl_pnt(idim);   ncp[1] = this->n_ctrl_pnt(jdim);
            ord[0] = (Order)this->order(idim); ord[1] = (Order)this->order(jdim);
            face->init(ncp, ord, DofObject::invalid_id);
            if ( this->_is_rational )
                face->set_to_rational();
            for (int i=0; i<this->n_ctrl_pnt(idim); i++) {
                for (int j=0; j<this->n_ctrl_pnt(jdim); j++) {
                    int I = i;
                    int J = this->n_ctrl_pnt(jdim)-1-j;
                    int K = 0;
                    face->set_ctrl_node(i,j) = this->get_ctrl_node(I, J, K);
                    if ( this->_is_rational )
                        face->set_weight(i,j) = this->weight(I, J, K);
                }
            }
            kv[0].resize(this->all_knots(idim));
            kv[0] = this->get_knot_vec(idim);
            kv[1].resize(this->all_knots(jdim));
            kv[1] = 1.0-this->get_knot_vec(jdim); kv[1].reverseSelf(0);
            break;
        case 1 :
            idim = 0;
            jdim = 2;
            ncp[0] = this->n_ctrl_pnt(idim);   ncp[1] = this->n_ctrl_pnt(jdim);
            ord[0] = (Order)this->order(idim); ord[1] = (Order)this->order(jdim);
            face->init(ncp, ord, DofObject::invalid_id);
            if ( this->_is_rational )
                face->set_to_rational();
            for (int i=0; i<this->n_ctrl_pnt(idim); i++) {
                for (int j=0; j<this->n_ctrl_pnt(jdim); j++) {
                    int I = i;
                    int J = 0;
                    int K = j;
                    face->set_ctrl_node(i,j) = this->get_ctrl_node(I, J, K);
                    if ( this->_is_rational )
                        face->set_weight(i,j) = this->weight(I, J, K);
                }
            }
            kv[0].resize(this->all_knots(idim));
            kv[0] = this->get_knot_vec(idim);
            kv[1].resize(this->all_knots(jdim));
            kv[1] = this->get_knot_vec(jdim);
            break;
        case 2 :
            idim = 1;
            jdim = 2;
            ncp[0] = this->n_ctrl_pnt(idim);   ncp[1] = this->n_ctrl_pnt(jdim);
            ord[0] = (Order)this->order(idim); ord[1] = (Order)this->order(jdim);
            face->init(ncp, ord, DofObject::invalid_id);
            if ( this->_is_rational )
                face->set_to_rational();
            for (int i=0; i<this->n_ctrl_pnt(idim); i++) {
                for (int j=0; j<this->n_ctrl_pnt(jdim); j++) {
                    int I = this->n_ctrl_pnt(0)-1;
                    int J = i;
                    int K = j;
                    face->set_ctrl_node(i,j) = this->get_ctrl_node(I, J, K);
                    if ( this->_is_rational )
                        face->set_weight(i,j) = this->weight(I, J, K);
                }
            }
            kv[0].resize(this->all_knots(idim));
            kv[0] = this->get_knot_vec(idim);
            kv[1].resize(this->all_knots(jdim));
            kv[1] = this->get_knot_vec(jdim);
            break;
        case 3 :
            idim = 0;
            jdim = 2;
            ncp[0] = this->n_ctrl_pnt(idim);   ncp[1] = this->n_ctrl_pnt(jdim);
            ord[0] = (Order)this->order(idim); ord[1] = (Order)this->order(jdim);
            face->init(ncp, ord, DofObject::invalid_id);
            if ( this->_is_rational )
                face->set_to_rational();
            for (int i=0; i<this->n_ctrl_pnt(idim); i++) {
                for (int j=0; j<this->n_ctrl_pnt(jdim); j++) {
                    int I = i;
                    int J = this->n_ctrl_pnt(1)-1;
                    int K = j;
                    face->set_ctrl_node(i,j) = this->get_ctrl_node(I, J, K);
                    if ( this->_is_rational )
                        face->set_weight(i,j) = this->weight(I, J, K);
                }
            }
            kv[0].resize(this->all_knots(idim));
            kv[0] = this->get_knot_vec(idim);
            kv[1].resize(this->all_knots(jdim));
            kv[1] = this->get_knot_vec(jdim);
            break;
            // idim = 0; // NOTE: opposite direction
            // jdim = 2;
            // ncp[0] = this->n_ctrl_pnt(idim);   ncp[1] = this->n_ctrl_pnt(jdim);
            // ord[0] = (Order)this->order(idim); ord[1] = (Order)this->order(jdim);
            // face->init(ncp, ord, DofObject::invalid_id);
            // if ( this->_is_rational )
            //     face->set_to_rational();
            // for (int i=0; i<this->n_ctrl_pnt(idim); i++) {
            //     for (int j=0; j<this->n_ctrl_pnt(jdim); j++) {
            //         int I = this->n_ctrl_pnt(idim)-1-i;
            //         int J = this->n_ctrl_pnt(1)-1;
            //         int K = j;
            //         face->set_ctrl_node(i,j) = this->get_ctrl_node(I, J, K);
            //         if ( this->_is_rational )
            //             face->set_weight(i,j) = this->weight(I, J, K);
            //     }
            // }
            // kv[0].resize(this->all_knots(idim));
            // kv[0] = 1.0-this->get_knot_vec(idim); kv[0].reverseSelf(0);
            // kv[1].resize(this->all_knots(jdim));
            // kv[1] = this->get_knot_vec(jdim);
            // break;
        case 4 :
            idim = 1;
            jdim = 2;
            ncp[0] = this->n_ctrl_pnt(idim);   ncp[1] = this->n_ctrl_pnt(jdim);
            ord[0] = (Order)this->order(idim); ord[1] = (Order)this->order(jdim);
            face->init(ncp, ord, DofObject::invalid_id);
            if ( this->_is_rational )
                face->set_to_rational();
            for (int i=0; i<this->n_ctrl_pnt(idim); i++) {
                for (int j=0; j<this->n_ctrl_pnt(jdim); j++) {
                    int I = 0;
                    int J = i;
                    int K = j;
                    face->set_ctrl_node(i,j) = this->get_ctrl_node(I, J, K);
                    if ( this->_is_rational )
                        face->set_weight(i,j) = this->weight(I, J, K);
                }
            }
            kv[0].resize(this->all_knots(idim));
            kv[0] = this->get_knot_vec(idim);
            kv[1].resize(this->all_knots(jdim));
            kv[1] = this->get_knot_vec(jdim);
            break;
            // idim = 1; // NOTE: opposite direction
            // jdim = 2;
            // ncp[0] = this->n_ctrl_pnt(idim);   ncp[1] = this->n_ctrl_pnt(jdim);
            // ord[0] = (Order)this->order(idim); ord[1] = (Order)this->order(jdim);
            // face->init(ncp, ord, DofObject::invalid_id);
            // if ( this->_is_rational )
            //     face->set_to_rational();
            // for (int i=0; i<this->n_ctrl_pnt(idim); i++) {
            //     for (int j=0; j<this->n_ctrl_pnt(jdim); j++) {
            //         int I = 0;
            //         int J = this->n_ctrl_pnt(idim)-1-i;
            //         int K = j;
            //         face->set_ctrl_node(i,j) = this->get_ctrl_node(I, J, K);
            //         if ( this->_is_rational )
            //             face->set_weight(i,j) = this->weight(I, J, K);
            //     }
            // }
            // kv[0].resize(this->all_knots(idim));
            // kv[0] = 1.0-this->get_knot_vec(idim); kv[0].reverseSelf(0);
            // kv[1].resize(this->all_knots(jdim));
            // kv[1] = this->get_knot_vec(jdim);
            // break;
        case 5 :
            idim = 0;
            jdim = 1;
            ncp[0] = this->n_ctrl_pnt(idim);   ncp[1] = this->n_ctrl_pnt(jdim);
            ord[0] = (Order)this->order(idim); ord[1] = (Order)this->order(jdim);
            face->init(ncp, ord, DofObject::invalid_id);
            if ( this->_is_rational )
                face->set_to_rational();
            for (int i=0; i<this->n_ctrl_pnt(idim); i++) {
                for (int j=0; j<this->n_ctrl_pnt(jdim); j++) {
                    int I = i;
                    int J = j;
                    int K =  this->n_ctrl_pnt(2)-1;
                    face->set_ctrl_node(i,j) = this->get_ctrl_node(I, J, K);
                    if ( this->_is_rational )
                        face->set_weight(i,j) = this->weight(I, J, K);
                }
            }
            kv[0].resize(this->all_knots(idim));
            kv[0] = this->get_knot_vec(idim);
            kv[1].resize(this->all_knots(jdim));
            kv[1] = this->get_knot_vec(jdim);
            break;
        default :
            libmesh_error();
        }
        //
        face->set_knot_vec(0, kv[0]);
        face->set_knot_vec(1, kv[1]);
        if ( this->_is_rational )
            face->set_to_rational();
        // ...else exception is caught...
    } else libmesh_error();
    //
    return std::move(face);
}
//=================================================================================================
