/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/paraview_io.h"
//=================================================================================================
#ifdef SMALLEST_NUMBER
#error "macro parameter \"SMALLEST_NUMBER\" has been defined elsewhere!"
#endif
#define SMALLEST_NUMBER   1.0e-24
//=================================================================================================
void ParaviewIO::write_nodal_data (const std::string& fname, EquationSystems& es) {
    const MeshBase& msh = MeshOutput<MeshBase>::mesh();
    //
    MeshBase::const_node_iterator cni_begin = msh.nodes_begin();
    MeshBase::const_node_iterator cni_end   = msh.nodes_end();
    MeshBase::const_element_iterator cei_begin = msh.active_elements_begin();
    MeshBase::const_element_iterator cei_end   = msh.active_elements_end();
    //
    // build the list of variable names that will be written
    std::vector<std::string> names;
    es.build_variable_names(names, libmesh_nullptr);
    // build the nodal solution values & get the variable names from the equation systems
    std::vector<Number> soln;
    es.build_solution_vector(soln);
    //
    unsigned int n_cell = 0;
    unsigned int n_point = 0;
    //
    std::vector<bool> nodes_bool;
    nodes_bool.assign(msh.n_nodes(), false);
    std::vector<int> nodes_vtk_id;
    nodes_vtk_id.assign(msh.n_nodes(), -1);
    //
    for (MeshBase::const_element_iterator cei=cei_begin; cei!=cei_end; cei++) {
        const Elem* elem = *cei;
        for (unsigned int i=0; i<elem->n_nodes(); i++)
            nodes_bool[elem->node_id(i)] = true;
        ++n_cell;
    }
    //
    for (MeshBase::const_node_iterator cni=cni_begin; cni!=cni_end; cni++) {
        const Node* node = *cni;
        if ( ! nodes_bool[node->id()] ) continue;
        nodes_vtk_id[node->id()] = n_point;
        ++n_point;
    }
    //
    // open and write-up in the VTU file
    std::ofstream  fvtu(fname.c_str());
    //
    fvtu << "<VTKFile type=\"UnstructuredGrid\" version=\"0.1\" byte_order=\"LittleEndian\">" << std::endl;
    fvtu << "  <UnstructuredGrid>" << std::endl;
    fvtu << "    <Piece  NumberOfPoints=\"" << n_point << "\" NumberOfCells=\"" << n_cell << "\">" << std::endl;
    fvtu << "      <Points>" << std::endl;
    fvtu << "        <DataArray type=\"Float64\" Name=\"position\" NumberOfComponents=\"3\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_node_iterator cni=cni_begin; cni!=cni_end; cni++) {
        if ( ! nodes_bool[(*cni)->id()] ) continue;
        //
        for (unsigned int i=0; i<SP3D; i++)
            fvtu << " " << (*cni)->operator()(i) << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "      </Points>" << std::endl;
    fvtu << "      <PointData>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"node_ID\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_node_iterator cni=cni_begin; cni!=cni_end; cni++) {
        if ( ! nodes_bool[(*cni)->id()] ) continue;
        //
        fvtu << " " << ((*cni)->id()+1) << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    const unsigned int n_es_vars = names.size();
    for (unsigned int j=0; j<n_es_vars; j++) {
        const char* vn = names[j].c_str();
        //
        fvtu << "        <DataArray type=\"Float64\" Name=\"" << vn << "\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        for (MeshBase::const_node_iterator cni=cni_begin; cni!=cni_end; cni++) {
            if ( ! nodes_bool[(*cni)->id()] ) continue;
            //
            const Real r = soln[((*cni)->id()*n_es_vars)+j];
            fvtu << " " << ( fabs(r)<=SMALLEST_NUMBER ? 0.0 : r ) << std::flush;
        }
        fvtu << std::endl;
        fvtu << "        </DataArray>" << std::endl;
    }
    fvtu << "      </PointData>" << std::endl;
    fvtu << "      <CellData>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"element_ID\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_element_iterator cei=cei_begin; cei!=cei_end; cei++) {
        const Elem* elem = *cei;
        fvtu << " " << (elem->id()+1) << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"region_ID\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_element_iterator cei=cei_begin; cei!=cei_end; cei++) {
        const Elem* elem = *cei;
        fvtu << " " << elem->subdomain_id() << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"processor_ID\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_element_iterator cei=cei_begin; cei!=cei_end; cei++) {
        const Elem* elem = *cei;
        fvtu << " " << elem->processor_id() << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "      </CellData>" << std::endl;
    fvtu << "      <Cells>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"connectivity\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_element_iterator cei=cei_begin; cei!=cei_end; cei++) {
        const Elem* elem = *cei;
        std::vector<unsigned int> vtk_cell_conn;
        this->_cell_connectivity(elem, vtk_cell_conn);
        for (unsigned int i=0; i<vtk_cell_conn.size(); i++)
            fvtu << " " << nodes_vtk_id[vtk_cell_conn[i]] << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"offsets\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    unsigned int offset = 0;
    for (MeshBase::const_element_iterator cei=cei_begin; cei!=cei_end; cei++) {
        const Elem* elem = *cei;
        offset += this->_cell_offset(elem);
        fvtu << " " << offset << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"types\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_element_iterator cei=cei_begin; cei!=cei_end; cei++) {
        const Elem* elem = *cei;
        fvtu << " " << this->_cell_type(elem) << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "      </Cells>" << std::endl;
    fvtu << "    </Piece>" << std::endl;
    fvtu << "  </UnstructuredGrid>" << std::endl;
    fvtu << "</VTKFile>" << std::endl;
}
//=================================================================================================
void ParaviewIO::_write_ascii ( const std::string& fname, const std::vector<Number>* soln,
                                const std::vector<std::string>* names )
{
    const MeshBase& msh = MeshOutput<MeshBase>::mesh();
    //
    MeshBase::const_element_iterator* p_cei_begin = NULL;
    MeshBase::const_element_iterator* p_cei_end   = NULL;
    MeshBase::const_element_iterator cei_a_begin = msh.active_elements_begin();
    MeshBase::const_element_iterator cei_a_end   = msh.active_elements_end();
    MeshBase::const_element_iterator cei_0_begin = msh.level_elements_begin(0);
    MeshBase::const_element_iterator cei_0_end   = msh.level_elements_end(0);
    if ( this->_write_refined_mesh ) {
        p_cei_begin = & cei_a_begin;
        p_cei_end   = & cei_a_end;
    } else {
        p_cei_begin = & cei_0_begin;
        p_cei_end   = & cei_0_end;
    }
    //
    unsigned int n_cell = 0;
    unsigned int n_point = 0;
    //
    std::vector<bool> nodes_bool;
    nodes_bool.assign(msh.n_nodes(), false);
    for (MeshBase::const_element_iterator cei=(*p_cei_begin); cei!=(*p_cei_end); cei++) {
        const Elem* elem = *cei;
        if ( this->_write_reduced_mesh ) {
            for (unsigned int i=0; i<elem->n_vertices(); i++)
                nodes_bool[elem->node_id(i)] = true;
        } else {
            for (unsigned int i=0; i<elem->n_nodes(); i++)
                nodes_bool[elem->node_id(i)] = true;
        }
        ++n_cell;
    }
    //
    std::vector<int> nodes_vtk_id;
    nodes_vtk_id.assign(msh.n_nodes(), -1);
    for (unsigned int i=0; i<nodes_bool.size(); i++) {
        if ( ! nodes_bool[i] ) continue;
        //
        nodes_vtk_id[i] = n_point;
        ++n_point;
    }
    // open and write VTU file
    std::ofstream  fvtu(fname.c_str());
    //
    fvtu << "<VTKFile type=\"UnstructuredGrid\" version=\"0.1\" byte_order=\"LittleEndian\">" << std::endl;
    fvtu << "  <UnstructuredGrid>" << std::endl;
    fvtu << "    <Piece  NumberOfPoints=\"" << n_point << "\" NumberOfCells=\"" << n_cell << "\">" << std::endl;
    fvtu << "      <Points>" << std::endl;
    fvtu << "        <DataArray type=\"Float64\" Name=\"position\" NumberOfComponents=\"3\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_node_iterator cni=msh.nodes_begin(); cni!=msh.nodes_end(); cni++) {
        if ( ! nodes_bool[(*cni)->id()] ) continue;
        //
        for (unsigned int i=0; i<SP3D; i++)
            fvtu << " " << (*cni)->operator()(i) << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "      </Points>" << std::endl;
    fvtu << "      <PointData>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"node_ID\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_node_iterator cni=msh.nodes_begin(); cni!=msh.nodes_end(); cni++) {
        if ( ! nodes_bool[(*cni)->id()] ) continue;
        //
        fvtu << " " << ((*cni)->id()+1) << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    const unsigned int n_es_vars = names->size();
    for (unsigned int j=0; j<n_es_vars; j++) {
        if ( this->_is_hidden_variable((*names)[j]) ) continue;
        //
        fvtu << "        <DataArray type=\"Float64\" Name=\"" << (*names)[j].c_str() << "\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
        for (MeshBase::const_node_iterator cni=msh.nodes_begin(); cni!=msh.nodes_end(); cni++) {
            if ( ! nodes_bool[(*cni)->id()] ) continue;
            //
            const Real result = (*soln)[((*cni)->id()*n_es_vars)+j];
            fvtu << " " << ( fabs(result)<=SMALLEST_NUMBER ? 0.0 : result ) << std::flush;
        }
        fvtu << std::endl;
        fvtu << "        </DataArray>" << std::endl;
    }
    fvtu << "      </PointData>" << std::endl;
    fvtu << "      <CellData>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"element_ID\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_element_iterator cei=(*p_cei_begin); cei!=(*p_cei_end); cei++) {
        const Elem* elem = *cei;
        fvtu << " " << (elem->id()+1) << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"region_ID\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_element_iterator cei=(*p_cei_begin); cei!=(*p_cei_end); cei++) {
        const Elem* elem = *cei;
        fvtu << " " << elem->subdomain_id() << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"processor_ID\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_element_iterator cei=(*p_cei_begin); cei!=(*p_cei_end); cei++) {
        const Elem* elem = *cei;
        fvtu << " " << elem->processor_id() << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "      </CellData>" << std::endl;
    fvtu << "      <Cells>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"connectivity\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_element_iterator cei=(*p_cei_begin); cei!=(*p_cei_end); cei++) {
        const Elem* elem = *cei;
        std::vector<unsigned int> vtk_cell_conn;
        if ( this->_write_reduced_mesh ) {
            UniquePtr<Elem> elem_red = Elem::build(Elem::first_order_equivalent_type(elem->type()));
            for (unsigned int i=0; i<elem->n_vertices(); i++)
                elem_red.get()->set_node(i) = const_cast<Node*>(elem->node_ptr(i));
            this->_cell_connectivity(elem_red.get(), vtk_cell_conn);
        } else {
            this->_cell_connectivity(elem, vtk_cell_conn);
        }
        for (unsigned int i=0; i<vtk_cell_conn.size(); i++) {
            fvtu << " " << nodes_vtk_id[vtk_cell_conn[i]] << std::flush;
        }
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"offsets\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    unsigned int offset = 0;
    for (MeshBase::const_element_iterator cei=(*p_cei_begin); cei!=(*p_cei_end); cei++) {
        const Elem* elem = *cei;
        if ( this->_write_reduced_mesh ) {
            UniquePtr<Elem> elem_red = Elem::build(Elem::first_order_equivalent_type(elem->type()));
            for (unsigned int i=0; i<elem->n_vertices(); i++)
                elem_red.get()->set_node(i) = const_cast<Node*>(elem->node_ptr(i));
            offset += this->_cell_offset(elem_red.get());
        } else {
            offset += this->_cell_offset(elem);
        }
        fvtu << " " << offset << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"types\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_element_iterator cei=(*p_cei_begin); cei!=(*p_cei_end); cei++) {
        const Elem* elem = *cei;
        if ( this->_write_reduced_mesh ) {
            UniquePtr<Elem> elem_red = Elem::build(Elem::first_order_equivalent_type(elem->type()));
            for (unsigned int i=0; i<elem->n_vertices(); i++)
                elem_red.get()->set_node(i) = const_cast<Node*>(elem->node_ptr(i));
            fvtu << " " << this->_cell_type(elem_red.get()) << std::flush;
        } else {
            fvtu << " " << this->_cell_type(elem) << std::flush;
        }
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "      </Cells>" << std::endl;
    fvtu << "    </Piece>" << std::endl;
    fvtu << "  </UnstructuredGrid>" << std::endl;
    fvtu << "</VTKFile>" << std::endl;
}
//=================================================================================================
void ParaviewIO::write_boundary (const std::string& fname) {
    if ( libMesh::global_processor_id() ) return;
    //
    const MeshBase& msh = MeshOutput<MeshBase>::mesh();
    //
    MeshBase::const_element_iterator cei_begin = msh.active_elements_begin();
    MeshBase::const_element_iterator cei_end   = msh.active_elements_end();
    //
    unsigned int n_cell = 0;
    unsigned int n_point = 0;
    //
    std::vector<bool> nodes_bool;
    nodes_bool.assign(msh.n_nodes(), false);
    std::vector<int> nodes_vtk_id;
    nodes_vtk_id.assign(msh.n_nodes(), -1);
    //
    for (MeshBase::const_element_iterator cei=cei_begin; cei!=cei_end; cei++) {
        const Elem* elem = *cei;
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            if ( elem->neighbor_ptr(s) ) continue;
            //
            UniquePtr<const Elem> side = elem->build_side_ptr(s);
            for (unsigned int i=0; i<side->n_nodes(); i++)
                nodes_bool[side->node_id(i)] = true;
            ++n_cell;
        }
    }
    //
    for (unsigned int i=0; i<nodes_bool.size(); i++) {
        if ( ! nodes_bool[i] ) continue;
        //
        nodes_vtk_id[i] = n_point;
        ++n_point;
    }
    // open and write VTU file
    std::ofstream  fvtu(fname.c_str());
    //
    fvtu << "<VTKFile type=\"UnstructuredGrid\" version=\"0.1\" byte_order=\"LittleEndian\">" << std::endl;
    fvtu << "  <UnstructuredGrid>" << std::endl;
    fvtu << "    <Piece  NumberOfPoints=\"" << n_point << "\" NumberOfCells=\"" << n_cell << "\">" << std::endl;
    fvtu << "      <Points>" << std::endl;
    fvtu << "        <DataArray type=\"Float64\" Name=\"position\" NumberOfComponents=\"3\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_node_iterator cni=msh.nodes_begin(); cni!=msh.nodes_end(); ++cni) {
        if ( ! nodes_bool[(*cni)->id()] ) continue;
        //
        for (unsigned int i=0; i<SP3D; i++)
            fvtu << " " << (*cni)->operator()(i) << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "      </Points>" << std::endl;
    fvtu << "      <PointData>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"node_ID\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_node_iterator cni=msh.nodes_begin(); cni!=msh.nodes_end(); ++cni) {
        if ( ! nodes_bool[(*cni)->id()] ) continue;
        //
        fvtu << " " << ((*cni)->id()+1) << std::flush;
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "      </PointData>" << std::endl;
    fvtu << "      <CellData>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"side_element_ID\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    unsigned int side_id = 1;
    for (MeshBase::const_element_iterator cei=cei_begin; cei!=cei_end; cei++) {
        const Elem* elem = *cei;
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            if ( elem->neighbor_ptr(s) ) continue;
            //
            fvtu << " " << side_id << std::flush;
            ++side_id;
        }
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"BC_ID\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_element_iterator cei=cei_begin; cei!=cei_end; cei++) {
        const Elem* elem = *cei;
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            if ( elem->neighbor_ptr(s) ) continue;
            //
            fvtu << " " << msh.boundary_info->boundary_id(elem, s) << std::flush;
        }
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "      </CellData>" << std::endl;
    fvtu << "      <Cells>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"connectivity\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_element_iterator cei=cei_begin; cei!=cei_end; cei++) {
        const Elem* elem = *cei;
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            if ( elem->neighbor_ptr(s) ) continue;
            //
            UniquePtr<const Elem> side = elem->build_side_ptr(s);
            std::vector<unsigned int> vtk_cell_conn;
            this->_cell_connectivity(side.get(), vtk_cell_conn);
            for (unsigned int i=0; i<vtk_cell_conn.size(); i++) {
                fvtu << " " << nodes_vtk_id[vtk_cell_conn[i]] << std::flush;
            }
        }
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"offsets\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    unsigned int offset = 0;
    for (MeshBase::const_element_iterator cei=cei_begin; cei!=cei_end; cei++) {
        const Elem* elem = *cei;
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            if ( elem->neighbor_ptr(s) ) continue;
            //
            UniquePtr<const Elem> side = elem->build_side_ptr(s);
            offset += this->_cell_offset(side.get());
            fvtu << " " << offset << std::flush;
        }
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "        <DataArray type=\"Int32\" Name=\"types\" NumberOfComponents=\"1\" format=\"ascii\">" << std::endl;
    for (MeshBase::const_element_iterator cei=cei_begin; cei!=cei_end; cei++) {
        const Elem* elem = *cei;
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            if ( elem->neighbor_ptr(s) ) continue;
            //
            UniquePtr<const Elem> side = elem->build_side_ptr(s);
            fvtu << " " << this->_cell_type(side.get()) << std::flush;
        }
    }
    fvtu << std::endl;
    fvtu << "        </DataArray>" << std::endl;
    fvtu << "      </Cells>" << std::endl;
    fvtu << "    </Piece>" << std::endl;
    fvtu << "  </UnstructuredGrid>" << std::endl;
    fvtu << "</VTKFile>" << std::endl;
    //
    //
    //
    const std::set<boundary_id_type>& bcs_list = msh.boundary_info->get_boundary_ids();
    for ( std::set<boundary_id_type>::const_iterator
          ci=bcs_list.begin(); ci!=bcs_list.end(); ci++ )
    {
        const unsigned int bc_id = *ci;
        //
        std::vector<bool> nodes__bb;
        nodes__bb.assign(msh.n_nodes(), false);
        std::vector<int> nodes__id;
        nodes__id.assign(msh.n_nodes(), -1);
        //
        n_point = 0;
        n_cell = 0;
        //
        for (MeshBase::const_element_iterator cei=cei_begin; cei!=cei_end; cei++) {
            const Elem* elem = *cei;
            for (unsigned int s=0; s<elem->n_sides(); s++) {
                if ( elem->neighbor_ptr(s) ) continue;
                if ( msh.boundary_info->boundary_id(elem, s) != bc_id ) continue;
                //
                UniquePtr<const Elem> side = elem->build_side_ptr(s);
                for (unsigned int i=0; i<side->n_nodes(); i++)
                    nodes__bb[side->node_id(i)] = true;
                ++n_cell;
            }
        }
        //
        for (unsigned int i=0; i<nodes__bb.size(); i++) {
            if ( ! nodes__bb[i] ) continue;
            //
            nodes__id[i] = n_point;
            ++n_point;
        }
        //
        char fname_msh[buffer_len];
        sprintf(fname_msh, "%s.%d.msh", fname.c_str(), bc_id);
        std::ofstream  fmsh(fname_msh);
        //
        fmsh << "$MeshFormat" << std::endl;
        fmsh << "2.0 0 8" << std::endl;
        fmsh << "$EndMeshFormat" << std::endl;
        fmsh << "$Nodes" << std::endl;
        fmsh << n_point << std::endl;
        for (MeshBase::const_node_iterator cni=msh.nodes_begin(); cni!=msh.nodes_end(); ++cni) {
            if ( ! nodes__bb[(*cni)->id()] ) continue;
            //
            fmsh << 1+nodes__id[(*cni)->id()] << " ";
            for (unsigned int i=0; i<SP3D; i++)
                fmsh << " " << (*cni)->operator()(i);
            fmsh << std::endl;
        }
        fmsh << "$EndNodes" << std::endl;
        fmsh << "$Elements" << std::endl;
        fmsh << n_cell << std::endl;
        n_cell = 0;
        for (MeshBase::const_element_iterator cei=cei_begin; cei!=cei_end; cei++) {
            const Elem* elem = *cei;
            for (unsigned int s=0; s<elem->n_sides(); s++) {
                if ( elem->neighbor_ptr(s) ) continue;
                if ( msh.boundary_info->boundary_id(elem, s) != bc_id ) continue;
                //
                UniquePtr<const Elem> side = elem->build_side_ptr(s);
                fmsh << 1+n_cell << " ";
                if      ( side->type() == TRI3  ) fmsh << " 2  2 " << bc_id << " 0 ";
                else if ( side->type() == TRI6  ) fmsh << " 9  2 " << bc_id << " 0 ";
                else if ( side->type() == QUAD4 ) fmsh << " 3  2 " << bc_id << " 0 ";
                else if ( side->type() == QUAD9 ) fmsh << " 10  2 " << bc_id << " 0 ";
                else if ( side->type() == QUAD8 ) fmsh << " 16  2 " << bc_id << " 0 ";
                else libmesh_error();
                for (unsigned int i=0; i<side->n_nodes(); i++)
                    fmsh << " " << 1+nodes__id[side->node_id(i)];
                fmsh << std::endl;
                ++n_cell;
            }
        }
        fmsh << "$EndElements" << std::endl;
        // end of boundary IDs list loop...
    }
}
//=================================================================================================
#undef SMALLEST_NUMBER
//=================================================================================================
