/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/rbf.h"
#include "feb3/rbf_poly.h"
#include "feb3/rbf_func.h"
#include <numeric>
#include <algorithm>
//=================================================================================================
// external global variables
extern
OFStream flog, ferr;
//=================================================================================================
#define CRUDE_SOLVE
#undef  CRUDE_SOLVE
//=================================================================================================
RBFmap3D::RBFmap3D (const RBFunction f, const PolynomialOrder o) {
    // check for the radial basis function
    switch ( f ) {
    case RBF_MQ :
        this->_radial_func = &multiquadric;
        break;
    case RBF_EG :
        this->_radial_func = &exp_gaussian;
        break;
    case RBF_WU_C2 :
        this->_radial_func = &wu_c2;
        break;
    case RBF_WU_C4 :
        this->_radial_func = &wu_c4;
        break;
    case RBF_WENDLAND_C2 :
        this->_radial_func = &wendland_c2;
        break;
    case RBF_WENDLAND_C4 :
        this->_radial_func = &wendland_c4;
        break;
    case RBF_WENDLAND_C6 :
        this->_radial_func = &wendland_c6;
        break;
    default :
        libmesh_error();
    }
    // check for the polynomial basis
    switch ( o ) {
    case PO_LINEAR :
        this->_polynomial = &linear_polynomial;
        this->_mbasis = 4;
        break;
    case PO_BILINEAR :
        this->_polynomial = &bilinear_polynomial;
        this->_mbasis = 7;
        break;
    case PO_QUADRATIC :
        this->_polynomial = &quadratic_polynomial;
        this->_mbasis = 10;
        break;
    case PO_BIQUADRATIC :
        this->_polynomial = &biquadratic_polynomial;
        this->_mbasis = 16;
        break;
    case PO_CUBIC :
        this->_polynomial = &cubic_polynomial;
        this->_mbasis = 19;
        break;
    default :
        libmesh_error();
    }
    // make sure to initialize sometime
    this->_is_init = false;
    this->_coeff = NULL;
    this->_pnt = NULL;
}
//=================================================================================================
void RBFmap3D::init ( const Real* coeff,
                      const std::vector<Point>& pnt, const std::vector<Real> soln[] ) {
    const int npnt = pnt.size(),
              mbas = this->_mbasis;
    if ( npnt < mbas ) WARNING_("less points inserted for such interpolation order");
    //
    #ifdef CRUDE_SOLVE
    Blitz_DyadicReal A(npnt+mbas, npnt+mbas);
    A = 0.0;
    #else
    DenseMatrix<Number> A(npnt+mbas, npnt+mbas);
    #endif
    {
        int k = 1;
        for (int i=0; i<npnt; i++) {
            A(i, i) = this->_radial_func(0.0, coeff);
            for (int j=k; j<npnt; j++) {
                A(i, j) = this->_radial_func(Point(pnt[i]-pnt[j]).norm(), coeff);
                A(j, i) = A(i, j);
            }
            ++k;
            //
            Blitz_VectorReal pv(mbas);
            this->_polynomial(pnt[i], pv);
            for (int j=0; j<mbas; j++) {
                A(i, npnt+j) = pv(j);
                A(npnt+j, i) = A(i, npnt+j);
            }
        }
    }
    //
    #ifdef CRUDE_SOLVE
    Blitz_DyadicReal A_inv;
    inverse(A, A_inv);
    #endif
    //
    for (unsigned int isoln=0; isoln<SP3D; isoln++) {
        #ifdef CRUDE_SOLVE
        Blitz_VectorReal b(npnt+mbas);
        #else
        DenseVector<Number> b(npnt+mbas);
        #endif
        // set right-hand-side vector
        for (int i=0; i<npnt; i++)
            b(i) = soln[isoln][i];
        for (int j=0; j<mbas; j++)
            b(j) = 0.0;
        #ifdef CRUDE_SOLVE
        Blitz_VectorReal x(npnt+mbas);
        x = sum(A_inv(blitz::tensor::i,blitz::tensor::j)*b(blitz::tensor::j), blitz::tensor::j);
        // copy solution into vector
        this->_alpha[isoln].resize(npnt+mbas);
        for (int l=0; l<npnt+mbas; l++)
            this->_alpha[isoln](l) = x(l);
        #else
        // solve system via LU decomposition
        A.lu_solve(b, this->_alpha[isoln]);
        #endif
    }
    // all are initialized
    this->_is_init = true;
    this->_coeff = coeff;
    this->_pnt = &pnt;
}
//=================================================================================================
void RBFmap3D::calc (const Point& pnt, Real soln[]) const {
    libmesh_assert( this->_is_init );
    const int npnt = this->_pnt->size(),
              mbas = this->_mbasis;
    //
    Blitz_VectorReal rv(npnt);
    for (int i=0; i<npnt; i++)
        rv(i) = this->_radial_func(Point(pnt-(*this->_pnt)[i]).norm(), this->_coeff);
    Blitz_VectorReal pv(mbas);
    this->_polynomial(pnt, pv);
    //
    for (unsigned int isoln=0; isoln<SP3D; isoln++) {
        soln[isoln] = 0.0;
        // add the "a" coefficents contribution
        for (int i=0; i<npnt; i++)
            soln[isoln] += this->_alpha[isoln](i) * rv(i);
        for (int j=0; j<mbas; j++)
            soln[isoln] += this->_alpha[isoln](npnt+j) * pv(j);
    }
}
//=================================================================================================
#undef  CRUDE_SOLVE
//=================================================================================================
