/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "libmesh/mesh.h"
#include "libmesh/equation_systems.h"
#include "libmesh/linear_implicit_system.h"
#include "libmesh/petsc_linear_solver.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/fe.h"
#include "libmesh/quadrature_gauss.h"
#include "libmesh/dof_map.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/dense_matrix.h"
#include "libmesh/dense_vector.h"
#include "libmesh/dense_submatrix.h"
#include "libmesh/dense_subvector.h"
#include "libmesh/numeric_vector.h"
#include "libmesh/boundary_info.h"
#include "feb3/reaction_diffusion_solver_Xd.h"
#include "feb3/fem-context.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
//=================================================================================================
void ReactionDiffusionSolverXD::_assemble_all () {
    this->_log.push("_assemble_all");

    EquationSystems& es = this->_dmn.get_equation_systems();
    const MeshBase& msh = es.get_mesh();
    LinearImplicitSystem& v_sys = es.get_system<LinearImplicitSystem>(this->_label+"::solution");
    // const System&       v_P_sys = es.get_system<System>(this->_label+"::previous_solution");
    System&              dc_sys = es.get_system<System>(this->_label+"::discontinuous_solution");
    //
    const FEM_Context v_ctx(v_sys);
    const FEM_Context dc_ctx(dc_sys);
    const MeshBase::const_element_iterator begin_el = msh.active_local_elements_begin();
    const MeshBase::const_element_iterator end_el = msh.active_local_elements_end();
    //
    libmesh_assert( SP3D == msh.mesh_dimension() );
    libmesh_assert( this->par_list.get<unsigned int>("dimension") == v_sys.n_vars() );
    libmesh_assert( this->par_list.get<unsigned int>("discontinuous variables") == dc_sys.n_vars() );
    //
    FEType fe_v_type = v_sys.variable_type(0);
    FEType fe_dc_type = dc_sys.variable_type(0);
    //
    const bool membrane_present = this->par_list.get<bool>("membrane elements present");
    const unsigned int ninvars_NCS = es.parameters.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = es.parameters.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = es.parameters.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = es.parameters.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    const Real current_time = es.parameters.get<Real>("current time"),
               time_step    = es.parameters.get<Real>("reaction-diffusion solver: time step");
    const Real alpha_parameter = 0.75,
               C0_ = time_step * alpha_parameter,
               C1_ = time_step * (1.0-alpha_parameter);
    libmesh_assert( nvars_RDS == v_sys.n_vars() );
    libmesh_assert( ndcvars_RDS == dc_sys.n_vars() );
    ASSERT_(!membrane_present, "fatal internal error occurred");
    //
    // zero-out the rhs-vector and matrix of the system
    v_sys.rhs->zero();
    v_sys.matrix->zero();
    //
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        //
        QGauss quad_volu(SP3D, es.parameters.get<Order>("Gauss quadrature order"));
        UniquePtr<FEBase> fe_v_volu = FEBase::build(SP3D, fe_v_type);
        fe_v_volu->attach_quadrature_rule(&quad_volu);
        UniquePtr<FEBase> fe_dc_volu = FEBase::build(SP3D, fe_dc_type);
        fe_dc_volu->attach_quadrature_rule(&quad_volu);
        //
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars_RDS];
        v_ctx.get_dof_indices(elem, dof_indices, dof_indices_var);
        std::vector<dof_id_type> dof_indices__disc;
        std::vector<dof_id_type>* dof_indices_var__disc = new std::vector<dof_id_type> [ndcvars_RDS];
        dc_ctx.get_dof_indices(elem, dof_indices__disc, dof_indices_var__disc);
        //
        std::vector<unsigned int> var_dofs;
        calc_var_dofs(nvars_RDS, dof_indices_var, var_dofs);
        std::vector<unsigned int> var_dofs__disc;
        calc_var_dofs(ndcvars_RDS, dof_indices_var__disc, var_dofs__disc);
        //
        DenseMatrix<Number> Me(dof_indices.size(), dof_indices.size());
        std::vector< DenseSubMatrix<Number> > Mesub(nvars_RDS*nvars_RDS, Me);
        DenseVector<Number> Ye(dof_indices.size());
        std::vector< DenseSubVector<Number> > Yesub(nvars_RDS, Ye);
        std::vector<unsigned int> offset;
        calc_offset(nvars_RDS, var_dofs, offset);
        for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
            for (unsigned int jvar=0; jvar<nvars_RDS; jvar++) {
                Mesub[ivar*nvars_RDS+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
            }
            Yesub[ivar].reposition(offset[ivar], var_dofs[ivar]);
        }
        // get access to the material region
        libmesh_assert( this->_dmn.regn.end() != this->_dmn.regn.find(elem->subdomain_id()) );
        const RegionInfo* ri_ptr = &(this->_dmn.regn.find(elem->subdomain_id())->second);
        // initialize and get access to the FE stuff
        const std::vector< Real >&                       JxW = fe_v_volu->get_JxW();
        const std::vector< std::vector<Real> >&          Phi = fe_v_volu->get_phi();
        const std::vector< std::vector<RealGradient> >& dPhi_dX = fe_v_volu->get_dphi();
        fe_v_volu->reinit(elem);
        //const std::vector< std::vector<Real> >&        Theta = fe_dc_volu->get_phi();
        //fe_dc_volu->reinit(elem);
        // get access to the element quadrature points data
        libmesh_assert( this->_dmn.elem_qi_local.end() != this->_dmn.elem_qi_local.find(elem->id()) );
        const QuadratureInfo* qpi_ptr = &(this->_dmn.elem_qi_local.find(elem->id())->second);
        // loop over volume integration quadrature points
        for (unsigned int qp=0; qp<quad_volu.n_points(); qp++) {
            //
            FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
            fld.current_time = current_time;
            fld.time_step = time_step;
            fld.refence_volume = elem->volume();
            for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                fld.biochem[ivar]      = qpi_ptr->biochem[qp][ivar];
                fld.grad_biochem[ivar] = qpi_ptr->grad_biochem[qp][ivar];
                fld.biochem_rate[ivar] = qpi_ptr->biochem_rate[qp][ivar];
            }
            for (unsigned int ivar=nvars_RDS; ivar<nvars_RDS+ndcvars_RDS; ivar++) {
                fld.biochem[ivar] = qpi_ptr->biochem[qp][ivar];
            }
            for (unsigned int ivar=0; ivar<ninvars_NCS; ivar++) {
                fld.biomech[ivar] = qpi_ptr->biomech[qp][ivar];
            }
            for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                fld.auxiliary[ivar]      = qpi_ptr->auxiliary[qp][ivar];
                fld.grad_auxiliary[ivar] = qpi_ptr->grad_auxiliary[qp][ivar];
                fld.auxiliary_rate[ivar] = qpi_ptr->auxiliary_rate[qp][ivar];
            }
            fld.deform_gradient = qpi_ptr->deform_grad[qp];
            for (unsigned int ispdm=0; ispdm<SP3D; ispdm++) {
                fld.fibre_ref[ispdm] = qpi_ptr->fibre_ref[ispdm][qp];
                fld.fibre_curr[ispdm] = unit(fld.deform_gradient * qpi_ptr->fibre_ref[ispdm][qp]);
            }
            fld.stress = qpi_ptr->stress[qp];
            fld.strain = qpi_ptr->strain[qp];
            fld.pressure = qpi_ptr->pressure[qp];
            fld.fluid_velocity = qpi_ptr->fluid_velocity[qp];
            #ifdef BIODYNAMO_ENABLED
            // fld.bdm_ptr = & qpi_ptr->bdm_d[qp];
            fld.bdm_ptr = &(this->_dmn.elem_rve_local.find(elem->id())->second[qp]);
            #endif
            //
            Blitz_VectorReal      Fv__Crd;
            Blitz_VectorGradient  Fv__Grad_Crd;
            Blitz_DyadicReal      Km__Sys;
            Blitz_VectorPoint     Fv__Velc;
            Blitz_DyadicReal      Km__Diff;
            Blitz_DyadicGradient  Km__Taxs;
            //
            ri_ptr->calc_biochem(fld, Fv__Crd, Fv__Grad_Crd, Km__Sys, Fv__Velc, Km__Diff, Km__Taxs);
            //
            for (unsigned int i=0; i<var_dofs[0]; i++) {
                //
                if ( C1_ )
                for (int a=0; a<nvars_RDS; a++)
                    Yesub[a](i) += JxW[qp] * (Phi[i][qp]*Fv__Crd(a)) * C1_;
                if ( C1_ )
                for (int a=0; a<nvars_RDS; a++)
                    Yesub[a](i) -= JxW[qp] * (dPhi_dX[i][qp]*Fv__Grad_Crd(a)) * C1_;
                // old capacity contribution
                for (int a=0; a<nvars_RDS; a++)
                    Yesub[a](i) += JxW[qp] * (Phi[i][qp]*fld.biochem[a]);
                //
                for (unsigned int j=0; j<var_dofs[0]; j++) {
                    // capacity contribution
                    for (int a=0; a<nvars_RDS; a++)
                        Mesub[a*nvars_RDS+a](i,j) += JxW[qp] * (Phi[i][qp]*Phi[j][qp]);
                    //
                    for (int a=0; a<nvars_RDS; a++)
                        for (int b=0; b<nvars_RDS; b++)
                            Mesub[a*nvars_RDS+b](i,j) -= JxW[qp] * (Phi[i][qp]*Km__Sys(a,b)*Phi[j][qp]) * C0_;
                    // convection contribution
                    for (int a=0; a<nvars_RDS; a++)
                        Mesub[a*nvars_RDS+a](i,j) += JxW[qp] * (Phi[i][qp]*Fv__Velc(a)*dPhi_dX[j][qp]) * C0_;
                    // diffusion contribution
                    for (int a=0; a<nvars_RDS; a++)
                        for (int b=0; b<nvars_RDS; b++)
                            Mesub[a*nvars_RDS+b](i,j) += JxW[qp] * (dPhi_dX[i][qp]*Km__Diff(a,b)*dPhi_dX[j][qp]) * C0_;
                    // ?-taxis contribution
                    for (int a=0; a<nvars_RDS; a++)
                        for (int b=0; b<nvars_RDS; b++)
                            Mesub[a*nvars_RDS+b](i,j) += JxW[qp] * (dPhi_dX[i][qp]*Km__Taxs(a,b)*Phi[j][qp]) * C0_;
                }
            }
            // end of volume integration quadrature loop...
        }
        // update the rhs-vector and matrix of the system
        v_sys.rhs->add_vector(Ye, dof_indices);
        v_sys.matrix->add_matrix(Me, dof_indices);
        // free memo
        delete [] dof_indices_var;
        delete [] dof_indices_var__disc;
        // end of active local elements loop...
    }
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    // close the rhs-vector and matrix of the system
    v_sys.rhs->close();
    v_sys.matrix->close();

    this->_log.pop("_assemble_all");
}
//=================================================================================================
