/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/reaction_diffusion_solver_Xd.h"
#include "libmesh/linear_implicit_system.h"
#include "libmesh/petsc_linear_solver.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/gmsh_io.h"
#include "libmesh/fe.h"
#include "libmesh/quadrature_gauss.h"
#include "libmesh/dof_map.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/dense_matrix.h"
#include "libmesh/dense_vector.h"
#include "libmesh/dense_submatrix.h"
#include "libmesh/dense_subvector.h"
#include "libmesh/numeric_vector.h"
#include "feb3/fem-context.h"
#include "feb3/paraview_io.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
extern // list of all systems used in the analysis
std::vector< std::pair<EquationSystems*, std::string> > systems_ls;
//=================================================================================================
#include "feb3/domain_data_inl.h"
#include "feb3/solution_io_inl.h"
//=================================================================================================
ReactionDiffusionSolverXD::ReactionDiffusionSolverXD (DomainData& dmn) :
    _log("ReactionDiffusionSolverXD"), _dmn(dmn) {
    // set the default label
    this->_label = "ReactionDiffusionSolverXD";
    // set parameters to a default value
    this->par_list.set<unsigned int>("dimension") = 0;
    this->par_list.set<Real>("time step")    = 1.0;
    this->par_list.set<int>("initial time") = 0;
    this->par_list.set<bool>("enforce to initialize solution") = true;
    this->par_list.set<bool>("explicit solve") = true;
    this->par_list.set<bool>("enforce solution") = false;
}
//=================================================================================================
void ReactionDiffusionSolverXD::init () {
    this->_log.push("init");

    EquationSystems& es = this->_dmn.get_equation_systems();
    Parameters& par = es.parameters;
    // set some system parameters
    this->_copy_params();
    // dimension of the reaction-diffusion solver
    const unsigned int nvars = this->par_list.get<unsigned int>("dimension");
    ASSERT_(nvars>0, "dimension of reaction-diffusion solver cannot be zero");
    // Lagrange basis order
    const Order lbasis = par.get<Order>("Lagrange basis order");
    //
    char var_name[buffer_len];
    int& hvar = par.set<int>("number of variables to hide in Paraview");
    // add the basic system
    do {
        LinearImplicitSystem& sys = es.add_system<LinearImplicitSystem>(this->_label+"::solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__v%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis, LAGRANGE);
        }
        sys.assemble_before_solve = false;
        if ( this->par_list.get<bool>("explicit solve") ) {
            sys.add_vector("rhs");
            sys.add_vector("reciprocal diag-mat");
        }
    } while ( false );
    // add another system
    do {
        System& sys = es.add_system<System>(this->_label+"::previous_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__v%d_PREV", this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis, LAGRANGE);
        }
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    } while ( false );
    // add another system
    do {
        System& sys = es.add_system<System>(this->_label+"::solution_time_derivative");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__Dv%d_Dt", this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis, LAGRANGE);
        }
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    } while ( false );
    // add another (auxiliary) system
    do {
        System& sys = es.add_system<System>(this->_label+"::discontinuous_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        const unsigned int nvars = this->par_list.get<unsigned int>("discontinuous variables");
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__C%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name, CONSTANT, MONOMIAL);
        }
    } while ( false );
    // add another (auxiliary) system
    do {
        System& sys = es.add_system<System>(this->_label+"::auxiliary_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        const unsigned int nvars = this->par_list.get<unsigned int>("auxiliary variables");
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__a%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis, LAGRANGE);
        }
    } while ( false );
    // add another (auxiliary) system
    do {
        System& sys = es.add_system<System>(this->_label+"::auxiliary_solution_time_derivative");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        const unsigned int nvars = this->par_list.get<unsigned int>("auxiliary variables");
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            sprintf(var_name, "%s__Da%d_Dt", this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis, LAGRANGE);
        }
        // hide the variables
        for (unsigned int ivar=0; ivar<sys.n_vars(); ivar++) {
            sprintf(var_name, "Paraview variable #%d to hide", hvar);
            par.set<std::string>(var_name) = sys.variable_name(ivar);
            hvar++;
        }
    } while ( false );

    this->_log.pop("init");
}
//=================================================================================================
void ReactionDiffusionSolverXD::run () {
    this->_log.push("run");

    EquationSystems& es = this->_dmn.get_equation_systems();
    //
    ParaviewIO pvtu(es.get_mesh());
    for (int l=0; l<es.parameters.get<int>("number of variables to hide in Paraview"); l++) {
        char var_name[buffer_len];
        sprintf(var_name, "Paraview variable #%d to hide", l);
        pvtu.hide_variable(es.parameters.get<std::string>(var_name));
    }
    //
    char cstr[buffer_len];
    const std::string results_dir  = es.parameters.get<std::string>("results directory"),
                      output_fname = es.parameters.get<std::string>("output file name"),
                      sim_title    = es.parameters.get<std::string>("simulation title");
    Real& current_time = es.parameters.set<Real>("current time");
    const int ini_time  = this->par_list.get<int>("initial time"),
              max_times = this->par_list.get<int>("maximum time steps"),
              soln_step = this->par_list.get<int>("solution increment"),
              rstr_step = this->par_list.get<int>("restart increment");
    const Real time_step = this->par_list.get<Real>("time step");
    //
    // set the initial conditions and initialise some member objects
    if ( this->par_list.get<bool>("enforce to initialize solution") ) {
        flog.print(" >> Enforce solution initialization\n", false);
        this->_set_initial();
    // ...else check if to read the restart file(s)
    } else if ( es.parameters.get<unsigned int>("input restart from file") ) {
        flog.print(" >> Load restart file(s)\n", false);
        this->load_restart();
    }
    //
    // update the corresponding quadrature information data for this solver
    this->_set_auxiliary();
    this->_set_quad_info();
    // ...and, finally, initialize the data associated with BioDynaMo (if necessary)
    this->_dmn.init_bdm();
    //
    // write a PVD file to store results
    if ( libMesh::global_processor_id() == 0 ) {
        flog.print(" >> Generate Paraview PVD-file\n", false);
        sprintf( cstr, "%s/%s-%s.pvd",
                 results_dir.c_str(), output_fname.c_str(), sim_title.c_str() );
        std::ofstream fpvd(cstr);
        fpvd.setf(std::ios_base::scientific|std::ios_base::unitbuf);
        //
        fpvd << "<?xml version=\"1.0\"?>" << std::endl;
        fpvd << "<VTKFile type=\"Collection\" version=\"0.1\" byte_order=\"LittleEndian\">" << std::endl;
        fpvd << "<Collection>" << std::endl;
        for (int time=ini_time; time<=max_times; time++) {
            if ( time == ini_time || time%soln_step == 0 || time == max_times ) {
                const Real vtk_time = time*time_step;
                sprintf( cstr, "paraview/%s-%s-%d.vtu",
                         output_fname.c_str(), sim_title.c_str(), time );
                fpvd << "<DataSet timestep=\"" << vtk_time << "\" group=\"\" part=\"0\" file=\"" << cstr << "\"/>" << std::endl;
            }
            // end of time steps loop...
        }
        fpvd << "</Collection>" << std::endl;
        fpvd << "</VTKFile>" << std::endl;
    }
    #ifdef BIODYNAMO_ENABLED
    {
        sprintf( cstr, "%s/%s-%s.BDM.pvd",
                 results_dir.c_str(), output_fname.c_str(), sim_title.c_str() );
        std::ofstream fpvd(cstr);
        fpvd.setf(std::ios_base::scientific|std::ios_base::unitbuf);
        //
        fpvd << "<?xml version=\"1.0\"?>" << std::endl;
        fpvd << "<VTKFile type=\"Collection\" version=\"0.1\" byte_order=\"LittleEndian\">" << std::endl;
        fpvd << "<Collection>" << std::endl;
        for (int time=ini_time; time<=max_times; time++) {
            if ( time == ini_time || time%soln_step == 0 || time == max_times ) {
                const Real vtk_time = time*time_step;
                sprintf( cstr, "paraview/%s-%s-%d.BDM.pvtu",
                         output_fname.c_str(), sim_title.c_str(), time );
                fpvd << "<DataSet timestep=\"" << vtk_time << "\" group=\"\" part=\"0\" file=\"" << cstr << "\"/>" << std::endl;
            }
            // end of time steps loop...
        }
        fpvd << "</Collection>" << std::endl;
        fpvd << "</VTKFile>" << std::endl;

    }
    #endif
    //
    // write initial results
    flog.print("  > Save solution\n", false);
    sprintf( cstr, "%s/paraview/%s-%s-%d.vtu",
             results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), ini_time );
    pvtu.write_equation_systems(cstr, es);
    // ...also write the BioDynaMo results
    sprintf( cstr, "%s-%s-%d.BDM",
             output_fname.c_str(), sim_title.c_str(), ini_time );
    this->_dmn.write_bdm(cstr);
    //
    // evaluate system lumped matrix (if necessary)
    if ( this->par_list.get<bool>("explicit solve") ) {
        flog.print(" >> Evaluate system lumped matrix\n", false);
        this->_diag_matrix();
    }
    //
    // run for all time steps
    for (int itime=ini_time; itime<max_times; itime++) {
        this->par_list.set<int>("current (pseudo-)time") = itime;
        current_time += time_step;
        //
        // solve the (reaction-diffusion) system
        sprintf( cstr, " >> Time %d out of %d (%.10f)                    \r",
                 itime+1, max_times, current_time );
        flog.print(cstr, false);
        // first operate the BioDynaMo solver (if necessary)
        this->_dmn.run_bdm();
        // ...and now the reaction-diffusion solver
        this->_solve();
        //
        // check if to write results
        if ( (itime+1)%soln_step == 0 || (itime+1) == max_times ) {
            sprintf( cstr, "%s/paraview/%s-%s-%d.vtu",
                     results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), itime+1 );
            pvtu.write_equation_systems(cstr, es);
            // ...also write the BioDynaMo results
            sprintf( cstr, "%s-%s-%d.BDM",
                     output_fname.c_str(), sim_title.c_str(), itime+1 );
            this->_dmn.write_bdm(cstr);
        }
        //
        // check if to write restart file
        if ( (itime+1)%rstr_step == 0 || (itime+1) == max_times ) {
            sprintf( cstr, "%s/restart/%s-%s-%d.res-%d",
                     results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), itime+1, libMesh::global_processor_id() );
            std::ofstream fout(cstr);
            this->save_restart(fout);
        }
        // end of time steps loop...
    }
    // end of simulation...
    clear_line_(std::cout);
    flog.print(" >> Simulation terminated\n", false);

    this->_log.pop("run");
}
//=================================================================================================
//
//
//
//=================================================================================================
#include "feb3/solution_io_inl.h"
//=================================================================================================
void ReactionDiffusionSolverXD::_copy_params () {
    this->_log.push("_copy_params");

    EquationSystems& es = this->_dmn.get_equation_systems();
    // set the label for the solver and the simulation title (if necessary)
    this->_label = es.parameters.get<std::string>("reaction-diffusion solver: label");
    if ( "_" == es.parameters.get<std::string>("simulation title") )
        es.parameters.set<std::string>("simulation title") = this->_label;
    //
    if ( es.parameters.get<unsigned int>("input restart from file") )
        this->par_list.set<bool>("enforce to initialize solution") = false;
    //
    this->par_list.set<unsigned int>("dimension") = es.parameters.get<unsigned int>("reaction-diffusion solver: dimension");
    this->par_list.set<bool>("membrane elements present") = es.parameters.get<bool>("membrane elements present");
    this->par_list.set<bool>("explicit solve") = es.parameters.get<bool>("reaction-diffusion solver: explicit");
    this->par_list.set<bool>("enforce solution") = es.parameters.get<bool>("reaction-diffusion solver: enforce solution");
    //
    this->par_list.set<Real>("time step")         = es.parameters.get<Real>("reaction-diffusion solver: time step");
    this->par_list.set<int>("maximum time steps") = es.parameters.get<int>("reaction-diffusion solver: maximum time steps");
    this->par_list.set<int>("solution increment") = es.parameters.get<int>("reaction-diffusion solver: solution increment");
    this->par_list.set<int>("restart increment") = es.parameters.get<int>("reaction-diffusion solver: restart increment");
    this->par_list.set<int>("initial time") = es.parameters.get<int>("reaction-diffusion solver: initial time");
    //
    for (unsigned int ivar=0; ivar<this->par_list.get<unsigned int>("dimension"); ivar++) {
        char field_name[buffer_len],
             field_name_es[buffer_len];
        sprintf(field_name, "healthy_domain_initial_field_%d_function", ivar);
        sprintf(field_name_es, "reaction-diffusion solver: healthy domain initial field %d function", ivar);
        this->par_list.set<int>(field_name) = es.parameters.get<int>(field_name_es);
        sprintf(field_name, "healthy_domain_initial_field_%d", ivar);
        sprintf(field_name_es, "reaction-diffusion solver: healthy domain initial field %d", ivar);
        this->par_list.set<Real>(field_name) = es.parameters.get<Real>(field_name_es);
        sprintf(field_name, "healthy_domain_initial_field_%d_deviation", ivar);
        sprintf(field_name_es, "reaction-diffusion solver: healthy domain initial field %d deviation", ivar);
        this->par_list.set<Real>(field_name) = es.parameters.get<Real>(field_name_es);
        sprintf(field_name, "damaged_domain_initial_field_%d_function", ivar);
        sprintf(field_name_es, "reaction-diffusion solver: damaged domain initial field %d function", ivar);
        this->par_list.set<int>(field_name) = es.parameters.get<int>(field_name_es);
        sprintf(field_name, "damaged_domain_initial_field_%d", ivar);
        sprintf(field_name_es, "reaction-diffusion solver: damaged domain initial field %d", ivar);
        this->par_list.set<Real>(field_name) = es.parameters.get<Real>(field_name_es);
        sprintf(field_name, "damaged_domain_initial_field_%d_deviation", ivar);
        sprintf(field_name_es, "reaction-diffusion solver: damaged domain initial field %d deviation", ivar);
        this->par_list.set<Real>(field_name) = es.parameters.get<Real>(field_name_es);
        sprintf(field_name, "healthy_to_damaged_domain_initial_field_%d_interface", ivar);
        sprintf(field_name_es, "reaction-diffusion solver: healthy/damaged domain initial field %d at interface", ivar);
        this->par_list.set<bool>(field_name) = es.parameters.get<bool>(field_name_es);
        sprintf(field_name, "constrained_field_%d_type", ivar);
        sprintf(field_name_es, "reaction-diffusion solver: constrained field %d type", ivar);
        this->par_list.set<int>(field_name) = es.parameters.get<int>(field_name_es);
    }
    //
    this->par_list.set<unsigned int>("discontinuous variables") =
        es.parameters.get<unsigned int>("reaction-diffusion solver: discontinuous variables");
    //
    for (unsigned int ivar=0; ivar<this->par_list.get<unsigned int>("discontinuous variables"); ivar++) {
        char field_name[buffer_len],
             field_name_es[buffer_len];
        sprintf(field_name, "discontinuous_initial_field_%d", ivar);
        sprintf(field_name_es, "reaction-diffusion solver: discontinuous initial field %d", ivar);
        this->par_list.set<Real>(field_name) = es.parameters.get<Real>(field_name_es);
        sprintf(field_name, "discontinuous_initial_field_%d_deviation", ivar);
        sprintf(field_name_es, "reaction-diffusion solver: discontinuous initial field %d deviation", ivar);
        this->par_list.set<Real>(field_name) = es.parameters.get<Real>(field_name_es);
    }
    //
    this->par_list.set<unsigned int>("auxiliary variables") =
        es.parameters.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    //
    // sanity check
    ASSERT_(this->par_list.get<unsigned int>("discontinuous variables"), "fatal internal error occurred");

    this->_log.pop("_copy_params");
}
//=================================================================================================
void ReactionDiffusionSolverXD::_set_initial () {
    this->_log.push("_set_initial");

    EquationSystems& es = this->_dmn.get_equation_systems();
    System& v_sys   = es.get_system<System>(this->_label+"::solution");
    System& DvDt_sys = es.get_system<System>(this->_label+"::solution_time_derivative");
    System& a_sys    = es.get_system<System>(this->_label+"::auxiliary_solution");
    System& DaDt_sys = es.get_system<System>(this->_label+"::auxiliary_solution_time_derivative");
    const MeshBase& msh = es.get_mesh();
    //
    const MeshBase::const_element_iterator begin_el = msh.active_local_elements_begin();
    const MeshBase::const_element_iterator end_el = msh.active_local_elements_end();
    const MeshBase::const_element_iterator begin_el_all = msh.active_elements_begin();
    const MeshBase::const_element_iterator end_el_all = msh.active_elements_end();
    const MeshBase::const_node_iterator    begin_nd = msh.nodes_begin();
    const MeshBase::const_node_iterator    end_nd = msh.nodes_end();
    //
    // zero-out the solution by default
    v_sys.solution->zero();
    DvDt_sys.solution->zero();
    // zero-out the auxiliary solution by default
    if ( this->par_list.get<unsigned int>("auxiliary variables") ) {
        a_sys.solution->zero();
        DaDt_sys.solution->zero();
    }
    //
    // get the initial values for the corresponding fields
    const unsigned int nvars = this->par_list.get<unsigned int>("dimension");
    std::vector<Real> initial_field_value__H(nvars),
                      initial_field_value__D(nvars);
    std::vector<Real> initial_field_devia__H(nvars),
                      initial_field_devia__D(nvars);
    std::vector<bool> initial_field_H_D_interface(nvars);
    for (unsigned int ivar=0; ivar<nvars; ivar++) {
        char field_name[buffer_len];
        sprintf(field_name, "healthy_domain_initial_field_%d", ivar);
        initial_field_value__H[ivar] = this->par_list.get<Real>(field_name);
        sprintf(field_name, "damaged_domain_initial_field_%d", ivar);
        initial_field_value__D[ivar] = this->par_list.get<Real>(field_name);
        sprintf(field_name, "healthy_domain_initial_field_%d_deviation", ivar);
        initial_field_devia__H[ivar] = this->par_list.get<Real>(field_name);
        sprintf(field_name, "damaged_domain_initial_field_%d_deviation", ivar);
        initial_field_devia__D[ivar] = this->par_list.get<Real>(field_name);
        sprintf(field_name, "healthy_to_damaged_domain_initial_field_%d_interface", ivar);
        initial_field_H_D_interface[ivar] = this->par_list.get<bool>(field_name);
    }
    // loop over local nodes (in damaged regions)
    for ( std::set<unsigned int>::const_iterator
          citer_nd=this->_dmn.damaged_node.begin(); citer_nd!=this->_dmn.damaged_node.end(); citer_nd++ )
    {
        const Node* node = msh.node_ptr(*citer_nd);
        //
        libmesh_assert( nvars == node->n_vars(v_sys.number()) );
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            const dof_id_type idof = node->dof_number(v_sys.number(), ivar, 0);
            libmesh_assert( node->n_comp(v_sys.number(), ivar) == 1 );
            // set the corresponding initial value
            if ( fabs(initial_field_devia__D[ivar]) < 1.0e-12 ) {
                v_sys.solution->set(idof, initial_field_value__D[ivar]);
            } else {
                char field_name[buffer_len];
                sprintf(field_name, "damaged_domain_initial_field_%d_function", ivar);
                if        ( -2 == this->par_list.get<int>(field_name) ) {
                    v_sys.solution->set(idof, uniform_distro(initial_field_value__D[ivar]-initial_field_devia__D[ivar],
                                                             initial_field_value__D[ivar]+initial_field_devia__D[ivar]));
                } else if ( -1 == this->par_list.get<int>(field_name) ) {
                    v_sys.solution->set(idof, normal_distro(initial_field_value__D[ivar], initial_field_devia__D[ivar]));
                } else if (  0 == this->par_list.get<int>(field_name) ) {
                    v_sys.solution->set(idof, initial_field_value__D[ivar]);
                } else if ( +1 == this->par_list.get<int>(field_name) ) {
                    const Real A = initial_field_value__D[ivar]*normal_distro(1., 0.02),
                               R2 = pow2((*node)(0)) + pow2((*node)(1)) + pow2((*node)(2));
                    v_sys.solution->set(idof, A*exp(-R2/(2.0*pow2(initial_field_devia__D[ivar]))));
                } else libmesh_error();
            }
        }
        // end of local nodes loop...
    }
    // loop over local nodes (in the interface => healthy/damaged region)
    for ( std::set<unsigned int>::const_iterator
          citer_nd=this->_dmn.interface_node.begin(); citer_nd!=this->_dmn.interface_node.end(); citer_nd++ )
    {
        const Node* node = msh.node_ptr(*citer_nd);
        //
        libmesh_assert( nvars == node->n_vars(v_sys.number()) );
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            const dof_id_type idof = node->dof_number(v_sys.number(), ivar, 0);
            libmesh_assert( node->n_comp(v_sys.number(), ivar) == 1 );
            if ( initial_field_H_D_interface[ivar] ) {
                // set the corresponding initial value
                if ( fabs(initial_field_devia__H[ivar]) < 1.0e-12 ) {
                    v_sys.solution->set(idof, initial_field_value__H[ivar]);
                } else {
                    char field_name[buffer_len];
                    sprintf(field_name, "healthy_domain_initial_field_%d_function", ivar);
                    if        ( -2 == this->par_list.get<int>(field_name) ) {
                        v_sys.solution->set(idof, uniform_distro(initial_field_value__H[ivar]-initial_field_devia__H[ivar],
                                                                 initial_field_value__H[ivar]+initial_field_devia__H[ivar]));
                    } else if ( -1 == this->par_list.get<int>(field_name) ) {
                        v_sys.solution->set(idof, normal_distro(initial_field_value__H[ivar], initial_field_devia__H[ivar]));
                    } else if (  0 == this->par_list.get<int>(field_name) ) {
                        v_sys.solution->set(idof, initial_field_value__H[ivar]);
                    } else if ( +1 == this->par_list.get<int>(field_name) ) {
                        const Real A = initial_field_value__H[ivar]*normal_distro(1., 0.02),
                                   R2 = pow2((*node)(0)) + pow2((*node)(1)) + pow2((*node)(2));
                        v_sys.solution->set(idof, A*exp(-R2/(2.0*pow2(initial_field_devia__H[ivar]))));
                    } else libmesh_error();
                }
            } else {
                // set the corresponding initial value
                if ( fabs(initial_field_devia__D[ivar]) < 1.0e-12 ) {
                    v_sys.solution->set(idof, initial_field_value__D[ivar]);
                } else {
                    char field_name[buffer_len];
                    sprintf(field_name, "damaged_domain_initial_field_%d_function", ivar);
                    if        ( -2 == this->par_list.get<int>(field_name) ) {
                        v_sys.solution->set(idof, uniform_distro(initial_field_value__D[ivar]-initial_field_devia__D[ivar],
                                                                 initial_field_value__D[ivar]+initial_field_devia__D[ivar]));
                    } else if ( -1 == this->par_list.get<int>(field_name) ) {
                        v_sys.solution->set(idof, normal_distro(initial_field_value__D[ivar], initial_field_devia__D[ivar]));
                    } else if (  0 == this->par_list.get<int>(field_name) ) {
                        v_sys.solution->set(idof, initial_field_value__D[ivar]);
                    } else if ( +1 == this->par_list.get<int>(field_name) ) {
                        const Real A = initial_field_value__D[ivar]*normal_distro(1., 0.02),
                                   R2 = pow2((*node)(0)) + pow2((*node)(1)) + pow2((*node)(2));
                        v_sys.solution->set(idof, A*exp(-R2/(2.0*pow2(initial_field_devia__D[ivar]))));
                    } else libmesh_error();
                }
            }
            // end of variables loop..
        }
        // end of local nodes loop...
    }
    // loop over local nodes (in healthy regions)
    for ( std::set<unsigned int>::const_iterator
          citer_nd=this->_dmn.healthy_node.begin(); citer_nd!=this->_dmn.healthy_node.end(); citer_nd++ )
    {
        const Node* node = msh.node_ptr(*citer_nd);
        //
        libmesh_assert( nvars == node->n_vars(v_sys.number()) );
        for (unsigned int ivar=0; ivar<nvars; ivar++) {
            const dof_id_type idof = node->dof_number(v_sys.number(), ivar, 0);
            libmesh_assert( node->n_comp(v_sys.number(), ivar) == 1 );
            // set the corresponding initial value
            if ( fabs(initial_field_devia__H[ivar]) < 1.0e-12 ) {
                v_sys.solution->set(idof, initial_field_value__H[ivar]);
            } else {
                char field_name[buffer_len];
                sprintf(field_name, "healthy_domain_initial_field_%d_function", ivar);
                if        ( -2 == this->par_list.get<int>(field_name) ) {
                    v_sys.solution->set(idof, uniform_distro(initial_field_value__H[ivar]-initial_field_devia__H[ivar],
                                                             initial_field_value__H[ivar]+initial_field_devia__H[ivar]));
                } else if ( -1 == this->par_list.get<int>(field_name) ) {
                    v_sys.solution->set(idof, normal_distro(initial_field_value__H[ivar], initial_field_devia__H[ivar]));
                } else if (  0 == this->par_list.get<int>(field_name) ) {
                    v_sys.solution->set(idof, initial_field_value__H[ivar]);
                } else if ( +1 == this->par_list.get<int>(field_name) ) {
                    const Real A = initial_field_value__H[ivar]*normal_distro(1., 0.02),
                               R2 = pow2((*node)(0)) + pow2((*node)(1)) + pow2((*node)(2));
                    v_sys.solution->set(idof, A*exp(-R2/(2.0*pow2(initial_field_devia__H[ivar]))));
                } else libmesh_error();
            }
        }
        // end of local nodes loop...
    }
    // check if to set some components of the solution from a data file
    if ( "NA" != es.parameters.get<std::string>("reaction-diffusion solver: predefined initial nodal field (file)") )
    {
        std::string name;
        std::ifstream fin;
        //
        name = es.parameters.get<std::string>("results directory") + "/temporary/"
             + es.parameters.get<std::string>("output file name") + "-node2node.txt";
        fin.open(name);
        //
        std::vector< Dual<dof_id_type> > updated_node_ID_ls;
        for (MeshBase::const_node_iterator n_citer=begin_nd; n_citer!=end_nd; n_citer++) {
            dof_id_type node_old_id, node_new_id;
            fin >> node_old_id >> node_new_id;
            //
            updated_node_ID_ls.push_back( Dual<dof_id_type>(node_new_id, node_old_id) );
        }
        // now close the file stream
        fin.close();
        //
        name = es.parameters.get<std::string>("reaction-diffusion solver: predefined initial nodal field (file)");
        fin.open(name);
        //
        int nfields;
        fin >> nfields;
        std::vector<std::string> field_ls(nfields);
        for (int ifield=0; ifield<nfields; ifield++) {
            fin >> field_ls[ifield];
            ASSERT_(field_ls[ifield].length()==3, "fatal internal error occurred");
        }
        //
        for (unsigned int inode=0; inode<msh.n_nodes(); inode++) {
            //
            const Node* node = msh.node_ptr( updated_node_ID_ls[inode].get_first() );
            ASSERT_(node, "fatal internal error occurred");
            ASSERT_(updated_node_ID_ls[inode].get_second()==inode, "fatal internal error occurred");
            //
            for (int ifield=0; ifield<nfields; ifield++) {
                Real field_value = 0.0;
                fin >> field_value;
                //
                char var[3];
                sprintf(var, "%c%c", field_ls[ifield][1], field_ls[ifield][2]);
                const unsigned int ivar = atoi(var);
                //
                if       ( 'v' == field_ls[ifield][0] ) {
                    ASSERT_(ivar>=0 && ivar<v_sys.n_vars(), "fatal internal error occurred");
                    //
                    const dof_id_type idof = node->dof_number(v_sys.number(), ivar, 0);
                    libmesh_assert( node->n_comp(v_sys.number(), ivar) == 1 );
                    //
                    v_sys.solution->set(idof, field_value);
                    //
                } else if ( 'V' == field_ls[ifield][0] ) {
                    ASSERT_(ivar>=0 && ivar<DvDt_sys.n_vars(), "fatal internal error occurred");
                    //
                    const dof_id_type idof = node->dof_number(DvDt_sys.number(), ivar, 0);
                    libmesh_assert( node->n_comp(DvDt_sys.number(), ivar) == 1 );
                    //
                    DvDt_sys.solution->set(idof, field_value);
                    //
                } else if ( 'a' == field_ls[ifield][0] ) {
                    ASSERT_(this->par_list.get<unsigned int>("auxiliary variables"), "fatal internal error occurred");
                    ASSERT_(ivar>=0 && ivar<a_sys.n_vars(), "fatal internal error occurred");
                    //
                    const dof_id_type idof = node->dof_number(a_sys.number(), ivar, 0);
                    libmesh_assert( node->n_comp(a_sys.number(), ivar) == 1 );
                    //
                    a_sys.solution->set(idof, field_value);
                    //
                } else if ( 'A' == field_ls[ifield][0] ) {
                    ASSERT_(this->par_list.get<unsigned int>("auxiliary variables"), "fatal internal error occurred");
                    ASSERT_(ivar>=0 && ivar<DaDt_sys.n_vars(), "fatal internal error occurred");
                    //
                    const dof_id_type idof = node->dof_number(DaDt_sys.number(), ivar, 0);
                    libmesh_assert( node->n_comp(DaDt_sys.number(), ivar) == 1 );
                    //
                    DaDt_sys.solution->set(idof, field_value);
                    //
                } else libmesh_error();
            }
        }
        // now close the file stream
        fin.close();
    }
    // close solution vectors and update the systems
    v_sys.solution->close();
    v_sys.update();
    DvDt_sys.solution->close();
    DvDt_sys.update();
    // close solution vectors and update the systems
    if ( this->par_list.get<unsigned int>("auxiliary variables") ) {
        a_sys.solution->close();
        a_sys.update();
        DaDt_sys.solution->close();
        DaDt_sys.update();
    }
    //
    System& dc_sys = es.get_system<System>(this->_label+"::discontinuous_solution");
    // zero-out the discontinuous solution by default
    dc_sys.solution->zero();
    // get the initial values for the corresponding discontinuous fields
    const unsigned int ndcvars = this->par_list.get<unsigned int>("discontinuous variables");
    std::vector<Real> initial_field_value__disc(ndcvars);
    std::vector<Real> initial_field_devia__disc(ndcvars);
    for (unsigned int ivar=0; ivar<ndcvars; ivar++) {
        char field_name[buffer_len];
        sprintf(field_name, "discontinuous_initial_field_%d", ivar);
        initial_field_value__disc[ivar] = this->par_list.get<Real>(field_name);
        sprintf(field_name, "discontinuous_initial_field_%d_deviation", ivar);
        initial_field_devia__disc[ivar] = this->par_list.get<Real>(field_name);
    }
    // get a reference to the Explicit system (discontinuous) variables
    unsigned int discontinuous_var[ndcvars];
    for (unsigned int l=0; l<ndcvars; l++) {
        char var_name[buffer_len];
        sprintf(var_name, "%s__C%d", this->_label.c_str(), l);
        discontinuous_var[l] = dc_sys.variable_number(var_name);
    }
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        //
        // local degrees of freedom for each 'gradient' variable
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [ndcvars];
        // now set the values of discontinuous values for this element
        for (unsigned int l=0; l<ndcvars; l++) {
            const unsigned int ivar = l;
            dc_sys.get_dof_map().dof_indices(elem, dof_indices_var[l], discontinuous_var[ivar]);
            const unsigned int idof = dof_indices_var[l][0];
            if ( (dc_sys.solution->first_local_index() <= idof) && (idof < dc_sys.solution->last_local_index()) ) {
                // set the corresponding initial value
                if ( fabs(initial_field_devia__disc[ivar]) <= 1.0e-6 ) {
                    dc_sys.solution->set(idof, initial_field_value__disc[ivar]);
                } else {
                    dc_sys.solution->set(idof, normal_distro(initial_field_value__disc[ivar], initial_field_devia__disc[ivar]));
                }
            }
        }
        // free memo
        delete [] dof_indices_var;
        // end of active local elements loop...
    }
    // check if to set some components of the solution from a data file
    if ( "NA" != es.parameters.get<std::string>("reaction-diffusion solver: predefined initial elemental field (file)") )
    {
        std::string name;
        std::ifstream fin;
        //
        name = es.parameters.get<std::string>("results directory") + "/temporary/"
             + es.parameters.get<std::string>("output file name") + "-elem2elem.txt";
        fin.open(name);
        //
        std::vector< Dual<dof_id_type> > updated_elem_ID_ls;
        for (MeshBase::const_element_iterator e_citer=begin_el_all; e_citer!=end_el_all; e_citer++) {
            dof_id_type elem_old_id, elem_new_id;
            fin >> elem_old_id >> elem_new_id;
            //
            updated_elem_ID_ls.push_back( Dual<dof_id_type>(elem_new_id, elem_old_id) );
        }
        // now close the file stream
        fin.close();
        //
        name = es.parameters.get<std::string>("reaction-diffusion solver: predefined initial elemental field (file)");
        fin.open(name);
        //
        int nfields;
        fin >> nfields;
        std::vector<std::string> field_ls(nfields);
        for (int ifield=0; ifield<nfields; ifield++) {
            fin >> field_ls[ifield];
            ASSERT_(field_ls[ifield].length()==3, "fatal internal error occurred");
        }
        //
        for (unsigned int ielem=0; ielem<msh.n_elem(); ielem++) {
            //
            const Elem* elem = msh.elem_ptr( updated_elem_ID_ls[ielem].get_first() );
            ASSERT_(elem, "fatal internal error occurred");
            ASSERT_(updated_elem_ID_ls[ielem].get_second()==ielem, "fatal internal error occurred");
            //
            // local degrees of freedom for each 'gradient' variable
            std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [ndcvars];
            //
            for (int ifield=0; ifield<nfields; ifield++) {
                Real field_value = 0.0;
                fin >> field_value;
                //
                char var[3];
                sprintf(var, "%c%c", field_ls[ifield][1], field_ls[ifield][2]);
                const unsigned int ivar = atoi(var);
                //
                if ( 'C' == field_ls[ifield][0] ) {
                    ASSERT_(ivar>=0 && ivar<dc_sys.n_vars(), "fatal internal error occurred");
                    dc_sys.get_dof_map().dof_indices(elem, dof_indices_var[ivar], discontinuous_var[ivar]);
                    const unsigned int idof = dof_indices_var[ivar][0];
                    //
                    dc_sys.solution->set(idof, field_value);
                } else libmesh_error();
            }
            // free memo
            delete [] dof_indices_var;
        }
        // now close the file stream
        fin.close();
    }
    // close solution vector and update the system
    dc_sys.solution->close();
    dc_sys.update();
    //
    System& v_P_sys = es.get_system<System>(this->_label+"::previous_solution");
    // copy & close solution vector and update the system
    v_P_sys.solution->operator=(*v_sys.solution);
    v_P_sys.solution->close();
    v_P_sys.update();
    //
    // update the corresponding quadrature information data for this solver
    this->_set_auxiliary();
    this->_set_quad_info();

    this->_log.pop("_set_initial");
}
//=================================================================================================
void ReactionDiffusionSolverXD::_diag_matrix () {
    this->_log.push("_diag_matrix");

    EquationSystems& es = this->_dmn.get_equation_systems();
    const MeshBase& msh = es.get_mesh();
    LinearImplicitSystem& v_sys = es.get_system<LinearImplicitSystem>(this->_label+"::solution");
    //
    const FEM_Context v_ctx(v_sys);
    const MeshBase::const_element_iterator begin_el = msh.active_local_elements_begin();
    const MeshBase::const_element_iterator end_el = msh.active_local_elements_end();
    //
    ASSERT_(SP3D==msh.mesh_dimension(), "fatal internal error occurred");
    ASSERT_(this->par_list.get<unsigned int>("dimension")==v_sys.n_vars(), "fatal internal error occurred");
    std::vector<unsigned short int> var;
    v_ctx.get_var_list(var);
    //
    FEType fe_v_type = v_sys.variable_type(var[0]);
    //
    const Real time_step = this->par_list.get<Real>("time step");
    const unsigned int nvars_RDS = es.parameters.get<unsigned int>("reaction-diffusion solver: dimension");
    const bool membrane_present = false;
    //
    if ( this->par_list.get<bool>("membrane elements present") )
        WARNING_("reaction-diffusion solver works improperly with membrane elements");
    //
    // zero-out the system (capacity) matrix and its reciprocal diagonal form
    v_sys.matrix->zero();
    v_sys.get_vector("reciprocal diag-mat").zero();
    //
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        //
        QGauss quad_volu(SP3D, es.parameters.get<Order>("Gauss quadrature order"));
        UniquePtr<FEBase> fe_v_volu = FEBase::build(SP3D, fe_v_type);
        fe_v_volu->attach_quadrature_rule(&quad_volu);
        //
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type>[nvars_RDS];
        v_ctx.get_dof_indices(elem, dof_indices, dof_indices_var);
        //
        std::vector<unsigned int> var_dofs;
        calc_var_dofs(nvars_RDS, dof_indices_var, var_dofs);
        // capacity matrix
        DenseMatrix<Number> Me(dof_indices.size(), dof_indices.size());
        std::vector< DenseSubMatrix<Number> > Mesub(nvars_RDS*nvars_RDS, Me);
        std::vector<unsigned int> offset;
        calc_offset(nvars_RDS, var_dofs, offset);
        for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
            for (unsigned int jvar=0; jvar<nvars_RDS; jvar++) {
                Mesub[ivar*nvars_RDS+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
            }
        }
        // get access to the material region
        //libmesh_assert( this->_dmn.regn.end() != this->_dmn.regn.find(elem->subdomain_id()) );
        //const RegionInfo* ri_ptr = &(this->_dmn.regn.find(elem->subdomain_id())->second);
        // initialize and get access to the FE stuff
        const std::vector< Real >&              JxW = fe_v_volu->get_JxW();
        const std::vector< std::vector<Real> >& Phi = fe_v_volu->get_phi();
        fe_v_volu->reinit(elem);
        // get access to the element quadrature points data
        //libmesh_assert( this->_dmn.elem_qi_local.end() != this->_dmn.elem_qi_local.find(elem->id()) );
        //const QuadratureInfo* qpi_ptr = &(this->_dmn.elem_qi_local.find(elem->id())->second);
        // loop over volume integration quadrature points
        for (unsigned int qp=0; qp<quad_volu.n_points(); qp++) {
            for (unsigned int i=0; i<var_dofs[0]; i++) {
                for (unsigned int j=0; j<var_dofs[0]; j++) {
                    for (int a=0; a<nvars_RDS; a++)
                        Mesub[a*nvars_RDS+a](i,j) += JxW[qp] * (Phi[i][qp]*Phi[j][qp])/time_step;
                }
            }
            // end of volume integration quadrature loop...
        }
        // row-sum method for lumping
        const unsigned int ndofs = dof_indices.size();
        for (unsigned int a=0; a<ndofs; a++) {
            Real Me_rs = 0.0;
            for (unsigned int b=0; b<ndofs; b++) {
                Me_rs += Me(a,b);
                Me(a,b) = 0.0;
            }
            Me(a,a) = Me_rs;
        }
        // update system matrix
        v_sys.matrix->add_matrix(Me, dof_indices);
        // free memo
        delete [] dof_indices_var;
        // end of active local elements loop...
    }
    //
    // loop over all active local finite elements
    if ( membrane_present )
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        // insert membrane contribution
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            // check if element has this side on boundary and get the boundary ID for side of element
            if ( elem->neighbor_ptr(s) != NULL ) continue;
            // create a side element
            UniquePtr<const Elem> side = elem->build_side_ptr(s);
            //
            QGauss quad_surf(SP2D, es.parameters.get<Order>("Gauss quadrature order"));
            UniquePtr<FEBase> fe_v_surf = FEBase::build(SP2D, fe_v_type);
            fe_v_surf->attach_quadrature_rule(&quad_surf);
            //
            std::vector<dof_id_type> dof_indices;
            std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars_RDS];
            std::vector<dof_id_type> vars;
            for (unsigned int i=0; i<nvars_RDS; i++) vars.push_back( i );
            v_ctx.get_dof_indices(side.get(), dof_indices, dof_indices_var, &vars);
            //
            std::vector<unsigned int> var_dofs;
            calc_var_dofs(nvars_RDS, dof_indices_var, var_dofs);
            // mass matrix
            DenseMatrix<Number> Me(dof_indices.size(), dof_indices.size());
            std::vector< DenseSubMatrix<Number> > Mesub(nvars_RDS*nvars_RDS, Me);
            std::vector<unsigned int> offset;
            calc_offset(nvars_RDS, var_dofs, offset);
            for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                for (unsigned int jvar=0; jvar<nvars_RDS; jvar++) {
                    Mesub[ivar*nvars_RDS+jvar].reposition(offset[ivar], offset[jvar], var_dofs[ivar], var_dofs[jvar]);
                }
            }
            // get the ID of the boundary
            const boundary_id_type bc_id = msh.boundary_info->boundary_id(elem, s);
            // check if membrane material region exists
            std::map<region_id_type, RegionInfo>::const_iterator citer_regn = this->_dmn.regn.find(bc_id);
            if ( this->_dmn.regn.end() != citer_regn && membrane_present ) {
                // get access to the membrane material region
                //const RegionInfo* ri_ptr = &(citer_regn->second);
                //libmesh_assert( ri_ptr->is_membrane );
                // initialize and get access to the FE stuff
                const std::vector<Real>&                JxW = fe_v_surf->get_JxW();
                const std::vector< std::vector<Real> >& Phi = fe_v_surf->get_phi();
                fe_v_surf->reinit(side.get());
                // get access to the element quadrature points data
                //libmesh_assert( this->_dmn.elem_side_qi_local.end() != this->_dmn.elem_side_qi_local.find(elem->id()) );
                //libmesh_assert( this->_dmn.elem_side_qi_local.find(elem->id())->second.end() != this->_dmn.elem_side_qi_local.find(elem->id())->second.find(s) );
                //const QuadratureInfo* qpi_ptr = &(this->_dmn.elem_side_qi_local.find(elem->id())->second.find(s)->second);
                // loop over surface integration quadrature points
                for (unsigned int qp=0; qp<quad_surf.n_points(); qp++) {
                    for (unsigned int i=0; i<var_dofs[0]; i++) {
                        for (unsigned int j=0; j<var_dofs[0]; j++) {
                            for (int a=0; a<nvars_RDS; a++)
                                Mesub[a*nvars_RDS+a](i,j) += JxW[qp] * (Phi[i][qp]*Phi[j][qp])/time_step;
                        }
                    }
                    // end of surface integration quadrature loop...
                }
                // end of if case...
            }
            // row-sum method for lumping
            const unsigned int ndofs = dof_indices.size();
            for (unsigned int a=0; a<ndofs; a++) {
                Real Me_rs = 0.0;
                for (unsigned int b=0; b<ndofs; b++) {
                    Me_rs += Me(a,b);
                    Me(a,b) = 0.0;
                }
                Me(a,a) = Me_rs;
            }
            // update system matrix
            v_sys.matrix->add_matrix(Me, dof_indices);
            // free memo
            delete [] dof_indices_var;
            // end of element sides loop...
        }
        // end of active local elements loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    // evaluate the reciprocal vector of the lumped (diagonal) system matrix
    v_sys.matrix->close();
    v_sys.matrix->get_diagonal(v_sys.get_vector("reciprocal diag-mat"));
    v_sys.get_vector("reciprocal diag-mat").reciprocal();
    v_sys.get_vector("reciprocal diag-mat").close();

    this->_log.pop("_diag_matrix");
}
//=================================================================================================
void ReactionDiffusionSolverXD::_solve () {
    this->_log.push("_solve");

    EquationSystems& es = this->_dmn.get_equation_systems();
    LinearImplicitSystem& v_sys = es.get_system<LinearImplicitSystem>(this->_label+"::solution");
    System&  v_P_sys = es.get_system<System>(this->_label+"::previous_solution");
    System& DvDt_sys = es.get_system<System>(this->_label+"::solution_time_derivative");
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // check if time-integration is explicit or implicit
    if ( this->par_list.get<bool>("explicit solve") ) {
        // assemble the rhs-vector
        this->_assemble_rhs();
        // calculate the effective rhs vector
        v_sys.get_vector("rhs").zero();
        v_sys.get_vector("rhs").pointwise_mult(*v_sys.rhs, v_sys.get_vector("reciprocal diag-mat"));
        v_sys.get_vector("rhs").close();
        // update the current solution
        v_sys.solution->add(v_sys.get_vector("rhs"));
        v_sys.solution->close();
        v_sys.update();
    } else {
        // assemble the rhs-vector and matrix
        this->_assemble_all();
        // now solve the linearised system
        v_sys.solve();
    }
    //
    DvDt_sys.solution->operator=(*v_sys.solution);
    DvDt_sys.solution->operator-=(*v_P_sys.solution);
    DvDt_sys.solution->operator*=(1.0/this->par_list.get<Real>("time step"));
    DvDt_sys.solution->close();
    DvDt_sys.update();
    //
    v_P_sys.solution->operator=(*v_sys.solution);
    v_P_sys.solution->close();
    v_P_sys.update();
    //
    // update the corresponding quadrature information data for this solver
    this->_set_auxiliary();
    this->_set_quad_info();
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("_solve");
}
//=================================================================================================
void ReactionDiffusionSolverXD::_set_auxiliary () {
    this->_log.push("_set_auxiliary");

    const Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    EquationSystems& es = this->_dmn.get_equation_systems();
    const System&  v_sys   = es.get_system<System>(this->_label+"::solution");
    // const System& DvDt_sys = es.get_system<System>(this->_label+"::solution_time_derivative");
    System&  a_sys   = es.get_system<System>(this->_label+"::auxiliary_solution");
    System& DaDt_sys = es.get_system<System>(this->_label+"::auxiliary_solution_time_derivative");
    //
    const MeshBase::const_element_iterator begin_el = es.get_mesh().active_elements_begin();
    const MeshBase::const_element_iterator end_el   = es.get_mesh().active_elements_end();

    const unsigned int ninvars_NCS = par.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = par.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = par.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = par.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    const Real current_time = par.get<Real>("current time"),
               time_step    = par.get<Real>("reaction-diffusion solver: time step");

    // check the problem we are running...
    if ( par.get<ProblemCase>("solver") == FEM_INVITRO_TUMOROID_SOLVER &&
         nvars_RDS == 10 && ndcvars_RDS == 1 && nvars_AUX == 5 ) {
        //
        std::set<dof_id_type> scanned_nodes;
        //
        std::vector<Number> rds_soln;
        v_sys.update_global_solution(rds_soln);
        //
        std::vector<Number> aux_soln;
        a_sys.update_global_solution(aux_soln);
        //
        std::vector<Number> aux_deriv_soln;
        DaDt_sys.update_global_solution(aux_deriv_soln);
        //
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        // loop over all active local finite elements
        for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
            const Elem* elem = *citer_el;
            //
            // get access to the material region
            libmesh_assert( this->_dmn.regn.end() != this->_dmn.regn.find(elem->subdomain_id()) );
            const RegionInfo* ri_ptr = &(this->_dmn.regn.find(elem->subdomain_id())->second);
            //
            for (unsigned int inode=0; inode<elem->n_nodes(); inode++) {
                const Node* node = elem->node_ptr(inode);
                //
                if ( libMesh::global_processor_id() != node->processor_id() ) continue;
                //
                if ( scanned_nodes.end() != scanned_nodes.find(node->id()) ) continue;
                scanned_nodes.insert(node->id());
                //
                FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
                fld.current_time = current_time;
                fld.time_step = time_step;
                //
                fld.a_flag = -1;
                //
                for (unsigned int I=0; I<nvars_RDS; I++)
                    fld.biochem[I] = rds_soln[node->dof_number(v_sys.number(), I, 0)];
                for (unsigned int I=0; I<nvars_AUX; I++)
                    fld.auxiliary[I] = aux_soln[node->dof_number(a_sys.number(), I, 0)];
                for (unsigned int I=0; I<nvars_AUX; I++)
                    fld.auxiliary_rate[I] = aux_deriv_soln[node->dof_number(DaDt_sys.number(), I, 0)];
                //
                Blitz_VectorReal      Fv__Crd;
                Blitz_VectorGradient  Fv__Grad_Crd;
                Blitz_DyadicReal      Km__Sys;
                Blitz_VectorPoint     Fv__Velc;
                Blitz_DyadicReal      Km__Diff;
                Blitz_DyadicGradient  Km__Taxs;
                //
                ri_ptr->calc_biochem(fld, Fv__Crd, Fv__Grad_Crd, Km__Sys, Fv__Velc, Km__Diff, Km__Taxs);
                //
                if ( ! current_time )
                    aux_soln[node->dof_number(a_sys.number(), 3, 0)] = fld.auxiliary[3];
                aux_soln[node->dof_number(a_sys.number(), 4, 0)] = fld.auxiliary[4];
                //
                if ( ! current_time )
                    aux_deriv_soln[node->dof_number(DaDt_sys.number(), 3, 0)] = fld.auxiliary_rate[3];
                aux_deriv_soln[node->dof_number(DaDt_sys.number(), 4, 0)] = fld.auxiliary_rate[4];
                // end of element nodes loop...
            }
            // end of active local finite elements loop...
        }
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        // close solution vectors and update the systems
        a_sys.solution->operator=(aux_soln);
        a_sys.solution->close();
        a_sys.update();
        DaDt_sys.solution->operator=(aux_deriv_soln);
        DaDt_sys.solution->close();
        DaDt_sys.update();
        // end of if case...
    }

    this->_log.pop("_set_auxiliary");
}
//=================================================================================================
void ReactionDiffusionSolverXD::_set_quad_info () {
    this->_log.push("_set_quad_info");

    const Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    EquationSystems& es = this->_dmn.get_equation_systems();
    System&  v_sys   = es.get_system<System>(this->_label+"::solution");
    System& DvDt_sys = es.get_system<System>(this->_label+"::solution_time_derivative");
    System& dc_sys = es.get_system<System>(this->_label+"::discontinuous_solution");
    System&  a_sys   = es.get_system<System>(this->_label+"::auxiliary_solution");
    System& DaDt_sys = es.get_system<System>(this->_label+"::auxiliary_solution_time_derivative");
    //
    const FEM_Context v_ctx(v_sys);
    const FEM_Context dc_ctx(dc_sys);
    const FEM_Context a_ctx(a_sys);
    const MeshBase::const_element_iterator begin_el = es.get_mesh().active_local_elements_begin();
    const MeshBase::const_element_iterator end_el   = es.get_mesh().active_local_elements_end();
    //
    FEType fe_v_type = v_sys.variable_type(0);
    FEType fe_dc_type = dc_sys.variable_type(0);
    FEType fe_a_type = a_sys.variable_type(0);
    //
    const bool membrane_present = this->par_list.get<bool>("membrane elements present");
    const unsigned int //ninvars_NCS = par.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = par.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = par.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = par.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    // sanity checks
    ASSERT_(false==membrane_present, "membrane finite elements have been removed from solver");
    ASSERT_(nvars_RDS==v_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(ndcvars_RDS==dc_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(nvars_AUX==a_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(nvars_AUX==DaDt_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(this->par_list.get<unsigned int>("dimension")==v_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(this->par_list.get<unsigned int>("discontinuous variables")==dc_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(this->par_list.get<unsigned int>("auxiliary variables")==a_sys.n_vars(), "fatal internal error occurred");
    //
    // obtain the global solution vectors
    std::vector<Number> v_sys_soln;
    v_sys.update_global_solution(v_sys_soln);
    std::vector<Number> DvDt_sys_soln;
    DvDt_sys.update_global_solution(DvDt_sys_soln);
    std::vector<Number> a_sys_soln;
    a_sys.update_global_solution(a_sys_soln);
    std::vector<Number> DaDt_sys_soln;
    DaDt_sys.update_global_solution(DaDt_sys_soln);
    //
    // loop over all active local finite elements
    for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
        const Elem* elem = *citer_el;
        //
        QGauss quad_volu(SP3D, par.get<Order>("Gauss quadrature order"));
        UniquePtr<FEBase> fe_v_volu = FEBase::build(SP3D, fe_v_type);
        fe_v_volu->attach_quadrature_rule(&quad_volu);
        UniquePtr<FEBase> fe_dc_volu = FEBase::build(SP3D, fe_dc_type);
        fe_dc_volu->attach_quadrature_rule(&quad_volu);
        UniquePtr<FEBase> fe_a_volu = FEBase::build(SP3D, fe_a_type);
        fe_a_volu->attach_quadrature_rule(&quad_volu);
        //
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars_RDS];
        v_ctx.get_dof_indices(elem, dof_indices, dof_indices_var);
        std::vector<dof_id_type> dof_indices__disc;
        std::vector<dof_id_type>* dof_indices_var__disc = new std::vector<dof_id_type> [ndcvars_RDS];
        dc_ctx.get_dof_indices(elem, dof_indices__disc, dof_indices_var__disc);
        std::vector<dof_id_type> dof_indices__aux;
        std::vector<dof_id_type>* dof_indices_var__aux = new std::vector<dof_id_type> [nvars_AUX];
        a_ctx.get_dof_indices(elem, dof_indices__aux, dof_indices_var__aux);
        //
        std::vector<unsigned int> var_dofs;
        calc_var_dofs(nvars_RDS, dof_indices_var, var_dofs);
        std::vector<unsigned int> var_dofs__disc;
        calc_var_dofs(ndcvars_RDS, dof_indices_var__disc, var_dofs__disc);
        std::vector<unsigned int> var_dofs__aux;
        calc_var_dofs(nvars_AUX, dof_indices_var__aux, var_dofs__aux);
        // initialize and get access to the FE stuff
        const std::vector< std::vector<Real> >&          Phi = fe_v_volu->get_phi();
        const std::vector< std::vector<RealGradient> >& dPhi_dX = fe_v_volu->get_dphi();
        fe_v_volu->reinit(elem);
        const std::vector< std::vector<Real> >&        Theta = fe_dc_volu->get_phi();
        fe_dc_volu->reinit(elem);
        const std::vector< std::vector<Real> >&          Psi = fe_a_volu->get_phi();
        const std::vector< std::vector<RealGradient> >& dPsi_dX = fe_a_volu->get_dphi();
        fe_a_volu->reinit(elem);
        // get access to the element quadrature points data
        libmesh_assert( this->_dmn.elem_qi_local.end() != this->_dmn.elem_qi_local.find(elem->id()) );
        QuadratureInfo* qpi_ptr = &(this->_dmn.elem_qi_local.find(elem->id())->second);
        // loop over volume integration quadrature points
        for (unsigned int qp=0; qp<quad_volu.n_points(); qp++) {
            //
            // update the quadrature information (continuous) data
            for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                qpi_ptr->biochem[qp][ivar] = 0.0;
                qpi_ptr->grad_biochem[qp][ivar] = RealGradient(0.0, 0.0, 0.0);
                qpi_ptr->biochem_rate[qp][ivar] = 0.0;
                //
                char field_name[buffer_len];
                sprintf(field_name, "constrained_field_%d_type", ivar);
                for (unsigned int l=0; l<var_dofs[ivar]; l++) {
                    if        ( -2 == this->par_list.get<int>(field_name) ||
                                -1 == this->par_list.get<int>(field_name) ) {
                        qpi_ptr->biochem[qp][ivar]      += Phi[l][qp]     * v_sys_soln[dof_indices_var[ivar][l]];
                        qpi_ptr->grad_biochem[qp][ivar] += dPhi_dX[l][qp] * v_sys_soln[dof_indices_var[ivar][l]];
                    } else if (  0 == this->par_list.get<int>(field_name) ||
                                10 == this->par_list.get<int>(field_name) ) {
                        qpi_ptr->biochem[qp][ivar]      += Phi[l][qp]     * apply_lbound(0.0, v_sys_soln[dof_indices_var[ivar][l]]);
                        qpi_ptr->grad_biochem[qp][ivar] += dPhi_dX[l][qp] * apply_lbound(0.0, v_sys_soln[dof_indices_var[ivar][l]]);
                    } else if (  2 == this->par_list.get<int>(field_name) ||
                                12 == this->par_list.get<int>(field_name) ) {
                        qpi_ptr->biochem[qp][ivar]      += Phi[l][qp]     * apply_bounds(0.0, v_sys_soln[dof_indices_var[ivar][l]], 1.0);
                        qpi_ptr->grad_biochem[qp][ivar] += dPhi_dX[l][qp] * apply_bounds(0.0, v_sys_soln[dof_indices_var[ivar][l]], 1.0);
                    } else {
                        ERROR_("error in the `"+std::string(field_name)+"' ="+std::to_string(this->par_list.get<int>(field_name)));
                    }
                    //
                    qpi_ptr->biochem_rate[qp][ivar] += Phi[l][qp] * DvDt_sys_soln[dof_indices_var[ivar][l]];
                }
                if ( -2 == this->par_list.get<int>(field_name) ) {
                    bool negate =false;
                    for (unsigned int l=0; l<var_dofs[ivar]; l++) {
                        if ( 0.0 > v_sys_soln[dof_indices_var[ivar][l]] )
                            negate = true;
                    }
                    if ( negate ) {
                        qpi_ptr->biochem[qp][ivar] = 0.0;
                        qpi_ptr->grad_biochem[qp][ivar] = RealGradient(0.0, 0.0, 0.0);
                        //
                        qpi_ptr->biochem_rate[qp][ivar] = 0.0;
                    }
                }
            }
            // update the quadrature information (discontinuous) data
            for (unsigned int ivar=nvars_RDS; ivar<nvars_RDS+ndcvars_RDS; ivar++) {
                qpi_ptr->biochem[qp][ivar] = 0.0;
                //
                unsigned int jvar=ivar-nvars_RDS;
                for (unsigned int l=0; l<var_dofs__disc[jvar]; l++) {
                    qpi_ptr->biochem[qp][ivar] += Theta[l][qp] * dc_sys.current_solution(dof_indices_var__disc[jvar][l]);
                }
            }
            //
            for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                qpi_ptr->auxiliary[qp][ivar] = 0.0;
                qpi_ptr->grad_auxiliary[qp][ivar] = RealGradient(0.0, 0.0, 0.0);
                qpi_ptr->auxiliary_rate[qp][ivar] = 0.0;
                //
                for (unsigned int l=0; l<var_dofs__aux[ivar]; l++) {
                    qpi_ptr->auxiliary[qp][ivar]      += Psi[l][qp]     * a_sys_soln[dof_indices_var__aux[ivar][l]];
                    qpi_ptr->grad_auxiliary[qp][ivar] += dPsi_dX[l][qp] * a_sys_soln[dof_indices_var__aux[ivar][l]];
                    qpi_ptr->auxiliary_rate[qp][ivar] += Psi[l][qp]  * DaDt_sys_soln[dof_indices_var__aux[ivar][l]];
                }
            }
            // end of volume integration quadrature loop...
        }
        // free memo
        delete [] dof_indices_var;
        delete [] dof_indices_var__disc;
        delete [] dof_indices_var__aux;
        // end of active local finite elements loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    // check if to enforce constraining the nodal solution for some state-variables
    if ( this->par_list.get<bool>("enforce solution") ) {
        // loop over all active local finite elements
        for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
            const Elem* elem = *citer_el;
            //
            std::vector<dof_id_type> dof_indices;
            std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [nvars_RDS];
            v_ctx.get_dof_indices(elem, dof_indices, dof_indices_var);
            //
            std::vector<unsigned int> var_dofs;
            calc_var_dofs(nvars_RDS, dof_indices_var, var_dofs);
            //
            // loop over system variables
            for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                char field_name[buffer_len];
                sprintf(field_name, "constrained_field_%d_type", ivar);
                for (unsigned int l=0; l<var_dofs[ivar]; l++) {
                    if        ( -2 == this->par_list.get<int>(field_name) ||
                                -1 == this->par_list.get<int>(field_name) ) {
                        ;
                    } else if (  0 == this->par_list.get<int>(field_name) ) {
                        ;
                    } else if ( 10 == this->par_list.get<int>(field_name) ) {
                        const Real soln = apply_lbound(0.0, v_sys_soln[dof_indices_var[ivar][l]]);
                        v_sys.solution->set(dof_indices_var[ivar][l], soln);
                    } else if (  2 == this->par_list.get<int>(field_name) ) {
                        ;
                    } else if ( 12 == this->par_list.get<int>(field_name) ) {
                        const Real soln = apply_bounds(0.0, v_sys_soln[dof_indices_var[ivar][l]], 1.0);
                        v_sys.solution->set(dof_indices_var[ivar][l], soln);
                    } else libmesh_error();
                    // end of nodal points (associated to variable) loop...
                }
                // end of variables loop...
            }
            // free memo
            delete [] dof_indices_var;
            // end of active local finite elements loop...
        }
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        // now update the system accordingly
        v_sys.solution->close();
        v_sys.update();
    }

    this->_log.pop("_set_quad_info");
}
//=================================================================================================
void ReactionDiffusionSolverXD::save_restart (std::ofstream& fout) const {
    this->_log.push("save_restart");

    for (unsigned isys=0; isys<systems_ls.size(); isys++) {
        const EquationSystems* es_ptr = systems_ls[isys].first;
        const System& sys = es_ptr->get_system<System>(systems_ls[isys].second);
        //
        sys.solution->close();
        std::vector<Number> soln;
        sys.solution->localize_to_one(soln);
        //
        if ( libMesh::global_processor_id() ) continue;
        //
        const unsigned int n = soln.size();
        fout.write((char*)&n, sizeof(unsigned int));
        for (unsigned int i=0; i<n; i++) {
            const Number component = soln[i];
            fout.write((char*)&component, sizeof(Number));
        }
    }
    //
    this->_dmn.save_quadrature(fout, QuadratureInfo::REFERENCE, QuadratureInfo::LEVEL_1);
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("save_restart");
}
//=================================================================================================
void ReactionDiffusionSolverXD::load_restart () {
    this->_log.push("load_restart");

    const Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    for (unsigned int ifile=0; ifile<par.get<unsigned int>("input restart from file"); ifile++) {
        const std::string str_name = "input restart file name #" + std::to_string(ifile);
        const std::string restart_file_name = par.get<std::string>(str_name.c_str());
        //
        std::ifstream fin(restart_file_name.c_str());
        // read the system solution from the first restart file only
        if ( 0 == ifile )
            for (unsigned isys=0; isys<systems_ls.size(); isys++) {
                EquationSystems* es_ptr = systems_ls[isys].first;
                System& sys = es_ptr->get_system<System>(systems_ls[isys].second);
                //
                unsigned int n = 0;
                fin.read((char*)&n, sizeof(unsigned int));
                for (unsigned int i=0; i<n; i++) {
                    Number component;
                    fin.read((char*)&component, sizeof(Number));
                    sys.solution->set(i, component);
                }
            }
        // read Quadrature data for this processor only
        if ( libMesh::global_processor_id() == ifile )
            this->_dmn.load_quadrature(fin, QuadratureInfo::REFERENCE, QuadratureInfo::LEVEL_1);
        // end of restart files loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    for (unsigned isys=0; isys<systems_ls.size(); isys++) {
        EquationSystems* es_ptr = systems_ls[isys].first;
        System& sys = es_ptr->get_system<System>(systems_ls[isys].second);
        //
        sys.solution->close();
        sys.update();
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("load_restart");
}
//=================================================================================================
