/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include <fstream>
#include "libmesh/gmsh_io.h"
#include "libmesh/mesh_tools.h"
#include "feb3/rve.h"
//=================================================================================================
// external variables
extern
libMesh::Parallel::Communicator* gcomm_ptr;
// local variables
static
std::map<unsigned int, std::vector<Real> > node_map; // node IDs start from "1" not "0"!
static
std::vector<unsigned int> node_s1, node_s2; // ...same as above
static
std::map<unsigned int, int> node_bc; // ...same as above; face labels are shown below
//=================================================================================================
std::string RVE::_gen_rve (const unsigned int nfib, const Real cmin, const Real cmax) {
    // write-up the 1D mesh in a temporary file to subsequently process
    const auto ch_set = charset();
    std::default_random_engine rng(std::random_device{}());
    std::uniform_int_distribution<> dist(0, ch_set.size()-1);
    auto randchar = [ch_set, &dist, &rng]() { return ch_set[dist(rng)]; };
    std::string fname = random_string(12, randchar) + std::string(".msh");
    //
    // permit zero-ranked process to broadcast the file name to save mesh
    MPI_Bcast((void*)fname.c_str(), fname.size(), MPI_CHAR, 0, MPI_COMM_WORLD);
    // exit this routine normally for non zero-ranked processes
    if ( libMesh::global_processor_id() ) {
        // now return the filename where the 1D mesh has been exported
        return fname;
    }
    //
    // initialise RVE parameters
    unsigned int NF = nfib;
    // initialise some local parameters
    const Real L = 0.2;
    const Real D2 = pow2(0.08);
    // initialise some extra local parameters
    const unsigned int NS = 2*NF+1;
    const unsigned int NT = 2*NF;
    // initialise RVE variables
    std::vector<Real> x(NS),    y(NS),    z(NS);
    std::vector<Real> x0(NS),   y0(NS),   z0(NS);
    std::vector<Real> dx(3*NS), dy(3*NS), dz(3*NS);
    std::vector<unsigned int> a(3*NT+1);
    // initialise a random number generator
    std::random_device rd;
    std::default_random_engine rgen(rd());
    // initialise uniform distribution
    std::uniform_real_distribution<Real> udist(0.0, 1.0);
    // create initial seeds
    for (unsigned int i=1; i<=NF; i++) {
        x0[i] = 2.0*udist(rgen);
        y0[i] = 2.0*udist(rgen);
        z0[i] = 2.0*udist(rgen);
        x[2*i-1] = x0[i];
        y[2*i-1] = y0[i];
        z[2*i-1] = z0[i];
        x[2*i] = x0[i];
        y[2*i] = y0[i];
        z[2*i] = z0[i];
        const Real dx1 = udist(rgen),
                   dy1 = udist(rgen),
                   dz1 = udist(rgen),
                   d1 = sqrt(pow2(dx1)+pow2(dy1)+pow2(dz1));
        dx[2*i-1] =  dx1/d1;
        dx[2*i+0] = -dx1/d1;
        dy[2*i-1] =  dy1/d1;
        dy[2*i+0] = -dy1/d1;
        dz[2*i-1] =  dz1/d1;
        dz[2*i+0] = -dz1/d1;
        a[2*i-1] = 2*i-1;
        a[2*i+0] = 2*i;
    }
    //
    unsigned int NL = 2*NF;
    // create vector of "living" segments
    std::vector<unsigned int> live;
    for (unsigned int i=0; i<=NL; i++) live.push_back(i);
    //
    // start adding monomer
    while (NL > 0) {
        unsigned int g = 1;
        // select fibre to update; +1 because matlab indices begin at 1 and floor(...) can equal 0
        unsigned int j = floor(udist(rgen)*NL) + 1;
        unsigned int k = live[j];
        // update fibre characteristics
        x[k] += L*dx[k];
        y[k] += L*dy[k];
        z[k] += L*dz[k];
        Real x1 = x[k];
        Real y1 = y[k];
        Real z1 = z[k];
        // check for fibre out of big box
        Real u[6];
        u[0] = (x1 - 2.0)/fabs(dx[k]);
        u[1] = (y1 - 2.0)/fabs(dy[k]);
        u[2] = (z1 - 2.0)/fabs(dz[k]);
        u[3] = -x1/fabs(dx[k]);
        u[4] = -y1/fabs(dy[k]);
        u[5] = -z1/fabs(dz[k]);
        Real Lu = u[0];
        // find max value of u
        for (unsigned int j=1; j<6; j++) {
            if (u[j] > Lu) Lu = u[j];
        }
        // fibre out of box
        if (Lu > 0.0) {
            x1 -= Lu*dx[k];
            y1 -= Lu*dy[k];
            z1 -= Lu*dz[k];
            // kill segment
            for (unsigned int m=j; m<=(NL-1); m++) {
                live[m] = live[m+1];
            }
            NL--;
            g = 0;
        }
        // check for fibre-fibre collision
        if ( 1 == g ) {
            unsigned int i = 0;
            if ( (k==1) || (k==2) ) i = 2;
            else i = 1;
            //
            while (i < NF) {
              Real xi0 = x[a[2*i-1]];
              Real yi0 = y[a[2*i-1]];
              Real zi0 = z[a[2*i-1]];
              Real xi1 = x[a[2*i]];
              Real yi1 = y[a[2*i]];
              Real zi1 = z[a[2*i]];
              Real h  = sqrt(pow2(xi1-xi0)+pow2(yi1-yi0)+pow2(zi1-zi0));
              Real h1 = sqrt(pow2(xi1-x1 )+pow2(yi1-y1 )+pow2(zi1-z1 ));
              Real h2 = sqrt(pow2(xi0-x1 )+pow2(yi0-y1 )+pow2(zi0-z1 ));
              //
              if ( (h1 < h) && (h2 < h) ) {
                  const Real aa = pow2(xi1-x1 )+pow2(yi1-y1 )+pow2(zi1-z1 ),
                             bb = pow2(xi0-xi1)+pow2(yi0-yi1)+pow2(zi0-zi1),
                             cc = pow2((xi1-x1)*(xi0-xi1)+(yi1-y1)*(yi0-yi1)+(zi1-z1)*(zi0-zi1));
                  const Real dist2 = (aa*bb-cc)/bb;
                  // collision!
                  if ( dist2 < D2 ) {
                      const Real tt = -((xi1-x1)*(xi0-xi1)+(yi1-y1)*(yi0-yi1)+(zi1-z1)*(zi0-zi1))/bb;
                      x1 = xi1 + (xi0-xi1)*tt;
                      y1 = yi1 + (yi0-yi1)*tt;
                      z1 = zi1 + (zi0-zi1)*tt;
                      // split struck fibre in two
                      NF++;
                      //
                      a[2*NF-1] = k;
                      a[2*NF] = a[2*i];
                      dx[2*NF] = dx[2*i];
                      dy[2*NF] = dy[2*i];
                      dz[2*NF] = dz[2*i];
                      dx[2*NF-1] = dx[2*i-1];
                      dy[2*NF-1] = dy[2*i-1];
                      dz[2*NF-1] = dz[2*i-1];
                      a[2*i] = k;
                      // kill segment in collision
                      for (unsigned int m=j; m<=(NL-1); m++) {
                          live[m] = live[m+1];
                      }
                      NL--;
                      // skip remainder of search
                      i = NF;
                  }
              }
              i++;
              // skip collision with self!
              if ( (a[2*i-1] == k) || (a[2*i] == k) ) i++;
              // end of while loop
            }
            // end of conditional
        }
        //
        g=1;
        x[k] = x1;
        y[k] = y1;
        z[k] = z1;
        // end of while loop
    }
    // coordinate containers
    std::vector<Real> temp(NF);
    std::vector< std::vector<Real> > Ax; Ax.assign(2, temp);
    std::vector< std::vector<Real> > Ay; Ay.assign(2, temp);
    std::vector< std::vector<Real> > Az; Az.assign(2, temp);
    std::vector<Real> b(2*NF);
    unsigned int k = 0;
    // select fibres with nodes in the box [x,y,z](-0.5,0.5)
    for (unsigned int i=1; i<=NF; i++) {
        //
        if ( (x[a[2*i]] > 0.51) &&
             (x[a[2*i]] < 1.49) &&
             (y[a[2*i]] > 0.51) &&
             (y[a[2*i]] < 1.49) &&
             (z[a[2*i]] > 0.51) &&
             (z[a[2*i]] < 1.49) )
        {
            Real u[6];
            u[0] = (x[a[2*i-1]] - 1.49)/fabs(dx[2*i-1]);
            u[1] = (y[a[2*i-1]] - 1.49)/fabs(dy[2*i-1]);
            u[2] = (z[a[2*i-1]] - 1.49)/fabs(dz[2*i-1]);
            u[3] = (0.51 - x[a[2*i-1]])/fabs(dx[2*i-1]);
            u[4] = (0.51 - y[a[2*i-1]])/fabs(dy[2*i-1]);
            u[5] = (0.51 - z[a[2*i-1]])/fabs(dz[2*i-1]);
            Real Lu = u[0];
            // find max value of u
            for (unsigned int j=1; j<6; j++) {
                if(u[j] > Lu) Lu = u[j];
            }
            // node a[2*i-1] is out of box?
            if ( Lu > 0.0 ) {
                u[0] = (x[a[2*i-1]] - 1.5)/fabs(dx[2*i-1]);
                u[1] = (y[a[2*i-1]] - 1.5)/fabs(dy[2*i-1]);
                u[2] = (z[a[2*i-1]] - 1.5)/fabs(dz[2*i-1]);
                u[3] = (0.5 - x[a[2*i-1]])/fabs(dx[2*i-1]);
                u[4] = (0.5 - y[a[2*i-1]])/fabs(dy[2*i-1]);
                u[5] = (0.5 - z[a[2*i-1]])/fabs(dz[2*i-1]);
                Real Lu1 = u[0];
                // find max value of u
                for (unsigned int j=1; j<6; j++) {
                  if (u[j] > Lu1) Lu1 = u[j];
                }
                x[a[2*i-1]] -= Lu1*dx[2*i-1];
                y[a[2*i-1]] -= Lu1*dy[2*i-1];
                z[a[2*i-1]] -= Lu1*dz[2*i-1];
            }
            k++;
            //
            Ax[0][k] = x[a[2*i-1]] - 1.0;
            Ax[1][k] = x[a[2*i]]   - 1.0;
            Ay[0][k] = y[a[2*i-1]] - 1.0;
            Ay[1][k] = y[a[2*i]]   - 1.0;
            Az[0][k] = z[a[2*i-1]] - 1.0;
            Az[1][k] = z[a[2*i]]   - 1.0;
            b[2*k-1] = a[2*i-1];
            b[2*k]   = a[2*i];
            // end of conditional
        } else if ( (x[a[2*i-1]] > 0.51) &&
                    (x[a[2*i-1]] < 1.49) &&
                    (y[a[2*i-1]] > 0.51) &&
                    (y[a[2*i-1]] < 1.49) &&
                    (z[a[2*i-1]] > 0.51) &&
                    (z[a[2*i-1]] < 1.49) )
        {
            Real u[6];
            u[0] = (x[a[2*i]] - 1.49)/fabs(dx[2*i]);
            u[1] = (y[a[2*i]] - 1.49)/fabs(dy[2*i]);
            u[2] = (z[a[2*i]] - 1.49)/fabs(dz[2*i]);
            u[3] = (0.51 - x[a[2*i]])/fabs(dx[2*i]);
            u[4] = (0.51 - y[a[2*i]])/fabs(dy[2*i]);
            u[5] = (0.51 - z[a[2*i]])/fabs(dz[2*i]);
            Real Lu = u[0];
            // find max value of u
            for (unsigned int j=1; j<6; j++) {
                if (u[j] > Lu) Lu = u[j];
            }
            // node a[2*i] is out of box?
            if ( Lu > 0.0 ) {
                u[0] = (x[a[2*i]] - 1.5)/fabs(dx[2*i]);
                u[1] = (y[a[2*i]] - 1.5)/fabs(dy[2*i]);
                u[2] = (z[a[2*i]] - 1.5)/fabs(dz[2*i]);
                u[3] = (0.5 - x[a[2*i]])/fabs(dx[2*i]);
                u[4] = (0.5 - y[a[2*i]])/fabs(dy[2*i]);
                u[5] = (0.5 - z[a[2*i]])/fabs(dz[2*i]);
                Real Lu1 = u[0];
                // find max value of u
                for (unsigned int j=1; j<6; j++) {
                  if (u[j] > Lu1) Lu1 = u[j];
                }
                x[a[2*i]] -= Lu1*dx[2*i];
                y[a[2*i]] -= Lu1*dy[2*i];
                z[a[2*i]] -= Lu1*dz[2*i];
            }
            //
            k++;
            Ax[0][k] = x[a[2*i-1]] - 1.0;
            Ax[1][k] = x[a[2*i]]   - 1.0;
            Ay[0][k] = y[a[2*i-1]] - 1.0;
            Ay[1][k] = y[a[2*i]]   - 1.0;
            Az[0][k] = z[a[2*i-1]] - 1.0;
            Az[1][k] = z[a[2*i]]   - 1.0;
            b[2*k-1] = a[2*i-1];
            b[2*k]   = a[2*i];
        }
        // end of for loop
    }
    // would close Lu file here...
    NF = k;
    std::vector<unsigned int> c;
    for (unsigned int i=0; i<=(2*NF); i++) c.push_back(0);
    unsigned int index = 0;
    unsigned int count = 1;
    c[1] = 1;
    for (unsigned int i=2; i<=(2*NF); i++) {
        for (unsigned int j=1; j<=(i-1); j++) {
            if (b[i]==b[j]) {
                c[i]=c[j];
                index=1;
            }
        }
        if (0==index) {
            count+=1;
            c[i]=count;
        }
        index=0;
        // end of for loop...
    }
    // transformation (if necessary) coefficients
    const Real alpha = (cmax-cmin),
               beta  = (cmax+cmin)*0.5,
               epsilon = 0.01*(cmax-cmin);
    //
    // Real proj_len = 0.0;
    std::vector<Real> node1_xcoord, node1_ycoord, node1_zcoord;
    std::vector<Real> node2_xcoord, node2_ycoord, node2_zcoord;
    // clean memo first
    node_s1.clear();
    node_s2.clear();
    // store node ID and position
    for (unsigned int i=1; i<=NF; i++) {
        // const Real BOX_MIN_COORD = -0.5,
        //            BOX_MAX_COORD =  0.5,
        //            BOX_DIF_COORD = BOX_MAX_COORD-BOX_MIN_COORD;
        // std::uniform_real_distribution<Real> cdist(0.001*BOX_DIF_COORD, 0.009*BOX_DIF_COORD);
        const Real BOX_MIN_COORD = -0.5,
                   BOX_MAX_COORD =  0.5;
        // check if fibre is still outside mesh...
        // if true, then pull the offending node back to the boundary
        for (unsigned ii=0; ii<2; ii++) {
            // FIXME: should do a linear interpolation
            if ( Ax[ii][i] > BOX_MAX_COORD ) Ax[ii][i] = BOX_MAX_COORD;// + cdist(rgen);
            if ( Ax[ii][i] < BOX_MIN_COORD ) Ax[ii][i] = BOX_MIN_COORD;// - cdist(rgen);
            if ( Ay[ii][i] > BOX_MAX_COORD ) Ay[ii][i] = BOX_MAX_COORD;// + cdist(rgen);
            if ( Ay[ii][i] < BOX_MIN_COORD ) Ay[ii][i] = BOX_MIN_COORD;// - cdist(rgen);
            if ( Az[ii][i] > BOX_MAX_COORD ) Az[ii][i] = BOX_MAX_COORD;// + cdist(rgen);
            if ( Az[ii][i] < BOX_MIN_COORD ) Az[ii][i] = BOX_MIN_COORD;// - cdist(rgen);
        }
        //
        node_s1.push_back(c[2*i-1]);
        node_s2.push_back(c[2*i+0]);
        node1_xcoord.push_back(Ax[0][i]);
        node1_ycoord.push_back(Ay[0][i]);
        node1_zcoord.push_back(Az[0][i]);
        node2_xcoord.push_back(Ax[1][i]);
        node2_ycoord.push_back(Ay[1][i]);
        node2_zcoord.push_back(Az[1][i]);
        // end of for loop...
    }
    //
    const unsigned int NODES = count;
    const unsigned int ELEMENTS = NF;
    // clean memo first
    node_map.clear();
    // let's sort the nodes out - use a map to order by ID and skip duplicates
    for (unsigned int i=0; i<NF; i++) {
        std::vector<Real> sv1;
        sv1.push_back(alpha*node1_xcoord[i]+beta);
        sv1.push_back(alpha*node1_ycoord[i]+beta);
        sv1.push_back(alpha*node1_zcoord[i]+beta);
        std::vector<Real> sv2;
        sv2.push_back(alpha*node2_xcoord[i]+beta);
        sv2.push_back(alpha*node2_ycoord[i]+beta);
        sv2.push_back(alpha*node2_zcoord[i]+beta);
        std::vector<Real> sv1_sv2;
        sv1_sv2.push_back(alpha*(node1_xcoord[i]-node2_xcoord[i]));
        sv1_sv2.push_back(alpha*(node1_ycoord[i]-node2_ycoord[i]));
        sv1_sv2.push_back(alpha*(node1_zcoord[i]-node2_zcoord[i]));
        //
        node_map.insert(std::make_pair(node_s1[i], sv1));
        node_map.insert(std::make_pair(node_s2[i], sv2));
        // end of for loop...
    }
    //
    // clean memo first
    node_bc.clear();
    // find the boundary nodes
    for ( std::map<unsigned int, std::vector<Real> >::const_iterator
          it=node_map.begin(); it!=node_map.end(); it++)
    {
        const std::vector<Real>& coord = it->second;
        int face = -1;
        if      ( coord[0] < cmin+epsilon ) face = 4;
        else if ( coord[0] > cmax-epsilon ) face = 2;
        else if ( coord[1] < cmin+epsilon ) face = 1;
        else if ( coord[1] > cmax-epsilon ) face = 3;
        else if ( coord[2] < cmin+epsilon ) face = 0;
        else if ( coord[2] > cmax-epsilon ) face = 5;
        else ;
        //
        if ( face >=0 ) {
            // std::cout << it->first << ' ' << face+1 << std::endl;
            node_bc.insert(std::make_pair(it->first, face));
        }
    }
    // sanity checks
    if (NODES!=node_map.size()) libmesh_error();
    if (ELEMENTS!=node_s1.size()) libmesh_error();
    if (ELEMENTS!=node_s2.size()) libmesh_error();
    if (node_bc.empty()) libmesh_error();
    //
    // now export the 1D mesh
    this->write(fname);
    // now return the filename where the 1D mesh has been exported
    return fname;
}
//=================================================================================================
void RVE::_setup_stuff () {
    const BoundingBox box = MeshTools::create_bounding_box(*this);
    const MeshBase::const_element_iterator begin_el = this->active_elements_begin();
    const MeshBase::const_element_iterator end_el   = this->active_elements_end();
    /*

           7        6
           o--------o
          /:       /|
         / :      / |
      4 /  :   5 /  |
       o--------o   |
       |   o....|...o 2
       |  .3    |  /
       | .      | /
       |.       |/
       o--------o
       0        1

    */
    // create the vertices of the RVE according and list them according to this schematic
    this->_vertex.clear();
    this->_vertex.push_back( Point(box.min()(0), box.min()(1), box.min()(2)) );
    this->_vertex.push_back( Point(box.max()(0), box.min()(1), box.min()(2)) );
    this->_vertex.push_back( Point(box.max()(0), box.max()(1), box.min()(2)) );
    this->_vertex.push_back( Point(box.min()(0), box.max()(1), box.min()(2)) );
    this->_vertex.push_back( Point(box.min()(0), box.min()(1), box.max()(2)) );
    this->_vertex.push_back( Point(box.max()(0), box.min()(1), box.max()(2)) );
    this->_vertex.push_back( Point(box.max()(0), box.max()(1), box.max()(2)) );
    this->_vertex.push_back( Point(box.min()(0), box.max()(1), box.max()(2)) );
    /*

        o--------o
       /:       /|
      / : 5    / |
     /  :   3 /  |
    o--------o 2 |
    | 4 o....|...o
    |  . 1   |  /
    | .   0  | /
    |.       |/
    o--------o

    */
    // create the normal vectors for each face according to this schematic
    this->_normal[0] = Point( 0.,  0., -1.);
    this->_normal[1] = Point( 0., -1.,  0.);
    this->_normal[2] = Point( 1.,  0.,  0.);
    this->_normal[3] = Point( 0.,  1.,  0.);
    this->_normal[4] = Point(-1.,  0.,  0.);
    this->_normal[5] = Point( 0.,  0.,  1.);
    // transformation coefficients & Jacobian of transformation
    const Real a = 0.5*(this->max_coord()-this->min_coord()),
               b = 0.5*(this->max_coord()+this->min_coord()),
               u0 = -1.0/sqrt(3.0), // 1st integration point [-1,+1]
               u1 = +1.0/sqrt(3.0); // 2nd integration point [-1,+1]
    this->_jac = a; // [a,b] <===> [-1,+1]
    //
    // setup the Gauss-Legendre integration points for all 6 faces of the RVE cube;
    // absissas are given below, while integration weights are equal to unity!
    this->_quad_pnt[0].clear();
    this->_quad_pnt[0].push_back( Point(a*u0+b, a*u0+b, this->min_coord()) );
    this->_quad_pnt[0].push_back( Point(a*u1+b, a*u0+b, this->min_coord()) );
    this->_quad_pnt[0].push_back( Point(a*u0+b, a*u1+b, this->min_coord()) );
    this->_quad_pnt[0].push_back( Point(a*u1+b, a*u1+b, this->min_coord()) );
    this->_quad_pnt[1].clear();
    this->_quad_pnt[1].push_back( Point(a*u0+b, this->min_coord(), a*u0+b) );
    this->_quad_pnt[1].push_back( Point(a*u1+b, this->min_coord(), a*u0+b) );
    this->_quad_pnt[1].push_back( Point(a*u0+b, this->min_coord(), a*u1+b) );
    this->_quad_pnt[1].push_back( Point(a*u1+b, this->min_coord(), a*u1+b) );
    this->_quad_pnt[2].clear();
    this->_quad_pnt[2].push_back( Point(this->max_coord(), a*u0+b, a*u0+b) );
    this->_quad_pnt[2].push_back( Point(this->max_coord(), a*u1+b, a*u0+b) );
    this->_quad_pnt[2].push_back( Point(this->max_coord(), a*u0+b, a*u1+b) );
    this->_quad_pnt[2].push_back( Point(this->max_coord(), a*u1+b, a*u1+b) );
    this->_quad_pnt[3].clear();
    this->_quad_pnt[3].push_back( Point(a*u0+b, this->max_coord(), a*u0+b) );
    this->_quad_pnt[3].push_back( Point(a*u1+b, this->max_coord(), a*u0+b) );
    this->_quad_pnt[3].push_back( Point(a*u0+b, this->max_coord(), a*u1+b) );
    this->_quad_pnt[3].push_back( Point(a*u1+b, this->max_coord(), a*u1+b) );
    this->_quad_pnt[4].clear();
    this->_quad_pnt[4].push_back( Point(this->min_coord(), a*u0+b, a*u0+b) );
    this->_quad_pnt[4].push_back( Point(this->min_coord(), a*u1+b, a*u0+b) );
    this->_quad_pnt[4].push_back( Point(this->min_coord(), a*u0+b, a*u1+b) );
    this->_quad_pnt[4].push_back( Point(this->min_coord(), a*u1+b, a*u1+b) );
    this->_quad_pnt[5].clear();
    this->_quad_pnt[5].push_back( Point(a*u0+b, a*u0+b, this->max_coord()) );
    this->_quad_pnt[5].push_back( Point(a*u1+b, a*u0+b, this->max_coord()) );
    this->_quad_pnt[5].push_back( Point(a*u0+b, a*u1+b, this->max_coord()) );
    this->_quad_pnt[5].push_back( Point(a*u1+b, a*u1+b, this->max_coord()) );
    //
    const Real p_tol = 0.02*(this->max_coord()-this->min_coord());
    const Point p_min = box.min();
    const Point p_max = box.max();
    //
    // zero-out the total fibre length
    this->_length = 0.0;
    // clear memory of the boundary nodes container
    this->_bnode.clear();
    this->_bnode_neighbor.clear();
    for (int f=0; f<6; f++) this->_bnode_neighbor_face[f].clear();
    // now iterate for all elements
    for (MeshBase::const_element_iterator ci_elem=begin_el; ci_elem!=end_el; ci_elem++)
    {
        const Elem* elem = (*ci_elem);
        libmesh_assert(libMesh::EDGE2==elem->type());
        //
        if (elem->n_nodes()!=2) libmesh_error();
        const Point& p0_ = elem->point(0);
        const Point& p1_ = elem->point(1);
        //
        int face =-1;
        std::map<dof_id_type, std::set<dof_id_type>>::iterator iter;
        if ( ! elem->neighbor_ptr(0) && ! elem->neighbor_ptr(1) ) {
            face = -1;
            if      ( p0_(0) < p_min(0)+p_tol ) face = 4;
            else if ( p0_(0) > p_max(0)-p_tol ) face = 2;
            else if ( p0_(1) < p_min(1)+p_tol ) face = 1;
            else if ( p0_(1) > p_max(1)-p_tol ) face = 3;
            else if ( p0_(2) < p_min(2)+p_tol ) face = 0;
            else if ( p0_(2) > p_max(2)-p_tol ) face = 5;
            else ;
            if (-1!=face) {
                this->_bnode.insert(std::make_pair(elem->node_id(0), face));
                //
                iter = this->_bnode_neighbor.find(elem->node_id(0));
                if ( this->_bnode_neighbor.end() == iter ) {
                    std::set<dof_id_type> ts;
                    ts.insert(elem->node_id(1));
                    this->_bnode_neighbor.insert( std::make_pair(elem->node_id(0), ts) );
                } else iter->second.insert(elem->node_id(1));
                //
                iter = this->_bnode_neighbor_face[face].find(elem->node_id(0));
                if ( this->_bnode_neighbor_face[face].end() == iter ) {
                    std::set<dof_id_type> ts;
                    ts.insert(elem->node_id(1));
                    this->_bnode_neighbor_face[face].insert( std::make_pair(elem->node_id(0), ts) );
                } else iter->second.insert(elem->node_id(1));
            }
            face = -1;
            if      ( p1_(0) < p_min(0)+p_tol ) face = 4;
            else if ( p1_(0) > p_max(0)-p_tol ) face = 2;
            else if ( p1_(1) < p_min(1)+p_tol ) face = 1;
            else if ( p1_(1) > p_max(1)-p_tol ) face = 3;
            else if ( p1_(2) < p_min(2)+p_tol ) face = 0;
            else if ( p1_(2) > p_max(2)-p_tol ) face = 5;
            else ;
            if (-1!=face) {
                this->_bnode.insert(std::make_pair(elem->node_id(1), face));
                //
                iter = this->_bnode_neighbor.find(elem->node_id(1));
                if ( this->_bnode_neighbor.end() == iter ) {
                    std::set<dof_id_type> ts;
                    ts.insert(elem->node_id(0));
                    this->_bnode_neighbor.insert( std::make_pair(elem->node_id(1), ts) );
                } else iter->second.insert(elem->node_id(0));
                //
                iter = this->_bnode_neighbor_face[face].find(elem->node_id(1));
                if ( this->_bnode_neighbor_face[face].end() == iter ) {
                    std::set<dof_id_type> ts;
                    ts.insert(elem->node_id(0));
                    this->_bnode_neighbor_face[face].insert( std::make_pair(elem->node_id(1), ts) );
                } else iter->second.insert(elem->node_id(0));
            }
        } else if ( ! elem->neighbor_ptr(0) && elem->neighbor_ptr(1) ) {
            face = -1;
            if      ( p0_(0) < p_min(0)+p_tol ) face = 4;
            else if ( p0_(0) > p_max(0)-p_tol ) face = 2;
            else if ( p0_(1) < p_min(1)+p_tol ) face = 1;
            else if ( p0_(1) > p_max(1)-p_tol ) face = 3;
            else if ( p0_(2) < p_min(2)+p_tol ) face = 0;
            else if ( p0_(2) > p_max(2)-p_tol ) face = 5;
            else ;
            if (-1!=face) {
                this->_bnode.insert(std::make_pair(elem->node_id(0), face));
                //
                iter = this->_bnode_neighbor.find(elem->node_id(0));
                if ( this->_bnode_neighbor.end() == iter ) {
                    std::set<dof_id_type> ts;
                    ts.insert(elem->node_id(1));
                    this->_bnode_neighbor.insert( std::make_pair(elem->node_id(0), ts) );
                } else iter->second.insert(elem->node_id(1));
                //
                iter = this->_bnode_neighbor_face[face].find(elem->node_id(0));
                if ( this->_bnode_neighbor_face[face].end() == iter ) {
                    std::set<dof_id_type> ts;
                    ts.insert(elem->node_id(1));
                    this->_bnode_neighbor_face[face].insert( std::make_pair(elem->node_id(0), ts) );
                } else iter->second.insert(elem->node_id(1));
            }
        } else if ( elem->neighbor_ptr(0) && ! elem->neighbor_ptr(1) ) {
            face = -1;
            if      ( p1_(0) < p_min(0)+p_tol ) face = 4;
            else if ( p1_(0) > p_max(0)-p_tol ) face = 2;
            else if ( p1_(1) < p_min(1)+p_tol ) face = 1;
            else if ( p1_(1) > p_max(1)-p_tol ) face = 3;
            else if ( p1_(2) < p_min(2)+p_tol ) face = 0;
            else if ( p1_(2) > p_max(2)-p_tol ) face = 5;
            else ;
            if (-1!=face) {
                this->_bnode.insert(std::make_pair(elem->node_id(1), face));
                //
                iter = this->_bnode_neighbor.find(elem->node_id(1));
                if ( this->_bnode_neighbor.end() == iter ) {
                    std::set<dof_id_type> ts;
                    ts.insert(elem->node_id(0));
                    this->_bnode_neighbor.insert( std::make_pair(elem->node_id(1), ts) );
                } else iter->second.insert(elem->node_id(0));
                //
                iter = this->_bnode_neighbor_face[face].find(elem->node_id(1));
                if ( this->_bnode_neighbor_face[face].end() == iter ) {
                    std::set<dof_id_type> ts;
                    ts.insert(elem->node_id(0));
                    this->_bnode_neighbor_face[face].insert( std::make_pair(elem->node_id(1), ts) );
                } else iter->second.insert(elem->node_id(0));
            }
        } else {
            ; // do nothing here...
        }
        // update the total fibre length
        this->_length += elem->volume();
    }
    // calculate the volume of the RVE
    this->_volume = pow3(this->max_coord()-this->min_coord());
    this->_side_surface = pow2(this->max_coord()-this->min_coord());
    this->_surface = 6.0*this->_side_surface;
    //
    // std::cout << std::endl
    //           << " >> RVE geometric features:\n"
    //           << "    domain = [" << this->min_coord() << ',' << this->max_coord() << "]^3\n"
    //           << "    volume = " << this->_volume << "\n"
    //           << "    surface = " << this->_surface << "\n"
    //           << "    total fibre length = " << this->_length << "\n"
    //           << std::endl;
    //
}
//=================================================================================================
void RVE::generate (const unsigned int nfib, const Real cmin, const Real cmax) {
    const std::string fname = this->_gen_rve(nfib, cmin, cmax);
    // now load and prepare the 1D mesh
    GmshIO(*this).read(fname);
    this->prepare_for_use();
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // and delete the temporary file
    remove(fname.c_str());
    // finally, set-up some important stuff for the RVE
    this->_setup_stuff();
}
//=================================================================================================
void RVE::write (const std::string& fname) const {
    // exit this routine normally for non zero-ranked processes
    if ( libMesh::global_processor_id() ) return;
    // open the file-stream where to save the mesh of the RVE
    std::ofstream fout(fname);
    //
    const unsigned int n_node = node_map.size(),
                       n_elem = node_s1.size(),
                       n_bc_node = node_bc.size();
    if (n_elem!=node_s2.size()) libmesh_error();
    // write data to Gmsh format file
    fout << "$MeshFormat" << std::endl;
    fout << "2.0 0 8" << std::endl;
    fout << "$EndMeshFormat" << std::endl;
    fout << "$Nodes" << std::endl;
    fout << n_node << std::endl;
    for ( std::map<unsigned int, std::vector<Real> >::const_iterator
          it=node_map.begin(); it!=node_map.end(); it++) {
        fout << it->first << "  "
             << (it->second)[0] << ' '
             << (it->second)[1] << ' '
             << (it->second)[2] << ' '
             << std::endl;
        // end of for loop...
    }
    fout << "$EndNodes" << std::endl;
    fout << "$Elements" << std::endl;
    fout << (n_bc_node+n_elem) << std::endl;
    unsigned int index = 1;
    for ( std::map<unsigned int, int>::const_iterator
          it=node_bc.begin(); it!=node_bc.end(); it++) {
        const unsigned int pid = 1 + index;
        fout << index << " 15  2 " << pid << ' ' << it->second << ' '
             << ' ' << it->first
             << std::endl;
        ++index;
    }
    for (unsigned int i=0; i<n_elem; i++) {
        const unsigned int pid = 100001 + index;
        fout << index << " 1  2 " << pid << ' ' << 0 << ' '
             << ' ' << node_s1[i] << ' ' << node_s2[i]
             << std::endl;
        ++index;
    }
    fout << "$EndElements" << std::endl;
    // close file stream normally
    fout.close();
    // end of routine...
}
//=================================================================================================
void RVE::reinit () {
    // prepare the 1D mesh
    this->prepare_for_use();
    // finally, set-up some important stuff for the RVE
    this->_setup_stuff();
}
//=================================================================================================
bool RVE::node_is_boundary (const dof_id_type& id, int& f) const {
    std::map<dof_id_type, int>::const_iterator ci = this->_bnode.find(id);
    f = -1;
    if ( this->_bnode.end() != ci ) {
        f = ci->second;
        return true;
    } else return false;
}
//=================================================================================================
