/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/tumour_angiogenesis_solver.h"
#include "libmesh/sparse_matrix.h"
#include "libmesh/linear_implicit_system.h"
#include "libmesh/petsc_linear_solver.h"
#include "feb3/quadrature.h"
#include "feb3/ida.h"
#include "feb3/fem-context.h"
#include "feb3/tumour_angiogenesis_constitutive.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
extern
TumourAngiogenesisConstitutive ta_constitutive;
//=================================================================================================
#include "feb3/domain_data_inl.h"
#include "feb3/solution_io_inl.h"
#include "feb3/vascular_network_aux_inl.h"
#include "feb3/vascular_network_io_inl.h"
//=================================================================================================
void TumourAngiogenesisSolver::_fluid_solver () {
    this->_log.push("_fluid_solver");

    const Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    ASSERT_( this->_dmn.mater.end() !=
             this->_dmn.mater.find(par.get<int>("tumour angiogenesis solver: vascular remodelling material ID")), "fatal internal error occurred");
    const MaterialInfo* mi__vnr = &(this->_dmn.mater.find(par.get<int>("tumour angiogenesis solver: vascular remodelling material ID"))->second);
    ASSERT_(NULL!=mi__vnr, "fatal internal error occurred");
    //
    const System& coupled_U_sys = this->_dmn_coupled.get_equation_systems().get_system<System>(this->_label+"::coupled::displacement_solution");
    const System& rds_sys = this->_dmn.get_equation_systems().get_system<System>(par.get<std::string>("reaction-diffusion solver: label")+"::solution");
    const System& aux_sys = this->_dmn.get_equation_systems().get_system<System>(par.get<std::string>("reaction-diffusion solver: label")+"::auxiliary_solution");

    std::map<dof_id_type, QuadratureInfo>* elem_qpi = ( libMesh::global_n_processors() == 1 )
                                                    ? &this->_dmn.elem_qi_local
                                                    : &this->_dmn.elem_qi;
    std::map<dof_id_type, std::map<unsigned int, QuadratureInfo> >* elem_side_qpi = ( libMesh::global_n_processors() == 1 )
                                                                                  ? &this->_dmn.elem_side_qi_local
                                                                                  : &this->_dmn.elem_side_qi;
    const Real current_time = par.get<Real>("current time");
    // const Real current_time = par.get<Real>("current time"),
    //            time_step = par.get<Real>("tumour angiogenesis solver: time step")
    //                      * par.get<int>("tumour angiogenesis solver: angiogenesis increment");
    const unsigned int ninvars_NCS = par.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = par.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = par.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = par.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    ASSERT_(ndcvars_RDS==9, "number of discontinuous variables adopted for the reaction-diffusion solver is invalid");
    ASSERT_(nvars_AUX==8, "number of auxiliary variables adopted for the reaction-diffusion solver is invalid");

    this->_log.push("_fluid_solver (TASK #1)");
    flog.print("  > Run fluid solver (1/6)          \r", false);
    ferr.print("  > Run fluid solver (1/6)          \r", true);
    // if only in parallel retrieve the quadrature info from all processors
    if ( libMesh::global_n_processors() > 1 )
    // and only if the "global" Quadrature data structure is empty
    if ( elem_qpi->empty() || elem_side_qpi->empty() ) {
        char fname[buffer_len];
        //
        flog.print("  > Run fluid solver (1-A/6)          \r", false);
        ferr.print("  > Run fluid solver (1-A/6)          \r", true);
        sprintf( fname, "%s/quadrature_data/%s.qpi.%d",
                 par.get<std::string>("results directory").c_str(),
                 par.get<std::string>("output file name").c_str(),
                 libMesh::global_processor_id() );
        this->_dmn.save_quadrature(fname, QuadratureInfo::ALL, QuadratureInfo::LEVEL_3);
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        //
        flog.print("  > Run fluid solver (1-B/6)          \r", false);
        ferr.print("  > Run fluid solver (1-B/6)          \r", true);
        std::vector<std::string> fname_ls(libMesh::global_n_processors());
        for (unsigned int iproc=0; iproc<libMesh::global_n_processors(); iproc++) {
            sprintf( fname, "%s/quadrature_data/%s.qpi.%d",
                     par.get<std::string>("results directory").c_str(),
                     par.get<std::string>("output file name").c_str(),
                     iproc );
            fname_ls[iproc] = std::string(fname);
        }
        this->_dmn.gather_quadrature(fname_ls, QuadratureInfo::ALL, QuadratureInfo::LEVEL_3);
    }
    //
    flog.print("  > Run fluid solver (1-C/6)          \r", false);
    ferr.print("  > Run fluid solver (1-C/6)          \r", true);
    std::vector<Number> coupled_displacement;
    coupled_U_sys.update_global_solution(coupled_displacement);
    //
    flog.print("  > Run fluid solver (1-D/6)          \r", false);
    ferr.print("  > Run fluid solver (1-D/6)          \r", true);
    std::set<dof_id_type> vasc_node_to_regress;
    for ( std::map<dof_id_type, VascularInfo>::const_iterator
          ci_ndvi=this->_dmn_vn.node_vi.begin(); ci_ndvi!=this->_dmn_vn.node_vi.end(); ci_ndvi++ )
    {
        const Node* node = this->_dmn_vn.get_mesh().node_ptr(ci_ndvi->first);
        ASSERT_(NULL!=node, "fatal internal error occurred");
        const VascularInfo* vi = &(ci_ndvi->second);
        ASSERT_(NULL!=vi, "fatal internal error occurred");
        //
        dof_id_type inoutlet_node_id = DofObject::invalid_id;
        std::set<dof_id_type> scanned_nodes;
        find_inlet_or_outlet( this->_dmn_vn.get_mesh(), node->id(),
                              this->_dmn_vn.node_vi, scanned_nodes, inoutlet_node_id);
        //
        if ( DofObject::invalid_id == inoutlet_node_id )
            vasc_node_to_regress.insert(node->id());
        // ...end of vascular nodes loop
    }
    //
    flog.print("  > Run fluid solver (1-E/6)          \r", false);
    ferr.print("  > Run fluid solver (1-E/6)          \r", true);
    std::vector<Number> ta_biochem_species;
    rds_sys.update_global_solution(ta_biochem_species);
    std::vector<Number> ta_auxiliary_species;
    aux_sys.update_global_solution(ta_auxiliary_species);
    // a label indicating that we may need to re-run the fluid solver
    int number_of_reruns = 0;
    rerun_fluid_solver: ;
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    this->_log.pop("_fluid_solver (TASK #1)");

    // flow solver system assembly (matrix and right-hand-side vector), where it is assumed for:
    // Hagen-Poiseuille flow in the vascular network, Darcy flow of bio-fluids in the extracellular space,
    // and Starling's law is considered for the transvascular flow.
    this->_log.push("_fluid_solver (TASK #2)");
    flog.print("  > Run fluid solver (2/6)          \r", false);
    ferr.print("  > Run fluid solver (2/6)          \r", true);
    do {
        EquationSystems& es_coupled = this->_dmn_coupled.get_equation_systems();
        LinearImplicitSystem& coupled_P_sys = es_coupled.get_system<LinearImplicitSystem>(this->_label+"::coupled::pressure_solution");
        ASSERT_(1==coupled_P_sys.n_vars(), "fatal internal error occurred");
        SparseMatrix<Number>*  system_matrix = coupled_P_sys.matrix;
        NumericVector<Number>* system_rhs    = coupled_P_sys.rhs;
        //
        const MeshBase& msh_coupled = es_coupled.get_mesh();
        libmesh_assert( SP1D == msh_coupled.mesh_dimension() );
        //
        const MeshBase& msh_3d = this->_dmn.get_mesh();
        libmesh_assert( SP3D == msh_3d.mesh_dimension() );
        //
        const FEM_Context p_ctx(coupled_P_sys);
        const MeshBase::const_element_iterator begin_el = msh_coupled.active_elements_begin();
        const MeshBase::const_element_iterator end_el = msh_coupled.active_elements_end();
        // const MeshBase::const_element_iterator begin_el = msh_coupled.active_local_elements_begin();
        // const MeshBase::const_element_iterator end_el = msh_coupled.active_local_elements_end();
        //
        Real system_matrix_max_local  = 0.0,
             system_matrix_max_global = 0.0;
        //
        system_matrix->zero();
        system_rhs->zero();
        //
        // loop over all active local finite elements
        flog.print("  > Run fluid solver (2-A/6)          \r", false);
        ferr.print("  > Run fluid solver (2-A/6)          \r", true);
        const Real critical_length = mi__vnr->get_biochem_param(14) / 10;
        for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
            const Elem* elem = *citer_el;
            libmesh_assert( elem->type() == libMesh::EDGE2 );
            const Node* node[] = { elem->node_ptr(0) ,
                                   elem->node_ptr(1) };
            libmesh_assert( NULL != node[0] );
            libmesh_assert( NULL != node[1] );
            const DomainData_SolidCoupledWithVascularNetwork::Type segm_type =
                static_cast<DomainData_SolidCoupledWithVascularNetwork::Type>(elem->subdomain_id());
            //
            // current (deformed) length of element segment
            const Point node_U[] = { Point( coupled_displacement[node[0]->dof_number(coupled_U_sys.number(), 0, 0)],
                                            coupled_displacement[node[0]->dof_number(coupled_U_sys.number(), 1, 0)],
                                            coupled_displacement[node[0]->dof_number(coupled_U_sys.number(), 2, 0)] ),
                                     Point( coupled_displacement[node[1]->dof_number(coupled_U_sys.number(), 0, 0)],
                                            coupled_displacement[node[1]->dof_number(coupled_U_sys.number(), 1, 0)],
                                            coupled_displacement[node[1]->dof_number(coupled_U_sys.number(), 2, 0)] ) };
            const Real segm_length = Point((*node[1])+node_U[1]-(*node[0])-node_U[0]).norm();
            ASSERT_(segm_length>critical_length, "fatal internal error occurred");
            //
            if ( DomainData_SolidCoupledWithVascularNetwork::VESSEL == segm_type ) {
                //
                ASSERT_(this->_dmn_coupled.cm_vasc_node__vasc_node.end()!=this->_dmn_coupled.cm_vasc_node__vasc_node.find(node[0]->id()), "fatal internal error occurred");
                ASSERT_(this->_dmn_coupled.cm_vasc_node__vasc_node.end()!=this->_dmn_coupled.cm_vasc_node__vasc_node.find(node[1]->id()), "fatal internal error occurred");
                const unsigned int id[] = { this->_dmn_coupled.cm_vasc_node__vasc_node.find(node[0]->id())->second,
                                            this->_dmn_coupled.cm_vasc_node__vasc_node.find(node[1]->id())->second };
                //
                // get access to the vascular info
                ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(id[0]), "fatal internal error occurred");
                ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(id[1]), "fatal internal error occurred");
                VascularInfo* vi[] = { &(this->_dmn_vn.node_vi.find(id[0])->second),
                                       &(this->_dmn_vn.node_vi.find(id[1])->second) };
                ASSERT_(NULL!=vi[0], "fatal internal error occurred");
                ASSERT_(NULL!=vi[1], "fatal internal error occurred");
                if      ( VascularInfo::BLOOD==vi[0]->flow ) ASSERT_(VascularInfo::BLOOD==vi[1]->flow, "fatal internal error occurred");
                else if ( VascularInfo::LYMPH==vi[0]->flow ) ASSERT_(VascularInfo::LYMPH==vi[1]->flow, "fatal internal error occurred");
                else libmesh_error();
                //
                const Real radius_min = VascularInfo::BLOOD == vi[0]->flow && VascularInfo::BLOOD == vi[1]->flow
                                      ? mi__vnr->get_biochem_param(52) : mi__vnr->get_biochem_param(102);
                //
                // ignore following computations if vascular node needs to be regressed
                if ( vasc_node_to_regress.end() != vasc_node_to_regress.find(id[0]) ) continue;
                if ( vasc_node_to_regress.end() != vasc_node_to_regress.find(id[1]) ) continue;
                bool skip_calculations = false;
                //
                if ( VascularInfo::COLLAPSED == vi[0]->status ) skip_calculations = true;
                if ( VascularInfo::COLLAPSED == vi[1]->status ) skip_calculations = true;
                //
                unsigned int I = 0;
                for (std::set<dof_id_type>::const_iterator ci=vi[0]->vsc_elem.begin(); ci!=vi[0]->vsc_elem.end(); ci++) {
                    const Elem& elem = this->_dmn_vn.get_mesh().elem_ref(*ci);
                    const dof_id_type conn_node_id = edge2_other_node(elem, id[0]);
                    //
                    ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(conn_node_id), "fatal internal error occurred");
                    const VascularInfo* conn_vi = &(this->_dmn_vn.node_vi.find(conn_node_id)->second);
                    ASSERT_(NULL!=conn_vi, "fatal internal error occurred");
                    //
                    if ( VascularInfo::COLLAPSED == conn_vi->status ) continue;
                    //
                    ASSERT_(3>I, "fatal internal error occurred");
                    ++I;
                }
                // if this (functional) vascular node is "alone" then shut it down!
                if ( 0 == I ) {
                    vi[0]->status = VascularInfo::COLLAPSED;
                    vi[0]->compression_scale = 1000.0;
                    //
                    const Real temp_load = 0.0;
                    const Real features[] = { current_time, 0.0, 0.0, 0.0, 0.0, temp_load,
                                              0.0, 0.0, 0.0, 0.0 };
                    //
                    // call the vessel wall remodelling routine to compute radius, thickness and pore size
                    tumour_angiogenesis::remodel_vascular_XYZ(*mi__vnr, *vi[0], features);
                    //
                    vi[0]->radius = radius_min / vi[0]->compression_scale;
                    //
                    // ignore following computations for collapsed vascular node
                    skip_calculations = true;
                }
                //
                I = 0;
                for (std::set<dof_id_type>::const_iterator ci=vi[1]->vsc_elem.begin(); ci!=vi[1]->vsc_elem.end(); ci++) {
                    const Elem& elem = this->_dmn_vn.get_mesh().elem_ref(*ci);
                    const dof_id_type conn_node_id = edge2_other_node(elem, id[1]);
                    //
                    ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(conn_node_id), "fatal internal error occurred");
                    const VascularInfo* conn_vi = &(this->_dmn_vn.node_vi.find(conn_node_id)->second);
                    ASSERT_(NULL!=conn_vi, "fatal internal error occurred");
                    //
                    if ( VascularInfo::COLLAPSED == conn_vi->status ) continue;
                    //
                    ASSERT_(3>I, "fatal internal error occurred");
                    ++I;
                }
                // if this (functional) vascular node is "alone" then shut it down!
                if ( 0 == I ) {
                    vi[1]->status = VascularInfo::COLLAPSED;
                    vi[1]->compression_scale = 1000.0;
                    //
                    const Real temp_load = 0.0;
                    const Real features[] = { current_time, 0.0, 0.0, 0.0, 0.0, temp_load,
                                              0.0, 0.0, 0.0, 0.0 };
                    //
                    // call the vessel wall remodelling routine to compute radius, thickness and pore size
                    tumour_angiogenesis::remodel_vascular_XYZ(*mi__vnr, *vi[1], features);
                    //
                    vi[1]->radius = radius_min / vi[1]->compression_scale;
                    //
                    // ignore following computations for collapsed vascular node
                    skip_calculations = true;
                }
                //
                // ignore following computations if vascular node has collapsed
                if ( skip_calculations ) continue;
                // ignore following computations if vascular element (segment) is not local
                if ( libMesh::global_processor_id() != elem->processor_id() ) continue;
                //
                // access the 3D-solid elements containing the vascular nodes
                const Elem* elem3d[] = { msh_3d.elem_ptr(vi[0]->elem3d_id),
                                         msh_3d.elem_ptr(vi[1]->elem3d_id) };
                ASSERT_(NULL!=elem3d[0], "fatal internal error occurred");
                ASSERT_(NULL!=elem3d[1], "fatal internal error occurred");
                //
                // get access to the material info
                ASSERT_(this->_dmn.regn.end()!=this->_dmn.regn.find(elem3d[0]->subdomain_id()), "fatal internal error occurred");
                ASSERT_(this->_dmn.regn.end()!=this->_dmn.regn.find(elem3d[1]->subdomain_id()), "fatal internal error occurred");
                const MaterialInfo* mi__dmn[] = { this->_dmn.regn.find(elem3d[0]->subdomain_id())->second.mat,
                                                  this->_dmn.regn.find(elem3d[1]->subdomain_id())->second.mat };
                ASSERT_(NULL!=mi__dmn[0], "fatal internal error occurred");
                ASSERT_(NULL!=mi__dmn[1], "fatal internal error occurred");
                //
                FieldInfo fld_0(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
                FieldInfo fld_1(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
                //
                const Real hydraulic_conductivity[] = { VascularInfo::BLOOD==vi[0]->flow ?
                                                        ta_constitutive.blood_vessel_hydraulic_conductivity(vi[0], mi__vnr, mi__dmn[0], &fld_0) :
                                                        ta_constitutive.lymph_vessel_hydraulic_conductivity(vi[0], mi__vnr, mi__dmn[0], &fld_0) ,
                                                        VascularInfo::BLOOD==vi[1]->flow ?
                                                        ta_constitutive.blood_vessel_hydraulic_conductivity(vi[1], mi__vnr, mi__dmn[1], &fld_1) :
                                                        ta_constitutive.lymph_vessel_hydraulic_conductivity(vi[1], mi__vnr, mi__dmn[1], &fld_1) };
                const Real vessel_cross_sectional_area[] = { vascular_cross_section(vi[0]),
                                                             vascular_cross_section(vi[1]) };
                //
                const Real mat[] = { hydraulic_conductivity[0]*vessel_cross_sectional_area[0]/segm_length,
                                     hydraulic_conductivity[1]*vessel_cross_sectional_area[1]/segm_length };
                ASSERT_(hydraulic_conductivity[0]>0.0, "fatal internal error occurred");
                ASSERT_(hydraulic_conductivity[1]>0.0, "fatal internal error occurred");
                ASSERT_(vessel_cross_sectional_area[0]>0.0, "fatal internal error occurred");
                ASSERT_(vessel_cross_sectional_area[1]>0.0, "fatal internal error occurred");
                // ASSERT_(mat[0]>0.0, "fatal internal error occurred");
                // ASSERT_(mat[1]>0.0, "fatal internal error occurred");
                //
                system_matrix->add( node[0]->dof_number(coupled_P_sys.number(),0,0), node[0]->dof_number(coupled_P_sys.number(),0,0),
                                    -mat[0] );
                system_matrix->add( node[0]->dof_number(coupled_P_sys.number(),0,0), node[1]->dof_number(coupled_P_sys.number(),0,0),
                                    +mat[0] );
                system_matrix->add( node[1]->dof_number(coupled_P_sys.number(),0,0), node[0]->dof_number(coupled_P_sys.number(),0,0),
                                    +mat[1] );
                system_matrix->add( node[1]->dof_number(coupled_P_sys.number(),0,0), node[1]->dof_number(coupled_P_sys.number(),0,0),
                                    -mat[1] );
                //
                // find the maximum value of the system matrix (if necessary)
                system_matrix_max_local = std::max(system_matrix_max_local, std::max(mat[0], mat[1]));
                //
            } else if ( DomainData_SolidCoupledWithVascularNetwork::STROMA == segm_type ) {
                //
                ASSERT_(this->_dmn_coupled.cm_strm_node__strm_node.end()!=this->_dmn_coupled.cm_strm_node__strm_node.find(node[0]->id()), "fatal internal error occurred");
                ASSERT_(this->_dmn_coupled.cm_strm_node__strm_node.end()!=this->_dmn_coupled.cm_strm_node__strm_node.find(node[1]->id()), "fatal internal error occurred");
                const unsigned int id[] = { this->_dmn_coupled.cm_strm_node__strm_node.find(node[0]->id())->second ,
                                            this->_dmn_coupled.cm_strm_node__strm_node.find(node[1]->id())->second };
                //
                // get access to the stomal node
                const Node* strm_node[] = { this->_dmn.get_mesh().node_ptr(id[0]) ,
                                            this->_dmn.get_mesh().node_ptr(id[1]) };
                ASSERT_(NULL!=strm_node[0], "fatal internal error occurred");
                ASSERT_(NULL!=strm_node[1], "fatal internal error occurred");
                //
                // ignore following computations if vascular element (segment) is not local
                if ( libMesh::global_processor_id() != elem->processor_id() ) continue;
                //
                // get access to the material info
                ASSERT_(this->_dmn_coupled.cm_strm_elem__subdmnid.end()!=this->_dmn_coupled.cm_strm_elem__subdmnid.find(elem->id()), "fatal internal error occurred");
                const subdomain_id_type rgn_id = this->_dmn_coupled.cm_strm_elem__subdmnid.find(elem->id())->second;
                ASSERT_(this->_dmn.regn.end()!=this->_dmn.regn.find(rgn_id), "fatal internal error occurred");
                const MaterialInfo* mi__dmn = this->_dmn.regn.find(rgn_id)->second.mat;
                ASSERT_(NULL!=mi__dmn, "fatal internal error occurred");
                //
                FieldInfo fld_0(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
                FieldInfo fld_1(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
                for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                    fld_0.biochem[ivar] = ta_biochem_species[strm_node[0]->dof_number(rds_sys.number(), ivar, 0)];
                    fld_1.biochem[ivar] = ta_biochem_species[strm_node[1]->dof_number(rds_sys.number(), ivar, 0)];
                }
                for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                    fld_0.auxiliary[ivar] = ta_auxiliary_species[strm_node[0]->dof_number(aux_sys.number(), ivar, 0)];
                    fld_1.auxiliary[ivar] = ta_auxiliary_species[strm_node[1]->dof_number(aux_sys.number(), ivar, 0)];
                }
                //
                const Point dir = ((*strm_node[1]) - (*strm_node[0])).unit();
                const Dyadic3D hydr_cond_tensor[] = { ta_constitutive.stroma_hydraulic_conductivity(mi__vnr, mi__dmn, &fld_0) ,
                                                      ta_constitutive.stroma_hydraulic_conductivity(mi__vnr, mi__dmn, &fld_1) };
                //
                const Real hydraulic_conductivity[] = { (hydr_cond_tensor[0]*dir).norm() ,
                                                        (hydr_cond_tensor[1]*dir).norm() };
                const Real tissue_cross_sectional_area[] = { ta_constitutive.stroma_cross_section(mi__vnr, mi__dmn, &fld_0) ,
                                                             ta_constitutive.stroma_cross_section(mi__vnr, mi__dmn, &fld_1) };
                //
                const Real mat[] = { hydraulic_conductivity[0]*tissue_cross_sectional_area[0]/segm_length ,
                                     hydraulic_conductivity[1]*tissue_cross_sectional_area[1]/segm_length };
                ASSERT_(mat[0]>0.0, "fatal internal error occurred");
                ASSERT_(mat[1]>0.0, "fatal internal error occurred");
                //
                system_matrix->add( node[0]->dof_number(coupled_P_sys.number(),0,0), node[0]->dof_number(coupled_P_sys.number(),0,0),
                                    -mat[0] );
                system_matrix->add( node[0]->dof_number(coupled_P_sys.number(),0,0), node[1]->dof_number(coupled_P_sys.number(),0,0),
                                    +mat[0] );
                system_matrix->add( node[1]->dof_number(coupled_P_sys.number(),0,0), node[0]->dof_number(coupled_P_sys.number(),0,0),
                                    +mat[1] );
                system_matrix->add( node[1]->dof_number(coupled_P_sys.number(),0,0), node[1]->dof_number(coupled_P_sys.number(),0,0),
                                    -mat[1] );
                //
                // find the maximum value of the system matrix (if necessary)
                system_matrix_max_local = std::max(system_matrix_max_local, std::max(mat[0], mat[1]));
                //
            } else if ( DomainData_SolidCoupledWithVascularNetwork::INTERFACE == segm_type ) {
                //
                ASSERT_(this->_dmn_coupled.cm_vasc_node__vasc_node.end()!=this->_dmn_coupled.cm_vasc_node__vasc_node.find(node[0]->id()), "fatal internal error occurred");
                ASSERT_(this->_dmn_coupled.cm_strm_node__strm_node.end()!=this->_dmn_coupled.cm_strm_node__strm_node.find(node[1]->id()), "fatal internal error occurred");
                const unsigned int id[] = { this->_dmn_coupled.cm_vasc_node__vasc_node.find(node[0]->id())->second ,
                                            this->_dmn_coupled.cm_strm_node__strm_node.find(node[1]->id())->second };
                //
                // get access to the vascular info
                ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(id[0]), "fatal internal error occurred");
                VascularInfo* vi = &(this->_dmn_vn.node_vi.find(id[0])->second);
                ASSERT_(NULL!=vi, "fatal internal error occurred");
                ASSERT_(VascularInfo::BLOOD==vi->flow || VascularInfo::LYMPH==vi->flow, "fatal internal error occurred");
                //
                const Real radius_min = VascularInfo::BLOOD == vi->flow
                                      ? mi__vnr->get_biochem_param(52) : mi__vnr->get_biochem_param(102);
                //
                // ignore following computations if vascular node needs to be regressed
                if ( vasc_node_to_regress.end() != vasc_node_to_regress.find(id[0]) ) continue;
                bool skip_calculations = false;
                //
                if ( VascularInfo::COLLAPSED == vi->status ) skip_calculations = true;
                //
                const Node* conn_node[] = { NULL, NULL, NULL };
                Point conn_node_U[] = { Point(0.0,0.0,0.0), Point(0.0,0.0,0.0), Point(0.0,0.0,0.0) };
                //
                unsigned int I = 0;
                for (std::set<dof_id_type>::const_iterator ci=vi->vsc_elem.begin(); ci!=vi->vsc_elem.end(); ci++) {
                    const Elem& elem = this->_dmn_vn.get_mesh().elem_ref(*ci);
                    const dof_id_type conn_node_id = edge2_other_node(elem, id[0]);
                    //
                    ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(conn_node_id), "fatal internal error occurred");
                    const VascularInfo* conn_vi = &(this->_dmn_vn.node_vi.find(conn_node_id)->second);
                    ASSERT_(NULL!=conn_vi, "fatal internal error occurred");
                    //
                    if ( VascularInfo::COLLAPSED == conn_vi->status ) continue;
                    //
                    conn_node[I] = msh_coupled.node_ptr( this->_dmn_coupled.vasc_node__cm_vasc_node.find(conn_node_id)->second );
                    ASSERT_(NULL!=conn_node[I], "fatal internal error occurred");
                    conn_node_U[I] = Point( coupled_displacement[conn_node[I]->dof_number(coupled_U_sys.number(), 0, 0)] ,
                                            coupled_displacement[conn_node[I]->dof_number(coupled_U_sys.number(), 1, 0)] ,
                                            coupled_displacement[conn_node[I]->dof_number(coupled_U_sys.number(), 2, 0)] );
                    ASSERT_(3>I, "fatal internal error occurred");
                    ++I;
                }
                // if this (functional) vascular node is "alone" then shut it down!
                if ( 0 == I ) {
                    vi->status = VascularInfo::COLLAPSED;
                    vi->compression_scale = 1000.0;
                    //
                    const Real temp_load = 0.0;
                    const Real features[] = { current_time, 0.0, 0.0, 0.0, 0.0, temp_load,
                                              0.0, 0.0, 0.0, 0.0 };
                    //
                    // call the vessel wall remodelling routine to compute radius, thickness and pore size
                    tumour_angiogenesis::remodel_vascular_XYZ(*mi__vnr, *vi, features);
                    //
                    vi->radius = radius_min / vi->compression_scale;
                    //
                    // ignore following computations for collapsed vascular node
                    skip_calculations = true;
                }
                //
                // ignore following computations if vascular node has collapsed
                if ( skip_calculations ) continue;
                // ignore following computations if vascular element (segment) is not local
                if ( libMesh::global_processor_id() != elem->processor_id() ) continue;
                //
                // access the 3D-solid element containing the vascular node
                const Elem* elem3d = msh_3d.elem_ptr(vi->elem3d_id);
                ASSERT_(NULL!=elem3d, "fatal internal error occurred");
                //
                // get access to the stomal node
                const Node* strm_node = this->_dmn.get_mesh().node_ptr(id[1]);
                ASSERT_(NULL!=strm_node, "fatal internal error occurred");
                //
                // get access to the material info
                ASSERT_(this->_dmn.regn.end()!=this->_dmn.regn.find(elem3d->subdomain_id()), "fatal internal error occurred");
                const MaterialInfo* mi__dmn = this->_dmn.regn.find(elem3d->subdomain_id())->second.mat;
                ASSERT_(NULL!=mi__dmn, "fatal internal error occurred");
                //
                FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
                for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                    fld.biochem[ivar] = ta_biochem_species[strm_node->dof_number(rds_sys.number(), ivar, 0)];
                }
                for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                    fld.auxiliary[ivar] = ta_auxiliary_species[strm_node->dof_number(aux_sys.number(), ivar, 0)];
                }
                //
                const Real hydraulic_permeability = VascularInfo::BLOOD==vi->flow
                                                  ? ta_constitutive.blood_vessel_hydraulic_permeability(vi, mi__vnr, mi__dmn, &fld)
                                                  : ta_constitutive.lymph_vessel_hydraulic_permeability(vi, mi__vnr, mi__dmn, &fld);
                const Real vessel_length = vascular_length(node[0], node_U[0], conn_node, conn_node_U);
                // ASSERT_(vessel_length>critical_length, "fatal internal error occurred");
                const Real vessel_surface_area = vascular_circumference(vi) * vessel_length;
                //
                // ignore following computations if vascular length is negligible
                if ( vessel_length <= critical_length ) continue;
                //
                const Real osmotic_pressure_diff = VascularInfo::BLOOD==vi->flow
                                                 ? ta_constitutive.osmotic_pressure_difference(vi, mi__vnr, mi__dmn, &fld)
                                                 : 0.0;
                ASSERT_(VascularInfo::LYMPH!=vi->flow, "osmotic_pressure_difference routine not set properly for lymphatic vessels");
                //
                const Real mat = hydraulic_permeability*vessel_surface_area;
                ASSERT_(mat>0.0, "fatal internal error occurred");
                const Real rhs = hydraulic_permeability*vessel_surface_area*osmotic_pressure_diff;
                //
                system_matrix->add( node[0]->dof_number(coupled_P_sys.number(),0,0), node[0]->dof_number(coupled_P_sys.number(),0,0),
                                    -mat );
                system_matrix->add( node[0]->dof_number(coupled_P_sys.number(),0,0), node[1]->dof_number(coupled_P_sys.number(),0,0),
                                    +mat );
                system_matrix->add( node[1]->dof_number(coupled_P_sys.number(),0,0), node[0]->dof_number(coupled_P_sys.number(),0,0),
                                    +mat );
                system_matrix->add( node[1]->dof_number(coupled_P_sys.number(),0,0), node[1]->dof_number(coupled_P_sys.number(),0,0),
                                    -mat );
                //
                system_rhs->add( node[0]->dof_number(coupled_P_sys.number(),0,0),
                                 -rhs );
                system_rhs->add( node[1]->dof_number(coupled_P_sys.number(),0,0),
                                 +rhs );
                //
                // find the maximum value of the system matrix (if necessary)
                system_matrix_max_local = std::max(system_matrix_max_local, mat);
                //
            } else libmesh_error();
            //
            // end of active local elements loop...
        }
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        // evaluate the penalty parameter for the system matrix
        MPI_Allreduce(&system_matrix_max_local, &system_matrix_max_global, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
        es_coupled.parameters.set<Real>("system matrix maximum value") = system_matrix_max_global;
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        // set the penalty parameter
        const Real penalty = es_coupled.parameters.get<Real>("system matrix maximum value")
                           * par.get<Real>("penalty parameter scale");
        //
        // loop over all active local finite elements
        flog.print("  > Run fluid solver (2-B/6)          \r", false);
        ferr.print("  > Run fluid solver (2-B/6)          \r", true);
        for (MeshBase::const_element_iterator citer_el=begin_el; citer_el!=end_el; citer_el++) {
            const Elem* elem = *citer_el;
            libmesh_assert( elem->type() == libMesh::EDGE2 );
            const Node* node[] = { elem->node_ptr(0) ,
                                   elem->node_ptr(1) };
            ASSERT_(NULL!=node[0], "fatal internal error occurred");
            ASSERT_(NULL!=node[1], "fatal internal error occurred");
            const DomainData_SolidCoupledWithVascularNetwork::Type segm_type =
                static_cast<DomainData_SolidCoupledWithVascularNetwork::Type>(elem->subdomain_id());
            //
            const unsigned int n_nodes = 2;
            //
            // ignore following computations if vascular element (segment) is not local
            if ( libMesh::global_processor_id() != elem->processor_id() ) continue;
            //
            if ( DomainData_SolidCoupledWithVascularNetwork::VESSEL == segm_type ) {
                //
                for (unsigned int l=0; l<n_nodes; l++) {
                    // retrieve the information of the vascular node
                    ASSERT_(this->_dmn_coupled.cm_vasc_node__vasc_node.end()!=this->_dmn_coupled.cm_vasc_node__vasc_node.find(node[l]->id()), "fatal internal error occurred");
                    const unsigned int id = this->_dmn_coupled.cm_vasc_node__vasc_node.find(node[l]->id())->second;
                    ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(id), "fatal internal error occurred");
                    const VascularInfo* vi = &(this->_dmn_vn.node_vi.find(id)->second);
                    ASSERT_(NULL!=vi, "fatal internal error occurred");
                    //
                    if ( vasc_node_to_regress.end() != vasc_node_to_regress.find(id) ||
                         VascularInfo::COLLAPSED == vi->status ) {
                        // NOTE prescribed vascular pressure
                        const Real pressure__vascular = 0.0;
                        //
                        system_matrix->add( node[l]->dof_number(coupled_P_sys.number(),0,0), node[l]->dof_number(coupled_P_sys.number(),0,0),
                                            penalty );
                        system_rhs->add( node[l]->dof_number(coupled_P_sys.number(),0,0),
                                         penalty*pressure__vascular );
                        //
                    } else
                    if ( this->_dmn_vn.node_inlet.end() != this->_dmn_vn.node_inlet.find(id) ) {
                        ASSERT_(VascularInfo::INLET==vi->type, "fatal internal error occurred");
                        ASSERT_((VascularInfo::BLOOD==vi->flow || VascularInfo::LYMPH==vi->flow), "fatal internal error occurred");
                        // NOTE prescribed vascular pressure at the inlet of the network
                        const Real pressure__vascular = VascularInfo::BLOOD == vi->flow
                                                      ? normal_distro(mi__vnr->get_biochem_param(22), mi__vnr->get_biochem_param(23))
                                                      : normal_distro(mi__vnr->get_biochem_param(26), mi__vnr->get_biochem_param(27));
                        //
                        system_matrix->add( node[l]->dof_number(coupled_P_sys.number(),0,0), node[l]->dof_number(coupled_P_sys.number(),0,0),
                                            penalty );
                        system_rhs->add( node[l]->dof_number(coupled_P_sys.number(),0,0),
                                         penalty*pressure__vascular );
                        //
                    } else if ( this->_dmn_vn.node_outlet.end() != this->_dmn_vn.node_outlet.find(id) ) {
                        ASSERT_(VascularInfo::OUTLET==vi->type, "fatal internal error occurred");
                        ASSERT_((VascularInfo::BLOOD==vi->flow || VascularInfo::LYMPH==vi->flow), "fatal internal error occurred");
                        // NOTE prescribed vascular pressure at the outlet of the network
                        const Real pressure__vascular = VascularInfo::BLOOD == vi->flow
                                                      ? normal_distro(mi__vnr->get_biochem_param(24), mi__vnr->get_biochem_param(25))
                                                      : normal_distro(mi__vnr->get_biochem_param(28), mi__vnr->get_biochem_param(29));
                        //
                        system_matrix->add( node[l]->dof_number(coupled_P_sys.number(),0,0), node[l]->dof_number(coupled_P_sys.number(),0,0),
                                            penalty );
                        system_rhs->add( node[l]->dof_number(coupled_P_sys.number(),0,0),
                                         penalty*pressure__vascular );
                        //
                    }
                    // end of element nodes loop...
                }
                //
            } else if ( DomainData_SolidCoupledWithVascularNetwork::STROMA == segm_type ) {
                //
                for (unsigned int l=0; l<n_nodes; l++) {
                    //
                    if ( this->_dmn_coupled.cm_strm_node__boundary.end()               !=
                         this->_dmn_coupled.cm_strm_node__boundary.find(node[l]->id()) ) {
                        // NOTE prescribed interstitial fluid pressure on the boundary of the analysis domain
                        const Real pressure__interstitial_fluid = normal_distro(mi__vnr->get_biochem_param(20), mi__vnr->get_biochem_param(21));
                        //
                        system_matrix->add( node[l]->dof_number(coupled_P_sys.number(),0,0), node[l]->dof_number(coupled_P_sys.number(),0,0),
                                            penalty );
                        system_rhs->add( node[l]->dof_number(coupled_P_sys.number(),0,0),
                                         penalty*pressure__interstitial_fluid );
                        //
                    }
                    // end of element nodes loop...
                }
                //
            } else if ( DomainData_SolidCoupledWithVascularNetwork::INTERFACE == segm_type ) {
                //
                {
                    // retrieve the information of the vascular node
                    ASSERT_(this->_dmn_coupled.cm_vasc_node__vasc_node.end()!=this->_dmn_coupled.cm_vasc_node__vasc_node.find(node[0]->id()), "fatal internal error occurred");
                    const unsigned int id = this->_dmn_coupled.cm_vasc_node__vasc_node.find(node[0]->id())->second;
                    ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(id), "fatal internal error occurred");
                    const VascularInfo* vi = &(this->_dmn_vn.node_vi.find(id)->second);
                    ASSERT_(NULL!=vi, "fatal internal error occurred");
                    //
                    if ( vasc_node_to_regress.end() != vasc_node_to_regress.find(id) ||
                         VascularInfo::COLLAPSED == vi->status ) {
                        // NOTE prescribed vascular pressure
                        const Real pressure__vascular = 0.0;
                        //
                        system_matrix->add( node[0]->dof_number(coupled_P_sys.number(),0,0), node[0]->dof_number(coupled_P_sys.number(),0,0),
                                            penalty );
                        system_rhs->add( node[0]->dof_number(coupled_P_sys.number(),0,0),
                                         penalty*pressure__vascular );
                        //
                    } else
                    if ( this->_dmn_vn.node_inlet.end() != this->_dmn_vn.node_inlet.find(id) ) {
                        ASSERT_(VascularInfo::INLET==vi->type, "fatal internal error occurred");
                        ASSERT_((VascularInfo::BLOOD==vi->flow || VascularInfo::LYMPH==vi->flow), "fatal internal error occurred");
                        // NOTE prescribed vascular pressure at the inlet of the network
                        const Real pressure__vascular = VascularInfo::BLOOD == vi->flow
                                                      ? normal_distro(mi__vnr->get_biochem_param(22), mi__vnr->get_biochem_param(23))
                                                      : normal_distro(mi__vnr->get_biochem_param(26), mi__vnr->get_biochem_param(27));
                        //
                        system_matrix->add( node[0]->dof_number(coupled_P_sys.number(),0,0), node[0]->dof_number(coupled_P_sys.number(),0,0),
                                            penalty );
                        system_rhs->add( node[0]->dof_number(coupled_P_sys.number(),0,0),
                                         penalty*pressure__vascular );
                        //
                    } else if ( this->_dmn_vn.node_outlet.end() != this->_dmn_vn.node_outlet.find(id) ) {
                        ASSERT_(VascularInfo::OUTLET==vi->type, "fatal internal error occurred");
                        ASSERT_((VascularInfo::BLOOD==vi->flow || VascularInfo::LYMPH==vi->flow), "fatal internal error occurred");
                        // NOTE prescribed vascular pressure at the outlet of the network
                        const Real pressure__vascular = VascularInfo::BLOOD == vi->flow
                                                      ? normal_distro(mi__vnr->get_biochem_param(24), mi__vnr->get_biochem_param(25))
                                                      : normal_distro(mi__vnr->get_biochem_param(28), mi__vnr->get_biochem_param(29));
                        //
                        system_matrix->add( node[0]->dof_number(coupled_P_sys.number(),0,0), node[0]->dof_number(coupled_P_sys.number(),0,0),
                                            penalty );
                        system_rhs->add( node[0]->dof_number(coupled_P_sys.number(),0,0),
                                         penalty*pressure__vascular );
                        //
                    }
                    //
                }
                //
                {
                    //
                    if ( this->_dmn_coupled.cm_strm_node__boundary.end()               !=
                         this->_dmn_coupled.cm_strm_node__boundary.find(node[1]->id()) ) {
                        // NOTE prescribed interstitial fluid pressure on the boundary of the analysis domain
                        const Real pressure__interstitial_fluid = normal_distro(mi__vnr->get_biochem_param(20), mi__vnr->get_biochem_param(21));
                        //
                        system_matrix->add( node[1]->dof_number(coupled_P_sys.number(),0,0), node[1]->dof_number(coupled_P_sys.number(),0,0),
                                            penalty );
                        system_rhs->add( node[1]->dof_number(coupled_P_sys.number(),0,0),
                                         penalty*pressure__interstitial_fluid );
                        //
                    }
                    //
                }
                //
            } else libmesh_error();
            // end of active local elements loop...
        }
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        //
        system_matrix->close();
        system_rhs->close();
        // end of do-while loop...
    } while ( false );
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    this->_log.pop("_fluid_solver (TASK #2)");

    // Solve the linear system of the coupled problem.
    this->_log.push("_fluid_solver (TASK #3)");
    flog.print("  > Run fluid solver (3/6)          \r", false);
    ferr.print("  > Run fluid solver (3/6)          \r", true);
    do {
        EquationSystems& es_coupled = this->_dmn_coupled.get_equation_systems();
        LinearImplicitSystem& coupled_P_sys = es_coupled.get_system<LinearImplicitSystem>(this->_label+"::coupled::pressure_solution");
        //
        // now solve the system
        coupled_P_sys.linear_solver->reuse_preconditioner(false);
        coupled_P_sys.solve();
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        // a sanity check (make sure that the max value is well below the max pressure applied in the problem)
        if ( coupled_P_sys.solution->linfty_norm() >= 1.0e+12 ) {
            WARNING_("an unexpected error has occurred in the fluid solver");
            ++number_of_reruns;
            //ASSERT_(4>number_of_reruns, "fatal internal error occurred");
            if ( number_of_reruns < 4 ) {
                goto rerun_fluid_solver;
            } else if ( number_of_reruns == 4 ) {
                this->_update_network();
                number_of_reruns = 0;
                goto rerun_fluid_solver;
            } else libmesh_error();
        }
        // end of do-while loop...
    } while ( false );
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    this->_log.pop("_fluid_solver (TASK #3)");

    // Project the vascular pressure into the vascular network and compute blood flow characteristics
    // such as fluid velocity and wall shear stress (WSS).
    this->_log.push("_fluid_solver (TASK #4)");
    flog.print("  > Run fluid solver (4/6)          \r", false);
    ferr.print("  > Run fluid solver (4/6)          \r", true);
    do {
        const EquationSystems& es_coupled = this->_dmn_coupled.get_equation_systems();
        const System& coupled_P_sys = es_coupled.get_system<System>(this->_label+"::coupled::pressure_solution");
        ASSERT_(1==coupled_P_sys.n_vars(), "fatal internal error occurred");
        //
        EquationSystems& es_vn = this->_dmn_vn.get_equation_systems();
        System& vn_flow_sys = es_vn.get_system<System>(this->_label+"::flow_solution");
        ASSERT_(8==vn_flow_sys.n_vars(), "fatal internal error occurred");
        //
        EquationSystems& es_3d = this->_dmn.get_equation_systems();
        System& s3d_cpl_sys = es_3d.get_system<System>(NavierCauchySolver3D::get_label()+"::"+this->_label+"::coupled_solution");
        ASSERT_(9==s3d_cpl_sys.n_vars(), "fatal internal error occurred");
        //
        const MeshBase& msh_coupled = es_coupled.get_mesh();
        libmesh_assert( SP1D == msh_coupled.mesh_dimension() );
        //
        const MeshBase& msh_3d = this->_dmn.get_mesh();
        libmesh_assert( SP3D == msh_3d.mesh_dimension() );
        //
        const MeshBase& msh_vn = this->_dmn_vn.get_mesh();
        libmesh_assert( SP1D == msh_vn.mesh_dimension() );
        //
        flog.print("  > Run fluid solver (4-A/6)          \r", false);
        ferr.print("  > Run fluid solver (4-A/6)          \r", true);
        std::vector<Number> coupled_pressure;
        coupled_P_sys.update_global_solution(coupled_pressure);
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        //
        // set the computed (from the fluid solver) vascular pressure for the vascular network system
        flog.print("  > Run fluid solver (4-B/6)          \r", false);
        ferr.print("  > Run fluid solver (4-B/6)          \r", true);
        for ( std::map<dof_id_type, VascularInfo>::iterator
              i_ndvi=this->_dmn_vn.node_vi.begin(); i_ndvi!=this->_dmn_vn.node_vi.end(); i_ndvi++ )
        {
            VascularInfo* vi = &(i_ndvi->second);
            ASSERT_(NULL!=vi, "fatal internal error occurred");
            const Elem* elem3d = msh_3d.elem_ptr(vi->elem3d_id);
            ASSERT_(NULL!=elem3d, "fatal internal error occurred");
            //
            // get access to the material info
            ASSERT_(this->_dmn.regn.end()!=this->_dmn.regn.find(elem3d->subdomain_id()), "fatal internal error occurred");
            const MaterialInfo* mi__dmn = this->_dmn.regn.find(elem3d->subdomain_id())->second.mat;
            ASSERT_(NULL!=mi__dmn, "fatal internal error occurred");
            //
            const Node* vn_node = msh_vn.node_ptr( i_ndvi->first );
            ASSERT_(NULL!=vn_node, "fatal internal error occurred");
            libmesh_assert( this->_dmn_coupled.vasc_node__cm_vasc_node.end() != this->_dmn_coupled.vasc_node__cm_vasc_node.find(vn_node->id()) );
            const Node* coupled_node = msh_coupled.node_ptr( this->_dmn_coupled.vasc_node__cm_vasc_node.find(vn_node->id())->second );
            ASSERT_(NULL!=coupled_node, "fatal internal error occurred");
            //
            if ( coupled_pressure[coupled_node->dof_number(coupled_P_sys.number(), 0, 0)] > 1.0e+12 )
                libmesh_error();
            if ( coupled_pressure[coupled_node->dof_number(coupled_P_sys.number(), 0, 0)] < 0.0e+00 )
                libmesh_error();
            //
            if ( vasc_node_to_regress.end() != vasc_node_to_regress.find(vn_node->id()) ) {
                vi->status = VascularInfo::COLLAPSED;
                vi->compression_scale = 1000.0;
            }
            //
            // evaluate the cross-sectional area of the lumen and its diameter
            const Real vessel_cross_sectional_area = vascular_cross_section(vi),
                       vessel_diameter = sqrt(4.0*vessel_cross_sectional_area/mPI);
            //
            const Node* coupled_conn_node[] = { NULL, NULL, NULL };
            Point coupled_conn_node_U[] = { Point(0.0,0.0,0.0), Point(0.0,0.0,0.0), Point(0.0,0.0,0.0) };
            Real coupled_conn_node_P[] = { 0.0, 0.0, 0.0 };
            //
            unsigned int n_conn_functional_vsc_nodes = 0;
            unsigned int I = 0;
            if ( VascularInfo::COLLAPSED != vi->status )
            for ( std::set<dof_id_type>::const_iterator
                  ci=vi->vsc_elem.begin(); ci!=vi->vsc_elem.end(); ci++ )
            {
                const Elem& elem = this->_dmn_vn.get_mesh().elem_ref(*ci);
                const dof_id_type conn_node_id = edge2_other_node(elem, vn_node->id());
                //
                ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(conn_node_id), "fatal internal error occurred");
                const VascularInfo* conn_vi = &(this->_dmn_vn.node_vi.find(conn_node_id)->second);
                ASSERT_(NULL!=conn_vi, "fatal internal error occurred");
                //
                if ( VascularInfo::COLLAPSED == conn_vi->status ) continue;
                ASSERT_(3>I, "fatal internal error occurred");
                //
                coupled_conn_node[I] = msh_coupled.node_ptr( this->_dmn_coupled.vasc_node__cm_vasc_node.find(conn_node_id)->second );
                ASSERT_(NULL!=coupled_conn_node[I], "fatal internal error occurred");
                coupled_conn_node_U[I] = Point( coupled_displacement[coupled_conn_node[I]->dof_number(coupled_U_sys.number(), 0, 0)] ,
                                                coupled_displacement[coupled_conn_node[I]->dof_number(coupled_U_sys.number(), 1, 0)] ,
                                                coupled_displacement[coupled_conn_node[I]->dof_number(coupled_U_sys.number(), 2, 0)] );
                coupled_conn_node_P[I] = coupled_pressure[coupled_conn_node[I]->dof_number(coupled_P_sys.number(), 0, 0)];
                ++I;
                ++n_conn_functional_vsc_nodes;
            }
            // if this (functional) vascular node is "alone" then shut it down!
            if ( 0 == I ) {
                vi->status = VascularInfo::COLLAPSED;
                vi->compression_scale = 1000.0;
            }
            //
            // some field-values variable
            FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
            //
            // obtain the "displacement" vector field at the vascular node
            const Point coupled_node_Displacement = Point( coupled_displacement[coupled_node->dof_number(coupled_U_sys.number(), 0, 0)] ,
                                                           coupled_displacement[coupled_node->dof_number(coupled_U_sys.number(), 1, 0)] ,
                                                           coupled_displacement[coupled_node->dof_number(coupled_U_sys.number(), 2, 0)] );
            //
            // obtain the "microvascular pressure" at the vascular node
            const Real coupled_node_Pressure = VascularInfo::COLLAPSED == vi->status
                                             ? 0.0
                                             : coupled_pressure[coupled_node->dof_number(coupled_P_sys.number(), 0, 0)];
            //
            // set the "dynamic viscosity" at the vascular node
            const Real coupled_node_Viscosity = VascularInfo::BLOOD == vi->flow
                                              ? blood_viscosity(mi__vnr)
                                              : lymph_viscosity(mi__vnr);
            ASSERT_(coupled_node_Viscosity>1.0e-6, "fatal internal error occurred");
            //
            // calculate the "intra-vascular flow rate"
            Real coupled_node_Q_VASC = 0.0;
            // loop for all connective vascular nodes to this vascular node
            if ( VascularInfo::COLLAPSED != vi->status )
            for (unsigned int l=0; l<n_conn_functional_vsc_nodes; l++) {
                ASSERT_(NULL!=coupled_conn_node[l], "fatal internal error occurred");
                //
                const Real hydraulic_conductivity = VascularInfo::BLOOD == vi->flow
                                                  ? ta_constitutive.blood_vessel_hydraulic_conductivity(vi, mi__vnr, mi__dmn, &fld)
                                                  : ta_constitutive.lymph_vessel_hydraulic_conductivity(vi, mi__vnr, mi__dmn, &fld);
                const Real pressure_gradient = (coupled_node_Pressure - coupled_conn_node_P[l])
                                             / Point((*coupled_node)+coupled_node_Displacement
                                                    -(*coupled_conn_node[l])-coupled_conn_node_U[l]).norm();
                //
                if ( 1 == n_conn_functional_vsc_nodes ) {
                    coupled_node_Q_VASC += hydraulic_conductivity*vessel_cross_sectional_area*fabs(pressure_gradient);
                } else {
                    if ( pressure_gradient < 0.0 )
                        coupled_node_Q_VASC -= hydraulic_conductivity*vessel_cross_sectional_area*pressure_gradient;
                }
                // end of connective vascular nodes loop...
            }
            //
            // calculate the "mean velocity" at the vascular node
            const Real coupled_node_MeanVelocity = VascularInfo::COLLAPSED == vi->status
                                                 ? 0.0
                                                 : fabs(coupled_node_Q_VASC)/vessel_cross_sectional_area;
            //
            // calculate the "wall shear stress" at the vascular node
            const Real coupled_node_WSS = VascularInfo::COLLAPSED == vi->status
                                        ? 0.0
                                        : (8.0*coupled_node_MeanVelocity*coupled_node_Viscosity/vessel_diameter);
            //
            // interpolate to obtain the "hydrostatic pressure" at the vascular node
            Real coupled_node_SHP = 0.0;
            {
                // get access to the element quadrature points data
                ASSERT_(elem_qpi->end()!=elem_qpi->find(elem3d->id()), "fatal internal error occurred");
                const QuadratureInfo* qpi_ptr = &(elem_qpi->find(elem3d->id())->second);
                ASSERT_(NULL!=qpi_ptr, "fatal internal error occurred");
                //
                InverseDistanceAlgorithm ida(2.5);
                ida.init(qpi_ptr->XYZ);
                ida.interp<Real>(*vn_node, qpi_ptr->pressure, coupled_node_SHP);
            }
            //
            // interpolate to obtain the "interstitial fluid pressure" at the vascular node
            Real coupled_node_IFP = 0.0;
            {
                std::vector<Point> nodal_X(elem3d->n_nodes());
                std::vector<Real> nodal_IFP(elem3d->n_nodes());
                // retrieve for all 'stroma' nodes the "interstitial fluid pressure" value
                for (unsigned int l=0; l<elem3d->n_nodes(); l++) {
                    const Node* node__elem3d = elem3d->node_ptr(l);
                    ASSERT_(NULL!=node__elem3d, "fatal internal error occurred");
                    //
                    ASSERT_(this->_dmn_coupled.strm_node__cm_strm_node.end()!=this->_dmn_coupled.strm_node__cm_strm_node.find(node__elem3d->id()), "fatal internal error occurred");
                    const dof_id_type coupled_nid = this->_dmn_coupled.strm_node__cm_strm_node.find(node__elem3d->id())->second;
                    const Node* node__coupled = msh_coupled.node_ptr(coupled_nid);
                    ASSERT_(NULL!=node__coupled, "fatal internal error occurred");
                    //
                    nodal_X[l] = (*node__elem3d);
                    nodal_IFP[l] = coupled_pressure[node__coupled->dof_number(coupled_P_sys.number(), 0, 0)];
                }
                //
                InverseDistanceAlgorithm ida(2.5);
                ida.init(nodal_X);
                ida.interp<Real>(*vn_node, nodal_IFP, coupled_node_IFP);
            }
            //
            // calculate the "effective vascular pressure"
            const Real coupled_node_PressureEffective = VascularInfo::BLOOD == vi->flow
                                                      ? coupled_node_Pressure - ta_constitutive.osmotic_pressure_difference(vi, mi__vnr, mi__dmn, &fld)
                                                      : coupled_node_Pressure;
            //
            // calculate the "trans-vascular flow rate"
            Real coupled_node_Q_TRAN = 0.0;
            // loop for all connective interstitial nodes to this vascular node
            if ( VascularInfo::COLLAPSED != vi->status )
            {
                //
                const Real hydraulic_permeability = VascularInfo::BLOOD == vi->flow
                                                  ? ta_constitutive.blood_vessel_hydraulic_permeability(vi, mi__vnr, mi__dmn, &fld)
                                                  : ta_constitutive.lymph_vessel_hydraulic_permeability(vi, mi__vnr, mi__dmn, &fld);
                const Real vessel_surface_area = vascular_circumference(vi)
                                               * vascular_length(coupled_node, coupled_node_Displacement, coupled_conn_node, coupled_conn_node_U);
                const Real pressure_difference = coupled_node_PressureEffective - coupled_node_IFP;
                //
                coupled_node_Q_TRAN = hydraulic_permeability*vessel_surface_area*pressure_difference;
                // end of connective interstitial nodes loop...
            }
            //
            // set the "blood pressure"
            set_system_solution_nodal_value<0>(vn_flow_sys, vn_node, coupled_node_Pressure);
            // set the "blood viscosity"
            set_system_solution_nodal_value<1>(vn_flow_sys, vn_node, coupled_node_Viscosity);
            // set the "wall shear stress"
            set_system_solution_nodal_value<2>(vn_flow_sys, vn_node, coupled_node_WSS);
            // set the "solid hydrostatic pressure"
            set_system_solution_nodal_value<3>(vn_flow_sys, vn_node, coupled_node_SHP);
            // set the "interstitial fluid pressure"
            set_system_solution_nodal_value<4>(vn_flow_sys, vn_node, coupled_node_IFP);
            // set the "trans-vascular flow rate"
            set_system_solution_nodal_value<5>(vn_flow_sys, vn_node, coupled_node_Q_TRAN);
            // set the "mean velocity"
            set_system_solution_nodal_value<6>(vn_flow_sys, vn_node, coupled_node_MeanVelocity);
            // set the "effective pressure"
            set_system_solution_nodal_value<7>(vn_flow_sys, vn_node, coupled_node_PressureEffective);
            // end of vascular nodes loop...
        }
        //
        // close the solution container and update the system
        flog.print("  > Run fluid solver (4-C/6)          \r", false);
        ferr.print("  > Run fluid solver (4-C/6)          \r", true);
        vn_flow_sys.solution->close();
        vn_flow_sys.update();
        //
        // set the computed (from the fluid solver) interstitial pressure for the solid system
        flog.print("  > Run fluid solver (4-D/6)          \r", false);
        ferr.print("  > Run fluid solver (4-D/6)          \r", true);
        for ( MeshBase::const_node_iterator
              n_citer=msh_3d.active_nodes_begin(); n_citer!=msh_3d.active_nodes_end(); n_citer++ )
        {
            // node @stroma
            const Node* node = *n_citer;
            //
            ASSERT_(this->_dmn_coupled.strm_node__cm_strm_node.end()!=this->_dmn_coupled.strm_node__cm_strm_node.find(node->id()), "fatal internal error occurred");
            const dof_id_type coupled_nid = this->_dmn_coupled.strm_node__cm_strm_node.find(node->id())->second;
            const Node* coupled_strm_node = msh_coupled.node_ptr(coupled_nid);
            ASSERT_(NULL!=coupled_strm_node, "fatal internal error occurred");
            //
            const Real coupled_strm_node_P = coupled_pressure[coupled_strm_node->dof_number(coupled_P_sys.number(), 0, 0)];
            //
            const Point coupled_strm_node_U = Point( coupled_displacement[coupled_strm_node->dof_number(coupled_U_sys.number(), 0, 0)] ,
                                                     coupled_displacement[coupled_strm_node->dof_number(coupled_U_sys.number(), 1, 0)] ,
                                                     coupled_displacement[coupled_strm_node->dof_number(coupled_U_sys.number(), 2, 0)] );
            //
            Point coupled_strm_node_V_TRAN;
            Real average_vessel_surface_area = 0.0;
            Point coupled_strm_node_V_INTE;
            Real average_tissue_cross_sectional_area = 0.0;
            //
            //
            std::map<dof_id_type, std::set<pair_dof_id_type> >::const_iterator mci__0 =
                this->_dmn_coupled.cm_strm_node__cm_vasc_node_segm.find(coupled_strm_node->id());
            std::map<dof_id_type, std::set<pair_dof_id_type> >::const_iterator mci__1 =
                this->_dmn_coupled.cm_strm_node__cm_strm_node_segm.find(coupled_strm_node->id());
            //
            //
            if ( this->_dmn_coupled.cm_strm_node__cm_vasc_node_segm.end() != mci__0 ) {
            // start of if case...
            int total_neighbor_vascular_nodes = 0;
            const std::set<pair_dof_id_type>& cm_vasc_node_segm = mci__0->second;
            for ( std::set<pair_dof_id_type>::const_iterator
                  ci=cm_vasc_node_segm.begin(); ci!=cm_vasc_node_segm.end(); ci++ )
            {
                const Node* coupled_vasc_node = msh_coupled.node_ptr(ci->first);
                ASSERT_(NULL!=coupled_vasc_node, "fatal internal error occurred");
                //
                const Real coupled_vasc_node_P = coupled_pressure[coupled_vasc_node->dof_number(coupled_P_sys.number(), 0, 0)];
                //
                const Point coupled_vasc_node_U = Point( coupled_displacement[coupled_vasc_node->dof_number(coupled_U_sys.number(), 0, 0)] ,
                                                         coupled_displacement[coupled_vasc_node->dof_number(coupled_U_sys.number(), 1, 0)] ,
                                                         coupled_displacement[coupled_vasc_node->dof_number(coupled_U_sys.number(), 2, 0)] );
                //
                ASSERT_(this->_dmn_coupled.cm_vasc_node__vasc_node.end()!=this->_dmn_coupled.cm_vasc_node__vasc_node.find(ci->first), "fatal internal error occurred");
                const unsigned int vn_id = this->_dmn_coupled.cm_vasc_node__vasc_node.find(ci->first)->second;
                ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(vn_id), "fatal internal error occurred");
                const VascularInfo* vi = &(this->_dmn_vn.node_vi.find(vn_id)->second);
                ASSERT_(NULL!=vi, "fatal internal error occurred");
                //
                if ( VascularInfo::COLLAPSED == vi->status ) continue;
                // get access to the material info
                ASSERT_(NULL!=msh_3d.elem_ptr(vi->elem3d_id), "fatal internal error occurred");
                const subdomain_id_type rgn_id = msh_3d.elem_ptr(vi->elem3d_id)->subdomain_id();
                ASSERT_(this->_dmn.regn.end()!=this->_dmn.regn.find(rgn_id), "fatal internal error occurred");
                const MaterialInfo* mi__dmn = this->_dmn.regn.find(rgn_id)->second.mat;
                ASSERT_(NULL!=mi__dmn, "fatal internal error occurred");
                //
                const Node* coupled_conn_vasc_node[] = { NULL, NULL, NULL };
                Point coupled_conn_vasc_node_U[] = { Point(0.0,0.0,0.0), Point(0.0,0.0,0.0), Point(0.0,0.0,0.0) };
                //
                unsigned int I = 0;
                for (std::set<dof_id_type>::const_iterator ci=vi->vsc_elem.begin(); ci!=vi->vsc_elem.end(); ci++) {
                    const Elem& elem = this->_dmn_vn.get_mesh().elem_ref(*ci);
                    const dof_id_type conn_node_id = edge2_other_node(elem, vn_id);
                    //
                    ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(conn_node_id), "fatal internal error occurred");
                    const VascularInfo* conn_vi = &(this->_dmn_vn.node_vi.find(conn_node_id)->second);
                    ASSERT_(NULL!=conn_vi, "fatal internal error occurred");
                    ASSERT_(3>I, "fatal internal error occurred");
                    //
                    if ( VascularInfo::COLLAPSED == conn_vi->status ) continue;
                    //
                    coupled_conn_vasc_node[I] = msh_coupled.node_ptr( this->_dmn_coupled.vasc_node__cm_vasc_node.find(conn_node_id)->second );
                    ASSERT_(NULL!=coupled_conn_vasc_node[I], "fatal internal error occurred");
                    coupled_conn_vasc_node_U[I] = Point( coupled_displacement[coupled_conn_vasc_node[I]->dof_number(coupled_U_sys.number(), 0, 0)] ,
                                                         coupled_displacement[coupled_conn_vasc_node[I]->dof_number(coupled_U_sys.number(), 1, 0)] ,
                                                         coupled_displacement[coupled_conn_vasc_node[I]->dof_number(coupled_U_sys.number(), 2, 0)] );
                    ++I;
                }
                //
                FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
                for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                    fld.biochem[ivar] = ta_biochem_species[node->dof_number(rds_sys.number(), ivar, 0)];
                }
                for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                    fld.auxiliary[ivar] = ta_auxiliary_species[node->dof_number(aux_sys.number(), ivar, 0)];
                }
                //
                const Real hydraulic_permeability = VascularInfo::BLOOD == vi->flow
                                                  ? ta_constitutive.blood_vessel_hydraulic_permeability(vi, mi__vnr, mi__dmn, &fld)
                                                  : ta_constitutive.lymph_vessel_hydraulic_permeability(vi, mi__vnr, mi__dmn, &fld);
                const Real vessel_surface_area = vascular_circumference(vi)
                                               * vascular_length(coupled_vasc_node, coupled_vasc_node_U, coupled_conn_vasc_node, coupled_conn_vasc_node_U);
                const Real pressure_difference = coupled_vasc_node_P - coupled_strm_node_P
                                               - ta_constitutive.osmotic_pressure_difference(vi, mi__vnr, mi__dmn, &fld);
                //
                const Point vector = Point((*coupled_strm_node)+coupled_strm_node_U
                                          -(*coupled_vasc_node)-coupled_vasc_node_U).unit();
                //
                coupled_strm_node_V_TRAN += (hydraulic_permeability*pressure_difference) * vector;
                //
                average_vessel_surface_area += vessel_surface_area;
                ++total_neighbor_vascular_nodes;
            }
            // now average the variable
            if ( total_neighbor_vascular_nodes )
                average_vessel_surface_area /= total_neighbor_vascular_nodes;
            //
            // end of if case...
            } else average_vessel_surface_area = 1.0;
            //
            //
            if ( this->_dmn_coupled.cm_strm_node__cm_strm_node_segm.end() != mci__1 ) {
            // start of if case...
            int total_neighbor_stromal_nodes = 0;
            const std::set<pair_dof_id_type>& cm_strm_node_segm = mci__1->second;
            for ( std::set<pair_dof_id_type>::const_iterator
                  ci=cm_strm_node_segm.begin(); ci!=cm_strm_node_segm.end(); ci++ )
            {
                const Node* coupled_strm_conn_node = msh_coupled.node_ptr(ci->first);
                ASSERT_(NULL!=coupled_strm_conn_node, "fatal internal error occurred");
                //
                const Real coupled_strm_conn_node_P = coupled_pressure[coupled_strm_conn_node->dof_number(coupled_P_sys.number(), 0, 0)];
                //
                const Point coupled_strm_conn_node_U = Point( coupled_displacement[coupled_strm_conn_node->dof_number(coupled_U_sys.number(), 0, 0)],
                                                              coupled_displacement[coupled_strm_conn_node->dof_number(coupled_U_sys.number(), 1, 0)],
                                                              coupled_displacement[coupled_strm_conn_node->dof_number(coupled_U_sys.number(), 2, 0)] );
                // get access to the material info
                ASSERT_(this->_dmn_coupled.cm_strm_elem__subdmnid.end()!=this->_dmn_coupled.cm_strm_elem__subdmnid.find(ci->second), "fatal internal error occurred");
                const subdomain_id_type rgn_id = this->_dmn_coupled.cm_strm_elem__subdmnid.find(ci->second)->second;
                ASSERT_(this->_dmn.regn.end()!=this->_dmn.regn.find(rgn_id), "fatal internal error occurred");
                const MaterialInfo* mi__dmn = this->_dmn.regn.find(rgn_id)->second.mat;
                ASSERT_(NULL!=mi__dmn, "fatal internal error occurred");
                //
                // get access to the stomal node
                ASSERT_(this->_dmn_coupled.cm_strm_node__strm_node.end()!=this->_dmn_coupled.cm_strm_node__strm_node.find(coupled_strm_conn_node->id()), "fatal internal error occurred");
                const Node* strm_conn_node = msh_3d.node_ptr( this->_dmn_coupled.cm_strm_node__strm_node.find(coupled_strm_conn_node->id())->second );
                ASSERT_(NULL!=strm_conn_node, "fatal internal error occurred");
                //
                FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
                for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                    fld.biochem[ivar] = ta_biochem_species[strm_conn_node->dof_number(rds_sys.number(), ivar, 0)];
                }
                for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                    fld.auxiliary[ivar] = ta_auxiliary_species[strm_conn_node->dof_number(aux_sys.number(), ivar, 0)];
                }
                //
                const Point dir = ((*coupled_strm_conn_node) - (*coupled_strm_node)).unit();
                const Dyadic3D hydr_cond_tensor = ta_constitutive.stroma_hydraulic_conductivity(mi__vnr, mi__dmn, &fld);
                //
                const Real hydraulic_conductivity = (hydr_cond_tensor*dir).norm();
                const Real tissue_cross_sectional_area = ta_constitutive.stroma_cross_section(mi__vnr, mi__dmn, &fld);
                const Real pressure_difference = coupled_strm_conn_node_P - coupled_strm_node_P;
                //
                const Point vector = Point((*coupled_strm_node)+coupled_strm_node_U-(*coupled_strm_conn_node)-coupled_strm_conn_node_U).unit();
                const Real segm_length = Point((*coupled_strm_node)+coupled_strm_node_U-(*coupled_strm_conn_node)-coupled_strm_conn_node_U).norm();
                //
                coupled_strm_node_V_INTE += (hydraulic_conductivity*pressure_difference/segm_length) * vector;
                //
                average_tissue_cross_sectional_area += tissue_cross_sectional_area;
                ++total_neighbor_stromal_nodes;
            }
            //
            if ( total_neighbor_stromal_nodes )
                average_tissue_cross_sectional_area /= total_neighbor_stromal_nodes;
            //
            // end of if case...
            } else libmesh_error();
            //
            //
            const Real coupled_strm_node_Q_TRAN = average_vessel_surface_area > 0.0
                                                ? coupled_strm_node_V_TRAN.norm()/average_vessel_surface_area : 0.0;
            const Real coupled_strm_node_Q_INTE = average_tissue_cross_sectional_area > 0.0
                                                ? coupled_strm_node_V_INTE.norm()/average_tissue_cross_sectional_area : 0.0;
            //
            // set the "interstitial fluid pressure"
            set_system_solution_nodal_value<0>(s3d_cpl_sys, node, coupled_strm_node_P);
            // set the "interstitial fluid velocity"
            set_system_solution_nodal_value<1>(s3d_cpl_sys, node, coupled_strm_node_V_INTE(0));
            set_system_solution_nodal_value<2>(s3d_cpl_sys, node, coupled_strm_node_V_INTE(1));
            set_system_solution_nodal_value<3>(s3d_cpl_sys, node, coupled_strm_node_V_INTE(2));
            // set the "interstitial flow rate"
            set_system_solution_nodal_value<4>(s3d_cpl_sys, node, coupled_strm_node_Q_INTE);
            // set the "trans-vascular fluid velocity"
            set_system_solution_nodal_value<5>(s3d_cpl_sys, node, coupled_strm_node_V_TRAN(0));
            set_system_solution_nodal_value<6>(s3d_cpl_sys, node, coupled_strm_node_V_TRAN(1));
            set_system_solution_nodal_value<7>(s3d_cpl_sys, node, coupled_strm_node_V_TRAN(2));
            // set the "trans-vascular flow rate"
            set_system_solution_nodal_value<8>(s3d_cpl_sys, node, coupled_strm_node_Q_TRAN);
            // end of active nodes loop...
        }
        //
        // close the solution container and update the system
        flog.print("  > Run fluid solver (4-E/6)          \r", false);
        ferr.print("  > Run fluid solver (4-E/6)          \r", true);
        s3d_cpl_sys.solution->close();
        s3d_cpl_sys.update();
        // end of do-while loop...
    } while ( false );
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    this->_log.pop("_fluid_solver (TASK #4)");

    // Initialise some important quadrature information, i.e. project from the coupled-system solution some field
    // variables associated with the fluid mechanics module of the model.
    this->_log.push("_fluid_solver (TASK #5)");
    flog.print("  > Run fluid solver (5/6)          \r", false);
    ferr.print("  > Run fluid solver (5/6)          \r", true);
    do {
        const EquationSystems& es_3d = this->_dmn.get_equation_systems();
        const System& s3d_cpl_sys = es_3d.get_system<System>(NavierCauchySolver3D::get_label()+"::"+this->_label+"::coupled_solution");
        ASSERT_(9==s3d_cpl_sys.n_vars(), "fatal internal error occurred");
        //
        const MeshBase& msh_3d = this->_dmn.get_mesh();
        libmesh_assert( SP3D == msh_3d.mesh_dimension() );
        //
        flog.print("  > Run fluid solver (5-A/6)          \r", false);
        ferr.print("  > Run fluid solver (5-A/6)          \r", true);
        //
        FEType fe_sc_type = s3d_cpl_sys.variable_type(0);
        //
        flog.print("  > Run fluid solver (5-B/6)          \r", false);
        ferr.print("  > Run fluid solver (5-B/6)          \r", true);
        // iterate for all active local elements
        for ( MeshBase::const_element_iterator
              citer_el=msh_3d.active_local_elements_begin(); citer_el!=msh_3d.active_local_elements_end(); citer_el++ )
        {
            const Elem* elem = *citer_el;
            ASSERT_(NULL!=elem, "fatal internal error occurred");
            //
            QGauss quad_volu(SP3D, par.get<Order>("Gauss quadrature order"));
            UniquePtr<FEBase> fe_sc_volu = FEBase::build(SP3D, fe_sc_type);
            fe_sc_volu->attach_quadrature_rule(&quad_volu);
            //
            const std::vector< std::vector<Real> >& Phi = fe_sc_volu->get_phi();
            fe_sc_volu->reinit(elem);
            //
            libmesh_assert( this->_dmn.elem_qi_local.end() != this->_dmn.elem_qi_local.find(elem->id()) );
            QuadratureInfo* qpi_ptr = &(this->_dmn.elem_qi_local.find(elem->id())->second);
            ASSERT_(NULL!=qpi_ptr, "fatal internal error occurred");
            //
            // nodal values of the interstitial fluid velocity
            std::vector<Point> IFV(elem->n_nodes());
            for (unsigned int l=0; l<elem->n_nodes(); l++) {
                const Node* node = elem->node_ptr(l);
                ASSERT_(NULL!=node, "fatal internal error occurred");
                //
                IFV[l](0) = s3d_cpl_sys.current_local_solution->el(node->dof_number(s3d_cpl_sys.number(), 1, 0));
                IFV[l](1) = s3d_cpl_sys.current_local_solution->el(node->dof_number(s3d_cpl_sys.number(), 2, 0));
                IFV[l](2) = s3d_cpl_sys.current_local_solution->el(node->dof_number(s3d_cpl_sys.number(), 3, 0));
            }
            //
            // loop over volume integration quadrature points
            for (unsigned int qp=0; qp<quad_volu.n_points(); qp++) {
                //
                qpi_ptr->fluid_velocity[qp] = Vector3D();
                for (unsigned int l=0; l<elem->n_nodes(); l++) {
                    qpi_ptr->fluid_velocity[qp].comp(0) += Phi[l][qp] * IFV[l](0);
                    qpi_ptr->fluid_velocity[qp].comp(1) += Phi[l][qp] * IFV[l](1);
                    qpi_ptr->fluid_velocity[qp].comp(2) += Phi[l][qp] * IFV[l](2);
                }
                // end of volume integration quadrature loop...
            }
            // end of active local elements loop...
        }
        // end of do-while loop...
    } while ( false );
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    this->_log.pop("_fluid_solver (TASK #5)");

    this->_log.push("_fluid_solver (TASK #6)");
    flog.print("  > Run fluid solver (6/6)          \r", false);
    ferr.print("  > Run fluid solver (6/6)          \r", true);
    // set the other systems (mainly the reaction-diffusion solver and some solution containers)
    // of the coupled system, etc.
    this->_set_other_systems();
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    this->_log.pop("_fluid_solver (TASK #6)");

    flog.print("  > Run fluid solver                  [ OK ]\r", false);

    this->_log.pop("_fluid_solver");
}
//=================================================================================================
