/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/tumour_angiogenesis_solver.h"
#include "feb3/quadrature.h"
#include "feb3/ida.h"
#include "libmesh/gmsh_io.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
// local variables
static
std::vector<Number> vn_flow_features;
static
std::vector<Number> rds_discontinuous;
static
std::vector<Number> drug_concentration;
static
std::set<dof_id_type> vasc_node_2_iterate,
                      vasc_node_2_notiterate;
static
std::map<dof_id_type, QuadratureInfo>* elem3d_qpi;
static
std::map<dof_id_type, std::map<unsigned int, QuadratureInfo> >* elem2d_qpi;
static
std::map<dof_id_type, Point>              new_node_to_add;
static
std::map<dof_id_type, Dual<dof_id_type> > new_elem_to_add;
static
std::map<dof_id_type, VascularInfo>       new_node_vi_to_add;
static
std::map<dof_id_type, Point>              existing_node_to_update;
static
std::map<dof_id_type, Dual<dof_id_type> > existing_elem_to_update;
static
std::map<dof_id_type, VascularInfo>       existing_node_vi_to_update;
//=================================================================================================
#include "feb3/domain_data_inl.h"
#include "feb3/solution_io_inl.h"
#include "feb3/vascular_network_aux_inl.h"
#include "feb3/vascular_network_io_inl.h"
//=================================================================================================
void TumourAngiogenesisSolver::_update_network () {
    this->_log.push("_update_network");

    this->_log.push("_update_network #1");
    flog.print("  > Update vascular network... 1/5                                     \r", false);
    ferr.print("  > Update vascular network... 1/5                                     \r", true );
    this->_prepare_vascular_update();
    this->_log.pop("_update_network #1");

    this->_log.push("_update_network #2");
    flog.print("  > Update vascular network... 2A/5                                    \r", false);
    ferr.print("  > Update vascular network... 2A/5                                    \r", true );
    this->_find_anastomoses();
    flog.print("  > Update vascular network... 2B/5                                    \r", false);
    ferr.print("  > Update vascular network... 2B/5                                    \r", true );
    this->_perform_sprouting();
    this->_log.pop("_update_network #2");

    this->_log.push("_update_network #3");
    flog.print("  > Update vascular network... 3A/5                                    \r", false);
    ferr.print("  > Update vascular network... 3A/5                                    \r", true );
    this->_distribute_results();
    flog.print("  > Update vascular network... 3B/5                                    \r", false);
    ferr.print("  > Update vascular network... 3B/5                                    \r", true );
    this->_update_domain_data();
    this->_log.pop("_update_network #3");

    this->_log.push("_update_network #4");
    flog.print("  > Update vascular network... 4A/5                                    \r", false);
    ferr.print("  > Update vascular network... 4A/5                                    \r", true );
    this->_prepare_meshes();
    flog.print("  > Update vascular network... 4B/5                                    \r", false);
    ferr.print("  > Update vascular network... 4B/5                                    \r", true );
    this->_reinit_systems();
    this->_log.pop("_update_network #4");

    this->_log.push("_update_network #5");
    flog.print("  > Update vascular network... 5A/5                                    \r", false);
    ferr.print("  > Update vascular network... 5A/5                                    \r", true );
    this->_project_displacement_solution();
    flog.print("  > Update vascular network... 5B/5                                    \r", false);
    ferr.print("  > Update vascular network... 5B/5                                    \r", true );
    this->_set_other_systems();
    this->_log.pop("_update_network #5");

    flog.print("  > Update vascular network... OK                                      \r", false);
    ferr.print("  > Update vascular network... OK                                      \r", true );

    this->_log.pop("_update_network");
}
//=================================================================================================
void TumourAngiogenesisSolver::_prepare_vascular_update () {
    this->_log.push("_prepare_vascular_update");

    const Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    ASSERT_(this->_dmn.mater.end()!=this->_dmn.mater.find(par.get<int>("tumour angiogenesis solver: vascular remodelling material ID")), "fatal internal error occurred");
    const MaterialInfo* mi__vnr = &(this->_dmn.mater.find(par.get<int>("tumour angiogenesis solver: vascular remodelling material ID"))->second);
    ASSERT_(NULL!=mi__vnr, "fatal internal error occurred");
    //
    const MeshBase& msh_vn = this->_dmn_vn.get_mesh();
    //
    const System& vn_flow_sys = this->_dmn_vn.get_equation_systems().get_system<System>(this->_label+"::flow_solution");
    ASSERT_(vn_flow_sys.n_vars()==8, "fatal internal error occurred");
    const System& rds_dc_sys = this->_dmn.get_equation_systems().get_system<System>(ReactionDiffusionSolverXD::get_label()+"::discontinuous_solution");
    ASSERT_(rds_dc_sys.n_vars()==9, "fatal internal error occurred");
    const System& rds_a_sys = this->_dmn.get_equation_systems().get_system<System>(ReactionDiffusionSolverXD::get_label()+"::auxiliary_solution");
    ASSERT_(rds_a_sys.n_vars()==8, "fatal internal error occurred");
    //
    // clear memory and initialise the solution vector
    vn_flow_features.clear();
    vn_flow_sys.update_global_solution(vn_flow_features);
    //
    // clear memory and initialise the solution vector
    rds_discontinuous.clear();
    rds_dc_sys.update_global_solution(rds_discontinuous);
    //
    // clear memory and initialise the solution vector
    drug_concentration.clear();
    rds_a_sys.update_global_solution(drug_concentration);
    //
    const Real current_time = par.get<Real>("current time");
    // check if membrane elements are present in analysis
    if ( par.get<bool>("membrane elements present") )
        ERROR_("this part of code is under development");
    //
    // initiallise this set of (blood) vascular nodes which could potentially sprout/branch
    if ( vasc_node_2_iterate.empty() ) {
        //
        const MeshBase::const_node_iterator msh_vn_nodes_begin = msh_vn.active_nodes_begin();
        const MeshBase::const_node_iterator msh_vn_nodes_end   = msh_vn.active_nodes_end();
        //
        int vascular_nodes_index = 0;
        for (MeshBase::const_node_iterator cn_iter=msh_vn_nodes_begin; cn_iter!=msh_vn_nodes_end; cn_iter++) {
            const Node* node = *cn_iter;
            ASSERT_(NULL!=node, "fatal internal error occurred");
            const dof_id_type node_id = node->id();
            //
            //status_bar(std::cout, ++vascular_nodes_index, msh_vn.n_nodes());
            //
            // get access to the vascular node info
            ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(node_id), "fatal internal error occurred");
            VascularInfo* vi = &(this->_dmn_vn.node_vi.find( node_id )->second);
            ASSERT_(NULL!=vi, "fatal internal error occurred");
            // also check if the flow regine for this vascular node is valid
            ASSERT_(VascularInfo::INVALID_FLOW!=vi->flow, "fatal internal error occurred");
            //
            const Real radius_min = VascularInfo::BLOOD == vi->flow && VascularInfo::BLOOD == vi->flow
                                  ? mi__vnr->get_biochem_param(52) : mi__vnr->get_biochem_param(102);
            //
            // skip vasculogenesis if not a blood vessel
            if ( VascularInfo::BLOOD != vi->flow ) {
                continue;
            }
            //
            // consider to avoid extending the network if node has collapsed already
            if ( VascularInfo::COLLAPSED == vi->status ) {
                continue;
            }
            //
            unsigned int I = 0;
            for (std::set<dof_id_type>::const_iterator ci=vi->vsc_elem.begin(); ci!=vi->vsc_elem.end(); ci++) {
                const Elem& elem = this->_dmn_vn.get_mesh().elem_ref(*ci);
                const dof_id_type conn_node_id = edge2_other_node(elem, node_id);
                //
                ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(conn_node_id), "fatal internal error occurred");
                const VascularInfo* conn_vi = &(this->_dmn_vn.node_vi.find(conn_node_id)->second);
                ASSERT_(NULL!=conn_vi, "fatal internal error occurred");
                //
                if ( VascularInfo::COLLAPSED == conn_vi->status ) continue;
                //
                ASSERT_(3>I, "fatal internal error occurred");
                ++I;
            }
            //
            // if this (functional) vascular node is "alone" then shut it down!
            if ( 0 == I ) {
                vi->status = VascularInfo::COLLAPSED;
                vi->compression_scale = 1000.0;
                //
                const Real temp_load = 0.0;
                const Real features[] = { current_time, 0.0, 0.0, 0.0, 0.0, temp_load,
                                          0.0, 0.0, 0.0, 0.0 };
                //
                // call the vessel wall remodelling routine to compute radius, thickness and pore size
                tumour_angiogenesis::remodel_vascular_XYZ(*mi__vnr, *vi, features);
                //
                vi->radius = radius_min / vi->compression_scale;
                //
                // ignore following computations for collapsed vascular node
                continue;
            }
            //
            // consider the following node if it is a sprout | tip vessel
            if ( VascularInfo::SPROUT == vi->type ||
                 VascularInfo::TIP    == vi->type ) {
                vasc_node_2_iterate.insert( node_id );
            }
            // end of active nodes loop...
        }
        //
    } else {
        //
        std::set<dof_id_type> vasc_node_2_iterate_2_remove;
        //
        int vascular_nodes_index = 0;
        for (std::set<dof_id_type>::const_iterator set_ci=vasc_node_2_iterate.begin(); set_ci!=vasc_node_2_iterate.end(); set_ci++) {
            const Node* node = msh_vn.node_ptr(*set_ci);
            ASSERT_(NULL!=node, "fatal internal error occurred");
            const dof_id_type node_id = node->id();
            // a simple flag to monitor if the info of this vascular node have changed
            bool node_vi_changed = false;
            //
            //status_bar(std::cout, ++vascular_nodes_index, vasc_node_2_iterate.size());
            //
            // get access to the vascular node info
            ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(node_id), "fatal internal error occurred");
            VascularInfo* vi = &(this->_dmn_vn.node_vi.find( node_id )->second);
            ASSERT_(NULL!=vi, "fatal internal error occurred");
            // also check if the flow regine for this vascular node is valid
            ASSERT_(VascularInfo::INVALID_FLOW!=vi->flow, "fatal internal error occurred");
            //
            const Real radius_min = VascularInfo::BLOOD == vi->flow && VascularInfo::BLOOD == vi->flow
                                  ? mi__vnr->get_biochem_param(52) : mi__vnr->get_biochem_param(102);
            //
            // skip vasculogenesis if not a blood vessel
            if ( VascularInfo::BLOOD != vi->flow ) {
                vasc_node_2_iterate_2_remove.insert( node_id );
            }
            //
            // consider to avoid extending the network if node has collapsed already
            if ( VascularInfo::COLLAPSED == vi->status ) {
                vasc_node_2_iterate_2_remove.insert( node_id );
            } else {
                //
                unsigned int I = 0;
                for (std::set<dof_id_type>::const_iterator ci=vi->vsc_elem.begin(); ci!=vi->vsc_elem.end(); ci++) {
                    const Elem& elem = this->_dmn_vn.get_mesh().elem_ref(*ci);
                    const dof_id_type conn_node_id = edge2_other_node(elem, node_id);
                    //
                    ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(conn_node_id), "fatal internal error occurred");
                    const VascularInfo* conn_vi = &(this->_dmn_vn.node_vi.find(conn_node_id)->second);
                    ASSERT_(NULL!=conn_vi, "fatal internal error occurred");
                    //
                    if ( VascularInfo::COLLAPSED == conn_vi->status ) continue;
                    //
                    ASSERT_(3>I, "fatal internal error occurred");
                    ++I;
                }
                //
                // if this (functional) vascular node is "alone" then shut it down!
                if ( 0 == I ) {
                    vi->status = VascularInfo::COLLAPSED;
                    vi->compression_scale = 1000.0;
                    //
                    const Real temp_load = 0.0;
                    const Real features[] = { current_time, 0.0, 0.0, 0.0, 0.0, temp_load,
                                              0.0, 0.0, 0.0, 0.0 };
                    //
                    // call the vessel wall remodelling routine to compute radius, thickness and pore size
                    tumour_angiogenesis::remodel_vascular_XYZ(*mi__vnr, *vi, features);
                    //
                    vi->radius = radius_min / vi->compression_scale;
                    //
                    // ignore following computations for collapsed vascular node
                    vasc_node_2_iterate_2_remove.insert( node_id );
                }
                //
            }
            //
            // ignore the following if node is not part of a sprout/tip vessel
            if ( ! ( VascularInfo::SPROUT == vi->type ||
                     VascularInfo::TIP    == vi->type ) ) {
                vasc_node_2_iterate_2_remove.insert( node_id );
            }
        }
        //
        for ( std::set<dof_id_type>::const_iterator
              set_ci=vasc_node_2_iterate_2_remove.begin(); set_ci!=vasc_node_2_iterate_2_remove.end(); set_ci++ ) {
            std::set<dof_id_type>::const_iterator set_cj = vasc_node_2_iterate.find( *set_ci );
            if ( vasc_node_2_iterate.end() != set_cj ) vasc_node_2_iterate.erase( set_cj );
        }
        //
    }
    //
    // clean the memory for this local container
    vasc_node_2_notiterate.clear();
    //
    // if only in parallel mode, then retrieve the quadrature info from all processors
    if ( libMesh::global_n_processors() > 1 ) {
        const std::string fname = par.get<std::string>("results directory: quadrature data") + "/"
                                + par.get<std::string>("output file name")
                                + ".qpi." + number2string(libMesh::global_processor_id());
        this->_dmn.save_quadrature(fname.c_str(), QuadratureInfo::ALL, QuadratureInfo::LEVEL_3);
        gcomm_ptr->barrier();
        //
        std::vector<std::string> fname_ls(libMesh::global_n_processors());
        for (unsigned int iproc=0; iproc<libMesh::global_n_processors(); iproc++) {
            fname_ls[iproc] = par.get<std::string>("results directory: quadrature data") + "/"
                            + par.get<std::string>("output file name")
                            + ".qpi." + number2string(iproc);
        }
        this->_dmn.gather_quadrature(fname_ls, QuadratureInfo::ALL, QuadratureInfo::LEVEL_3);
        gcomm_ptr->barrier();
    }
    //
    elem3d_qpi = ( libMesh::global_n_processors() == 1 ? &this->_dmn.elem_qi_local : &this->_dmn.elem_qi );
    elem2d_qpi = ( libMesh::global_n_processors() == 1 ? &this->_dmn.elem_side_qi_local : &this->_dmn.elem_side_qi );
    //
    // enforce to clean the memory
    new_node_to_add.clear();
    new_elem_to_add.clear();
    new_node_vi_to_add.clear();
    existing_node_to_update.clear();
    existing_elem_to_update.clear();
    existing_node_vi_to_update.clear();

    this->_log.pop("_prepare_vascular_update");
}
//=================================================================================================
void TumourAngiogenesisSolver::_distribute_results () {
    this->_log.push("_distribute_results");

    // if in parallel mode then the 0-rank processor should distribute the info wrt
    // the newly created nodes of the vascular network
    if ( 1 == libMesh::global_n_processors() ) {
        this->_log.pop("_distribute_results");
        return;
    }
    //
    const Parameters& par = this->_dmn.get_equation_systems().parameters;
    const std::string fname = par.get<std::string>("results directory: temporary") + "/"
                            + par.get<std::string>("output file name") + ".vn.dat";
    //
    if ( libMesh::global_processor_id() == 0 ) {
        std::ofstream file(fname);
        write_current_vascular_network( new_node_to_add, new_elem_to_add, new_node_vi_to_add,
                                        existing_node_to_update, existing_elem_to_update, this->_dmn_vn.node_vi,
                                        this->_dmn_vn.branch__node, file );
    }
    gcomm_ptr->barrier();
    //
    if ( libMesh::global_processor_id() != 0 ) {
        std::ifstream file(fname);
        read_current_vascular_network( new_node_to_add, new_elem_to_add, new_node_vi_to_add,
                                       existing_node_to_update, existing_elem_to_update, existing_node_vi_to_update,
                                       this->_dmn_vn.branch__node, file );
    }
    gcomm_ptr->barrier();

    this->_log.pop("_distribute_results");
}
//=================================================================================================
void TumourAngiogenesisSolver::_update_domain_data () {
    this->_log.push("_update_domain_data");

    const MeshBase::const_element_iterator msh_3d_elements_begin = this->_dmn.get_mesh().active_elements_begin();
    const MeshBase::const_element_iterator msh_3d_elements_end   = this->_dmn.get_mesh().active_elements_end();
    //
    MeshBase::const_element_iterator cei = msh_3d_elements_begin;
    std::map< dof_id_type, Point >::const_iterator mdp_ci;
    std::map<dof_id_type, VascularInfo>::const_iterator mdvi_ci;
    std::map< dof_id_type, Dual<dof_id_type> >::const_iterator mddd_ci;
    //
    // memory to clean first and initialise first
    this->_dmn_vn.elem3d__node.clear();
    for (cei=msh_3d_elements_begin; cei!=msh_3d_elements_end; cei++)
        this->_dmn_vn.elem3d__node.insert(std::make_pair((*cei)->id(), std::set<dof_id_type>()));
    //
    // add the new points in the vascular network (one-dimensional) mesh
    flog.print("  > Update domain data - vascular network (1/7)                        \r", false);
    ferr.print("  > Update domain data - vascular network (1/7)                        \r", true );
    for (mdp_ci=new_node_to_add.begin(); mdp_ci!=new_node_to_add.end(); mdp_ci++) {
        const dof_id_type& vascular_node_id = mdp_ci->first;
        const Point& vascular_node_xyz = mdp_ci->second;
        //
        this->_dmn_vn.get_mesh().add_point(vascular_node_xyz, vascular_node_id);
    }
    //
    // update the vascular info of the nodes
    flog.print("  > Update domain data - vascular network (2/7)                        \r", false);
    ferr.print("  > Update domain data - vascular network (2/7)                        \r", true );
    for (mdvi_ci=new_node_vi_to_add.begin(); mdvi_ci!=new_node_vi_to_add.end(); mdvi_ci++) {
        const dof_id_type& vascular_node_id = mdvi_ci->first;
        const VascularInfo& vascular_info = mdvi_ci->second;
        //
        this->_dmn_vn.node_vi.insert(std::make_pair(vascular_node_id, vascular_info));
    }
    //
    // update the list of in/outlet vascular nodes
    flog.print("  > Update domain data - vascular network (3/7)                        \r", false);
    ferr.print("  > Update domain data - vascular network (3/7)                        \r", true );
    for (mdvi_ci=this->_dmn_vn.node_vi.begin(); mdvi_ci!=this->_dmn_vn.node_vi.end(); mdvi_ci++) {
        const dof_id_type& vascular_node_id = mdvi_ci->first;
        //
        if ( mdvi_ci->second.type == VascularInfo::INLET )
            this->_dmn_vn.node_inlet.insert(vascular_node_id);
        else if ( mdvi_ci->second.type == VascularInfo::OUTLET )
            this->_dmn_vn.node_outlet.insert(vascular_node_id);
    }
    //
    // add the new two-node edge elements in the vascular network one-dimensional mesh
    flog.print("  > Update domain data - vascular network (4/7)                        \r", false);
    ferr.print("  > Update domain data - vascular network (4/7)                        \r", true );
    for (mddd_ci=new_elem_to_add.begin(); mddd_ci!=new_elem_to_add.end(); mddd_ci++) {
        const dof_id_type& vascular_node_0_id = mddd_ci->second.get_first();
        const dof_id_type& vascular_node_1_id = mddd_ci->second.get_second();
        Elem* vessel = Elem::build(EDGE2).release();
        vessel->set_id(mddd_ci->first);
        vessel->set_node(0) = this->_dmn_vn.get_mesh().node_ptr(vascular_node_0_id);
        vessel->set_node(1) = this->_dmn_vn.get_mesh().node_ptr(vascular_node_1_id);
        vessel->subdomain_id() = this->_dmn_vn.node_vi.find(vascular_node_0_id)->second.flow;
        //
        this->_dmn_vn.get_mesh().add_elem(vessel);
    }
    //
    // update the existing points of the vascular network one-dimensional mesh
    flog.print("  > Update domain data - vascular network (5/7)                        \r", false);
    ferr.print("  > Update domain data - vascular network (5/7)                        \r", true );
    for (mdp_ci=existing_node_to_update.begin(); mdp_ci!=existing_node_to_update.end(); mdp_ci++) {
        const dof_id_type& vascular_node_id = mdp_ci->first;
        Node* this_node = this->_dmn_vn.get_mesh().node_ptr(vascular_node_id);
        (*this_node)(0) = mdp_ci->second(0);
        (*this_node)(1) = mdp_ci->second(1);
        (*this_node)(2) = mdp_ci->second(2);
    }
    //
    // update for the existing points their vascular information data
    flog.print("  > Update domain data - vascular network (6/7)                        \r", false);
    ferr.print("  > Update domain data - vascular network (6/7)                        \r", true );
    for (mdvi_ci=existing_node_vi_to_update.begin(); mdvi_ci!=existing_node_vi_to_update.end(); mdvi_ci++) {
        const dof_id_type& vascular_node_id = mdvi_ci->first;
        //
        std::map<dof_id_type, VascularInfo>::iterator j = this->_dmn_vn.node_vi.find(vascular_node_id);
        if ( this->_dmn_vn.node_vi.end() != j ) j->second = mdvi_ci->second;
        else libmesh_error();
    }
    //
    // update the container of associated 3D elements and vascular nodes
    flog.print("  > Update domain data - vascular network (7/7)                        \r", false);
    ferr.print("  > Update domain data - vascular network (7/7)                        \r", true );
    for (mdvi_ci=this->_dmn_vn.node_vi.begin(); mdvi_ci!=this->_dmn_vn.node_vi.end(); mdvi_ci++) {
        const dof_id_type& vascular_node_id = mdvi_ci->first;
        const dof_id_type& element3d_id = mdvi_ci->second.elem3d_id;
        //
        this->_dmn_vn.elem3d__node.find(element3d_id)->second.insert(vascular_node_id);
    }
    //
    // memory to clean first and initialise later
    this->_dmn_coupled.cm_strm_node__cm_vasc_node_segm.clear();
    //
    // add the new points of the vascular network
    flog.print("  > Update domain data - coupled-problem (1/5)                         \r", false);
    ferr.print("  > Update domain data - coupled-problem (1/5)                         \r", true );
    for (mdp_ci=new_node_to_add.begin(); mdp_ci!=new_node_to_add.end(); mdp_ci++) {
        const Point& vascular_node_xyz = mdp_ci->second;
        const dof_id_type& cm_vascular_node_id = static_cast<dof_id_type>(mdp_ci->first + this->_dmn.get_mesh().n_nodes());
        //
        this->_dmn_coupled.get_mesh().add_point(vascular_node_xyz, cm_vascular_node_id);
        //
        this->_dmn_coupled.vasc_node__cm_vasc_node.insert( std::make_pair(mdp_ci->first, cm_vascular_node_id) );
        this->_dmn_coupled.cm_vasc_node__vasc_node.insert( std::make_pair(cm_vascular_node_id, mdp_ci->first) );
    }
    //
    // add the new two-node edge elements of the vascular network
    flog.print("  > Update domain data - coupled-problem (2/5)                         \r", false);
    ferr.print("  > Update domain data - coupled-problem (2/5)                         \r", true );
    for (mddd_ci=new_elem_to_add.begin(); mddd_ci!=new_elem_to_add.end(); mddd_ci++) {
        const dof_id_type& cm_vascular_node_0_id = this->_dmn_coupled.vasc_node__cm_vasc_node.find(mddd_ci->second.get_first() )->second;
        const dof_id_type& cm_vascular_node_1_id = this->_dmn_coupled.vasc_node__cm_vasc_node.find(mddd_ci->second.get_second())->second;
        Elem* vessel = Elem::build(EDGE2).release();
        vessel->set_id( this->_dmn_coupled.get_mesh().n_elem() );
        vessel->set_node(0) = this->_dmn_coupled.get_mesh().node_ptr(cm_vascular_node_0_id);
        vessel->set_node(1) = this->_dmn_coupled.get_mesh().node_ptr(cm_vascular_node_1_id);
        vessel->subdomain_id() = static_cast<subdomain_id_type>(DomainData_SolidCoupledWithVascularNetwork::VESSEL);
        //
        this->_dmn_coupled.get_mesh().add_elem(vessel);
        //
        this->_dmn_coupled.cm_vasc_elem__vasc_elem.insert( std::make_pair(vessel->id(), mddd_ci->first) );
        this->_dmn_coupled.vasc_elem__cm_vasc_elem.insert( std::make_pair(mddd_ci->first, vessel->id()) );
    }
    //
    // add the new two-node edge elements of the interface between vascular network nodes and stroma nodes
    flog.print("  > Update domain data - coupled-problem (3/5)                         \r", false);
    ferr.print("  > Update domain data - coupled-problem (3/5)                         \r", true );
    for (mdvi_ci=new_node_vi_to_add.begin(); mdvi_ci!=new_node_vi_to_add.end(); mdvi_ci++) {
        const dof_id_type& cm_vascular_node_id = this->_dmn_coupled.vasc_node__cm_vasc_node.find(mdvi_ci->first)->second;
        const Elem* elem = this->_dmn.get_mesh().elem_ptr( mdvi_ci->second.elem3d_id );
        ASSERT_(NULL!=elem, "fatal internal error occurred");
        //
        std::set<dof_id_type> cm_elem_interface;
        for (unsigned int i=0; i<elem->n_nodes(); i++) {
            const dof_id_type& cm_stromal_node_id = this->_dmn_coupled.strm_node__cm_strm_node.find(elem->node_id(i))->second;
            Elem* interface = Elem::build(EDGE2).release();
            interface->set_id( this->_dmn_coupled.get_mesh().n_elem() );
            interface->set_node(0) = this->_dmn_coupled.get_mesh().node_ptr( cm_vascular_node_id );
            interface->set_node(1) = this->_dmn_coupled.get_mesh().node_ptr( cm_stromal_node_id  );
            interface->subdomain_id() = static_cast<subdomain_id_type>(DomainData_SolidCoupledWithVascularNetwork::INTERFACE);
            //
            this->_dmn_coupled.get_mesh().add_elem(interface);
            // insert the interface (segment) element in the list
            cm_elem_interface.insert(interface->id());
            //
            // add this vascular node to the corresponding (neighborhood) coupled-mesh stromal nodes
            std::map<dof_id_type, std::set<pair_dof_id_type> >::iterator iter =
                this->_dmn_coupled.cm_strm_node__cm_vasc_node_segm.find( cm_stromal_node_id );
            if ( this->_dmn_coupled.cm_strm_node__cm_vasc_node_segm.end() == iter ) {
                std::set<pair_dof_id_type> ini_set;
                ini_set.insert( pair_dof_id_type(cm_vascular_node_id, interface->id()) );
                //
                this->_dmn_coupled.cm_strm_node__cm_vasc_node_segm.insert( std::make_pair(cm_stromal_node_id, ini_set) );
            } else {
                iter->second.insert( pair_dof_id_type(cm_vascular_node_id, interface->id()) );
            }
        }
        // insert for this vascular node the list of associated interface elements
        this->_dmn_coupled.cm_vasc_node__cm_intf_elem.insert( std::make_pair(cm_vascular_node_id, cm_elem_interface) );
    }
    //
    // update the coordinates of the existing vascular network
    flog.print("  > Update domain data - coupled-problem (4/5)                         \r", false);
    ferr.print("  > Update domain data - coupled-problem (4/5)                         \r", true );
    for (mdp_ci=existing_node_to_update.begin(); mdp_ci!=existing_node_to_update.end(); mdp_ci++) {
        const dof_id_type& cm_vascular_node_id = this->_dmn_coupled.vasc_node__cm_vasc_node.find(mdp_ci->first)->second;
        Node* this_node = this->_dmn_coupled.get_mesh().node_ptr(cm_vascular_node_id);
        (*this_node)(0) = mdp_ci->second(0);
        (*this_node)(1) = mdp_ci->second(1);
        (*this_node)(2) = mdp_ci->second(2);
    }
    //
    // update the existing two-node edge elements of the interface between vascular network nodes and stroma nodes
    flog.print("  > Update domain data - coupled-problem (5/5)                         \r", false);
    ferr.print("  > Update domain data - coupled-problem (5/5)                         \r", true );
    for (mdp_ci=existing_node_to_update.begin(); mdp_ci!=existing_node_to_update.end(); mdp_ci++) {
        const dof_id_type& cm_vascular_node_id = this->_dmn_coupled.vasc_node__cm_vasc_node.find(mdp_ci->first)->second;
        const std::set<dof_id_type>& interface_elem = this->_dmn_coupled.cm_vasc_node__cm_intf_elem.find(cm_vascular_node_id)->second;
        //
        const Elem* elem = this->_dmn.get_mesh().elem_ptr( this->_dmn_vn.node_vi.find(mdp_ci->first)->second.elem3d_id );
        ASSERT_(NULL!=elem, "fatal internal error occurred");
        ASSERT_(elem->n_nodes()==interface_elem.size(), "fatal internal error occurred");
        //
        std::set<dof_id_type>::const_iterator ci_intf = interface_elem.begin();
        for (unsigned int i=0; i<elem->n_nodes(); i++) {
            const dof_id_type& cm_stromal_node_id = this->_dmn_coupled.strm_node__cm_strm_node.find(elem->node_id(i))->second;
            Elem* interface = this->_dmn_coupled.get_mesh().elem_ptr(*ci_intf);
            interface->set_node(0) = this->_dmn_coupled.get_mesh().node_ptr( cm_vascular_node_id ); // NOTE: this ID should have not changed
            interface->set_node(1) = this->_dmn_coupled.get_mesh().node_ptr( cm_stromal_node_id );
            //
            // add this vascular node to the corresponding (neighborhood) coupled-mesh stromal nodes
            std::map<dof_id_type, std::set<pair_dof_id_type> >::iterator iter =
                this->_dmn_coupled.cm_strm_node__cm_vasc_node_segm.find( cm_stromal_node_id );
            if ( this->_dmn_coupled.cm_strm_node__cm_vasc_node_segm.end() == iter ) {
                std::set<pair_dof_id_type> ini_set;
                ini_set.insert( pair_dof_id_type(cm_vascular_node_id, interface->id()) );
                //
                this->_dmn_coupled.cm_strm_node__cm_vasc_node_segm.insert( std::make_pair(cm_stromal_node_id, ini_set) );
            } else {
                iter->second.insert( pair_dof_id_type(cm_vascular_node_id, interface->id()) );
            }
            // increment this iterator
            ++ci_intf;
        }
    }
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    // enforce to clean the memory
    new_node_to_add.clear();
    new_elem_to_add.clear();
    new_node_vi_to_add.clear();
    existing_node_to_update.clear();
    existing_elem_to_update.clear();
    existing_node_vi_to_update.clear();

    this->_log.pop("_update_domain_data");
}
//=================================================================================================
void TumourAngiogenesisSolver::_find_anastomoses () {
    this->_log.push("_find_anastomoses");

    // only the 0-rank processor will compute the elongation (extension) of the tip vessels or
    // the newly created sprouts that emerge from branching or even the formation of
    // anastomoses, and then generate the vascular node if necessary
    if ( libMesh::global_processor_id() ) {
        this->_log.pop("_find_anastomoses");
        return;
    }

    const Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    const System& vn_flow_sys = this->_dmn_vn.get_equation_systems().get_system<System>(this->_label+"::flow_solution");
    //
    const MeshBase& msh_vn = this->_dmn_vn.get_mesh();
    const MeshBase& msh_3d = this->_dmn.get_mesh();
    //
    const MeshBase::const_element_iterator msh_3d_elements_begin = msh_3d.active_elements_begin();
    const MeshBase::const_element_iterator msh_3d_elements_end   = msh_3d.active_elements_end();
    //
    const MaterialInfo* mi__vnr = &(this->_dmn.mater.find(par.get<int>("tumour angiogenesis solver: vascular remodelling material ID"))->second);
    //
    const Real current_time = par.get<Real>("current time"),
               time_step = par.get<Real>("tumour angiogenesis solver: time step")
                         * par.get<int>("tumour angiogenesis solver: angiogenesis increment");
    const unsigned int ninvars_NCS = par.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = par.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = par.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = par.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    // WARNING: following parameters are for blood vessel angiogenesis, not for lympangiogenesis
    const Real capillary_min_radius         = mi__vnr->get_biochem_param(52),
               capillary_max_radius         = mi__vnr->get_biochem_param(53),
               capillary_min_thickness      = mi__vnr->get_biochem_param(56),
               capillary_max_pore_radius    = mi__vnr->get_biochem_param(61),
               capillary_max_pores_fraction = mi__vnr->get_biochem_param(65),
               collapsed_vascular_node_survival = mi__vnr->get_biochem_param(200);
    ASSERT_(collapsed_vascular_node_survival>0 && collapsed_vascular_node_survival<100, "parameter to designate the survival propability of a just collapsed vascular node is not valid");
    const Real mean_velocity__threshold = mi__vnr->get_biochem_param(10),
               extension_min_length     = mi__vnr->get_biochem_param(14),
               sprouts_safe_margin_mean = mi__vnr->get_biochem_param(15),
               sprouts_safe_margin_std  = mi__vnr->get_biochem_param(16),
               anastomosis_tolerance    = capillary_max_radius,
               tip_extension_max_length = mi__vnr->get_biochem_param(18);
    ASSERT_(anastomosis_tolerance<=sprouts_safe_margin_mean+sprouts_safe_margin_std, "anastomosis tolerance is less/equal to the sprouts safe margin in the tumour angiogenesis solver");
    ASSERT_(extension_min_length<=par.get<Real>("finite element grid: min edge length"), "min extension length for angiogenesis algorithm is not valid");

    std::set<dof_id_type> vasc_node_2_iterate_2_add;
    std::set<dof_id_type> vasc_node_2_iterate_2_remove;

    int vascular_nodes_index = 0;
    for ( std::set<dof_id_type>::const_iterator
          set_ci=vasc_node_2_iterate.begin(); set_ci!=vasc_node_2_iterate.end(); set_ci++ )
    {
        const Node* node = msh_vn.node_ptr(*set_ci);
        ASSERT_(NULL!=node, "fatal internal error occurred");
        const dof_id_type node_id = node->id();
        // a simple flag to monitor if the info of this vascular node have changed
        bool node_vi_changed = false;
        //
        status_bar(std::cout, ++vascular_nodes_index, vasc_node_2_iterate.size());
        //
        // get access to the vascular node info
        ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(node_id), "fatal internal error occurred");
        VascularInfo* vi = &(this->_dmn_vn.node_vi.find( node_id )->second);
        ASSERT_(NULL!=vi, "fatal internal error occurred");
        // also check if the flow regine for this vascular node is valid
        ASSERT_(VascularInfo::INVALID_FLOW!=vi->flow, "fatal internal error occurred");
        //
        // skip vasculogenesis if not a blood vessel
        if ( VascularInfo::BLOOD != vi->flow ) {
            vasc_node_2_iterate_2_remove.insert( node_id );
            continue;
        }
        //
        // consider to avoid extending the network if node has collapsed already
        if ( VascularInfo::COLLAPSED == vi->status ) {
            vasc_node_2_iterate_2_remove.insert( node_id );
            continue;
        }
        //
        // ignore the following if node is not part of a sprout/tip vessel
        if ( ! ( VascularInfo::SPROUT == vi->type ||
                 VascularInfo::TIP    == vi->type ) ) {
            vasc_node_2_iterate_2_remove.insert( node_id );
            continue;
        }
        //
        const MaterialInfo* mi = NULL;
        const QuadratureInfo* qpi_ptr = NULL;
        {
            //
            // get access to the 3D element
            const Elem* elem3d = msh_3d.elem_ptr(vi->elem3d_id);
            ASSERT_(NULL!=elem3d, "fatal internal error occurred");
            //
            // get access to the region and material info
            ASSERT_(this->_dmn.regn.end()!=this->_dmn.regn.find(elem3d->subdomain_id()), "fatal internal error occurred");
            const RegionInfo* rgi = &(this->_dmn.regn.find(elem3d->subdomain_id())->second);
            ASSERT_(NULL!=rgi, "fatal internal error occurred");
            mi = rgi->mat;
            //
            // get access to the 3D element quadrature points data
            ASSERT_(elem3d_qpi->end()!=elem3d_qpi->find(elem3d->id()), "fatal internal error occurred");
            qpi_ptr = &(elem3d_qpi->find(elem3d->id())->second);
            ASSERT_(NULL!=qpi_ptr, "fatal internal error occurred");
            //
        }
        //
        // get the "mean velocity" at the vascular node
        const Real mean_velocity = get_system_solution_nodal_value<6>(vn_flow_sys, node, vn_flow_features);
        //
        // calculate the vascular elongation
        Displacement dX(0.0, 0.0, 0.0);
        {
            //
            // use the inverse distance algorithm to interpolate the data
            InverseDistanceAlgorithm ida(2.5);
            ida.init(qpi_ptr->XYZ);
            std::vector<Real> weight;
            ida.calc(*node, weight);
            //
            FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
            // do initialise these variables first
            fld.current_time = current_time;
            fld.time_step = time_step;
            fld.deform_gradient.comp = 0.0;
            //
            // interpolate the quadrature data to initialise field data properly
            for (unsigned int ipnt=0; ipnt<weight.size(); ipnt++) {
                for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                    fld.biochem[ivar]      += weight[ipnt] * qpi_ptr->biochem[ipnt][ivar];
                    fld.grad_biochem[ivar] += weight[ipnt] * qpi_ptr->grad_biochem[ipnt][ivar];
                    fld.biochem_rate[ivar] += weight[ipnt] * qpi_ptr->biochem_rate[ipnt][ivar];
                }
                for (unsigned int ivar=nvars_RDS; ivar<nvars_RDS+ndcvars_RDS; ivar++) {
                    fld.biochem[ivar] += weight[ipnt] * qpi_ptr->biochem[ipnt][ivar];
                }
                fld.stress.comp += weight[ipnt] * qpi_ptr->stress[ipnt].comp;
                fld.deform_gradient.comp += weight[ipnt] * qpi_ptr->deform_grad[ipnt].comp;
                for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                    fld.auxiliary[ivar]      += weight[ipnt] * qpi_ptr->auxiliary[ipnt][ivar];
                    fld.grad_auxiliary[ivar] += weight[ipnt] * qpi_ptr->grad_auxiliary[ipnt][ivar];
                    fld.auxiliary_rate[ivar] += weight[ipnt] * qpi_ptr->auxiliary_rate[ipnt][ivar];
                }
            }
            //
            // a simple way to flag the vascular node as well-perfused or not
            fld.div_velocity = ( mean_velocity < mean_velocity__threshold ? 0.0 : 1.0 );
            //
            const Vector3D reference_coords = set_vector((*node)(0), (*node)(1), (*node)(2)),
                           current_coords = mul_im_m(fld.deform_gradient, reference_coords);
            //
            dX = (*node);
            if      ( 4 == nvars_RDS && 9 == ndcvars_RDS && 8 == nvars_AUX )
                tumour_angiogenesis::growth_vascular_498(*mi, fld, *vi, dX);
            else ERROR_("fatal internal error occurred");
            //
        }
        //
        // ignore the following computations if the displacement increment is negligible
        if ( dX.norm() < extension_min_length ) {
            vasc_node_2_notiterate.insert( node_id );
            continue;
        }
        //
        // sprouting/branching vascular node coordinates and new vascular node coordinates
        const Point origin = *node;
        Point destination = origin + dX;
        //
        dof_id_type target_node_id = DofObject::invalid_id;
        //
        // check for anastomosis between different branches
        for ( std::set<dof_id_type>::const_iterator
              set_cj=vasc_node_2_iterate.begin(); set_cj!=vasc_node_2_iterate.end(); set_cj++ )
        {
            const Node* existing_node = msh_vn.node_ptr(*set_cj);
            ASSERT_(NULL!=existing_node, "fatal internal error occurred");
            const dof_id_type existing_node_id = existing_node->id();
            //
            // skip for self-node check...
            if ( node_id == existing_node_id ) continue;
            //
            // get access to the target vascular node info
            libmesh_assert(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(existing_node_id));
            const VascularInfo* vi_t = &(this->_dmn_vn.node_vi.find( existing_node_id )->second);
            libmesh_assert(NULL!=vi_t);
            //
            // skip vasculogenesis if not a blood vessel
            if ( VascularInfo::BLOOD != vi_t->flow ) continue;
            //
            // consider to avoid anastomosis in the network if target node has collapsed already
            if ( VascularInfo::COLLAPSED == vi_t->status ) continue;
            //
            // ignore anastomosis if target node is not part of a sprout/mature/tip vessel
            if ( ! ( VascularInfo::SPROUT == vi_t->type ||
                     VascularInfo::MATURE == vi_t->type ||
                     VascularInfo::TIP    == vi_t->type ) ) continue;
            //
            // ignore if both vascular network nodes lie in the same branch/parent branch
            if ( vi_t->branch_id == vi->branch_id ) continue;
            if ( vi_t->branch_id        == vi->parent_branch_id ||
                 vi_t->parent_branch_id == vi->branch_id        ) continue;
            //
            // check the distance between the destination point and an existing node
            if ( Point(destination-(*existing_node)).norm() <= anastomosis_tolerance ) {
                target_node_id = existing_node_id;
                break;
            }
            // end of (existing) vascular nodes loop...
        }
        //
        // check below if an anastomosis occurs
        if ( DofObject::invalid_id == target_node_id ) continue;
        //
        // ...in that case just flag these two nodes to skip in the sprouting that follows
        vasc_node_2_notiterate.insert( node_id );
        vasc_node_2_notiterate.insert( target_node_id );

        // set the ID of the new vascular element (segment)
        const dof_id_type current_segm_id = msh_vn.n_elem() + new_elem_to_add.size();

        node_vi_changed = true; // set this boolean flag
        //
        // access the target vascular node
        const Node* target_node = msh_vn.node_ptr(target_node_id);
        ASSERT_(NULL!=target_node, "fatal internal error occurred");
        // retrieve vascular information of target node; check first
        ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(target_node_id), "fatal internal error occurred");
        VascularInfo* vi_t = &(this->_dmn_vn.node_vi.find( target_node_id )->second);
        ASSERT_(NULL!=vi_t, "fatal internal error occurred");
        //
        unsigned int n_conn__node = 0;
        for (std::set<dof_id_type>::const_iterator s_citer=vi->vsc_elem.begin(); s_citer!=vi->vsc_elem.end(); s_citer++) {
            dof_id_type conn_node_id = DofObject::invalid_id;
            if ( new_elem_to_add.end() == new_elem_to_add.find(*s_citer) ) {
                const Elem& elem = msh_vn.elem_ref(*s_citer);
                conn_node_id = edge2_other_node(elem, node_id);
            } else {
                const Dual<dof_id_type>& dl = new_elem_to_add.find(*s_citer)->second;
                if      ( dl.get_first()  == node_id ) conn_node_id = dl.get_second();
                else if ( dl.get_second() == node_id ) conn_node_id = dl.get_first();
                else libmesh_error();
            }
            ASSERT_(DofObject::invalid_id!=conn_node_id, "fatal internal error occurred");
            //
            if ( this->_dmn_vn.node_vi.end() != this->_dmn_vn.node_vi.find(conn_node_id) ) {
                const VascularInfo& conn_vi = this->_dmn_vn.node_vi.find(conn_node_id)->second;
                if ( VascularInfo::COLLAPSED == conn_vi.status ) continue;
            } else if ( new_node_vi_to_add.end() != new_node_vi_to_add.find(conn_node_id) ) {
                const VascularInfo& conn_vi = new_node_vi_to_add.find(conn_node_id)->second;
                if ( VascularInfo::COLLAPSED == conn_vi.status ) libmesh_error();
            } else libmesh_error();
            //
            ASSERT_(3>n_conn__node, "fatal internal error occurred");
            ++n_conn__node;
        }
        //
        unsigned int n_conn__target_node = 0;
        for (std::set<dof_id_type>::const_iterator s_citer=vi_t->vsc_elem.begin(); s_citer!=vi_t->vsc_elem.end(); s_citer++) {
            dof_id_type conn_node_id = DofObject::invalid_id;
            if ( new_elem_to_add.end() == new_elem_to_add.find(*s_citer) ) {
                const Elem& elem = msh_vn.elem_ref(*s_citer);
                conn_node_id = edge2_other_node(elem, target_node_id);
            } else {
                const Dual<dof_id_type>& dl = new_elem_to_add.find(*s_citer)->second;
                if      ( dl.get_first()  == target_node_id ) conn_node_id = dl.get_second();
                else if ( dl.get_second() == target_node_id ) conn_node_id = dl.get_first();
                else libmesh_error();
            }
            ASSERT_(DofObject::invalid_id!=conn_node_id, "fatal internal error occurred");
            //
            if ( this->_dmn_vn.node_vi.end() != this->_dmn_vn.node_vi.find(conn_node_id) ) {
                const VascularInfo& conn_vi = this->_dmn_vn.node_vi.find(conn_node_id)->second;
                if ( VascularInfo::COLLAPSED == conn_vi.status ) continue;
            } else if ( new_node_vi_to_add.end() != new_node_vi_to_add.find(conn_node_id) ) {
                const VascularInfo& conn_vi = new_node_vi_to_add.find(conn_node_id)->second;
                if ( VascularInfo::COLLAPSED == conn_vi.status ) libmesh_error();
            } else libmesh_error();
            //
            ASSERT_(3>n_conn__target_node, "fatal internal error occurred");
            ++n_conn__target_node;
        }
        //
        if        ( VascularInfo::SPROUT == vi->type ||
                    VascularInfo::MATURE == vi->type ) {
            if ( 2 != n_conn__node ) libmesh_error();
            vi->type = VascularInfo::BRANCH;
        } else if ( VascularInfo::TIP    == vi->type ) {
            if ( 1 != n_conn__node ) libmesh_error();
            vi->type = VascularInfo::SPROUT;
        } else libmesh_error();
        vi->vsc_elem.insert( current_segm_id );
        //
        if        ( VascularInfo::SPROUT == vi_t->type ||
                    VascularInfo::MATURE == vi_t->type ) {
            if ( 2 != n_conn__target_node ) libmesh_error();
            vi_t->type = VascularInfo::BRANCH;
        } else if ( VascularInfo::TIP    == vi_t->type ) {
            if ( 1 != n_conn__target_node ) libmesh_error();
            vi_t->type = VascularInfo::SPROUT;
        } else libmesh_error();
        vi_t->vsc_elem.insert( current_segm_id );
        // define the vertices of the vascular segment to insert
        const Dual<dof_id_type> edge2_nodes(node_id, target_node_id);
        //
        // update the container
        new_elem_to_add.insert( std::make_pair(current_segm_id, edge2_nodes) );
        //
        // end of active nodes loop...
    }
    // sanity checks
    ASSERT_(new_node_to_add.size()==new_node_vi_to_add.size(), "fatal internal error occurred");
    ASSERT_(existing_elem_to_update.empty(), "fatal internal error occurred");
    //
    // now update this local container of (blood) vessel nodes to iterate on the subsequent
    // vascular network update call
    for ( std::set<dof_id_type>::const_iterator
          set_ci=vasc_node_2_iterate_2_add.begin(); set_ci!=vasc_node_2_iterate_2_add.end(); set_ci++ ) {
        vasc_node_2_iterate.insert( *set_ci );
    }
    for ( std::set<dof_id_type>::const_iterator
          set_ci=vasc_node_2_iterate_2_remove.begin(); set_ci!=vasc_node_2_iterate_2_remove.end(); set_ci++ ) {
        std::set<dof_id_type>::const_iterator set_cj = vasc_node_2_iterate.find( *set_ci );
        if ( vasc_node_2_iterate.end() != set_cj ) vasc_node_2_iterate.erase( set_cj );
    }

    this->_log.pop("_find_anastomoses");
}
//=================================================================================================
void TumourAngiogenesisSolver::_perform_sprouting () {
    this->_log.push("_perform_sprouting");

    // only the 0-rank processor will compute the elongation (extension) of the tip vessels or
    // the newly created sprouts that emerge from branching or even the formation of
    // anastomoses, and then generate the vascular node if necessary
    if ( libMesh::global_processor_id() ) {
        this->_log.pop("_perform_sprouting");
        return;
    }

    System& rds_dc_sys = this->_dmn.get_equation_systems().get_system<System>(ReactionDiffusionSolverXD::get_label()+"::discontinuous_solution");
    //
    const FEM_Context rds_dc_ctx(rds_dc_sys);

    System& rds_a_sys = this->_dmn.get_equation_systems().get_system<System>(ReactionDiffusionSolverXD::get_label()+"::auxiliary_solution");
    //
    const FEM_Context rds_a_ctx(rds_a_sys);

    const System& vn_flow_sys = this->_dmn_vn.get_equation_systems().get_system<System>(this->_label+"::flow_solution");

    const Parameters& par = this->_dmn.get_equation_systems().parameters;

    const MeshBase& msh_vn = this->_dmn_vn.get_mesh();
    const MeshBase& msh_3d = this->_dmn.get_mesh();

    const MeshBase::const_element_iterator msh_3d_elements_begin = msh_3d.active_elements_begin();
    const MeshBase::const_element_iterator msh_3d_elements_end   = msh_3d.active_elements_end();

    const MaterialInfo* mi__vnr = &(this->_dmn.mater.find(par.get<int>("tumour angiogenesis solver: vascular remodelling material ID"))->second);

    const Real current_time = par.get<Real>("current time"),
               time_step = par.get<Real>("tumour angiogenesis solver: time step")
                         * par.get<int>("tumour angiogenesis solver: angiogenesis increment");
    const unsigned int ninvars_NCS = par.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = par.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = par.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = par.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    const Real capillary_min_radius         = mi__vnr->get_biochem_param(52),
               capillary_max_radius         = mi__vnr->get_biochem_param(53),
               capillary_min_thickness      = mi__vnr->get_biochem_param(56),
               capillary_max_pore_radius    = mi__vnr->get_biochem_param(61),
               capillary_max_pores_fraction = mi__vnr->get_biochem_param(65),
               collapsed_vascular_node_survival = mi__vnr->get_biochem_param(200);
    ASSERT_(collapsed_vascular_node_survival>0 && collapsed_vascular_node_survival<100, "parameter to designate the survival propability of a just collapsed vascular node is not valid");
    const Real mean_velocity__threshold = mi__vnr->get_biochem_param(10),
               extension_min_length     = mi__vnr->get_biochem_param(14),
               sprouts_safe_margin_mean = mi__vnr->get_biochem_param(15),
               sprouts_safe_margin_std  = mi__vnr->get_biochem_param(16),
               anastomosis_tolerance    = capillary_max_radius,
               tip_extension_max_length = mi__vnr->get_biochem_param(18);
    ASSERT_(anastomosis_tolerance<=sprouts_safe_margin_mean+sprouts_safe_margin_std, "anastomosis tolerance is less/equal to the sprouts safe margin in the tumour angiogenesis solver");
    ASSERT_(extension_min_length<=par.get<Real>("finite element grid: min edge length"), "min extension length for angiogenesis algorithm is not valid");
    ASSERT_(ndcvars_RDS==9, "number of discontinuous variables adopted for the reaction-diffusion solver is invalid");
    ASSERT_(nvars_AUX==8, "number of auxiliary variables adopted for the reaction-diffusion solver is invalid");
    //
    unsigned int& branches_total_number = this->_dmn_vn.get_equation_systems().parameters.set<unsigned int>("total number of branches");

    std::set<dof_id_type> vasc_node_2_iterate_2_add;
    std::set<dof_id_type> vasc_node_2_iterate_2_remove;

    int vascular_nodes_index = 0;
    for ( std::set<dof_id_type>::const_iterator
          set_ci=vasc_node_2_iterate.begin(); set_ci!=vasc_node_2_iterate.end(); set_ci++ )
    {
        const Node* node = msh_vn.node_ptr(*set_ci);
        ASSERT_(NULL!=node, "fatal internal error occurred");
        const dof_id_type node_id = node->id();
        // a simple flag to monitor if the info of this vascular node have changed
        bool node_vi_changed = false;
        //
        status_bar(std::cout, ++vascular_nodes_index, vasc_node_2_iterate.size());
        //
        // get access to the vascular node info
        ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(node_id), "fatal internal error occurred");
        VascularInfo* vi = &(this->_dmn_vn.node_vi.find( node_id )->second);
        ASSERT_(NULL!=vi, "fatal internal error occurred");
        // also check if the flow regine for this vascular node is valid
        ASSERT_(VascularInfo::INVALID_FLOW!=vi->flow, "fatal internal error occurred");
        //
        // skip vasculogenesis if not a blood vessel
        if ( VascularInfo::BLOOD != vi->flow ) {
            vasc_node_2_iterate_2_remove.insert( node_id );
            continue;
        }
        //
        // consider to avoid extending the network if node has collapsed already
        if ( VascularInfo::COLLAPSED == vi->status ) {
            vasc_node_2_iterate_2_remove.insert( node_id );
            continue;
        }
        //
        // ignore the following if node is not part of a sprout/tip vessel
        if ( ! ( VascularInfo::SPROUT == vi->type || VascularInfo::TIP == vi->type ) ) {
            vasc_node_2_iterate_2_remove.insert( node_id );
            continue;
        }
        //
        // now (after executing the routine to find anastomoses) if this node has been flagged
        // to be ignored (only for this time step) then just skip the sunsequent computations
        if ( vasc_node_2_notiterate.end() != vasc_node_2_notiterate.find(node_id) ) {
            continue;
        }
        //
        const MaterialInfo* mi = NULL;
        const QuadratureInfo* qpi_ptr = NULL;
        {
            //
            // get access to the 3D element
            const Elem* elem3d = msh_3d.elem_ptr(vi->elem3d_id);
            ASSERT_(NULL!=elem3d, "fatal internal error occurred");
            //
            // get access to the region and material info
            ASSERT_(this->_dmn.regn.end()!=this->_dmn.regn.find(elem3d->subdomain_id()), "fatal internal error occurred");
            const RegionInfo* rgi = &(this->_dmn.regn.find(elem3d->subdomain_id())->second);
            ASSERT_(NULL!=rgi, "fatal internal error occurred");
            mi = rgi->mat;
            //
            // get access to the 3D element quadrature points data
            ASSERT_(elem3d_qpi->end()!=elem3d_qpi->find(elem3d->id()), "fatal internal error occurred");
            qpi_ptr = &(elem3d_qpi->find(elem3d->id())->second);
            ASSERT_(NULL!=qpi_ptr, "fatal internal error occurred");
            //
        }
        //
        // get the "mean velocity" at the vascular node
        const Real mean_velocity = get_system_solution_nodal_value<6>(vn_flow_sys, node, vn_flow_features);
        //
        // calculate the vascular elongation
        Displacement dX(0.0, 0.0, 0.0);
        {
            //
            // use the inverse distance algorithm to interpolate the data
            InverseDistanceAlgorithm ida(2.5);
            ida.init(qpi_ptr->XYZ);
            std::vector<Real> weight;
            ida.calc(*node, weight);
            //
            FieldInfo fld(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
            // do initialise these variables first
            fld.current_time = current_time;
            fld.time_step = time_step;
            fld.deform_gradient.comp = 0.0;
            //
            // interpolate the quadrature data to initialise field data properly
            for (unsigned int ipnt=0; ipnt<weight.size(); ipnt++) {
                for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                    fld.biochem[ivar]      += weight[ipnt] * qpi_ptr->biochem[ipnt][ivar];
                    fld.grad_biochem[ivar] += weight[ipnt] * qpi_ptr->grad_biochem[ipnt][ivar];
                    fld.biochem_rate[ivar] += weight[ipnt] * qpi_ptr->biochem_rate[ipnt][ivar];
                }
                for (unsigned int ivar=nvars_RDS; ivar<nvars_RDS+ndcvars_RDS; ivar++) {
                    fld.biochem[ivar] += weight[ipnt] * qpi_ptr->biochem[ipnt][ivar];
                }
                fld.stress.comp += weight[ipnt] * qpi_ptr->stress[ipnt].comp;
                fld.deform_gradient.comp += weight[ipnt] * qpi_ptr->deform_grad[ipnt].comp;
                for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                    fld.auxiliary[ivar]      += weight[ipnt] * qpi_ptr->auxiliary[ipnt][ivar];
                    fld.grad_auxiliary[ivar] += weight[ipnt] * qpi_ptr->grad_auxiliary[ipnt][ivar];
                    fld.auxiliary_rate[ivar] += weight[ipnt] * qpi_ptr->auxiliary_rate[ipnt][ivar];
                }
            }
            //
            // a simple way to flag the vascular node as well-perfused or not
            fld.div_velocity = ( mean_velocity < mean_velocity__threshold ? 0.0 : 1.0 );
            //
            const Vector3D reference_coords = set_vector((*node)(0), (*node)(1), (*node)(2)),
                           current_coords = mul_im_m(fld.deform_gradient, reference_coords);
            //
            dX = (*node);
            if      ( 4 == nvars_RDS && 9 == ndcvars_RDS && 8 == nvars_AUX )
                tumour_angiogenesis::growth_vascular_498(*mi, fld, *vi, dX);
            else ERROR_("fatal internal error occurred");
            //
        }
        //
        // ignore the following computations if the displacement increment is negligible
        if ( dX.norm() < extension_min_length ) libmesh_error();
        //
        // sprouting/branching vascular node coordinates and new vascular node coordinates
        const Point origin = *node;
        Point destination = origin + dX;
        //
        // before proceeding with the subsequent computations make sure the
        // new-to-add vascular node should not collapse, else skip it...
        bool new_vasc_node_has_collapsed = false;
        {
            // extension length without considering the deformation of the vascular network
            const Real extension_length = Point(destination-origin).norm();
            //
            // temporary vascular-info data
            VascularInfo vasc_info;
            vasc_info.type = VascularInfo::TIP;
            vasc_info.flow = vi->flow;
            vasc_info.time_generate = current_time;
            vasc_info.time_mature = mi__vnr->get_biochem_param(11);
            if        ( VascularInfo::BRANCH == vi->type ) {
                vasc_info.length = extension_length;
            } else {
                vasc_info.length = extension_length + vi->length;
            }
            vasc_info.total_length = extension_length + vi->total_length;
            vasc_info.radius         = capillary_min_radius         * uniform_distro(0.90, 1.10);
            vasc_info.thickness      = capillary_min_thickness      * uniform_distro(0.90, 1.10);
            vasc_info.pore_radius    = capillary_max_pore_radius    * uniform_distro(0.90, 1.10);
            vasc_info.pores_fraction = capillary_max_pores_fraction * uniform_distro(0.90, 1.10);
            vasc_info.vsc_elem.insert( msh_vn.n_elem()+new_elem_to_add.size() );
            //
            // obtain the "vascular pressure" at the vascular node
            const Real vascular_press = get_system_solution_nodal_value<0>(vn_flow_sys, node, vn_flow_features);
            // obtain the "blood viscosity" at the vascular node
            const Real blood_viscosity = get_system_solution_nodal_value<1>(vn_flow_sys, node, vn_flow_features);
            // obtain the "wall shear stress" at the vascular node
            const Real wall_shear_stress = get_system_solution_nodal_value<2>(vn_flow_sys, node, vn_flow_features);
            // obtain the (solid) "hydrostatic pressure" at the vascular node
            const Real solid_hydrostatic_press = get_system_solution_nodal_value<3>(vn_flow_sys, node, vn_flow_features);
            // obtain the (solid) "interstitial fluid pressure" at the vascular node
            const Real interstitial_fluid_press = get_system_solution_nodal_value<4>(vn_flow_sys, node, vn_flow_features);
            // obtain the "extra-vascular flow rate" at the vascular node
            //const Real extravascular_flow_rate = get_system_solution_nodal_value<5>(vn_flow_sys, node, vn_flow_features);
            // obtain the "mean velocity" at the vascular node
            const Real mean_blood_veloc = get_system_solution_nodal_value<6>(vn_flow_sys, node, vn_flow_features);
            // obtain the "effective pressure" at the vascular node
            //const Real effective_press = get_system_solution_nodal_value<7>(vn_flow_sys, node, vn_flow_features);
            // calculate the "pressure ratio" at the vascular node
            if ( vascular_press < 1.0e-12 ) {
                vasc_info.status = VascularInfo::COLLAPSED;
                vasc_info.compression_scale = 1000.0;
            } else if ( vascular_press > 1.0e+12 ) {
                libmesh_error();
            } else {
                vasc_info.press_ratio = (interstitial_fluid_press - solid_hydrostatic_press) / vascular_press;
            }
            // set the temperature loading and drug concentrations to zero (for the time being!!!)
            const Real temp_load = 0.0;
            const Real drug0_p[] = { 0.0, 0.0 };
            const Real drug1_p[] = { 0.0, 0.0 };
            const Real features[] = { current_time, vascular_press, blood_viscosity, wall_shear_stress, mean_blood_veloc, temp_load,
                                      drug0_p[0], drug0_p[1], drug1_p[0], drug1_p[1] };
            //
            // call the vessel wall remodelling routine to compute radius, thickness and pore size
            tumour_angiogenesis::remodel_vascular_XYZ(*mi__vnr, vasc_info, features);
            //
            const Real radius_min = VascularInfo::BLOOD == vasc_info.flow
                                  ? mi__vnr->get_biochem_param(52) : mi__vnr->get_biochem_param(102);
            //
            // check if this newly-created vessel node has collapsed
            if ( VascularInfo::COLLAPSED == vasc_info.status ) {
                // ...and check if it has any chance of survival
                if        ( 1000.0 != vasc_info.compression_scale ) {
                    if ( uniform_distro(0.1, 100.0) <= collapsed_vascular_node_survival )
                        vasc_info.status = VascularInfo::COMPRESSED;
                    else goto ONE_STEP_BELOW;
                } else if ( 1000.0 == vasc_info.compression_scale ) {
                    // NOTE: a conditional jump may have occurred...
                    ONE_STEP_BELOW : ;
                    //
                    vasc_info.compression_scale = 1000.0;
                    //
                    const Real new_features[] = { current_time, 0.0, 0.0, 0.0, 0.0, temp_load,
                                                  drug0_p[0], drug0_p[1], drug1_p[0], drug1_p[1] };
                    //
                    // call the vessel wall remodelling routine to set radius, thickness and pore size
                    tumour_angiogenesis::remodel_vascular_XYZ(*mi__vnr, vasc_info, new_features);
                }
            }
            // now update the lumen size of the vessel wall
            if ( VascularInfo::COLLAPSED == vasc_info.status ) {
                vasc_info.radius = radius_min / vasc_info.compression_scale;
            } else {
                vasc_info.radius = vasc_info.radius / vasc_info.compression_scale;
            }
            //
            // if new vascular node has collapsed then enforce jump
            if ( VascularInfo::COLLAPSED == vasc_info.status ) {
                new_vasc_node_has_collapsed = true;
            }
            //
        }
        //
        if ( ! new_vasc_node_has_collapsed )
        //
        {
        //
        // a flag to check if new node needs correction or not
        bool check_destination = true;
        //
        // various features that modulate branching of an existing or newly-created vessel
        const Real branching_time = mean_velocity < mean_velocity__threshold
                                  ? uniform_distro(mi->get_biochem_param(90), mi->get_biochem_param(91))
                                  : uniform_distro(mi->get_biochem_param(92), mi->get_biochem_param(93));
        const Real branching_length =  mean_velocity < mean_velocity__threshold
                                    ? uniform_distro(mi->get_biochem_param(94), mi->get_biochem_param(95))
                                    : uniform_distro(mi->get_biochem_param(96), mi->get_biochem_param(97));

        std::set<dof_id_type> conn_vsc_nodes;

        unsigned int n_conn__node = 0;
        for (std::set<dof_id_type>::const_iterator s_citer=vi->vsc_elem.begin(); s_citer!=vi->vsc_elem.end(); s_citer++) {
            dof_id_type conn_node_id = DofObject::invalid_id;
            if ( new_elem_to_add.end() == new_elem_to_add.find(*s_citer) ) {
                const Elem& elem = msh_vn.elem_ref(*s_citer);
                conn_node_id = edge2_other_node(elem, node_id);
            } else {
                const Dual<dof_id_type>& dl = new_elem_to_add.find(*s_citer)->second;
                if      ( dl.get_first()  == node_id ) conn_node_id = dl.get_second();
                else if ( dl.get_second() == node_id ) conn_node_id = dl.get_first();
                else libmesh_error();
            }
            ASSERT_(DofObject::invalid_id!=conn_node_id, "fatal internal error occurred");
            //
            if ( this->_dmn_vn.node_vi.end() != this->_dmn_vn.node_vi.find(conn_node_id) ) {
                const VascularInfo& conn_vi = this->_dmn_vn.node_vi.find(conn_node_id)->second;
                if ( VascularInfo::COLLAPSED == conn_vi.status ) continue;
            } else if ( new_node_vi_to_add.end() != new_node_vi_to_add.find(conn_node_id) ) {
                const VascularInfo& conn_vi = new_node_vi_to_add.find(conn_node_id)->second;
                if ( VascularInfo::COLLAPSED == conn_vi.status ) libmesh_error();
            } else libmesh_error();
            //
            conn_vsc_nodes.insert(conn_node_id);
            //
            ASSERT_(3>n_conn__node, "fatal internal error occurred");
            ++n_conn__node;
        }

        // set the ID of the new vascular node
        const dof_id_type current_node_id = msh_vn.n_nodes() + new_node_to_add.size();
        // set the ID of the new vascular element (segment)
        const dof_id_type current_segm_id = msh_vn.n_elem() + new_elem_to_add.size();

        if ( VascularInfo::TIP == vi->type ) {
            ASSERT_(1==n_conn__node, "fatal internal error occurred");

            check_destination = true; // set this boolean flag

            node_vi_changed = true; // set this boolean flag

            vi->type = VascularInfo::SPROUT;
            vi->vsc_elem.insert( current_segm_id );

        } else if ( VascularInfo::SPROUT == vi->type ) {
            ASSERT_(2==n_conn__node, "fatal internal error occurred");

            check_destination = false; // set this boolean flag

            // check with respect to time & length if we should branch or not; the vascular
            // segment of a vascular node starts from the corresponding initial point of the branch
            if ( 0.0 != vi->time_generate ) {
                const Real vessel_time   = current_time - vi->time_generate,
                           vessel_length = vi->length;
                //
                if ( vessel_time < branching_time || vessel_length < branching_length ) goto JUMP_END_PSR;
            }

            // check if the supposed-to-branch node is in close proximity to another branched node;
            // if yes then do not allow this vascular node to branch
            if ( 0.0 == vi->time_branch ) {
                ASSERT_(this->_dmn_vn.branch__node.end()!=this->_dmn_vn.branch__node.find(vi->branch_id), "fatal internal error occurred");
                //
                const std::set<dof_id_type>& this_branch_nodes = this->_dmn_vn.branch__node.find(vi->branch_id)->second;
                std::set<dof_id_type>::const_iterator s_citer = this_branch_nodes.begin();
                for ( ; s_citer!=this_branch_nodes.end(); s_citer++) {
                    const dof_id_type other_node_id = (*s_citer);
                    //
                    if ( other_node_id == node->id() ) continue; // skip for self-node check...
                    //
                    Point other_node_coords;
                    VascularInfo* other_node_vi = NULL;
                    if ( new_node_vi_to_add.end() != new_node_vi_to_add.find(other_node_id) ) {
                        //
                        ASSERT_(new_node_to_add.end()!=new_node_to_add.find(other_node_id), "fatal internal error occurred");
                        other_node_coords = new_node_to_add.find(other_node_id)->second;
                        other_node_vi = &(new_node_vi_to_add.find(other_node_id)->second);
                        //
                    } else {
                        //
                        ASSERT_(msh_vn.query_node_ptr(other_node_id), "fatal internal error occurred");
                        other_node_coords = msh_vn.point(other_node_id);
                        ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(other_node_id), "fatal internal error occurred");
                        other_node_vi = &(this->_dmn_vn.node_vi.find(other_node_id)->second);
                        //
                    }
                    //
                    if ( VascularInfo::COLLAPSED == other_node_vi->status ) continue; // skip if other node has collapsed...
                    if ( VascularInfo::BRANCH != other_node_vi->type ) continue; // skip if other node is not branch...
                    //
                    // avoid branching this node if there is another one that is very close
                    const Real distance = Point(other_node_coords-(*node)).norm();
                    if ( distance < branching_length ) goto JUMP_END_PSR;
                }
                //
                const Real scale = uniform_distro(0.99, 1.49); // NOTE parameters!!!
                const Point nv = Point( uniform_distro(-1.0, +1.0) ,
                                        uniform_distro(-1.0, +1.0) ,
                                        uniform_distro(-1.0, +1.0) );
                const Real dX_mag = scale * dX.norm();
                const Point dY = dX_mag * dX.cross(nv).unit();
                destination = origin + dY;

                check_destination = false; // set this boolean flag
                // ...else an exception is caught
            } else ERROR_("fatal internal error occurred");

            node_vi_changed = true; // set this boolean flag

            vi->type = VascularInfo::BRANCH;
            vi->time_branch = current_time;
            vi->length = 0.0;
            vi->vsc_elem.insert( current_segm_id );

        } else libmesh_error();

        // check for all active 3D finite elements where the new vascular node is contained
        Elem* elem3d = NULL;
        for (MeshBase::const_element_iterator e_citer=msh_3d_elements_begin; e_citer!=msh_3d_elements_end; e_citer++) {
            if ( (*e_citer)->contains_point(destination) ) {
                elem3d = (*e_citer);
                break;
            }
        }
        // if node is contained in NO element then probably got outside the analysis domain
        if ( ! elem3d ) {
            vasc_node_2_iterate_2_remove.insert( node_id );
            continue;
        }

        // extension length without considering the deformation of the vascular network
        const Real extension_length = Point(destination-origin).norm();

        // check if (previously-defined) sprout node is could be reused instead as the new tip node
        // if only the total length is not beyon a certain threshold
        if ( VascularInfo::SPROUT == vi->type && node_vi_changed ) {

            node_vi_changed = true; // set this boolean flag

            // reset the original vascular node information
            vi->type = VascularInfo::TIP;
            vi->vsc_elem.erase( vi->vsc_elem.find(current_segm_id) );

            ASSERT_(1==conn_vsc_nodes.size(), "fatal internal error occurred");
            const Node* prev_node = msh_vn.node_ptr(*conn_vsc_nodes.begin());

            const Real total_extension_length = extension_length
                                              + Point(*prev_node-origin).norm();
            //
            // check if the total elongation length is below some threshold; if yes,
            // then just elongate the existing tip vascular node
            if ( total_extension_length < tip_extension_max_length ) {
                // update the containers
                existing_node_to_update.insert( std::make_pair(node_id, destination) );
                //
                Node* this_node = this->_dmn_vn.get_mesh().node_ptr(node_id);
                (*this_node)(0) = destination(0);
                (*this_node)(1) = destination(1);
                (*this_node)(2) = destination(2);
                //
                vi->type = VascularInfo::TIP;
                vi->time_generate = current_time;
                vi->time_mature = mi__vnr->get_biochem_param(11);
                vi->length += extension_length;
                vi->total_length += extension_length;
                vi->radius         = capillary_min_radius         * uniform_distro(0.90, 1.10);
                vi->thickness      = capillary_min_thickness      * uniform_distro(0.90, 1.10);
                vi->pore_radius    = capillary_max_pore_radius    * uniform_distro(0.90, 1.10);
                vi->pores_fraction = capillary_max_pores_fraction * uniform_distro(0.90, 1.10);
                // set the 3D finite element that contains this new vascular node
                vi->elem3d_id = elem3d->id();
                // therefore no new to add a new tip vascular node in the network
                goto JUMP_END_PSR;
                // end of if case...
            }
            // bring back the vascular node to its original status
            vi->type = VascularInfo::SPROUT;
            vi->vsc_elem.insert( current_segm_id );

            // ...and continue as normal
        }

        VascularInfo vasc_info;
        //
        vasc_info.type = VascularInfo::TIP;
        vasc_info.flow = vi->flow;
        vasc_info.time_generate = current_time;
        vasc_info.time_mature = mi__vnr->get_biochem_param(11);
        if        ( VascularInfo::BRANCH == vi->type ) {
            vasc_info.length = extension_length;
        } else {
            vasc_info.length = extension_length + vi->length;
        }
        vasc_info.total_length = extension_length + vi->total_length;
        vasc_info.radius         = capillary_min_radius         * uniform_distro(0.90, 1.10);
        vasc_info.thickness      = capillary_min_thickness      * uniform_distro(0.90, 1.10);
        vasc_info.pore_radius    = capillary_max_pore_radius    * uniform_distro(0.90, 1.10);
        vasc_info.pores_fraction = capillary_max_pores_fraction * uniform_distro(0.90, 1.10);
        if        ( VascularInfo::BRANCH == vi->type ) {
            vasc_info.level = vi->level + 1;
            vasc_info.branch_id = branches_total_number;
            vasc_info.parent_branch_id = vi->branch_id;
            //
            ASSERT_(this->_dmn_vn.branch__node.end()==this->_dmn_vn.branch__node.find(vasc_info.branch_id), "fatal internal error occurred");
            std::set<dof_id_type> nodls;
            nodls.insert( node_id ); // insert also the originating node of this branch...
            nodls.insert( current_node_id );
            this->_dmn_vn.branch__node.insert( std::make_pair(vasc_info.branch_id, nodls) );
            // increment this integer indicator
            ++branches_total_number;
        } else if ( VascularInfo::SPROUT == vi->type ) {
            vasc_info.level = vi->level;
            vasc_info.branch_id = vi->branch_id;
            vasc_info.parent_branch_id = vi->parent_branch_id;
            //
            ASSERT_(this->_dmn_vn.branch__node.end()!=this->_dmn_vn.branch__node.find(vasc_info.branch_id), "fatal internal error occurred");
            this->_dmn_vn.branch__node.find(vasc_info.branch_id)->second.insert( current_node_id );
        } else libmesh_error();
        // set the 3D finite element that contains this new vascular node
        vasc_info.elem3d_id = elem3d->id();
        // initialise the containing vascular segment (element)
        vasc_info.vsc_elem.insert( current_segm_id );
        // define the vertices of the vascular segment to insert
        const Dual<dof_id_type> edge2_nodes(node_id, current_node_id);
        //
        // update the containers
        new_node_to_add.insert( std::make_pair(current_node_id, destination) );
        new_elem_to_add.insert( std::make_pair(current_segm_id, edge2_nodes) );
        new_node_vi_to_add.insert( std::make_pair(current_node_id, vasc_info) );
        // update this container too
        vasc_node_2_iterate_2_add.insert( current_node_id );
        //
        }
        //
        // NOTE: a conditional jump may have occurred...
        JUMP_END_PSR : ;
        // end of active nodes loop...
    }
    // sanity checks
    ASSERT_(new_node_to_add.size()==new_node_vi_to_add.size(), "fatal internal error occurred");
    ASSERT_(existing_elem_to_update.empty(), "fatal internal error occurred");
    //
    // now update this local container of (blood) vessel nodes to iterate on the subsequent
    // vascular network update call
    for ( std::set<dof_id_type>::const_iterator
          set_ci=vasc_node_2_iterate_2_add.begin(); set_ci!=vasc_node_2_iterate_2_add.end(); set_ci++ ) {
        vasc_node_2_iterate.insert( *set_ci );
    }
    for ( std::set<dof_id_type>::const_iterator
          set_ci=vasc_node_2_iterate_2_remove.begin(); set_ci!=vasc_node_2_iterate_2_remove.end(); set_ci++ ) {
        std::set<dof_id_type>::const_iterator set_cj = vasc_node_2_iterate.find( *set_ci );
        if ( vasc_node_2_iterate.end() != set_cj ) vasc_node_2_iterate.erase( set_cj );
    }

    this->_log.pop("_perform_sprouting");
}
//=================================================================================================
void TumourAngiogenesisSolver::_prepare_meshes () {
    this->_log.push("_prepare_meshes");

    const Parameters& par = this->_dmn.get_equation_systems().parameters;
    const int itime = par.get<int>("tumour angiogenesis solver: i-time");
    const std::string fname = par.get<std::string>("results directory: temporary") + "/"
                            + par.get<std::string>("output file name")
                            + "-" + number2string(itime+1) + "-coupled.msh";
    //
    // update the mesh of the vascular network
    flog.print("  > Prepare the vascular network mesh for use                          \r", false);
    ferr.print("  > Prepare the vascular network mesh for use                          \r", true );
    this->_dmn_vn.get_mesh().prepare_for_use(true);
    //
    // output the coupled-problem mesh
    if ( libMesh::global_processor_id() == 0 )
        GmshIO(this->_dmn_coupled.get_mesh()).write(fname.c_str());
    gcomm_ptr->barrier();
    //
    // clear the mesh to re-initialise back again
    this->_dmn_coupled.get_mesh().clear();
    gcomm_ptr->barrier();
    // read the coupled-problem mesh and then delete the Gmsh file
    GmshIO(this->_dmn_coupled.get_mesh()).read(fname.c_str());
    gcomm_ptr->barrier();
    if ( libMesh::global_processor_id() == 0 )
        remove(fname.c_str());
    //
    // prepare the coupled-problem mesh for use
    flog.print("  > Prepare the coupled-problem mesh for use                           \r", false);
    ferr.print("  > Prepare the coupled-problem mesh for use                           \r", true );
    this->_dmn_coupled.get_mesh().prepare_for_use(true);

    this->_log.pop("_prepare_meshes");
}
//=================================================================================================
void TumourAngiogenesisSolver::_reinit_systems () {
    this->_log.push("_reinit_systems");

    // re-initialize the equation systems for the vascular network
    flog.print("  > Re-initialize the vascular network associated equation systems     \r", false);
    ferr.print("  > Re-initialize the vascular network associated equation systems     \r", true );
    this->_dmn_vn.get_equation_systems().reinit();
    // ...and the coupled-problem equation systems
    flog.print("  > Re-initialize the coupled-problem equation systems                \r", false);
    ferr.print("  > Re-initialize the coupled-problem equation systems                \r", true );
    this->_dmn_coupled.get_equation_systems().reinit();
    // synchronize all processes before proceeding
    gcomm_ptr->barrier();

    this->_log.pop("_reinit_systems");
}
//=================================================================================================
void TumourAngiogenesisSolver::_project_displacement_solution () {
    this->_log.push("_project_displacement_solution");

    // calculate the displacement field of each node of the vascular network by projecting the
    // corresponding solution from the 3D solid domain
    //
    const EquationSystems& es_3d = this->_dmn.get_equation_systems();
    const System& solid_u_sys = es_3d.get_system<System>(NavierCauchySolver3D::get_label()+"::solution");
    ASSERT_(SP3D==solid_u_sys.n_vars(), "fatal internal error occurred");
    //
    std::vector<Number> displacement;
    solid_u_sys.update_global_solution(displacement);
    //
    EquationSystems& es_vn = this->_dmn_vn.get_equation_systems();
    System& vn_u_sys = es_vn.get_system<System>(this->_label+"::displacement_solution");
    ASSERT_(SP3D==vn_u_sys.n_vars(), "fatal internal error occurred");
    //
    EquationSystems& es_coupled = this->_dmn_coupled.get_equation_systems();
    System& coupled_u_sys = es_coupled.get_system<System>(this->_label+"::coupled::displacement_solution");
    ASSERT_(SP3D==coupled_u_sys.n_vars(), "fatal internal error occurred");
    //
    std::map<dof_id_type, VascularInfo>::const_iterator ci_node_vi = this->_dmn_vn.node_vi.begin();
    for ( ; ci_node_vi!=this->_dmn_vn.node_vi.end(); ci_node_vi++) {
        const Node* vn_node = es_vn.get_mesh().node_ptr(ci_node_vi->first);
        ASSERT_(NULL!=vn_node, "fatal internal error occurred");
        const Elem* elem3d = es_3d.get_mesh().elem_ptr(ci_node_vi->second.elem3d_id);
        ASSERT_(NULL!=elem3d, "fatal internal error occurred");

        std::vector<Point> nodal_X(elem3d->n_nodes());
        std::vector<Displacement> nodal_U(elem3d->n_nodes());
        for (unsigned int inode=0; inode<elem3d->n_nodes(); inode++) {
            const Node* elem_node = elem3d->node_ptr(inode);
            ASSERT_(NULL!=elem_node, "fatal internal error occurred");
            const Real u0 = displacement[elem_node->dof_number(solid_u_sys.number(), 0, 0)],
                       u1 = displacement[elem_node->dof_number(solid_u_sys.number(), 1, 0)],
                       u2 = displacement[elem_node->dof_number(solid_u_sys.number(), 2, 0)];
            nodal_X[inode] = Point(*elem_node);
            nodal_U[inode] = Displacement(u0, u1, u2);
        }

        InverseDistanceAlgorithm ida(2.5);
        ida.init(nodal_X);
        Displacement vn_U;
        ida.interp<Displacement>(*vn_node, nodal_U, vn_U);

        set_system_solution_nodal_value<0>(vn_u_sys, vn_node, vn_U(0));
        set_system_solution_nodal_value<1>(vn_u_sys, vn_node, vn_U(1));
        set_system_solution_nodal_value<2>(vn_u_sys, vn_node, vn_U(2));

        const std::map<dof_id_type, dof_id_type>::const_iterator ci_mdd = this->_dmn_coupled.vasc_node__cm_vasc_node.find(vn_node->id());
        ASSERT_(this->_dmn_coupled.vasc_node__cm_vasc_node.end()!=ci_mdd, "fatal internal error occurred");
        const dof_id_type coupled_nid = ci_mdd->second;
        const Node* coupled_node = es_coupled.get_mesh().node_ptr(coupled_nid);
        ASSERT_(NULL!=coupled_node, "fatal internal error occurred");

        set_system_solution_nodal_value<0>(coupled_u_sys, coupled_node, vn_U(0));
        set_system_solution_nodal_value<1>(coupled_u_sys, coupled_node, vn_U(1));
        set_system_solution_nodal_value<2>(coupled_u_sys, coupled_node, vn_U(2));
        // end of vascular network and coupled system `vascular' nodes loop...
    }
    //
    std::map<dof_id_type, dof_id_type>::const_iterator ci_nid = this->_dmn_coupled.cm_strm_node__strm_node.begin();
    for ( ; ci_nid!=this->_dmn_coupled.cm_strm_node__strm_node.end(); ci_nid++) {
        const Node* coupled_node = es_coupled.get_mesh().node_ptr(ci_nid->first);
        ASSERT_(NULL!=coupled_node, "fatal internal error occurred");
        const Node* node = es_3d.get_mesh().node_ptr(ci_nid->second);
        ASSERT_(NULL!=node, "fatal internal error occurred");

        const Displacement U( displacement[node->dof_number(solid_u_sys.number(), 0, 0)],
                              displacement[node->dof_number(solid_u_sys.number(), 1, 0)],
                              displacement[node->dof_number(solid_u_sys.number(), 2, 0)] );

        set_system_solution_nodal_value<0>(coupled_u_sys, coupled_node, U(0));
        set_system_solution_nodal_value<1>(coupled_u_sys, coupled_node, U(1));
        set_system_solution_nodal_value<2>(coupled_u_sys, coupled_node, U(2));
        // end of coupled system 'stroma' nodes loop...
    }
    //
    // close the solution container and update the system
    vn_u_sys.solution->close();
    vn_u_sys.update();
    // close the solution container and update the system
    coupled_u_sys.solution->close();
    coupled_u_sys.update();

    this->_log.pop("_project_displacement_solution");
}
//=================================================================================================
