/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/tumour_angiogenesis_solver.h"
#include "libmesh/linear_implicit_system.h"
#include "libmesh/gmsh_io.h"
#include "feb3/quadrature.h"
#include "feb3/ida.h"
#include "feb3/fem-context.h"
#include "feb3/paraview_io.h"
#include "feb3/tumour_angiogenesis_constitutive.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
extern // list of all systems used in the analysis
std::vector< std::pair<EquationSystems*, std::string> > systems_ls;
// a global variable
TumourAngiogenesisConstitutive ta_constitutive;
//=================================================================================================
#include "feb3/domain_data_inl.h"
#include "feb3/solution_io_inl.h"
//=================================================================================================
TumourAngiogenesisSolver::TumourAngiogenesisSolver ( DomainData& dmn,
                                                     DomainData_VascularNetwork& dmn_vn,
                                                     DomainData_SolidCoupledWithVascularNetwork& dmn_coupled ) :
    NavierCauchySolver3D(dmn), ReactionDiffusionSolverXD(dmn), DrugDeliverySolver(dmn, dmn_vn),
    _log("TumourAngiogenesisSolver"),
    _dmn(dmn), _dmn_vn(dmn_vn), _dmn_coupled(dmn_coupled) {
    // set the default label
    this->_label = "TumourAngiogenesisSolver";
}
//=================================================================================================
void TumourAngiogenesisSolver::init () {
    this->_log.push("init");

    // initialize the solvers respectively
    NavierCauchySolver3D::init();
    ReactionDiffusionSolverXD::init();
    // access the equations systems
    EquationSystems& es_3d = this->_dmn.get_equation_systems();
    EquationSystems& es_vn = this->_dmn_vn.get_equation_systems();
    EquationSystems& es_coupled = this->_dmn_coupled.get_equation_systems();
    //
    Parameters& par = es_3d.parameters;
    // set the label for the solver and the simulation title
    this->_label = par.get<std::string>("tumour angiogenesis solver: label");
    par.set<std::string>("simulation title") = this->_label;
    // Lagrange basis order
    const Order lbasis = par.get<Order>("Lagrange basis order");
    //
    char var_name[buffer_len];
    // add the basic system
    do {
        LinearImplicitSystem& sys = es_coupled.add_system<LinearImplicitSystem>(this->_label+"::coupled::pressure_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        sprintf(var_name, "%s_C__p", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        //
        sys.assemble_before_solve = false;
    } while ( false );
    // add another system
    do {
        System& sys = es_coupled.add_system<System>(this->_label+"::coupled::displacement_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<SP3D; ivar++) {
            sprintf(var_name, "%s_C__u%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name, FIRST, LAGRANGE);
        }
    } while ( false );
    // add another system
    do {
        System& sys = es_vn.add_system<System>(this->_label+"::flow_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        sprintf(var_name, "%s__pressure", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__dynamic_viscocity", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__wall_shear_stress", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__pressure_stromal", NavierCauchySolver3D::get_label().c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__pressure_interstitial", NavierCauchySolver3D::get_label().c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__Q_TRAN", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__mean_velocity", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__pressure_effective", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
    } while ( false );
    // add another system
    do {
        System& sys = es_3d.add_system<System>(NavierCauchySolver3D::get_label()+"::"+this->_label+"::coupled_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        sprintf(var_name, "%s_%s__IFP", NavierCauchySolver3D::get_label().c_str(), this->_label.c_str());
        sys.add_variable(var_name, lbasis, LAGRANGE);
        for (unsigned int ivar=0; ivar<SP3D; ivar++) {
            sprintf(var_name, "%s_%s__v%d_INTE", NavierCauchySolver3D::get_label().c_str(), this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis, LAGRANGE);
        }
        sprintf(var_name, "%s_%s__Q_INTE", NavierCauchySolver3D::get_label().c_str(), this->_label.c_str());
        sys.add_variable(var_name, lbasis, LAGRANGE);
        for (unsigned int ivar=0; ivar<SP3D; ivar++) {
            sprintf(var_name, "%s_%s__v%d_TRAN", NavierCauchySolver3D::get_label().c_str(), this->_label.c_str(), ivar);
            sys.add_variable(var_name, lbasis, LAGRANGE);
        }
        sprintf(var_name, "%s_%s__Q_TRAN", NavierCauchySolver3D::get_label().c_str(), this->_label.c_str());
        sys.add_variable(var_name, lbasis, LAGRANGE);
    } while ( false );
    // add another system
    do {
        System& sys = es_3d.add_system<System>(NavierCauchySolver3D::get_label()+"::"+this->_label+"::coupled_discontinuous_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        sprintf(var_name, "%s_%s__Volume_Stroma", NavierCauchySolver3D::get_label().c_str(), this->_label.c_str());
        sys.add_variable(var_name, CONSTANT, MONOMIAL);
        sprintf(var_name, "%s_%s__HydrPerm_BloodVessel", NavierCauchySolver3D::get_label().c_str(), this->_label.c_str());
        sys.add_variable(var_name, CONSTANT, MONOMIAL);
        sprintf(var_name, "%s_%s__HydrPerm_LymphVessel", NavierCauchySolver3D::get_label().c_str(), this->_label.c_str());
        sys.add_variable(var_name, CONSTANT, MONOMIAL);
        sprintf(var_name, "%s_%s__Surface_BloodVessel", NavierCauchySolver3D::get_label().c_str(), this->_label.c_str());
        sys.add_variable(var_name, CONSTANT, MONOMIAL);
        sprintf(var_name, "%s_%s__Surface_LymphVessel", NavierCauchySolver3D::get_label().c_str(), this->_label.c_str());
        sys.add_variable(var_name, CONSTANT, MONOMIAL);
    } while ( false );
    // add another system
    do {
        System& sys = es_vn.add_system<System>(this->_label+"::displacement_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        for (unsigned int ivar=0; ivar<SP3D; ivar++) {
            sprintf(var_name, "%s__u%d", this->_label.c_str(), ivar);
            sys.add_variable(var_name, FIRST, LAGRANGE);
        }
    } while ( false );
    // add another system
    do {
        System& sys = es_vn.add_system<System>(this->_label+"::auxiliary_solution");
        systems_ls.push_back( std::make_pair(&sys.get_equation_systems(), sys.name()) );
        //
        sprintf(var_name, "%s__level", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__flow", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__type", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__status", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__time_generate", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__time_branch", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__time_mature", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__tissue_ID", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__length", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__total_length", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__radius", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__thickness", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__pore_radius", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__pores_fraction", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__branch_ID", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__parent_branch_ID", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__pressure_ratio", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__pressure_ratio_critical", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
        sprintf(var_name, "%s__compression_scale", this->_label.c_str());
        sys.add_variable(var_name, FIRST, LAGRANGE);
    } while ( false );
    // initialize the constitutive functions
    if ( 4 == par.get<unsigned int>("reaction-diffusion solver: dimension")               &&
         9 == par.get<unsigned int>("reaction-diffusion solver: discontinuous variables") &&
         8 == par.get<unsigned int>("reaction-diffusion solver: auxiliary variables")     ) {
        ta_constitutive.blood_vessel_hydraulic_conductivity = (&blood_vessel_hydraulic_conductivity_498);
        ta_constitutive.lymph_vessel_hydraulic_conductivity = (&lymph_vessel_hydraulic_conductivity_498);
        ta_constitutive.blood_vessel_hydraulic_permeability = (&blood_vessel_hydraulic_permeability_498);
        ta_constitutive.lymph_vessel_hydraulic_permeability = (&lymph_vessel_hydraulic_permeability_498);
        ta_constitutive.vascular_hydraulic_permeability_2_drug = (&vascular_hydraulic_permeability_2_drug_498);
        ta_constitutive.vascular_reflection_coefficient_2_drug = (&vascular_reflection_coefficient_2_drug_498);
        ta_constitutive.osmotic_pressure_difference   = (&osmotic_pressure_difference_498);
        ta_constitutive.blood_vessel_stroma_tissue_ratio = (&blood_vessel_stroma_ratio_498);
        ta_constitutive.lymph_vessel_stroma_tissue_ratio = (&lymph_vessel_stroma_ratio_498);
        ta_constitutive.stroma_cross_section          = (&stroma_cross_section_498);
        ta_constitutive.stroma_hydraulic_conductivity = (&stroma_hydraulic_conductivity_498);
        ta_constitutive.stroma_diffusion_coefficient_2_drug = (&stroma_diffusion_coefficient_2_drug_498);
    } else libmesh_error();
    // initialise the drug delivery (simulation) solver
    DrugDeliverySolver::init();

    this->_log.pop("init");
}
//=================================================================================================
#ifdef _4DEBUGGING_
    #error "this shouldn't have been defined somewhere else!"
#else
    #define _4DEBUGGING_
    #undef _4DEBUGGING_
#endif
//=================================================================================================
void TumourAngiogenesisSolver::run () {
    this->_log.push("run");

    Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    ParaviewIO pvtu(this->_dmn.get_equation_systems().get_mesh()),
               pvtu_vn(this->_dmn_vn.get_equation_systems().get_mesh());
    ParaviewIO pvtu_c(this->_dmn_coupled.get_equation_systems().get_mesh());
    for (int l=0; l<par.get<int>("number of variables to hide in Paraview"); l++) {
        char var_name[buffer_len];
        sprintf(var_name, "Paraview variable #%d to hide", l);
        pvtu.hide_variable(par.get<std::string>(var_name));
        pvtu_vn.hide_variable(par.get<std::string>(var_name));
        pvtu_c.hide_variable(par.get<std::string>(var_name));
    }
    //
    char cstr[buffer_len];
    const std::string results_dir  = par.get<std::string>("results directory"),
                      output_fname = par.get<std::string>("output file name"),
                      sim_title    = par.get<std::string>("simulation title");
    Real& current_time = par.set<Real>("current time");
    const int ini_time  = par.get<int>("tumour angiogenesis solver: initial time"),
              max_times = par.get<int>("tumour angiogenesis solver: maximum time steps"),
              soln_step = par.get<int>("tumour angiogenesis solver: solution increment"),
              ang_step  = par.get<int>("tumour angiogenesis solver: angiogenesis increment"),
              rds_step  = par.get<int>("tumour angiogenesis solver: biochemics increment"),
              ncs_step  = par.get<int>("tumour angiogenesis solver: biomechanics increment"),
              dds_step  = par.get<int>("tumour angiogenesis solver: drug delivery increment"),
              drds_step = par.get<int>("tumour angiogenesis solver: drug biochemics increment"),
              rstr_step = par.get<int>("tumour angiogenesis solver: restart increment");
    const Real time_step = par.get<Real>("tumour angiogenesis solver: time step");
    par.set<int>("tumour angiogenesis solver: i-time") = ini_time;
    //
    // set the initial conditions, initialise some member objects, and the vascular network
    flog.print(" >> Solution initialization                                \n", false);
    ferr.print(" >> Solution initialization                                \n", true);
    this->_set_initial();
    //
    // write a PVD file to store results
    if ( libMesh::global_processor_id() == 0 ) {
        flog.print(" >> Generate Paraview PVD-files                            \n", false);
        sprintf( cstr, "%s/%s-%s.pvd",
                 results_dir.c_str(), output_fname.c_str(), sim_title.c_str() );
        std::ofstream fpvd(cstr);
        fpvd.setf(std::ios_base::scientific|std::ios_base::unitbuf);
        //
        fpvd << "<?xml version=\"1.0\"?>" << std::endl;
        fpvd << "<VTKFile type=\"Collection\" version=\"0.1\" byte_order=\"LittleEndian\">" << std::endl;
        fpvd << "<Collection>" << std::endl;
        for (int itime=ini_time; itime<=max_times; itime++) {
            if ( itime == ini_time || itime%soln_step == 0 || itime == max_times ) {
                const Real vtk_time = itime*time_step;
                sprintf( cstr, "paraview/%s-%s-%d.vtu",
                         output_fname.c_str(), sim_title.c_str(), itime );
                fpvd << "<DataSet timestep=\"" << vtk_time << "\" group=\"\" part=\"0\" file=\"" << cstr << "\"/>" << std::endl;
            }
            // end of time steps loop...
        }
        fpvd << "</Collection>" << std::endl;
        fpvd << "</VTKFile>" << std::endl;
    }
    //
    // write a PVD file to store results
    if ( libMesh::global_processor_id() == 0 ) {
        sprintf( cstr, "%s/%s-%s-vn.pvd",
                 results_dir.c_str(), output_fname.c_str(), sim_title.c_str() );
        std::ofstream fpvd(cstr);
        fpvd.setf(std::ios_base::scientific|std::ios_base::unitbuf);
        //
        fpvd << "<?xml version=\"1.0\"?>" << std::endl;
        fpvd << "<VTKFile type=\"Collection\" version=\"0.1\" byte_order=\"LittleEndian\">" << std::endl;
        fpvd << "<Collection>" << std::endl;
        for (int itime=ini_time; itime<=max_times; itime++) {
            if ( itime == ini_time || itime%soln_step == 0 || itime == max_times ) {
                const Real vtk_time = itime*time_step;
                sprintf( cstr, "paraview/%s-%s-vn-%d.vtu",
                         output_fname.c_str(), sim_title.c_str(), itime );
                fpvd << "<DataSet timestep=\"" << vtk_time << "\" group=\"\" part=\"0\" file=\"" << cstr << "\"/>" << std::endl;
            }
            // end of time steps loop...
        }
        fpvd << "</Collection>" << std::endl;
        fpvd << "</VTKFile>" << std::endl;
    }
    //
    // write a PVD file to store results
    #ifdef _4DEBUGGING_
    if ( libMesh::global_processor_id() == 0 ) {
        sprintf( cstr, "%s/%s-%s-c.pvd",
                 results_dir.c_str(), output_fname.c_str(), sim_title.c_str() );
        std::ofstream fpvd(cstr);
        fpvd.setf(std::ios_base::scientific|std::ios_base::unitbuf);
        //
        fpvd << "<?xml version=\"1.0\"?>" << std::endl;
        fpvd << "<VTKFile type=\"Collection\" version=\"0.1\" byte_order=\"LittleEndian\">" << std::endl;
        fpvd << "<Collection>" << std::endl;
        for (int itime=ini_time; itime<=max_times; itime++) {
            if ( itime == ini_time || itime%soln_step == 0 || itime == max_times ) {
                const Real vtk_time = itime*time_step;
                sprintf( cstr, "temporary/%s-%s-c-%d.vtu",
                         output_fname.c_str(), sim_title.c_str(), itime );
                fpvd << "<DataSet timestep=\"" << vtk_time << "\" group=\"\" part=\"0\" file=\"" << cstr << "\"/>" << std::endl;
            }
            // end of time steps loop...
        }
        fpvd << "</Collection>" << std::endl;
        fpvd << "</VTKFile>" << std::endl;
    }
    #endif
    //
    // write initial results
    do {
        flog.print("  > Save solution                                          \r", false);
        ferr.print("  > Save solution                                          \r", true);
        sprintf( cstr, "%s/paraview/%s-%s-%d.vtu",
                 results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), ini_time );
        pvtu.write_equation_systems(cstr, this->_dmn.get_equation_systems());
        //
        flog.print("  > Save vascular network solution                         \r", false);
        ferr.print("  > Save vascular network solution                         \r", true);
        sprintf( cstr, "%s/paraview/%s-%s-vn-%d.vtu",
                 results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), ini_time );
        pvtu_vn.write_equation_systems(cstr, this->_dmn_vn.get_equation_systems());
        //
        #ifdef _4DEBUGGING_
        flog.print("  > Save coupled-problem solution                          \r", false);
        ferr.print("  > Save coupled-problem solution                          \r", true);
        sprintf( cstr, "%s/temporary/%s-%s-c-%d.vtu",
                 results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), ini_time );
        pvtu_c.write_equation_systems(cstr, this->_dmn_coupled.get_equation_systems());
        #endif
        //
        flog.print("  > Save restart vascular mesh                             \r", false);
        ferr.print("  > Save restart vascular mesh                             \r", true);
        sprintf( cstr, "%s/restart/%s-%s-%d-vn.msh",
                 results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), ini_time );
        this->_dmn_vn.save_mesh(cstr);
        //
        if ( 0 == libMesh::global_processor_id() ) {
            flog.print("  > Save restart vascular network data                     \r", false);
            ferr.print("  > Save restart vascular network data                     \r", true);
            sprintf( cstr, "%s/restart/%s-%s-%d-vn.dmn",
                     results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), ini_time );
            this->_dmn_vn.save_all(cstr);
        }
    } while ( false );
    //
    // evaluate system lumped matrix
    if ( ReactionDiffusionSolverXD::par_list.get<bool>("explicit solve") ) {
        flog.print(" >> Evaluate system lumped matrix                          \n", false);
        ferr.print(" >> Evaluate system lumped matrix                          \n", true);
        ReactionDiffusionSolverXD::_diag_matrix();
    }
    //
    // run for all time steps
    ReactionDiffusionSolverXD::par_list.set<Real>("time step") = time_step * rds_step;
    NavierCauchySolver3D::par_list.set<int>("maximum increments") = max_times / ncs_step;
    const int nloads = NavierCauchySolver3D::par_list.get<int>("maximum increments"),
              niters = NavierCauchySolver3D::par_list.get<int>("maximum iterations");
    int load = 0;
    for (int itime=0; itime<ini_time; itime++) {
        // check if to run solid solver
        if ( (itime+1)%ncs_step == 0 )
            // increment this index
            ++load;
    }
    for (int itime=ini_time; itime<max_times; itime++) {
        par.set<int>("tumour angiogenesis solver: i-time") = itime;
        current_time += time_step;
        //
        // solve the system
        sprintf( cstr, " >> Time %d out of %d (%.10e)                                                 \r",
                 itime+1, max_times, current_time );
        flog.print(cstr, false);
        ferr.print(cstr, true);
        //
        // check if to run biochemics solver
        if ( (itime+1)%rds_step == 0 ) {
            sprintf( cstr, "  > Biochemics solver; time %d out of %d (%.10e)                              \r",
                     itime+1, max_times, current_time );
            flog.print(cstr, false);
            ferr.print(cstr, true);
            ReactionDiffusionSolverXD::_solve();
        }
        //
        // check if to run solid biomechanics solver
        if ( (itime+1)%ncs_step == 0 || 0 == itime ) {
            flog.print("\n", false);
            // solve the system
            NavierCauchySolver3D::par_list.set<int>("increment step") = load;
            NavierCauchySolver3D::par_list.set<Real>("previous pseudo-time") = (load+0.0)/nloads;
            NavierCauchySolver3D::par_list.set<Real>("current pseudo-time") = (load+1.0)/nloads;
            //
            // run for all iteration steps
            sprintf(cstr, "  > Pseudo-time %d out of %d\n", load+1, nloads);
            flog.print(cstr, false);
            ferr.print(cstr, true);
            for (int iter=0; iter<niters; iter++) {
                NavierCauchySolver3D::par_list.set<int>("iteration step") = iter;
                //
                // assemble the linearized system
                NavierCauchySolver3D::par_list.get<bool>("mixed formulation") ?
                NavierCauchySolver3D::_assemble_system__FWD_UP() :
                NavierCauchySolver3D::_assemble_system__FWD_U();
                //
                // solve and update solution via Newton-Raphson
                NavierCauchySolver3D::_solve();
                //
                // calculate convergence criteria
                const int status = NavierCauchySolver3D::_check_convergence();
                // check if to exit loop, abort or continue
                if      ( status == 1 ) break;
                else if ( status <  0 ) {
                    //
                    flog.print("  > Save solution                                                             \r", false);
                    ferr.print("  > Save solution                                                             \r", true);
                    sprintf( cstr, "%s/paraview/%s-%s-%d.vtu",
                             results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), itime+1 );
                    pvtu.write_equation_systems(cstr, this->_dmn.get_equation_systems());
                    //
                    flog.print("  > Save vascular network solution                                            \r", false);
                    ferr.print("  > Save vascular network solution                                            \r", true);
                    sprintf( cstr, "%s/paraview/%s-%s-vn-%d.vtu",
                             results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), itime+1 );
                    pvtu_vn.write_equation_systems(cstr, this->_dmn_vn.get_equation_systems());
                    //
                    goto END_LABEL;
                }
                // end of iterations loop...
            }
            // make sure that all processes have finished
            gcomm_ptr->barrier();
            //
            // update quadrature point set data
            flog.print("  > Update quadrature data                                                    \r", false);
            ferr.print("  > Update quadrature data                                                    \r", true);
            NavierCauchySolver3D::par_list.get<bool>("mixed formulation") ?
            NavierCauchySolver3D::_set_quad_info__FWD_UP() :
            NavierCauchySolver3D::_set_quad_info__FWD_U();
            //
            flog.print("  > Project solution to other systems                                         \r", false);
            ferr.print("  > Project solution to other systems                                         \r", true);
            this->_project_displacement_solution();
            this->_remodel_vasculature();
            //
            sprintf( cstr, "  > Biofluid-flow solver; time %d out of %d (%.10e)                           \r",
                     itime+1, max_times, current_time );
            flog.print(cstr, false);
            ferr.print(cstr, true);
            this->_fluid_solver();
            //
            flog.print("  > Remodel the vascular network                                              \r", false);
            ferr.print("  > Remodel the vascular network                                              \r", true);
            this->_remodel_vasculature();
            //
            // since the vascular network has changed ensure this parameter is set "true"
            par.set<bool>("update vascular parameters for drug delivery solver") = true;
            // increment this index
            if ( 0 != itime ) ++load;
            // end of if case...
        }
        //
        // check if to update the vascular network
        if ( (itime+1)%ang_step == 0 ) {
            //
            flog.print("  > Update the vascular network                                               \r", false);
            ferr.print("  > Update the vascular network                                               \r", true);
            this->_update_network();
            //
            sprintf( cstr, "  > Biofluid-flow solver; time %d out of %d (%.10e)                           \r",
                     itime+1, max_times, current_time );
            flog.print(cstr, false);
            ferr.print(cstr, true);
            this->_fluid_solver();
            //
            flog.print("  > Remodel the vascular network                                              \r", false);
            ferr.print("  > Remodel the vascular network                                              \r", true);
            this->_remodel_vasculature();
            //
            // since the vascular network has changed ensure this parameter is set "true"
            par.set<bool>("update vascular parameters for drug delivery solver") = true;
        }
        //
        // check if to run (intravascular) drug delivery solver
        if ( (itime+1)%dds_step == 0 || itime == ini_time ) {
            sprintf( cstr, "  > Inject the drug; time %d out of %d (%.10e)                                \r",
                     itime+1, max_times, current_time );
            flog.print(cstr, false);
            ferr.print(cstr, true);
            DrugDeliverySolver::injection();
            //
            // since the vascular network has changed ensure this parameter is set "true"
            par.set<bool>("update vascular parameters for drug delivery solver") = true;
        }
        //
        // check if to run biochemics drug delivery solver
        if ( (itime+1)%drds_step == 0 ) {
            sprintf( cstr, "  > Drug-delivery solver; time %d out of %d (%.10e)                           \r",
                     itime+1, max_times, current_time );
            flog.print(cstr, false);
            ferr.print(cstr, true);
            DrugDeliverySolver::delivery();
        }
        //
        // update (maybe again) the quadrature information
        if ( (itime+1)%rds_step == 0 ||
             (itime+1)%ncs_step == 0 ||
             (itime+1)%ang_step == 0 ||
             (itime+1)%soln_step == 0 ||
             (itime+1)%rstr_step == 0 || (itime+1) == max_times ) {
            flog.print("  > Project drug-delivery solver solution                                     \r", false);
            ferr.print("  > Project drug-delivery solver solution                                     \r", true);
            this->_project_drug_solution();
            flog.print("  > Set auxiliary and quadrature data                                         \r", false);
            ferr.print("  > Set auxiliary and quadrature data                                         \r", true);
            ReactionDiffusionSolverXD::_set_auxiliary();
            ReactionDiffusionSolverXD::_set_quad_info();
        }
        //
        if ( (itime+1)%ncs_step  == 0 ||
             (itime+1)%soln_step == 0 ||
             (itime+1)%rstr_step == 0 || (itime+1) == max_times ) {
            // calculate the gradient solution (if necessary)
            if ( NavierCauchySolver3D::par_list.get<bool>("calculate gradients") ) {
                flog.print("  > Calculate gradients                                                       \r", false);
                ferr.print("  > Calculate gradients                                                       \r", true);
                NavierCauchySolver3D::_calc_gradients();
            }
        }
        //
        // check if to write results
        if ( (itime+1)%soln_step == 0 || (itime+1) == max_times ) {
            //
            flog.print("  > Save solution                                                             \r", false);
            ferr.print("  > Save solution                                                             \r", true);
            sprintf( cstr, "%s/paraview/%s-%s-%d.vtu",
                     results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), itime+1 );
            pvtu.write_equation_systems(cstr, this->_dmn.get_equation_systems());
            //
            flog.print("  > Save vascular network solution                                            \r", false);
            ferr.print("  > Save vascular network solution                                            \r", true);
            sprintf( cstr, "%s/paraview/%s-%s-vn-%d.vtu",
                     results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), itime+1 );
            pvtu_vn.write_equation_systems(cstr, this->_dmn_vn.get_equation_systems());
            //
            #ifdef _4DEBUGGING_
            flog.print("  > Save coupled-problem solution                                             \r", false);
            ferr.print("  > Save coupled-problem solution                                             \r", true);
            sprintf( cstr, "%s/temporary/%s-%s-c-%d.vtu",
                     results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), itime+1 );
            pvtu_c.write_equation_systems(cstr, this->_dmn_coupled.get_equation_systems());
            #endif
            // make sure that all processes have finished
            gcomm_ptr->barrier();
        }
        //
        // check if solid solver has run
        if ( (itime+1)%ncs_step == 0 ) {
            // reset (zero-out) incremental solution for subsequent increment step
            NavierCauchySolver3D::_reset_increment();
        }
        //
        // check if to write restart file
        if ( (itime+1)%rstr_step == 0 || (itime+1) == max_times ) {
            flog.print("  > Save restart solution                                                     \r", false);
            ferr.print("  > Save restart solution                                                     \r", true);
            sprintf( cstr, "%s/restart/%s-%s-%d.res-%d",
                     results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), itime+1, libMesh::global_processor_id() );
            std::ofstream fout(cstr);
            this->_save_restart(fout);
            //
            flog.print("  > Save restart vascular mesh                                                \r", false);
            ferr.print("  > Save restart vascular mesh                                                \r", true);
            sprintf( cstr, "%s/restart/%s-%s-%d-vn.msh",
                     results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), itime+1 );
            this->_dmn_vn.save_mesh(cstr);
            //
            if ( 0 == libMesh::global_processor_id() ) {
                //
                flog.print("  > Save restart vascular network data                                        \r", false);
                ferr.print("  > Save restart vascular network data                                        \r", true);
                sprintf( cstr, "%s/restart/%s-%s-%d-vn.dmn",
                         results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), itime+1 );
                this->_dmn_vn.save_all(cstr);
            }
            // make sure that all processes have finished
            gcomm_ptr->barrier();
        }
        // end of time steps loop...
    }
    // end of simulation...
    END_LABEL: ;
    clear_line_(std::cout);
    flog.print(" >> Simulation terminated\n", false);

    this->_log.pop("run");
}
//=================================================================================================
//
//
//
//=================================================================================================
#include "feb3/vascular_network_aux_inl.h"
#include "feb3/vascular_network_io_inl.h"
//=================================================================================================
void TumourAngiogenesisSolver::_set_initial () {
    this->_log.push("_set_initial");

    EquationSystems& es_3d = this->_dmn.get_equation_systems();
    // initialise parameters of the solid solver
    NavierCauchySolver3D::par_list.set<bool>("inverse formulation") = false;
    NavierCauchySolver3D::par_list.set<bool>("save intermediate solution") = false;
    NavierCauchySolver3D::par_list.set<bool>("calculate gradients") = true;
    NavierCauchySolver3D::par_list.set<bool>("save updated mesh") = false;
    // set the initial conditions and initialise some member objects
    NavierCauchySolver3D::_set_initial();
    NavierCauchySolver3D::par_list.set<bool>("enforce to initialize solution") = false;
    ReactionDiffusionSolverXD::_set_initial();
    ReactionDiffusionSolverXD::par_list.set<bool>("enforce to initialize solution") = false;
    DrugDeliverySolver::set_initial();
    // initialise other systems of (or associated with) this solver
    // also initialise the vascular network and then execute the fluid solver
    const unsigned int nrf = es_3d.parameters.get<unsigned int>("input restart from file");
    const Real current_time = es_3d.parameters.get<Real>("current time");
    if ( 0 == es_3d.parameters.get<int>("tumour angiogenesis solver: i-time") ) {
        ASSERT_(0==nrf, "fatal internal error occurred");
        ASSERT_(0.0==current_time, "fatal internal error occurred");
        //
        flog.print("  > Initialise the systems                                         \r", false);
        ferr.print("  > Initialise the systems                                         \r", true );
        this->_set_other_systems();
        //
        flog.print("  > Run the fluid solver                                           \r", false);
        ferr.print("  > Run the fluid solver                                           \r", true );
        this->_fluid_solver();
        //
        flog.print("  > Initialise more systems                                        \r", false);
        ferr.print("  > Initialise more systems                                        \r", true );
        DrugDeliverySolver::injection();
        this->_set_other_systems();
    // else try to read the restart files and initialise accordingly
    } else {
        ASSERT_(0!=nrf, "fatal internal error occurred");
        ASSERT_(0.0!=current_time, "fatal internal error occurred");
        ASSERT_(libMesh::global_n_processors()==nrf, "fatal internal error occurred");
        //
        flog.print("  > Load restart files and prepare data                            \r", false);
        ferr.print("  > Load restart files and prepare data                            \r", true );
        this->_load_restart();
        this->_prepare_vascular_update();
        this->_project_displacement_solution();
        //
        flog.print("  > Project solution to other systems                              \r", false);
        ferr.print("  > Project solution to other systems                              \r", true );
        this->_set_other_systems();
        //
        flog.print("  > Run the fluid solver                                           \r", false);
        ferr.print("  > Run the fluid solver                                           \r", true );
        this->_fluid_solver();
        //
        flog.print("  > Initialise more systems                                        \r", false);
        ferr.print("  > Initialise more systems                                        \r", true );
        DrugDeliverySolver::injection();
        this->_set_other_systems();
    }
    // perform a brute-force initialisation (conditional check internally)
    // applicable for the reaction-diffusion solver systems only!
    flog.print("  > Enforced-initialise some systems                               \r", false);
    ferr.print("  > Enforced-initialise some systems                               \r", true );
    this->_brute_force_init();

    this->_log.pop("_set_initial");
}
//=================================================================================================
void TumourAngiogenesisSolver::_brute_force_init () {
    this->_log.push("_brute_force_init");

    EquationSystems& es_3d = this->_dmn.get_equation_systems();
    const Parameters& params = es_3d.parameters;
    //
    const MeshBase::const_element_iterator begin_el = es_3d.get_mesh().active_elements_begin();
    const MeshBase::const_element_iterator end_el = es_3d.get_mesh().active_elements_end();
    const MeshBase::const_node_iterator    begin_nd = es_3d.get_mesh().nodes_begin();
    const MeshBase::const_node_iterator    end_nd = es_3d.get_mesh().nodes_end();
    //
    System& rds_sys    = es_3d.get_system<System>(ReactionDiffusionSolverXD::get_label()+"::solution");
    System& rds_a_sys  = es_3d.get_system<System>(ReactionDiffusionSolverXD::get_label()+"::auxiliary_solution");
    System& rds_dc_sys = es_3d.get_system<System>(ReactionDiffusionSolverXD::get_label()+"::discontinuous_solution");
    //
    const unsigned int nvars_RDS   = params.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = params.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = params.get<unsigned int>("reaction-diffusion solver: auxiliary variables");
    ASSERT_(nvars_AUX==ReactionDiffusionSolverXD::par_list.get<unsigned int>("auxiliary variables"), "fatal internal error occurred");
    // check if to set some components of the solution from a data file
    if ( "NA" != params.get<std::string>("reaction-diffusion solver: predefined initial nodal field (file)") )
    {
        std::string name;
        std::ifstream fin;
        //
        name = params.get<std::string>("results directory") + "/temporary/"
             + params.get<std::string>("output file name") + "-node2node.txt";
        fin.open(name);
        //
        std::vector< Dual<dof_id_type> > updated_node_ID_ls;
        for (MeshBase::const_node_iterator n_citer=begin_nd; n_citer!=end_nd; n_citer++) {
            dof_id_type node_old_id, node_new_id;
            fin >> node_old_id >> node_new_id;
            //
            updated_node_ID_ls.push_back( Dual<dof_id_type>(node_new_id, node_old_id) );
        }
        //
        fin.close();
        //
        name = params.get<std::string>("reaction-diffusion solver: predefined initial nodal field (file)");
        fin.open(name);
        //
        int nfields;
        fin >> nfields;
        std::vector<std::string> field_ls(nfields);
        for (int ifield=0; ifield<nfields; ifield++) {
            fin >> field_ls[ifield];
            ASSERT_(field_ls[ifield].length()==3, "fatal internal error occurred");
        }
        //
        for (unsigned int inode=0; inode<es_3d.get_mesh().n_nodes(); inode++) {
            //
            const Node* node = es_3d.get_mesh().node_ptr( updated_node_ID_ls[inode].get_first() );
            ASSERT_(node, "fatal internal error occurred");
            ASSERT_(updated_node_ID_ls[inode].get_second()==inode, "fatal internal error occurred");
            //
            for (int ifield=0; ifield<nfields; ifield++) {
                Real field_value = 0.0;
                fin >> field_value;
                //
                char var[3];
                sprintf(var, "%c%c", field_ls[ifield][1], field_ls[ifield][2]);
                const unsigned int ivar = atoi(var);
                //
                if       ( 'v' == field_ls[ifield][0] ) {
                    ASSERT_(ivar>=0 && ivar<rds_sys.n_vars() && ivar<nvars_RDS, "fatal internal error occurred");
                    //
                    const dof_id_type idof = node->dof_number(rds_sys.number(), ivar, 0);
                    libmesh_assert( node->n_comp(rds_sys.number(), ivar) == 1 );
                    //
                    rds_sys.solution->set(idof, field_value);
                    //
                } else if ( 'a' == field_ls[ifield][0] ) {
                    ASSERT_(ivar>=0 && ivar<rds_a_sys.n_vars() && ivar<nvars_AUX, "fatal internal error occurred");
                    //
                    const dof_id_type idof = node->dof_number(rds_a_sys.number(), ivar, 0);
                    libmesh_assert( node->n_comp(rds_a_sys.number(), ivar) == 1 );
                    //
                    rds_a_sys.solution->set(idof, field_value);
                    //
                } else libmesh_error();
            }
        }
        // now close the file stream
        fin.close();
        // close solution vector and update the systems
        rds_sys.solution->close();
        rds_sys.update();
        if ( ReactionDiffusionSolverXD::par_list.get<unsigned int>("auxiliary variables") ) {
            rds_a_sys.solution->close();
            rds_a_sys.update();
        }
        // end of if-scope...
    }
    //
    ASSERT_(ndcvars_RDS==ReactionDiffusionSolverXD::par_list.get<unsigned int>("discontinuous variables"), "fatal internal error occurred");
    const unsigned int ndcvars = ReactionDiffusionSolverXD::par_list.get<unsigned int>("discontinuous variables");
    // get a reference to the Explicit system (discontinuous) variables
    unsigned int discontinuous_var[ndcvars];
    for (unsigned int l=0; l<ndcvars; l++) {
        char var_name[buffer_len];
        sprintf(var_name, "%s__C%d", ReactionDiffusionSolverXD::get_label().c_str(), l);
        discontinuous_var[l] = rds_dc_sys.variable_number(var_name);
    }
    // check if to set some components of the solution from a data file
    if ( "NA" != params.get<std::string>("reaction-diffusion solver: predefined initial elemental field (file)") )
    {
        std::string name;
        std::ifstream fin;
        //
        name = params.get<std::string>("results directory") + "/temporary/"
             + params.get<std::string>("output file name") + "-elem2elem.txt";
        fin.open(name);
        //
        std::vector< Dual<dof_id_type> > updated_elem_ID_ls;
        for (MeshBase::const_element_iterator e_citer=begin_el; e_citer!=end_el; e_citer++) {
            dof_id_type elem_old_id, elem_new_id;
            fin >> elem_old_id >> elem_new_id;
            //
            updated_elem_ID_ls.push_back( Dual<dof_id_type>(elem_new_id, elem_old_id) );
        }
        // now close the file stream
        fin.close();
        //
        name = params.get<std::string>("reaction-diffusion solver: predefined initial elemental field (file)");
        fin.open(name);
        //
        int nfields;
        fin >> nfields;
        std::vector<std::string> field_ls(nfields);
        for (int ifield=0; ifield<nfields; ifield++) {
            fin >> field_ls[ifield];
            ASSERT_(field_ls[ifield].length()==3, "fatal internal error occurred");
        }
        //
        for (unsigned int ielem=0; ielem<es_3d.get_mesh().n_elem(); ielem++) {
            //
            const Elem* elem = es_3d.get_mesh().elem_ptr( updated_elem_ID_ls[ielem].get_first() );
            ASSERT_(elem, "fatal internal error occurred");
            ASSERT_(updated_elem_ID_ls[ielem].get_second()==ielem, "fatal internal error occurred");
            //
            // local degrees of freedom for each 'gradient' variable
            std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [ndcvars];
            //
            for (int ifield=0; ifield<nfields; ifield++) {
                Real field_value = 0.0;
                fin >> field_value;
                //
                char var[3];
                sprintf(var, "%c%c", field_ls[ifield][1], field_ls[ifield][2]);
                const unsigned int ivar = atoi(var);
                //
                if ( 'C' == field_ls[ifield][0] ) {
                    ASSERT_(ivar>=0 && ivar<rds_dc_sys.n_vars(), "fatal internal error occurred");
                    rds_dc_sys.get_dof_map().dof_indices(elem, dof_indices_var[ivar], discontinuous_var[ivar]);
                    const unsigned int idof = dof_indices_var[ivar][0];
                    //
                    rds_dc_sys.solution->set(idof, field_value);
                } else libmesh_error();
            }
            // free memo
            delete [] dof_indices_var;
        }
        // now close the file stream
        fin.close();
        // close solution vector and update the system
        rds_dc_sys.solution->close();
        rds_dc_sys.update();
        // end of if-scope...
    }

    this->_log.pop("_brute_force_init");
}
//=================================================================================================
void TumourAngiogenesisSolver::_set_other_systems () {
    this->_log.push("_set_other_systems");

    flog.print("  > Run remodelling vasculature (4-B/4) [BLOCK 0]                    \r", false);
    // Project solution and miscela information from the vascular network to the 3D-domain solvers,
    // in this case the ReactionDiffusionSolverXD solver (continuous/discontinuous fields).
    EquationSystems& es_vn = this->_dmn_vn.get_equation_systems();
    EquationSystems& es_3d = this->_dmn.get_equation_systems();

    const System& vn_flow_sys = es_vn.get_system<System>(this->_label+"::flow_solution");
    System& aux_sys = es_vn.get_system<System>(this->_label+"::auxiliary_solution");
    System& rds_sys = es_3d.get_system<System>(ReactionDiffusionSolverXD::get_label()+"::solution");
    System& rds_dc_sys = es_3d.get_system<System>(ReactionDiffusionSolverXD::get_label()+"::discontinuous_solution");
    const System& vn_displ_sys = es_vn.get_system<System>(this->_label+"::displacement_solution");
    System& s3d_cpl_dc_sys = es_3d.get_system<System>(NavierCauchySolver3D::get_label()+"::"+this->_label+"::coupled_discontinuous_solution");

    const bool fix_tissue_O2_max_level = es_3d.parameters.get<bool>("fix tissue 02 maximum level");
    const bool membrane_present = es_3d.parameters.get<bool>("membrane elements present");
    const unsigned int ninvars_NCS = es_3d.parameters.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = es_3d.parameters.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = es_3d.parameters.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = es_3d.parameters.get<unsigned int>("reaction-diffusion solver: auxiliary variables"),
                       nvars_VNAUX = 19,
                       ndcvars_CPL = 5;
    ASSERT_(nvars_RDS==rds_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(ndcvars_RDS==rds_dc_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(nvars_VNAUX==aux_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(ndcvars_CPL==s3d_cpl_dc_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(8==vn_flow_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(SP3D==vn_displ_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(!membrane_present, "fatal internal error occurred");

    MeshBase::const_element_iterator msh_3d_active_local_elem_begin = es_3d.get_mesh().active_local_elements_begin();
    MeshBase::const_element_iterator msh_3d_active_local_elem_end   = es_3d.get_mesh().active_local_elements_end();
    //
    std::map<dof_id_type, QuadratureInfo>* elem_qpi = &this->_dmn.elem_qi_local;
    //std::map<dof_id_type, std::map<unsigned int, QuadratureInfo> >* elem_side_qpi = &this->_dmn.elem_side_qi_local;
    //
    ASSERT_( this->_dmn.mater.end() !=
             this->_dmn.mater.find(es_3d.parameters.get<int>("tumour angiogenesis solver: vascular remodelling material ID")), "fatal internal error occurred");
    const MaterialInfo* mi__vnr = &(this->_dmn.mater.find(es_3d.parameters.get<int>("tumour angiogenesis solver: vascular remodelling material ID"))->second);
    ASSERT_(NULL!=mi__vnr, "fatal internal error occurred");
    //
    const FEM_Context rds_ctx(rds_sys);
    const FEM_Context rds_dc_ctx(rds_dc_sys);
    const FEM_Context s3d_cpl_dc_ctx(s3d_cpl_dc_sys);

    // check if membrane elements are present in analysis
    if ( es_3d.parameters.get<bool>("membrane elements present") )
        ERROR_("this part of code is under development");

    flog.print("  > Run remodelling vasculature (4-B/4) [BLOCK 1]                    \r", false);
    // enforce zero stresses and strains at zero current time - no prestressing assumed
    if ( 0 == es_3d.parameters.get<Real>("current time") )
    // loop over all active local finite elements
    for ( MeshBase::const_element_iterator
          ce_citer=msh_3d_active_local_elem_begin; ce_citer!=msh_3d_active_local_elem_end; ce_citer++ )
    {
        const Elem* elem3d = *ce_citer;
        ASSERT_(NULL!=elem3d, "fatal internal error occurred");
        //
        // get access to the element quadrature points data
        ASSERT_(elem_qpi->end()!=elem_qpi->find(elem3d->id()), "fatal internal error occurred");
        QuadratureInfo* qpi_ptr = &(elem_qpi->find(elem3d->id())->second);
        ASSERT_(NULL!=qpi_ptr, "fatal internal error occurred");
        const unsigned int n_quad_pnt = qpi_ptr->XYZ.size();
        //
        // loop over volume integration quadrature points
        for (unsigned int qp=0; qp<n_quad_pnt; qp++) {
            qpi_ptr->deform_grad[qp] = identity_matrix;
            qpi_ptr->stress[qp] = zero_matrix;
            // end of quadrature points loop...
        }
        // end of active local elements loop...
    }
    // enforce zero stresses and strains at zero current time - no prestressing assumed
    if ( 0 == es_3d.parameters.get<Real>("current time") )
    // now compute the gradient solutions from the quadrature points
    NavierCauchySolver3D::_calc_gradients();
    //
    flog.print("  > Run remodelling vasculature (4-B/4) [BLOCK 2]                    \r", false);
    // min "blood-flow velocity" to assume a blood vessel well-perfused
    const Real perfused_vessel_min_velocity = mi__vnr->get_biochem_param(17);
    //
    std::vector<Number> ta_biochem_species;
    rds_sys.update_global_solution(ta_biochem_species);
    //
    std::vector<Number> vn_flow_features;
    vn_flow_sys.update_global_solution(vn_flow_features);
    //
    std::vector<Number> vn_displacement;
    vn_displ_sys.update_global_solution(vn_displacement);
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // check for the reaction-diffusion solver dimensionality
    flog.print("  > Run remodelling vasculature (4-B/4) [BLOCK 3]                    \r", false);
    if ( 4 == nvars_RDS && 9 == ndcvars_RDS && 8 == nvars_AUX ) {
    //
    // get a reference to the Explicit system (discontinuous) variables
    unsigned int discontinuous_var[ndcvars_RDS];
    for (unsigned int l=0; l<ndcvars_RDS; l++) {
        char var_name[buffer_len];
        sprintf(var_name, "%s__C%d", ReactionDiffusionSolverXD::get_label().c_str(), l);
        discontinuous_var[l] = rds_dc_sys.variable_number(var_name);
    }
    //
    // loop over all active local finite elements
    for ( MeshBase::const_element_iterator
          ce_citer=msh_3d_active_local_elem_begin; ce_citer!=msh_3d_active_local_elem_end; ce_citer++ )
    {
        const Elem* elem3d = *ce_citer;
        //
        // local degrees of freedom for each 'gradient' variable
        const std::vector<dof_id_type> some_vars = {0, 1, 2, 3, 4, 5, 6, 7};
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [ndcvars_RDS];
        // now set the values of discontinuous values for this element
        for (auto ivar : some_vars) {
            const unsigned int l = ivar;
            //
            rds_dc_sys.get_dof_map().dof_indices(elem3d, dof_indices_var[l], discontinuous_var[ivar]);
            //
            const unsigned int idof = dof_indices_var[l][0];
            // check if this discontinuous DoF is local...
            if ( rds_dc_sys.solution->first_local_index() <= idof && idof < rds_dc_sys.solution->last_local_index() )
                // set the corresponding initial value to zero!
                rds_dc_sys.solution->set(idof, 0.0);
        }
        // free memo
        delete [] dof_indices_var;
        // end of active local elements loop...
    }
    //
    // now iterate for all nodes of the vascular network
    for ( std::map<dof_id_type, VascularInfo>::const_iterator
          ci_node_vi=this->_dmn_vn.node_vi.begin(); ci_node_vi!=this->_dmn_vn.node_vi.end(); ci_node_vi++ )
    {
        // node of vascular network and 3D element that contains the node
        const Node* node = this->_dmn_vn.get_mesh().node_ptr(ci_node_vi->first);
        ASSERT_(NULL!=node, "fatal internal error occurred");
        const Elem* elem3d = es_3d.get_mesh().elem_ptr(ci_node_vi->second.elem3d_id);
        ASSERT_(NULL!=elem3d, "fatal internal error occurred");
        // information corresponding to this vascular node
        const VascularInfo* vi = &this->_dmn_vn.node_vi.find(ci_node_vi->first)->second;
        //
        // obtain the "mean velocity" at the vascular node
        const Real mean_velocity = get_system_solution_nodal_value<6>(vn_flow_sys, node, vn_flow_features);
        // and set the "maximum blood-flow velocity" at the vascular node
        const Real max_velocity = 2.0 * mean_velocity;
        //
        // set the "level"
        {
            const unsigned int ivar = 0;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, vi->level);
        }
        // set the "flow"
        {
            const unsigned int ivar = 1;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, vi->flow);
        }
        // set the "type"
        {
            const unsigned int ivar = 2;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, vi->type);
        }
        // set the "status"
        {
            const unsigned int ivar = 3;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, vi->status);
        }
        // set the "time_generate"
        {
            const unsigned int ivar = 4;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, vi->time_generate);
        }
        // set the "time_branch"
        {
            const unsigned int ivar = 5;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, vi->time_branch);
        }
        // set the "time_mature"
        {
            const unsigned int ivar = 6;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, vi->time_mature);
        }
        // set the "region ID"
        {
            const unsigned int ivar = 7;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, elem3d->subdomain_id());
        }
        // set the "length"
        {
            const unsigned int ivar = 8;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, vi->length);
        }
        // set the "total length"
        {
            const unsigned int ivar = 9;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, vi->total_length);
        }
        // set the "radius"
        {
            const unsigned int ivar = 10;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, vi->radius);
        }
        // set the "thickness"
        {
            const unsigned int ivar = 11;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, vi->thickness);
        }
        // set the mean "pore_radius"
        {
            const unsigned int ivar = 12;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, vi->pore_radius);
        }
        // set the mean "pores_fraction"
        {
            const unsigned int ivar = 13;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, vi->pores_fraction);
        }
        // set the "branch ID"
        {
            const unsigned int ivar = 14;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, vi->branch_id);
        }
        // set the "parent-branch ID"
        {
            const unsigned int ivar = 15;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, vi->parent_branch_id);
        }
        // set the "pressure ratio"
        {
            const unsigned int ivar = 16;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, VascularInfo::COLLAPSED == vi->status ? 1000.0 : vi->press_ratio);
        }
        // set the "pressure ratio" critical value
        {
            const unsigned int ivar = 17;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, VascularInfo::COLLAPSED == vi->status ? 0.0 : vi->press_ratio_cr);
        }
        // set the "compression_scale"
        {
            const unsigned int ivar = 18;
            const dof_id_type idof = node->dof_number(aux_sys.number(), ivar, 0);
            if ( aux_sys.solution->first_local_index() <= idof && idof < aux_sys.solution->last_local_index() )
                aux_sys.solution->set(idof, VascularInfo::COLLAPSED == vi->status ? 1000.0 : vi->compression_scale);
        }
        //
        // get access to the material info
        ASSERT_(this->_dmn.regn.end()!=this->_dmn.regn.find(elem3d->subdomain_id()), "fatal internal error occurred");
        const MaterialInfo* mi__dmn = this->_dmn.regn.find(elem3d->subdomain_id())->second.mat;
        ASSERT_(NULL!=mi__dmn, "fatal internal error occurred");
        //
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [ndcvars_RDS];
        rds_dc_ctx.get_dof_indices(elem3d, dof_indices, dof_indices_var);
        //
        // set the discontinuous state-variables
        //
        Real blood_vessel__density = 0.0;
        Real lymph_vessel__density = 0.0;
        if ( VascularInfo::BLOOD == vi->flow ) {
            const unsigned int ivar = 0;
            const unsigned int idof = dof_indices_var[ivar][0];
            // check if this discontinuous DoF is local...
            if ( rds_dc_sys.solution->first_local_index() <= idof && idof < rds_dc_sys.solution->last_local_index() ) {
                blood_vessel__density = VascularInfo::COLLAPSED != vi->status
                                      ? ( 1.0 + rds_dc_sys.solution->el(idof) )  // NOTE parameters!!!
                                      : ( 0.0 + rds_dc_sys.solution->el(idof) ); // NOTE parameters!!!
                rds_dc_sys.solution->set(idof, blood_vessel__density);
            }
        } else if ( VascularInfo::LYMPH == vi->flow ) {
            const unsigned int ivar = 1;
            const unsigned int idof = dof_indices_var[ivar][0];
            // check if this discontinuous DoF is local...
            if ( rds_dc_sys.solution->first_local_index() <= idof && idof < rds_dc_sys.solution->last_local_index() ) {
                lymph_vessel__density = VascularInfo::COLLAPSED != vi->status
                                      ? ( 1.0 + rds_dc_sys.solution->el(idof) )  // NOTE parameters!!!
                                      : ( 0.0 + rds_dc_sys.solution->el(idof) ); // NOTE parameters!!!
                rds_dc_sys.solution->set(idof, lymph_vessel__density);
            }
        } else libmesh_error();
        //
        Real blood_vessel__tip = 0.0;
        Real lymph_vessel__tip = 0.0;
        if ( VascularInfo::BLOOD == vi->flow ) {
            const unsigned int ivar = 2;
            const unsigned int idof = dof_indices_var[ivar][0];
            // check if this discontinuous DoF is local...
            if ( rds_dc_sys.solution->first_local_index() <= idof && idof < rds_dc_sys.solution->last_local_index() ) {
                blood_vessel__tip = VascularInfo::TIP == vi->type
                                  ? ( 1.0 + rds_dc_sys.solution->el(idof) )  // NOTE parameters!!!
                                  : ( 0.0 + rds_dc_sys.solution->el(idof) ); // NOTE parameters!!!
                rds_dc_sys.solution->set(idof, blood_vessel__tip);
            }
        } else if ( VascularInfo::LYMPH == vi->flow ) {
            const unsigned int ivar = 3;
            const unsigned int idof = dof_indices_var[ivar][0];
            // check if this discontinuous DoF is local...
            if ( rds_dc_sys.solution->first_local_index() <= idof && idof < rds_dc_sys.solution->last_local_index() ) {
                lymph_vessel__tip = VascularInfo::TIP == vi->type
                                  ? ( 1.0 + rds_dc_sys.solution->el(idof) )  // NOTE parameters!!!
                                  : ( 0.0 + rds_dc_sys.solution->el(idof) ); // NOTE parameters!!!
                rds_dc_sys.solution->set(idof, lymph_vessel__tip);
            }
        } else libmesh_error();
        //
        Real blood_vessel__functional = 0.0;
        Real lymph_vessel__functional = 0.0;
        if ( VascularInfo::BLOOD == vi->flow ) {
            const unsigned int ivar = 4;
            const unsigned int idof = dof_indices_var[ivar][0];
            // check if this discontinuous DoF is local...
            if ( rds_dc_sys.solution->first_local_index() <= idof && idof < rds_dc_sys.solution->last_local_index() ) {
                if ( VascularInfo::COLLAPSED == vi->status )
                    blood_vessel__functional = std::max(0.0, rds_dc_sys.solution->el(idof)); // NOTE parameters!!!
                else if ( VascularInfo::COMPRESSED == vi->status )
                    blood_vessel__functional = std::max(0.5, rds_dc_sys.solution->el(idof)); // NOTE parameters!!!
                else
                    blood_vessel__functional = std::max(1.0, rds_dc_sys.solution->el(idof)); // NOTE parameters!!!
                rds_dc_sys.solution->set(idof, blood_vessel__functional);
            }
        } else if ( VascularInfo::LYMPH == vi->flow ) {
            const unsigned int ivar = 5;
            const unsigned int idof = dof_indices_var[ivar][0];
            // check if this discontinuous DoF is local...
            if ( rds_dc_sys.solution->first_local_index() <= idof && idof < rds_dc_sys.solution->last_local_index() ) {
                if ( VascularInfo::COLLAPSED == vi->status )
                    lymph_vessel__functional = std::max(0.0, rds_dc_sys.solution->el(idof)); // NOTE parameters!!!
                else if ( VascularInfo::COMPRESSED == vi->status )
                    lymph_vessel__functional = std::max(0.5, rds_dc_sys.solution->el(idof)); // NOTE parameters!!!
                else
                    lymph_vessel__functional = std::max(1.0, rds_dc_sys.solution->el(idof)); // NOTE parameters!!!
                rds_dc_sys.solution->set(idof, lymph_vessel__functional);
            }
        } else libmesh_error();
        //
        Real blood_vessel__perfused = 0.0;
        Real lymph_vessel__perfused = 0.0;
        if ( VascularInfo::BLOOD == vi->flow ) {
            const unsigned int ivar = 6;
            const unsigned int idof = dof_indices_var[ivar][0];
            // check if this discontinuous DoF is local...
            if ( rds_dc_sys.solution->first_local_index() <= idof && idof < rds_dc_sys.solution->last_local_index() ) {
                if ( VascularInfo::COLLAPSED == vi->status )
                    blood_vessel__perfused = std::max(0.0, rds_dc_sys.solution->el(idof)); // NOTE parameters!!!
                else if ( VascularInfo::COMPRESSED == vi->status )
                    blood_vessel__perfused = std::max( max_velocity >= perfused_vessel_min_velocity ? 0.5 : 0.0, // NOTE parameters!!!
                                                       rds_dc_sys.solution->el(idof) );
                else
                    blood_vessel__perfused = std::max( max_velocity >= perfused_vessel_min_velocity ? 1.0 : 0.0, // NOTE parameters!!!
                                                       rds_dc_sys.solution->el(idof) );
                rds_dc_sys.solution->set(idof, blood_vessel__perfused);
            }
        } else if ( VascularInfo::LYMPH == vi->flow ) {
            const unsigned int ivar = 7;
            const unsigned int idof = dof_indices_var[ivar][0];
            // check if this discontinuous DoF is local...
            if ( rds_dc_sys.solution->first_local_index() <= idof && idof < rds_dc_sys.solution->last_local_index() ) {
                if ( VascularInfo::COLLAPSED == vi->status )
                    lymph_vessel__perfused = std::max(0.0, rds_dc_sys.solution->el(idof)); // NOTE parameters!!!
                else if ( VascularInfo::COMPRESSED == vi->status )
                    lymph_vessel__perfused = std::max( max_velocity >= perfused_vessel_min_velocity ? 0.5 : 0.0, // NOTE parameters!!!
                                                       rds_dc_sys.solution->el(idof) );
                else
                    lymph_vessel__perfused = std::max( max_velocity >= perfused_vessel_min_velocity ? 1.0 : 0.0, // NOTE parameters!!!
                                                       rds_dc_sys.solution->el(idof) );
                rds_dc_sys.solution->set(idof, lymph_vessel__perfused);
            }
        } else libmesh_error();
        // free memo
        delete [] dof_indices_var;
        // end of vascular network nodes loop...
    }
    //
    // ...an exception is caught!
    } else libmesh_error();
    //
    // close the solution container and update the system
    flog.print("  > Run remodelling vasculature (4-B/4) [BLOCK 4]                    \r", false);
    aux_sys.solution->close();
    aux_sys.update();
    // close the solution container and update the system
    flog.print("  > Run remodelling vasculature (4-B/4) [BLOCK 5]                    \r", false);
    rds_dc_sys.solution->close();
    rds_dc_sys.update();
    //
    flog.print("  > Run remodelling vasculature (4-B/4) [BLOCK 6]                    \r", false);
    std::vector<Number> rds_dc_soln;
    rds_dc_sys.update_global_solution(rds_dc_soln);
    //
    // now enforce for oxygen concentration to be maximum when at least one functional vascular node
    // is located inside this 3D element
    flog.print("  > Run remodelling vasculature (4-B/4) [BLOCK 7]                    \r", false);
    //***************************
    //***************************
    const Real tissue_O2_min_level = 0.0,
               tissue_O2_max_level = es_3d.parameters.get<Real>("fix tissue 02 maximum level: value");
    if ( fix_tissue_O2_max_level )
    //***************************
    //***************************
    for ( MeshBase::const_element_iterator
          ce_citer=msh_3d_active_local_elem_begin; ce_citer!=msh_3d_active_local_elem_end; ce_citer++ )
    {
        const Elem* elem3d = *ce_citer;
        //
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [ndcvars_RDS];
        rds_dc_ctx.get_dof_indices(elem3d, dof_indices, dof_indices_var);
        //
        const unsigned int ivar = 6;
        const unsigned int idof = dof_indices_var[ivar][0];
        const Real blood_vessel__perfused = rds_dc_soln[idof];
        //
        // loop for all element nodes
        for (unsigned int l=0; l<elem3d->n_nodes(); l++) {
            const unsigned int ivar = 1;
            const dof_id_type idof = elem3d->node_ptr(l)->dof_number(rds_sys.number(), ivar, 0);
            libmesh_assert( elem3d->node_ptr(l)->n_comp(rds_sys.number(), ivar) == 1 );
            //
            const Real tissue_O2_level__preset = ta_biochem_species[idof];
            const Real blood_vessel_O2__supply = blood_vessel__perfused;
            //
            const Real tissue_O2_level =
                std::max( tissue_O2_level__preset,
                          ( blood_vessel_O2__supply>0.0 ? tissue_O2_max_level : tissue_O2_min_level ) );
            // enforce for oxygen concentration to be maximum when at least one functional
            // vascular node is located inside this 3D element
            rds_sys.solution->set(idof, tissue_O2_level);
            // end of element nodes loop...
        }
        // free memo
        delete [] dof_indices_var;
        // end of active local elements loop...
    }
    //***************************
    //***************************
    //
    // close the solution container and update the system
    flog.print("  > Run remodelling vasculature (4-B/4) [BLOCK 8]                    \r", false);
    rds_sys.solution->close();
    rds_sys.update();
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    // calculate the values of the coupled discontinuous-variables system
    flog.print("  > Run remodelling vasculature (4-B/4) [BLOCK 9]                    \r", false);
    for ( MeshBase::const_element_iterator
          ce_citer=msh_3d_active_local_elem_begin; ce_citer!=msh_3d_active_local_elem_end; ce_citer++ )
    {
        const Elem* elem3d = *ce_citer;
        ASSERT_(NULL!=elem3d, "fatal internal error occurred");
        //
        std::vector<dof_id_type> dof_indices;
        std::vector<dof_id_type>* dof_indices_var = new std::vector<dof_id_type> [ndcvars_CPL];
        s3d_cpl_dc_ctx.get_dof_indices(elem3d, dof_indices, dof_indices_var);
        //
        // get access to the material info
        ASSERT_(this->_dmn.regn.end()!=this->_dmn.regn.find(elem3d->subdomain_id()), "fatal internal error occurred");
        const MaterialInfo* mi__dmn = this->_dmn.regn.find(elem3d->subdomain_id())->second.mat;
        ASSERT_(NULL!=mi__dmn, "fatal internal error occurred");
        //
        // get access to the element quadrature points data
        ASSERT_(elem_qpi->end()!=elem_qpi->find(elem3d->id()), "fatal internal error occurred");
        const QuadratureInfo* qpi_ptr = &(elem_qpi->find(elem3d->id())->second);
        ASSERT_(NULL!=qpi_ptr, "fatal internal error occurred");
        const unsigned int n_quad_pnt = qpi_ptr->XYZ.size();
        //
        Real tissue_deformed_volume = elem3d->volume();
        // compute the average deformation gradient determinant
        {
            Real det_F = 0.0;
            for (unsigned int iqp=0; iqp<n_quad_pnt; iqp++)
                det_F += determinant(qpi_ptr->deform_grad[iqp]);
            det_F /= n_quad_pnt;
            // update the tissue (finite element) volume
            tissue_deformed_volume /= det_F;
        }
        //
        FieldInfo fld_strm(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
        // compute the average biochemical fields
        {
            Point p;
            for (unsigned int iqp=0; iqp<n_quad_pnt; iqp++)
                p += qpi_ptr->xyz[iqp];
            p /= n_quad_pnt;
            //
            InverseDistanceAlgorithm ida(2.5);
            ida.init(qpi_ptr->xyz);
            //
            for (unsigned int ivar=0; ivar<nvars_RDS; ivar++) {
                std::vector<Real> v;
                for (unsigned int iqp=0; iqp<n_quad_pnt; iqp++)
                    v.push_back(qpi_ptr->biochem[iqp][ivar]);
                //
                ida.interp(p, v, fld_strm.biochem[ivar]);
            }
            //
            for (unsigned int ivar=0; ivar<nvars_AUX; ivar++) {
                std::vector<Real> a;
                for (unsigned int iqp=0; iqp<n_quad_pnt; iqp++)
                    a.push_back(qpi_ptr->auxiliary[iqp][ivar]);
                //
                ida.interp(p, a, fld_strm.auxiliary[ivar]);
            }
        }
        //
        Real blood_vessel_hydr_perm = 0.0;
        Real lymph_vessel_hydr_perm = 0.0;
        Real blood_vessel_surf_area = 0.0;
        Real lymph_vessel_surf_area = 0.0;
        // now iterate for all nodes of the vascular network
        unsigned int n_blood_vessel_nodes = 0;
        unsigned int n_lymph_vessel_nodes = 0;
        for ( std::map<dof_id_type, VascularInfo>::const_iterator
              ci_node_vi=this->_dmn_vn.node_vi.begin(); ci_node_vi!=this->_dmn_vn.node_vi.end(); ci_node_vi++ ) {
            // node of vascular network and 3D element that contains the node
            const Node* node = this->_dmn_vn.get_mesh().node_ptr(ci_node_vi->first);
            ASSERT_(NULL!=node, "fatal internal error occurred");
            // information corresponding to this vascular node
            const VascularInfo* vi = &(ci_node_vi->second);
            ASSERT_(NULL!=vi, "fatal internal error occurred");
            //
            // skip computations if this vascular node does not fall into this tissue 3D finite element
            if ( elem3d->id() != vi->elem3d_id ) continue;
            //
            // ignore the following if vascular segment has collapsed
            if ( VascularInfo::COLLAPSED == vi->status ) continue;
            //
            // obtain the "vascular pressure" at the vascular node
            /* const Real vascular_pressure = get_system_solution_nodal_value<0>(vn_flow_sys, node, vn_flow_features); */
            // obtain the "effective pressure" at the vascular node
            /* const Real effective_pressure = get_system_solution_nodal_value<7>(vn_flow_sys, node, vn_flow_features); */
            // obtain the displacement vector of this vascular node
            const Point node_U = Point( get_system_solution_nodal_value<0>(vn_displ_sys, node, vn_displacement),
                                        get_system_solution_nodal_value<1>(vn_displ_sys, node, vn_displacement),
                                        get_system_solution_nodal_value<2>(vn_displ_sys, node, vn_displacement) );
            // obtain the coordinates of the connecting (adjacent) vascular nodes
            const Node* conn_node[] = { NULL, NULL, NULL };
            // obtain the displacement vector of the connecting (adjacent) vascular nodes
            Point conn_node_U[] = { Point(), Point(), Point() };
            //
            unsigned int I = 0;
            for (std::set<dof_id_type>::const_iterator ci=vi->vsc_elem.begin(); ci!=vi->vsc_elem.end(); ci++) {
                const Elem& elem = this->_dmn_vn.get_mesh().elem_ref(*ci);
                const dof_id_type conn_node_id = edge2_other_node(elem, node->id());
                //
                ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(conn_node_id), "fatal internal error occurred");
                const VascularInfo* conn_vi = &(this->_dmn_vn.node_vi.find(conn_node_id)->second);
                ASSERT_(NULL!=conn_vi, "fatal internal error occurred");
                //
                if ( VascularInfo::COLLAPSED == conn_vi->status ) continue;
                //
                conn_node[I] = this->_dmn_vn.get_mesh().node_ptr(conn_node_id);
                ASSERT_(NULL!=conn_node[I], "fatal internal error occurred");
                conn_node_U[I] = Point( get_system_solution_nodal_value<0>(vn_displ_sys, conn_node[I], vn_displacement) ,
                                        get_system_solution_nodal_value<1>(vn_displ_sys, conn_node[I], vn_displacement) ,
                                        get_system_solution_nodal_value<2>(vn_displ_sys, conn_node[I], vn_displacement) );
                ASSERT_(3>I, "fatal internal error occurred");
                ++I;
            }
            // if this (functional) vascular node is "alone" then shut it down!
            if ( 0 == I ) continue;
            //
            FieldInfo fld_vasc(nvars_RDS+ndcvars_RDS, ninvars_NCS, nvars_AUX);
            //
            if ( VascularInfo::BLOOD == vi->flow ) {
                blood_vessel_hydr_perm += ta_constitutive.blood_vessel_hydraulic_permeability(vi, mi__vnr, mi__dmn, &fld_vasc);
                blood_vessel_surf_area += vascular_circumference(vi) * (0.5*vascular_length(node, node_U, conn_node, conn_node_U));
                // increment this index
                ++n_blood_vessel_nodes;
            }
            //
            if ( VascularInfo::LYMPH == vi->flow ) {
                lymph_vessel_hydr_perm += ta_constitutive.lymph_vessel_hydraulic_permeability(vi, mi__vnr, mi__dmn, &fld_vasc);
                lymph_vessel_surf_area += vascular_circumference(vi) * (0.5*vascular_length(node, node_U, conn_node, conn_node_U));
                // increment this index
                ++n_lymph_vessel_nodes;
            }
            //
            // end of vascular network nodes loop...
        }
        // get the average value of the following quantities (very crude!)
        if ( n_blood_vessel_nodes )
            blood_vessel_hydr_perm /= n_blood_vessel_nodes;
        if ( n_lymph_vessel_nodes )
            lymph_vessel_hydr_perm /= n_lymph_vessel_nodes;
        //
        // set the discontinuous state-variables
        {
            const unsigned int ivar = 0;
            s3d_cpl_dc_sys.solution->set(dof_indices_var[ivar][0], tissue_deformed_volume);
        }
        if ( n_blood_vessel_nodes ) {
            const unsigned int ivar = 1;
            s3d_cpl_dc_sys.solution->set(dof_indices_var[ivar][0], blood_vessel_hydr_perm);
        }
        if ( n_lymph_vessel_nodes ) {
            const unsigned int ivar = 2;
            s3d_cpl_dc_sys.solution->set(dof_indices_var[ivar][0], lymph_vessel_hydr_perm);
        }
        if ( n_blood_vessel_nodes ) {
            const unsigned int ivar = 3;
            s3d_cpl_dc_sys.solution->set(dof_indices_var[ivar][0], blood_vessel_surf_area);
        }
        if ( n_lymph_vessel_nodes ) {
            const unsigned int ivar = 4;
            s3d_cpl_dc_sys.solution->set(dof_indices_var[ivar][0], lymph_vessel_surf_area);
        }
        // free memo
        delete [] dof_indices_var;
        // end of active local elements loop...
    }
    //
    // close the solution container and update the system
    flog.print("  > Run remodelling vasculature (4-B/4) [BLOCK 10]                    \r", false);
    s3d_cpl_dc_sys.solution->close();
    s3d_cpl_dc_sys.update();
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("_set_other_systems");
}
//=================================================================================================
void TumourAngiogenesisSolver::_inspect_network () {
    this->_log.push("_inspect_network");

    MeshBase::const_node_iterator msh_vn_nodes_begin = this->_dmn_vn.get_mesh().active_nodes_begin();
    MeshBase::const_node_iterator msh_vn_nodes_end   = this->_dmn_vn.get_mesh().active_nodes_end();

    // re-iterate for all vascular nodes of the network
    for (MeshBase::const_node_iterator cn_iter=msh_vn_nodes_begin; cn_iter!=msh_vn_nodes_end; cn_iter++) {
        const Node* node = *cn_iter;
        ASSERT_(NULL!=node, "fatal internal error occurred");
        ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(node->id()), "fatal internal error occurred");
        VascularInfo* vi = &(this->_dmn_vn.node_vi.find( node->id() )->second);
        ASSERT_(NULL!=vi, "fatal internal error occurred");
        //
        if ( VascularInfo::INLET  == vi->type ||
             VascularInfo::OUTLET == vi->type ) {
            if ( vi->vsc_elem.size() != 1 ) libmesh_error();
            else continue;
        }
        //
        unsigned int n_conn_functional_vsc_nodes = 0;
        for (std::set<dof_id_type>::const_iterator ci=vi->vsc_elem.begin(); ci!=vi->vsc_elem.end(); ci++) {
            const Elem& elem = this->_dmn_vn.get_mesh().elem_ref(*ci);
            const dof_id_type conn_node_id = edge2_other_node(elem, node->id());
            //
            ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(conn_node_id), "fatal internal error occurred");
            const VascularInfo* conn_vi = &(this->_dmn_vn.node_vi.find(conn_node_id)->second);
            ASSERT_(NULL!=conn_vi, "fatal internal error occurred");
            //
            if ( VascularInfo::COLLAPSED == conn_vi->status ) continue;
            //
            ASSERT_(3>n_conn_functional_vsc_nodes, "fatal internal error occurred");
            ++n_conn_functional_vsc_nodes;
        }
        //
        // enforce to ignore "collapsing" a vascular node that is set as inlet or outlet
        if ( VascularInfo::INLET  == vi->type ||
             VascularInfo::OUTLET == vi->type ) {
            //
            if        ( 0 == n_conn_functional_vsc_nodes) {
                ERROR_("a vascular inlet/outlet has collapsed; check your results");
            } else if ( 1 == n_conn_functional_vsc_nodes ) {
                ; // nothing to do here...
            } else libmesh_error();
            //
        } else if ( VascularInfo::COLLAPSED == vi->status ) {
            //
            dof_id_type inoutlet_node_id = DofObject::invalid_id;
            std::set<dof_id_type> scanned_vasc_nodes;
            //
            find_inlet_or_outlet( this->_dmn_vn.get_mesh(), node->id(), this->_dmn_vn.node_vi,
                                  scanned_vasc_nodes, inoutlet_node_id );
            //
            if ( DofObject::invalid_id == inoutlet_node_id )
                enforce_collapse_nodes(scanned_vasc_nodes, this->_dmn_vn.node_vi);
            //
        } else {
            //
            if        ( 0 == n_conn_functional_vsc_nodes ) {
                vi->status = VascularInfo::COLLAPSED;
                vi->time_branch = 0.0;
            } else if ( 1 == n_conn_functional_vsc_nodes ) {
                vi->type = VascularInfo::TIP;
                vi->time_branch = 0.0;
            } else if ( 2 == n_conn_functional_vsc_nodes ) {
                vi->type = VascularInfo::SPROUT;
                vi->time_branch = 0.0;
            } else if ( 3 == n_conn_functional_vsc_nodes ) {
                ; // nothing to do here...
            } else libmesh_error();
            //
        }
        // end of vascular network nodes loop...
    }

    this->_log.pop("_inspect_network");
}
//=================================================================================================
void TumourAngiogenesisSolver::_check_network () {
    this->_log.push("_check_network");

    MeshBase::const_node_iterator msh_vn_nodes_begin = this->_dmn_vn.get_mesh().active_nodes_begin();
    MeshBase::const_node_iterator msh_vn_nodes_end   = this->_dmn_vn.get_mesh().active_nodes_end();
    MeshBase::const_element_iterator msh_vn_elems_begin = this->_dmn_vn.get_mesh().active_elements_begin();
    MeshBase::const_element_iterator msh_vn_elems_end   = this->_dmn_vn.get_mesh().active_elements_end();
    //
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("_check_network");
}
//=================================================================================================
void TumourAngiogenesisSolver::_remodel_vasculature () {
    this->_log.push("_remodel_vasculature");

    // Check for any of the existing or newly created blood vessels (capillaries) whether they have
    // compressed or collapsed due to cancer mass development, and compute the vessel radius.

    const System& vn_flow_sys = this->_dmn_vn.get_equation_systems().get_system<System>(this->_label+"::flow_solution");
    ASSERT_(8==vn_flow_sys.n_vars(), "fatal internal error occurred");
    //
    std::vector<Number> vn_flow_features;
    vn_flow_sys.update_global_solution(vn_flow_features);

    System& rds_dc_sys = this->_dmn.get_equation_systems().get_system<System>(ReactionDiffusionSolverXD::get_label()+"::discontinuous_solution");
    //
    // obtain the current discontinuous solution of the reaction-diffusion solver
    std::vector<Number> rds_discontinuous;
    rds_dc_sys.update_global_solution(rds_discontinuous);
    //
    const FEM_Context rds_dc_ctx(rds_dc_sys);

    System& rds_a_sys = this->_dmn.get_equation_systems().get_system<System>(ReactionDiffusionSolverXD::get_label()+"::auxiliary_solution");
    // obtain the current solution of the (interstitial) drug concentration
    std::vector<Number> drug_concentration;
    rds_a_sys.update_global_solution(drug_concentration);
    //
    const FEM_Context rds_a_ctx(rds_a_sys);

    const MeshBase::const_node_iterator msh_vn_nodes_begin = this->_dmn_vn.get_mesh().active_nodes_begin();
    const MeshBase::const_node_iterator msh_vn_nodes_end   = this->_dmn_vn.get_mesh().active_nodes_end();

    const Parameters& par = this->_dmn.get_equation_systems().parameters;

    const MaterialInfo* mi__vnr = &(this->_dmn.mater.find(par.get<int>("tumour angiogenesis solver: vascular remodelling material ID"))->second);

    const Real current_time = par.get<Real>("current time");
    const unsigned int ninvars_NCS = par.get<unsigned int>("Navier-Cauchy solver: internal variables"),
                       nvars_RDS   = par.get<unsigned int>("reaction-diffusion solver: dimension"),
                       ndcvars_RDS = par.get<unsigned int>("reaction-diffusion solver: discontinuous variables"),
                       nvars_AUX   = par.get<unsigned int>("reaction-diffusion solver: auxiliary variables");

    ASSERT_(ndcvars_RDS==9, "number of discontinuous variables adopted for the reaction-diffusion solver is invalid");
    ASSERT_(nvars_AUX==8, "number of auxiliary variables adopted for the reaction-diffusion solver is invalid");
    // a parameter to designate the survival propability of a just collapsed vascular node
    const Real collapsed_vascular_node_survival = mi__vnr->get_biochem_param(200);
    ASSERT_(collapsed_vascular_node_survival>0 && collapsed_vascular_node_survival<100, "fatal internal error occurred");

    if ( libMesh::global_processor_id() == 0 )
    // allow only the 0-rank processor to do the job
    for (int iloop=0; iloop<2; iloop++)
    {
    //
    // iterate for all vascular nodes of the network
    this->_log.push("_remodel_vasculature (TASK #1)");
    flog.print("  > Run remodelling vasculature (1/4)          \r", false);
    // iterate now for all vascular nodes
    for (MeshBase::const_node_iterator cn_iter=msh_vn_nodes_begin; cn_iter!=msh_vn_nodes_end; cn_iter++) {
        const Node* node = *cn_iter;
        ASSERT_(NULL!=node, "fatal internal error occurred");
        ASSERT_(this->_dmn_vn.node_vi.end()!=this->_dmn_vn.node_vi.find(node->id()), "fatal internal error occurred");
        VascularInfo* vi = &(this->_dmn_vn.node_vi.find( node->id() )->second);
        ASSERT_(NULL!=vi, "fatal internal error occurred");
        //
        if ( VascularInfo::INLET  == vi->type ||
             VascularInfo::OUTLET == vi->type ) continue;
        //
        // check for an uncompressed node if the "vascular pressure" is positive, else it has
        // surely collapsed and hence should be removed from the network
        if ( VascularInfo::COLLAPSED != vi->status ) {
            const Real vascular_press = get_system_solution_nodal_value<0>(vn_flow_sys, node, vn_flow_features);
            //
            if ( vascular_press <= 0.0 ) vi->status = VascularInfo::COLLAPSED;
        }
        //
        const Real radius_min = VascularInfo::BLOOD == vi->flow
                              ? mi__vnr->get_biochem_param(52) : mi__vnr->get_biochem_param(102);
        //
        // check if vascular node has already collapsed
        if ( VascularInfo::COLLAPSED == vi->status ) {
            //
            vi->compression_scale = 1000.0;
            //
            const Real temp_load = 0.0;
            //
            const Real features[] = { current_time, 0.0, 0.0, 0.0, 0.0, temp_load,
                                      0.0, 0.0, 0.0, 0.0 };
            //
            // call the vessel wall remodelling routine to set radius, thickness and pore size
            tumour_angiogenesis::remodel_vascular_XYZ(*mi__vnr, *vi, features);
            // now update the lumen size of the vessel wall
            vi->radius = radius_min / vi->compression_scale;
            //
        // ...else it is under physiological conditions...
        } else {
            //
            std::vector<dof_id_type> dof_indices;
            std::vector<dof_id_type>* dof_indices_dc_var = new std::vector<dof_id_type> [ndcvars_RDS];
            std::vector<dof_id_type>* dof_indices_a_var = new std::vector<dof_id_type> [nvars_AUX];
            //
            const Elem* elem3d = this->_dmn.get_mesh().elem_ptr(vi->elem3d_id);
            ASSERT_(NULL!=elem3d, "fatal internal error occurred");
            //
            rds_dc_ctx.get_dof_indices(elem3d, dof_indices, dof_indices_dc_var);
            rds_a_ctx.get_dof_indices(elem3d, dof_indices, dof_indices_a_var);
            //
            const Real temp_load = rds_discontinuous[dof_indices_dc_var[8][0]];
            //
            Real drug0_p[] = { 0.0, 0.0 };
            Real drug1_p[] = { 0.0, 0.0 };
            for (unsigned int p=0; p<2; p++) {
                for (unsigned int l=0; l<elem3d->n_nodes(); l++) {
                    drug0_p[p] += drug_concentration[dof_indices_a_var[p+0][l]];
                    drug1_p[p] += drug_concentration[dof_indices_a_var[p+4][l]];
                }
                drug0_p[p] = apply_lbound(0.0, drug0_p[p]/elem3d->n_nodes());
                drug1_p[p] = apply_lbound(0.0, drug1_p[p]/elem3d->n_nodes());
            }
            //
            // obtain the "vascular pressure" at the vascular node
            const Real vascular_press = get_system_solution_nodal_value<0>(vn_flow_sys, node, vn_flow_features);
            // obtain the "blood viscosity" at the vascular node
            const Real blood_viscosity = get_system_solution_nodal_value<1>(vn_flow_sys, node, vn_flow_features);
            // obtain the "wall shear stress" at the vascular node
            const Real wall_shear_stress = get_system_solution_nodal_value<2>(vn_flow_sys, node, vn_flow_features);
            // obtain the (solid) "hydrostatic pressure" at the vascular node
            const Real solid_hydrostatic_press = get_system_solution_nodal_value<3>(vn_flow_sys, node, vn_flow_features);
            // obtain the (solid) "interstitial fluid pressure" at the vascular node
            const Real interstitial_fluid_press = get_system_solution_nodal_value<4>(vn_flow_sys, node, vn_flow_features);
            // obtain the "extra-vascular flow rate" at the vascular node
            /* const Real extravascular_flow_rate = get_system_solution_nodal_value<5>(vn_flow_sys, node, vn_flow_features); */
            // obtain the "mean velocity" at the vascular node
            const Real mean_velocity = get_system_solution_nodal_value<6>(vn_flow_sys, node, vn_flow_features);
            // obtain the "effective pressure" at the vascular node
            /* const Real effective_press = get_system_solution_nodal_value<7>(vn_flow_sys, node, vn_flow_features); */
            // calculate the "pressure ratio" at the vascular node
            if ( vascular_press < 1.0e-12 ) {
                vi->status = VascularInfo::COLLAPSED;
                vi->compression_scale = 1000.0;
            } else if ( vascular_press > 1.0e+12 ) {
                libmesh_error();
            } else {
                vi->press_ratio = (interstitial_fluid_press - solid_hydrostatic_press) / vascular_press;
            }
            //
            const Real features[] = { current_time, vascular_press, blood_viscosity, wall_shear_stress, mean_velocity, temp_load,
                                      drug0_p[0], drug0_p[1], drug1_p[0], drug1_p[1] };
            //
            // call the vessel wall remodelling routine to compute radius, thickness and pore size
            tumour_angiogenesis::remodel_vascular_XYZ(*mi__vnr, *vi, features);
            //
            // check if this newly collapsed vessel node has any chance of survival
            if ( VascularInfo::COLLAPSED == vi->status ) {
                if        ( 1000.0 != vi->compression_scale ) {
                    if ( uniform_distro(0.1, 100.0) <= collapsed_vascular_node_survival )
                        vi->status = VascularInfo::COMPRESSED;
                    else goto ONE_STEP_BELOW;
                } else if ( 1000.0 == vi->compression_scale ) {
                    // NOTE: a conditional jump may have occurred...
                    ONE_STEP_BELOW : ;
                    //
                    vi->compression_scale = 1000.0;
                    //
                    const Real new_features[] = { current_time, 0.0, 0.0, 0.0, 0.0, temp_load,
                                                  0.0, 0.0, 0.0, 0.0 };
                    //
                    // call the vessel wall remodelling routine to set radius, thickness and pore size
                    tumour_angiogenesis::remodel_vascular_XYZ(*mi__vnr, *vi, new_features);
                }
            }
            // now update the lumen size of the vessel wall
            if ( VascularInfo::COLLAPSED == vi->status && 1000.0 == vi->compression_scale ) {
                vi->radius = radius_min / vi->compression_scale;
            } else if ( VascularInfo::COLLAPSED == vi->status || 1000.0 == vi->compression_scale ) {
                libmesh_error();
            } else {
                vi->radius = vi->radius / vi->compression_scale;
            }
            // free memo
            delete [] dof_indices_dc_var;
            delete [] dof_indices_a_var;
            // end if case...
        }
        //
        // check if vascular node has collapsed now
        if ( VascularInfo::COLLAPSED == vi->status ) {
            // enforce to remove from the list of branches this collapsed vascular node
            for ( std::map<dof_id_type, std::set<dof_id_type> >::iterator
                  it_b2n=this->_dmn_vn.branch__node.begin(); it_b2n!=this->_dmn_vn.branch__node.end(); it_b2n++ ) {
                std::set<dof_id_type>* branch__vascular_nodes = &(it_b2n->second);
                //
                std::set<dof_id_type>::iterator it = branch__vascular_nodes->find( node->id() );
                if ( branch__vascular_nodes->end() != it ) branch__vascular_nodes->erase(it);
                // end of branches loop...
            }
            // end if case...
        }
        // end of vascular network nodes loop...
    }
    this->_log.pop("_remodel_vasculature (TASK #1)");

    // inspect the vascular nodes of the network wrt to their state
    this->_log.push("_remodel_vasculature (TASK #2)");
    flog.print("  > Run remodelling vasculature (2/4)          \r", false);
    this->_inspect_network();
    this->_log.pop("_remodel_vasculature (TASK #2)");
    //
    // end of loop for 0-rank processor...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    // distribute the changes of the vascular-netwrk nodal points due to the above remodelling update
    this->_log.push("_remodel_vasculature (TASK #3)");
    flog.print("  > Run remodelling vasculature (3/4)          \r", false);
    if ( libMesh::global_n_processors() > 1 )
    {
        char name[buffer_len];
        sprintf( name, "%s/temporary/%s.vn.dat",
                 par.get<std::string>("results directory").c_str(),
                 par.get<std::string>("output file name").c_str() );
        //
        // check also 3rd task in the "_update_network" routine for correspondence
        std::map<dof_id_type, Point>                  tmp_map_1;
        std::map<dof_id_type, Dual<dof_id_type> >     tmp_map_2;
        std::map<dof_id_type, VascularInfo>           tmp_map_3;
        std::map<dof_id_type, Point>                  tmp_map_4;
        std::map<dof_id_type, Dual<dof_id_type> >     tmp_map_5;
        std::map<dof_id_type, VascularInfo>         & tmp_map_6 = this->_dmn_vn.node_vi;
        std::map<dof_id_type, std::set<dof_id_type> > tmp_map_7;
        //
        flog.print("  > Run remodelling vasculature (3-A/4)          \r", false);
        if ( libMesh::global_processor_id() == 0 ) {
            std::ofstream file(name);
            write_current_vascular_network( tmp_map_1, tmp_map_2, tmp_map_3,
                                            tmp_map_4, tmp_map_5, tmp_map_6,
                                            tmp_map_7, file );
        }
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        //
        flog.print("  > Run remodelling vasculature (3-B/4)          \r", false);
        if ( libMesh::global_processor_id() != 0 ) {
            std::ifstream file(name);
            read_current_vascular_network( tmp_map_1, tmp_map_2, tmp_map_3,
                                           tmp_map_4, tmp_map_5, tmp_map_6,
                                           tmp_map_7, file );
        }
        // make sure that all processes have finished
        gcomm_ptr->barrier();
        // end of if-case...
    }
    this->_log.pop("_remodel_vasculature (TASK #3)");

    // now distribute the vascular network calculations to the corresponding systems
    this->_log.push("_remodel_vasculature (TASK #4)");
    flog.print("  > Run remodelling vasculature (4-A/4)          \r", false);
    this->_check_network();
    flog.print("  > Run remodelling vasculature (4-B/4)          \r", false);
    this->_set_other_systems();
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    this->_log.pop("_remodel_vasculature (TASK #4)");

    this->_log.pop("_remodel_vasculature");
}
//=================================================================================================
void TumourAngiogenesisSolver::_save_restart (std::ofstream& fout) const {
    this->_log.push("_save_restart");

    for (unsigned isys=0; isys<systems_ls.size(); isys++) {
        const EquationSystems* es_ptr = systems_ls[isys].first;
        const System& sys = es_ptr->get_system<System>(systems_ls[isys].second);
        //
        sys.solution->close();
        std::vector<Number> soln;
        sys.solution->localize_to_one(soln);
        //
        if ( libMesh::global_processor_id() ) continue;
        //
        const unsigned int n = soln.size();
        fout.write((char*)&n, sizeof(unsigned int));
        for (unsigned int i=0; i<n; i++) {
            const Number component = soln[i];
            fout.write((char*)&component, sizeof(Number));
        }
    }
    //
    this->_dmn.save_quadrature(fout, QuadratureInfo::ALL, QuadratureInfo::LEVEL_3);
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("_save_restart");
}
//=================================================================================================
void TumourAngiogenesisSolver::_load_restart () {
    this->_log.push("_load_restart");

    const Parameters& par = this->_dmn.get_equation_systems().parameters;
    //
    for (unsigned int ifile=0; ifile<par.get<unsigned int>("input restart from file"); ifile++) {
        const std::string str_name = "input restart file name #" + std::to_string(ifile);
        const std::string restart_file_name = par.get<std::string>(str_name.c_str());
        //
        std::ifstream fin(restart_file_name.c_str());
        // read the system solution from the first restart file only
        if ( 0 == ifile )
            for (unsigned isys=0; isys<systems_ls.size(); isys++) {
                EquationSystems* es_ptr = systems_ls[isys].first;
                System& sys = es_ptr->get_system<System>(systems_ls[isys].second);
                //
                unsigned int n = 0;
                fin.read((char*)&n, sizeof(unsigned int));
                for (unsigned int i=0; i<n; i++) {
                    Number component;
                    fin.read((char*)&component, sizeof(Number));
                    sys.solution->set(i, component);
                }
            }
        // read Quadrature data for this processor only
        if ( libMesh::global_processor_id() == ifile )
            this->_dmn.load_quadrature(fin, QuadratureInfo::ALL, QuadratureInfo::LEVEL_3);
        // end of restart files loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    //
    for (unsigned isys=0; isys<systems_ls.size(); isys++) {
        EquationSystems* es_ptr = systems_ls[isys].first;
        System& sys = es_ptr->get_system<System>(systems_ls[isys].second);
        //
        sys.solution->close();
        sys.update();
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("_load_restart");
}
//=================================================================================================
void TumourAngiogenesisSolver::_project_drug_solution () {
    this->_log.push("_project_drug_solution");

    EquationSystems& es_3d = this->_dmn.get_equation_systems();

    System& rds_a_sys = es_3d.get_system<System>(ReactionDiffusionSolverXD::get_label()+"::auxiliary_solution");
    System& rds_DaDt_sys = es_3d.get_system<System>(ReactionDiffusionSolverXD::get_label()+"::auxiliary_solution_time_derivative");

    const System& dd_g_sys = es_3d.get_system<System>(DrugDeliverySolver::get_label()+"::solution");
    const System& dd_DgDt_sys = es_3d.get_system<System>(DrugDeliverySolver::get_label()+"::solution_time_derivative");

    ASSERT_(dd_g_sys.n_vars()==rds_a_sys.n_vars(), "fatal internal error occurred");
    ASSERT_(dd_DgDt_sys.n_vars()==rds_DaDt_sys.n_vars(), "fatal internal error occurred");

    ASSERT_( es_3d.parameters.get<unsigned int>("reaction-diffusion solver: dimension") == 4,
             "adopted dimension for the reaction-diffusion solver is invalid" );
    ASSERT_( es_3d.parameters.get<unsigned int>("reaction-diffusion solver: discontinuous variables") == 9,
             "number of discontinuous variables inserted for the reaction-diffusion solver is invalid" );
    ASSERT_( es_3d.parameters.get<unsigned int>("reaction-diffusion solver: auxiliary variables") == 8,
             "number of auxiliary variables adopted for the reaction-diffusion solver is invalid" );

    // close solution vectors first
    dd_g_sys.solution->close();
    dd_DgDt_sys.solution->close();
    // copy & close solution vector and update the system
    rds_a_sys.solution->operator=(*dd_g_sys.solution);
    rds_a_sys.solution->close();
    rds_a_sys.update();
    // copy & close solution vector and update the system
    rds_DaDt_sys.solution->operator=(*dd_DgDt_sys.solution);
    rds_DaDt_sys.solution->close();
    rds_DaDt_sys.update();
    // make sure that all processes have finished
    gcomm_ptr->barrier();

    this->_log.pop("_project_drug_solution");
}
//=================================================================================================
