/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "libmesh/equation_systems.h"
#include "libmesh/mesh.h"
#include "feb3/rbf.h"
#include "feb3/domain_data.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
//=================================================================================================
void restore_mesh (DomainData& dmn, const std::vector< std::vector<Real> >& solu_u, const IndexPair& solu_idx) {
    MeshBase& msh = dmn.get_equation_systems().get_mesh();
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // loop over all nodes
    MeshBase::node_iterator iter_nd_end = msh.nodes_end();
    for ( MeshBase::node_iterator
          iter_nd=msh.nodes_begin(); iter_nd!=iter_nd_end; iter_nd++ ) {
        const unsigned int node_id = (*iter_nd)->id();
        const unsigned int index = solu_idx.get_index(node_id);
        libmesh_assert( node_id == solu_idx.get_id(index) );
        //
        Node* node_ptr = msh.node_ptr(node_id);
        //
        node_ptr->operator()(0) -= solu_u[index][0];
        node_ptr->operator()(1) -= solu_u[index][1];
        node_ptr->operator()(2) -= solu_u[index][2];
        // end of nodes list loop...
    }
    // make sure that all processes have finished
    gcomm_ptr->barrier();
}
//=================================================================================================
#define BOUNDARY_NODES_REMESH
// #undef  BOUNDARY_NODES_REMESH
#define HEALTHY_DOMAIN_WITH_RBF_REMESHING
// #undef  HEALTHY_DOMAIN_WITH_RBF_REMESHING
//=================================================================================================
void update_mesh_rbf (DomainData& dmn, std::vector< std::vector<Real> >& solu_u, const IndexPair& solu_idx) {
    MeshBase& msh = dmn.get_equation_systems().get_mesh();
    // list of nodes scanned (IDs) and some more info too
    std::set<unsigned int> boundary_nodes_scanned;
    std::vector<Point> boundary_nodes_xyz;
    std::vector<Real>  boundary_nodes_u[SP3D];
    // loop over all active finite elements
    for ( MeshBase::element_iterator
          citer_el=msh.active_elements_begin(); citer_el!=msh.active_elements_end(); citer_el++ ) {
        const Elem* elem = *citer_el;
        // run for all element sides
        for (unsigned int s=0; s<elem->n_sides(); s++) {
            if ( elem->neighbor_ptr(s) != NULL ) continue;
            // create a side element
            UniquePtr<const Elem> side = elem->build_side_ptr(s);
            // find the boundary nodes to this element
            std::vector<unsigned int> element_boundary_nodes;
            for (unsigned int ns=0; ns<side->n_nodes(); ns++) {
                bool is_found = false;
                for (unsigned int ne=0; ne<elem->n_nodes() && !is_found; ne++) {
                    if ( elem->node_id(ne) == side->node_id(ns) ) {
                        element_boundary_nodes.push_back( ne );
                        is_found = true;
                    }
                }
                libmesh_assert( is_found );
            }
            const unsigned int n_b_nodes = element_boundary_nodes.size();
            if ( n_b_nodes != side->n_nodes() ) libmesh_error();
            //
            for (unsigned int l=0; l<n_b_nodes; l++) {
                const unsigned int ne = element_boundary_nodes[l];
                const unsigned int node_id = elem->node_id(ne);
                const unsigned int index = solu_idx.get_index(node_id);
                //
                if ( boundary_nodes_scanned.end() != boundary_nodes_scanned.find(node_id) ) continue;
                else boundary_nodes_scanned.insert(node_id);
                //
                Node* node_ptr = msh.node_ptr(node_id);
                boundary_nodes_xyz.push_back( *node_ptr );
                boundary_nodes_u[0].push_back( solu_u[index][0] );
                boundary_nodes_u[1].push_back( solu_u[index][1] );
                boundary_nodes_u[2].push_back( solu_u[index][2] );
            }
            // end of element sides loop...
        }
        // end of active elements loop...
    }
    // list of nodes scanned (IDs) and some more info too
    std::set<unsigned int> interface_nodes_scanned;
    std::vector<Point> interface_nodes_xyz;
    std::vector<Real>  interface_nodes_u[SP3D];
    // loop over interface nodes
    for ( std::set<unsigned int>::const_iterator
          citer_nd=dmn.interface_node.begin(); citer_nd!=dmn.interface_node.end(); citer_nd++ ) {
        const unsigned int node_id = *citer_nd;
        const unsigned int index = solu_idx.get_index(node_id);
        //
        if ( boundary_nodes_scanned.end() != boundary_nodes_scanned.find(node_id) ) continue;
        if ( interface_nodes_scanned.end() != interface_nodes_scanned.find(node_id) ) continue;
        else interface_nodes_scanned.insert(node_id);
        //
        Node* node_ptr = msh.node_ptr(node_id);
        interface_nodes_xyz.push_back( *node_ptr );
        interface_nodes_u[0].push_back( solu_u[index][0] );
        interface_nodes_u[1].push_back( solu_u[index][1] );
        interface_nodes_u[2].push_back( solu_u[index][2] );
        // end of interface nodes loop...
    }
    // list of nodes scanned (IDs) and some more info too
    std::set<unsigned int> healthy_nodes_scanned;
    // loop over healthy nodes
    #ifndef HEALTHY_DOMAIN_WITH_RBF_REMESHING
    for ( std::set<unsigned int>::const_iterator
          citer_nd=dmn.healthy_node.begin(); citer_nd!=dmn.healthy_node.end(); citer_nd++ ) {
        const unsigned int node_id = *citer_nd;
        Node* node_ptr = msh.node_ptr(node_id);
        //
        if ( boundary_nodes_scanned.end() != boundary_nodes_scanned.find(node_id) ) continue;
        if ( healthy_nodes_scanned.end() != healthy_nodes_scanned.find(node_id) ) continue;
        else healthy_nodes_scanned.insert(node_id);
        // end of healthy nodes loop...
    }
    #endif
    // make sure that all processes have finished
    gcomm_ptr->barrier();
    // create the RBF-mesh update object
    RBFmap3D rbf(RBF_MQ, PO_BILINEAR);
    // set the parameters for the radial basis functions
    const Real rbf_parameters[] = { 0.0001, -0.35 };
    // initialize the RBF-mesh update object
    #ifdef BOUNDARY_NODES_REMESH
    rbf.init(rbf_parameters, boundary_nodes_xyz, boundary_nodes_u);
    #else
    rbf.init(rbf_parameters, interface_nodes_xyz, interface_nodes_u);
    #endif
    // loop over all nodes
    MeshBase::node_iterator iter_nd_end = msh.nodes_end();
    for ( MeshBase::node_iterator
          iter_nd=msh.nodes_begin(); iter_nd!=iter_nd_end; iter_nd++ ) {
        const unsigned int node_id = (*iter_nd)->id();
        const unsigned int index = solu_idx.get_index(node_id);
        //
        Node* node_ptr = msh.node_ptr(node_id);
        //
        if ( boundary_nodes_scanned.end() != boundary_nodes_scanned.find(node_id) ||
             healthy_nodes_scanned.end() != healthy_nodes_scanned.find(node_id)  ||
             interface_nodes_scanned.end() != interface_nodes_scanned.find(node_id) ) {
            //
            node_ptr->operator()(0) += solu_u[index][0];
            node_ptr->operator()(1) += solu_u[index][1];
            node_ptr->operator()(2) += solu_u[index][2];
        } else {
            const Point pnt = (*node_ptr);
            Real soln[] = { 0.0, 0.0, 0.0 };
            rbf.calc(pnt, soln);
            //
            node_ptr->operator()(0) += soln[0];
            node_ptr->operator()(1) += soln[1];
            node_ptr->operator()(2) += soln[2];
            //
            solu_u[index][0] = soln[0];
            solu_u[index][1] = soln[1];
            solu_u[index][2] = soln[2];
        }
        // end of nodes list loop...
    }
}
//=================================================================================================
#undef  BOUNDARY_NODES_REMESH
#undef  HEALTHY_DOMAIN_WITH_RBF_REMESHING
//=================================================================================================
void update_mesh_lagrange (DomainData& dmn, std::vector< std::vector<Real> >& solu_u, const IndexPair& solu_idx) {
    MeshBase& msh = dmn.get_equation_systems().get_mesh();
    // loop over all nodes
    MeshBase::node_iterator iter_nd_end = msh.nodes_end();
    for ( MeshBase::node_iterator
          iter_nd=msh.nodes_begin(); iter_nd!=iter_nd_end; iter_nd++ ) {
        const unsigned int node_id = (*iter_nd)->id();
        const unsigned int index = solu_idx.get_index(node_id);
        //
        Node* node_ptr = msh.node_ptr(node_id);
        //
        node_ptr->operator()(0) += solu_u[index][0];
        node_ptr->operator()(1) += solu_u[index][1];
        node_ptr->operator()(2) += solu_u[index][2];
        // end of nodes list loop...
    }
}
//=================================================================================================
