/*************************************************************************************************

    Copyright (C) 2009-2019    Peter Wijeratne (pwijeratne@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include <iostream>
#include <cmath>
#include <random>
#include <sstream>
#include <fstream>
#include <map>
//=================================================================================================
static
void _gen_cap_net ( const double& L, const double& R, const unsigned int& N,
                    const bool& bifurcate, const int& vessel_arrangement,
                    const bool& exclude, const bool& mix );
static
std::map<int, std::vector<double> > propSeg (double& x0, double& y0, double& z0,
					    std::default_random_engine& rgen,
					    std::uniform_real_distribution<double>& udist_x,
					    std::uniform_real_distribution<double>& udist_y,
					    std::uniform_real_distribution<double>& udist_z,
					    const double& L, std::ofstream& omsh, int count,
					    std::vector<std::string>& top_ids,
					    std::map<std::string, std::vector<int> >& ind_map, int bIts,
					    std::vector<std::vector<float> >& allPoints);
static
bool inHole (int i, int j, std::vector<std::pair<int, int> >& holes, const int& N);
//=================================================================================================
void generate_vascular_network ()
{
  //////////////////////////////////////
  // USER INPUT BEGIN
  /////
  // cube dimensions / total length!
  const double L = 11.9e-3;
  // tumour radius
  const double R = 0.5e-3;
  // number of seeds per side
  const unsigned int N = 30;
  // bifurcations?
  const bool bifurcate = true;
  // vessel arrangement: 0 = regular array; 1 = uniform random; 2 = circular hole; 3 = percolation (cf. Baish et al 2010 doi:10.1073/pnas.1018154108)
  // NOTE: arrangement 3 will not work for anything smaller than a 4x4 array (ie. N >= 4)
  const int vessel_arrangement = 0;
  // set to true if you want to exclude the tumour
  const bool exclude = true;
  // set to true if you want a uniform mix of inlets and outlets on each surface
  // else it will set all nodes on one surface as inlets and the other as outlets
  const bool mix = true;
  /////
  // USER INPUT END
  /////////////////////////////////////

  // generate and write the vascular network (of capillaries)
  _gen_cap_net(L, R, N, bifurcate, vessel_arrangement, exclude, mix);
}
//=================================================================================================
//
//
//
//=================================================================================================
void _gen_cap_net ( const double& L, const double& R, const unsigned int& N,
                    const bool& bifurcate, const int& vessel_arrangement,
                    const bool& exclude, const bool& mix )
{

  // write data to Gmsh .geo file format
  std::stringstream filename;
  filename << "cap_net.geo" << std::flush;
  std::ofstream omsh(filename.str().c_str());
  //
  omsh << "Geometry.Tolerance = 1.0e-8;" << std::endl;
  omsh << "Mesh.ElementOrder = 1;" << std::endl;
  //
  // output a list of parameters: radius of tumour, length of domain, nodal multiplier
  omsh << "RAD = " << R << ";" << std::endl;
  omsh << "LEN = " << L << ";" << std::endl;
  omsh << "NS = 2;" << std::endl;

  // first determine seed point grid
  const double lx = L/N, ly = L/N;

  // spline index
  unsigned int count = 0;

  // keep track of the bottom and top Point IDs
  std::vector<std::string> bot_ids;
  std::vector<std::string> top_ids;

  // keep track of the spline number and number of points / spline
  std::map<std::string, std::vector<int> > ind_map;

  // all points - for collision detection
  std::vector<std::vector<float> > allPoints;

  // initialise a random number generator
  std::random_device rd;
  std::default_random_engine rgen(rd());

  // initialise percolation holes
  std::vector<std::pair<int, int> > holes;
  std::uniform_int_distribution<int> udist_N(0, N);
  std::uniform_int_distribution<int> udist_R(0, N/4);
  for(int i=0; i<4; i++){
    // currently just choose 4 points points arbitrarily, 1 / edge
    int centre = udist_N(rgen);
    int radius = 0;
    // set radius range
    while(radius < N/10)
      radius = udist_R(rgen);
    holes.push_back(std::make_pair(centre, radius));
  }

  // loop over grid elements
  for(unsigned int i=0; i<N; i++){

    for(unsigned int j=0; j<N; j++){

      // vessel arrangement checks
      if(vessel_arrangement == 1 && rand()%2)
        continue;
      if(vessel_arrangement == 2 && 2*std::sqrt(i*i+j*j) < N)
        continue;
      if(vessel_arrangement == 3 && inHole(i, j, holes, N))
        continue;

      double xmin = -L/2.0 + i*lx;
      double xmax = -L/2.0 + (i+1)*lx;
      double ymin = -L/2.0 + j*ly;
      double ymax = -L/2.0 + (j+1)*ly;

      if(exclude){
        // don't generate a capillary if this element is entirely enclosed by a square around the tumour
        if(xmin > -R && xmax < R && ymin > -R && ymax < R)
          continue;
      }

      count++;

      // initialise uniform distribution for the x-coordinate
      std::uniform_real_distribution<double> udist_x(xmin, xmax);
      // initialise uniform distribution for the y-coordinate
      std::uniform_real_distribution<double> udist_y(ymin, ymax);
      // initialise uniform distribution for the z-coordinate
      std::uniform_real_distribution<double> udist_z(0.0, L);

      // seed point coordinates
      double x0 = udist_x(rgen);
      double y0 = udist_y(rgen);
      double z0 = -L/2.0;

      std::vector<float> tempPoint;
      tempPoint.push_back(x0);
      tempPoint.push_back(y0);
      tempPoint.push_back(z0);
      allPoints.push_back(tempPoint);

      omsh << "Point(" << count << "00) = {" << x0 << ", " << y0 << ", " << z0 << "};" << std::endl;

      // save the bottom Point ID
      std::stringstream bot_str;
      bot_str << count << "00";
      bot_ids.push_back(bot_str.str().c_str());

      // first propagate a single vessel along this grid point, and return bifurcation points
      std::map<int, std::vector<double> > bifurcation_map;
      bifurcation_map = propSeg(x0, y0, z0,
				rgen,
				udist_x,
				udist_y,
				udist_z,
				L, omsh, count,
				top_ids,
				ind_map, 0,
				allPoints);
      if(bifurcate){
        // now iterate over bifurcation points along this vessel and propagate
        int bIts = 1;
        for(std::map<int, std::vector<double> >::iterator mIt=bifurcation_map.begin(); mIt!= bifurcation_map.end(); mIt++){
          omsh << "Point(" << count << bIts << "0) = {" << mIt->second[0] << ", " << mIt->second[1] << ", " << mIt->second[2] << "};" << std::endl;
          std::map<int, std::vector<double> > dummy_map;
          dummy_map = propSeg(mIt->second[0], mIt->second[1], mIt->second[2],
        		      rgen,
        		      udist_x,
        		      udist_y,
        		      udist_z,
        		      L, omsh, count,
        		      top_ids,
        		      ind_map, bIts,
        		      allPoints);
          bIts++;
          // could have a conditional loop here that propagates the next level of bifurcations...
          /*
            for(std::map<int, std::vector<double> >::iterator mIt1=dummy_map.begin(); mIt1!= dummy_map.end(); mIt1++){
            omsh << "Point(" << count << bIts << "0) = {" << mIt1->second[0] << ", " << mIt1->second[1] << ", " << mIt1->second[2] << "};" << std::endl;
            std::map<int, std::vector<double> > dummy_map1;
            dummy_map1 = propSeg(mIt1->second[0], mIt1->second[1], mIt1->second[2],
            rgen,
            udist_x,
            udist_y,
            udist_z,
            L, omsh, count,
            top_ids,
            ind_map, bIts,
            allPoints);
            bIts++;
            // could have a loop here that propagates the next level of bifurcations...
            }
          */
        }
      } // end of bifurcate check
    }
  }
  omsh << "Transfinite Line \"*\" = 2;" << std::endl;

  // write a transfinite line with each spline
  for(std::map<std::string, std::vector<int> >::iterator it = ind_map.begin(); it!= ind_map.end(); it++){
    omsh << "Transfinite Line {" << it->first << "0} = " << it->second[1] << "*NS;" << std::endl;
  }

  // set physical numbers for the inlet and outlet vessels
  // for now split equally... could also do it randomly
  unsigned int N_inl = (int)ind_map.size()/2;
  //  std::map<int, int>::iterator it = ind_map.begin();
  std::map<std::string, std::vector<int> >::iterator it = ind_map.begin();
  unsigned int ind = 0;
  // inlets...
  omsh << "Physical Line(1001) = {";
  for(it; it!= ind_map.end(); it++){
    ind++;
    //
    omsh << it->first << "0";
    if(ind > N_inl){
      it++;
      break;
    }
    omsh << ",";
  }
  omsh << "};" << std::endl;
  // outlets...
  omsh << "Physical Line(1002) = {";
  for(it; it!= ind_map.end(); it++){
    omsh << it->first << "0";
    // some insane acrobatics to determine if we're at the last entry...
    std::map<std::string, std::vector<int> >::iterator dum_it = it;
    dum_it++;
    if(dum_it == ind_map.end())
      break;
    omsh << ",";
  }
  omsh << "};" << std::endl;

  // check that we've counted the number of inlets and outlets correctly
  if( bot_ids.size() != top_ids.size() )
    std::cout << "Odd number of inlets and outlets: " << (bot_ids.size() + top_ids.size()) << ". If there are no bifurcations, this shouldn't happen..." << std::endl;

  // set physical numbers for the inlet and outlet points

  if(mix){
    // use a random generator to get a mix of inlets and outlets on each surface
    std::random_device rd;
    std::default_random_engine rgen(rd());
    std::uniform_real_distribution<double> udist(0.0, 1.0);
    // sort the inlets and outlets into vectors
    std::vector<std::string> inl_ids;
    std::vector<std::string> out_ids;
    for(unsigned int i=0; i<bot_ids.size(); i++){
      if(udist(rgen) > 0.5){
        inl_ids.push_back(bot_ids[i]);
        out_ids.push_back(top_ids[i]);
      }
      else{
        inl_ids.push_back(top_ids[i]);
        out_ids.push_back(bot_ids[i]);
      }
    }
    // now print the inlets...
    omsh << "Physical Point(101) = {";
    for(unsigned int i=0; i<inl_ids.size(); i++){

      omsh << inl_ids[i];

      if(i+1 == inl_ids.size())
        break;

      omsh << ",";
    }
    omsh << "};" << std::endl;

    // outlets...
    omsh << "Physical Point(102) = {";
    for(unsigned int i=0; i<out_ids.size(); i++){

    omsh << out_ids[i];

    if(i+1 == out_ids.size())
      break;

    omsh << ",";
    }
    omsh << "};" << std::endl;

  }
  else{
    // set all points on one surface as inlets and the other as outlets
    // inlets...
    omsh << "Physical Point(101) = {";
    for(unsigned int i=0; i<bot_ids.size(); i++){

      omsh << bot_ids[i];

      if(i+1 == bot_ids.size())
        break;

      omsh << ",";
    }
    omsh << "};" << std::endl;

    // outlets...
    omsh << "Physical Point(102) = {";
    for(unsigned int i=0; i<top_ids.size(); i++){

    omsh << top_ids[i];

    if(i+1 == top_ids.size())
      break;

    omsh << ",";
    }
    omsh << "};" << std::endl;
  }

  // close readout
  omsh.close();
    }
//=================================================================================================
std::map<int, std::vector<double> > propSeg ( double& x0, double& y0, double& z0,
					    std::default_random_engine& rgen,
					    std::uniform_real_distribution<double>& udist_x,
					    std::uniform_real_distribution<double>& udist_y,
					    std::uniform_real_distribution<double>& udist_z,
					    const double& L, std::ofstream& omsh, int count,
					    std::vector<std::string>& top_ids,
					    std::map<std::string, std::vector<int> >& ind_map, int bIt,
					    std::vector<std::vector<float> >& allPoints )
{
  // now generate another point along the z-axis
  const unsigned int N_z = 10;
  // keep moving up the z-axis until we hit the boundary
  bool propagate = true;
  const double tol = 1E-6;
  unsigned int k = 0;

  double xk = x0, yk = y0, zk = z0;

  std::map<int, std::vector<double> > bifurcation_points;

  while(propagate){

    k++;

    // move within a defined radius of the last point
    const double rad = 1.0;
    // generate a new displacement
    double ux = 0.0, uy = 0.0, uz = 0.0;

    bool reroll = true;

    while(reroll){
      ux = udist_x(rgen)*rad;
      uy = udist_y(rgen)*rad;
      uz = udist_z(rgen)*0.5;

      double xtemp = xk + ux;
      double ytemp = yk + uy;
      double ztemp = zk + uz;

      // if both x and y coordinates are within the boundary, we can exit the loop
      if( (std::fabs(xtemp - L/2.0) > tol && std::fabs(ytemp - L/2.0) > tol) )
        reroll = false;

      // however, if we're too close to the last point in z then reroll; this prevents kinked and out-of-bounds splines
      if( uz / L < 0.1)
        reroll = true;

      // also if too close to any existing point
      for(int pIt=0; pIt<allPoints.size(); pIt++){
        if( std::fabs(allPoints[pIt][0]-xtemp) < 1E-6 &&
            std::fabs(allPoints[pIt][1]-ytemp) < 1E-6 &&
            std::fabs(allPoints[pIt][2]-ztemp) < 1E-6 )
          reroll = true;
      }

      //      if(exclude){
	// if either x-y coordinate is now within a square of half length R about the origin, i.e. too close to the tumour, reroll
      //	if( std::fabs(xtemp) < R && std::fabs(ytemp) < R)
      //	  reroll = true;
      //      }

    }

    xk = ux;
    yk = uy;
    zk += uz;

    // stop propagating along the z-axis when the vessel approaches the boundary
    if(zk > 0.8*L/2.0){
      zk = L/2.0;
      propagate = false;
    }

    // write new coordinates
    //    omsh << "Point(" << count << "0" << k << ") = {" << xk << ", " << yk << ", " << zk << "};" << std::endl;
    omsh << "Point(" << count << bIt << k << ") = {" << xk << ", " << yk << ", " << zk << "};" << std::endl;

    std::vector<float> tempPoint;
    tempPoint.push_back(xk);
    tempPoint.push_back(yk);
    tempPoint.push_back(zk);
    allPoints.push_back(tempPoint);

    // check if point bifurcates
    if(rand()%2 && propagate){
      std::vector<double> new_start;
      new_start.push_back(xk);
      new_start.push_back(yk);
      new_start.push_back(zk);
      bifurcation_points.insert(std::pair<int, std::vector<double> >(k, new_start));
    }

    // save last point's ID
    if(!propagate){
      std::stringstream top_str;
      top_str << count << "0" << k;
      top_ids.push_back(top_str.str().c_str());
      break;
    }

  }
  // write a spline with these points
  //  omsh << "Spline(" << count << "00) = {" << count << 0 << 0 << ":" << count << 0 << k << "};" << std::endl;
  omsh << "Spline(" << count << bIt << "0) = {" << count << bIt << 0 << ":" << count << bIt << k << "};" << std::endl;

  //  ind_map.insert(std::make_pair(count, k+1));
  std::vector<int> indexes;
  indexes.push_back(bIt);
  indexes.push_back(k+1);
  std::string strout = std::to_string(count)+std::to_string(bIt);
  //  ind_map.insert(std::make_pair(count, indexes));
  ind_map.insert(std::make_pair(strout, indexes));

  return bifurcation_points;
}
//=================================================================================================
bool inHole (int i, int j, std::vector<std::pair<int, int> >& holes, const int& N)
{
  bool cut=false;
  int count = 0;
  for(std::vector<std::pair<int, int> >::iterator it=holes.begin(); it!=holes.end(); it++){
    int centre = it->first;
    int radius = it->second;
    // bottom edge
    if(count==0){
      // move point to current coordinate system
      int i_tick = i-centre;
      int j_tick = j;
      if(std::sqrt(i_tick*i_tick+j_tick*j_tick) < radius)
        cut=true;
    }
    // left edge
    else if(count==1){
      // move point to current coordinate system
      int i_tick = i;
      int j_tick = j-centre;
      if(std::sqrt(i_tick*i_tick+j_tick*j_tick) < radius)
        cut=true;
    }
    // top edge
    else if(count==2){
      // move point to current coordinate system
      int i_tick = i-centre;
      int j_tick = N-1-j;
      if(std::sqrt(i_tick*i_tick+j_tick*j_tick) < radius)
        cut=true;
    }
    // right edge
    else if(count==3){
      // move point to current coordinate system
      int i_tick = N-1-i;
      int j_tick = j-centre;
      if(std::sqrt(i_tick*i_tick+j_tick*j_tick) < radius)
        cut=true;
    }
    count++;
  }
  return cut;
}
//=================================================================================================
