/*************************************************************************************************

    Copyright (C) 2009-2019    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/utils.h"
#include "feb3/domain_data.h"
#include "libmesh/mesh.h"
#include "libmesh/equation_systems.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
//=================================================================================================
void prepare_vascular_network () {
    Mesh msh(*gcomm_ptr, SP3D);
    EquationSystems es(msh);
    DomainData dmn(es);
    //
    Mesh msh_vn(*gcomm_ptr, SP1D);
    EquationSystems es_vn(msh_vn);
    DomainData_VascularNetwork dmn_vn(es_vn, es);
    //
    Parameters& par = es.parameters;
    //
    par.set<std::string>("results directory")                  = "prepare_vascular_network-res";
    par.set<std::string>("results directory: paraview")        = par.get<std::string>("results directory") + "/paraview";
    par.set<std::string>("results directory: quadrature data") = par.get<std::string>("results directory") + "/quadrature_data";
    par.set<std::string>("results directory: nodal data")      = par.get<std::string>("results directory") + "/nodal_data";
    par.set<std::string>("results directory: temporary")       = par.get<std::string>("results directory") + "/temporary";
    par.set<std::string>("results directory: restart")         = par.get<std::string>("results directory") + "/restart";
    par.set<std::string>("output file name") = "out";
    par.set<bool>("delete unwanted elements/nodes") = false;
    par.set<bool>("initialise RVEs") = false;
    par.set<int>("nodal force evaluation: number of external surfaces") = 0;
    par.set<int>("nodal force evaluation: number of internal surfaces") = 0;
    par.set<int>("nodal force evaluation: number of regions") = 0;
    par.set<int>("nodal force evaluation: number of regions") = 0;
    par.set<std::string>("input mesh file")                         = "tumouroid_angiogenesis-tissue.msh";
    par.set<std::string>("input vascular network file")             = "tumouroid_angiogenesis-vasculature.msh";
    par.set<std::string>("input vascular network information file") = "tumouroid_angiogenesis-vasculature.info";
    par.set<std::string>("input vascular network domain data file") = "NA";
    par.set<unsigned int>("input Quadrature from file") = 0;
    par.set<bool>("FEM framework") = true;
    par.set<bool>("IGA framework") = false;
    par.set<bool>("membrane elements present") = false;
    par.set<Order>("Lagrange basis order") = FIRST;
    par.set<Order>("Gauss quadrature order") = FOURTH;
    par.set<bool>("enforce input mesh order") = true;
    par.set<bool>("skip input mesh renumbering") = false;
    par.set<bool>("input vascular network domain data: skip renumbering") = true;
    // par.set<bool>("input vascular network domain data: skip renumbering") = false;
    par.set<bool>("input vascular network domain data: skip branching") = true;
    par.set<bool>("input vascular network domain data: skip branching") = false;
    //
    par.set<unsigned int>("reaction-diffusion solver: dimension") = 0;
    par.set<unsigned int>("reaction-diffusion solver: discontinuous variables") = 0;
    par.set<unsigned int>("Navier-Cauchy solver: internal variables") = 0;
    par.set<unsigned int>("reaction-diffusion solver: auxiliary variables") = 0;
    //
    par.set<std::string>("input mesh file")                         = "ASCII_STL_LS3.msh";
    par.set<std::string>("input vascular network file")             = "LS3_ascii.msh";
    //
    {
        const std::string Rpath[] = { "mkdir " + par.get<std::string>("results directory"),
                                      "mkdir " + par.get<std::string>("results directory: paraview"),
                                      "mkdir " + par.get<std::string>("results directory: paraview") + "/MS",
                                      "mkdir " + par.get<std::string>("results directory: quadrature data"),
                                      "mkdir " + par.get<std::string>("results directory: nodal data"),
                                      "mkdir " + par.get<std::string>("results directory: temporary"),
                                      "mkdir " + par.get<std::string>("results directory: restart") };
        for (int i=0; i<7; i++)
            std::system(Rpath[i].c_str());
    }
    //
    es.add_system<System>("a system");
    es.get_system<System>("a system").add_variable("a variable", par.get<Order>("Lagrange basis order"), LAGRANGE);
    //
    es_vn.add_system<System>("a system");
    es_vn.get_system<System>("a system").add_variable("a variable", par.get<Order>("Lagrange basis order"), LAGRANGE);
    //
    dmn.regn.insert( std::make_pair(3001, RegionInfo()) );
    dmn.regn.insert( std::make_pair(3002, RegionInfo()) );
    dmn.bcs.insert( std::make_pair(2001, BoundaryConditionInfo()) );
    dmn.bcs.insert( std::make_pair(2002, BoundaryConditionInfo()) );
    dmn.bcs.insert( std::make_pair(2003, BoundaryConditionInfo()) );
    dmn.bcs.insert( std::make_pair(2004, BoundaryConditionInfo()) );
    dmn.bcs.insert( std::make_pair(2005, BoundaryConditionInfo()) );
    dmn.bcs.insert( std::make_pair(2006, BoundaryConditionInfo()) );
    //
    dmn.prepare_mesh();
    dmn_vn.prepare_mesh();
    //
    es.init();
    es.print_info();
    es.print_info(flog.stream());
    es_vn.init();
    es_vn.print_info();
    es_vn.print_info(flog.stream());
    //
    dmn.init();
    dmn_vn.init();
    //
    std::cout << " >> Meshes prepared successfully.\n" << std::endl;
}
//=================================================================================================
