/*************************************************************************************************

    Copyright (C) 2009-2019    Peter Wijeratne (pwijeratne@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include <fstream>
#include <vector>
#include <random>
#include <string>
#include "libmesh/libmesh.h"
#include "libmesh/point.h"
#include "libmesh/mesh.h"
#include "feb3/utils.h"
//=================================================================================================
using namespace libMesh;
//=================================================================================================
namespace test {
//=================================================================================================
class RVE {
public:
    RVE () : L(0.20), D(0.1) {}
    ~RVE () {}
    // public member functions
    void generate (const unsigned int nfib, const Real cmin =0., const Real cmax =1.);
    void ouput (const std::string& fname ="rve.msh") const;
private:
    // private member objects
    std::map<unsigned int, std::vector<Real> > _node_map;
    std::vector<unsigned int> _node1, _node2;
    std::map<unsigned int, unsigned int> _node_bc;
public:
    // public member objects
    Real L, D; // RVE parameters
};
//=================================================================================================
void RVE::generate (const unsigned int nfib, const Real cmin, const Real cmax)
{
    // initialise RVE parameters
    unsigned int NF = nfib;
    const Real D2 = pow2(this->D);
    // initialise some extra local parameters
    const unsigned int NS = 2*NF+1;
    const unsigned int NT = 2*NF;
    // initialise RVE variables
    std::vector<Real> x(NS),    y(NS),    z(NS);
    std::vector<Real> x0(NS),   y0(NS),   z0(NS);
    std::vector<Real> dx(3*NS), dy(3*NS), dz(3*NS);
    std::vector<unsigned int> a(3*NT+1);
    // initialise a random number generator
    std::random_device rd;
    std::default_random_engine rgen(rd());
    // initialise uniform and normal distributions
    std::uniform_real_distribution<Real> udist(0.0, 1.0);
    std::normal_distribution<Real> ndist(0.65,0.1);
    // create initial seeds
    for (unsigned int i=1; i<=NF; i++) {
        x0[i] = 2.0*udist(rgen);
        y0[i] = 2.0*udist(rgen);
        z0[i] = 2.0*udist(rgen);
        x[2*i-1] = x0[i];
        y[2*i-1] = y0[i];
        z[2*i-1] = z0[i];
        x[2*i] = x0[i];
        y[2*i] = y0[i];
        z[2*i] = z0[i];
        const Real dx1 = udist(rgen),
                   dy1 = udist(rgen),
                   dz1 = udist(rgen),
                   /////////////////////// dz1 = ndist(rgen),
                   d1 = sqrt(pow2(dx1)+pow2(dy1)+pow2(dz1));
        dx[2*i-1] =  dx1/d1;
        dx[2*i+0] = -dx1/d1;
        dy[2*i-1] =  dy1/d1;
        dy[2*i+0] = -dy1/d1;
        dz[2*i-1] =  dz1/d1;
        dz[2*i+0] = -dz1/d1;
        a[2*i-1] = 2*i-1;
        a[2*i+0] = 2*i;
    }
    //
    unsigned int NL = 2*NF;
    // create vector of "living" segments
    std::vector<unsigned int> live;
    for (unsigned int i=0; i<=NL; i++) live.push_back(i);
    //
    // start adding monomer
    while (NL > 0) {
        unsigned int g = 1;
        // select fibre to update; +1 because matlab indices begin at 1 and floor(...) can equal 0
        unsigned int j = floor(udist(rgen)*NL) + 1;
        unsigned int k = live[j];
        // update fibre characteristics
        x[k] += this->L*dx[k];
        y[k] += this->L*dy[k];
        z[k] += this->L*dz[k];
        Real x1 = x[k];
        Real y1 = y[k];
        Real z1 = z[k];
        // check for fibre out of big box
        Real u[6];
        u[0] = (x1 - 2.0)/fabs(dx[k]);
        u[1] = (y1 - 2.0)/fabs(dy[k]);
        u[2] = (z1 - 2.0)/fabs(dz[k]);
        u[3] = -x1/fabs(dx[k]);
        u[4] = -y1/fabs(dy[k]);
        u[5] = -z1/fabs(dz[k]);
        Real Lu = u[0];
        // find max value of u
        for (unsigned int j=1; j<6; j++) {
            if (u[j] > Lu) Lu = u[j];
        }
        // fibre out of box
        if (Lu > 0.0) {
            x1 -= Lu*dx[k];
            y1 -= Lu*dy[k];
            z1 -= Lu*dz[k];
            // kill segment
            for (unsigned int m=j; m<=(NL-1); m++) {
              live[m] = live[m+1];
            }
            NL--;
            g = 0;
        }
        // check for fibre-fibre collision
        if ( 1 == g ) {
            unsigned int i = 0;
            if ( (k==1) || (k==2) ) i = 2;
            else i = 1;
            //
            while (i < NF) {
              Real xi0 = x[a[2*i-1]];
              Real yi0 = y[a[2*i-1]];
              Real zi0 = z[a[2*i-1]];
              Real xi1 = x[a[2*i]];
              Real yi1 = y[a[2*i]];
              Real zi1 = z[a[2*i]];
              Real h  = sqrt( pow2(xi1-xi0) + pow2(yi1-yi0) + pow2(zi1-zi0) );
              Real h1 = sqrt( pow2(xi1-x1 ) + pow2(yi1-y1 ) + pow2(zi1-z1 ) );
              Real h2 = sqrt( pow2(xi0-x1 ) + pow2(yi0-y1 ) + pow2(zi0-z1 ) );
              //
              if ( (h1 < h) && (h2 < h) ) {
                  const Real aa = pow2(xi1-x1 )+pow2(yi1-y1 )+pow2(zi1-z1 ),
                             bb = pow2(xi0-xi1)+pow2(yi0-yi1)+pow2(zi0-zi1),
                             cc = pow2((xi1-x1)*(xi0-xi1)+(yi1-y1)*(yi0-yi1)+(zi1-z1)*(zi0-zi1));
                  const Real dist2 = (aa*bb-cc)/bb;
                  // collision!
                  if ( dist2 < D2 ) {
                      const Real tt = -((xi1-x1)*(xi0-xi1)+(yi1-y1)*(yi0-yi1)+(zi1-z1)*(zi0-zi1))/bb;
                      x1 = xi1 + (xi0-xi1)*tt;
                      y1 = yi1 + (yi0-yi1)*tt;
                      z1 = zi1 + (zi0-zi1)*tt;
                      // split struck fibre in two
                      NF++;
                      //
                      a[2*NF-1] = k;
                      a[2*NF] = a[2*i];
                      dx[2*NF] = dx[2*i];
                      dy[2*NF] = dy[2*i];
                      dz[2*NF] = dz[2*i];
                      dx[2*NF-1] = dx[2*i-1];
                      dy[2*NF-1] = dy[2*i-1];
                      dz[2*NF-1] = dz[2*i-1];
                      a[2*i] = k;
                      // kill segment in collision
                      for (unsigned int m=j; m<=(NL-1); m++) {
                          live[m] = live[m+1];
                      }
                      NL--;
                      // skip remainder of search
                      i = NF;
                  }
              }
              i++;
              // skip collision with self!
              if ( (a[2*i-1] == k) || (a[2*i] == k) ) i++;
              // end of while loop
            }
            // end of conditional
        }
        //
        g=1;
        x[k] = x1;
        y[k] = y1;
        z[k] = z1;
        // end of while loop
    }
    // coordinate containers
    std::vector<Real> temp(NF);
    std::vector< std::vector<Real> > Ax; Ax.assign(2, temp);
    std::vector< std::vector<Real> > Ay; Ay.assign(2, temp);
    std::vector< std::vector<Real> > Az; Az.assign(2, temp);
    std::vector<Real> b(2*NF);
    unsigned int k = 0;
    // select fibres with nodes in the box [x,y,z](-0.5,0.5)
    for (unsigned int i=1; i<=NF; i++) {
        //
        if ( (x[a[2*i]] > 0.51) &&
             (x[a[2*i]] < 1.49) &&
             (y[a[2*i]] > 0.51) &&
             (y[a[2*i]] < 1.49) &&
             (z[a[2*i]] > 0.51) &&
             (z[a[2*i]] < 1.49) )
        {
            Real u[6];
            u[0] = (x[a[2*i-1]] - 1.49)/fabs(dx[2*i-1]);
            u[1] = (y[a[2*i-1]] - 1.49)/fabs(dy[2*i-1]);
            u[2] = (z[a[2*i-1]] - 1.49)/fabs(dz[2*i-1]);
            u[3] = (0.51 - x[a[2*i-1]])/fabs(dx[2*i-1]);
            u[4] = (0.51 - y[a[2*i-1]])/fabs(dy[2*i-1]);
            u[5] = (0.51 - z[a[2*i-1]])/fabs(dz[2*i-1]);
            Real Lu = u[0];
            // find max value of u
            for (unsigned int j=1; j<6; j++) {
                if(u[j] > Lu) Lu = u[j];
            }
            // node a[2*i-1] is out of box?
            if ( Lu > 0.0 ) {
                u[0] = (x[a[2*i-1]] - 1.5)/fabs(dx[2*i-1]);
                u[1] = (y[a[2*i-1]] - 1.5)/fabs(dy[2*i-1]);
                u[2] = (z[a[2*i-1]] - 1.5)/fabs(dz[2*i-1]);
                u[3] = (0.5 - x[a[2*i-1]])/fabs(dx[2*i-1]);
                u[4] = (0.5 - y[a[2*i-1]])/fabs(dy[2*i-1]);
                u[5] = (0.5 - z[a[2*i-1]])/fabs(dz[2*i-1]);
                Real Lu1 = u[0];
                // find max value of u
                for (unsigned int j=1; j<6; j++) {
                  if (u[j] > Lu1) Lu1 = u[j];
                }
                x[a[2*i-1]] -= Lu1*dx[2*i-1];
                y[a[2*i-1]] -= Lu1*dy[2*i-1];
                z[a[2*i-1]] -= Lu1*dz[2*i-1];
            }
            k++;
            //
            Ax[0][k] = x[a[2*i-1]] - 1.0;
            Ax[1][k] = x[a[2*i]]   - 1.0;
            Ay[0][k] = y[a[2*i-1]] - 1.0;
            Ay[1][k] = y[a[2*i]]   - 1.0;
            Az[0][k] = z[a[2*i-1]] - 1.0;
            Az[1][k] = z[a[2*i]]   - 1.0;
            b[2*k-1] = a[2*i-1];
            b[2*k]   = a[2*i];
            // end of conditional
        } else if ( (x[a[2*i-1]] > 0.51) &&
                    (x[a[2*i-1]] < 1.49) &&
                    (y[a[2*i-1]] > 0.51) &&
                    (y[a[2*i-1]] < 1.49) &&
                    (z[a[2*i-1]] > 0.51) &&
                    (z[a[2*i-1]] < 1.49) )
        {
            Real u[6];
            u[0] = (x[a[2*i]] - 1.49)/fabs(dx[2*i]);
            u[1] = (y[a[2*i]] - 1.49)/fabs(dy[2*i]);
            u[2] = (z[a[2*i]] - 1.49)/fabs(dz[2*i]);
            u[3] = (0.51 - x[a[2*i]])/fabs(dx[2*i]);
            u[4] = (0.51 - y[a[2*i]])/fabs(dy[2*i]);
            u[5] = (0.51 - z[a[2*i]])/fabs(dz[2*i]);
            Real Lu = u[0];
            // find max value of u
            for (unsigned int j=1; j<6; j++) {
                if (u[j] > Lu) Lu = u[j];
            }
            // node a[2*i] is out of box?
            if ( Lu > 0.0 ) {
                u[0] = (x[a[2*i]] - 1.5)/fabs(dx[2*i]);
                u[1] = (y[a[2*i]] - 1.5)/fabs(dy[2*i]);
                u[2] = (z[a[2*i]] - 1.5)/fabs(dz[2*i]);
                u[3] = (0.5 - x[a[2*i]])/fabs(dx[2*i]);
                u[4] = (0.5 - y[a[2*i]])/fabs(dy[2*i]);
                u[5] = (0.5 - z[a[2*i]])/fabs(dz[2*i]);
                Real Lu1 = u[0];
                // find max value of u
                for (unsigned int j=1; j<6; j++) {
                  if (u[j] > Lu1) Lu1 = u[j];
                }
                x[a[2*i]] -= Lu1*dx[2*i];
                y[a[2*i]] -= Lu1*dy[2*i];
                z[a[2*i]] -= Lu1*dz[2*i];
            }
            //
            k++;
            Ax[0][k] = x[a[2*i-1]] - 1.0;
            Ax[1][k] = x[a[2*i]]   - 1.0;
            Ay[0][k] = y[a[2*i-1]] - 1.0;
            Ay[1][k] = y[a[2*i]]   - 1.0;
            Az[0][k] = z[a[2*i-1]] - 1.0;
            Az[1][k] = z[a[2*i]]   - 1.0;
            b[2*k-1] = a[2*i-1];
            b[2*k]   = a[2*i];
        }
        // end of for loop
    }
    // would close Lu file here...
    NF = k;
    std::vector<unsigned int> c;
    for (unsigned int i=0; i<=(2*NF); i++) c.push_back(0);
    unsigned int index = 0;
    unsigned int count = 1;
    c[1] = 1;
    for (unsigned int i=2; i<=(2*NF); i++) {
        for (unsigned int j=1; j<=(i-1); j++) {
            if (b[i]==b[j]) {
                c[i]=c[j];
                index=1;
            }
        }
        if (0==index) {
            count+=1;
            c[i]=count;
        }
        index=0;
        // end of for loop...
    }
    // clean memo first
    this->_node1.clear();
    this->_node2.clear();
    this->_node_map.clear();
    this->_node_bc.clear();
    // transformation (if necessary) coefficients
    const Real alpha = (cmax-cmin),
               beta  = (cmax+cmin)*0.5,
               epsilon = (cmax-cmin)*1.0e-6;
    //
    // std::vector<Real> length;
    // Real len = 0.0;
    // Real proj_len = 0.0;
    std::vector<Real> node1_xcoord, node1_ycoord, node1_zcoord;
    std::vector<Real> node2_xcoord, node2_ycoord, node2_zcoord;
    // store node ID and position
    for (unsigned int i=1; i<=NF; i++) {
        const Real BOX_MIN_COORD = -0.5,
                   BOX_MAX_COORD =  0.5;
        // check if fibre is still outside mesh...
        // if true, then pull the offending node back to the boundary
        for (unsigned ii=0; ii<2; ii++) {
            // FIXME: should do a linear interpolation
            if ( Ax[ii][i] > BOX_MAX_COORD ) Ax[ii][i] = BOX_MAX_COORD;
            if ( Ax[ii][i] < BOX_MIN_COORD ) Ax[ii][i] = BOX_MIN_COORD;
            if ( Ay[ii][i] > BOX_MAX_COORD ) Ay[ii][i] = BOX_MAX_COORD;
            if ( Ay[ii][i] < BOX_MIN_COORD ) Ay[ii][i] = BOX_MIN_COORD;
            if ( Az[ii][i] > BOX_MAX_COORD ) Az[ii][i] = BOX_MAX_COORD;
            if ( Az[ii][i] < BOX_MIN_COORD ) Az[ii][i] = BOX_MIN_COORD;
        }
        //
        this->_node1.push_back(c[2*i-1]);
        this->_node2.push_back(c[2*i+0]);
        node1_xcoord.push_back(Ax[0][i]);
        node1_ycoord.push_back(Ay[0][i]);
        node1_zcoord.push_back(Az[0][i]);
        node2_xcoord.push_back(Ax[1][i]);
        node2_ycoord.push_back(Ay[1][i]);
        node2_zcoord.push_back(Az[1][i]);
        // length.push_back(sqrt( pow2(Ax[0][i]-Ax[1][i])
        //                      + pow2(Ay[0][i]-Ay[1][i])
        //                      + pow2(Az[0][i]-Az[1][i])));
        // len += length[i-1];
        // proj_len += sqrt( pow2(Ax[0][i]-Ax[1][i]) + pow2(Ay[0][i]-Ay[1][i]));
        // end of for loop...
    }
    //
    const unsigned int NODES = count;
    const unsigned int ELEMENTS = NF;
    // let's sort the nodes out - use a map to order by ID and skip duplicates
    for (unsigned int i=0; i<NF; i++) {
        std::vector<Real> temp_vec1;
        temp_vec1.push_back(alpha*node1_xcoord[i]+beta);
        temp_vec1.push_back(alpha*node1_ycoord[i]+beta);
        temp_vec1.push_back(alpha*node1_zcoord[i]+beta);
        std::vector<Real> temp_vec2;
        temp_vec2.push_back(alpha*node2_xcoord[i]+beta);
        temp_vec2.push_back(alpha*node2_ycoord[i]+beta);
        temp_vec2.push_back(alpha*node2_zcoord[i]+beta);
        this->_node_map.insert(std::make_pair(this->_node1[i], temp_vec1));
        this->_node_map.insert(std::make_pair(this->_node2[i], temp_vec2));
        // end of for loop...
    }
    /*

        o--------o
       /:       /|
      / : 5    / |
     /  :   3 /  |
    o--------o 2 |
    | 4 o....|...o
    |  . 1   |  /
    | .   0  | /
    |.       |/
    o--------o

    */
    // find the boundary nodes
    for ( std::map<unsigned int, std::vector<Real> >::const_iterator
          it=this->_node_map.begin(); it!=this->_node_map.end(); it++) {
        const std::vector<Real>& coord = it->second;
        if      ( coord[0] < cmin-epsilon ) this->_node_bc.insert(std::make_pair(it->first, 3));
        else if ( coord[0] > cmax-epsilon ) this->_node_bc.insert(std::make_pair(it->first, 1));
        else if ( coord[1] < cmin-epsilon ) this->_node_bc.insert(std::make_pair(it->first, 4));
        else if ( coord[1] > cmax-epsilon ) this->_node_bc.insert(std::make_pair(it->first, 2));
        else if ( coord[2] < cmin-epsilon ) this->_node_bc.insert(std::make_pair(it->first, 0));
        else if ( coord[2] > cmax-epsilon ) this->_node_bc.insert(std::make_pair(it->first, 5));
        else ;
    }
    // sanity checks
    if (NODES!=this->_node_map.size()) libmesh_error();
    if (ELEMENTS!=this->_node1.size()) libmesh_error();
    if (ELEMENTS!=this->_node2.size()) libmesh_error();
    if (this->_node_bc.empty()) libmesh_error();
}
//=================================================================================================
void RVE::ouput (const std::string& fname) const
{
    std::ofstream fout(fname);
    //
    const unsigned int n_node = this->_node_map.size(),
                       n_elem = this->_node1.size(),
                       n_bc_node = this->_node_bc.size();
    if (n_elem!=this->_node2.size()) libmesh_error();
    // write data to Gmsh format file
    fout << "$MeshFormat" << std::endl;
    fout << "2.2 0 8" << std::endl;
    fout << "$EndMeshFormat" << std::endl;
    fout << "$Nodes" << std::endl;
    fout << n_node << std::endl;
    for ( std::map<unsigned int, std::vector<Real> >::const_iterator
          it=this->_node_map.begin(); it!=this->_node_map.end(); it++) {
        fout << it->first << "  "
             << (it->second)[0] << ' '
             << (it->second)[1] << ' '
             << (it->second)[2] << ' '
             << std::endl;
        // end of for loop...
    }
    fout << "$EndNodes" << std::endl;
    fout << "$Elements" << std::endl;
    fout << (n_bc_node+n_elem) << std::endl;
    unsigned int index = 1;
    for ( std::map<unsigned int, unsigned int>::const_iterator
          it=this->_node_bc.begin(); it!=this->_node_bc.end(); it++) {
        const unsigned int pid = 1 + index;
        fout << index << " 15  2 " << pid << ' ' << it->second << "  "
             << it->first << ' '
             << std::endl;
        ++index;
    }
    for (unsigned int i=0; i<n_elem; i++) {
        const unsigned int pid = 100001 + index;
        fout << index << " 1  2 " << pid << ' ' << 0 << "  "
             << this->_node1[i] << ' '
             << this->_node2[i] << ' '
             << std::endl;
        ++index;
    }
    fout << "$EndElements" << std::endl;
    // close file stream normally
    fout.close();
    // end of routine...
}
//=================================================================================================
} // end of namespace: test
//=================================================================================================
void gen_rve ( const unsigned int N_FIBR, const Point& centre, const Real& L_RVE,
               const std::string& filename )
{
    // here we produce an RVE of length `L_RVE` centred around point p,
    // having a rather isotropic distribution of fibres.

    // initialise the RVE parameters
    unsigned int NF = N_FIBR;
    const unsigned int NF_const = N_FIBR;
    const Real L = 0.1;
    const Real D2 = pow2(0.08);
    const unsigned int NS = 2*NF_const+1;
    unsigned int NL = 2*NF;
    const unsigned int NT = 2*NF_const;
    // initialise RVE variables
    std::vector<Real> x; x.assign(NS, 0.0);
    std::vector<Real> y; y.assign(NS, 0.0);
    std::vector<Real> z; z.assign(NS, 0.0);
    std::vector<Real> x0; x0.assign(NS, 0.0);
    std::vector<Real> y0; y0.assign(NS, 0.0);
    std::vector<Real> z0; z0.assign(NS, 0.0);
    std::vector<Real> dx; dx.assign(3*NS, 0.0);
    std::vector<Real> dy; dy.assign(3*NS, 0.0);
    std::vector<Real> dz; dz.assign(3*NS, 0.0);
    std::vector<unsigned int> a; a.assign(3*NT+1, 0);
    // RVE boundaries
    const Real x_max = centre(0) + 0.5*L_RVE;
    const Real x_min = centre(0) - 0.5*L_RVE;
    const Real y_max = centre(1) + 0.5*L_RVE;
    const Real y_min = centre(1) - 0.5*L_RVE;
    const Real z_max = centre(2) + 0.5*L_RVE;
    const Real z_min = centre(2) - 0.5*L_RVE;
    const Real btol = 1.0e-6;
    //
    // initialise a random number generator
    std::random_device rd;
    std::default_random_engine rgen(rd());
    // initialise uniform and normal distributions
    std::uniform_real_distribution<Real> udist(0.0, 1.0);
    std::normal_distribution<Real> ndist(0.65,0.1);
    //
    // create initial seeds
    for (unsigned int i=1; i<=NF; i++) {
        x0[i] = 2.0*udist(rgen);
        y0[i] = 2.0*udist(rgen);
        z0[i] = 2.0*udist(rgen);
        x[2*i-1] = x0[i];
        y[2*i-1] = y0[i];
        z[2*i-1] = z0[i];
        x[2*i] = x0[i];
        y[2*i] = y0[i];
        z[2*i] = z0[i];
        //
        const Real dx1 = udist(rgen) - 0.5;
        const Real dy1 = udist(rgen) - 0.5;
        const Real dz1 = udist(rgen) - 0.5;
        const Real d1 = sqrt(pow2(dx1) + pow2(dy1) + pow2(dz1));
        //
        dx[2*i-1] =  dx1/d1;
        dx[2*i]   = -dx1/d1;
        dy[2*i-1] =  dy1/d1;
        dy[2*i]   = -dy1/d1;
        dz[2*i-1] =  dz1/d1;
        dz[2*i]   = -dz1/d1;
        a[2*i-1] = 2*i-1;
        a[2*i]   = 2*i;
    }
    // create vector of "living" segments
    std::vector<unsigned int> live;
    for (unsigned int i=0; i<=NL; i++) live.push_back(i);
    //
    // start adding monomer
    while (NL > 0) {
        //
        unsigned int g = 1;
        // select fibre to update; +1 because matlab indices begin at 1 and floor(...) can equal 0
        unsigned int j = floor(udist(rgen)*NL) + 1;
        unsigned int k = live[j];
        // update fibre characteristics
        x[k] += L*dx[k];
        y[k] += L*dy[k];
        z[k] += L*dz[k];
        //
        Real x1 = x[k];
        Real y1 = y[k];
        Real z1 = z[k];
        // check for fibre out of big box
        Real u[6];
        u[0] = (x1 - 2.0)/fabs(dx[k]);
        u[1] = (y1 - 2.0)/fabs(dy[k]);
        u[2] = (z1 - 2.0)/fabs(dz[k]);
        u[3] = -x1/fabs(dx[k]);
        u[4] = -y1/fabs(dy[k]);
        u[5] = -z1/fabs(dz[k]);
        Real Lu = u[0];
        // find max value of u
        for (unsigned int j=1; j<6; j++) {
            if (u[j] > Lu) Lu = u[j];
        }
        // fibre is out of box
        if (Lu > 0.0) {
            x1 -= Lu*dx[k];
            y1 -= Lu*dy[k];
            z1 -= Lu*dz[k];
            // kill segment
            for (unsigned int m=j; m<=(NL-1); m++)
                live[m] = live[m+1];
            //
            NL--;
            g = 0;
        }
        // check for fibre-fibre collision
        if ( 1 == g ) {
            unsigned int i = 0;
            if ( (1==k) || (2==k) ) i=2;
            else i=1;
            //
            while (i < NF) {
                // coordinates of ith fiber nodes
                const Real xi0 = x[a[2*i-1]];
                const Real yi0 = y[a[2*i-1]];
                const Real zi0 = z[a[2*i-1]];
                const Real xi1 = x[a[2*i]];
                const Real yi1 = y[a[2*i]];
                const Real zi1 = z[a[2*i]];
                // h - length of ith fiber
                // h1 - distance between current node (k) and even node of ith fiber
                // h2 - distance between current node (k) and odd node of ith fiber
                const Real h = sqrt( pow2(xi1-xi0) + pow2(yi1-yi0) + pow2(zi1-zi0) );
                const Real h1 = sqrt( pow2(xi1-x1) + pow2(yi1-y1) + pow2(zi1-z1) );
                const Real h2 = sqrt( pow2(xi0-x1) + pow2(yi0-y1) + pow2(zi0-z1) );
                // formula based on distance between a point and a line in 3D
                // http://mathworld.wolfram.com/Point-LineDistance3-Dimensional.html
                if ( (h1 < h) && (h2 < h) ) {
                    const Real aa = pow2(xi1-x1)+pow2(yi1-y1)+pow2(zi1-z1);
                    const Real bb = pow2(xi0-xi1)+pow2(yi0-yi1)+pow2(zi0-zi1);
                    const Real cc = pow2((xi1-x1)*(xi0-xi1)+(yi1-y1)*(yi0-yi1)+(zi1-z1)*(zi0-zi1));
                    const Real dist2 = (aa*bb-cc)/bb;
                    // collision!
                    if ( dist2 < D2 ) {
                        const Real tt = -((xi1-x1)*(xi0-xi1)+(yi1-y1)*(yi0-yi1)+(zi1-z1)*(zi0-zi1))/bb;
                        //
                        x1 = xi1+(xi0-xi1)*tt;
                        y1 = yi1+(yi0-yi1)*tt;
                        z1 = zi1+(zi0-zi1)*tt;
                        // split struck fibre in two
                        NF++;
                        //
                        a[2*NF-1] = k;
                        a[2*NF  ] = a[2*i];
                        //
                        dx[2*NF] = dx[2*i];
                        dy[2*NF] = dy[2*i];
                        dz[2*NF] = dz[2*i];
                        dx[2*NF-1] = dx[2*i-1];
                        dy[2*NF-1] = dy[2*i-1];
                        dz[2*NF-1] = dz[2*i-1];
                        //
                        a[2*i] = k;
                        // kill segment in collision
                        for (unsigned int m=j; m<=(NL-1); m++) {
                            live[m] = live[m+1];
                        }
                        //
                        NL--;
                        // skip remainder of search
                        i = NF;
                    }
                }
                //
                i++;
                // skip collision with self!
                if ( (a[2*i-1] == k) || (a[2*i] == k) ) i++;
                // end of while loop
            }
            // end of conditional
        }
        //
        g = 1;
        x[k] = x1;
        y[k] = y1;
        z[k] = z1;
        // end of while loop
    }
    //
    // coordinate containers
    std::vector<Real> temp; temp.assign(NF, 0.0);
    std::vector< std::vector<Real> > Ax; Ax.assign(2, temp);
    std::vector< std::vector<Real> > Ay; Ay.assign(2, temp);
    std::vector< std::vector<Real> > Az; Az.assign(2, temp);
    std::vector<Real> b; b.assign(2*NF, 0.0);
    unsigned int k = 0;
    // if we wanted to write the Lu stuff to file, we'd open one here...
    //
    // select fibres with nodes in the box [x,y,z](-0.5,0.5)
    for (unsigned int i=1; i<=NF; i++) {
        //
        if ( (x[a[2*i]] > 0.51) && (x[a[2*i]] < 1.49) &&
             (y[a[2*i]] > 0.51) && (y[a[2*i]] < 1.49) &&
             (z[a[2*i]] > 0.51) && (z[a[2*i]] < 1.49) )
        {
            //
            Real u[6];
            u[0] = (x[a[2*i-1]] - 1.49)/fabs(dx[2*i-1]);
            u[1] = (y[a[2*i-1]] - 1.49)/fabs(dy[2*i-1]);
            u[2] = (z[a[2*i-1]] - 1.49)/fabs(dz[2*i-1]);
            u[3] = (0.51 - x[a[2*i-1]])/fabs(dx[2*i-1]);
            u[4] = (0.51 - y[a[2*i-1]])/fabs(dy[2*i-1]);
            u[5] = (0.51 - z[a[2*i-1]])/fabs(dz[2*i-1]);
            Real Lu = u[0];
            // find max value of u
            for (unsigned int j=1; j<6; j++) {
                if ( u[j] > Lu ) Lu = u[j];
            }
            // node a[2*i-1] is out of box?
            if ( Lu > 0.0 ) {
                u[0] = (x[a[2*i-1]] - 1.5)/fabs(dx[2*i-1]);
                u[1] = (y[a[2*i-1]] - 1.5)/fabs(dy[2*i-1]);
                u[2] = (z[a[2*i-1]] - 1.5)/fabs(dz[2*i-1]);
                u[3] = (0.5 - x[a[2*i-1]])/fabs(dx[2*i-1]);
                u[4] = (0.5 - y[a[2*i-1]])/fabs(dy[2*i-1]);
                u[5] = (0.5 - z[a[2*i-1]])/fabs(dz[2*i-1]);
                Real Lu1 = u[0];
                // find max value of u
                for (unsigned int j=1; j<6; j++) {
                    if ( u[j] > Lu1 ) Lu1 = u[j];
                }
                x[a[2*i-1]] -= Lu1*dx[2*i-1];
                y[a[2*i-1]] -= Lu1*dy[2*i-1];
                z[a[2*i-1]] -= Lu1*dz[2*i-1];
                // print stuff here...
            }
            //
            k++;
            //
            Ax[0][k] = x[a[2*i-1]] - 1.0;
            Ax[1][k] = x[a[2*i]]   - 1.0;
            Ay[0][k] = y[a[2*i-1]] - 1.0;
            Ay[1][k] = y[a[2*i]]   - 1.0;
            Az[0][k] = z[a[2*i-1]] - 1.0;
            Az[1][k] = z[a[2*i]]   - 1.0;
            b[2*k-1] = a[2*i-1];
            b[2*k]   = a[2*i];
            //
        } else if ( (x[a[2*i-1]] > 0.51) && (x[a[2*i-1]] < 1.49) &&
                    (y[a[2*i-1]] > 0.51) && (y[a[2*i-1]] < 1.49) &&
                    (z[a[2*i-1]] > 0.51) && (z[a[2*i-1]] < 1.49) )
        {
            //
            Real u[6];
            u[0] = (x[a[2*i]] - 1.49)/fabs(dx[2*i]);
            u[1] = (y[a[2*i]] - 1.49)/fabs(dy[2*i]);
            u[2] = (z[a[2*i]] - 1.49)/fabs(dz[2*i]);
            u[3] = (0.51 - x[a[2*i]])/fabs(dx[2*i]);
            u[4] = (0.51 - y[a[2*i]])/fabs(dy[2*i]);
            u[5] = (0.51 - z[a[2*i]])/fabs(dz[2*i]);
            Real Lu = u[0];
            // find max value of u
            for (unsigned int j=1; j<6; j++) {
                if ( u[j] > Lu ) Lu = u[j];
            }
            // node a[2*i] is out of box?
            if( Lu > 0.0 ) {
                u[0] = (x[a[2*i]] - 1.5)/fabs(dx[2*i]);
                u[1] = (y[a[2*i]] - 1.5)/fabs(dy[2*i]);
                u[2] = (z[a[2*i]] - 1.5)/fabs(dz[2*i]);
                u[3] = (0.5 - x[a[2*i]])/fabs(dx[2*i]);
                u[4] = (0.5 - y[a[2*i]])/fabs(dy[2*i]);
                u[5] = (0.5 - z[a[2*i]])/fabs(dz[2*i]);
                Real Lu1 = u[0];
                // find max value of u
                for (unsigned int j=1; j<6; j++) {
                    if ( u[j] > Lu1 ) Lu1 = u[j];
                }
                //
                x[a[2*i]] -= Lu1*dx[2*i];
                y[a[2*i]] -= Lu1*dy[2*i];
                z[a[2*i]] -= Lu1*dz[2*i];
            }
            //
            k++;
            //
            Ax[0][k] = x[a[2*i-1]] - 1.0;
            Ax[1][k] = x[a[2*i]]   - 1.0;
            Ay[0][k] = y[a[2*i-1]] - 1.0;
            Ay[1][k] = y[a[2*i]]   - 1.0;
            Az[0][k] = z[a[2*i-1]] - 1.0;
            Az[1][k] = z[a[2*i]]   - 1.0;
            b[2*k-1] = a[2*i-1];
            b[2*k]   = a[2*i];
            //
        }
        // end of for loop...
    }
    // would close Lu file here...
    NF = k;
    //
    //
    //
    std::vector<unsigned int> c;
    for (unsigned int i=0; i<=(2*NF); i++) { c.push_back(0); }
    unsigned int idex = 0;
    unsigned int count = 1;
    c[1] = 1;
    for (unsigned int i=2; i<=(2*NF); i++) {
        for (unsigned int j=1; j<=(i-1); j++) {
            if ( b[i] == b[j] ) {
                c[i] = c[j];
                idex = 1;
            }
        }
        if ( 0 == idex ) {
            count += 1;
            c[i] = count;
        }
        idex = 0;
    }
    //
    //
    //
    Real len = 0.0;
    Real proj_len = 0.0;
    Real om11 = 0.0;
    Real om12 = 0.0;
    Real om13 = 0.0;
    Real om22 = 0.0;
    Real om23 = 0.0;
    Real om33 = 0.0;
    std::vector<Real> length;
    std::vector<unsigned int> node1;
    std::vector<unsigned int> node2;
    std::vector<Real> node1_xcoord;
    std::vector<Real> node1_ycoord;
    std::vector<Real> node1_zcoord;
    std::vector<Real> node2_xcoord;
    std::vector<Real> node2_ycoord;
    std::vector<Real> node2_zcoord;
    // an array of sets  - one per surface - to store the boundary node ids
    std::set<unsigned int> varr[6];
    //
    const Real L_x = L_RVE;
    const Real L_y = L_RVE;
    const Real L_z = L_RVE;
    //
    // store node id and position
    for (unsigned int i=1; i<=NF; i++) {
        // FIXME: should do a linear interpolation
        // check if fibre is still outside mesh... if it is,
        // then pull the offending node back to the boundary
        for (unsigned ii=0; ii<2; ii++) {
            if ( Ax[ii][i] >  0.5 ) Ax[ii][i] =  0.5;
            if ( Ax[ii][i] < -0.5 ) Ax[ii][i] = -0.5;
            if ( Ay[ii][i] >  0.5 ) Ay[ii][i] =  0.5;
            if ( Ay[ii][i] < -0.5 ) Ay[ii][i] = -0.5;
            if ( Az[ii][i] >  0.5 ) Az[ii][i] =  0.5;
            if ( Az[ii][i] < -0.5 ) Az[ii][i] = -0.5;
        }
        //
        node1.push_back(c[2*i-1]);
        node2.push_back(c[2*i]);
        node1_xcoord.push_back(centre(0) + Ax[0][i]*L_x);
        node1_ycoord.push_back(centre(1) + Ay[0][i]*L_y);
        node1_zcoord.push_back(centre(2) + Az[0][i]*L_z);
        node2_xcoord.push_back(centre(0) + Ax[1][i]*L_x);
        node2_ycoord.push_back(centre(1) + Ay[1][i]*L_y);
        node2_zcoord.push_back(centre(2) + Az[1][i]*L_z);
        //
        // FIXME: this will copy node to > 1 surface if it's on an interface
        // identify boundary nodes
        //
        if ( fabs(node1_zcoord[i-1] - z_min) < btol) varr[0].insert(node1[i-1]);
        if ( fabs(node1_ycoord[i-1] - y_min) < btol) varr[1].insert(node1[i-1]);
        if ( fabs(node1_xcoord[i-1] - x_max) < btol) varr[2].insert(node1[i-1]);
        if ( fabs(node1_ycoord[i-1] - y_max) < btol) varr[3].insert(node1[i-1]);
        if ( fabs(node1_xcoord[i-1] - x_min) < btol) varr[4].insert(node1[i-1]);
        if ( fabs(node1_zcoord[i-1] - z_max) < btol) varr[5].insert(node1[i-1]);

        if ( fabs(node2_zcoord[i-1] - z_min) < btol) varr[0].insert(node2[i-1]);
        if ( fabs(node2_ycoord[i-1] - y_min) < btol) varr[1].insert(node2[i-1]);
        if ( fabs(node2_xcoord[i-1] - x_max) < btol) varr[2].insert(node2[i-1]);
        if ( fabs(node2_ycoord[i-1] - y_max) < btol) varr[3].insert(node2[i-1]);
        if ( fabs(node2_xcoord[i-1] - x_min) < btol) varr[4].insert(node2[i-1]);
        if ( fabs(node2_zcoord[i-1] - z_max) < btol) varr[5].insert(node2[i-1]);
        //
        length.push_back(sqrt( pow2(node1_xcoord[i-1] - node2_xcoord[i-1])
                             + pow2(node1_ycoord[i-1] - node2_ycoord[i-1])
                             + pow2(node1_zcoord[i-1] - node2_zcoord[i-1]) ));
        len += length[i-1];
        proj_len += sqrt( pow2(node1_xcoord[i-1] - node2_xcoord[i-1])
                        + pow2(node1_ycoord[i-1] - node2_ycoord[i-1]) );
        //
        // calculate components of the orientation tensor
        const Real cosa = (node2_xcoord[i-1] - node1_xcoord[i-1])/length[i-1];
        const Real cosb = (node2_ycoord[i-1] - node1_ycoord[i-1])/length[i-1];
        const Real cosg = (node2_zcoord[i-1] - node1_zcoord[i-1])/length[i-1];
        om11 += length[i-1]*cosg*cosg;
        om12 += length[i-1]*cosg*cosb;
        om13 += length[i-1]*cosg*cosa;
        om22 += length[i-1]*cosb*cosb;
        om23 += length[i-1]*cosa*cosb;
        om33 += length[i-1]*cosa*cosa;
        // end of for loop
    }
    //
    // FIXME: this can create duplicate entries in different sets
    // fill boundary node ids for each face
    std::vector< std::set<unsigned int> > bnode_ids;
    for (unsigned i=0; i<6; i++) {
        bnode_ids.push_back(varr[i]);
    }
    // set face qp's
    // the faces are indexed like this:
    /*
    HEX8:
        o--------o
       /:       /|
      / : 5    / |
     /  :   3 /  |
    o--------o 2 |
    | 4 o....|...o
    |  . 1   |  /
    | .   0  | /
    |.       |/
    o--------o

    */
    //
    // note the vector must be filled in the same order as the bnode_ids (and hence
    // the boundary norms returned by libmesh) ie. the faces must correspond
    Point qps[6];
    qps[0] = Point(0.5*x_max, 0.5*y_max, z_min);
    qps[1] = Point(0.5*x_max, y_min, 0.5*z_max);
    qps[2] = Point(x_max, 0.5*y_max, 0.5*z_max);
    qps[3] = Point(0.5*x_max, y_max, 0.5*z_max);
    qps[4] = Point(x_min, 0.5*y_max, 0.5*z_max);
    qps[5] = Point(0.5*x_max, 0.5*y_max, z_max);
    //
    std::vector< Point > face_qps;
    for (unsigned i=0; i<6; i++) {
        face_qps.push_back(qps[i]);
    }
    //
    om11 /= len;
    om12 /= len;
    om13 /= len;
    om22 /= len;
    om23 /= len;
    om33 /= len;
    //
    const unsigned int NODES = count;
    const unsigned int GSSIZE = 3*NODES;
    const unsigned int ELEMENTS = NF;
    // let's sort the nodes out - use a map to order by id and remove duplicates
    std::map<const unsigned int, std::vector<Real> > node_map;
    for (unsigned int i=0; i<NF; i++) {
        std::vector<Real> temp_vec1;
        temp_vec1.push_back(node1_xcoord[i]);
        temp_vec1.push_back(node1_ycoord[i]);
        temp_vec1.push_back(node1_zcoord[i]);
        //
        std::vector<Real> temp_vec2;
        temp_vec2.push_back(node2_xcoord[i]);
        temp_vec2.push_back(node2_ycoord[i]);
        temp_vec2.push_back(node2_zcoord[i]);
        //
        node_map.insert(std::make_pair(node1[i], temp_vec1));
        node_map.insert(std::make_pair(node2[i], temp_vec2));
    }
    //
    Real totL = 0.0;
    for (unsigned int i=0; i<ELEMENTS; i++) {
        std::vector<Real> n1_xyz = node_map.find(node1[i])->second;
        std::vector<Real> n2_xyz = node_map.find(node2[i])->second;
        //
        totL += sqrt( pow2(n1_xyz[0]-n2_xyz[0])
                    + pow2(n1_xyz[1]-n2_xyz[1])
                    + pow2(n1_xyz[2]-n2_xyz[2]) );
    }
    //
    // write data to Gmsh format file
    std::ofstream omsh(filename.c_str());
    omsh << "$MeshFormat" << std::endl;
    omsh << "2.2 0 8" << std::endl;
    omsh << "$EndMeshFormat" << std::endl;
    omsh << "$Nodes" << std::endl;
    omsh << node_map.size() << std::endl;
    for ( std::map<const unsigned int, std::vector<Real> >::iterator
          it=node_map.begin(); it!=node_map.end(); it++ ) {
        omsh << it->first
             << ' ' << (it->second)[0]
             << ' ' << (it->second)[1]
             << ' ' << (it->second)[2]
             << std::endl;
    }
    omsh << "$EndNodes" << std::endl;
    omsh << "$Elements" << std::endl;
    omsh << ELEMENTS << std::endl;
    for(unsigned int i=0; i<ELEMENTS; i++) {
        omsh << i+1
             << ' ' << 1
             << ' ' << " 2 99 2 "
             << ' ' << node1[i]
             << ' ' << node2[i]
             << std::endl;
    }
    omsh << "$EndElements" << std::endl;
    omsh.close();
    // completed the task generating the cubic RVE...
}
//=================================================================================================
void test_generate_rve () {
    test::RVE a;
    a.generate(1000, 0.0, 300.0);
    a.ouput("rve.msh");
    //
    gen_rve(1000, Point(150.0, 150.0, 150.0), 300.0, "rve~test.msh");
}
//=================================================================================================
