/*************************************************************************************************

    Copyright (C) 2009-2019    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/utils.h"
#include "feb3/domain_data.h"
#include "libmesh/mesh.h"
#include "libmesh/equation_systems.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
//=================================================================================================
typedef std::pair<int, int>  SegmentID_NodeID;
//=================================================================================================
struct NodalData {
    NodalData () : is_active(false) { xyz=Point(); rad=0.0; }
    ~NodalData () {}
    //
    bool is_active;
    Point xyz;
    Real rad;
    std::vector<SegmentID_NodeID> segm;
};
struct SegmentData {
    SegmentData () : is_active(false) { conn[0]=conn[1]=0; }
    ~SegmentData () {}
    //
    bool is_active;
    int conn[2];
};
// local parameters
static
const std::string in_fname = "vascular_network.in";
static
const std::string fname = "vascular_network.res";
// local variables
static
PerfLog plog("translate_vascular_network");
static
Real scale_param = 1.0,
     min_segm_length = 0.0;
static
Real ratio_x_trans = 0.0,
     ratio_y_trans = 0.0,
     ratio_z_trans = 0.0;
static
std::vector<NodalData> nodes;
static
std::vector<SegmentData> segms;
static
std::map<int, VascularInfo::Type> map_b_nodes;
static
std::map<int, dof_id_type> map_active_nodes_ids;
static
std::map<int, dof_id_type> map_active_segms_ids;
//=================================================================================================
static
void read_mesh (const std::string& name, const bool scale) {
    plog.push("read_mesh");
    //
    std::ifstream fin(name);
    //
    unsigned int index, n_nodes, n_segms, n_b_nodes;
    Real min_x(0.0), max_x(0.0),
         min_y(0.0), max_y(0.0),
         min_z(0.0), max_z(0.0);
    //
    nodes.clear();
    segms.clear();
    map_b_nodes.clear();
    //
    // read the geometric scaling parameter and the minimum length of a vasculat segment
    fin >> scale_param;
    fin >> min_segm_length;
    //
    // read the parameters that would define how further to translate the mesh of the vascular tree
    fin >> ratio_x_trans >> ratio_y_trans >> ratio_z_trans;
    //
    std::cout << ' ' << scale_param
              << ' ' << min_segm_length
              << ' ' << ratio_x_trans << ' ' << ratio_y_trans << ' ' << ratio_z_trans << std::endl;
    //
    fin >> n_nodes;
    for (index=0; index<n_nodes; index++) {
        int n_id;
        fin >> n_id;
        if (n_id!=index) libmesh_error();
        //
        NodalData nd;
        nd.is_active = true;
        fin >> nd.xyz(0) >> nd.xyz(1) >> nd.xyz(2) >> nd.rad;
        //
        if ( scale ) nd.xyz *= scale_param;
        if ( scale ) nd.rad *= scale_param;
        //
        min_x = std::min(min_x, nd.xyz(0)); max_x = std::max(max_x, nd.xyz(0));
        min_y = std::min(min_y, nd.xyz(1)); max_y = std::max(max_y, nd.xyz(1));
        min_z = std::min(min_z, nd.xyz(2)); max_z = std::max(max_z, nd.xyz(2));
        //
        nodes.push_back(nd);
    }
    std::cout << " >> Read " << n_nodes << " vascular nodes." << std::endl;
    //
    fin >> n_segms;
    for (index=0; index<n_segms; index++) {
        int s_id;
        fin >> s_id;
        if (s_id!=index) libmesh_error();
        //
        SegmentData sc;
        sc.is_active = true;
        fin >> sc.conn[0] >> sc.conn[1];
        if (sc.conn[0]>=n_nodes) libmesh_error();
        if (sc.conn[1]>=n_nodes) libmesh_error();
        //
        segms.push_back(sc);
    }
    std::cout << " >> Read " << n_segms << " vascular segments." << std::endl;
    //
    fin >> n_b_nodes;
    for (index=0; index<n_b_nodes; index++) {
        int n_id;
        fin >> n_id;
        if (n_id>=n_nodes) libmesh_error();
        //
        int bc;
        fin >> bc;
        VascularInfo::Type t;
        if      (101==bc) t = VascularInfo::INLET;
        else if (102==bc) t = VascularInfo::OUTLET;
        else if (106==bc) t = VascularInfo::TIP;
        else libmesh_error();
        //
        map_b_nodes.insert( std::make_pair(n_id, t) );
    }
    std::cout << " >> Read " << n_b_nodes << " vascular boundary nodes." << std::endl;
    //
    const Real dl_x = (max_x-min_x)*ratio_x_trans, rc_x = dl_x - min_x,
               dl_y = (max_y-min_y)*ratio_y_trans, rc_y = dl_y - min_y,
               dl_z = (max_z-min_z)*ratio_z_trans, rc_z = dl_z - min_z;
    //
    if ( scale )
    for (unsigned int i=0; i<nodes.size(); i++) {
        nodes[i].xyz(0) += rc_x;
        nodes[i].xyz(1) += rc_y;
        nodes[i].xyz(2) += rc_z;
    }
    //
    //
    //
    if ( scale )
    {
        //
        std::ofstream fout(fname+".stats");
        //
        fout << "rc_x = " << rc_x << ";" << std::endl;
        fout << "rc_y = " << rc_y << ";" << std::endl;
        fout << "rc_z = " << rc_z << ";" << std::endl;
        //
        max_x = 0.0;
        max_y = 0.0;
        max_z = 0.0;
        for (unsigned int i=0; i<nodes.size(); i++) {
            const NodalData& nd = nodes[i];
            max_x = std::max(max_x, nd.xyz(0));
            max_y = std::max(max_y, nd.xyz(1));
            max_z = std::max(max_z, nd.xyz(2));
        }
        min_x = max_x;
        min_y = max_y;
        min_z = max_z;
        for (unsigned int i=0; i<nodes.size(); i++) {
            const NodalData& nd = nodes[i];
            min_x = std::min(min_x, nd.xyz(0));
            min_y = std::min(min_y, nd.xyz(1));
            min_z = std::min(min_z, nd.xyz(2));
        }
        //
        fout << "min_x = " << min_x << "; max_x = " << max_x << "; LEN_x = " << min_x+max_x << ";" << std::endl;
        fout << "min_y = " << min_y << "; max_y = " << max_y << "; LEN_y = " << min_y+max_y << ";" << std::endl;
        fout << "min_z = " << min_z << "; max_z = " << max_z << "; LEN_z = " << min_z+max_z << ";" << std::endl;
        //
    }
    //
    plog.pop("read_mesh");
}
//=================================================================================================
static
void prepare_mesh () {
    plog.push("prepare_mesh");
    //
    dof_id_type index = 0;
    map_active_nodes_ids.clear();
    map_active_segms_ids.clear();
    //
    index = 0;
    for (unsigned int n=0; n<nodes.size(); n++) {
        if ( nodes[n].is_active ) continue;
        //
        std::map<int, VascularInfo::Type>::iterator mi = map_b_nodes.find(n);
        if ( map_b_nodes.end()!=mi ) map_b_nodes.erase(mi);
    }
    //
    index = 0;
    for (unsigned int n=0; n<nodes.size(); n++) {
        if ( ! nodes[n].is_active ) continue;
        //
        map_active_nodes_ids.insert( std::make_pair(n, index) );
        ++index;
    }
    if (!index) libmesh_error();
    //
    index = 0;
    for (unsigned int s=0; s<segms.size(); s++) {
        if ( ! segms[s].is_active ) continue;
        //
        map_active_segms_ids.insert( std::make_pair(s, index) );
        ++index;
    }
    if (!index) libmesh_error();
    //

    //
    plog.pop("prepare_mesh");
}
//=================================================================================================
static
void write_mesh (const std::string& name) {
    plog.push("write_mesh");
    //
    std::ofstream fout(name);
    //
    fout << 1.0 << std::endl
         << min_segm_length << std::endl
         << 0.0 << ' ' << 0.0 << ' ' << 0.0 << std::endl;
    //
    const unsigned n_nodes = map_active_nodes_ids.size(),
                   n_segms = map_active_segms_ids.size(),
                   n_b_nodes = map_b_nodes.size();
    //
    fout << n_nodes << std::endl;
    for ( std::map<int, dof_id_type>::const_iterator
          mci=map_active_nodes_ids.begin(); mci!=map_active_nodes_ids.end(); mci++ ) {
        fout << mci->second;
        //
        const int N = mci->first;
        //
        fout << ' ' << nodes[N].xyz(0) << ' ' << nodes[N].xyz(1) << ' ' << nodes[N].xyz(2);
        fout << ' ' << nodes[N].rad;
        fout << std::endl;
    }
    std::cout << " >> Wrote " << n_nodes << " vascular nodes." << std::endl;
    //
    fout << n_segms << std::endl;
    for ( std::map<int, dof_id_type>::const_iterator
          mci=map_active_segms_ids.begin(); mci!=map_active_segms_ids.end(); mci++ ) {
        fout << mci->second;
        //
        const int S = mci->first;
        //
        std::map<int, dof_id_type>::const_iterator mcj = map_active_nodes_ids.find(segms[S].conn[0]);
        if ( map_active_nodes_ids.end() == mcj ) libmesh_error();
        std::map<int, dof_id_type>::const_iterator mck = map_active_nodes_ids.find(segms[S].conn[1]);
        if ( map_active_nodes_ids.end() == mck ) libmesh_error();
        //
        fout << ' ' << mcj->second << ' ' << mck->second;
        fout << std::endl;
    }
    std::cout << " >> Wrote " << n_segms << " vascular segments." << std::endl;
    //
    fout << n_b_nodes << std::endl;
    std::cout << " >> Wrote " << n_b_nodes << " vascular boundary nodes." << std::endl;
    for ( std::map<int, VascularInfo::Type>::const_iterator
          mci=map_b_nodes.begin(); mci!=map_b_nodes.end(); mci++ ) {
        std::map<int, dof_id_type>::const_iterator mcj = map_active_nodes_ids.find(mci->first);
        if ( map_active_nodes_ids.end() == mcj ) libmesh_error();
        //
        fout << mcj->second;
        fout << ' ' << mci->second;
        fout << std::endl;
    }
    //
    plog.pop("write_mesh");
}
//=================================================================================================
static
void write_data (const std::string& name) {
    plog.push("write_data");
    //
    std::ofstream fout(name);
    //
    dof_id_type index = 0;
    //
    fout << map_active_nodes_ids.size() << std::endl;
    index = 1;
    for ( std::map<int, dof_id_type>::const_iterator
          mci=map_active_nodes_ids.begin(); mci!=map_active_nodes_ids.end(); mci++ ) {
        //
        fout << mci->second+1;
        if ( index != mci->second+1 ) libmesh_error();
        //
        const int N = mci->first;
        //
        fout << ' ' << nodes[N].rad;
        fout << std::endl;
        //
        ++index;
    }
    //
    plog.pop("write_data");
}
//=================================================================================================
static
void init_data (bool run_full_check =false) {
    plog.push("init_data");
    //
    for (unsigned int n=0; n<nodes.size(); n++) {
        nodes[n].segm.clear();
    }
    //
    for (unsigned int s=0; s<segms.size(); s++) {
        SegmentData& sd = segms[s];
        //
        if ( sd.conn[0] == sd.conn[1] )
            sd.is_active = false;
        //
        if ( ! sd.is_active ) continue;
        //
        for (unsigned int n=0; n<2; n++) {
            NodalData& nd = nodes[sd.conn[n]];
            //
            if ( run_full_check ) {
                if ( ! nd.is_active ) libmesh_error();
            }
            //
            const int n_conn = (0==n ? sd.conn[1] : sd.conn[0]);
            nd.segm.push_back( std::make_pair(s, n_conn) );
        }
    }
    //
    for (unsigned int n=0; n<nodes.size(); n++) {
        const NodalData& nd = nodes[n];
        //
        if ( ! nd.is_active ) {
            std::map<int, VascularInfo::Type>::iterator mi = map_b_nodes.find(n);
            if ( map_b_nodes.end()!=mi ) map_b_nodes.erase(mi);
            //
            continue;
        }
        //
        if ( run_full_check ) {
            if ( nd.segm.empty() ) {
                libmesh_error();
            } else if ( 4 <= nd.segm.size() ) {
                libmesh_error();
            } else if ( 1 == nd.segm.size() ) {
                std::map<int, VascularInfo::Type>::iterator mi = map_b_nodes.find(n);
                if ( map_b_nodes.end() != mi ) {
                    map_b_nodes.insert( std::make_pair(n, VascularInfo::TIP) );
                } else {
                    VascularInfo::Type& t = mi->second;
                    //
                    if      (VascularInfo::INLET  == t) ; // nothing to do here...
                    else if (VascularInfo::OUTLET == t) ; // nothing to do here...
                    else if (VascularInfo::TIP    == t) ; // nothing to do here...
                    else t = VascularInfo::TIP;
                }
            } else {
                std::map<int, VascularInfo::Type>::iterator mi = map_b_nodes.find(n);
                if ( map_b_nodes.end()!=mi ) map_b_nodes.erase(mi);
            }
        }
    }
    //
    plog.pop("init_data");
}
//=================================================================================================
static
void fix_branches () {
    plog.push("fix_branches");
    //
    std::cout << " >> Fix branches of vascular network mesh (1/2)." << std::endl;
    for (int ITER=0; ITER<5; ITER++)
    for (unsigned int n=0; n<nodes.size(); n++) {
        const NodalData& nd = nodes[n];
        //
        if ( ! nd.is_active ) continue;
        if ( nd.segm.size() <= 3 ) continue;
        //
        for (unsigned int S=0; S<(nd.segm.size()-1); S++) {
            unsigned int n_0, n_1;
            //
            SegmentData& sd_0 = segms[nd.segm[S+0].first];
            if ( ! sd_0.is_active ) libmesh_error();
            //
            if      (n == sd_0.conn[0]) n_0 = sd_0.conn[1];
            else if (n == sd_0.conn[1]) n_0 = sd_0.conn[0];
            else libmesh_error();
            //
            SegmentData& sd_1 = segms[nd.segm[S+1].first];
            if ( ! sd_1.is_active ) libmesh_error();
            //
            if      (n == sd_1.conn[0]) n_1 = sd_1.conn[1];
            else if (n == sd_1.conn[1]) n_1 = sd_1.conn[0];
            else libmesh_error();
            //
            if        ( nodes[n_0].segm.size() >  2 &&
                        nodes[n_1].segm.size() >  2 ) {
                ;
            } else if ( nodes[n_0].segm.size() <= 2 ) {
                //
                sd_1.conn[0] = n_0;
                sd_1.conn[1] = n_1;
                //
                break;
                //
            } else if ( nodes[n_1].segm.size() <= 2 ) {
                //
                sd_0.conn[0] = n_0;
                sd_0.conn[1] = n_1;
                //
                break;
                //
            } else libmesh_error();
        }
        //
        init_data();
    }
    //
    std::vector<NodalData> nodes_2add;
    unsigned int n_new = nodes.size();
    //
    std::cout << " >> Fix branches of vascular network mesh (2/2)." << std::endl;
    for (unsigned int n=0; n<nodes.size(); n++) {
        const NodalData& nd = nodes[n];
        //
        if ( ! nd.is_active ) continue;
        if ( nd.segm.size() <= 3 ) continue;
        //
        NodalData nd_new;
        nd_new.is_active = true;
        nd_new.xyz(0) = nd.xyz(0) * uniform_distro(0.999, 1.001);
        nd_new.xyz(1) = nd.xyz(1) * uniform_distro(0.999, 1.001);
        nd_new.xyz(2) = nd.xyz(2) * uniform_distro(0.999, 1.001);
        nd_new.rad    = nd.rad    * uniform_distro(0.999, 1.001);
        //
        SegmentData& sd_1 = segms[nd.segm[2].first];
        if ( ! sd_1.is_active ) libmesh_error();
        if      (n == sd_1.conn[0]) sd_1.conn[0] = n_new;
        else if (n == sd_1.conn[1]) sd_1.conn[1] = n_new;
        else libmesh_error();
        //
        SegmentData& sd_2 = segms[nd.segm[3].first];
        if ( ! sd_2.is_active ) libmesh_error();
        if      (n == sd_2.conn[0]) sd_2.conn[0] = n_new;
        else if (n == sd_2.conn[1]) sd_2.conn[1] = n_new;
        else libmesh_error();
        //
        nodes_2add.push_back(nd_new);
        //
        ++n_new;
    }
    //
    nodes.insert(nodes.end(), nodes_2add.begin(), nodes_2add.end());
    //
    init_data();
    //
    plog.pop("fix_branches");
}
//=================================================================================================
static
void coarsen_mesh() {
    plog.push("coarsen_mesh");
    //
    std::set<int> nodes_scanned;
    //
    std::cout << " >> Coarsen vascular network mesh." << std::endl;
    for (unsigned int n=0; n<nodes.size(); n++) {
        NodalData& nd = nodes[n];
        //
        if ( ! nd.is_active ) continue;
        if ( nd.segm.size() != 2 ) continue;
        //
        if ( nodes_scanned.end()!=nodes_scanned.find(n) ) continue;
        //
        SegmentData& sd_1 = segms[nd.segm[0].first];
        SegmentData& sd_2 = segms[nd.segm[1].first];
        const unsigned int n1 = nd.segm[0].second;
        const unsigned int n2 = nd.segm[1].second;
        const NodalData& nd_1 = nodes[n1];
        const NodalData& nd_2 = nodes[n2];
        if ( ! nd_1.is_active ) libmesh_error();
        if ( ! nd_2.is_active ) libmesh_error();
        if ( nd_1.segm.size() != 2 ) continue;
        if ( nd_2.segm.size() != 2 ) continue;
        //
        Point a = nd.xyz - nd_1.xyz;
        Point b = nd_2.xyz - nd.xyz;
        const Point a_ = a.unit();
        const Point b_ = b.unit();
        //
        if ( acos(a_*b_) > 0.2*mPI_2 ) continue;
        //
        SegmentData sd_new;
        sd_new.is_active = true;
        sd_new.conn[0] = n1;
        sd_new.conn[1] = n2;
        //
        nodes_scanned.insert(n);
        nodes_scanned.insert(n1);
        nodes_scanned.insert(n2);
        //
        nd.is_active = false;
        sd_1.is_active = false;
        sd_2.is_active = false;
        //
        segms.push_back(sd_new);
        //
        init_data();
    }
    //
    //
    for (unsigned int s=0; s<segms.size(); s++) {
        SegmentData& sd = segms[s];
        //
        if ( ! sd.is_active ) continue;
        //
        const NodalData& nd_0 = nodes[sd.conn[0]];
        const NodalData& nd_1 = nodes[sd.conn[1]];
        //
        if ( ! nd_0.is_active ) libmesh_error();
        if ( ! nd_1.is_active ) libmesh_error();
        //
        const Real length = Point(nd_0.xyz-nd_1.xyz).norm();
        //
        // if ( nd_0.segm.size() == 1 ) continue;
        // if ( nd_1.segm.size() == 1 ) continue;
        //
        if ( length > min_segm_length ) continue;
        libmesh_error();
        //
        if ( nd_0.segm.size() == 2 )
            map_b_nodes.insert( std::make_pair(sd.conn[0], VascularInfo::TIP) );
        if ( nd_1.segm.size() == 2 )
            map_b_nodes.insert( std::make_pair(sd.conn[1], VascularInfo::TIP) );
        //
        sd.is_active = false;
        //
        init_data();
    }
    //
    plog.pop("coarsen_mesh");
}
//=================================================================================================
static
void write_gmsh (const std::string& name) {
    plog.push("write_gmsh");
    //
    std::ofstream fout(name);
    //
    dof_id_type index = 0;
    //
    fout << "$MeshFormat" << std::endl;
    fout << 2.2 << ' ' << 0 << ' ' << 8 << std::endl;
    fout << "$EndMeshFormat" << std::endl;
    //
    fout << "$Nodes" << std::endl;
    fout << map_active_nodes_ids.size() << std::endl;
    index = 1;
    for ( std::map<int, dof_id_type>::const_iterator
          mci=map_active_nodes_ids.begin(); mci!=map_active_nodes_ids.end(); mci++ ) {
        //
        fout << mci->second+1;
        if ( index != mci->second+1 ) libmesh_error();
        //
        const int N = mci->first;
        const NodalData& nd = nodes[N];
        fout << ' ' << nd.xyz(0)
             << ' ' << nd.xyz(1)
             << ' ' << nd.xyz(2);
        fout << std::endl;
        //
        ++index;
    }
    fout << "$EndNodes" << std::endl;
    //
    fout << "$Elements" << std::endl;
    fout << map_active_segms_ids.size()+map_b_nodes.size() << std::endl;
    index = 1;
    for ( std::map<int, dof_id_type>::const_iterator
          mci=map_active_segms_ids.begin(); mci!=map_active_segms_ids.end(); mci++ ) {
        //
        fout << mci->second+1;
        if ( index != mci->second+1 ) libmesh_error();
        //
        const int S = mci->first;
        //
        if ( segms[S].conn[0] == segms[S].conn[1] ) libmesh_error();
        std::map<int, dof_id_type>::const_iterator mcj = map_active_nodes_ids.find(segms[S].conn[0]);
        if ( map_active_nodes_ids.end() == mcj ) libmesh_error();
        std::map<int, dof_id_type>::const_iterator mck = map_active_nodes_ids.find(segms[S].conn[1]);
        if ( map_active_nodes_ids.end() == mck ) libmesh_error();
        //
        fout << " 1   2 1001 0   ";
        fout << mcj->second+1 << ' ' << mck->second+1;
        fout << std::endl;
        //
        ++index;
    }
    for ( std::map<int, VascularInfo::Type>::const_iterator
          mci=map_b_nodes.begin(); mci!=map_b_nodes.end(); mci++ ) {
        //
        fout << index;
        //
        std::map<int, dof_id_type>::const_iterator mcj = map_active_nodes_ids.find(mci->first);
        if ( map_active_nodes_ids.end() == mcj ) libmesh_error();
        //
        fout << " 15   2 " << mci->second << " 0   ";
        fout << mcj->second+1;
        fout << std::endl;
        //
        ++index;
    }
    fout << "$EndElements" << std::endl;
    //
    std::cout << " >> Written vascular network mesh." << std::endl;
    //
    plog.pop("write_gmsh");
}
//=================================================================================================
void translate_vascular_network () {
    //
    std::string tmp_fname = "~vn~";
    std::cout << std::endl;
    for (int I=1; I<=12; I++) {
        const std::string in = (1==I ? in_fname : tmp_fname);
        const bool scale = (1==I ? true : false);
        //
        std::cout << " ** Iteration: " << I << std::endl;
        read_mesh(in, scale);
        init_data();
        if ( I<10 ) coarsen_mesh();
        else fix_branches();
        init_data();
        prepare_mesh();
        write_mesh(tmp_fname);
        // if (1==I) write_gmsh(tmp_fname+".msh");
    }
    for (int I=1; I<=3; I++) {
        init_data(true);
        prepare_mesh();
    }
    write_mesh(fname+".in");
    write_gmsh(fname+".msh");
    write_data(fname+".vi");
    remove(tmp_fname.c_str());
    //
    std::cout << " >> Mesh translated successfully.\n" << std::endl;
}
//=================================================================================================
