/*************************************************************************************************

    Copyright (C) 2009-2018    Vasileios Vavourakis (vasvav@gmail.com)

    This program is free software; you can redistribute it and/or modify it under the terms of
    the GNU General Public License as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
    without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
    the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with this program; if
    not, write to:
        Free Software Foundation, Inc.,
        59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 *************************************************************************************************/

//=================================================================================================
#include "feb3/wound_healing_solver.h"
#include "libmesh/gmsh_io.h"
#include "libmesh/vtk_io.h"
#include "feb3/paraview_io.h"
//=================================================================================================
// external variables
extern
OFStream flog, ferr;
extern
GlobalTolerance tol;
extern
libMesh::Parallel::Communicator* gcomm_ptr;
// external functions
extern
void update_mesh_lagrange (DomainData&, std::vector< std::vector<Real> >&, const IndexPair& );
extern
void restore_mesh (DomainData&, const std::vector< std::vector<Real> >&, const IndexPair& );
//=================================================================================================
#include "feb3/domain_data_inl.h"
#include "feb3/solution_io_inl.h"
//=================================================================================================
WoundHealingSolver::WoundHealingSolver (DomainData& dmn) :
    NavierCauchySolver3D(dmn), ReactionDiffusionSolverXD(dmn),
    _log("WoundHealingSolver") {
    // set the default label
    this->_label = "WoundHealingSolver";
}
//=================================================================================================
void WoundHealingSolver::init () {
    this->_log.push("init");

    // initialize the solvers respectively
    NavierCauchySolver3D::init();
    ReactionDiffusionSolverXD::init();

    this->_log.pop("init");
}
//=================================================================================================
void WoundHealingSolver::run () {
    this->_log.push("run");

    EquationSystems& es = NavierCauchySolver3D::_dmn.get_equation_systems();
    std::vector< std::vector<Real> > solu_u;
    IndexPair solu_idx;
    //
    // set the initial conditions and initialise some member objects
    NavierCauchySolver3D::_set_initial();
    NavierCauchySolver3D::par_list.set<bool>("enforce to initialize solution") = false;
    ReactionDiffusionSolverXD::_set_initial();
    ReactionDiffusionSolverXD::par_list.set<bool>("enforce to initialize solution") = false;
    // enforce these parameters too
    NavierCauchySolver3D::par_list.set<bool>("save updated mesh") = false;
    NavierCauchySolver3D::par_list.set<Real>("relaxing parameter") = 0.999;
    //
    // save the FE-grid
    if ( libMesh::global_processor_id() == 0 ) {
        char fname[buffer_len];
        flog.print(" >> Save mesh\n", false);
        sprintf( fname, "%s/%s-%s.vtu",
                 es.parameters.get<std::string>("results directory").c_str(),
                 es.parameters.get<std::string>("output file name").c_str(),
                 stringfy_number(es.parameters.get<Real>("current time")).c_str() );
        VTKIO(es.get_mesh()).write(fname);
    }
    //
    if ( es.parameters.get<bool>("Navier-Cauchy solver: apply body force") ) {
        // set parameters
        es.parameters.set<std::string>("wound healing solver: apply gravity title") = "gravity_loading_supine";
        es.parameters.set<bool>("wound healing solver: apply gravity (increase)") = true;
        // run simulation
        this->_apply_gravity();
    }
    //
    if ( es.parameters.get<bool>("wound healing solver: apply wound closure") ) {
        // run simulation
        this->_wound_closure();
    }
    //
    do {
        const bool membrane_present = NavierCauchySolver3D::par_list.get<bool>("membrane elements present");
        // set parameters
        ReactionDiffusionSolverXD::par_list.set<bool>("membrane elements present") = false;
        // run simulation
        this->_wound_healing();
        // set parameters
        ReactionDiffusionSolverXD::par_list.set<bool>("membrane elements present") = membrane_present;

    } while ( false );
    //
    if ( es.parameters.get<bool>("Navier-Cauchy solver: apply body force") ) {
        // set parameters
        es.parameters.set<std::string>("wound healing solver: apply gravity title") = "gravity_unloading_supine";
        es.parameters.set<bool>("wound healing solver: apply gravity (increase)") = false;
        // run simulation
        this->_apply_gravity();
        // end of if case...
    }
    //
    if ( es.parameters.get<bool>("wound healing solver: apply gravity (upright)") ) {
        // set parameters
        es.parameters.set<std::string>("wound healing solver: apply gravity title") = "gravity_loading_upright";
        es.parameters.set<bool>("wound healing solver: apply gravity (increase)") = true;
        // reset gravity load
        NavierCauchySolver3D::_dmn.bforce.value.comp =
            es.parameters.get<Point>("wound healing solver: apply gravity (upright): direction")(0),
            es.parameters.get<Point>("wound healing solver: apply gravity (upright): direction")(1),
            es.parameters.get<Point>("wound healing solver: apply gravity (upright): direction")(2);
        // run simulation
        this->_apply_gravity();
    }
    // update the FE-grid
    flog.print(" >> Update mesh\n", false);
    NavierCauchySolver3D::get_displacements("::solution", solu_u, solu_idx);
    update_mesh_lagrange(NavierCauchySolver3D::_dmn, solu_u, solu_idx);
    // save the FE-grid
    if ( libMesh::global_processor_id() == 0 ) {
        char fname[buffer_len];
        flog.print(" >> Save mesh\n", false);
        sprintf( fname, "%s/%s-%s.vtu",
                 es.parameters.get<std::string>("results directory").c_str(),
                 es.parameters.get<std::string>("output file name").c_str(),
                 stringfy_number(es.parameters.get<Real>("current time")).c_str() );
        VTKIO(es.get_mesh()).write(fname);
        // end of if case...
    }
    // end of all simulations...

    this->_log.pop("run");
}
//=================================================================================================
void WoundHealingSolver::_apply_gravity () {
    this->_log.push("_apply_gravity");

    DomainData& dmn = NavierCauchySolver3D::_dmn;
    EquationSystems& es = dmn.get_equation_systems();
    //
    flog.print("\n >>> Run gravity simulation <<<\n\n", false);
    //
    for (int isurf=0; isurf<es.parameters.get<int>("wound healing solver: wound closure: number of BC surfaces"); isurf++) {
        char field_name[buffer_len];
        sprintf(field_name, "wound healing solver: wound closure: BC surface %d", isurf);
        const boundary_id_type id = es.parameters.get<boundary_id_type>(field_name);
        // activate interface boundary condition surface from simulations
        std::map<boundary_id_type, BoundaryConditionInfo>::iterator iter_map = NavierCauchySolver3D::_dmn.bcs.find(id);
        if ( NavierCauchySolver3D::_dmn.bcs.end() == iter_map ) libmesh_error();
        iter_map->second.set_active_function(&flat_function);
        iter_map->second.active_params.assign(1, 0.0);
        iter_map->second.active_params[0] = 0.0;
        iter_map->second.set_time_function(&flat_function);
        iter_map->second.time_params.assign(1, 0.0);
        iter_map->second.time_params[0] = 0.0;
    }
    //
    dmn.bforce.set_time_function(ramp_function);
    if ( es.parameters.get<bool>("wound healing solver: apply gravity (increase)") ) {
        dmn.bforce.time_params.assign(4, 0.0);
        dmn.bforce.time_params[0] = 1.0;
        dmn.bforce.time_params[1] = 0.0;
        dmn.bforce.time_params[2] = 1.0;
        dmn.bforce.time_params[3] = 1.0;
    } else {
        dmn.bforce.time_params.assign(4, 0.0);
        dmn.bforce.time_params[0] = 1.0;
        dmn.bforce.time_params[1] = -1.0;
        dmn.bforce.time_params[2] = 1.0;
        dmn.bforce.time_params[3] = 1.0;
    }
    //
    dmn.nforce.set_time_function__employ(&flat_function);
    dmn.nforce.time_params__employ.assign(1, 0.0);
    dmn.nforce.time_params__employ[0] = es.parameters.get<bool>("wound healing solver: apply wound closure") ? 1.0 : 0.0;
    dmn.nforce.set_time_function__evaluate(&flat_function);
    dmn.nforce.time_params__evaluate.assign(1, 0.0);
    dmn.nforce.time_params__evaluate[0] = 0.0;
    //
    NavierCauchySolver3D::par_list.set<bool>("full Newton-Raphson") = es.parameters.get<bool>("Navier-Cauchy solver: full Newton-Raphson");
    NavierCauchySolver3D::par_list.set<bool>("load control")      = true;
    NavierCauchySolver3D::par_list.set<bool>("apply body force")  = true;
    NavierCauchySolver3D::par_list.set<bool>("apply nodal force") = es.parameters.get<bool>("wound healing solver: apply wound closure");
    NavierCauchySolver3D::par_list.set<int>("maximum increments") = es.parameters.get<int>("Navier-Cauchy solver: maximum increments");
    NavierCauchySolver3D::par_list.set<int>("solution increment") = es.parameters.get<int>("Navier-Cauchy solver: solution increment");
    NavierCauchySolver3D::par_list.set<int>("maximum iterations") = es.parameters.get<int>("Navier-Cauchy solver: maximum iterations");
    NavierCauchySolver3D::par_list.set<int>("minimum iterations") = es.parameters.get<int>("Navier-Cauchy solver: minimum iterations");
    NavierCauchySolver3D::par_list.set<bool>("save updated mesh")          = false;
    NavierCauchySolver3D::par_list.set<bool>("save intermediate solution") = true;
    //
    es.parameters.set<std::string>("simulation title") = es.parameters.get<std::string>("wound healing solver: apply gravity title");
    //
    NavierCauchySolver3D::run();
    // end of simulation...

    this->_log.pop("_apply_gravity");
}
//=================================================================================================
void WoundHealingSolver::_wound_closure () {
    this->_log.push("_wound_closure");

    DomainData& dmn = NavierCauchySolver3D::_dmn;
    EquationSystems& es = dmn.get_equation_systems();
    //
    flog.print("\n >>> Run wound-closure simulation <<<\n\n", false);
    //
    for (int isurf=0; isurf<es.parameters.get<int>("wound healing solver: wound closure: number of BC surfaces"); isurf++) {
        char field_name[buffer_len];
        sprintf(field_name, "wound healing solver: wound closure: BC surface %d", isurf);
        const boundary_id_type id = es.parameters.get<boundary_id_type>(field_name);
        // activate interface boundary condition surface from simulations
        std::map<boundary_id_type, BoundaryConditionInfo>::iterator iter_map = NavierCauchySolver3D::_dmn.bcs.find(id);
        if ( NavierCauchySolver3D::_dmn.bcs.end() == iter_map ) libmesh_error();
        iter_map->second.set_active_function(&flat_function);
        iter_map->second.active_params.assign(1, 0.0);
        iter_map->second.active_params[0] = 1.0;
        iter_map->second.set_time_function(&linear_function);
        iter_map->second.time_params.assign(1, 0.0);
        iter_map->second.time_params[0] = 1.0;
    }
    //
    dmn.bforce.set_time_function(&flat_function);
    dmn.bforce.time_params.assign(1, 0.0);
    dmn.bforce.time_params[0] = es.parameters.get<bool>("Navier-Cauchy solver: apply body force") ? 1.0 : 0.0;
    //
    dmn.nforce.set_time_function__employ(&flat_function);
    dmn.nforce.time_params__employ.assign(1, 0.0);
    dmn.nforce.time_params__employ[0] = 0.0;
    dmn.nforce.set_time_function__evaluate(&flat_function);
    dmn.nforce.time_params__evaluate.assign(1, 0.0);
    dmn.nforce.time_params__evaluate[0] = 1.0;
    //
    NavierCauchySolver3D::par_list.set<bool>("full Newton-Raphson") = true;
    NavierCauchySolver3D::par_list.set<bool>("load control")      = false;
    NavierCauchySolver3D::par_list.set<bool>("apply body force")  = es.parameters.get<bool>("Navier-Cauchy solver: apply body force");
    NavierCauchySolver3D::par_list.set<bool>("apply nodal force") = true;
    NavierCauchySolver3D::par_list.set<int>("maximum increments") = 5;
    NavierCauchySolver3D::par_list.set<int>("solution increment") = 5;
    NavierCauchySolver3D::par_list.set<int>("maximum iterations") = es.parameters.get<int>("Navier-Cauchy solver: maximum iterations");
    NavierCauchySolver3D::par_list.set<int>("minimum iterations") = es.parameters.get<int>("Navier-Cauchy solver: minimum iterations");
    NavierCauchySolver3D::par_list.set<bool>("save updated mesh")          = false;
    NavierCauchySolver3D::par_list.set<bool>("save intermediate solution") = false;
    //
    es.parameters.set<std::string>("simulation title") = "wound_closure";
    //
    NavierCauchySolver3D::run();
    // end of simulation...

    this->_log.pop("_wound_closure");
}
//=================================================================================================
void WoundHealingSolver::_wound_healing () {
    this->_log.push("_wound_healing");

    DomainData& dmn = ReactionDiffusionSolverXD::_dmn;
    EquationSystems& es = dmn.get_equation_systems();
    //
    ParaviewIO pvtu(dmn.get_mesh());
    for (int l=0; l<es.parameters.get<int>("number of variables to hide in Paraview"); l++) {
        char var_name[buffer_len];
        sprintf(var_name, "Paraview variable #%d to hide", l);
        pvtu.hide_variable(es.parameters.get<std::string>(var_name));
    }
    //
    flog.print("\n >>> Run wound-healing simulation <<<\n\n", false);
    //
    for (int isurf=0; isurf<es.parameters.get<int>("wound healing solver: wound closure: number of BC surfaces"); isurf++) {
        char field_name[buffer_len];
        sprintf(field_name, "wound healing solver: wound closure: BC surface %d", isurf);
        const boundary_id_type id = es.parameters.get<boundary_id_type>(field_name);
        // activate interface boundary condition surface from simulations
        std::map<boundary_id_type, BoundaryConditionInfo>::iterator iter_map = NavierCauchySolver3D::_dmn.bcs.find(id);
        if ( NavierCauchySolver3D::_dmn.bcs.end() == iter_map ) libmesh_error();
        iter_map->second.set_active_function(&flat_function);
        iter_map->second.active_params.assign(1, 0.0);
        iter_map->second.active_params[0] = 0.0;
        iter_map->second.set_time_function(&flat_function);
        iter_map->second.time_params.assign(1, 0.0);
        iter_map->second.time_params[0] = 0.0;
    }
    //
    dmn.bforce.set_time_function(&flat_function);
    dmn.bforce.time_params.assign(1, 0.0);
    dmn.bforce.time_params[0] = es.parameters.get<bool>("Navier-Cauchy solver: apply body force") ? 1.0 : 0.0;
    //
    dmn.nforce.set_time_function__employ(&flat_function);
    dmn.nforce.time_params__employ.assign(1, 0.0);
    dmn.nforce.time_params__employ[0] = es.parameters.get<bool>("wound healing solver: apply wound closure") ? 1.0 : 0.0;
    dmn.nforce.set_time_function__evaluate(&flat_function);
    dmn.nforce.time_params__evaluate.assign(1, 0.0);
    dmn.nforce.time_params__evaluate[0] = 0.0;
    //
    NavierCauchySolver3D::par_list.set<bool>("full Newton-Raphson") = true;
    NavierCauchySolver3D::par_list.set<bool>("load control")      = true;
    NavierCauchySolver3D::par_list.set<bool>("apply body force")  = es.parameters.get<bool>("Navier-Cauchy solver: apply body force");
    NavierCauchySolver3D::par_list.set<bool>("apply nodal force") = es.parameters.get<bool>("wound healing solver: apply wound closure");
    NavierCauchySolver3D::par_list.set<int>("maximum increments") = ReactionDiffusionSolverXD::par_list.get<int>("maximum time steps")
                                                                  / es.parameters.get<int>("wound healing solver: solid solver increment");
    NavierCauchySolver3D::par_list.set<int>("solution increment") = es.parameters.get<int>("wound healing solver: solid solver increment");
    NavierCauchySolver3D::par_list.set<int>("maximum iterations") = es.parameters.get<int>("Navier-Cauchy solver: maximum iterations");
    NavierCauchySolver3D::par_list.set<int>("minimum iterations") = es.parameters.get<int>("Navier-Cauchy solver: minimum iterations");
    NavierCauchySolver3D::par_list.set<bool>("save updated mesh")          = false;
    NavierCauchySolver3D::par_list.set<bool>("save intermediate solution") = false;
    //
    es.parameters.set<std::string>("simulation title") = "wound_healing";
    //
    char fname[buffer_len];
    Real& current_time = es.parameters.set<Real>("current time");
    const std::string results_dir  = es.parameters.get<std::string>("results directory"),
                      output_fname = es.parameters.get<std::string>("output file name"),
                      sim_title    = es.parameters.get<std::string>("simulation title");
    //
    // write a PVD file to store results
    if ( libMesh::global_processor_id() == 0 ) {
        flog.print(" >> Generate Paraview PVD-file\n", false);
        sprintf( fname, "%s/%s-%s.pvd",
                 results_dir.c_str(), output_fname.c_str(), sim_title.c_str() );
        std::ofstream fpvd(fname);
        fpvd.setf(std::ios_base::scientific|std::ios_base::unitbuf);
        //
        fpvd << "<?xml version=\"1.0\"?>" << std::endl;
        fpvd << "<VTKFile type=\"Collection\" version=\"0.1\" byte_order=\"LittleEndian\">" << std::endl;
        fpvd << "<Collection>" << std::endl;
        for (int time=0; time<=ReactionDiffusionSolverXD::par_list.get<int>("maximum time steps"); time++) {
            if ( time == 0                                                                    ||
                 time%es.parameters.get<int>("wound healing solver: solution increment") == 0 ||
                 time == ReactionDiffusionSolverXD::par_list.get<int>("maximum time steps")   ) {
                const Real t_step = ReactionDiffusionSolverXD::par_list.get<Real>("time step"),
                           t_time = time*t_step;
                sprintf( fname, "paraview/%s-%s-%s.vtu",
                         output_fname.c_str(), sim_title.c_str(), stringfy_number(t_time).c_str() );
                fpvd << "<DataSet timestep=\"" << t_time << "\" group=\"\" part=\"0\" file=\"" << fname << "\"/>" << std::endl;
            }
            // end of time steps loop...
        }
        fpvd << "</Collection>" << std::endl;
        fpvd << "</VTKFile>" << std::endl;
    }
    //
    // write initial results
    flog.print("  > Save solution\n", false);
    sprintf( fname, "%s/paraview/%s-%s-%s.vtu",
             results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), stringfy_number(current_time).c_str() );
    pvtu.write_equation_systems(fname, es);
    //
    // evaluate system lumped matrix
    flog.print(" >> Evaluate system lumped matrix\n", false);
    ReactionDiffusionSolverXD::_diag_matrix();
    //
    // run for all time steps
    const int ntimes = ReactionDiffusionSolverXD::par_list.get<int>("maximum time steps"),
              soln_step = es.parameters.get<int>("wound healing solver: solution increment");
    const Real time_step = ReactionDiffusionSolverXD::par_list.get<Real>("time step");
    const int ncs_step = es.parameters.get<int>("wound healing solver: solid solver increment"),
              nloads = NavierCauchySolver3D::par_list.get<int>("maximum increments");
    int load = 0;
    for (int itime=0; itime<ntimes; itime++) {
        current_time += time_step;
        es.parameters.set<Real>("current time") = current_time;
        //
        // solve the system
        {
            const char end_of_line[] = "           \r";
            char txt[buffer_len];
            sprintf( txt, " >> Time %d out of %d (%.10f) %s",
                     itime+1, ntimes, current_time, end_of_line );
            flog.print(txt, false);
        }
        ReactionDiffusionSolverXD::_solve();
        //
        // check if to run solid solver
        if ( (itime+1)%ncs_step == 0 ) {
            flog.print("\n", false);
            // solve the system
            NavierCauchySolver3D::par_list.set<int>("increment step") = load;
            NavierCauchySolver3D::par_list.set<Real>("previous pseudo-time") = (load+0.0)/nloads;
            NavierCauchySolver3D::par_list.set<Real>("current pseudo-time") = (load+1.0)/nloads;
            //
            if ( NavierCauchySolver3D::par_list.get<bool>("apply nodal force") ) {
                char txt[buffer_len];
                sprintf( txt, "  > Pseudo-time %d out of %d (%.10f)\t Fn(Eval):%d\t Fn(Appl):%d\n",
                         load+1, nloads,
                         NavierCauchySolver3D::par_list.get<Real>("current pseudo-time"),
                         dmn.nforce.calc_time_function__evaluate(NavierCauchySolver3D::par_list.get<Real>("current pseudo-time")),
                         (bool)dmn.nforce.calc_time_function__employ(NavierCauchySolver3D::par_list.get<Real>("current pseudo-time")) );
                flog.print(txt, false);
            } else {
                char txt[buffer_len];
                sprintf( txt, "  > Pseudo-time %d out of %d (%.10f)\n",
                         load+1, nloads, NavierCauchySolver3D::par_list.get<Real>("current pseudo-time") );
                flog.print(txt, false);
            }
            //
            // run for all iteration steps
            const int niters = NavierCauchySolver3D::par_list.get<int>("maximum iterations");
            for (int iter=0; iter<niters; iter++) {
                NavierCauchySolver3D::par_list.set<int>("iteration step") = iter;
                //
                // assemble the linearized system
                NavierCauchySolver3D::par_list.get<bool>("mixed formulation") ?
                NavierCauchySolver3D::_assemble_system__FWD_UP() : NavierCauchySolver3D::_assemble_system__FWD_U();
                //
                // solve and update solution via Newton-Raphson
                NavierCauchySolver3D::_solve();
                //
                // calculate convergence criteria
                const int status = NavierCauchySolver3D::_check_convergence();
                // check if to exit loop, abort or continue
                if      ( status ==  1 ) break;
                else if ( status <   0 ) {
                    //
                    // reset solution after failure
                    NavierCauchySolver3D::_reset_increment(true);
                    //
                    // calculate the gradient solution (if necessary)
                    if ( NavierCauchySolver3D::par_list.get<bool>("calculate gradients") ) {
                        flog.print("  > Calculate gradient solution\n", false);
                        NavierCauchySolver3D::_calc_gradients();
                    }
                    //
                    flog.print("  > Save solution\n", false);
                    sprintf( fname, "%s/paraview/%s-%s-%s.vtu",
                             results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), stringfy_number(current_time).c_str() );
                    pvtu.write_equation_systems(fname, es);
                    // now jump code...
                    goto END_LABEL;
                }
                // end of iterations loop...
            }
            // make sure that all processes have finished
            gcomm_ptr->barrier();
            //
            // update quadrature point set data
            flog.print("  > Update quadrature data\n", false);
            NavierCauchySolver3D::par_list.get<bool>("mixed formulation") ?
            NavierCauchySolver3D::_set_quad_info__FWD_UP() : NavierCauchySolver3D::_set_quad_info__FWD_U();
            // increment this index
            ++load;
            // end of if case...
        }
        //
        // check if to write results
        if ( (itime+1)%soln_step == 0 || (itime+1) == ntimes ) {
            if ( (itime+1)%ncs_step == 0 ) {
                // calculate the gradient solution (if necessary)
                if ( NavierCauchySolver3D::par_list.get<bool>("calculate gradients") ) {
                    flog.print("  > Calculate gradient solution\n", false);
                    NavierCauchySolver3D::_calc_gradients();
                }
            }
            flog.print("  > Save solution\r", false);
            sprintf( fname, "%s/paraview/%s-%s-%s.vtu",
                     results_dir.c_str(), output_fname.c_str(), sim_title.c_str(), stringfy_number(current_time).c_str() );
            pvtu.write_equation_systems(fname, es);
        }
        //
        // check if solid solver has run
        if ( (itime+1)%ncs_step == 0 ) {
            // reset (zero-out) incremental solution for subsequent increment step
            NavierCauchySolver3D::_reset_increment();
        }
        // end of time steps loop...
    }
    // end of simulation...
    END_LABEL: ;
    clear_line_(std::cout);
    flog.print(" >> Simulation terminated\n", false);

    this->_log.pop("_wound_healing");
}
//=================================================================================================
